'''
This directory contains libraries and code for reading file formats
using vendorized dependencies, not necessarily for MS vendor binary
files. These vendorized libraries may still require other platform-dependent
resources such as installed DLLs or 3rd party programs.

Each reader has had it's author's license prepended.
'''
