/* Generated by Cython 0.29.13 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [
            "C:\\Users\\Asus\\Miniconda2\\envs\\py37\\lib\\site-packages\\numpy\\core\\include\\numpy\\arrayobject.h",
            "C:\\Users\\Asus\\Miniconda2\\envs\\py37\\lib\\site-packages\\numpy\\core\\include\\numpy\\npy_math.h",
            "C:\\Users\\Asus\\Miniconda2\\envs\\py37\\lib\\site-packages\\numpy\\core\\include\\numpy\\ufuncobject.h"
        ],
        "include_dirs": [
            "C:\\Users\\Asus\\Miniconda2\\envs\\py37\\lib\\site-packages\\numpy\\core\\include",
            "C:\\Users\\Asus\\Miniconda2\\envs\\py37\\lib\\site-packages\\ms_peak_picker\\_c",
            "c:\\users\\asus\\dev\\brainpy\\brainpy\\_c"
        ],
        "name": "ms_deisotope._c.feature_map.feature_processor",
        "sources": [
            "ms_deisotope/_c/feature_map/feature_processor.pyx"
        ]
    },
    "module_name": "ms_deisotope._c.feature_map.feature_processor"
}
END: Cython Metadata */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_13"
#define CYTHON_HEX_VERSION 0x001D0DF0
#define CYTHON_FUTURE_DIVISION 0
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
#if PY_VERSION_HEX >= 0x030800A4 && PY_VERSION_HEX < 0x030800B2
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, 0, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
  #define PyObject_Unicode             PyObject_Str
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? PyMethod_New(func, self) : (Py_INCREF(func), func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif


#define __PYX_ERR(f_index, lineno, Ln_error) \
{ \
  __pyx_filename = __pyx_f[f_index]; __pyx_lineno = lineno; __pyx_clineno = __LINE__; goto Ln_error; \
}

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__ms_deisotope___c__feature_map__feature_processor
#define __PYX_HAVE_API__ms_deisotope___c__feature_map__feature_processor
/* Early includes */
#include <string.h>
#include <stdlib.h>
#include "numpy/npy_math.h"
#include <stdio.h>
#include "numpy/arrayobject.h"
#include "numpy/ufuncobject.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;

/* Header.proto */
#if !defined(CYTHON_CCOMPLEX)
  #if defined(__cplusplus)
    #define CYTHON_CCOMPLEX 1
  #elif defined(_Complex_I)
    #define CYTHON_CCOMPLEX 1
  #else
    #define CYTHON_CCOMPLEX 0
  #endif
#endif
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #include <complex>
  #else
    #include <complex.h>
  #endif
#endif
#if CYTHON_CCOMPLEX && !defined(__cplusplus) && defined(__sun__) && defined(__GNUC__)
  #undef _Complex_I
  #define _Complex_I 1.0fj
#endif


static const char *__pyx_f[] = {
  "ms_deisotope\\_c\\feature_map\\feature_processor.pyx",
  "stringsource",
  "__init__.pxd",
  "type.pxd",
  "peak_set.pxd",
  "composition.pxd",
  "isotopic_distribution.pxd",
  "peak_index.pxd",
  "ms_deisotope\\_c\\averagine.pxd",
  "ms_deisotope\\_c\\deconvoluter_base.pxd",
  "ms_deisotope\\_c\\scoring.pxd",
  "ms_deisotope\\_c\\feature_map\\lcms_feature.pxd",
  "ms_deisotope\\_c\\feature_map\\feature_map.pxd",
  "ms_deisotope\\_c\\peak_set.pxd",
  "ms_deisotope\\_c\\feature_map\\feature_fit.pxd",
};
/* BufferFormatStructs.proto */
#define IS_UNSIGNED(type) (((type) -1) > 0)
struct __Pyx_StructField_;
#define __PYX_BUF_FLAGS_PACKED_STRUCT (1 << 0)
typedef struct {
  const char* name;
  struct __Pyx_StructField_* fields;
  size_t size;
  size_t arraysize[8];
  int ndim;
  char typegroup;
  char is_unsigned;
  int flags;
} __Pyx_TypeInfo;
typedef struct __Pyx_StructField_ {
  __Pyx_TypeInfo* type;
  const char* name;
  size_t offset;
} __Pyx_StructField;
typedef struct {
  __Pyx_StructField* field;
  size_t parent_offset;
} __Pyx_BufFmt_StackElem;
typedef struct {
  __Pyx_StructField root;
  __Pyx_BufFmt_StackElem* head;
  size_t fmt_offset;
  size_t new_count, enc_count;
  size_t struct_alignment;
  int is_complex;
  char enc_type;
  char new_packmode;
  char enc_packmode;
  char is_valid_array;
} __Pyx_BufFmt_Context;


/* "brainpy/_c/composition.pxd":3
 * from brainpy.mass_dict import nist_mass as __nist_mass
 * 
 * ctypedef long count_type             # <<<<<<<<<<<<<<
 * 
 * cdef dict nist_mass
 */
typedef long __pyx_t_7brainpy_2_c_11composition_count_type;

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":776
 * # in Cython to enable them only on the right systems.
 * 
 * ctypedef npy_int8       int8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 */
typedef npy_int8 __pyx_t_5numpy_int8_t;

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":777
 * 
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t
 */
typedef npy_int16 __pyx_t_5numpy_int16_t;

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":778
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int64      int64_t
 * #ctypedef npy_int96      int96_t
 */
typedef npy_int32 __pyx_t_5numpy_int32_t;

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":779
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_int96      int96_t
 * #ctypedef npy_int128     int128_t
 */
typedef npy_int64 __pyx_t_5numpy_int64_t;

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":783
 * #ctypedef npy_int128     int128_t
 * 
 * ctypedef npy_uint8      uint8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 */
typedef npy_uint8 __pyx_t_5numpy_uint8_t;

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":784
 * 
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t
 */
typedef npy_uint16 __pyx_t_5numpy_uint16_t;

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":785
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint64     uint64_t
 * #ctypedef npy_uint96     uint96_t
 */
typedef npy_uint32 __pyx_t_5numpy_uint32_t;

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":786
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_uint96     uint96_t
 * #ctypedef npy_uint128    uint128_t
 */
typedef npy_uint64 __pyx_t_5numpy_uint64_t;

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":790
 * #ctypedef npy_uint128    uint128_t
 * 
 * ctypedef npy_float32    float32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_float64    float64_t
 * #ctypedef npy_float80    float80_t
 */
typedef npy_float32 __pyx_t_5numpy_float32_t;

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":791
 * 
 * ctypedef npy_float32    float32_t
 * ctypedef npy_float64    float64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_float80    float80_t
 * #ctypedef npy_float128   float128_t
 */
typedef npy_float64 __pyx_t_5numpy_float64_t;

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":800
 * # The int types are mapped a bit surprising --
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t
 */
typedef npy_long __pyx_t_5numpy_int_t;

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":801
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   longlong_t
 * 
 */
typedef npy_longlong __pyx_t_5numpy_long_t;

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":802
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_ulong      uint_t
 */
typedef npy_longlong __pyx_t_5numpy_longlong_t;

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":804
 * ctypedef npy_longlong   longlong_t
 * 
 * ctypedef npy_ulong      uint_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t
 */
typedef npy_ulong __pyx_t_5numpy_uint_t;

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":805
 * 
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 */
typedef npy_ulonglong __pyx_t_5numpy_ulong_t;

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":806
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_intp       intp_t
 */
typedef npy_ulonglong __pyx_t_5numpy_ulonglong_t;

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":808
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 * ctypedef npy_intp       intp_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uintp      uintp_t
 * 
 */
typedef npy_intp __pyx_t_5numpy_intp_t;

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":809
 * 
 * ctypedef npy_intp       intp_t
 * ctypedef npy_uintp      uintp_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_double     float_t
 */
typedef npy_uintp __pyx_t_5numpy_uintp_t;

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":811
 * ctypedef npy_uintp      uintp_t
 * 
 * ctypedef npy_double     float_t             # <<<<<<<<<<<<<<
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t
 */
typedef npy_double __pyx_t_5numpy_float_t;

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":812
 * 
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longdouble longdouble_t
 * 
 */
typedef npy_double __pyx_t_5numpy_double_t;

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":813
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cfloat      cfloat_t
 */
typedef npy_longdouble __pyx_t_5numpy_longdouble_t;
/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< float > __pyx_t_float_complex;
  #else
    typedef float _Complex __pyx_t_float_complex;
  #endif
#else
    typedef struct { float real, imag; } __pyx_t_float_complex;
#endif
static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float, float);

/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< double > __pyx_t_double_complex;
  #else
    typedef double _Complex __pyx_t_double_complex;
  #endif
#else
    typedef struct { double real, imag; } __pyx_t_double_complex;
#endif
static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double, double);


/*--- Type declarations ---*/
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase;
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak;
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet;
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed;
struct __pyx_obj_7brainpy_2_c_11composition_PyComposition;
struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak;
struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex;
struct __pyx_obj_12ms_deisotope_2_c_9averagine_Averagine;
struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern;
struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache;
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase;
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase;
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase;
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_LeastSquaresFitter;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_GTestFitter;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledGTestFitter;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_FunctionScorer;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_DotProductFitter;
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList;
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase;
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode;
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase;
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature;
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature;
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator;
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage;
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass;
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap;
struct __pyx_obj_12ms_deisotope_2_c_8peak_set__Index;
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_EnvelopePair;
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope;
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak;
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution;
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet;
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed;
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_map_coord;
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit;
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator;
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer;
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase;
struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet__has_peak;
struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet_all_peaks_for;
struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_cell;
struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_list;
struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_has_peak_hinted;

/* "ms_peak_picker/_c/peak_set.pxd":40
 *     cdef FittedPeak _get_nearest_peak(self, double mz, double* errout)
 * 
 *     cdef FittedPeak _has_peak(self, double mz, double tolerance=*)             # <<<<<<<<<<<<<<
 * 
 *     cpdef PeakSet clone(self)
 */
struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet__has_peak {
  int __pyx_n;
  double tolerance;
};

/* "ms_peak_picker/_c/peak_set.pxd":48
 *     cdef int _between_bounds(self, double m1, double m2, size_t* startp, size_t* endp)
 * 
 *     cpdef tuple all_peaks_for(self, double mz, double error_tolerance=*)             # <<<<<<<<<<<<<<
 * 
 *     cdef FittedPeak getitem(self, size_t i)
 */
struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet_all_peaks_for {
  int __pyx_n;
  double error_tolerance;
};

/* "ms_peak_picker/_c/peak_set.pxd":62
 * 
 * 
 * cdef struct index_cell:             # <<<<<<<<<<<<<<
 *     double center_value
 *     size_t start
 */
struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_cell {
  double center_value;
  size_t start;
  size_t end;
};

/* "ms_peak_picker/_c/peak_set.pxd":68
 * 
 * 
 * cdef struct index_list:             # <<<<<<<<<<<<<<
 *     index_cell* index
 *     size_t size
 */
struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_list {
  struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_cell *index;
  size_t size;
  double low;
  double high;
};

/* "ms_peak_picker/_c/peak_set.pxd":88
 *     cpdef _allocate_index(self)
 * 
 *     cpdef FittedPeak has_peak_hinted(self, double mz, double tolerance=*, size_t hint_start=*, size_t hint_end=*)             # <<<<<<<<<<<<<<
 * 
 * cdef size_t double_binary_search_ppm(double* array, double value, double tolerance, size_t n)
 */
struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_has_peak_hinted {
  int __pyx_n;
  double tolerance;
  size_t hint_start;
  size_t hint_end;
};
struct __pyx_opt_args_7brainpy_2_c_11composition_neutral_mass;
struct __pyx_opt_args_7brainpy_2_c_11composition_mass_charge_ratio;
struct __pyx_t_7brainpy_2_c_11composition_Isotope;
struct __pyx_t_7brainpy_2_c_11composition_IsotopeMap;
struct __pyx_t_7brainpy_2_c_11composition_Element;
struct __pyx_t_7brainpy_2_c_11composition_ElementHashBucket;
struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable;
struct __pyx_t_7brainpy_2_c_11composition_Composition;
struct __pyx_opt_args_7brainpy_2_c_11composition_13PyComposition__set_impl;
struct __pyx_opt_args_7brainpy_2_c_11composition_13PyComposition_pop;

/* "brainpy/_c/composition.pxd":10
 * cdef double PROTON = nist_mass["H+"][0][0]
 * 
 * cdef double neutral_mass(double mz,  int z, double charge_carrier=*) nogil             # <<<<<<<<<<<<<<
 * cdef double mass_charge_ratio(double neutral_mass, int z, double charge_carrier=*) nogil
 * cdef char* _parse_isotope_string(char* label, int* isotope_num, char* element_name) nogil
 */
struct __pyx_opt_args_7brainpy_2_c_11composition_neutral_mass {
  int __pyx_n;
  double charge_carrier;
};

/* "brainpy/_c/composition.pxd":11
 * 
 * cdef double neutral_mass(double mz,  int z, double charge_carrier=*) nogil
 * cdef double mass_charge_ratio(double neutral_mass, int z, double charge_carrier=*) nogil             # <<<<<<<<<<<<<<
 * cdef char* _parse_isotope_string(char* label, int* isotope_num, char* element_name) nogil
 * 
 */
struct __pyx_opt_args_7brainpy_2_c_11composition_mass_charge_ratio {
  int __pyx_n;
  double charge_carrier;
};

/* "brainpy/_c/composition.pxd":17
 * # Isotope and IsotopeMap Declarations
 * 
 * cdef struct Isotope:             # <<<<<<<<<<<<<<
 *     double mass
 *     double abundance
 */
struct __pyx_t_7brainpy_2_c_11composition_Isotope {
  double mass;
  double abundance;
  int neutrons;
  int neutron_shift;
};

/* "brainpy/_c/composition.pxd":23
 *     int neutron_shift
 * 
 * cdef struct IsotopeMap:             # <<<<<<<<<<<<<<
 *     Isotope* bins
 *     size_t size
 */
struct __pyx_t_7brainpy_2_c_11composition_IsotopeMap {
  struct __pyx_t_7brainpy_2_c_11composition_Isotope *bins;
  size_t size;
};

/* "brainpy/_c/composition.pxd":36
 * # Element Declarations
 * 
 * cdef struct Element:             # <<<<<<<<<<<<<<
 *     char* symbol
 *     IsotopeMap* isotopes
 */
struct __pyx_t_7brainpy_2_c_11composition_Element {
  char *symbol;
  struct __pyx_t_7brainpy_2_c_11composition_IsotopeMap *isotopes;
  int monoisotopic_isotope_index;
};

/* "brainpy/_c/composition.pxd":56
 * # ElementHashTable and ElementHashBucket Declarations
 * 
 * cdef struct ElementHashBucket:             # <<<<<<<<<<<<<<
 *     Element** elements
 *     size_t used
 */
struct __pyx_t_7brainpy_2_c_11composition_ElementHashBucket {
  struct __pyx_t_7brainpy_2_c_11composition_Element **elements;
  size_t used;
  size_t size;
};

/* "brainpy/_c/composition.pxd":65
 * 
 * 
 * cdef struct ElementHashTable:             # <<<<<<<<<<<<<<
 *     ElementHashBucket* buckets
 *     size_t size
 */
struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable {
  struct __pyx_t_7brainpy_2_c_11composition_ElementHashBucket *buckets;
  size_t size;
};

/* "brainpy/_c/composition.pxd":94
 * # Composition Declarations
 * 
 * cdef struct Composition:             # <<<<<<<<<<<<<<
 *     char** elements
 *     count_type* counts
 */
struct __pyx_t_7brainpy_2_c_11composition_Composition {
  char **elements;
  __pyx_t_7brainpy_2_c_11composition_count_type *counts;
  size_t size;
  size_t used;
  int max_variants;
};

/* "brainpy/_c/composition.pxd":130
 *     @staticmethod
 *     cdef PyComposition _create(Composition* base)
 *     cdef void _set_impl(self, Composition* composition, bint free_existing=*)             # <<<<<<<<<<<<<<
 * 
 *     cdef void _initialize_from_formula(self, str formula)
 */
struct __pyx_opt_args_7brainpy_2_c_11composition_13PyComposition__set_impl {
  int __pyx_n;
  int free_existing;
};

/* "brainpy/_c/composition.pxd":144
 *     cpdef list items(self)
 * 
 *     cpdef pop(self, str key, object default=*)             # <<<<<<<<<<<<<<
 * 
 *     cdef count_type getitem(self, str key)
 */
struct __pyx_opt_args_7brainpy_2_c_11composition_13PyComposition_pop {
  int __pyx_n;
  PyObject *__pyx_default;
};
struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector;

/* "brainpy/_c/double_vector.pxd":1
 * cdef struct DoubleVector:             # <<<<<<<<<<<<<<
 *     double* v
 *     size_t used
 */
struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector {
  double *v;
  size_t used;
  size_t size;
};
struct __pyx_t_7brainpy_2_c_18isotopic_constants_PolynomialParameters;
struct __pyx_t_7brainpy_2_c_18isotopic_constants_PhiConstants;
struct __pyx_t_7brainpy_2_c_18isotopic_constants_IsotopicConstants;

/* "brainpy/_c/isotopic_constants.pxd":6
 * from brainpy._c.double_vector cimport DoubleVector
 * 
 * ctypedef DoubleVector dvec             # <<<<<<<<<<<<<<
 * 
 * 
 */
typedef struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector __pyx_t_7brainpy_2_c_18isotopic_constants_dvec;

/* "brainpy/_c/isotopic_constants.pxd":9
 * 
 * 
 * cdef struct PolynomialParameters:             # <<<<<<<<<<<<<<
 *     dvec* elementary_symmetric_polynomial
 *     dvec* power_sum
 */
struct __pyx_t_7brainpy_2_c_18isotopic_constants_PolynomialParameters {
  __pyx_t_7brainpy_2_c_18isotopic_constants_dvec *elementary_symmetric_polynomial;
  __pyx_t_7brainpy_2_c_18isotopic_constants_dvec *power_sum;
};

/* "brainpy/_c/isotopic_constants.pxd":14
 * 
 * 
 * cdef struct PhiConstants:             # <<<<<<<<<<<<<<
 *     int order
 *     Element* element
 */
struct __pyx_t_7brainpy_2_c_18isotopic_constants_PhiConstants {
  int order;
  struct __pyx_t_7brainpy_2_c_11composition_Element *element;
  struct __pyx_t_7brainpy_2_c_18isotopic_constants_PolynomialParameters *element_coefficients;
  struct __pyx_t_7brainpy_2_c_18isotopic_constants_PolynomialParameters *mass_coefficients;
};

/* "brainpy/_c/isotopic_constants.pxd":21
 * 
 * 
 * cdef struct IsotopicConstants:             # <<<<<<<<<<<<<<
 *     int order
 *     PhiConstants** constants
 */
struct __pyx_t_7brainpy_2_c_18isotopic_constants_IsotopicConstants {
  int order;
  struct __pyx_t_7brainpy_2_c_18isotopic_constants_PhiConstants **constants;
  size_t size;
  size_t used;
};
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak;
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList;
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache;
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_IsotopicDistribution;
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_make_isotopic_distribution;
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_guess_npeaks;
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution__isotopic_variants;
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_isotopic_variants;

/* "brainpy/_c/isotopic_distribution.pxd":7
 * 
 * 
 * cdef struct Peak:             # <<<<<<<<<<<<<<
 *     double mz
 *     double intensity
 */
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak {
  double mz;
  double intensity;
  int charge;
};

/* "brainpy/_c/isotopic_distribution.pxd":13
 * 
 * 
 * cdef struct PeakList:             # <<<<<<<<<<<<<<
 *     Peak* peaks
 *     size_t used
 */
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList {
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak *peaks;
  size_t used;
  size_t size;
};

/* "brainpy/_c/isotopic_distribution.pxd":35
 * 
 * 
 * cdef struct ElementCache:             # <<<<<<<<<<<<<<
 *     Element** elements
 *     size_t used
 */
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache {
  struct __pyx_t_7brainpy_2_c_11composition_Element **elements;
  size_t used;
  size_t size;
  struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable *source;
};

/* "brainpy/_c/isotopic_distribution.pxd":49
 * 
 * 
 * cdef struct IsotopicDistribution:             # <<<<<<<<<<<<<<
 *     Composition* composition
 *     IsotopicConstants* _isotopic_constants
 */
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_IsotopicDistribution {
  struct __pyx_t_7brainpy_2_c_11composition_Composition *composition;
  struct __pyx_t_7brainpy_2_c_18isotopic_constants_IsotopicConstants *_isotopic_constants;
  int order;
  double average_mass;
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak *monoisotopic_peak;
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache *cache;
};

/* "brainpy/_c/isotopic_distribution.pxd":58
 * 
 * 
 * cdef IsotopicDistribution* make_isotopic_distribution(Composition* composition, int order, ElementCache* cache=*) nogil             # <<<<<<<<<<<<<<
 * cdef void free_isotopic_distribution(IsotopicDistribution* distribution)  nogil
 * 
 */
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_make_isotopic_distribution {
  int __pyx_n;
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache *cache;
};

/* "brainpy/_c/isotopic_distribution.pxd":77
 * 
 * cdef int max_variants(Composition* composition, ElementCache* cache) nogil
 * cdef int guess_npeaks(Composition* composition_struct, size_t max_npeaks, ElementCache* cache=*) nogil             # <<<<<<<<<<<<<<
 * 
 * cpdef list _isotopic_variants(object composition, object npeaks=*, int charge=*, double charge_carrier=*)
 */
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_guess_npeaks {
  int __pyx_n;
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache *cache;
};

/* "brainpy/_c/isotopic_distribution.pxd":79
 * cdef int guess_npeaks(Composition* composition_struct, size_t max_npeaks, ElementCache* cache=*) nogil
 * 
 * cpdef list _isotopic_variants(object composition, object npeaks=*, int charge=*, double charge_carrier=*)             # <<<<<<<<<<<<<<
 * 
 * 
 */
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution__isotopic_variants {
  int __pyx_n;
  PyObject *npeaks;
  int charge;
  double charge_carrier;
};

/* "brainpy/_c/isotopic_distribution.pxd":82
 * 
 * 
 * cdef PeakList* isotopic_variants(Composition* composition, int npeaks, int charge=*, double charge_carrier=*) nogil             # <<<<<<<<<<<<<<
 */
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_isotopic_variants {
  int __pyx_n;
  int charge;
  double charge_carrier;
};

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":815
 * ctypedef npy_longdouble longdouble_t
 * 
 * ctypedef npy_cfloat      cfloat_t             # <<<<<<<<<<<<<<
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t
 */
typedef npy_cfloat __pyx_t_5numpy_cfloat_t;

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":816
 * 
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t             # <<<<<<<<<<<<<<
 * ctypedef npy_clongdouble clongdouble_t
 * 
 */
typedef npy_cdouble __pyx_t_5numpy_cdouble_t;

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":817
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cdouble     complex_t
 */
typedef npy_clongdouble __pyx_t_5numpy_clongdouble_t;

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":819
 * ctypedef npy_clongdouble clongdouble_t
 * 
 * ctypedef npy_cdouble     complex_t             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 */
typedef npy_cdouble __pyx_t_5numpy_complex_t;
struct __pyx_opt_args_14ms_peak_picker_2_c_10peak_index_9PeakIndex__has_peak;
struct __pyx_opt_args_14ms_peak_picker_2_c_10peak_index_9PeakIndex_all_peaks_for;

/* "ms_peak_picker/_c/peak_index.pxd":11
 *         public PeakSet peaks
 * 
 *     cdef FittedPeak _has_peak(self, double mz, double tolerance=*)             # <<<<<<<<<<<<<<
 *     cdef PeakSet _between(self, double start, double stop)
 * 
 */
struct __pyx_opt_args_14ms_peak_picker_2_c_10peak_index_9PeakIndex__has_peak {
  int __pyx_n;
  double tolerance;
};

/* "ms_peak_picker/_c/peak_index.pxd":14
 *     cdef PeakSet _between(self, double start, double stop)
 * 
 *     cpdef tuple all_peaks_for(self, double mz, double error_tolerance=*)             # <<<<<<<<<<<<<<
 * 
 *     cdef size_t get_size(self)
 */
struct __pyx_opt_args_14ms_peak_picker_2_c_10peak_index_9PeakIndex_all_peaks_for {
  int __pyx_n;
  double error_tolerance;
};
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_mass_charge_ratio;
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_neutral_mass;
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine_scale;
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine_isotopic_cluster;
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine__isotopic_cluster;
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_ignore_below;
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_truncate_after;
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_scale;
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern__scale;
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_14AveragineCache_has_mz_charge_pair;
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_14AveragineCache_isotopic_cluster;
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_isotopic_shift;

/* "ms_deisotope/_c/averagine.pxd":7
 * cdef double PROTON
 * 
 * cdef double mass_charge_ratio(double neutral_mass, int z, double charge_carrier=*)             # <<<<<<<<<<<<<<
 * cdef double neutral_mass(double mz,  int z, double charge_carrier=*)
 * cdef void slide(double mz, list peaklist)
 */
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_mass_charge_ratio {
  int __pyx_n;
  double charge_carrier;
};

/* "ms_deisotope/_c/averagine.pxd":8
 * 
 * cdef double mass_charge_ratio(double neutral_mass, int z, double charge_carrier=*)
 * cdef double neutral_mass(double mz,  int z, double charge_carrier=*)             # <<<<<<<<<<<<<<
 * cdef void slide(double mz, list peaklist)
 * cdef dict scale_dict(dict data, double factor)
 */
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_neutral_mass {
  int __pyx_n;
  double charge_carrier;
};

/* "ms_deisotope/_c/averagine.pxd":17
 *         public dict base_composition
 * 
 *     cpdef dict scale(self, double mz, int charge=*, double charge_carrier=*)             # <<<<<<<<<<<<<<
 *     cpdef TheoreticalIsotopicPattern isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)
 *     cdef TheoreticalIsotopicPattern _isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)
 */
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine_scale {
  int __pyx_n;
  int charge;
  double charge_carrier;
};

/* "ms_deisotope/_c/averagine.pxd":18
 * 
 *     cpdef dict scale(self, double mz, int charge=*, double charge_carrier=*)
 *     cpdef TheoreticalIsotopicPattern isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)             # <<<<<<<<<<<<<<
 *     cdef TheoreticalIsotopicPattern _isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)
 * 
 */
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine_isotopic_cluster {
  int __pyx_n;
  int charge;
  double charge_carrier;
  double truncate_after;
  double ignore_below;
};

/* "ms_deisotope/_c/averagine.pxd":19
 *     cpdef dict scale(self, double mz, int charge=*, double charge_carrier=*)
 *     cpdef TheoreticalIsotopicPattern isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)
 *     cdef TheoreticalIsotopicPattern _isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)             # <<<<<<<<<<<<<<
 * 
 * 
 */
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine__isotopic_cluster {
  int __pyx_n;
  int charge;
  double charge_carrier;
  double truncate_after;
  double ignore_below;
};

/* "ms_deisotope/_c/averagine.pxd":39
 *     cdef bint _eq_inst(self, TheoreticalIsotopicPattern other)
 * 
 *     cpdef TheoreticalIsotopicPattern ignore_below(self, double ignore_below=*)             # <<<<<<<<<<<<<<
 *     cpdef TheoreticalIsotopicPattern truncate_after(self, double truncate_after=*)
 *     cpdef TheoreticalIsotopicPattern shift(self, double mz)
 */
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_ignore_below {
  int __pyx_n;
  double ignore_below;
};

/* "ms_deisotope/_c/averagine.pxd":40
 * 
 *     cpdef TheoreticalIsotopicPattern ignore_below(self, double ignore_below=*)
 *     cpdef TheoreticalIsotopicPattern truncate_after(self, double truncate_after=*)             # <<<<<<<<<<<<<<
 *     cpdef TheoreticalIsotopicPattern shift(self, double mz)
 *     cpdef TheoreticalIsotopicPattern scale(self, list experimental_distribution, str method=*)
 */
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_truncate_after {
  int __pyx_n;
  double truncate_after;
};

/* "ms_deisotope/_c/averagine.pxd":42
 *     cpdef TheoreticalIsotopicPattern truncate_after(self, double truncate_after=*)
 *     cpdef TheoreticalIsotopicPattern shift(self, double mz)
 *     cpdef TheoreticalIsotopicPattern scale(self, list experimental_distribution, str method=*)             # <<<<<<<<<<<<<<
 *     cpdef TheoreticalIsotopicPattern scale_raw(self, double scale_factor)
 *     cpdef double drop_last_peak(self)
 */
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_scale {
  int __pyx_n;
  PyObject *method;
};

/* "ms_deisotope/_c/averagine.pxd":53
 * 
 *     @cython.final
 *     cdef inline TheoreticalIsotopicPattern _scale(self, list experimental_distribution, str method=*)             # <<<<<<<<<<<<<<
 * 
 *     @cython.final
 */
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern__scale {
  int __pyx_n;
  PyObject *method;
};

/* "ms_deisotope/_c/averagine.pxd":88
 *     cdef double _make_cache_key(self, double mz)
 * 
 *     cdef TheoreticalIsotopicPattern has_mz_charge_pair(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)             # <<<<<<<<<<<<<<
 *     cpdef TheoreticalIsotopicPattern isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)
 * 
 */
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_14AveragineCache_has_mz_charge_pair {
  int __pyx_n;
  int charge;
  double charge_carrier;
  double truncate_after;
  double ignore_below;
};

/* "ms_deisotope/_c/averagine.pxd":89
 * 
 *     cdef TheoreticalIsotopicPattern has_mz_charge_pair(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)
 *     cpdef TheoreticalIsotopicPattern isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)             # <<<<<<<<<<<<<<
 * 
 * 
 */
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_14AveragineCache_isotopic_cluster {
  int __pyx_n;
  int charge;
  double charge_carrier;
  double truncate_after;
  double ignore_below;
};

/* "ms_deisotope/_c/averagine.pxd":92
 * 
 * 
 * cpdef double isotopic_shift(int charge=*)             # <<<<<<<<<<<<<<
 */
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_isotopic_shift {
  int __pyx_n;
  int charge;
};
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_match_theoretical_isotopic_distribution;
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_subtraction;
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak;
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak;
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak_inplace;
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak_inplace;
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_fit_theoretical_distribution;
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase__fit_peaks_at_charges;
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_fit_theoretical_distribution;
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase__fit_peaks_at_charges;
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base__get_all_peak_charge_pairs;

/* "ms_deisotope/_c/deconvoluter_base.pxd":42
 *     cdef FittedPeak _has_peak(self, double mz, double error_tolerance)
 * 
 *     cpdef list match_theoretical_isotopic_distribution(self, list theoretical_distribution, double error_tolerance=*)             # <<<<<<<<<<<<<<
 *     cpdef scale_theoretical_distribution(self, TheoreticalIsotopicPattern theoretical_distribution, list experimental_distribution)
 *     cpdef IsotopicFitRecord _evaluate_theoretical_distribution(self, list experimental, TheoreticalIsotopicPattern theoretical, FittedPeak peak, int charge)
 */
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_match_theoretical_isotopic_distribution {
  int __pyx_n;
  double error_tolerance;
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":45
 *     cpdef scale_theoretical_distribution(self, TheoreticalIsotopicPattern theoretical_distribution, list experimental_distribution)
 *     cpdef IsotopicFitRecord _evaluate_theoretical_distribution(self, list experimental, TheoreticalIsotopicPattern theoretical, FittedPeak peak, int charge)
 *     cpdef subtraction(self, TheoreticalIsotopicPattern isotopic_cluster, double error_tolerance=*)             # <<<<<<<<<<<<<<
 *     cpdef list fit_incremental_truncation(self, IsotopicFitRecord seed_fit, double lower_bound)
 * 
 */
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_subtraction {
  int __pyx_n;
  double error_tolerance;
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":50
 *     cpdef bint _check_fit(self, IsotopicFitRecord fit)
 * 
 *     cpdef list _find_next_putative_peak(self, double mz, int charge, int step=*, double tolerance=*)             # <<<<<<<<<<<<<<
 *     cpdef list _find_previous_putative_peak(self, double mz, int charge, int step=*, double tolerance=*)
 * 
 */
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak {
  int __pyx_n;
  int step;
  double tolerance;
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":51
 * 
 *     cpdef list _find_next_putative_peak(self, double mz, int charge, int step=*, double tolerance=*)
 *     cpdef list _find_previous_putative_peak(self, double mz, int charge, int step=*, double tolerance=*)             # <<<<<<<<<<<<<<
 * 
 *     cdef int _find_next_putative_peak_inplace(self, double mz, int charge, set result, int step=*, double tolerance=*)
 */
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak {
  int __pyx_n;
  int step;
  double tolerance;
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":53
 *     cpdef list _find_previous_putative_peak(self, double mz, int charge, int step=*, double tolerance=*)
 * 
 *     cdef int _find_next_putative_peak_inplace(self, double mz, int charge, set result, int step=*, double tolerance=*)             # <<<<<<<<<<<<<<
 *     cdef int _find_previous_putative_peak_inplace(self, double mz, int charge, set result, int step=*, double tolerance=*)
 * 
 */
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak_inplace {
  int __pyx_n;
  int step;
  double tolerance;
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":54
 * 
 *     cdef int _find_next_putative_peak_inplace(self, double mz, int charge, set result, int step=*, double tolerance=*)
 *     cdef int _find_previous_putative_peak_inplace(self, double mz, int charge, set result, int step=*, double tolerance=*)             # <<<<<<<<<<<<<<
 * 
 * 
 */
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak_inplace {
  int __pyx_n;
  int step;
  double tolerance;
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":61
 *         public AveragineCache averagine
 * 
 *     cpdef IsotopicFitRecord fit_theoretical_distribution(self, FittedPeak peak, double error_tolerance, int charge,             # <<<<<<<<<<<<<<
 *                                                          double charge_carrier=*, double truncate_after=*,
 *                                                          double ignore_below=*)
 */
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_fit_theoretical_distribution {
  int __pyx_n;
  double charge_carrier;
  double truncate_after;
  double ignore_below;
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":65
 *                                                          double ignore_below=*)
 * 
 *     cpdef set _fit_peaks_at_charges(self, set peak_charge_set, double error_tolerance, double charge_carrier=*, double truncate_after=*,             # <<<<<<<<<<<<<<
 *                                     double ignore_below=*)
 * 
 */
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase__fit_peaks_at_charges {
  int __pyx_n;
  double charge_carrier;
  double truncate_after;
  double ignore_below;
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":73
 *         public list averagines
 * 
 *     cpdef IsotopicFitRecord fit_theoretical_distribution(self, FittedPeak peak, double error_tolerance, int charge,             # <<<<<<<<<<<<<<
 *                                                          AveragineCache  averagine, double charge_carrier=*, double truncate_after=*,
 *                                                          double ignore_below=*)
 */
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_fit_theoretical_distribution {
  int __pyx_n;
  double charge_carrier;
  double truncate_after;
  double ignore_below;
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":76
 *                                                          AveragineCache  averagine, double charge_carrier=*, double truncate_after=*,
 *                                                          double ignore_below=*)
 *     cpdef set _fit_peaks_at_charges(self, set peak_charge_set, double error_tolerance, double charge_carrier=*, double truncate_after=*,             # <<<<<<<<<<<<<<
 *                                     double ignore_below=*)
 * 
 */
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase__fit_peaks_at_charges {
  int __pyx_n;
  double charge_carrier;
  double truncate_after;
  double ignore_below;
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":83
 * 
 * 
 * cpdef set _get_all_peak_charge_pairs(DeconvoluterBase self, FittedPeak peak, double error_tolerance=*,             # <<<<<<<<<<<<<<
 *                                      object charge_range=*,
 *                                      int left_search_limit=*, int right_search_limit=*,
 */
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base__get_all_peak_charge_pairs {
  int __pyx_n;
  double error_tolerance;
  PyObject *charge_range;
  int left_search_limit;
  int right_search_limit;
  int recalculate_starting_peak;
  int use_quick_charge;
};
struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord__create;
struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_score_peak;
struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_21InterferenceDetection_detect_interference;

/* "ms_deisotope/_c/scoring.pxd":29
 * 
 *     @staticmethod
 *     cdef IsotopicFitRecord _create(FittedPeak seed_peak, double score, int charge, TheoreticalIsotopicPattern theoretical,             # <<<<<<<<<<<<<<
 *                                    list experimental, object data=*, int missed_peaks=*)
 * 
 */
struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord__create {
  int __pyx_n;
  PyObject *data;
  int missed_peaks;
};

/* "ms_deisotope/_c/scoring.pxd":75
 *     cdef public double mass_error_tolerance
 * 
 *     cdef double score_peak(self, FittedPeak obs, TheoreticalPeak theo, double mass_error_tolerance=*, double minimum_signal_to_noise=*) nogil             # <<<<<<<<<<<<<<
 * 
 * 
 */
struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_score_peak {
  int __pyx_n;
  double mass_error_tolerance;
  double minimum_signal_to_noise;
};

/* "ms_deisotope/_c/scoring.pxd":92
 *     cdef public PeakSet peaklist
 * 
 *     cdef double detect_interference(self, list experimental_peaks, double lower=*, double upper=*)             # <<<<<<<<<<<<<<
 * 
 * 
 */
struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_21InterferenceDetection_detect_interference {
  int __pyx_n;
  double lower;
  double upper;
};
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature__invalidate;
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage__bootstrap;
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_bootstrap;

/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":93
 *     cpdef insert_node(self, LCMSFeatureTreeNode node)
 *     cpdef insert(self, PeakBase peak, double time)
 *     cpdef _invalidate(self, bint reaverage=*)             # <<<<<<<<<<<<<<
 * 
 * 
 */
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature__invalidate {
  int __pyx_n;
  int reaverage;
};

/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":139
 *     cpdef double recompute(self)
 *     cpdef RunningWeightedAverage update(self, iterable)
 *     cpdef double _bootstrap(self, size_t n=*, size_t k=*)             # <<<<<<<<<<<<<<
 *     cpdef RunningWeightedAverage bootstrap(self, size_t n=*, size_t k=*)
 *     cpdef RunningWeightedAverage subsample(self, size_t k)
 */
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage__bootstrap {
  int __pyx_n;
  size_t n;
  size_t k;
};

/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":140
 *     cpdef RunningWeightedAverage update(self, iterable)
 *     cpdef double _bootstrap(self, size_t n=*, size_t k=*)
 *     cpdef RunningWeightedAverage bootstrap(self, size_t n=*, size_t k=*)             # <<<<<<<<<<<<<<
 *     cpdef RunningWeightedAverage subsample(self, size_t k)
 * 
 */
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_bootstrap {
  int __pyx_n;
  size_t n;
  size_t k;
};
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet__has_peak;
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet_has_peak;
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet_all_peaks_for;
struct __pyx_t_12ms_deisotope_2_c_8peak_set_index_cell;
struct __pyx_t_12ms_deisotope_2_c_8peak_set_index_list;

/* "ms_deisotope/_c/peak_set.pxd":82
 *     cpdef DeconvolutedPeakSet copy(self)
 * 
 *     cdef DeconvolutedPeak _has_peak(self, double neutral_mass, double error_tolerance=*, bint use_mz=*)             # <<<<<<<<<<<<<<
 * 
 *     cpdef DeconvolutedPeak has_peak(self, double neutral_mass, double error_tolerance=*, bint use_mz=*)
 */
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet__has_peak {
  int __pyx_n;
  double error_tolerance;
  int use_mz;
};

/* "ms_deisotope/_c/peak_set.pxd":84
 *     cdef DeconvolutedPeak _has_peak(self, double neutral_mass, double error_tolerance=*, bint use_mz=*)
 * 
 *     cpdef DeconvolutedPeak has_peak(self, double neutral_mass, double error_tolerance=*, bint use_mz=*)             # <<<<<<<<<<<<<<
 * 
 *     cpdef tuple all_peaks_for(self, double neutral_mass, double tolerance=*)
 */
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet_has_peak {
  int __pyx_n;
  double error_tolerance;
  int use_mz;
};

/* "ms_deisotope/_c/peak_set.pxd":86
 *     cpdef DeconvolutedPeak has_peak(self, double neutral_mass, double error_tolerance=*, bint use_mz=*)
 * 
 *     cpdef tuple all_peaks_for(self, double neutral_mass, double tolerance=*)             # <<<<<<<<<<<<<<
 * 
 *     cdef DeconvolutedPeak _get_nearest_peak(self, double neutral_mass, double* errout)
 */
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet_all_peaks_for {
  int __pyx_n;
  double tolerance;
};

/* "ms_deisotope/_c/peak_set.pxd":113
 * cdef size_t INTERVAL_INDEX_SIZE
 * 
 * cdef struct index_cell:             # <<<<<<<<<<<<<<
 *     double center_value
 *     size_t start
 */
struct __pyx_t_12ms_deisotope_2_c_8peak_set_index_cell {
  double center_value;
  size_t start;
  size_t end;
};

/* "ms_deisotope/_c/peak_set.pxd":119
 * 
 * 
 * cdef struct index_list:             # <<<<<<<<<<<<<<
 *     index_cell* index
 *     size_t size
 */
struct __pyx_t_12ms_deisotope_2_c_8peak_set_index_list {
  struct __pyx_t_12ms_deisotope_2_c_8peak_set_index_cell *index;
  size_t size;
  double low;
  double high;
};
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor__conform_envelopes;
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer_conform;
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_conform_envelopes;
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_create_theoretical_distribution;
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_find_all_features;
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_find_features;
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_match_theoretical_isotopic_distribution;
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_single_feature_set;
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_theoretical_distribution_on_features;
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_feature_set;
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_composition;

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":150
 * 
 * 
 * cdef tuple _conform_envelopes(list experimental, list base_theoretical, size_t* n_missing_out,             # <<<<<<<<<<<<<<
 *                               double minimum_theoretical_abundance=0.05):
 *     cdef:
 */
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor__conform_envelopes {
  int __pyx_n;
  double minimum_theoretical_abundance;
};

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":181
 *         self.n_missing = 0
 * 
 *     cdef void conform(self, double minimum_theoretical_abundance=0.05):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double total = 0
 */
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer_conform {
  int __pyx_n;
  double minimum_theoretical_abundance;
};

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":215
 * 
 * 
 * cpdef tuple conform_envelopes(list experimental, list base_theoretical,             # <<<<<<<<<<<<<<
 *                               double minimum_theoretical_abundance=0.01):
 *     cdef:
 */
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_conform_envelopes {
  int __pyx_n;
  double minimum_theoretical_abundance;
};

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":246
 *         public AveragineCache averagine
 * 
 *     cpdef list create_theoretical_distribution(self, double mz, int charge, double charge_carrier=PROTON, double truncate_after=0.8,             # <<<<<<<<<<<<<<
 *                                                double ignore_below=0.05):
 *         cdef:
 */
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_create_theoretical_distribution {
  int __pyx_n;
  double charge_carrier;
  double truncate_after;
  double ignore_below;
};

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":255
 *         return base_tid
 * 
 *     cpdef list find_all_features(self, double mz, double error_tolerance=2e-5):             # <<<<<<<<<<<<<<
 *         return self.feature_map._find_all(mz, error_tolerance)
 * 
 */
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_find_all_features {
  int __pyx_n;
  double error_tolerance;
};

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":258
 *         return self.feature_map._find_all(mz, error_tolerance)
 * 
 *     cpdef list find_features(self, double mz, double error_tolerance=2e-5, LCMSFeature interval=None):             # <<<<<<<<<<<<<<
 *         cdef:
 *             list f, overlapped
 */
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_find_features {
  int __pyx_n;
  double error_tolerance;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *interval;
};

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":288
 *         return f
 * 
 *     cpdef list match_theoretical_isotopic_distribution(self, list theoretical_distribution, double error_tolerance=2e-5, LCMSFeature interval=None):             # <<<<<<<<<<<<<<
 *         """Given a list of theoretical peaks, find their counterparts in :attr:`peaklist` within `error_tolerance`
 *         ppm error. If no experimental peak is found, a placeholder will be used in its stead.
 */
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_match_theoretical_isotopic_distribution {
  int __pyx_n;
  double error_tolerance;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *interval;
};

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":346
 *         return acc / count
 * 
 *     cpdef LCMSFeatureSetFit _fit_single_feature_set(self, list features, list base_tid, double error_tolerance,             # <<<<<<<<<<<<<<
 *                                                     int charge, double charge_carrier=PROTON, int max_missed_peaks=1,
 *                                                     double threshold_scale=0.3):
 */
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_single_feature_set {
  int __pyx_n;
  double charge_carrier;
  int max_missed_peaks;
  double threshold_scale;
};

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":402
 *         return fit
 * 
 *     cdef list _fit_theoretical_distribution_on_features(self, double mz, double error_tolerance, int charge,             # <<<<<<<<<<<<<<
 *                                                         TheoreticalIsotopicPattern base_tid,
 *                                                         double charge_carrier=PROTON,
 */
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_theoretical_distribution_on_features {
  int __pyx_n;
  double charge_carrier;
  double truncate_after;
  int max_missed_peaks;
  double threshold_scale;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *feature;
};

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":509
 *         return feature_fits
 * 
 *     cpdef list _fit_feature_set(self, double mz, double error_tolerance, int charge, int left_search=1,             # <<<<<<<<<<<<<<
 *                                 int right_search=1, double charge_carrier=PROTON, double truncate_after=0.8,
 *                                 int max_missed_peaks=1, double threshold_scale=0.3, LCMSFeature feature=None):
 */
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_feature_set {
  int __pyx_n;
  int left_search;
  int right_search;
  double charge_carrier;
  double truncate_after;
  int max_missed_peaks;
  double threshold_scale;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *feature;
};

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":523
 *         return feature_fits
 * 
 *     cpdef list _fit_composition(self, dict composition, double error_tolerance, int charge, double charge_carrier=PROTON,             # <<<<<<<<<<<<<<
 *                                 double truncate_after=0.8, double ignore_below=0.05, int max_missed_peaks=1,
 *                                 double threshold_scale=0.3, LCMSFeature feature=None):
 */
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_composition {
  int __pyx_n;
  double charge_carrier;
  double truncate_after;
  double ignore_below;
  int max_missed_peaks;
  double threshold_scale;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *feature;
};

/* "ms_peak_picker/_c/peak_set.pxd":2
 * 
 * cdef class PeakBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mz
 */
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase {
  PyObject_HEAD
  struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_vtab;
  double mz;
  double intensity;
  double area;
};


/* "ms_peak_picker/_c/peak_set.pxd":12
 * 
 * 
 * cdef class FittedPeak(PeakBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double signal_to_noise
 */
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase __pyx_base;
  double signal_to_noise;
  double full_width_at_half_max;
  double left_width;
  double right_width;
  long peak_count;
  long index;
};


/* "ms_peak_picker/_c/peak_set.pxd":29
 *                             double area)
 * 
 * cdef class PeakSet(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public tuple peaks
 */
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet {
  PyObject_HEAD
  struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_vtab;
  PyObject *peaks;
  int indexed;
};


/* "ms_peak_picker/_c/peak_set.pxd":81
 * 
 * 
 * cdef class PeakSetIndexed(PeakSet):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double* mz_index
 */
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet __pyx_base;
  double *mz_index;
  struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_list *interval_index;
};


/* "brainpy/_c/composition.pxd":123
 * 
 * 
 * cdef class PyComposition(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         Composition* impl
 */
struct __pyx_obj_7brainpy_2_c_11composition_PyComposition {
  PyObject_HEAD
  struct __pyx_vtabstruct_7brainpy_2_c_11composition_PyComposition *__pyx_vtab;
  struct __pyx_t_7brainpy_2_c_11composition_Composition *impl;
  double cached_mass;
  int _clean;
};


/* "brainpy/_c/isotopic_distribution.pxd":63
 * 
 * @freelist(100000)
 * cdef class TheoreticalPeak(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mz
 */
struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak {
  PyObject_HEAD
  struct __pyx_vtabstruct_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_vtab;
  double mz;
  double intensity;
  int charge;
};


/* "ms_peak_picker/_c/peak_index.pxd":5
 * 
 * 
 * cdef class PeakIndex(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public np.ndarray mz_array
 */
struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex {
  PyObject_HEAD
  struct __pyx_vtabstruct_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_vtab;
  PyArrayObject *mz_array;
  PyArrayObject *intensity_array;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *peaks;
};


/* "ms_deisotope/_c/averagine.pxd":12
 * cdef dict scale_dict(dict data, double factor)
 * 
 * cdef class Averagine(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double base_mass
 */
struct __pyx_obj_12ms_deisotope_2_c_9averagine_Averagine {
  PyObject_HEAD
  struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_Averagine *__pyx_vtab;
  double base_mass;
  PyObject *base_composition;
};


/* "ms_deisotope/_c/averagine.pxd":23
 * 
 * @cython.final
 * cdef class TheoreticalIsotopicPattern(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list peaklist
 */
struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern {
  PyObject_HEAD
  struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *__pyx_vtab;
  PyObject *peaklist;
  double origin;
  double offset;
};


/* "ms_deisotope/_c/averagine.pxd":79
 * 
 * @cython.final
 * cdef class AveragineCache(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public dict backend
 */
struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache {
  PyObject_HEAD
  struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_AveragineCache *__pyx_vtab;
  PyObject *backend;
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_Averagine *averagine;
  double cache_truncation;
  int enabled;
};


/* "ms_deisotope/_c/deconvoluter_base.pxd":25
 * 
 * 
 * cdef class DeconvoluterBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public bint use_subtraction
 */
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase {
  PyObject_HEAD
  struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_vtab;
  int use_subtraction;
  PyObject *scale_method;
  int merge_isobaric_peaks;
  double minimum_intensity;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *peaklist;
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *scorer;
  int verbose;
  PyObject *_slice_cache;
};


/* "ms_deisotope/_c/deconvoluter_base.pxd":57
 * 
 * 
 * cdef class AveragineDeconvoluterBase(DeconvoluterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public AveragineCache averagine
 */
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase {
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase __pyx_base;
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *averagine;
};


/* "ms_deisotope/_c/deconvoluter_base.pxd":69
 * 
 * 
 * cdef class MultiAveragineDeconvoluterBase(DeconvoluterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list averagines
 */
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase {
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase __pyx_base;
  PyObject *averagines;
};


/* "ms_deisotope/_c/deconvoluter_base.pxd":91
 * 
 * 
 * cdef class ChargeIterator(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public int lower
 */
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator {
  PyObject_HEAD
  struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_vtab;
  int lower;
  int upper;
  int sign;
  int *values;
  size_t size;
  size_t index;
};


/* "ms_deisotope/_c/scoring.pxd":16
 * 
 * 
 * cdef class IsotopicFitRecord(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public FittedPeak seed_peak
 */
struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord {
  PyObject_HEAD
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_vtab;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *seed_peak;
  double score;
  int charge;
  PyObject *experimental;
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *theoretical;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *monoisotopic_peak;
  int missed_peaks;
  PyObject *data;
  Py_hash_t _hash;
};


/* "ms_deisotope/_c/scoring.pxd":38
 * 
 * 
 * cdef class FitSelectorBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double minimum_score
 */
struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase {
  PyObject_HEAD
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FitSelectorBase *__pyx_vtab;
  double minimum_score;
};


/* "ms_deisotope/_c/scoring.pxd":50
 * 
 * 
 * cdef class IsotopicFitterBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public FitSelectorBase select
 */
struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase {
  PyObject_HEAD
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_vtab;
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *select;
};


/* "ms_deisotope/_c/scoring.pxd":62
 * 
 * 
 * cdef class LeastSquaresFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_12ms_deisotope_2_c_7scoring_LeastSquaresFitter {
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;
};


/* "ms_deisotope/_c/scoring.pxd":66
 * 
 * 
 * cdef class GTestFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_12ms_deisotope_2_c_7scoring_GTestFitter {
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;
};


/* "ms_deisotope/_c/scoring.pxd":69
 *     pass
 * 
 * cdef class ScaledGTestFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledGTestFitter {
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;
};


/* "ms_deisotope/_c/scoring.pxd":72
 *     pass
 * 
 * cdef class MSDeconVFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef public double mass_error_tolerance
 * 
 */
struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter {
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;
  double mass_error_tolerance;
};


/* "ms_deisotope/_c/scoring.pxd":78
 * 
 * 
 * cdef class PenalizedMSDeconVFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mass_error_tolerance
 */
struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter {
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;
  double mass_error_tolerance;
  double penalty_factor;
};


/* "ms_deisotope/_c/scoring.pxd":84
 * 
 * 
 * cdef class FunctionScorer(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         object function
 */
struct __pyx_obj_12ms_deisotope_2_c_7scoring_FunctionScorer {
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;
  PyObject *function;
};


/* "ms_deisotope/_c/scoring.pxd":89
 * 
 * 
 * cdef class InterferenceDetection(object):             # <<<<<<<<<<<<<<
 *     cdef public PeakSet peaklist
 * 
 */
struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection {
  PyObject_HEAD
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_InterferenceDetection *__pyx_vtab;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *peaklist;
};


/* "ms_deisotope/_c/scoring.pxd":95
 * 
 * 
 * cdef class DistinctPatternFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public InterferenceDetection interference_detector
 */
struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter {
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection *interference_detector;
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledGTestFitter *g_test_scaled;
  double peak_count_scale;
  double domain_scale;
};


/* "ms_deisotope/_c/scoring.pxd":102
 *         public double domain_scale
 * 
 * cdef class ScaledPenalizedMSDeconvFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double scale_factor
 */
struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter {
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;
  double scale_factor;
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter *scorer;
};


/* "ms_deisotope/_c/scoring.pxd":112
 * 
 * 
 * cdef class DotProductFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     pass
 */
struct __pyx_obj_12ms_deisotope_2_c_7scoring_DotProductFitter {
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;
};


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":4
 * 
 * 
 * cdef class LCMSFeatureTreeList(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list roots
 */
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList {
  PyObject_HEAD
  struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_vtab;
  PyObject *roots;
  PyObject *_node_id_hash;
};


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":23
 * 
 * 
 * cdef class LCMSFeatureTreeNodeBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double time
 */
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase {
  PyObject_HEAD
  struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_vtab;
  double time;
  PyObject *members;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *_most_abundant_member;
  PyObject *node_id;
};


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":44
 * 
 * 
 * cdef class LCMSFeatureTreeNode(LCMSFeatureTreeNodeBase):             # <<<<<<<<<<<<<<
 *     cdef:
 * 
 */
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode {
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase __pyx_base;
  double _mz;
};


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":56
 * 
 * 
 * cdef class FeatureBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public LCMSFeatureTreeList nodes
 */
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase {
  PyObject_HEAD
  struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_vtab;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *nodes;
  double _mz;
  double _start_time;
  double _end_time;
};


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":72
 * 
 * 
 * cdef class LCMSFeature(FeatureBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double _total_intensity
 */
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature {
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase __pyx_base;
  double _total_intensity;
  double _last_mz;
  PyObject *_times;
  PyObject *_peaks;
  PyObject *adducts;
  PyObject *used_as_adduct;
  PyObject *created_at;
  PyObject *feature_id;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *_peak_averager;
};


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":96
 * 
 * 
 * cdef class EmptyFeature(FeatureBase):             # <<<<<<<<<<<<<<
 *     @staticmethod
 *     cdef EmptyFeature _create(double mz)
 */
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature {
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase __pyx_base;
};


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":101
 * 
 * 
 * cdef class FeatureSetIterator(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list features
 */
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator {
  PyObject_HEAD
  struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_vtab;
  PyObject *features;
  PyObject *real_features;
  double start_time;
  double end_time;
  double last_time_seen;
  size_t *index_list;
};


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":129
 * 
 * 
 * cdef class RunningWeightedAverage(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list accumulator
 */
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage {
  PyObject_HEAD
  struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *__pyx_vtab;
  PyObject *accumulator;
  double current_mean;
  size_t current_count;
  double total_weight;
};


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":148
 * 
 * 
 * cdef class RunningWeightedAverageNeutralMass(RunningWeightedAverage):             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass {
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage __pyx_base;
};


/* "ms_deisotope/_c/feature_map/feature_map.pxd":5
 * 
 * 
 * cdef class LCMSFeatureMap(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list features
 */
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap {
  PyObject_HEAD
  struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_vtab;
  PyObject *features;
};


/* "ms_deisotope/_c/peak_set.pxd":5
 * 
 * 
 * cdef class _Index:             # <<<<<<<<<<<<<<
 *     cdef:
 *         public size_t neutral_mass
 */
struct __pyx_obj_12ms_deisotope_2_c_8peak_set__Index {
  PyObject_HEAD
  struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set__Index *__pyx_vtab;
  size_t neutral_mass;
  size_t mz;
};


/* "ms_deisotope/_c/peak_set.pxd":14
 * 
 * 
 * cdef class EnvelopePair:             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mz
 */
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_EnvelopePair {
  PyObject_HEAD
  struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_EnvelopePair *__pyx_vtab;
  double mz;
  double intensity;
};


/* "ms_deisotope/_c/peak_set.pxd":25
 * 
 * 
 * cdef class Envelope:             # <<<<<<<<<<<<<<
 *     cdef public tuple pairs
 * 
 */
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope {
  PyObject_HEAD
  struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_Envelope *__pyx_vtab;
  PyObject *pairs;
};


/* "ms_deisotope/_c/peak_set.pxd":40
 * 
 * 
 * cdef class DeconvolutedPeak(PeakBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double neutral_mass
 */
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase __pyx_base;
  double neutral_mass;
  double signal_to_noise;
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set__Index *_index;
  double full_width_at_half_max;
  int charge;
  double a_to_a2_ratio;
  double most_abundant_mass;
  double average_mass;
  double score;
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *envelope;
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *fit;
  int chosen_for_msms;
};


/* "ms_deisotope/_c/peak_set.pxd":66
 * 
 * 
 * cdef class DeconvolutedPeakSolution(DeconvolutedPeak):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public object solution
 */
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution {
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak __pyx_base;
  PyObject *solution;
};


/* "ms_deisotope/_c/peak_set.pxd":70
 *         public object solution
 * 
 * cdef class DeconvolutedPeakSet(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public tuple peaks
 */
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet {
  PyObject_HEAD
  struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *__pyx_vtab;
  PyObject *peaks;
  PyObject *_mz_ordered;
  int indexed;
};


/* "ms_deisotope/_c/peak_set.pxd":99
 * 
 * 
 * cdef class DeconvolutedPeakSetIndexed(DeconvolutedPeakSet):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double* neutral_mass_array
 */
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed {
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet __pyx_base;
  double *neutral_mass_array;
  double *mz_array;
  struct __pyx_t_12ms_deisotope_2_c_8peak_set_index_list *interval_index;
  size_t _size;
};


/* "ms_deisotope/_c/feature_map/feature_fit.pxd":8
 * import numpy as np
 * 
 * cdef class map_coord:             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mz
 */
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_map_coord {
  PyObject_HEAD
  double mz;
  double time;
};


/* "ms_deisotope/_c/feature_map/feature_fit.pxd":14
 * 
 * 
 * cdef class LCMSFeatureSetFit(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list features
 */
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit {
  PyObject_HEAD
  struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit *__pyx_vtab;
  PyObject *features;
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *theoretical;
  PyObject *supporters;
  double score;
  double mz;
  double neutral_mass;
  int charge;
  PyObject *data;
  size_t missing_features;
  size_t n_points;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *monoisotopic_feature;
  PyArrayObject *scores;
  PyArrayObject *times;
};


/* "ms_deisotope/_c/feature_map/feature_processor.pyx":60
 * 
 * 
 * cdef class CartesianProductIterator(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list collections
 */
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator {
  PyObject_HEAD
  struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_vtab;
  PyObject *collections;
  size_t size;
  int *lengths;
  int *indices;
  int done;
  size_t total_combinations;
};


/* "ms_deisotope/_c/feature_map/feature_processor.pyx":166
 * 
 * 
 * cdef class envelope_conformer:             # <<<<<<<<<<<<<<
 *     cdef:
 *         list experimental
 */
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer {
  PyObject_HEAD
  struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *__pyx_vtab;
  PyObject *experimental;
  PyObject *theoretical;
  size_t n_missing;
};


/* "ms_deisotope/_c/feature_map/feature_processor.pyx":240
 * 
 * 
 * cdef class LCMSFeatureProcessorBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public LCMSFeatureMap feature_map
 */
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase {
  PyObject_HEAD
  struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_vtab;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *feature_map;
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *scorer;
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *averagine;
};



/* "ms_peak_picker/_c/peak_set.pxd":2
 * 
 * cdef class PeakBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mz
 */

struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakBase {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *(*clone)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *, int __pyx_skip_dispatch);
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *(*copy)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakBase;


/* "ms_peak_picker/_c/peak_set.pxd":12
 * 
 * 
 * cdef class FittedPeak(PeakBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double signal_to_noise
 */

struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_FittedPeak {
  struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakBase __pyx_base;
  int (*_eq)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, int __pyx_skip_dispatch);
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*_create)(double, double, double, double, double, double, long, long, double);
};
static struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_FittedPeak;


/* "ms_peak_picker/_c/peak_set.pxd":29
 *                             double area)
 * 
 * cdef class PeakSet(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public tuple peaks
 */

struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*_create)(PyObject *);
  PyObject *(*reindex)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch);
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*_get_nearest_peak)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, double, double *);
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*_has_peak)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, double, struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet__has_peak *__pyx_optional_args);
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*clone)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch);
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*copy)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch);
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*_between)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, double, double);
  int (*_between_bounds)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, double, double, size_t *, size_t *);
  PyObject *(*all_peaks_for)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet_all_peaks_for *__pyx_optional_args);
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*getitem)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, size_t);
  size_t (*get_size)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *);
  int (*_eq)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSet;


/* "ms_peak_picker/_c/peak_set.pxd":81
 * 
 * 
 * cdef class PeakSetIndexed(PeakSet):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double* mz_index
 */

struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed {
  struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet __pyx_base;
  PyObject *(*_allocate_index)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *, int __pyx_skip_dispatch);
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*has_peak_hinted)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_has_peak_hinted *__pyx_optional_args);
};
static struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed;


/* "brainpy/_c/composition.pxd":123
 * 
 * 
 * cdef class PyComposition(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         Composition* impl
 */

struct __pyx_vtabstruct_7brainpy_2_c_11composition_PyComposition {
  struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *(*_create)(struct __pyx_t_7brainpy_2_c_11composition_Composition *);
  void (*_set_impl)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, struct __pyx_t_7brainpy_2_c_11composition_Composition *, struct __pyx_opt_args_7brainpy_2_c_11composition_13PyComposition__set_impl *__pyx_optional_args);
  void (*_initialize_from_formula)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, PyObject *);
  double (*mass)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, int __pyx_skip_dispatch);
  int (*__pyx___equality_pycomposition)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, int __pyx_skip_dispatch);
  int (*__pyx___equality_dict)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, PyObject *, int __pyx_skip_dispatch);
  struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *(*copy)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, int __pyx_skip_dispatch);
  PyObject *(*update)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*keys)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, int __pyx_skip_dispatch);
  PyObject *(*values)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, int __pyx_skip_dispatch);
  PyObject *(*items)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, int __pyx_skip_dispatch);
  PyObject *(*pop)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_7brainpy_2_c_11composition_13PyComposition_pop *__pyx_optional_args);
  __pyx_t_7brainpy_2_c_11composition_count_type (*getitem)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, PyObject *);
  void (*setitem)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, PyObject *, __pyx_t_7brainpy_2_c_11composition_count_type);
  void (*increment)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, PyObject *, __pyx_t_7brainpy_2_c_11composition_count_type);
  void (*add_from)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *);
  void (*subtract_from)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *);
  void (*scale_by)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, long);
};
static struct __pyx_vtabstruct_7brainpy_2_c_11composition_PyComposition *__pyx_vtabptr_7brainpy_2_c_11composition_PyComposition;


/* "brainpy/_c/isotopic_distribution.pxd":63
 * 
 * @freelist(100000)
 * cdef class TheoreticalPeak(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mz
 */

struct __pyx_vtabstruct_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak {
  struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *(*_create)(double, double, int);
  int (*_eq)(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *, struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *, int __pyx_skip_dispatch);
  struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *(*clone)(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_vtabptr_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak;


/* "ms_peak_picker/_c/peak_index.pxd":5
 * 
 * 
 * cdef class PeakIndex(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public np.ndarray mz_array
 */

struct __pyx_vtabstruct_14ms_peak_picker_2_c_10peak_index_PeakIndex {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*_has_peak)(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *, double, struct __pyx_opt_args_14ms_peak_picker_2_c_10peak_index_9PeakIndex__has_peak *__pyx_optional_args);
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*_between)(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *, double, double);
  PyObject *(*all_peaks_for)(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_14ms_peak_picker_2_c_10peak_index_9PeakIndex_all_peaks_for *__pyx_optional_args);
  size_t (*get_size)(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *);
  struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *(*clone)(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *, int __pyx_skip_dispatch);
  struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *(*copy)(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_vtabptr_14ms_peak_picker_2_c_10peak_index_PeakIndex;


/* "ms_deisotope/_c/averagine.pxd":12
 * cdef dict scale_dict(dict data, double factor)
 * 
 * cdef class Averagine(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double base_mass
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_Averagine {
  PyObject *(*scale)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_Averagine *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine_scale *__pyx_optional_args);
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*isotopic_cluster)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_Averagine *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine_isotopic_cluster *__pyx_optional_args);
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*_isotopic_cluster)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_Averagine *, double, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine__isotopic_cluster *__pyx_optional_args);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_Averagine *__pyx_vtabptr_12ms_deisotope_2_c_9averagine_Averagine;


/* "ms_deisotope/_c/averagine.pxd":23
 * 
 * @cython.final
 * cdef class TheoreticalIsotopicPattern(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list peaklist
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern {
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*_create)(PyObject *, double, double);
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*clone)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch);
  int (*_eq)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, PyObject *, int __pyx_skip_dispatch);
  int (*_eq_inst)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *);
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*ignore_below)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_ignore_below *__pyx_optional_args);
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*truncate_after)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_truncate_after *__pyx_optional_args);
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*shift)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, double, int __pyx_skip_dispatch);
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*scale)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_scale *__pyx_optional_args);
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*scale_raw)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, double, int __pyx_skip_dispatch);
  double (*drop_last_peak)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch);
  double (*total)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch);
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*normalize)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch);
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*_scale)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, PyObject *, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern__scale *__pyx_optional_args);
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*clone_shift)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, double);
  struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *(*get)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, Py_ssize_t);
  size_t (*get_size)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *);
  double (*get_monoisotopic_mz)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *);
  PyObject *(*get_processed_peaks)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *);
  size_t (*_basepeak_index)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch);
  PyObject *(*incremental_truncation)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, double, int __pyx_skip_dispatch);
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*clone_drop_last)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *__pyx_vtabptr_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern;


/* "ms_deisotope/_c/averagine.pxd":79
 * 
 * @cython.final
 * cdef class AveragineCache(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public dict backend
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_AveragineCache {
  double (*_make_cache_key)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *, double);
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*has_mz_charge_pair)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *, double, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_14AveragineCache_has_mz_charge_pair *__pyx_optional_args);
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*isotopic_cluster)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_14AveragineCache_isotopic_cluster *__pyx_optional_args);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_AveragineCache *__pyx_vtabptr_12ms_deisotope_2_c_9averagine_AveragineCache;


/* "ms_deisotope/_c/deconvoluter_base.pxd":25
 * 
 * 
 * cdef class DeconvoluterBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public bint use_subtraction
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*between)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, double, int __pyx_skip_dispatch);
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*has_peak)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, double, int __pyx_skip_dispatch);
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*_has_peak)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, double);
  PyObject *(*match_theoretical_isotopic_distribution)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_match_theoretical_isotopic_distribution *__pyx_optional_args);
  PyObject *(*scale_theoretical_distribution)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, PyObject *, int __pyx_skip_dispatch);
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *(*_evaluate_theoretical_distribution)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, PyObject *, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, int, int __pyx_skip_dispatch);
  PyObject *(*subtraction)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_subtraction *__pyx_optional_args);
  PyObject *(*fit_incremental_truncation)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, double, int __pyx_skip_dispatch);
  int (*_check_fit)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch);
  PyObject *(*_find_next_putative_peak)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, int, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak *__pyx_optional_args);
  PyObject *(*_find_previous_putative_peak)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, int, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak *__pyx_optional_args);
  int (*_find_next_putative_peak_inplace)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, int, PyObject *, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak_inplace *__pyx_optional_args);
  int (*_find_previous_putative_peak_inplace)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, int, PyObject *, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak_inplace *__pyx_optional_args);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase;


/* "ms_deisotope/_c/deconvoluter_base.pxd":57
 * 
 * 
 * cdef class AveragineDeconvoluterBase(DeconvoluterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public AveragineCache averagine
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase {
  struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase __pyx_base;
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *(*fit_theoretical_distribution)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, double, int, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_fit_theoretical_distribution *__pyx_optional_args);
  PyObject *(*_fit_peaks_at_charges)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *, PyObject *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase__fit_peaks_at_charges *__pyx_optional_args);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase;


/* "ms_deisotope/_c/deconvoluter_base.pxd":69
 * 
 * 
 * cdef class MultiAveragineDeconvoluterBase(DeconvoluterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list averagines
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase {
  struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase __pyx_base;
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *(*fit_theoretical_distribution)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, double, int, struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_fit_theoretical_distribution *__pyx_optional_args);
  PyObject *(*_fit_peaks_at_charges)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *, PyObject *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase__fit_peaks_at_charges *__pyx_optional_args);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase;


/* "ms_deisotope/_c/deconvoluter_base.pxd":91
 * 
 * 
 * cdef class ChargeIterator(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public int lower
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator {
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *(*_create)(int, int);
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *(*from_tuple)(PyObject *);
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *(*from_quickcharge)(PyObject *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *);
  PyObject *(*sequence_from_quickcharge)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, int __pyx_skip_dispatch);
  PyObject *(*make_sequence)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *, int __pyx_skip_dispatch);
  void (*release_sequence)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *);
  void (*set_bounds)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *, int, int);
  int (*has_more)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *, int __pyx_skip_dispatch);
  PyObject *(*reset)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *, int __pyx_skip_dispatch);
  int (*get_next_value)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *, int __pyx_skip_dispatch);
  size_t (*get_size)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator;


/* "ms_deisotope/_c/scoring.pxd":16
 * 
 * 
 * cdef class IsotopicFitRecord(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public FittedPeak seed_peak
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitRecord {
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *(*_create)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, double, int, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, PyObject *, struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord__create *__pyx_optional_args);
  int (*_eq)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch);
  int (*_ne)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch);
  int (*_lt)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch);
  int (*_gt)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_IsotopicFitRecord;


/* "ms_deisotope/_c/scoring.pxd":38
 * 
 * 
 * cdef class FitSelectorBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double minimum_score
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FitSelectorBase {
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *(*best)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *, PyObject *, int __pyx_skip_dispatch);
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *(*_best_from_set)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *, PyObject *);
  int (*reject)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch);
  int (*reject_score)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *, double, int __pyx_skip_dispatch);
  int (*is_maximizing)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FitSelectorBase *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_FitSelectorBase;


/* "ms_deisotope/_c/scoring.pxd":50
 * 
 * 
 * cdef class IsotopicFitterBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public FitSelectorBase select
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase {
  double (*_evaluate)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, PyObject *, PyObject *, int __pyx_skip_dispatch);
  int (*reject)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch);
  int (*reject_score)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *, double, int __pyx_skip_dispatch);
  int (*is_maximizing)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *, int __pyx_skip_dispatch);
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *(*_configure)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *, struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, PyObject *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_IsotopicFitterBase;


/* "ms_deisotope/_c/scoring.pxd":62
 * 
 * 
 * cdef class LeastSquaresFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_LeastSquaresFitter {
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_LeastSquaresFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_LeastSquaresFitter;


/* "ms_deisotope/_c/scoring.pxd":66
 * 
 * 
 * cdef class GTestFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_GTestFitter {
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_GTestFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_GTestFitter;


/* "ms_deisotope/_c/scoring.pxd":69
 *     pass
 * 
 * cdef class ScaledGTestFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_ScaledGTestFitter {
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_ScaledGTestFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_ScaledGTestFitter;


/* "ms_deisotope/_c/scoring.pxd":72
 *     pass
 * 
 * cdef class MSDeconVFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef public double mass_error_tolerance
 * 
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_MSDeconVFitter {
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;
  double (*score_peak)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *, struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_score_peak *__pyx_optional_args);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_MSDeconVFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_MSDeconVFitter;


/* "ms_deisotope/_c/scoring.pxd":78
 * 
 * 
 * cdef class PenalizedMSDeconVFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mass_error_tolerance
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter {
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter;


/* "ms_deisotope/_c/scoring.pxd":84
 * 
 * 
 * cdef class FunctionScorer(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         object function
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FunctionScorer {
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FunctionScorer *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_FunctionScorer;


/* "ms_deisotope/_c/scoring.pxd":89
 * 
 * 
 * cdef class InterferenceDetection(object):             # <<<<<<<<<<<<<<
 *     cdef public PeakSet peaklist
 * 
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_InterferenceDetection {
  double (*detect_interference)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection *, PyObject *, struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_21InterferenceDetection_detect_interference *__pyx_optional_args);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_InterferenceDetection *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_InterferenceDetection;


/* "ms_deisotope/_c/scoring.pxd":95
 * 
 * 
 * cdef class DistinctPatternFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public InterferenceDetection interference_detector
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_DistinctPatternFitter {
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_DistinctPatternFitter;


/* "ms_deisotope/_c/scoring.pxd":102
 *         public double domain_scale
 * 
 * cdef class ScaledPenalizedMSDeconvFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double scale_factor
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter {
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;
  double (*_calculate_scale_factor)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch);
  void (*scale_fitted_peaks)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *, PyObject *, double);
  void (*scale_theoretical_peaks)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *, PyObject *, double);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter;


/* "ms_deisotope/_c/scoring.pxd":112
 * 
 * 
 * cdef class DotProductFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     pass
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_DotProductFitter {
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_DotProductFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_DotProductFitter;


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":4
 * 
 * 
 * cdef class LCMSFeatureTreeList(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list roots
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList {
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *(*_create)(PyObject *);
  void (*_invalidate)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *);
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *(*_make_node)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *, double, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*find_time)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *, double, int __pyx_skip_dispatch);
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *(*_find_time)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *, double, size_t *);
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *(*getitem)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *, size_t);
  size_t (*get_size)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList;


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":23
 * 
 * 
 * cdef class LCMSFeatureTreeNodeBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double time
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase {
  double (*_total_intensity_members)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *, int __pyx_skip_dispatch);
  double (*max_intensity)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *, int __pyx_skip_dispatch);
  double (*total_intensity)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *, int __pyx_skip_dispatch);
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *(*_find_most_abundant_member)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *);
  PyObject *(*_calculate_most_abundant_member)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *, int __pyx_skip_dispatch);
  PyObject *(*_recalculate)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *, int __pyx_skip_dispatch);
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *(*getitem)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *, size_t);
  size_t (*get_members_size)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase;


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":44
 * 
 * 
 * cdef class LCMSFeatureTreeNode(LCMSFeatureTreeNodeBase):             # <<<<<<<<<<<<<<
 *     cdef:
 * 
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode {
  struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase __pyx_base;
  int (*_eq)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *, int __pyx_skip_dispatch);
  int (*_ne)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *, int __pyx_skip_dispatch);
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*getpeak)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *, size_t);
  double (*get_mz)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode;


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":56
 * 
 * 
 * cdef class FeatureBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public LCMSFeatureTreeList nodes
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase {
  double (*get_mz)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *);
  double (*get_start_time)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *);
  double (*get_end_time)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *);
  size_t (*get_size)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *);
  PyObject *(*find_time)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *, double, int __pyx_skip_dispatch);
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *(*_find_time)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *, double, size_t *);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase;


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":72
 * 
 * 
 * cdef class LCMSFeature(FeatureBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double _total_intensity
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature {
  struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase __pyx_base;
  int (*_eq)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, PyObject *, int __pyx_skip_dispatch);
  int (*_ne)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, PyObject *, int __pyx_skip_dispatch);
  int (*overlaps_in_time)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *, int __pyx_skip_dispatch);
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *(*getitem)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, size_t);
  void (*_feed_peak_averager)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *);
  PyObject *(*insert_node)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *, int __pyx_skip_dispatch);
  PyObject *(*insert)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *, double, int __pyx_skip_dispatch);
  PyObject *(*_invalidate)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature__invalidate *__pyx_optional_args);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature;


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":96
 * 
 * 
 * cdef class EmptyFeature(FeatureBase):             # <<<<<<<<<<<<<<
 *     @staticmethod
 *     cdef EmptyFeature _create(double mz)
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature {
  struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase __pyx_base;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature *(*_create)(double);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature *__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature;


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":101
 * 
 * 
 * cdef class FeatureSetIterator(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list features
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator {
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *(*_create)(PyObject *);
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *(*_create_with_threshold)(PyObject *, PyObject *, double);
  void (*_initialize)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *, PyObject *);
  PyObject *(*init_indices)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *, int __pyx_skip_dispatch);
  double (*get_next_time)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *, int __pyx_skip_dispatch);
  double (*get_current_time)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *, int __pyx_skip_dispatch);
  int (*has_more)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *, int __pyx_skip_dispatch);
  PyObject *(*get_peaks_for_time)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *, double, int __pyx_skip_dispatch);
  PyObject *(*get_next_value)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *, int __pyx_skip_dispatch);
  size_t (*get_size)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *);
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *(*getitem)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *, size_t);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator;


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":129
 * 
 * 
 * cdef class RunningWeightedAverage(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list accumulator
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage {
  PyObject *(*add)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *, int __pyx_skip_dispatch);
  double (*recompute)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *, int __pyx_skip_dispatch);
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *(*update)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *, PyObject *, int __pyx_skip_dispatch);
  double (*_bootstrap)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage__bootstrap *__pyx_optional_args);
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *(*bootstrap)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_bootstrap *__pyx_optional_args);
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *(*subsample)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *, size_t, int __pyx_skip_dispatch);
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *(*_create)(PyObject *);
  void (*_update)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *, PyObject *);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage;


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":148
 * 
 * 
 * cdef class RunningWeightedAverageNeutralMass(RunningWeightedAverage):             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass {
  struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage __pyx_base;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass *(*_create)(PyObject *);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass *__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass;


/* "ms_deisotope/_c/feature_map/feature_map.pxd":5
 * 
 * 
 * cdef class LCMSFeatureMap(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list features
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap {
  PyObject *(*_find_all)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *, double, double, int __pyx_skip_dispatch);
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *(*_search)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *, double, double, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap;


/* "ms_deisotope/_c/peak_set.pxd":5
 * 
 * 
 * cdef class _Index:             # <<<<<<<<<<<<<<
 *     cdef:
 *         public size_t neutral_mass
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set__Index {
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set__Index *(*_create)(size_t, size_t);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set__Index *__pyx_vtabptr_12ms_deisotope_2_c_8peak_set__Index;


/* "ms_deisotope/_c/peak_set.pxd":14
 * 
 * 
 * cdef class EnvelopePair:             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mz
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_EnvelopePair {
  int (*_eq)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_EnvelopePair *, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_EnvelopePair *, int __pyx_skip_dispatch);
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_EnvelopePair *(*_create)(double, double);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_EnvelopePair *__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_EnvelopePair;


/* "ms_deisotope/_c/peak_set.pxd":25
 * 
 * 
 * cdef class Envelope:             # <<<<<<<<<<<<<<
 *     cdef public tuple pairs
 * 
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_Envelope {
  size_t (*get_size)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *);
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_EnvelopePair *(*getitem)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *, size_t);
  int (*_eq)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *, int __pyx_skip_dispatch);
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *(*clone)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *, int __pyx_skip_dispatch);
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *(*_create)(PyObject *);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_Envelope *__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_Envelope;


/* "ms_deisotope/_c/peak_set.pxd":40
 * 
 * 
 * cdef class DeconvolutedPeak(PeakBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double neutral_mass
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak {
  struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakBase __pyx_base;
  int (*_eq)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *, int __pyx_skip_dispatch);
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *(*_create_simple)(double, double, int, double, double, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak;


/* "ms_deisotope/_c/peak_set.pxd":66
 * 
 * 
 * cdef class DeconvolutedPeakSolution(DeconvolutedPeak):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public object solution
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution {
  struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak __pyx_base;
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution *__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution;


/* "ms_deisotope/_c/peak_set.pxd":70
 *         public object solution
 * 
 * cdef class DeconvolutedPeakSet(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public tuple peaks
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet {
  PyObject *(*reindex)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, int __pyx_skip_dispatch);
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *(*clone)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, int __pyx_skip_dispatch);
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *(*copy)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, int __pyx_skip_dispatch);
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *(*_has_peak)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, double, struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet__has_peak *__pyx_optional_args);
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *(*has_peak)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet_has_peak *__pyx_optional_args);
  PyObject *(*all_peaks_for)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet_all_peaks_for *__pyx_optional_args);
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *(*_get_nearest_peak)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, double, double *);
  size_t (*get_size)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *);
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *(*getitem)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, size_t);
  PyObject *(*getslice)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, size_t, size_t);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet;


/* "ms_deisotope/_c/peak_set.pxd":99
 * 
 * 
 * cdef class DeconvolutedPeakSetIndexed(DeconvolutedPeakSet):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double* neutral_mass_array
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed {
  struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet __pyx_base;
  void (*_build_index_arrays)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed *);
  int (*_interval_for)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed *, double, double, size_t *, size_t *);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed *__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed;


/* "ms_deisotope/_c/feature_map/feature_fit.pxd":14
 * 
 * 
 * cdef class LCMSFeatureSetFit(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list features
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit {
  int (*_eq)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit *, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit *, int __pyx_skip_dispatch);
  int (*_ne)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit *, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit *, int __pyx_skip_dispatch);
  int (*_lt)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit *, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit *, int __pyx_skip_dispatch);
  int (*_gt)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit *, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit *, int __pyx_skip_dispatch);
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit *(*_create)(PyObject *, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, double, int, size_t, PyObject *, PyObject *, double, size_t, PyArrayObject *, PyArrayObject *);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit *__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit;


/* "ms_deisotope/_c/feature_map/feature_processor.pyx":60
 * 
 * 
 * cdef class CartesianProductIterator(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list collections
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator {
  int (*has_more)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *, int __pyx_skip_dispatch);
  PyObject *(*compose_next_value)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *, int __pyx_skip_dispatch);
  PyObject *(*advance)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *, int __pyx_skip_dispatch);
  PyObject *(*get_next_value)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator;


/* "ms_deisotope/_c/feature_map/feature_processor.pyx":166
 * 
 * 
 * cdef class envelope_conformer:             # <<<<<<<<<<<<<<
 *     cdef:
 *         list experimental
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer {
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *(*_create)(void);
  void (*acquire)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *, PyObject *, PyObject *);
  void (*conform)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer_conform *__pyx_optional_args);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer;


/* "ms_deisotope/_c/feature_map/feature_processor.pyx":240
 * 
 * 
 * cdef class LCMSFeatureProcessorBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public LCMSFeatureMap feature_map
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase {
  PyObject *(*create_theoretical_distribution)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *, double, int, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_create_theoretical_distribution *__pyx_optional_args);
  PyObject *(*find_all_features)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_find_all_features *__pyx_optional_args);
  PyObject *(*find_features)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_find_features *__pyx_optional_args);
  PyObject *(*match_theoretical_isotopic_distribution)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_match_theoretical_isotopic_distribution *__pyx_optional_args);
  PyObject *(*conform_envelopes)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *, PyObject *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*_conform_envelopes)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *, PyObject *, PyObject *, size_t *);
  double (*_find_thresholded_score)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *, struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector *, double);
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit *(*_fit_single_feature_set)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *, PyObject *, PyObject *, double, int, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_single_feature_set *__pyx_optional_args);
  PyObject *(*_fit_theoretical_distribution_on_features)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *, double, double, int, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_theoretical_distribution_on_features *__pyx_optional_args);
  PyObject *(*_fit_feature_set)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *, double, double, int, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_feature_set *__pyx_optional_args);
  PyObject *(*_fit_composition)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *, PyObject *, double, int, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_composition *__pyx_optional_args);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* IsLittleEndian.proto */
static CYTHON_INLINE int __Pyx_Is_Little_Endian(void);

/* BufferFormatCheck.proto */
static const char* __Pyx_BufFmt_CheckString(__Pyx_BufFmt_Context* ctx, const char* ts);
static void __Pyx_BufFmt_Init(__Pyx_BufFmt_Context* ctx,
                              __Pyx_BufFmt_StackElem* stack,
                              __Pyx_TypeInfo* type);

/* BufferGetAndValidate.proto */
#define __Pyx_GetBufferAndValidate(buf, obj, dtype, flags, nd, cast, stack)\
    ((obj == Py_None || obj == NULL) ?\
    (__Pyx_ZeroBuffer(buf), 0) :\
    __Pyx__GetBufferAndValidate(buf, obj, dtype, flags, nd, cast, stack))
static int  __Pyx__GetBufferAndValidate(Py_buffer* buf, PyObject* obj,
    __Pyx_TypeInfo* dtype, int flags, int nd, int cast, __Pyx_BufFmt_StackElem* stack);
static void __Pyx_ZeroBuffer(Py_buffer* buf);
static CYTHON_INLINE void __Pyx_SafeReleaseBuffer(Py_buffer* info);
static Py_ssize_t __Pyx_minusones[] = { -1, -1, -1, -1, -1, -1, -1, -1 };
static Py_ssize_t __Pyx_zeros[] = { 0, 0, 0, 0, 0, 0, 0, 0 };

/* BufferFallbackError.proto */
static void __Pyx_RaiseBufferFallbackError(void);

/* BufferIndexError.proto */
static void __Pyx_RaiseBufferIndexError(int axis);

#define __Pyx_BufPtrCContig1d(type, buf, i0, s0) ((type)buf + i0)
/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        Py_SIZE(list) = len+1;
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        Py_SIZE(list) = len+1;
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
#define __Pyx_GetModuleGlobalNameUncached(var, name)  {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely((Py_TYPE(obj) == type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* SetItemInt.proto */
#define __Pyx_SetItemInt(o, i, v, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_SetItemInt_Fast(o, (Py_ssize_t)i, v, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list assignment index out of range"), -1) :\
               __Pyx_SetItemInt_Generic(o, to_py_func(i), v)))
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v);
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v,
                                               int is_list, int wraparound, int boundscheck);

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* HasAttr.proto */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* RaiseNoneIterError.proto */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyObject *dict);

/* pyobject_as_double.proto */
static double __Pyx__PyObject_AsDouble(PyObject* obj);
#if CYTHON_COMPILING_IN_PYPY
#define __Pyx_PyObject_AsDouble(obj)\
(likely(PyFloat_CheckExact(obj)) ? PyFloat_AS_DOUBLE(obj) :\
 likely(PyInt_CheckExact(obj)) ?\
 PyFloat_AsDouble(obj) : __Pyx__PyObject_AsDouble(obj))
#else
#define __Pyx_PyObject_AsDouble(obj)\
((likely(PyFloat_CheckExact(obj))) ?\
 PyFloat_AS_DOUBLE(obj) : __Pyx__PyObject_AsDouble(obj))
#endif

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* BufferStructDeclare.proto */
typedef struct {
  Py_ssize_t shape, strides, suboffsets;
} __Pyx_Buf_DimInfo;
typedef struct {
  size_t refcount;
  Py_buffer pybuffer;
} __Pyx_Buffer;
typedef struct {
  __Pyx_Buffer *rcbuffer;
  char *data;
  __Pyx_Buf_DimInfo diminfo[8];
} __Pyx_LocalBuf_ND;

#if PY_MAJOR_VERSION < 3
    static int __Pyx_GetBuffer(PyObject *obj, Py_buffer *view, int flags);
    static void __Pyx_ReleaseBuffer(Py_buffer *view);
#else
    #define __Pyx_GetBuffer PyObject_GetBuffer
    #define __Pyx_ReleaseBuffer PyBuffer_Release
#endif


/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_Py_intptr_t(Py_intptr_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* Print.proto */
static int __Pyx_Print(PyObject*, PyObject *, int);
#if CYTHON_COMPILING_IN_PYPY || PY_MAJOR_VERSION >= 3
static PyObject* __pyx_print = 0;
static PyObject* __pyx_print_kwargs = 0;
#endif

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* RealImag.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #define __Pyx_CREAL(z) ((z).real())
    #define __Pyx_CIMAG(z) ((z).imag())
  #else
    #define __Pyx_CREAL(z) (__real__(z))
    #define __Pyx_CIMAG(z) (__imag__(z))
  #endif
#else
    #define __Pyx_CREAL(z) ((z).real)
    #define __Pyx_CIMAG(z) ((z).imag)
#endif
#if defined(__cplusplus) && CYTHON_CCOMPLEX\
        && (defined(_WIN32) || defined(__clang__) || (defined(__GNUC__) && (__GNUC__ >= 5 || __GNUC__ == 4 && __GNUC_MINOR__ >= 4 )) || __cplusplus >= 201103)
    #define __Pyx_SET_CREAL(z,x) ((z).real(x))
    #define __Pyx_SET_CIMAG(z,y) ((z).imag(y))
#else
    #define __Pyx_SET_CREAL(z,x) __Pyx_CREAL(z) = (x)
    #define __Pyx_SET_CIMAG(z,y) __Pyx_CIMAG(z) = (y)
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_float(a, b)   ((a)==(b))
    #define __Pyx_c_sum_float(a, b)  ((a)+(b))
    #define __Pyx_c_diff_float(a, b) ((a)-(b))
    #define __Pyx_c_prod_float(a, b) ((a)*(b))
    #define __Pyx_c_quot_float(a, b) ((a)/(b))
    #define __Pyx_c_neg_float(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_float(z) ((z)==(float)0)
    #define __Pyx_c_conj_float(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (::std::abs(z))
        #define __Pyx_c_pow_float(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_float(z) ((z)==0)
    #define __Pyx_c_conj_float(z)    (conjf(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (cabsf(z))
        #define __Pyx_c_pow_float(a, b)  (cpowf(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex);
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex);
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex, __pyx_t_float_complex);
    #endif
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_double(a, b)   ((a)==(b))
    #define __Pyx_c_sum_double(a, b)  ((a)+(b))
    #define __Pyx_c_diff_double(a, b) ((a)-(b))
    #define __Pyx_c_prod_double(a, b) ((a)*(b))
    #define __Pyx_c_quot_double(a, b) ((a)/(b))
    #define __Pyx_c_neg_double(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_double(z) ((z)==(double)0)
    #define __Pyx_c_conj_double(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (::std::abs(z))
        #define __Pyx_c_pow_double(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_double(z) ((z)==0)
    #define __Pyx_c_conj_double(z)    (conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (cabs(z))
        #define __Pyx_c_pow_double(a, b)  (cpow(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex);
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex);
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex, __pyx_t_double_complex);
    #endif
#endif

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__NPY_TYPES(enum NPY_TYPES value);

/* CIntFromPy.proto */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* PrintOne.proto */
static int __Pyx_PrintOne(PyObject* stream, PyObject *o);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* VoidPtrImport.proto */
static int __Pyx_ImportVoidPtr(PyObject *module, const char *name, void **p, const char *sig);

/* FunctionImport.proto */
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static int __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_has_more(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_compose_next_value(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_advance(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_get_next_value(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer__create(void); /* proto*/
static void __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer_acquire(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *__pyx_v_self, PyObject *__pyx_v_experimental, PyObject *__pyx_v_theoretical); /* proto*/
static void __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer_conform(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *__pyx_v_self, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer_conform *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_create_theoretical_distribution(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, double __pyx_v_mz, int __pyx_v_charge, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_create_theoretical_distribution *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_find_all_features(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, double __pyx_v_mz, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_find_all_features *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_find_features(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, double __pyx_v_mz, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_find_features *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_match_theoretical_isotopic_distribution(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, PyObject *__pyx_v_theoretical_distribution, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_match_theoretical_isotopic_distribution *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_conform_envelopes(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, PyObject *__pyx_v_experimental, PyObject *__pyx_v_base_theoretical, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__conform_envelopes(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, PyObject *__pyx_v_experimental, PyObject *__pyx_v_base_theoretical, size_t *__pyx_v_n_missing); /* proto*/
static double __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__find_thresholded_score(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector *__pyx_v_scores, double __pyx_v_percentage); /* proto*/
static struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_single_feature_set(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, PyObject *__pyx_v_features, PyObject *__pyx_v_base_tid, CYTHON_UNUSED double __pyx_v_error_tolerance, int __pyx_v_charge, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_single_feature_set *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_theoretical_distribution_on_features(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, double __pyx_v_mz, double __pyx_v_error_tolerance, int __pyx_v_charge, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *__pyx_v_base_tid, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_theoretical_distribution_on_features *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_feature_set(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, double __pyx_v_mz, double __pyx_v_error_tolerance, int __pyx_v_charge, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_feature_set *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_composition(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, PyObject *__pyx_v_composition, double __pyx_v_error_tolerance, int __pyx_v_charge, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_composition *__pyx_optional_args); /* proto*/

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdlib' */

/* Module declarations from 'cython' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'cpython.buffer' */

/* Module declarations from 'cpython' */

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cpython.list' */

/* Module declarations from 'ms_peak_picker._c.peak_set' */
static PyTypeObject *__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase = 0;
static PyTypeObject *__pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak = 0;
static PyTypeObject *__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet = 0;
static PyTypeObject *__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed = 0;
static size_t *__pyx_vp_14ms_peak_picker_2_c_8peak_set_INTERVAL_INDEX_SIZE = 0;
#define __pyx_v_14ms_peak_picker_2_c_8peak_set_INTERVAL_INDEX_SIZE (*__pyx_vp_14ms_peak_picker_2_c_8peak_set_INTERVAL_INDEX_SIZE)

/* Module declarations from 'brainpy._c.composition' */
static PyTypeObject *__pyx_ptype_7brainpy_2_c_11composition_PyComposition = 0;
static PyObject **__pyx_vp_7brainpy_2_c_11composition_nist_mass = 0;
#define __pyx_v_7brainpy_2_c_11composition_nist_mass (*__pyx_vp_7brainpy_2_c_11composition_nist_mass)
static double *__pyx_vp_7brainpy_2_c_11composition_PROTON = 0;
#define __pyx_v_7brainpy_2_c_11composition_PROTON (*__pyx_vp_7brainpy_2_c_11composition_PROTON)
static struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable **__pyx_vp_7brainpy_2_c_11composition__ElementTable = 0;
#define __pyx_v_7brainpy_2_c_11composition__ElementTable (*__pyx_vp_7brainpy_2_c_11composition__ElementTable)

/* Module declarations from 'brainpy._c.double_vector' */
static struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector *(*__pyx_f_7brainpy_2_c_13double_vector_make_double_vector)(void); /*proto*/
static int (*__pyx_f_7brainpy_2_c_13double_vector_double_vector_append)(struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector *, double); /*proto*/
static void (*__pyx_f_7brainpy_2_c_13double_vector_free_double_vector)(struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector *); /*proto*/
static PyObject *(*__pyx_f_7brainpy_2_c_13double_vector_double_vector_to_list)(struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector *); /*proto*/

/* Module declarations from 'brainpy._c.isotopic_constants' */
static size_t *__pyx_vp_7brainpy_2_c_18isotopic_constants_DEFAULT_ISOTOPIC_CONSTANTS_SIZE = 0;
#define __pyx_v_7brainpy_2_c_18isotopic_constants_DEFAULT_ISOTOPIC_CONSTANTS_SIZE (*__pyx_vp_7brainpy_2_c_18isotopic_constants_DEFAULT_ISOTOPIC_CONSTANTS_SIZE)

/* Module declarations from 'brainpy._c.isotopic_distribution' */
static PyTypeObject *__pyx_ptype_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak = 0;
static PyObject *(*__pyx_f_7brainpy_2_c_21isotopic_distribution__isotopic_variants)(PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution__isotopic_variants *__pyx_optional_args); /*proto*/

/* Module declarations from 'cpython.ref' */

/* Module declarations from 'cpython.mem' */

/* Module declarations from 'numpy' */

/* Module declarations from 'numpy' */
static PyTypeObject *__pyx_ptype_5numpy_dtype = 0;
static PyTypeObject *__pyx_ptype_5numpy_flatiter = 0;
static PyTypeObject *__pyx_ptype_5numpy_broadcast = 0;
static PyTypeObject *__pyx_ptype_5numpy_ndarray = 0;
static PyTypeObject *__pyx_ptype_5numpy_ufunc = 0;
static CYTHON_INLINE char *__pyx_f_5numpy__util_dtypestring(PyArray_Descr *, char *, char *, int *); /*proto*/
static CYTHON_INLINE int __pyx_f_5numpy_import_array(void); /*proto*/

/* Module declarations from 'ms_peak_picker._c.peak_index' */
static PyTypeObject *__pyx_ptype_14ms_peak_picker_2_c_10peak_index_PeakIndex = 0;

/* Module declarations from 'ms_deisotope._c.averagine' */
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_9averagine_Averagine = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_9averagine_AveragineCache = 0;
static double *__pyx_vp_12ms_deisotope_2_c_9averagine_PROTON = 0;
#define __pyx_v_12ms_deisotope_2_c_9averagine_PROTON (*__pyx_vp_12ms_deisotope_2_c_9averagine_PROTON)
static double (*__pyx_f_12ms_deisotope_2_c_9averagine_neutral_mass)(double, int, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_neutral_mass *__pyx_optional_args); /*proto*/

/* Module declarations from 'ms_deisotope._c.deconvoluter_base' */
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator = 0;

/* Module declarations from 'ms_deisotope._c.scoring' */
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitRecord = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_FitSelectorBase = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitterBase = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_LeastSquaresFitter = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_GTestFitter = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledGTestFitter = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_MSDeconVFitter = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_FunctionScorer = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_InterferenceDetection = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_DistinctPatternFitter = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_DotProductFitter = 0;
static double *__pyx_vp_12ms_deisotope_2_c_7scoring_INFINITY = 0;
#define __pyx_v_12ms_deisotope_2_c_7scoring_INFINITY (*__pyx_vp_12ms_deisotope_2_c_7scoring_INFINITY)

/* Module declarations from 'ms_deisotope._c.feature_map.lcms_feature' */
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass = 0;

/* Module declarations from 'ms_deisotope._c.feature_map.feature_map' */
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap = 0;

/* Module declarations from 'ms_deisotope._c.peak_set' */
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_8peak_set__Index = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_8peak_set_EnvelopePair = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_8peak_set_Envelope = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed = 0;
static size_t *__pyx_vp_12ms_deisotope_2_c_8peak_set_INTERVAL_INDEX_SIZE = 0;
#define __pyx_v_12ms_deisotope_2_c_8peak_set_INTERVAL_INDEX_SIZE (*__pyx_vp_12ms_deisotope_2_c_8peak_set_INTERVAL_INDEX_SIZE)

/* Module declarations from 'ms_deisotope._c.feature_map.feature_fit' */
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_11feature_map_11feature_fit_map_coord = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit = 0;

/* Module declarations from 'ms_deisotope._c.feature_map.feature_processor' */
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase = 0;
static PyObject *__pyx_v_12ms_deisotope_2_c_11feature_map_17feature_processor_zeros = 0;
static double __pyx_v_12ms_deisotope_2_c_11feature_map_17feature_processor_INF;
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_double_vector_to_ndarray(struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector *); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor__conform_envelopes(PyObject *, PyObject *, size_t *, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor__conform_envelopes *__pyx_optional_args); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_conform_envelopes(PyObject *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_conform_envelopes *__pyx_optional_args); /*proto*/
static int __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_has_no_valid_features(PyObject *); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor___pyx_unpickle_envelope_conformer__set_state(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *, PyObject *); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor___pyx_unpickle_LCMSFeatureProcessorBase__set_state(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *, PyObject *); /*proto*/
static __Pyx_TypeInfo __Pyx_TypeInfo_double = { "double", NULL, sizeof(double), { 0 }, 0, 'R', 0, 0 };
#define __Pyx_MODULE_NAME "ms_deisotope._c.feature_map.feature_processor"
extern int __pyx_module_is_main_ms_deisotope___c__feature_map__feature_processor;
int __pyx_module_is_main_ms_deisotope___c__feature_map__feature_processor = 0;

/* Implementation of 'ms_deisotope._c.feature_map.feature_processor' */
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_map;
static PyObject *__pyx_builtin_StopIteration;
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_RuntimeError;
static PyObject *__pyx_builtin_ImportError;
static const char __pyx_k__6[] = "-";
static const char __pyx_k_mz[] = "mz";
static const char __pyx_k_np[] = "np";
static const char __pyx_k_end[] = "end";
static const char __pyx_k_inf[] = "inf";
static const char __pyx_k_map[] = "map";
static const char __pyx_k_new[] = "__new__";
static const char __pyx_k_dict[] = "__dict__";
static const char __pyx_k_file[] = "file";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_numpy[] = "numpy";
static const char __pyx_k_print[] = "print";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_zeros[] = "zeros";
static const char __pyx_k_charge[] = "charge";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_pickle[] = "pickle";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_update[] = "update";
static const char __pyx_k_advance[] = "advance";
static const char __pyx_k_feature[] = "feature";
static const char __pyx_k_base_tid[] = "base_tid";
static const char __pyx_k_end_time[] = "end_time";
static const char __pyx_k_features[] = "features";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_has_more[] = "has_more";
static const char __pyx_k_interval[] = "interval";
static const char __pyx_k_pyx_type[] = "__pyx_type";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_0_2f_0_2f[] = "%0.2f-%0.2f";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_pyx_state[] = "__pyx_state";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_pyx_result[] = "__pyx_result";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_start_time[] = "start_time";
static const char __pyx_k_ImportError[] = "ImportError";
static const char __pyx_k_PickleError[] = "PickleError";
static const char __pyx_k_collections[] = "collections";
static const char __pyx_k_composition[] = "composition";
static const char __pyx_k_left_search[] = "left_search";
static const char __pyx_k_RuntimeError[] = "RuntimeError";
static const char __pyx_k_experimental[] = "experimental";
static const char __pyx_k_ignore_below[] = "ignore_below";
static const char __pyx_k_pyx_checksum[] = "__pyx_checksum";
static const char __pyx_k_right_search[] = "right_search";
static const char __pyx_k_stringsource[] = "stringsource";
static const char __pyx_k_StopIteration[] = "StopIteration";
static const char __pyx_k_find_features[] = "find_features";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_truncated_tid[] = "truncated_tid";
static const char __pyx_k_charge_carrier[] = "charge_carrier";
static const char __pyx_k_get_next_value[] = "get_next_value";
static const char __pyx_k_truncate_after[] = "truncate_after";
static const char __pyx_k_error_tolerance[] = "error_tolerance";
static const char __pyx_k_fit_composition[] = "_fit_composition";
static const char __pyx_k_fit_feature_set[] = "_fit_feature_set";
static const char __pyx_k_pyx_PickleError[] = "__pyx_PickleError";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_threshold_scale[] = "threshold_scale";
static const char __pyx_k_base_theoretical[] = "base_theoretical";
static const char __pyx_k_max_missed_peaks[] = "max_missed_peaks";
static const char __pyx_k_conform_envelopes[] = "conform_envelopes";
static const char __pyx_k_find_all_features[] = "find_all_features";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_compose_next_value[] = "compose_next_value";
static const char __pyx_k_envelope_conformer[] = "envelope_conformer";
static const char __pyx_k_fit_single_feature_set[] = "_fit_single_feature_set";
static const char __pyx_k_CartesianProductIterator[] = "CartesianProductIterator";
static const char __pyx_k_LCMSFeatureProcessorBase[] = "LCMSFeatureProcessorBase";
static const char __pyx_k_theoretical_distribution[] = "theoretical_distribution";
static const char __pyx_k_ndarray_is_not_C_contiguous[] = "ndarray is not C contiguous";
static const char __pyx_k_minimum_theoretical_abundance[] = "minimum_theoretical_abundance";
static const char __pyx_k_pyx_unpickle_LCMSFeatureProces[] = "__pyx_unpickle_LCMSFeatureProcessorBase";
static const char __pyx_k_pyx_unpickle_envelope_conforme[] = "__pyx_unpickle_envelope_conformer";
static const char __pyx_k_create_theoretical_distribution[] = "create_theoretical_distribution";
static const char __pyx_k_d_combinations_have_been_consid[] = "\t%d combinations have been considered for m/z %0.3f at charge %d in %s (%0.2f%%)";
static const char __pyx_k_numpy_core_multiarray_failed_to[] = "numpy.core.multiarray failed to import";
static const char __pyx_k_unknown_dtype_code_in_numpy_pxd[] = "unknown dtype code in numpy.pxd (%d)";
static const char __pyx_k_Format_string_allocated_too_shor[] = "Format string allocated too short, see comment in numpy.pxd";
static const char __pyx_k_Incompatible_checksums_s_vs_0xa9[] = "Incompatible checksums (%s vs 0xa9723cd = (averagine, feature_map, scorer))";
static const char __pyx_k_Incompatible_checksums_s_vs_0xba[] = "Incompatible checksums (%s vs 0xba9b7dc = (experimental, n_missing, theoretical))";
static const char __pyx_k_Non_native_byte_order_not_suppor[] = "Non-native byte order not supported";
static const char __pyx_k_match_theoretical_isotopic_distr[] = "match_theoretical_isotopic_distribution";
static const char __pyx_k_ms_deisotope__c_feature_map_feat[] = "ms_deisotope._c.feature_map.feature_processor";
static const char __pyx_k_ndarray_is_not_Fortran_contiguou[] = "ndarray is not Fortran contiguous";
static const char __pyx_k_no_default___reduce___due_to_non[] = "no default __reduce__ due to non-trivial __cinit__";
static const char __pyx_k_numpy_core_umath_failed_to_impor[] = "numpy.core.umath failed to import";
static const char __pyx_k_Format_string_allocated_too_shor_2[] = "Format string allocated too short.";
static PyObject *__pyx_kp_s_0_2f_0_2f;
static PyObject *__pyx_n_s_CartesianProductIterator;
static PyObject *__pyx_kp_u_Format_string_allocated_too_shor;
static PyObject *__pyx_kp_u_Format_string_allocated_too_shor_2;
static PyObject *__pyx_n_s_ImportError;
static PyObject *__pyx_kp_s_Incompatible_checksums_s_vs_0xa9;
static PyObject *__pyx_kp_s_Incompatible_checksums_s_vs_0xba;
static PyObject *__pyx_n_s_LCMSFeatureProcessorBase;
static PyObject *__pyx_kp_u_Non_native_byte_order_not_suppor;
static PyObject *__pyx_n_s_PickleError;
static PyObject *__pyx_n_s_RuntimeError;
static PyObject *__pyx_n_s_StopIteration;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_kp_s__6;
static PyObject *__pyx_n_s_advance;
static PyObject *__pyx_n_s_base_theoretical;
static PyObject *__pyx_n_s_base_tid;
static PyObject *__pyx_n_s_charge;
static PyObject *__pyx_n_s_charge_carrier;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_collections;
static PyObject *__pyx_n_s_compose_next_value;
static PyObject *__pyx_n_s_composition;
static PyObject *__pyx_n_s_conform_envelopes;
static PyObject *__pyx_n_s_create_theoretical_distribution;
static PyObject *__pyx_kp_s_d_combinations_have_been_consid;
static PyObject *__pyx_n_s_dict;
static PyObject *__pyx_n_s_end;
static PyObject *__pyx_n_s_end_time;
static PyObject *__pyx_n_s_envelope_conformer;
static PyObject *__pyx_n_s_error_tolerance;
static PyObject *__pyx_n_s_experimental;
static PyObject *__pyx_n_s_feature;
static PyObject *__pyx_n_s_features;
static PyObject *__pyx_n_s_file;
static PyObject *__pyx_n_s_find_all_features;
static PyObject *__pyx_n_s_find_features;
static PyObject *__pyx_n_s_fit_composition;
static PyObject *__pyx_n_s_fit_feature_set;
static PyObject *__pyx_n_s_fit_single_feature_set;
static PyObject *__pyx_n_s_get_next_value;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_n_s_has_more;
static PyObject *__pyx_n_s_ignore_below;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_inf;
static PyObject *__pyx_n_s_interval;
static PyObject *__pyx_n_s_left_search;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_map;
static PyObject *__pyx_n_s_match_theoretical_isotopic_distr;
static PyObject *__pyx_n_s_max_missed_peaks;
static PyObject *__pyx_n_s_minimum_theoretical_abundance;
static PyObject *__pyx_n_s_ms_deisotope__c_feature_map_feat;
static PyObject *__pyx_n_s_mz;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_kp_u_ndarray_is_not_C_contiguous;
static PyObject *__pyx_kp_u_ndarray_is_not_Fortran_contiguou;
static PyObject *__pyx_n_s_new;
static PyObject *__pyx_kp_s_no_default___reduce___due_to_non;
static PyObject *__pyx_n_s_np;
static PyObject *__pyx_n_s_numpy;
static PyObject *__pyx_kp_s_numpy_core_multiarray_failed_to;
static PyObject *__pyx_kp_s_numpy_core_umath_failed_to_impor;
static PyObject *__pyx_n_s_pickle;
static PyObject *__pyx_n_s_print;
static PyObject *__pyx_n_s_pyx_PickleError;
static PyObject *__pyx_n_s_pyx_checksum;
static PyObject *__pyx_n_s_pyx_result;
static PyObject *__pyx_n_s_pyx_state;
static PyObject *__pyx_n_s_pyx_type;
static PyObject *__pyx_n_s_pyx_unpickle_LCMSFeatureProces;
static PyObject *__pyx_n_s_pyx_unpickle_envelope_conforme;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_reduce_cython;
static PyObject *__pyx_n_s_reduce_ex;
static PyObject *__pyx_n_s_right_search;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_setstate_cython;
static PyObject *__pyx_n_s_start_time;
static PyObject *__pyx_kp_s_stringsource;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_theoretical_distribution;
static PyObject *__pyx_n_s_threshold_scale;
static PyObject *__pyx_n_s_truncate_after;
static PyObject *__pyx_n_s_truncated_tid;
static PyObject *__pyx_kp_u_unknown_dtype_code_in_numpy_pxd;
static PyObject *__pyx_n_s_update;
static PyObject *__pyx_n_s_zeros;
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator___cinit__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self, PyObject *__pyx_v_collections); /* proto */
static void __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_2__dealloc__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_4get_lengths(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_6get_indices(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_8has_more(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_10compose_next_value(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_12advance(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_14get_next_value(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_16__iter__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_18__next__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_11collections___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_11collections_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_11collections_4__del__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_4size___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_4size_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_4done___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_4done_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_18total_combinations___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_18total_combinations_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_20__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_22__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer___reduce_cython__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer_2__setstate_cython__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_conform_envelopes(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_experimental, PyObject *__pyx_v_base_theoretical, double __pyx_v_minimum_theoretical_abundance); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_create_theoretical_distribution(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, double __pyx_v_mz, int __pyx_v_charge, double __pyx_v_charge_carrier, double __pyx_v_truncate_after, double __pyx_v_ignore_below); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_2find_all_features(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, double __pyx_v_mz, double __pyx_v_error_tolerance); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_4find_features(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, double __pyx_v_mz, double __pyx_v_error_tolerance, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_interval); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_6match_theoretical_isotopic_distribution(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, PyObject *__pyx_v_theoretical_distribution, double __pyx_v_error_tolerance, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_interval); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_8conform_envelopes(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, PyObject *__pyx_v_experimental, PyObject *__pyx_v_base_theoretical); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_10_fit_single_feature_set(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, PyObject *__pyx_v_features, PyObject *__pyx_v_base_tid, double __pyx_v_error_tolerance, int __pyx_v_charge, double __pyx_v_charge_carrier, int __pyx_v_max_missed_peaks, double __pyx_v_threshold_scale); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_12_fit_feature_set(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, double __pyx_v_mz, double __pyx_v_error_tolerance, int __pyx_v_charge, int __pyx_v_left_search, int __pyx_v_right_search, double __pyx_v_charge_carrier, double __pyx_v_truncate_after, int __pyx_v_max_missed_peaks, double __pyx_v_threshold_scale, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_feature); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_14_fit_composition(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, PyObject *__pyx_v_composition, double __pyx_v_error_tolerance, int __pyx_v_charge, double __pyx_v_charge_carrier, double __pyx_v_truncate_after, double __pyx_v_ignore_below, int __pyx_v_max_missed_peaks, double __pyx_v_threshold_scale, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_feature); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_11feature_map___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_11feature_map_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_11feature_map_4__del__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_6scorer___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_6scorer_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_6scorer_4__del__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_9averagine___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_9averagine_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_9averagine_4__del__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_16__reduce_cython__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_18__setstate_cython__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_2__pyx_unpickle_envelope_conformer(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_4__pyx_unpickle_LCMSFeatureProcessorBase(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_5numpy_7ndarray___getbuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /* proto */
static void __pyx_pf_5numpy_7ndarray_2__releasebuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info); /* proto */
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_int_177677261;
static PyObject *__pyx_int_195672028;
static double __pyx_k__3;
static double __pyx_k__4;
static double __pyx_k__5;
static double __pyx_k__7;
static double __pyx_k__8;
static PyObject *__pyx_tuple_;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__9;
static PyObject *__pyx_tuple__10;
static PyObject *__pyx_tuple__11;
static PyObject *__pyx_tuple__12;
static PyObject *__pyx_tuple__13;
static PyObject *__pyx_tuple__14;
static PyObject *__pyx_tuple__15;
static PyObject *__pyx_tuple__16;
static PyObject *__pyx_tuple__18;
static PyObject *__pyx_codeobj__17;
static PyObject *__pyx_codeobj__19;
/* Late includes */

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":46
 * 
 * 
 * cdef object double_vector_to_ndarray(dvec* vec):             # <<<<<<<<<<<<<<
 *     cdef:
 *         np.ndarray[double, ndim=1, mode='c'] out
 */

static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_double_vector_to_ndarray(struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector *__pyx_v_vec) {
  PyArrayObject *__pyx_v_out = 0;
  npy_intp __pyx_v_shape[1];
  size_t __pyx_v_i;
  size_t __pyx_v_n;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_out;
  __Pyx_Buffer __pyx_pybuffer_out;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyArrayObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  size_t __pyx_t_11;
  size_t __pyx_t_12;
  size_t __pyx_t_13;
  __Pyx_RefNannySetupContext("double_vector_to_ndarray", 0);
  __pyx_pybuffer_out.pybuffer.buf = NULL;
  __pyx_pybuffer_out.refcount = 0;
  __pyx_pybuffernd_out.data = NULL;
  __pyx_pybuffernd_out.rcbuffer = &__pyx_pybuffer_out;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":52
 *         size_t i, n
 * 
 *     n = shape[0] = vec.used             # <<<<<<<<<<<<<<
 *     # out = np.PyArray_SimpleNew(1, shape, np.NPY_DOUBLE)
 *     out = zeros(shape[0])
 */
  __pyx_t_1 = __pyx_v_vec->used;
  __pyx_v_n = __pyx_t_1;
  (__pyx_v_shape[0]) = __pyx_t_1;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":54
 *     n = shape[0] = vec.used
 *     # out = np.PyArray_SimpleNew(1, shape, np.NPY_DOUBLE)
 *     out = zeros(shape[0])             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         out[i] = vec.v[i]
 */
  __pyx_t_3 = __Pyx_PyInt_From_Py_intptr_t((__pyx_v_shape[0])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_12ms_deisotope_2_c_11feature_map_17feature_processor_zeros);
  __pyx_t_4 = __pyx_v_12ms_deisotope_2_c_11feature_map_17feature_processor_zeros; __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 54, __pyx_L1_error)
  __pyx_t_6 = ((PyArrayObject *)__pyx_t_2);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_out.rcbuffer->pybuffer);
    __pyx_t_7 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_out.rcbuffer->pybuffer, (PyObject*)__pyx_t_6, &__Pyx_TypeInfo_double, PyBUF_FORMAT| PyBUF_C_CONTIGUOUS| PyBUF_WRITABLE, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_7 < 0)) {
      PyErr_Fetch(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_out.rcbuffer->pybuffer, (PyObject*)__pyx_v_out, &__Pyx_TypeInfo_double, PyBUF_FORMAT| PyBUF_C_CONTIGUOUS| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_8); Py_XDECREF(__pyx_t_9); Py_XDECREF(__pyx_t_10);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      }
      __pyx_t_8 = __pyx_t_9 = __pyx_t_10 = 0;
    }
    __pyx_pybuffernd_out.diminfo[0].strides = __pyx_pybuffernd_out.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_out.diminfo[0].shape = __pyx_pybuffernd_out.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 54, __pyx_L1_error)
  }
  __pyx_t_6 = 0;
  __pyx_v_out = ((PyArrayObject *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":55
 *     # out = np.PyArray_SimpleNew(1, shape, np.NPY_DOUBLE)
 *     out = zeros(shape[0])
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         out[i] = vec.v[i]
 *     return out
 */
  __pyx_t_1 = __pyx_v_n;
  __pyx_t_11 = __pyx_t_1;
  for (__pyx_t_12 = 0; __pyx_t_12 < __pyx_t_11; __pyx_t_12+=1) {
    __pyx_v_i = __pyx_t_12;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":56
 *     out = zeros(shape[0])
 *     for i in range(n):
 *         out[i] = vec.v[i]             # <<<<<<<<<<<<<<
 *     return out
 * 
 */
    __pyx_t_13 = __pyx_v_i;
    __pyx_t_7 = -1;
    if (unlikely(__pyx_t_13 >= (size_t)__pyx_pybuffernd_out.diminfo[0].shape)) __pyx_t_7 = 0;
    if (unlikely(__pyx_t_7 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_7);
      __PYX_ERR(0, 56, __pyx_L1_error)
    }
    *__Pyx_BufPtrCContig1d(double *, __pyx_pybuffernd_out.rcbuffer->pybuffer.buf, __pyx_t_13, __pyx_pybuffernd_out.diminfo[0].strides) = (__pyx_v_vec->v[__pyx_v_i]);
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":57
 *     for i in range(n):
 *         out[i] = vec.v[i]
 *     return out             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_out));
  __pyx_r = ((PyObject *)__pyx_v_out);
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":46
 * 
 * 
 * cdef object double_vector_to_ndarray(dvec* vec):             # <<<<<<<<<<<<<<
 *     cdef:
 *         np.ndarray[double, ndim=1, mode='c'] out
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_out.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.double_vector_to_ndarray", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_out.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_out);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":69
 *         public size_t total_combinations
 * 
 *     def __cinit__(self, collections):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, n
 */

/* Python wrapper */
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_collections = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_collections,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_collections)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(0, 69, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_collections = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 69, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.CartesianProductIterator.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator___cinit__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *)__pyx_v_self), __pyx_v_collections);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator___cinit__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self, PyObject *__pyx_v_collections) {
  size_t __pyx_v_i;
  PyObject *__pyx_v_sublist = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  size_t __pyx_t_5;
  PyObject *__pyx_t_6;
  int __pyx_t_7;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":74
 *             list sublist
 * 
 *         self.collections = list(map(list, collections))             # <<<<<<<<<<<<<<
 *         self.size = PyList_GET_SIZE(self.collections)
 *         self.lengths = <int*>malloc(sizeof(int) * self.size)
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 74, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(((PyObject *)(&PyList_Type)));
  __Pyx_GIVEREF(((PyObject *)(&PyList_Type)));
  PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)(&PyList_Type)));
  __Pyx_INCREF(__pyx_v_collections);
  __Pyx_GIVEREF(__pyx_v_collections);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_collections);
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_map, __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 74, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PySequence_List(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 74, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->collections);
  __Pyx_DECREF(__pyx_v_self->collections);
  __pyx_v_self->collections = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":75
 * 
 *         self.collections = list(map(list, collections))
 *         self.size = PyList_GET_SIZE(self.collections)             # <<<<<<<<<<<<<<
 *         self.lengths = <int*>malloc(sizeof(int) * self.size)
 *         self.indices = <int*>malloc(sizeof(int) * self.size)
 */
  __pyx_t_1 = __pyx_v_self->collections;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_self->size = PyList_GET_SIZE(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":76
 *         self.collections = list(map(list, collections))
 *         self.size = PyList_GET_SIZE(self.collections)
 *         self.lengths = <int*>malloc(sizeof(int) * self.size)             # <<<<<<<<<<<<<<
 *         self.indices = <int*>malloc(sizeof(int) * self.size)
 *         self.done = False
 */
  __pyx_v_self->lengths = ((int *)malloc(((sizeof(int)) * __pyx_v_self->size)));

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":77
 *         self.size = PyList_GET_SIZE(self.collections)
 *         self.lengths = <int*>malloc(sizeof(int) * self.size)
 *         self.indices = <int*>malloc(sizeof(int) * self.size)             # <<<<<<<<<<<<<<
 *         self.done = False
 *         self.total_combinations = 1
 */
  __pyx_v_self->indices = ((int *)malloc(((sizeof(int)) * __pyx_v_self->size)));

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":78
 *         self.lengths = <int*>malloc(sizeof(int) * self.size)
 *         self.indices = <int*>malloc(sizeof(int) * self.size)
 *         self.done = False             # <<<<<<<<<<<<<<
 *         self.total_combinations = 1
 * 
 */
  __pyx_v_self->done = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":79
 *         self.indices = <int*>malloc(sizeof(int) * self.size)
 *         self.done = False
 *         self.total_combinations = 1             # <<<<<<<<<<<<<<
 * 
 *         for i in range(self.size):
 */
  __pyx_v_self->total_combinations = 1;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":81
 *         self.total_combinations = 1
 * 
 *         for i in range(self.size):             # <<<<<<<<<<<<<<
 *             sublist = <list>PyList_GetItem(self.collections, i)
 *             self.lengths[i] = PyList_GET_SIZE(sublist)
 */
  __pyx_t_3 = __pyx_v_self->size;
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
    __pyx_v_i = __pyx_t_5;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":82
 * 
 *         for i in range(self.size):
 *             sublist = <list>PyList_GetItem(self.collections, i)             # <<<<<<<<<<<<<<
 *             self.lengths[i] = PyList_GET_SIZE(sublist)
 *             self.total_combinations *= self.lengths[i]
 */
    __pyx_t_1 = __pyx_v_self->collections;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_6 = PyList_GetItem(__pyx_t_1, __pyx_v_i); if (unlikely(__pyx_t_6 == ((PyObject *)NULL))) __PYX_ERR(0, 82, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = ((PyObject *)__pyx_t_6);
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_sublist, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":83
 *         for i in range(self.size):
 *             sublist = <list>PyList_GetItem(self.collections, i)
 *             self.lengths[i] = PyList_GET_SIZE(sublist)             # <<<<<<<<<<<<<<
 *             self.total_combinations *= self.lengths[i]
 *             if self.lengths[i] == 0:
 */
    (__pyx_v_self->lengths[__pyx_v_i]) = PyList_GET_SIZE(__pyx_v_sublist);

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":84
 *             sublist = <list>PyList_GetItem(self.collections, i)
 *             self.lengths[i] = PyList_GET_SIZE(sublist)
 *             self.total_combinations *= self.lengths[i]             # <<<<<<<<<<<<<<
 *             if self.lengths[i] == 0:
 *                 self.done = True
 */
    __pyx_v_self->total_combinations = (__pyx_v_self->total_combinations * (__pyx_v_self->lengths[__pyx_v_i]));

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":85
 *             self.lengths[i] = PyList_GET_SIZE(sublist)
 *             self.total_combinations *= self.lengths[i]
 *             if self.lengths[i] == 0:             # <<<<<<<<<<<<<<
 *                 self.done = True
 *             self.indices[i] = 0
 */
    __pyx_t_7 = (((__pyx_v_self->lengths[__pyx_v_i]) == 0) != 0);
    if (__pyx_t_7) {

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":86
 *             self.total_combinations *= self.lengths[i]
 *             if self.lengths[i] == 0:
 *                 self.done = True             # <<<<<<<<<<<<<<
 *             self.indices[i] = 0
 * 
 */
      __pyx_v_self->done = 1;

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":85
 *             self.lengths[i] = PyList_GET_SIZE(sublist)
 *             self.total_combinations *= self.lengths[i]
 *             if self.lengths[i] == 0:             # <<<<<<<<<<<<<<
 *                 self.done = True
 *             self.indices[i] = 0
 */
    }

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":87
 *             if self.lengths[i] == 0:
 *                 self.done = True
 *             self.indices[i] = 0             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
    (__pyx_v_self->indices[__pyx_v_i]) = 0;
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":69
 *         public size_t total_combinations
 * 
 *     def __cinit__(self, collections):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, n
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.CartesianProductIterator.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_sublist);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":89
 *             self.indices[i] = 0
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         free(self.lengths)
 *         free(self.indices)
 */

/* Python wrapper */
static void __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_2__dealloc__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_2__dealloc__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":90
 * 
 *     def __dealloc__(self):
 *         free(self.lengths)             # <<<<<<<<<<<<<<
 *         free(self.indices)
 * 
 */
  free(__pyx_v_self->lengths);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":91
 *     def __dealloc__(self):
 *         free(self.lengths)
 *         free(self.indices)             # <<<<<<<<<<<<<<
 * 
 *     def get_lengths(self):
 */
  free(__pyx_v_self->indices);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":89
 *             self.indices[i] = 0
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         free(self.lengths)
 *         free(self.indices)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":93
 *         free(self.indices)
 * 
 *     def get_lengths(self):             # <<<<<<<<<<<<<<
 *         return [self.lengths[i] for i in range(self.size)]
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_5get_lengths(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_4get_lengths[] = "CartesianProductIterator.get_lengths(self)";
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_5get_lengths(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_lengths (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_4get_lengths(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_4get_lengths(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self) {
  size_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("get_lengths", 0);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":94
 * 
 *     def get_lengths(self):
 *         return [self.lengths[i] for i in range(self.size)]             # <<<<<<<<<<<<<<
 * 
 *     def get_indices(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 94, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_v_self->size;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;
    __pyx_t_5 = __Pyx_PyInt_From_int((__pyx_v_self->lengths[__pyx_v_i])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 94, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_5))) __PYX_ERR(0, 94, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":93
 *         free(self.indices)
 * 
 *     def get_lengths(self):             # <<<<<<<<<<<<<<
 *         return [self.lengths[i] for i in range(self.size)]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.CartesianProductIterator.get_lengths", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":96
 *         return [self.lengths[i] for i in range(self.size)]
 * 
 *     def get_indices(self):             # <<<<<<<<<<<<<<
 *         return [self.indices[i] for i in range(self.size)]
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_7get_indices(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_6get_indices[] = "CartesianProductIterator.get_indices(self)";
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_7get_indices(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_indices (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_6get_indices(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_6get_indices(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self) {
  size_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("get_indices", 0);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":97
 * 
 *     def get_indices(self):
 *         return [self.indices[i] for i in range(self.size)]             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint has_more(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 97, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_v_self->size;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;
    __pyx_t_5 = __Pyx_PyInt_From_int((__pyx_v_self->indices[__pyx_v_i])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 97, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_5))) __PYX_ERR(0, 97, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":96
 *         return [self.lengths[i] for i in range(self.size)]
 * 
 *     def get_indices(self):             # <<<<<<<<<<<<<<
 *         return [self.indices[i] for i in range(self.size)]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.CartesianProductIterator.get_indices", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":99
 *         return [self.indices[i] for i in range(self.size)]
 * 
 *     cpdef bint has_more(self):             # <<<<<<<<<<<<<<
 *         return not self.done
 * 
 */

static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_9has_more(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static int __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_has_more(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self, int __pyx_skip_dispatch) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  __Pyx_RefNannySetupContext("has_more", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_has_more); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 99, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_9has_more)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 99, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 99, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_5;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":100
 * 
 *     cpdef bint has_more(self):
 *         return not self.done             # <<<<<<<<<<<<<<
 * 
 *     cpdef list compose_next_value(self):
 */
  __pyx_r = (!(__pyx_v_self->done != 0));
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":99
 *         return [self.indices[i] for i in range(self.size)]
 * 
 *     cpdef bint has_more(self):             # <<<<<<<<<<<<<<
 *         return not self.done
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("ms_deisotope._c.feature_map.feature_processor.CartesianProductIterator.has_more", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_9has_more(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_8has_more[] = "CartesianProductIterator.has_more(self) -> bool";
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_9has_more(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("has_more (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_8has_more(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_8has_more(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("has_more", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_has_more(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 99, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.CartesianProductIterator.has_more", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":102
 *         return not self.done
 * 
 *     cpdef list compose_next_value(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             int i, ix
 */

static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_11compose_next_value(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_compose_next_value(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self, int __pyx_skip_dispatch) {
  int __pyx_v_i;
  int __pyx_v_ix;
  PyObject *__pyx_v_result = 0;
  PyObject *__pyx_v_sublist = 0;
  PyObject *__pyx_v_value = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  size_t __pyx_t_6;
  size_t __pyx_t_7;
  int __pyx_t_8;
  PyObject *__pyx_t_9;
  int __pyx_t_10;
  __Pyx_RefNannySetupContext("compose_next_value", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_compose_next_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 102, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_11compose_next_value)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 102, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 102, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":107
 *             list result, sublist
 *             object value
 *         if self.done:             # <<<<<<<<<<<<<<
 *             return None
 *         result = []
 */
  __pyx_t_5 = (__pyx_v_self->done != 0);
  if (__pyx_t_5) {

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":108
 *             object value
 *         if self.done:
 *             return None             # <<<<<<<<<<<<<<
 *         result = []
 *         for i in range(self.size):
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = ((PyObject*)Py_None); __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":107
 *             list result, sublist
 *             object value
 *         if self.done:             # <<<<<<<<<<<<<<
 *             return None
 *         result = []
 */
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":109
 *         if self.done:
 *             return None
 *         result = []             # <<<<<<<<<<<<<<
 *         for i in range(self.size):
 *             sublist = <list>PyList_GetItem(self.collections, i)
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":110
 *             return None
 *         result = []
 *         for i in range(self.size):             # <<<<<<<<<<<<<<
 *             sublist = <list>PyList_GetItem(self.collections, i)
 *             ix = self.indices[i]
 */
  __pyx_t_6 = __pyx_v_self->size;
  __pyx_t_7 = __pyx_t_6;
  for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {
    __pyx_v_i = __pyx_t_8;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":111
 *         result = []
 *         for i in range(self.size):
 *             sublist = <list>PyList_GetItem(self.collections, i)             # <<<<<<<<<<<<<<
 *             ix = self.indices[i]
 *             value = <object>PyList_GetItem(sublist, ix)
 */
    __pyx_t_1 = __pyx_v_self->collections;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_9 = PyList_GetItem(__pyx_t_1, __pyx_v_i); if (unlikely(__pyx_t_9 == ((PyObject *)NULL))) __PYX_ERR(0, 111, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = ((PyObject *)__pyx_t_9);
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_sublist, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":112
 *         for i in range(self.size):
 *             sublist = <list>PyList_GetItem(self.collections, i)
 *             ix = self.indices[i]             # <<<<<<<<<<<<<<
 *             value = <object>PyList_GetItem(sublist, ix)
 *             result.append(value)
 */
    __pyx_v_ix = (__pyx_v_self->indices[__pyx_v_i]);

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":113
 *             sublist = <list>PyList_GetItem(self.collections, i)
 *             ix = self.indices[i]
 *             value = <object>PyList_GetItem(sublist, ix)             # <<<<<<<<<<<<<<
 *             result.append(value)
 *         return result
 */
    __pyx_t_9 = PyList_GetItem(__pyx_v_sublist, __pyx_v_ix); if (unlikely(__pyx_t_9 == ((PyObject *)NULL))) __PYX_ERR(0, 113, __pyx_L1_error)
    __pyx_t_1 = ((PyObject *)__pyx_t_9);
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":114
 *             ix = self.indices[i]
 *             value = <object>PyList_GetItem(sublist, ix)
 *             result.append(value)             # <<<<<<<<<<<<<<
 *         return result
 * 
 */
    __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_result, __pyx_v_value); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 114, __pyx_L1_error)
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":115
 *             value = <object>PyList_GetItem(sublist, ix)
 *             result.append(value)
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     cpdef advance(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":102
 *         return not self.done
 * 
 *     cpdef list compose_next_value(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             int i, ix
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.CartesianProductIterator.compose_next_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_sublist);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_11compose_next_value(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_10compose_next_value[] = "CartesianProductIterator.compose_next_value(self) -> list";
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_11compose_next_value(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("compose_next_value (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_10compose_next_value(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_10compose_next_value(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("compose_next_value", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_compose_next_value(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.CartesianProductIterator.compose_next_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":117
 *         return result
 * 
 *     cpdef advance(self):             # <<<<<<<<<<<<<<
 *         for i in range(self.size - 1, -1, -1):
 *             if self.indices[i] == self.lengths[i] - 1:
 */

static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_13advance(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_advance(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self, int __pyx_skip_dispatch) {
  size_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_t_6;
  size_t __pyx_t_7;
  __Pyx_RefNannySetupContext("advance", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_advance); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 117, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_13advance)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 117, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":118
 * 
 *     cpdef advance(self):
 *         for i in range(self.size - 1, -1, -1):             # <<<<<<<<<<<<<<
 *             if self.indices[i] == self.lengths[i] - 1:
 *                 self.indices[i] = 0
 */
  for (__pyx_t_5 = (__pyx_v_self->size - 1) + 1; __pyx_t_5 > -1L + 1; ) { __pyx_t_5-=1;
    __pyx_v_i = __pyx_t_5;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":119
 *     cpdef advance(self):
 *         for i in range(self.size - 1, -1, -1):
 *             if self.indices[i] == self.lengths[i] - 1:             # <<<<<<<<<<<<<<
 *                 self.indices[i] = 0
 *                 if i == 0:
 */
    __pyx_t_6 = (((__pyx_v_self->indices[__pyx_v_i]) == ((__pyx_v_self->lengths[__pyx_v_i]) - 1)) != 0);
    if (__pyx_t_6) {

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":120
 *         for i in range(self.size - 1, -1, -1):
 *             if self.indices[i] == self.lengths[i] - 1:
 *                 self.indices[i] = 0             # <<<<<<<<<<<<<<
 *                 if i == 0:
 *                     self.done = True
 */
      (__pyx_v_self->indices[__pyx_v_i]) = 0;

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":121
 *             if self.indices[i] == self.lengths[i] - 1:
 *                 self.indices[i] = 0
 *                 if i == 0:             # <<<<<<<<<<<<<<
 *                     self.done = True
 *             else:
 */
      __pyx_t_6 = ((__pyx_v_i == 0) != 0);
      if (__pyx_t_6) {

        /* "ms_deisotope/_c/feature_map/feature_processor.pyx":122
 *                 self.indices[i] = 0
 *                 if i == 0:
 *                     self.done = True             # <<<<<<<<<<<<<<
 *             else:
 *                 self.indices[i] += 1
 */
        __pyx_v_self->done = 1;

        /* "ms_deisotope/_c/feature_map/feature_processor.pyx":121
 *             if self.indices[i] == self.lengths[i] - 1:
 *                 self.indices[i] = 0
 *                 if i == 0:             # <<<<<<<<<<<<<<
 *                     self.done = True
 *             else:
 */
      }

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":119
 *     cpdef advance(self):
 *         for i in range(self.size - 1, -1, -1):
 *             if self.indices[i] == self.lengths[i] - 1:             # <<<<<<<<<<<<<<
 *                 self.indices[i] = 0
 *                 if i == 0:
 */
      goto __pyx_L5;
    }

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":124
 *                     self.done = True
 *             else:
 *                 self.indices[i] += 1             # <<<<<<<<<<<<<<
 *                 break
 * 
 */
    /*else*/ {
      __pyx_t_7 = __pyx_v_i;
      (__pyx_v_self->indices[__pyx_t_7]) = ((__pyx_v_self->indices[__pyx_t_7]) + 1);

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":125
 *             else:
 *                 self.indices[i] += 1
 *                 break             # <<<<<<<<<<<<<<
 * 
 *     cpdef list get_next_value(self):
 */
      goto __pyx_L4_break;
    }
    __pyx_L5:;
  }
  __pyx_L4_break:;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":117
 *         return result
 * 
 *     cpdef advance(self):             # <<<<<<<<<<<<<<
 *         for i in range(self.size - 1, -1, -1):
 *             if self.indices[i] == self.lengths[i] - 1:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.CartesianProductIterator.advance", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_13advance(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_12advance[] = "CartesianProductIterator.advance(self)";
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_13advance(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("advance (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_12advance(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_12advance(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("advance", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_advance(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.CartesianProductIterator.advance", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":127
 *                 break
 * 
 *     cpdef list get_next_value(self):             # <<<<<<<<<<<<<<
 *         cdef list value
 *         if self.done:
 */

static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_15get_next_value(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_get_next_value(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_value = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  __Pyx_RefNannySetupContext("get_next_value", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_next_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 127, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_15get_next_value)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 127, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 127, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":129
 *     cpdef list get_next_value(self):
 *         cdef list value
 *         if self.done:             # <<<<<<<<<<<<<<
 *             return None
 *         else:
 */
  __pyx_t_5 = (__pyx_v_self->done != 0);
  if (__pyx_t_5) {

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":130
 *         cdef list value
 *         if self.done:
 *             return None             # <<<<<<<<<<<<<<
 *         else:
 *             value = self.compose_next_value()
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = ((PyObject*)Py_None); __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":129
 *     cpdef list get_next_value(self):
 *         cdef list value
 *         if self.done:             # <<<<<<<<<<<<<<
 *             return None
 *         else:
 */
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":132
 *             return None
 *         else:
 *             value = self.compose_next_value()             # <<<<<<<<<<<<<<
 *             self.advance()
 *             return value
 */
  /*else*/ {
    __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *)__pyx_v_self->__pyx_vtab)->compose_next_value(__pyx_v_self, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 132, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_value = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":133
 *         else:
 *             value = self.compose_next_value()
 *             self.advance()             # <<<<<<<<<<<<<<
 *             return value
 * 
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *)__pyx_v_self->__pyx_vtab)->advance(__pyx_v_self, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 133, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":134
 *             value = self.compose_next_value()
 *             self.advance()
 *             return value             # <<<<<<<<<<<<<<
 * 
 *     def __iter__(self):
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_value);
    __pyx_r = __pyx_v_value;
    goto __pyx_L0;
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":127
 *                 break
 * 
 *     cpdef list get_next_value(self):             # <<<<<<<<<<<<<<
 *         cdef list value
 *         if self.done:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.CartesianProductIterator.get_next_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_15get_next_value(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_14get_next_value[] = "CartesianProductIterator.get_next_value(self) -> list";
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_15get_next_value(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_next_value (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_14get_next_value(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_14get_next_value(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("get_next_value", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_get_next_value(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.CartesianProductIterator.get_next_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":136
 *             return value
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         return self
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_17__iter__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_17__iter__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_16__iter__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_16__iter__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__", 0);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":137
 * 
 *     def __iter__(self):
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def __next__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __pyx_r = ((PyObject *)__pyx_v_self);
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":136
 *             return value
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         return self
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":139
 *         return self
 * 
 *     def __next__(self):             # <<<<<<<<<<<<<<
 *         if self.done:
 *             raise StopIteration()
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_19__next__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_19__next__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__next__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_18__next__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_18__next__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self) {
  PyObject *__pyx_v_value = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  __Pyx_RefNannySetupContext("__next__", 0);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":140
 * 
 *     def __next__(self):
 *         if self.done:             # <<<<<<<<<<<<<<
 *             raise StopIteration()
 *         value = self.compose_next_value()
 */
  __pyx_t_1 = (__pyx_v_self->done != 0);
  if (unlikely(__pyx_t_1)) {

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":141
 *     def __next__(self):
 *         if self.done:
 *             raise StopIteration()             # <<<<<<<<<<<<<<
 *         value = self.compose_next_value()
 *         if value is None:
 */
    __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_builtin_StopIteration); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 141, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 141, __pyx_L1_error)

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":140
 * 
 *     def __next__(self):
 *         if self.done:             # <<<<<<<<<<<<<<
 *             raise StopIteration()
 *         value = self.compose_next_value()
 */
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":142
 *         if self.done:
 *             raise StopIteration()
 *         value = self.compose_next_value()             # <<<<<<<<<<<<<<
 *         if value is None:
 *             raise StopIteration()
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *)__pyx_v_self->__pyx_vtab)->compose_next_value(__pyx_v_self, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_value = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":143
 *             raise StopIteration()
 *         value = self.compose_next_value()
 *         if value is None:             # <<<<<<<<<<<<<<
 *             raise StopIteration()
 *         else:
 */
  __pyx_t_1 = (__pyx_v_value == ((PyObject*)Py_None));
  __pyx_t_3 = (__pyx_t_1 != 0);
  if (unlikely(__pyx_t_3)) {

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":144
 *         value = self.compose_next_value()
 *         if value is None:
 *             raise StopIteration()             # <<<<<<<<<<<<<<
 *         else:
 *             self.advance()
 */
    __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_builtin_StopIteration); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 144, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 144, __pyx_L1_error)

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":143
 *             raise StopIteration()
 *         value = self.compose_next_value()
 *         if value is None:             # <<<<<<<<<<<<<<
 *             raise StopIteration()
 *         else:
 */
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":146
 *             raise StopIteration()
 *         else:
 *             self.advance()             # <<<<<<<<<<<<<<
 *         return value
 * 
 */
  /*else*/ {
    __pyx_t_2 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *)__pyx_v_self->__pyx_vtab)->advance(__pyx_v_self, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 146, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":147
 *         else:
 *             self.advance()
 *         return value             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_value);
  __pyx_r = __pyx_v_value;
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":139
 *         return self
 * 
 *     def __next__(self):             # <<<<<<<<<<<<<<
 *         if self.done:
 *             raise StopIteration()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.CartesianProductIterator.__next__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":62
 * cdef class CartesianProductIterator(object):
 *     cdef:
 *         public list collections             # <<<<<<<<<<<<<<
 *         public size_t size
 *         int* lengths
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_11collections_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_11collections_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_11collections___get__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_11collections___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->collections);
  __pyx_r = __pyx_v_self->collections;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_11collections_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_11collections_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_11collections_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_11collections_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(PyList_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(0, 62, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->collections);
  __Pyx_DECREF(__pyx_v_self->collections);
  __pyx_v_self->collections = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.CartesianProductIterator.collections.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_11collections_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_11collections_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_11collections_4__del__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_11collections_4__del__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->collections);
  __Pyx_DECREF(__pyx_v_self->collections);
  __pyx_v_self->collections = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":63
 *     cdef:
 *         public list collections
 *         public size_t size             # <<<<<<<<<<<<<<
 *         int* lengths
 *         int* indices
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_4size_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_4size_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_4size___get__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_4size___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.CartesianProductIterator.size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_4size_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_4size_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_4size_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_4size_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_size_t(__pyx_v_value); if (unlikely((__pyx_t_1 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 63, __pyx_L1_error)
  __pyx_v_self->size = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.CartesianProductIterator.size.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":66
 *         int* lengths
 *         int* indices
 *         public bint done             # <<<<<<<<<<<<<<
 *         public size_t total_combinations
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_4done_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_4done_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_4done___get__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_4done___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->done); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.CartesianProductIterator.done.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_4done_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_4done_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_4done_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_4done_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 66, __pyx_L1_error)
  __pyx_v_self->done = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.CartesianProductIterator.done.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":67
 *         int* indices
 *         public bint done
 *         public size_t total_combinations             # <<<<<<<<<<<<<<
 * 
 *     def __cinit__(self, collections):
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_18total_combinations_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_18total_combinations_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_18total_combinations___get__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_18total_combinations___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->total_combinations); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.CartesianProductIterator.total_combinations.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_18total_combinations_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_18total_combinations_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_18total_combinations_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_18total_combinations_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_size_t(__pyx_v_value); if (unlikely((__pyx_t_1 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 67, __pyx_L1_error)
  __pyx_v_self->total_combinations = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.CartesianProductIterator.total_combinations.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_21__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_20__reduce_cython__[] = "CartesianProductIterator.__reduce_cython__(self)";
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_21__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_20__reduce_cython__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_20__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.CartesianProductIterator.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_23__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_22__setstate_cython__[] = "CartesianProductIterator.__setstate_cython__(self, __pyx_state)";
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_23__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_22__setstate_cython__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_22__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.CartesianProductIterator.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":150
 * 
 * 
 * cdef tuple _conform_envelopes(list experimental, list base_theoretical, size_t* n_missing_out,             # <<<<<<<<<<<<<<
 *                               double minimum_theoretical_abundance=0.05):
 *     cdef:
 */

static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor__conform_envelopes(PyObject *__pyx_v_experimental, PyObject *__pyx_v_base_theoretical, size_t *__pyx_v_n_missing_out, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor__conform_envelopes *__pyx_optional_args) {
  double __pyx_v_minimum_theoretical_abundance = ((double)0.05);
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *__pyx_v_conformer = 0;
  PyObject *__pyx_v_cleaned_eid = 0;
  PyObject *__pyx_v_tid = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer_conform __pyx_t_2;
  size_t __pyx_t_3;
  __Pyx_RefNannySetupContext("_conform_envelopes", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_minimum_theoretical_abundance = __pyx_optional_args->minimum_theoretical_abundance;
    }
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":156
 *         list cleaned_eid, tid
 * 
 *     conformer = envelope_conformer._create()             # <<<<<<<<<<<<<<
 *     conformer.acquire(experimental, base_theoretical)
 *     conformer.conform(minimum_theoretical_abundance)
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer__create()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_conformer = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":157
 * 
 *     conformer = envelope_conformer._create()
 *     conformer.acquire(experimental, base_theoretical)             # <<<<<<<<<<<<<<
 *     conformer.conform(minimum_theoretical_abundance)
 * 
 */
  ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *)__pyx_v_conformer->__pyx_vtab)->acquire(__pyx_v_conformer, __pyx_v_experimental, __pyx_v_base_theoretical);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":158
 *     conformer = envelope_conformer._create()
 *     conformer.acquire(experimental, base_theoretical)
 *     conformer.conform(minimum_theoretical_abundance)             # <<<<<<<<<<<<<<
 * 
 *     cleaned_eid = conformer.experimental
 */
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.minimum_theoretical_abundance = __pyx_v_minimum_theoretical_abundance;
  ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *)__pyx_v_conformer->__pyx_vtab)->conform(__pyx_v_conformer, &__pyx_t_2); 

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":160
 *     conformer.conform(minimum_theoretical_abundance)
 * 
 *     cleaned_eid = conformer.experimental             # <<<<<<<<<<<<<<
 *     tid = conformer.theoretical
 *     n_missing_out[0] = conformer.n_missing
 */
  __pyx_t_1 = __pyx_v_conformer->experimental;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_cleaned_eid = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":161
 * 
 *     cleaned_eid = conformer.experimental
 *     tid = conformer.theoretical             # <<<<<<<<<<<<<<
 *     n_missing_out[0] = conformer.n_missing
 *     return (cleaned_eid, tid)
 */
  __pyx_t_1 = __pyx_v_conformer->theoretical;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_tid = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":162
 *     cleaned_eid = conformer.experimental
 *     tid = conformer.theoretical
 *     n_missing_out[0] = conformer.n_missing             # <<<<<<<<<<<<<<
 *     return (cleaned_eid, tid)
 * 
 */
  __pyx_t_3 = __pyx_v_conformer->n_missing;
  (__pyx_v_n_missing_out[0]) = __pyx_t_3;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":163
 *     tid = conformer.theoretical
 *     n_missing_out[0] = conformer.n_missing
 *     return (cleaned_eid, tid)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_cleaned_eid);
  __Pyx_GIVEREF(__pyx_v_cleaned_eid);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_cleaned_eid);
  __Pyx_INCREF(__pyx_v_tid);
  __Pyx_GIVEREF(__pyx_v_tid);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_tid);
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":150
 * 
 * 
 * cdef tuple _conform_envelopes(list experimental, list base_theoretical, size_t* n_missing_out,             # <<<<<<<<<<<<<<
 *                               double minimum_theoretical_abundance=0.05):
 *     cdef:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor._conform_envelopes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_conformer);
  __Pyx_XDECREF(__pyx_v_cleaned_eid);
  __Pyx_XDECREF(__pyx_v_tid);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":173
 * 
 *     @staticmethod
 *     cdef envelope_conformer _create():             # <<<<<<<<<<<<<<
 *         return envelope_conformer.__new__(envelope_conformer)
 * 
 */

static struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer__create(void) {
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("_create", 0);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":174
 *     @staticmethod
 *     cdef envelope_conformer _create():
 *         return envelope_conformer.__new__(envelope_conformer)             # <<<<<<<<<<<<<<
 * 
 *     cdef void acquire(self, list experimental, list theoretical):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer(((PyTypeObject *)__pyx_ptype_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_GOTREF(((PyObject *)__pyx_t_1));
  __pyx_r = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":173
 * 
 *     @staticmethod
 *     cdef envelope_conformer _create():             # <<<<<<<<<<<<<<
 *         return envelope_conformer.__new__(envelope_conformer)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.envelope_conformer._create", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":176
 *         return envelope_conformer.__new__(envelope_conformer)
 * 
 *     cdef void acquire(self, list experimental, list theoretical):             # <<<<<<<<<<<<<<
 *         self.experimental = experimental
 *         self.theoretical = theoretical
 */

static void __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer_acquire(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *__pyx_v_self, PyObject *__pyx_v_experimental, PyObject *__pyx_v_theoretical) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("acquire", 0);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":177
 * 
 *     cdef void acquire(self, list experimental, list theoretical):
 *         self.experimental = experimental             # <<<<<<<<<<<<<<
 *         self.theoretical = theoretical
 *         self.n_missing = 0
 */
  __Pyx_INCREF(__pyx_v_experimental);
  __Pyx_GIVEREF(__pyx_v_experimental);
  __Pyx_GOTREF(__pyx_v_self->experimental);
  __Pyx_DECREF(__pyx_v_self->experimental);
  __pyx_v_self->experimental = __pyx_v_experimental;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":178
 *     cdef void acquire(self, list experimental, list theoretical):
 *         self.experimental = experimental
 *         self.theoretical = theoretical             # <<<<<<<<<<<<<<
 *         self.n_missing = 0
 * 
 */
  __Pyx_INCREF(__pyx_v_theoretical);
  __Pyx_GIVEREF(__pyx_v_theoretical);
  __Pyx_GOTREF(__pyx_v_self->theoretical);
  __Pyx_DECREF(__pyx_v_self->theoretical);
  __pyx_v_self->theoretical = __pyx_v_theoretical;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":179
 *         self.experimental = experimental
 *         self.theoretical = theoretical
 *         self.n_missing = 0             # <<<<<<<<<<<<<<
 * 
 *     cdef void conform(self, double minimum_theoretical_abundance=0.05):
 */
  __pyx_v_self->n_missing = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":176
 *         return envelope_conformer.__new__(envelope_conformer)
 * 
 *     cdef void acquire(self, list experimental, list theoretical):             # <<<<<<<<<<<<<<
 *         self.experimental = experimental
 *         self.theoretical = theoretical
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":181
 *         self.n_missing = 0
 * 
 *     cdef void conform(self, double minimum_theoretical_abundance=0.05):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double total = 0
 */

static void __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer_conform(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *__pyx_v_self, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer_conform *__pyx_optional_args) {
  double __pyx_v_minimum_theoretical_abundance = ((double)0.05);
  double __pyx_v_total;
  size_t __pyx_v_n_missing;
  size_t __pyx_v_i;
  PyObject *__pyx_v_cleaned_eid = 0;
  PyObject *__pyx_v_tid = 0;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_fpeak = 0;
  struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_v_tpeak = 0;
  struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_v_peak = 0;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  size_t __pyx_t_4;
  PyObject *__pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("conform", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_minimum_theoretical_abundance = __pyx_optional_args->minimum_theoretical_abundance;
    }
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":183
 *     cdef void conform(self, double minimum_theoretical_abundance=0.05):
 *         cdef:
 *             double total = 0             # <<<<<<<<<<<<<<
 *             size_t n_missing = 0
 *             size_t i = 0
 */
  __pyx_v_total = 0.0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":184
 *         cdef:
 *             double total = 0
 *             size_t n_missing = 0             # <<<<<<<<<<<<<<
 *             size_t i = 0
 *             list cleaned_eid
 */
  __pyx_v_n_missing = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":185
 *             double total = 0
 *             size_t n_missing = 0
 *             size_t i = 0             # <<<<<<<<<<<<<<
 *             list cleaned_eid
 *             list tid
 */
  __pyx_v_i = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":191
 *             TheoreticalPeak tpeak, peak
 * 
 *         cleaned_eid = []             # <<<<<<<<<<<<<<
 * 
 *         for i in range(PyList_GET_SIZE(self.experimental)):
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_cleaned_eid = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":193
 *         cleaned_eid = []
 * 
 *         for i in range(PyList_GET_SIZE(self.experimental)):             # <<<<<<<<<<<<<<
 *             fpeak = <FittedPeak>PyList_GET_ITEM(self.experimental, i)
 *             if fpeak is None:
 */
  __pyx_t_1 = __pyx_v_self->experimental;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyList_GET_SIZE(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":194
 * 
 *         for i in range(PyList_GET_SIZE(self.experimental)):
 *             fpeak = <FittedPeak>PyList_GET_ITEM(self.experimental, i)             # <<<<<<<<<<<<<<
 *             if fpeak is None:
 *                 peak = <TheoreticalPeak>PyList_GET_ITEM(self.theoretical, i)
 */
    __pyx_t_1 = __pyx_v_self->experimental;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_5 = PyList_GET_ITEM(__pyx_t_1, __pyx_v_i);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = ((PyObject *)__pyx_t_5);
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_fpeak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":195
 *         for i in range(PyList_GET_SIZE(self.experimental)):
 *             fpeak = <FittedPeak>PyList_GET_ITEM(self.experimental, i)
 *             if fpeak is None:             # <<<<<<<<<<<<<<
 *                 peak = <TheoreticalPeak>PyList_GET_ITEM(self.theoretical, i)
 *                 fpeak = FittedPeak._create(peak.mz, 1, 1, -1, -1, 0, 1, 0, 0)
 */
    __pyx_t_6 = (((PyObject *)__pyx_v_fpeak) == Py_None);
    __pyx_t_7 = (__pyx_t_6 != 0);
    if (__pyx_t_7) {

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":196
 *             fpeak = <FittedPeak>PyList_GET_ITEM(self.experimental, i)
 *             if fpeak is None:
 *                 peak = <TheoreticalPeak>PyList_GET_ITEM(self.theoretical, i)             # <<<<<<<<<<<<<<
 *                 fpeak = FittedPeak._create(peak.mz, 1, 1, -1, -1, 0, 1, 0, 0)
 *                 if peak.intensity > minimum_theoretical_abundance:
 */
      __pyx_t_1 = __pyx_v_self->theoretical;
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_5 = PyList_GET_ITEM(__pyx_t_1, __pyx_v_i);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = ((PyObject *)__pyx_t_5);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_peak, ((struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":197
 *             if fpeak is None:
 *                 peak = <TheoreticalPeak>PyList_GET_ITEM(self.theoretical, i)
 *                 fpeak = FittedPeak._create(peak.mz, 1, 1, -1, -1, 0, 1, 0, 0)             # <<<<<<<<<<<<<<
 *                 if peak.intensity > minimum_theoretical_abundance:
 *                     n_missing += 1
 */
      __pyx_t_1 = ((PyObject *)__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_FittedPeak->_create(__pyx_v_peak->mz, 1.0, 1.0, -1.0, -1.0, 0.0, 1, 0, 0.0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 197, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF_SET(__pyx_v_fpeak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":198
 *                 peak = <TheoreticalPeak>PyList_GET_ITEM(self.theoretical, i)
 *                 fpeak = FittedPeak._create(peak.mz, 1, 1, -1, -1, 0, 1, 0, 0)
 *                 if peak.intensity > minimum_theoretical_abundance:             # <<<<<<<<<<<<<<
 *                     n_missing += 1
 *             total += fpeak.intensity
 */
      __pyx_t_7 = ((__pyx_v_peak->intensity > __pyx_v_minimum_theoretical_abundance) != 0);
      if (__pyx_t_7) {

        /* "ms_deisotope/_c/feature_map/feature_processor.pyx":199
 *                 fpeak = FittedPeak._create(peak.mz, 1, 1, -1, -1, 0, 1, 0, 0)
 *                 if peak.intensity > minimum_theoretical_abundance:
 *                     n_missing += 1             # <<<<<<<<<<<<<<
 *             total += fpeak.intensity
 *             cleaned_eid.append(fpeak)
 */
        __pyx_v_n_missing = (__pyx_v_n_missing + 1);

        /* "ms_deisotope/_c/feature_map/feature_processor.pyx":198
 *                 peak = <TheoreticalPeak>PyList_GET_ITEM(self.theoretical, i)
 *                 fpeak = FittedPeak._create(peak.mz, 1, 1, -1, -1, 0, 1, 0, 0)
 *                 if peak.intensity > minimum_theoretical_abundance:             # <<<<<<<<<<<<<<
 *                     n_missing += 1
 *             total += fpeak.intensity
 */
      }

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":195
 *         for i in range(PyList_GET_SIZE(self.experimental)):
 *             fpeak = <FittedPeak>PyList_GET_ITEM(self.experimental, i)
 *             if fpeak is None:             # <<<<<<<<<<<<<<
 *                 peak = <TheoreticalPeak>PyList_GET_ITEM(self.theoretical, i)
 *                 fpeak = FittedPeak._create(peak.mz, 1, 1, -1, -1, 0, 1, 0, 0)
 */
    }

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":200
 *                 if peak.intensity > minimum_theoretical_abundance:
 *                     n_missing += 1
 *             total += fpeak.intensity             # <<<<<<<<<<<<<<
 *             cleaned_eid.append(fpeak)
 *             i += 1
 */
    __pyx_v_total = (__pyx_v_total + __pyx_v_fpeak->__pyx_base.intensity);

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":201
 *                     n_missing += 1
 *             total += fpeak.intensity
 *             cleaned_eid.append(fpeak)             # <<<<<<<<<<<<<<
 *             i += 1
 * 
 */
    __pyx_t_8 = __Pyx_PyList_Append(__pyx_v_cleaned_eid, ((PyObject *)__pyx_v_fpeak)); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 201, __pyx_L1_error)

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":202
 *             total += fpeak.intensity
 *             cleaned_eid.append(fpeak)
 *             i += 1             # <<<<<<<<<<<<<<
 * 
 *         tid = []
 */
    __pyx_v_i = (__pyx_v_i + 1);
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":204
 *             i += 1
 * 
 *         tid = []             # <<<<<<<<<<<<<<
 *         for i in range(PyList_GET_SIZE(self.theoretical)):
 *             tpeak = <TheoreticalPeak>PyList_GET_ITEM(self.theoretical, i)
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_tid = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":205
 * 
 *         tid = []
 *         for i in range(PyList_GET_SIZE(self.theoretical)):             # <<<<<<<<<<<<<<
 *             tpeak = <TheoreticalPeak>PyList_GET_ITEM(self.theoretical, i)
 *             peak = tpeak.clone()
 */
  __pyx_t_1 = __pyx_v_self->theoretical;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyList_GET_SIZE(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":206
 *         tid = []
 *         for i in range(PyList_GET_SIZE(self.theoretical)):
 *             tpeak = <TheoreticalPeak>PyList_GET_ITEM(self.theoretical, i)             # <<<<<<<<<<<<<<
 *             peak = tpeak.clone()
 *             peak.intensity *= total
 */
    __pyx_t_1 = __pyx_v_self->theoretical;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_5 = PyList_GET_ITEM(__pyx_t_1, __pyx_v_i);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = ((PyObject *)__pyx_t_5);
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_tpeak, ((struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":207
 *         for i in range(PyList_GET_SIZE(self.theoretical)):
 *             tpeak = <TheoreticalPeak>PyList_GET_ITEM(self.theoretical, i)
 *             peak = tpeak.clone()             # <<<<<<<<<<<<<<
 *             peak.intensity *= total
 *             tid.append(peak)
 */
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *)__pyx_v_tpeak->__pyx_vtab)->clone(__pyx_v_tpeak, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 207, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_peak, ((struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":208
 *             tpeak = <TheoreticalPeak>PyList_GET_ITEM(self.theoretical, i)
 *             peak = tpeak.clone()
 *             peak.intensity *= total             # <<<<<<<<<<<<<<
 *             tid.append(peak)
 *         self.experimental = cleaned_eid
 */
    __pyx_v_peak->intensity = (__pyx_v_peak->intensity * __pyx_v_total);

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":209
 *             peak = tpeak.clone()
 *             peak.intensity *= total
 *             tid.append(peak)             # <<<<<<<<<<<<<<
 *         self.experimental = cleaned_eid
 *         self.theoretical = tid
 */
    __pyx_t_8 = __Pyx_PyList_Append(__pyx_v_tid, ((PyObject *)__pyx_v_peak)); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 209, __pyx_L1_error)
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":210
 *             peak.intensity *= total
 *             tid.append(peak)
 *         self.experimental = cleaned_eid             # <<<<<<<<<<<<<<
 *         self.theoretical = tid
 *         self.n_missing = n_missing
 */
  __Pyx_INCREF(__pyx_v_cleaned_eid);
  __Pyx_GIVEREF(__pyx_v_cleaned_eid);
  __Pyx_GOTREF(__pyx_v_self->experimental);
  __Pyx_DECREF(__pyx_v_self->experimental);
  __pyx_v_self->experimental = __pyx_v_cleaned_eid;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":211
 *             tid.append(peak)
 *         self.experimental = cleaned_eid
 *         self.theoretical = tid             # <<<<<<<<<<<<<<
 *         self.n_missing = n_missing
 * 
 */
  __Pyx_INCREF(__pyx_v_tid);
  __Pyx_GIVEREF(__pyx_v_tid);
  __Pyx_GOTREF(__pyx_v_self->theoretical);
  __Pyx_DECREF(__pyx_v_self->theoretical);
  __pyx_v_self->theoretical = __pyx_v_tid;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":212
 *         self.experimental = cleaned_eid
 *         self.theoretical = tid
 *         self.n_missing = n_missing             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_v_self->n_missing = __pyx_v_n_missing;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":181
 *         self.n_missing = 0
 * 
 *     cdef void conform(self, double minimum_theoretical_abundance=0.05):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double total = 0
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_WriteUnraisable("ms_deisotope._c.feature_map.feature_processor.envelope_conformer.conform", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_cleaned_eid);
  __Pyx_XDECREF(__pyx_v_tid);
  __Pyx_XDECREF((PyObject *)__pyx_v_fpeak);
  __Pyx_XDECREF((PyObject *)__pyx_v_tpeak);
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);
  __Pyx_RefNannyFinishContext();
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer___reduce_cython__[] = "envelope_conformer.__reduce_cython__(self)";
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer___reduce_cython__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer___reduce_cython__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.experimental, self.n_missing, self.theoretical)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->n_missing); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_self->experimental);
  __Pyx_GIVEREF(__pyx_v_self->experimental);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_self->experimental);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->theoretical);
  __Pyx_GIVEREF(__pyx_v_self->theoretical);
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_self->theoretical);
  __pyx_t_1 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.experimental, self.n_missing, self.theoretical)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_2 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v__dict = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "(tree fragment)":7
 *     state = (self.experimental, self.n_missing, self.theoretical)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_3 = (__pyx_v__dict != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v__dict);
    __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.experimental is not None or self.theoretical is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.experimental, self.n_missing, self.theoretical)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.experimental is not None or self.theoretical is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_envelope_conformer, (type(self), 0xba9b7dc, None), state
 */
  /*else*/ {
    __pyx_t_3 = (__pyx_v_self->experimental != ((PyObject*)Py_None));
    __pyx_t_5 = (__pyx_t_3 != 0);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->theoretical != ((PyObject*)Py_None));
    __pyx_t_3 = (__pyx_t_5 != 0);
    __pyx_t_4 = __pyx_t_3;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_4;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.experimental is not None or self.theoretical is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_envelope_conformer, (type(self), 0xba9b7dc, None), state
 *     else:
 */
  __pyx_t_4 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_4) {

    /* "(tree fragment)":13
 *         use_setstate = self.experimental is not None or self.theoretical is not None
 *     if use_setstate:
 *         return __pyx_unpickle_envelope_conformer, (type(self), 0xba9b7dc, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_envelope_conformer, (type(self), 0xba9b7dc, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pyx_unpickle_envelope_conforme); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_195672028);
    __Pyx_GIVEREF(__pyx_int_195672028);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_195672028);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_2, 2, Py_None);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_2);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_v_state);
    __pyx_t_1 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.experimental is not None or self.theoretical is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_envelope_conformer, (type(self), 0xba9b7dc, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_envelope_conformer, (type(self), 0xba9b7dc, None), state
 *     else:
 *         return __pyx_unpickle_envelope_conformer, (type(self), 0xba9b7dc, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_envelope_conformer__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_pyx_unpickle_envelope_conforme); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_195672028);
    __Pyx_GIVEREF(__pyx_int_195672028);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_195672028);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_state);
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2);
    __pyx_t_6 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.envelope_conformer.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_envelope_conformer, (type(self), 0xba9b7dc, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_envelope_conformer__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer_2__setstate_cython__[] = "envelope_conformer.__setstate_cython__(self, __pyx_state)";
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer_2__setstate_cython__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer_2__setstate_cython__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_envelope_conformer, (type(self), 0xba9b7dc, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_envelope_conformer__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor___pyx_unpickle_envelope_conformer__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_envelope_conformer, (type(self), 0xba9b7dc, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_envelope_conformer__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.envelope_conformer.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":215
 * 
 * 
 * cpdef tuple conform_envelopes(list experimental, list base_theoretical,             # <<<<<<<<<<<<<<
 *                               double minimum_theoretical_abundance=0.01):
 *     cdef:
 */

static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_1conform_envelopes(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_conform_envelopes(PyObject *__pyx_v_experimental, PyObject *__pyx_v_base_theoretical, CYTHON_UNUSED int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_conform_envelopes *__pyx_optional_args) {
  double __pyx_v_minimum_theoretical_abundance = ((double)0.01);
  size_t __pyx_v_n_missing;
  PyObject *__pyx_v_out = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor__conform_envelopes __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("conform_envelopes", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_minimum_theoretical_abundance = __pyx_optional_args->minimum_theoretical_abundance;
    }
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":221
 *         tuple out
 * 
 *     out = _conform_envelopes(experimental, base_theoretical, &n_missing, minimum_theoretical_abundance)             # <<<<<<<<<<<<<<
 *     out += (n_missing,)
 *     return out
 */
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.minimum_theoretical_abundance = __pyx_v_minimum_theoretical_abundance;
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor__conform_envelopes(__pyx_v_experimental, __pyx_v_base_theoretical, (&__pyx_v_n_missing), &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 221, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_out = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":222
 * 
 *     out = _conform_envelopes(experimental, base_theoretical, &n_missing, minimum_theoretical_abundance)
 *     out += (n_missing,)             # <<<<<<<<<<<<<<
 *     return out
 * 
 */
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_n_missing); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 222, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 222, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_v_out, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 222, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF_SET(__pyx_v_out, ((PyObject*)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":223
 *     out = _conform_envelopes(experimental, base_theoretical, &n_missing, minimum_theoretical_abundance)
 *     out += (n_missing,)
 *     return out             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_out);
  __pyx_r = __pyx_v_out;
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":215
 * 
 * 
 * cpdef tuple conform_envelopes(list experimental, list base_theoretical,             # <<<<<<<<<<<<<<
 *                               double minimum_theoretical_abundance=0.01):
 *     cdef:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.conform_envelopes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_out);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_1conform_envelopes(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_conform_envelopes[] = "conform_envelopes(list experimental, list base_theoretical, double minimum_theoretical_abundance=0.01) -> tuple";
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_1conform_envelopes(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_experimental = 0;
  PyObject *__pyx_v_base_theoretical = 0;
  double __pyx_v_minimum_theoretical_abundance;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("conform_envelopes (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_experimental,&__pyx_n_s_base_theoretical,&__pyx_n_s_minimum_theoretical_abundance,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_experimental)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_base_theoretical)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("conform_envelopes", 0, 2, 3, 1); __PYX_ERR(0, 215, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_minimum_theoretical_abundance);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "conform_envelopes") < 0)) __PYX_ERR(0, 215, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_experimental = ((PyObject*)values[0]);
    __pyx_v_base_theoretical = ((PyObject*)values[1]);
    if (values[2]) {
      __pyx_v_minimum_theoretical_abundance = __pyx_PyFloat_AsDouble(values[2]); if (unlikely((__pyx_v_minimum_theoretical_abundance == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 216, __pyx_L3_error)
    } else {
      __pyx_v_minimum_theoretical_abundance = ((double)0.01);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("conform_envelopes", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 215, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.conform_envelopes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_experimental), (&PyList_Type), 1, "experimental", 1))) __PYX_ERR(0, 215, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_base_theoretical), (&PyList_Type), 1, "base_theoretical", 1))) __PYX_ERR(0, 215, __pyx_L1_error)
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_conform_envelopes(__pyx_self, __pyx_v_experimental, __pyx_v_base_theoretical, __pyx_v_minimum_theoretical_abundance);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_conform_envelopes(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_experimental, PyObject *__pyx_v_base_theoretical, double __pyx_v_minimum_theoretical_abundance) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_conform_envelopes __pyx_t_2;
  __Pyx_RefNannySetupContext("conform_envelopes", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.minimum_theoretical_abundance = __pyx_v_minimum_theoretical_abundance;
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_conform_envelopes(__pyx_v_experimental, __pyx_v_base_theoretical, 0, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 215, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.conform_envelopes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":226
 * 
 * 
 * cdef int has_no_valid_features(list features):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i, n, cnt
 */

static int __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_has_no_valid_features(PyObject *__pyx_v_features) {
  size_t __pyx_v_i;
  size_t __pyx_v_n;
  size_t __pyx_v_cnt;
  PyObject *__pyx_v_feature = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  PyObject *__pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  __Pyx_RefNannySetupContext("has_no_valid_features", 0);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":231
 *         object feature
 * 
 *     n = PyList_GET_SIZE(features)             # <<<<<<<<<<<<<<
 *     cnt = 0
 *     for i in range(n):
 */
  __pyx_v_n = PyList_GET_SIZE(__pyx_v_features);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":232
 * 
 *     n = PyList_GET_SIZE(features)
 *     cnt = 0             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         feature = <object>PyList_GET_ITEM(features, i)
 */
  __pyx_v_cnt = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":233
 *     n = PyList_GET_SIZE(features)
 *     cnt = 0
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         feature = <object>PyList_GET_ITEM(features, i)
 *         if feature is None:
 */
  __pyx_t_1 = __pyx_v_n;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":234
 *     cnt = 0
 *     for i in range(n):
 *         feature = <object>PyList_GET_ITEM(features, i)             # <<<<<<<<<<<<<<
 *         if feature is None:
 *             cnt += 1
 */
    __pyx_t_4 = PyList_GET_ITEM(__pyx_v_features, __pyx_v_i);
    __pyx_t_5 = ((PyObject *)__pyx_t_4);
    __Pyx_INCREF(__pyx_t_5);
    __Pyx_XDECREF_SET(__pyx_v_feature, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":235
 *     for i in range(n):
 *         feature = <object>PyList_GET_ITEM(features, i)
 *         if feature is None:             # <<<<<<<<<<<<<<
 *             cnt += 1
 *     return cnt == n
 */
    __pyx_t_6 = (__pyx_v_feature == Py_None);
    __pyx_t_7 = (__pyx_t_6 != 0);
    if (__pyx_t_7) {

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":236
 *         feature = <object>PyList_GET_ITEM(features, i)
 *         if feature is None:
 *             cnt += 1             # <<<<<<<<<<<<<<
 *     return cnt == n
 * 
 */
      __pyx_v_cnt = (__pyx_v_cnt + 1);

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":235
 *     for i in range(n):
 *         feature = <object>PyList_GET_ITEM(features, i)
 *         if feature is None:             # <<<<<<<<<<<<<<
 *             cnt += 1
 *     return cnt == n
 */
    }
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":237
 *         if feature is None:
 *             cnt += 1
 *     return cnt == n             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = (__pyx_v_cnt == __pyx_v_n);
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":226
 * 
 * 
 * cdef int has_no_valid_features(list features):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i, n, cnt
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_feature);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":246
 *         public AveragineCache averagine
 * 
 *     cpdef list create_theoretical_distribution(self, double mz, int charge, double charge_carrier=PROTON, double truncate_after=0.8,             # <<<<<<<<<<<<<<
 *                                                double ignore_below=0.05):
 *         cdef:
 */

static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_1create_theoretical_distribution(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_create_theoretical_distribution(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, double __pyx_v_mz, int __pyx_v_charge, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_create_theoretical_distribution *__pyx_optional_args) {
  double __pyx_v_charge_carrier = __pyx_k__3;
  double __pyx_v_truncate_after = ((double)0.8);
  double __pyx_v_ignore_below = ((double)0.05);
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *__pyx_v_base_tid = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_14AveragineCache_isotopic_cluster __pyx_t_12;
  __Pyx_RefNannySetupContext("create_theoretical_distribution", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_charge_carrier = __pyx_optional_args->charge_carrier;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_truncate_after = __pyx_optional_args->truncate_after;
        if (__pyx_optional_args->__pyx_n > 2) {
          __pyx_v_ignore_below = __pyx_optional_args->ignore_below;
        }
      }
    }
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_create_theoretical_distribution); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 246, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_1create_theoretical_distribution)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_mz); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 246, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_charge); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 246, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_5 = PyFloat_FromDouble(__pyx_v_charge_carrier); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 246, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_6 = PyFloat_FromDouble(__pyx_v_truncate_after); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 246, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_7 = PyFloat_FromDouble(__pyx_v_ignore_below); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 246, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_8 = __pyx_t_1; __pyx_t_9 = NULL;
        __pyx_t_10 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
          __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
          if (likely(__pyx_t_9)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
            __Pyx_INCREF(__pyx_t_9);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_8, function);
            __pyx_t_10 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_8)) {
          PyObject *__pyx_temp[6] = {__pyx_t_9, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6, __pyx_t_7};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_10, 5+__pyx_t_10); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 246, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_8)) {
          PyObject *__pyx_temp[6] = {__pyx_t_9, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6, __pyx_t_7};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_10, 5+__pyx_t_10); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 246, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        } else
        #endif
        {
          __pyx_t_11 = PyTuple_New(5+__pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 246, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_11);
          if (__pyx_t_9) {
            __Pyx_GIVEREF(__pyx_t_9); PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_9); __pyx_t_9 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_11, 0+__pyx_t_10, __pyx_t_3);
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_11, 1+__pyx_t_10, __pyx_t_4);
          __Pyx_GIVEREF(__pyx_t_5);
          PyTuple_SET_ITEM(__pyx_t_11, 2+__pyx_t_10, __pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_6);
          PyTuple_SET_ITEM(__pyx_t_11, 3+__pyx_t_10, __pyx_t_6);
          __Pyx_GIVEREF(__pyx_t_7);
          PyTuple_SET_ITEM(__pyx_t_11, 4+__pyx_t_10, __pyx_t_7);
          __pyx_t_3 = 0;
          __pyx_t_4 = 0;
          __pyx_t_5 = 0;
          __pyx_t_6 = 0;
          __pyx_t_7 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_11, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 246, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        }
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 246, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":250
 *         cdef:
 *             TheoreticalIsotopicPattern base_tid
 *         base_tid = self.averagine.isotopic_cluster(             # <<<<<<<<<<<<<<
 *             mz, charge, truncate_after=truncate_after, ignore_below=ignore_below,
 *             charge_carrier=charge_carrier)
 */
  __pyx_t_12.__pyx_n = 4;
  __pyx_t_12.charge = __pyx_v_charge;
  __pyx_t_12.charge_carrier = __pyx_v_charge_carrier;
  __pyx_t_12.truncate_after = __pyx_v_truncate_after;
  __pyx_t_12.ignore_below = __pyx_v_ignore_below;
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_AveragineCache *)__pyx_v_self->averagine->__pyx_vtab)->isotopic_cluster(__pyx_v_self->averagine, __pyx_v_mz, 0, &__pyx_t_12)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 250, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_base_tid = ((struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":253
 *             mz, charge, truncate_after=truncate_after, ignore_below=ignore_below,
 *             charge_carrier=charge_carrier)
 *         return base_tid             # <<<<<<<<<<<<<<
 * 
 *     cpdef list find_all_features(self, double mz, double error_tolerance=2e-5):
 */
  __Pyx_XDECREF(__pyx_r);
  if (!(likely(PyList_CheckExact(((PyObject *)__pyx_v_base_tid)))||((((PyObject *)__pyx_v_base_tid)) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(((PyObject *)__pyx_v_base_tid))->tp_name), 0))) __PYX_ERR(0, 253, __pyx_L1_error)
  __Pyx_INCREF(((PyObject *)__pyx_v_base_tid));
  __pyx_r = ((PyObject*)__pyx_v_base_tid);
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":246
 *         public AveragineCache averagine
 * 
 *     cpdef list create_theoretical_distribution(self, double mz, int charge, double charge_carrier=PROTON, double truncate_after=0.8,             # <<<<<<<<<<<<<<
 *                                                double ignore_below=0.05):
 *         cdef:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase.create_theoretical_distribution", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_base_tid);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_1create_theoretical_distribution(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_create_theoretical_distribution[] = "LCMSFeatureProcessorBase.create_theoretical_distribution(self, double mz, int charge, double charge_carrier=PROTON, double truncate_after=0.8, double ignore_below=0.05) -> list";
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_1create_theoretical_distribution(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_mz;
  int __pyx_v_charge;
  double __pyx_v_charge_carrier;
  double __pyx_v_truncate_after;
  double __pyx_v_ignore_below;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("create_theoretical_distribution (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_mz,&__pyx_n_s_charge,&__pyx_n_s_charge_carrier,&__pyx_n_s_truncate_after,&__pyx_n_s_ignore_below,0};
    PyObject* values[5] = {0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mz)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_charge)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("create_theoretical_distribution", 0, 2, 5, 1); __PYX_ERR(0, 246, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_charge_carrier);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_truncate_after);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ignore_below);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "create_theoretical_distribution") < 0)) __PYX_ERR(0, 246, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_mz = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_mz == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 246, __pyx_L3_error)
    __pyx_v_charge = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_charge == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 246, __pyx_L3_error)
    if (values[2]) {
      __pyx_v_charge_carrier = __pyx_PyFloat_AsDouble(values[2]); if (unlikely((__pyx_v_charge_carrier == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 246, __pyx_L3_error)
    } else {
      __pyx_v_charge_carrier = __pyx_k__3;
    }
    if (values[3]) {
      __pyx_v_truncate_after = __pyx_PyFloat_AsDouble(values[3]); if (unlikely((__pyx_v_truncate_after == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 246, __pyx_L3_error)
    } else {
      __pyx_v_truncate_after = ((double)0.8);
    }
    if (values[4]) {
      __pyx_v_ignore_below = __pyx_PyFloat_AsDouble(values[4]); if (unlikely((__pyx_v_ignore_below == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 247, __pyx_L3_error)
    } else {
      __pyx_v_ignore_below = ((double)0.05);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("create_theoretical_distribution", 0, 2, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 246, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase.create_theoretical_distribution", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_create_theoretical_distribution(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)__pyx_v_self), __pyx_v_mz, __pyx_v_charge, __pyx_v_charge_carrier, __pyx_v_truncate_after, __pyx_v_ignore_below);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_create_theoretical_distribution(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, double __pyx_v_mz, int __pyx_v_charge, double __pyx_v_charge_carrier, double __pyx_v_truncate_after, double __pyx_v_ignore_below) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_create_theoretical_distribution __pyx_t_2;
  __Pyx_RefNannySetupContext("create_theoretical_distribution", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 3;
  __pyx_t_2.charge_carrier = __pyx_v_charge_carrier;
  __pyx_t_2.truncate_after = __pyx_v_truncate_after;
  __pyx_t_2.ignore_below = __pyx_v_ignore_below;
  __pyx_t_1 = __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase->create_theoretical_distribution(__pyx_v_self, __pyx_v_mz, __pyx_v_charge, 1, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase.create_theoretical_distribution", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":255
 *         return base_tid
 * 
 *     cpdef list find_all_features(self, double mz, double error_tolerance=2e-5):             # <<<<<<<<<<<<<<
 *         return self.feature_map._find_all(mz, error_tolerance)
 * 
 */

static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_3find_all_features(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_find_all_features(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, double __pyx_v_mz, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_find_all_features *__pyx_optional_args) {
  double __pyx_v_error_tolerance = ((double)2e-5);
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("find_all_features", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_error_tolerance = __pyx_optional_args->error_tolerance;
    }
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_find_all_features); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 255, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_3find_all_features)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_mz); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 255, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_error_tolerance); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 255, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_5 = __pyx_t_1; __pyx_t_6 = NULL;
        __pyx_t_7 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_6)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_6);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
            __pyx_t_7 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_5)) {
          PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_3, __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 255, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
          PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_3, __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 255, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 255, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          if (__pyx_t_6) {
            __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_t_3);
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_t_4);
          __pyx_t_3 = 0;
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 255, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        }
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 255, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":256
 * 
 *     cpdef list find_all_features(self, double mz, double error_tolerance=2e-5):
 *         return self.feature_map._find_all(mz, error_tolerance)             # <<<<<<<<<<<<<<
 * 
 *     cpdef list find_features(self, double mz, double error_tolerance=2e-5, LCMSFeature interval=None):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *)__pyx_v_self->feature_map->__pyx_vtab)->_find_all(__pyx_v_self->feature_map, __pyx_v_mz, __pyx_v_error_tolerance, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 256, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":255
 *         return base_tid
 * 
 *     cpdef list find_all_features(self, double mz, double error_tolerance=2e-5):             # <<<<<<<<<<<<<<
 *         return self.feature_map._find_all(mz, error_tolerance)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase.find_all_features", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_3find_all_features(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_2find_all_features[] = "LCMSFeatureProcessorBase.find_all_features(self, double mz, double error_tolerance=2e-5) -> list";
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_3find_all_features(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_mz;
  double __pyx_v_error_tolerance;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("find_all_features (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_mz,&__pyx_n_s_error_tolerance,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mz)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_error_tolerance);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "find_all_features") < 0)) __PYX_ERR(0, 255, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_mz = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_mz == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 255, __pyx_L3_error)
    if (values[1]) {
      __pyx_v_error_tolerance = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_error_tolerance == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 255, __pyx_L3_error)
    } else {
      __pyx_v_error_tolerance = ((double)2e-5);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("find_all_features", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 255, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase.find_all_features", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_2find_all_features(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)__pyx_v_self), __pyx_v_mz, __pyx_v_error_tolerance);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_2find_all_features(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, double __pyx_v_mz, double __pyx_v_error_tolerance) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_find_all_features __pyx_t_2;
  __Pyx_RefNannySetupContext("find_all_features", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.error_tolerance = __pyx_v_error_tolerance;
  __pyx_t_1 = __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase->find_all_features(__pyx_v_self, __pyx_v_mz, 1, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase.find_all_features", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":258
 *         return self.feature_map._find_all(mz, error_tolerance)
 * 
 *     cpdef list find_features(self, double mz, double error_tolerance=2e-5, LCMSFeature interval=None):             # <<<<<<<<<<<<<<
 *         cdef:
 *             list f, overlapped
 */

static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_5find_features(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_find_features(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, double __pyx_v_mz, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_find_features *__pyx_optional_args) {
  double __pyx_v_error_tolerance = ((double)2e-5);
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_interval = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)Py_None);
  PyObject *__pyx_v_f = 0;
  PyObject *__pyx_v_overlapped = 0;
  double __pyx_v_search_width;
  double __pyx_v_hit_width;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_el = 0;
  size_t __pyx_v_i;
  size_t __pyx_v_n;
  CYTHON_UNUSED int __pyx_v_has_miss;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_find_all_features __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  size_t __pyx_t_12;
  size_t __pyx_t_13;
  size_t __pyx_t_14;
  PyObject *__pyx_t_15;
  int __pyx_t_16;
  __Pyx_RefNannySetupContext("find_features", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_error_tolerance = __pyx_optional_args->error_tolerance;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_interval = __pyx_optional_args->interval;
      }
    }
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_find_features); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 258, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_5find_features)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_mz); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 258, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_error_tolerance); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 258, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_5 = __pyx_t_1; __pyx_t_6 = NULL;
        __pyx_t_7 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_6)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_6);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
            __pyx_t_7 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_5)) {
          PyObject *__pyx_temp[4] = {__pyx_t_6, __pyx_t_3, __pyx_t_4, ((PyObject *)__pyx_v_interval)};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 3+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 258, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
          PyObject *__pyx_temp[4] = {__pyx_t_6, __pyx_t_3, __pyx_t_4, ((PyObject *)__pyx_v_interval)};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 3+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 258, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_8 = PyTuple_New(3+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 258, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          if (__pyx_t_6) {
            __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_t_3);
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_t_4);
          __Pyx_INCREF(((PyObject *)__pyx_v_interval));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_interval));
          PyTuple_SET_ITEM(__pyx_t_8, 2+__pyx_t_7, ((PyObject *)__pyx_v_interval));
          __pyx_t_3 = 0;
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 258, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        }
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 258, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":265
 *             size_t i, n
 *             bint has_miss
 *         has_miss = False             # <<<<<<<<<<<<<<
 *         f = self.find_all_features(mz, error_tolerance)
 *         if not f:
 */
  __pyx_v_has_miss = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":266
 *             bint has_miss
 *         has_miss = False
 *         f = self.find_all_features(mz, error_tolerance)             # <<<<<<<<<<<<<<
 *         if not f:
 *             f = [None]
 */
  __pyx_t_9.__pyx_n = 1;
  __pyx_t_9.error_tolerance = __pyx_v_error_tolerance;
  __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)__pyx_v_self->__pyx_vtab)->find_all_features(__pyx_v_self, __pyx_v_mz, 0, &__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_f = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":267
 *         has_miss = False
 *         f = self.find_all_features(mz, error_tolerance)
 *         if not f:             # <<<<<<<<<<<<<<
 *             f = [None]
 *             has_miss = True
 */
  __pyx_t_10 = (__pyx_v_f != Py_None)&&(PyList_GET_SIZE(__pyx_v_f) != 0);
  __pyx_t_11 = ((!__pyx_t_10) != 0);
  if (__pyx_t_11) {

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":268
 *         f = self.find_all_features(mz, error_tolerance)
 *         if not f:
 *             f = [None]             # <<<<<<<<<<<<<<
 *             has_miss = True
 *         elif interval is not None:
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 268, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyList_SET_ITEM(__pyx_t_1, 0, Py_None);
    __Pyx_DECREF_SET(__pyx_v_f, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":269
 *         if not f:
 *             f = [None]
 *             has_miss = True             # <<<<<<<<<<<<<<
 *         elif interval is not None:
 *             overlapped = []
 */
    __pyx_v_has_miss = 1;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":267
 *         has_miss = False
 *         f = self.find_all_features(mz, error_tolerance)
 *         if not f:             # <<<<<<<<<<<<<<
 *             f = [None]
 *             has_miss = True
 */
    goto __pyx_L3;
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":270
 *             f = [None]
 *             has_miss = True
 *         elif interval is not None:             # <<<<<<<<<<<<<<
 *             overlapped = []
 *             n = PyList_GET_SIZE(f)
 */
  __pyx_t_11 = (((PyObject *)__pyx_v_interval) != Py_None);
  __pyx_t_10 = (__pyx_t_11 != 0);
  if (__pyx_t_10) {

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":271
 *             has_miss = True
 *         elif interval is not None:
 *             overlapped = []             # <<<<<<<<<<<<<<
 *             n = PyList_GET_SIZE(f)
 *             search_width = interval.get_end_time() - interval.get_start_time()
 */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 271, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_overlapped = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":272
 *         elif interval is not None:
 *             overlapped = []
 *             n = PyList_GET_SIZE(f)             # <<<<<<<<<<<<<<
 *             search_width = interval.get_end_time() - interval.get_start_time()
 *             if search_width == 0:
 */
    __pyx_v_n = PyList_GET_SIZE(__pyx_v_f);

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":273
 *             overlapped = []
 *             n = PyList_GET_SIZE(f)
 *             search_width = interval.get_end_time() - interval.get_start_time()             # <<<<<<<<<<<<<<
 *             if search_width == 0:
 *                 return [None]
 */
    __pyx_v_search_width = (((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_interval->__pyx_base.__pyx_vtab)->__pyx_base.get_end_time(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_interval)) - ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_interval->__pyx_base.__pyx_vtab)->__pyx_base.get_start_time(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_interval)));

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":274
 *             n = PyList_GET_SIZE(f)
 *             search_width = interval.get_end_time() - interval.get_start_time()
 *             if search_width == 0:             # <<<<<<<<<<<<<<
 *                 return [None]
 *             for i in range(n):
 */
    __pyx_t_10 = ((__pyx_v_search_width == 0.0) != 0);
    if (__pyx_t_10) {

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":275
 *             search_width = interval.get_end_time() - interval.get_start_time()
 *             if search_width == 0:
 *                 return [None]             # <<<<<<<<<<<<<<
 *             for i in range(n):
 *                 el = <LCMSFeature>PyList_GET_ITEM(f, i)
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 275, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      PyList_SET_ITEM(__pyx_t_1, 0, Py_None);
      __pyx_r = ((PyObject*)__pyx_t_1);
      __pyx_t_1 = 0;
      goto __pyx_L0;

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":274
 *             n = PyList_GET_SIZE(f)
 *             search_width = interval.get_end_time() - interval.get_start_time()
 *             if search_width == 0:             # <<<<<<<<<<<<<<
 *                 return [None]
 *             for i in range(n):
 */
    }

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":276
 *             if search_width == 0:
 *                 return [None]
 *             for i in range(n):             # <<<<<<<<<<<<<<
 *                 el = <LCMSFeature>PyList_GET_ITEM(f, i)
 *                 if el.overlaps_in_time(interval):
 */
    __pyx_t_12 = __pyx_v_n;
    __pyx_t_13 = __pyx_t_12;
    for (__pyx_t_14 = 0; __pyx_t_14 < __pyx_t_13; __pyx_t_14+=1) {
      __pyx_v_i = __pyx_t_14;

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":277
 *                 return [None]
 *             for i in range(n):
 *                 el = <LCMSFeature>PyList_GET_ITEM(f, i)             # <<<<<<<<<<<<<<
 *                 if el.overlaps_in_time(interval):
 *                     hit_width = el.get_end_time() - el.get_start_time()
 */
      __pyx_t_15 = PyList_GET_ITEM(__pyx_v_f, __pyx_v_i);
      __pyx_t_1 = ((PyObject *)__pyx_t_15);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_el, ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":278
 *             for i in range(n):
 *                 el = <LCMSFeature>PyList_GET_ITEM(f, i)
 *                 if el.overlaps_in_time(interval):             # <<<<<<<<<<<<<<
 *                     hit_width = el.get_end_time() - el.get_start_time()
 *                     if (hit_width / search_width) < 0.05:
 */
      __pyx_t_10 = (((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_el->__pyx_base.__pyx_vtab)->overlaps_in_time(__pyx_v_el, ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_interval), 0) != 0);
      if (__pyx_t_10) {

        /* "ms_deisotope/_c/feature_map/feature_processor.pyx":279
 *                 el = <LCMSFeature>PyList_GET_ITEM(f, i)
 *                 if el.overlaps_in_time(interval):
 *                     hit_width = el.get_end_time() - el.get_start_time()             # <<<<<<<<<<<<<<
 *                     if (hit_width / search_width) < 0.05:
 *                         continue
 */
        __pyx_v_hit_width = (((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_el->__pyx_base.__pyx_vtab)->__pyx_base.get_end_time(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_el)) - ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_el->__pyx_base.__pyx_vtab)->__pyx_base.get_start_time(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_el)));

        /* "ms_deisotope/_c/feature_map/feature_processor.pyx":280
 *                 if el.overlaps_in_time(interval):
 *                     hit_width = el.get_end_time() - el.get_start_time()
 *                     if (hit_width / search_width) < 0.05:             # <<<<<<<<<<<<<<
 *                         continue
 *                     overlapped.append(el)
 */
        if (unlikely(__pyx_v_search_width == 0)) {
          PyErr_SetString(PyExc_ZeroDivisionError, "float division");
          __PYX_ERR(0, 280, __pyx_L1_error)
        }
        __pyx_t_10 = (((__pyx_v_hit_width / __pyx_v_search_width) < 0.05) != 0);
        if (__pyx_t_10) {

          /* "ms_deisotope/_c/feature_map/feature_processor.pyx":281
 *                     hit_width = el.get_end_time() - el.get_start_time()
 *                     if (hit_width / search_width) < 0.05:
 *                         continue             # <<<<<<<<<<<<<<
 *                     overlapped.append(el)
 *             f = overlapped
 */
          goto __pyx_L5_continue;

          /* "ms_deisotope/_c/feature_map/feature_processor.pyx":280
 *                 if el.overlaps_in_time(interval):
 *                     hit_width = el.get_end_time() - el.get_start_time()
 *                     if (hit_width / search_width) < 0.05:             # <<<<<<<<<<<<<<
 *                         continue
 *                     overlapped.append(el)
 */
        }

        /* "ms_deisotope/_c/feature_map/feature_processor.pyx":282
 *                     if (hit_width / search_width) < 0.05:
 *                         continue
 *                     overlapped.append(el)             # <<<<<<<<<<<<<<
 *             f = overlapped
 *             if PyList_GET_SIZE(f) == 0:
 */
        __pyx_t_16 = __Pyx_PyList_Append(__pyx_v_overlapped, ((PyObject *)__pyx_v_el)); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(0, 282, __pyx_L1_error)

        /* "ms_deisotope/_c/feature_map/feature_processor.pyx":278
 *             for i in range(n):
 *                 el = <LCMSFeature>PyList_GET_ITEM(f, i)
 *                 if el.overlaps_in_time(interval):             # <<<<<<<<<<<<<<
 *                     hit_width = el.get_end_time() - el.get_start_time()
 *                     if (hit_width / search_width) < 0.05:
 */
      }
      __pyx_L5_continue:;
    }

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":283
 *                         continue
 *                     overlapped.append(el)
 *             f = overlapped             # <<<<<<<<<<<<<<
 *             if PyList_GET_SIZE(f) == 0:
 *                 f = [None]
 */
    __Pyx_INCREF(__pyx_v_overlapped);
    __Pyx_DECREF_SET(__pyx_v_f, __pyx_v_overlapped);

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":284
 *                     overlapped.append(el)
 *             f = overlapped
 *             if PyList_GET_SIZE(f) == 0:             # <<<<<<<<<<<<<<
 *                 f = [None]
 *         return f
 */
    __pyx_t_10 = ((PyList_GET_SIZE(__pyx_v_f) == 0) != 0);
    if (__pyx_t_10) {

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":285
 *             f = overlapped
 *             if PyList_GET_SIZE(f) == 0:
 *                 f = [None]             # <<<<<<<<<<<<<<
 *         return f
 * 
 */
      __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 285, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      PyList_SET_ITEM(__pyx_t_1, 0, Py_None);
      __Pyx_DECREF_SET(__pyx_v_f, ((PyObject*)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":284
 *                     overlapped.append(el)
 *             f = overlapped
 *             if PyList_GET_SIZE(f) == 0:             # <<<<<<<<<<<<<<
 *                 f = [None]
 *         return f
 */
    }

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":270
 *             f = [None]
 *             has_miss = True
 *         elif interval is not None:             # <<<<<<<<<<<<<<
 *             overlapped = []
 *             n = PyList_GET_SIZE(f)
 */
  }
  __pyx_L3:;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":286
 *             if PyList_GET_SIZE(f) == 0:
 *                 f = [None]
 *         return f             # <<<<<<<<<<<<<<
 * 
 *     cpdef list match_theoretical_isotopic_distribution(self, list theoretical_distribution, double error_tolerance=2e-5, LCMSFeature interval=None):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_f);
  __pyx_r = __pyx_v_f;
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":258
 *         return self.feature_map._find_all(mz, error_tolerance)
 * 
 *     cpdef list find_features(self, double mz, double error_tolerance=2e-5, LCMSFeature interval=None):             # <<<<<<<<<<<<<<
 *         cdef:
 *             list f, overlapped
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase.find_features", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_f);
  __Pyx_XDECREF(__pyx_v_overlapped);
  __Pyx_XDECREF((PyObject *)__pyx_v_el);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_5find_features(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_4find_features[] = "LCMSFeatureProcessorBase.find_features(self, double mz, double error_tolerance=2e-5, LCMSFeature interval=None) -> list";
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_5find_features(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_mz;
  double __pyx_v_error_tolerance;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_interval = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("find_features (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_mz,&__pyx_n_s_error_tolerance,&__pyx_n_s_interval,0};
    PyObject* values[3] = {0,0,0};
    values[2] = (PyObject *)((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mz)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_error_tolerance);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_interval);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "find_features") < 0)) __PYX_ERR(0, 258, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_mz = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_mz == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 258, __pyx_L3_error)
    if (values[1]) {
      __pyx_v_error_tolerance = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_error_tolerance == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 258, __pyx_L3_error)
    } else {
      __pyx_v_error_tolerance = ((double)2e-5);
    }
    __pyx_v_interval = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)values[2]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("find_features", 0, 1, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 258, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase.find_features", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_interval), __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature, 1, "interval", 0))) __PYX_ERR(0, 258, __pyx_L1_error)
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_4find_features(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)__pyx_v_self), __pyx_v_mz, __pyx_v_error_tolerance, __pyx_v_interval);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_4find_features(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, double __pyx_v_mz, double __pyx_v_error_tolerance, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_interval) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_find_features __pyx_t_2;
  __Pyx_RefNannySetupContext("find_features", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 2;
  __pyx_t_2.error_tolerance = __pyx_v_error_tolerance;
  __pyx_t_2.interval = __pyx_v_interval;
  __pyx_t_1 = __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase->find_features(__pyx_v_self, __pyx_v_mz, 1, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase.find_features", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":288
 *         return f
 * 
 *     cpdef list match_theoretical_isotopic_distribution(self, list theoretical_distribution, double error_tolerance=2e-5, LCMSFeature interval=None):             # <<<<<<<<<<<<<<
 *         """Given a list of theoretical peaks, find their counterparts in :attr:`peaklist` within `error_tolerance`
 *         ppm error. If no experimental peak is found, a placeholder will be used in its stead.
 */

static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_7match_theoretical_isotopic_distribution(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_match_theoretical_isotopic_distribution(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, PyObject *__pyx_v_theoretical_distribution, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_match_theoretical_isotopic_distribution *__pyx_optional_args) {
  double __pyx_v_error_tolerance = ((double)2e-5);
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_interval = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)Py_None);
  PyObject *__pyx_v_experimental_distribution = 0;
  PyObject *__pyx_v_found = 0;
  size_t __pyx_v_i;
  size_t __pyx_v_n;
  struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_v_p = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  size_t __pyx_t_8;
  size_t __pyx_t_9;
  size_t __pyx_t_10;
  PyObject *__pyx_t_11;
  struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_find_features __pyx_t_12;
  int __pyx_t_13;
  __Pyx_RefNannySetupContext("match_theoretical_isotopic_distribution", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_error_tolerance = __pyx_optional_args->error_tolerance;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_interval = __pyx_optional_args->interval;
      }
    }
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_match_theoretical_isotopic_distr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 288, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_7match_theoretical_isotopic_distribution)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_error_tolerance); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 288, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_theoretical_distribution, __pyx_t_3, ((PyObject *)__pyx_v_interval)};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 288, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_theoretical_distribution, __pyx_t_3, ((PyObject *)__pyx_v_interval)};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 288, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 288, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(__pyx_v_theoretical_distribution);
          __Pyx_GIVEREF(__pyx_v_theoretical_distribution);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_theoretical_distribution);
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __Pyx_INCREF(((PyObject *)__pyx_v_interval));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_interval));
          PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, ((PyObject *)__pyx_v_interval));
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 288, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 288, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":309
 *             TheoreticalPeak p
 * 
 *         experimental_distribution = []             # <<<<<<<<<<<<<<
 *         n = PyList_GET_SIZE(theoretical_distribution)
 *         for i in range(n):
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 309, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_experimental_distribution = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":310
 * 
 *         experimental_distribution = []
 *         n = PyList_GET_SIZE(theoretical_distribution)             # <<<<<<<<<<<<<<
 *         for i in range(n):
 *             p = <TheoreticalPeak>PyList_GET_ITEM(theoretical_distribution, i)
 */
  __pyx_v_n = PyList_GET_SIZE(__pyx_v_theoretical_distribution);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":311
 *         experimental_distribution = []
 *         n = PyList_GET_SIZE(theoretical_distribution)
 *         for i in range(n):             # <<<<<<<<<<<<<<
 *             p = <TheoreticalPeak>PyList_GET_ITEM(theoretical_distribution, i)
 *             found = self.find_features(p.mz, error_tolerance, interval=interval)
 */
  __pyx_t_8 = __pyx_v_n;
  __pyx_t_9 = __pyx_t_8;
  for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_9; __pyx_t_10+=1) {
    __pyx_v_i = __pyx_t_10;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":312
 *         n = PyList_GET_SIZE(theoretical_distribution)
 *         for i in range(n):
 *             p = <TheoreticalPeak>PyList_GET_ITEM(theoretical_distribution, i)             # <<<<<<<<<<<<<<
 *             found = self.find_features(p.mz, error_tolerance, interval=interval)
 *             experimental_distribution.append(found)
 */
    __pyx_t_11 = PyList_GET_ITEM(__pyx_v_theoretical_distribution, __pyx_v_i);
    __pyx_t_1 = ((PyObject *)__pyx_t_11);
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_p, ((struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":313
 *         for i in range(n):
 *             p = <TheoreticalPeak>PyList_GET_ITEM(theoretical_distribution, i)
 *             found = self.find_features(p.mz, error_tolerance, interval=interval)             # <<<<<<<<<<<<<<
 *             experimental_distribution.append(found)
 *         return experimental_distribution
 */
    __pyx_t_12.__pyx_n = 2;
    __pyx_t_12.error_tolerance = __pyx_v_error_tolerance;
    __pyx_t_12.interval = __pyx_v_interval;
    __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)__pyx_v_self->__pyx_vtab)->find_features(__pyx_v_self, __pyx_v_p->mz, 0, &__pyx_t_12); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 313, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_found, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":314
 *             p = <TheoreticalPeak>PyList_GET_ITEM(theoretical_distribution, i)
 *             found = self.find_features(p.mz, error_tolerance, interval=interval)
 *             experimental_distribution.append(found)             # <<<<<<<<<<<<<<
 *         return experimental_distribution
 * 
 */
    __pyx_t_13 = __Pyx_PyList_Append(__pyx_v_experimental_distribution, __pyx_v_found); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 314, __pyx_L1_error)
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":315
 *             found = self.find_features(p.mz, error_tolerance, interval=interval)
 *             experimental_distribution.append(found)
 *         return experimental_distribution             # <<<<<<<<<<<<<<
 * 
 *     cpdef tuple conform_envelopes(self, list experimental, list base_theoretical):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_experimental_distribution);
  __pyx_r = __pyx_v_experimental_distribution;
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":288
 *         return f
 * 
 *     cpdef list match_theoretical_isotopic_distribution(self, list theoretical_distribution, double error_tolerance=2e-5, LCMSFeature interval=None):             # <<<<<<<<<<<<<<
 *         """Given a list of theoretical peaks, find their counterparts in :attr:`peaklist` within `error_tolerance`
 *         ppm error. If no experimental peak is found, a placeholder will be used in its stead.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase.match_theoretical_isotopic_distribution", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_experimental_distribution);
  __Pyx_XDECREF(__pyx_v_found);
  __Pyx_XDECREF((PyObject *)__pyx_v_p);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_7match_theoretical_isotopic_distribution(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_6match_theoretical_isotopic_distribution[] = "LCMSFeatureProcessorBase.match_theoretical_isotopic_distribution(self, list theoretical_distribution, double error_tolerance=2e-5, LCMSFeature interval=None) -> list\nGiven a list of theoretical peaks, find their counterparts in :attr:`peaklist` within `error_tolerance`\n        ppm error. If no experimental peak is found, a placeholder will be used in its stead.\n\n        Parameters\n        ----------\n        theoretical_distribution : list of TheoreticalPeak\n            The theoretical isotopic pattern to match\n        error_tolerance : float, optional\n            Parts-per-million error tolerance to permit in searching for matches\n\n        Returns\n        -------\n        list of list of LCMSFeature\n            The list of matched features for each theoretical m/z\n        ";
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_7match_theoretical_isotopic_distribution(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_theoretical_distribution = 0;
  double __pyx_v_error_tolerance;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_interval = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("match_theoretical_isotopic_distribution (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_theoretical_distribution,&__pyx_n_s_error_tolerance,&__pyx_n_s_interval,0};
    PyObject* values[3] = {0,0,0};
    values[2] = (PyObject *)((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_theoretical_distribution)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_error_tolerance);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_interval);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "match_theoretical_isotopic_distribution") < 0)) __PYX_ERR(0, 288, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_theoretical_distribution = ((PyObject*)values[0]);
    if (values[1]) {
      __pyx_v_error_tolerance = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_error_tolerance == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 288, __pyx_L3_error)
    } else {
      __pyx_v_error_tolerance = ((double)2e-5);
    }
    __pyx_v_interval = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)values[2]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("match_theoretical_isotopic_distribution", 0, 1, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 288, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase.match_theoretical_isotopic_distribution", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_theoretical_distribution), (&PyList_Type), 1, "theoretical_distribution", 1))) __PYX_ERR(0, 288, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_interval), __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature, 1, "interval", 0))) __PYX_ERR(0, 288, __pyx_L1_error)
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_6match_theoretical_isotopic_distribution(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)__pyx_v_self), __pyx_v_theoretical_distribution, __pyx_v_error_tolerance, __pyx_v_interval);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_6match_theoretical_isotopic_distribution(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, PyObject *__pyx_v_theoretical_distribution, double __pyx_v_error_tolerance, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_interval) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_match_theoretical_isotopic_distribution __pyx_t_2;
  __Pyx_RefNannySetupContext("match_theoretical_isotopic_distribution", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 2;
  __pyx_t_2.error_tolerance = __pyx_v_error_tolerance;
  __pyx_t_2.interval = __pyx_v_interval;
  __pyx_t_1 = __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase->match_theoretical_isotopic_distribution(__pyx_v_self, __pyx_v_theoretical_distribution, 1, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase.match_theoretical_isotopic_distribution", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":317
 *         return experimental_distribution
 * 
 *     cpdef tuple conform_envelopes(self, list experimental, list base_theoretical):             # <<<<<<<<<<<<<<
 *         return conform_envelopes(experimental, base_theoretical)
 * 
 */

static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_9conform_envelopes(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_conform_envelopes(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, PyObject *__pyx_v_experimental, PyObject *__pyx_v_base_theoretical, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("conform_envelopes", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_conform_envelopes); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 317, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_9conform_envelopes)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_experimental, __pyx_v_base_theoretical};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 317, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_experimental, __pyx_v_base_theoretical};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 317, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 317, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          if (__pyx_t_4) {
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
          }
          __Pyx_INCREF(__pyx_v_experimental);
          __Pyx_GIVEREF(__pyx_v_experimental);
          PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_v_experimental);
          __Pyx_INCREF(__pyx_v_base_theoretical);
          __Pyx_GIVEREF(__pyx_v_base_theoretical);
          PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_v_base_theoretical);
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 317, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (!(likely(PyTuple_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 317, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":318
 * 
 *     cpdef tuple conform_envelopes(self, list experimental, list base_theoretical):
 *         return conform_envelopes(experimental, base_theoretical)             # <<<<<<<<<<<<<<
 * 
 *     cdef tuple _conform_envelopes(self, list experimental, list base_theoretical, size_t* n_missing):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_conform_envelopes(__pyx_v_experimental, __pyx_v_base_theoretical, 0, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":317
 *         return experimental_distribution
 * 
 *     cpdef tuple conform_envelopes(self, list experimental, list base_theoretical):             # <<<<<<<<<<<<<<
 *         return conform_envelopes(experimental, base_theoretical)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase.conform_envelopes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_9conform_envelopes(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_8conform_envelopes[] = "LCMSFeatureProcessorBase.conform_envelopes(self, list experimental, list base_theoretical) -> tuple";
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_9conform_envelopes(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_experimental = 0;
  PyObject *__pyx_v_base_theoretical = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("conform_envelopes (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_experimental,&__pyx_n_s_base_theoretical,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_experimental)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_base_theoretical)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("conform_envelopes", 1, 2, 2, 1); __PYX_ERR(0, 317, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "conform_envelopes") < 0)) __PYX_ERR(0, 317, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_experimental = ((PyObject*)values[0]);
    __pyx_v_base_theoretical = ((PyObject*)values[1]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("conform_envelopes", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 317, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase.conform_envelopes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_experimental), (&PyList_Type), 1, "experimental", 1))) __PYX_ERR(0, 317, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_base_theoretical), (&PyList_Type), 1, "base_theoretical", 1))) __PYX_ERR(0, 317, __pyx_L1_error)
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_8conform_envelopes(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)__pyx_v_self), __pyx_v_experimental, __pyx_v_base_theoretical);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_8conform_envelopes(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, PyObject *__pyx_v_experimental, PyObject *__pyx_v_base_theoretical) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("conform_envelopes", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_conform_envelopes(__pyx_v_self, __pyx_v_experimental, __pyx_v_base_theoretical, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 317, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase.conform_envelopes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":320
 *         return conform_envelopes(experimental, base_theoretical)
 * 
 *     cdef tuple _conform_envelopes(self, list experimental, list base_theoretical, size_t* n_missing):             # <<<<<<<<<<<<<<
 *         return _conform_envelopes(experimental, base_theoretical, n_missing)
 * 
 */

static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__conform_envelopes(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, PyObject *__pyx_v_experimental, PyObject *__pyx_v_base_theoretical, size_t *__pyx_v_n_missing) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("_conform_envelopes", 0);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":321
 * 
 *     cdef tuple _conform_envelopes(self, list experimental, list base_theoretical, size_t* n_missing):
 *         return _conform_envelopes(experimental, base_theoretical, n_missing)             # <<<<<<<<<<<<<<
 * 
 *     cdef double _find_thresholded_score(self, dvec* scores, double percentage):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor__conform_envelopes(__pyx_v_experimental, __pyx_v_base_theoretical, __pyx_v_n_missing, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":320
 *         return conform_envelopes(experimental, base_theoretical)
 * 
 *     cdef tuple _conform_envelopes(self, list experimental, list base_theoretical, size_t* n_missing):             # <<<<<<<<<<<<<<
 *         return _conform_envelopes(experimental, base_theoretical, n_missing)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase._conform_envelopes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":323
 *         return _conform_envelopes(experimental, base_theoretical, n_missing)
 * 
 *     cdef double _find_thresholded_score(self, dvec* scores, double percentage):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, n, count
 */

static double __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__find_thresholded_score(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector *__pyx_v_scores, double __pyx_v_percentage) {
  size_t __pyx_v_i;
  size_t __pyx_v_n;
  size_t __pyx_v_count;
  double __pyx_v_maximum;
  double __pyx_v_val;
  double __pyx_v_threshold;
  double __pyx_v_acc;
  double __pyx_r;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  int __pyx_t_4;
  __Pyx_RefNannySetupContext("_find_thresholded_score", 0);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":328
 *             double maximum, val, threshold, acc
 * 
 *         maximum = -INF             # <<<<<<<<<<<<<<
 *         count = 0
 *         acc = 0
 */
  __pyx_v_maximum = (-__pyx_v_12ms_deisotope_2_c_11feature_map_17feature_processor_INF);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":329
 * 
 *         maximum = -INF
 *         count = 0             # <<<<<<<<<<<<<<
 *         acc = 0
 *         n = scores.used
 */
  __pyx_v_count = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":330
 *         maximum = -INF
 *         count = 0
 *         acc = 0             # <<<<<<<<<<<<<<
 *         n = scores.used
 *         for i in range(n):
 */
  __pyx_v_acc = 0.0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":331
 *         count = 0
 *         acc = 0
 *         n = scores.used             # <<<<<<<<<<<<<<
 *         for i in range(n):
 *             val = scores.v[i]
 */
  __pyx_t_1 = __pyx_v_scores->used;
  __pyx_v_n = __pyx_t_1;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":332
 *         acc = 0
 *         n = scores.used
 *         for i in range(n):             # <<<<<<<<<<<<<<
 *             val = scores.v[i]
 *             if val > maximum:
 */
  __pyx_t_1 = __pyx_v_n;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":333
 *         n = scores.used
 *         for i in range(n):
 *             val = scores.v[i]             # <<<<<<<<<<<<<<
 *             if val > maximum:
 *                 maximum = val
 */
    __pyx_v_val = (__pyx_v_scores->v[__pyx_v_i]);

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":334
 *         for i in range(n):
 *             val = scores.v[i]
 *             if val > maximum:             # <<<<<<<<<<<<<<
 *                 maximum = val
 *         threshold = maximum * percentage
 */
    __pyx_t_4 = ((__pyx_v_val > __pyx_v_maximum) != 0);
    if (__pyx_t_4) {

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":335
 *             val = scores.v[i]
 *             if val > maximum:
 *                 maximum = val             # <<<<<<<<<<<<<<
 *         threshold = maximum * percentage
 *         for i in range(n):
 */
      __pyx_v_maximum = __pyx_v_val;

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":334
 *         for i in range(n):
 *             val = scores.v[i]
 *             if val > maximum:             # <<<<<<<<<<<<<<
 *                 maximum = val
 *         threshold = maximum * percentage
 */
    }
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":336
 *             if val > maximum:
 *                 maximum = val
 *         threshold = maximum * percentage             # <<<<<<<<<<<<<<
 *         for i in range(n):
 *             val = scores.v[i]
 */
  __pyx_v_threshold = (__pyx_v_maximum * __pyx_v_percentage);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":337
 *                 maximum = val
 *         threshold = maximum * percentage
 *         for i in range(n):             # <<<<<<<<<<<<<<
 *             val = scores.v[i]
 *             if val > threshold:
 */
  __pyx_t_1 = __pyx_v_n;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":338
 *         threshold = maximum * percentage
 *         for i in range(n):
 *             val = scores.v[i]             # <<<<<<<<<<<<<<
 *             if val > threshold:
 *                 acc += val
 */
    __pyx_v_val = (__pyx_v_scores->v[__pyx_v_i]);

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":339
 *         for i in range(n):
 *             val = scores.v[i]
 *             if val > threshold:             # <<<<<<<<<<<<<<
 *                 acc += val
 *                 count += 1
 */
    __pyx_t_4 = ((__pyx_v_val > __pyx_v_threshold) != 0);
    if (__pyx_t_4) {

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":340
 *             val = scores.v[i]
 *             if val > threshold:
 *                 acc += val             # <<<<<<<<<<<<<<
 *                 count += 1
 *         if count == 0:
 */
      __pyx_v_acc = (__pyx_v_acc + __pyx_v_val);

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":341
 *             if val > threshold:
 *                 acc += val
 *                 count += 1             # <<<<<<<<<<<<<<
 *         if count == 0:
 *             return 0
 */
      __pyx_v_count = (__pyx_v_count + 1);

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":339
 *         for i in range(n):
 *             val = scores.v[i]
 *             if val > threshold:             # <<<<<<<<<<<<<<
 *                 acc += val
 *                 count += 1
 */
    }
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":342
 *                 acc += val
 *                 count += 1
 *         if count == 0:             # <<<<<<<<<<<<<<
 *             return 0
 *         return acc / count
 */
  __pyx_t_4 = ((__pyx_v_count == 0) != 0);
  if (__pyx_t_4) {

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":343
 *                 count += 1
 *         if count == 0:
 *             return 0             # <<<<<<<<<<<<<<
 *         return acc / count
 * 
 */
    __pyx_r = 0.0;
    goto __pyx_L0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":342
 *                 acc += val
 *                 count += 1
 *         if count == 0:             # <<<<<<<<<<<<<<
 *             return 0
 *         return acc / count
 */
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":344
 *         if count == 0:
 *             return 0
 *         return acc / count             # <<<<<<<<<<<<<<
 * 
 *     cpdef LCMSFeatureSetFit _fit_single_feature_set(self, list features, list base_tid, double error_tolerance,
 */
  if (unlikely(__pyx_v_count == 0)) {
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    __PYX_ERR(0, 344, __pyx_L1_error)
  }
  __pyx_r = (__pyx_v_acc / __pyx_v_count);
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":323
 *         return _conform_envelopes(experimental, base_theoretical, n_missing)
 * 
 *     cdef double _find_thresholded_score(self, dvec* scores, double percentage):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, n, count
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase._find_thresholded_score", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":346
 *         return acc / count
 * 
 *     cpdef LCMSFeatureSetFit _fit_single_feature_set(self, list features, list base_tid, double error_tolerance,             # <<<<<<<<<<<<<<
 *                                                     int charge, double charge_carrier=PROTON, int max_missed_peaks=1,
 *                                                     double threshold_scale=0.3):
 */

static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_11_fit_single_feature_set(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_single_feature_set(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, PyObject *__pyx_v_features, PyObject *__pyx_v_base_tid, CYTHON_UNUSED double __pyx_v_error_tolerance, int __pyx_v_charge, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_single_feature_set *__pyx_optional_args) {
  double __pyx_v_charge_carrier = __pyx_k__4;
  int __pyx_v_max_missed_peaks = ((int)1);
  double __pyx_v_threshold_scale = ((double)0.3);
  double __pyx_v_score;
  double __pyx_v_final_score;
  double __pyx_v_score_acc;
  double __pyx_v_neutral_mass;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *__pyx_v_conformer = 0;
  struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector *__pyx_v_score_vec;
  struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector *__pyx_v_time_vec;
  PyArrayObject *__pyx_v_scores_array = 0;
  PyArrayObject *__pyx_v_times_array = 0;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_feat_iter = 0;
  size_t __pyx_v_feat_i;
  size_t __pyx_v_feat_n;
  size_t __pyx_v_n_missing;
  size_t __pyx_v_missing_features;
  size_t __pyx_v_counter;
  PyObject *__pyx_v_eid = NULL;
  PyObject *__pyx_v_cleaned_eid = NULL;
  PyObject *__pyx_v_tid = NULL;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_f = NULL;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit *__pyx_v_fit = NULL;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  int __pyx_t_13;
  size_t __pyx_t_14;
  size_t __pyx_t_15;
  size_t __pyx_t_16;
  PyObject *__pyx_t_17;
  double __pyx_t_18;
  struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_neutral_mass __pyx_t_19;
  __Pyx_RefNannySetupContext("_fit_single_feature_set", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_charge_carrier = __pyx_optional_args->charge_carrier;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_max_missed_peaks = __pyx_optional_args->max_missed_peaks;
        if (__pyx_optional_args->__pyx_n > 2) {
          __pyx_v_threshold_scale = __pyx_optional_args->threshold_scale;
        }
      }
    }
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_fit_single_feature_set); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 346, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_11_fit_single_feature_set)) {
        __Pyx_XDECREF(((PyObject *)__pyx_r));
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_error_tolerance); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 346, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_charge); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 346, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_5 = PyFloat_FromDouble(__pyx_v_charge_carrier); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 346, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_max_missed_peaks); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 346, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_7 = PyFloat_FromDouble(__pyx_v_threshold_scale); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 346, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_8 = __pyx_t_1; __pyx_t_9 = NULL;
        __pyx_t_10 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
          __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
          if (likely(__pyx_t_9)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
            __Pyx_INCREF(__pyx_t_9);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_8, function);
            __pyx_t_10 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_8)) {
          PyObject *__pyx_temp[8] = {__pyx_t_9, __pyx_v_features, __pyx_v_base_tid, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6, __pyx_t_7};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_10, 7+__pyx_t_10); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 346, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_8)) {
          PyObject *__pyx_temp[8] = {__pyx_t_9, __pyx_v_features, __pyx_v_base_tid, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6, __pyx_t_7};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_10, 7+__pyx_t_10); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 346, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        } else
        #endif
        {
          __pyx_t_11 = PyTuple_New(7+__pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 346, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_11);
          if (__pyx_t_9) {
            __Pyx_GIVEREF(__pyx_t_9); PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_9); __pyx_t_9 = NULL;
          }
          __Pyx_INCREF(__pyx_v_features);
          __Pyx_GIVEREF(__pyx_v_features);
          PyTuple_SET_ITEM(__pyx_t_11, 0+__pyx_t_10, __pyx_v_features);
          __Pyx_INCREF(__pyx_v_base_tid);
          __Pyx_GIVEREF(__pyx_v_base_tid);
          PyTuple_SET_ITEM(__pyx_t_11, 1+__pyx_t_10, __pyx_v_base_tid);
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_11, 2+__pyx_t_10, __pyx_t_3);
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_11, 3+__pyx_t_10, __pyx_t_4);
          __Pyx_GIVEREF(__pyx_t_5);
          PyTuple_SET_ITEM(__pyx_t_11, 4+__pyx_t_10, __pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_6);
          PyTuple_SET_ITEM(__pyx_t_11, 5+__pyx_t_10, __pyx_t_6);
          __Pyx_GIVEREF(__pyx_t_7);
          PyTuple_SET_ITEM(__pyx_t_11, 6+__pyx_t_10, __pyx_t_7);
          __pyx_t_3 = 0;
          __pyx_t_4 = 0;
          __pyx_t_5 = 0;
          __pyx_t_6 = 0;
          __pyx_t_7 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_11, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 346, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        }
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit))))) __PYX_ERR(0, 346, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":358
 *             size_t feat_i, feat_n, n_missing, missing_features, counter
 * 
 *         conformer = envelope_conformer._create()             # <<<<<<<<<<<<<<
 *         feat_iter = FeatureSetIterator._create(features)
 *         score_acc = 0
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer__create()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 358, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_conformer = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":359
 * 
 *         conformer = envelope_conformer._create()
 *         feat_iter = FeatureSetIterator._create(features)             # <<<<<<<<<<<<<<
 *         score_acc = 0
 *         counter = 0
 */
  __pyx_t_1 = ((PyObject *)__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator->_create(__pyx_v_features)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 359, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_feat_iter = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":360
 *         conformer = envelope_conformer._create()
 *         feat_iter = FeatureSetIterator._create(features)
 *         score_acc = 0             # <<<<<<<<<<<<<<
 *         counter = 0
 *         score_vec = make_double_vector()
 */
  __pyx_v_score_acc = 0.0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":361
 *         feat_iter = FeatureSetIterator._create(features)
 *         score_acc = 0
 *         counter = 0             # <<<<<<<<<<<<<<
 *         score_vec = make_double_vector()
 *         time_vec = make_double_vector()
 */
  __pyx_v_counter = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":362
 *         score_acc = 0
 *         counter = 0
 *         score_vec = make_double_vector()             # <<<<<<<<<<<<<<
 *         time_vec = make_double_vector()
 *         while feat_iter.has_more():
 */
  __pyx_v_score_vec = __pyx_f_7brainpy_2_c_13double_vector_make_double_vector();

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":363
 *         counter = 0
 *         score_vec = make_double_vector()
 *         time_vec = make_double_vector()             # <<<<<<<<<<<<<<
 *         while feat_iter.has_more():
 *             eid = feat_iter.get_next_value()
 */
  __pyx_v_time_vec = __pyx_f_7brainpy_2_c_13double_vector_make_double_vector();

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":364
 *         score_vec = make_double_vector()
 *         time_vec = make_double_vector()
 *         while feat_iter.has_more():             # <<<<<<<<<<<<<<
 *             eid = feat_iter.get_next_value()
 *             if eid is None:
 */
  while (1) {
    __pyx_t_12 = (((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *)__pyx_v_feat_iter->__pyx_vtab)->has_more(__pyx_v_feat_iter, 0) != 0);
    if (!__pyx_t_12) break;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":365
 *         time_vec = make_double_vector()
 *         while feat_iter.has_more():
 *             eid = feat_iter.get_next_value()             # <<<<<<<<<<<<<<
 *             if eid is None:
 *                 continue
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *)__pyx_v_feat_iter->__pyx_vtab)->get_next_value(__pyx_v_feat_iter, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 365, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_eid, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":366
 *         while feat_iter.has_more():
 *             eid = feat_iter.get_next_value()
 *             if eid is None:             # <<<<<<<<<<<<<<
 *                 continue
 *             conformer.acquire(eid, base_tid)
 */
    __pyx_t_12 = (__pyx_v_eid == ((PyObject*)Py_None));
    __pyx_t_13 = (__pyx_t_12 != 0);
    if (__pyx_t_13) {

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":367
 *             eid = feat_iter.get_next_value()
 *             if eid is None:
 *                 continue             # <<<<<<<<<<<<<<
 *             conformer.acquire(eid, base_tid)
 *             conformer.conform()
 */
      goto __pyx_L3_continue;

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":366
 *         while feat_iter.has_more():
 *             eid = feat_iter.get_next_value()
 *             if eid is None:             # <<<<<<<<<<<<<<
 *                 continue
 *             conformer.acquire(eid, base_tid)
 */
    }

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":368
 *             if eid is None:
 *                 continue
 *             conformer.acquire(eid, base_tid)             # <<<<<<<<<<<<<<
 *             conformer.conform()
 *             cleaned_eid = conformer.experimental
 */
    ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *)__pyx_v_conformer->__pyx_vtab)->acquire(__pyx_v_conformer, __pyx_v_eid, __pyx_v_base_tid);

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":369
 *                 continue
 *             conformer.acquire(eid, base_tid)
 *             conformer.conform()             # <<<<<<<<<<<<<<
 *             cleaned_eid = conformer.experimental
 *             tid = conformer.theoretical
 */
    ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *)__pyx_v_conformer->__pyx_vtab)->conform(__pyx_v_conformer, NULL);

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":370
 *             conformer.acquire(eid, base_tid)
 *             conformer.conform()
 *             cleaned_eid = conformer.experimental             # <<<<<<<<<<<<<<
 *             tid = conformer.theoretical
 *             n_missing = conformer.n_missing
 */
    __pyx_t_1 = __pyx_v_conformer->experimental;
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_cleaned_eid, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":371
 *             conformer.conform()
 *             cleaned_eid = conformer.experimental
 *             tid = conformer.theoretical             # <<<<<<<<<<<<<<
 *             n_missing = conformer.n_missing
 *             if n_missing > max_missed_peaks:
 */
    __pyx_t_1 = __pyx_v_conformer->theoretical;
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_tid, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":372
 *             cleaned_eid = conformer.experimental
 *             tid = conformer.theoretical
 *             n_missing = conformer.n_missing             # <<<<<<<<<<<<<<
 *             if n_missing > max_missed_peaks:
 *                 continue
 */
    __pyx_t_14 = __pyx_v_conformer->n_missing;
    __pyx_v_n_missing = __pyx_t_14;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":373
 *             tid = conformer.theoretical
 *             n_missing = conformer.n_missing
 *             if n_missing > max_missed_peaks:             # <<<<<<<<<<<<<<
 *                 continue
 *             score = self.scorer._evaluate(None, cleaned_eid, tid)
 */
    __pyx_t_13 = ((__pyx_v_n_missing > __pyx_v_max_missed_peaks) != 0);
    if (__pyx_t_13) {

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":374
 *             n_missing = conformer.n_missing
 *             if n_missing > max_missed_peaks:
 *                 continue             # <<<<<<<<<<<<<<
 *             score = self.scorer._evaluate(None, cleaned_eid, tid)
 *             if npy_isnan(score):
 */
      goto __pyx_L3_continue;

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":373
 *             tid = conformer.theoretical
 *             n_missing = conformer.n_missing
 *             if n_missing > max_missed_peaks:             # <<<<<<<<<<<<<<
 *                 continue
 *             score = self.scorer._evaluate(None, cleaned_eid, tid)
 */
    }

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":375
 *             if n_missing > max_missed_peaks:
 *                 continue
 *             score = self.scorer._evaluate(None, cleaned_eid, tid)             # <<<<<<<<<<<<<<
 *             if npy_isnan(score):
 *                 continue
 */
    __pyx_v_score = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *)__pyx_v_self->scorer->__pyx_vtab)->_evaluate(__pyx_v_self->scorer, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)Py_None), __pyx_v_cleaned_eid, __pyx_v_tid, 0);

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":376
 *                 continue
 *             score = self.scorer._evaluate(None, cleaned_eid, tid)
 *             if npy_isnan(score):             # <<<<<<<<<<<<<<
 *                 continue
 *             score_acc += score
 */
    __pyx_t_13 = (npy_isnan(__pyx_v_score) != 0);
    if (__pyx_t_13) {

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":377
 *             score = self.scorer._evaluate(None, cleaned_eid, tid)
 *             if npy_isnan(score):
 *                 continue             # <<<<<<<<<<<<<<
 *             score_acc += score
 *             counter += 1
 */
      goto __pyx_L3_continue;

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":376
 *                 continue
 *             score = self.scorer._evaluate(None, cleaned_eid, tid)
 *             if npy_isnan(score):             # <<<<<<<<<<<<<<
 *                 continue
 *             score_acc += score
 */
    }

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":378
 *             if npy_isnan(score):
 *                 continue
 *             score_acc += score             # <<<<<<<<<<<<<<
 *             counter += 1
 *             double_vector_append(score_vec, score)
 */
    __pyx_v_score_acc = (__pyx_v_score_acc + __pyx_v_score);

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":379
 *                 continue
 *             score_acc += score
 *             counter += 1             # <<<<<<<<<<<<<<
 *             double_vector_append(score_vec, score)
 *             double_vector_append(time_vec, feat_iter.get_current_time())
 */
    __pyx_v_counter = (__pyx_v_counter + 1);

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":380
 *             score_acc += score
 *             counter += 1
 *             double_vector_append(score_vec, score)             # <<<<<<<<<<<<<<
 *             double_vector_append(time_vec, feat_iter.get_current_time())
 *         # Compute feature score from score_vec
 */
    (void)(__pyx_f_7brainpy_2_c_13double_vector_double_vector_append(__pyx_v_score_vec, __pyx_v_score));

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":381
 *             counter += 1
 *             double_vector_append(score_vec, score)
 *             double_vector_append(time_vec, feat_iter.get_current_time())             # <<<<<<<<<<<<<<
 *         # Compute feature score from score_vec
 *         final_score = self._find_thresholded_score(score_vec, threshold_scale)
 */
    (void)(__pyx_f_7brainpy_2_c_13double_vector_double_vector_append(__pyx_v_time_vec, ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *)__pyx_v_feat_iter->__pyx_vtab)->get_current_time(__pyx_v_feat_iter, 0)));
    __pyx_L3_continue:;
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":383
 *             double_vector_append(time_vec, feat_iter.get_current_time())
 *         # Compute feature score from score_vec
 *         final_score = self._find_thresholded_score(score_vec, threshold_scale)             # <<<<<<<<<<<<<<
 *         missing_features = 0
 *         feat_n = PyList_GET_SIZE(features)
 */
  __pyx_v_final_score = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)__pyx_v_self->__pyx_vtab)->_find_thresholded_score(__pyx_v_self, __pyx_v_score_vec, __pyx_v_threshold_scale);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":384
 *         # Compute feature score from score_vec
 *         final_score = self._find_thresholded_score(score_vec, threshold_scale)
 *         missing_features = 0             # <<<<<<<<<<<<<<
 *         feat_n = PyList_GET_SIZE(features)
 *         for feat_i in range(feat_n):
 */
  __pyx_v_missing_features = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":385
 *         final_score = self._find_thresholded_score(score_vec, threshold_scale)
 *         missing_features = 0
 *         feat_n = PyList_GET_SIZE(features)             # <<<<<<<<<<<<<<
 *         for feat_i in range(feat_n):
 *             f = <LCMSFeature>PyList_GET_ITEM(features, feat_i)
 */
  __pyx_v_feat_n = PyList_GET_SIZE(__pyx_v_features);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":386
 *         missing_features = 0
 *         feat_n = PyList_GET_SIZE(features)
 *         for feat_i in range(feat_n):             # <<<<<<<<<<<<<<
 *             f = <LCMSFeature>PyList_GET_ITEM(features, feat_i)
 *             if f is None:
 */
  __pyx_t_14 = __pyx_v_feat_n;
  __pyx_t_15 = __pyx_t_14;
  for (__pyx_t_16 = 0; __pyx_t_16 < __pyx_t_15; __pyx_t_16+=1) {
    __pyx_v_feat_i = __pyx_t_16;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":387
 *         feat_n = PyList_GET_SIZE(features)
 *         for feat_i in range(feat_n):
 *             f = <LCMSFeature>PyList_GET_ITEM(features, feat_i)             # <<<<<<<<<<<<<<
 *             if f is None:
 *                 missing_features += 1
 */
    __pyx_t_17 = PyList_GET_ITEM(__pyx_v_features, __pyx_v_feat_i);
    __pyx_t_1 = ((PyObject *)__pyx_t_17);
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_f, ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":388
 *         for feat_i in range(feat_n):
 *             f = <LCMSFeature>PyList_GET_ITEM(features, feat_i)
 *             if f is None:             # <<<<<<<<<<<<<<
 *                 missing_features += 1
 *         f = <LCMSFeature>PyList_GET_ITEM(features, 0)
 */
    __pyx_t_13 = (((PyObject *)__pyx_v_f) == Py_None);
    __pyx_t_12 = (__pyx_t_13 != 0);
    if (__pyx_t_12) {

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":389
 *             f = <LCMSFeature>PyList_GET_ITEM(features, feat_i)
 *             if f is None:
 *                 missing_features += 1             # <<<<<<<<<<<<<<
 *         f = <LCMSFeature>PyList_GET_ITEM(features, 0)
 *         neutral_mass = calc_neutral_mass(
 */
      __pyx_v_missing_features = (__pyx_v_missing_features + 1);

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":388
 *         for feat_i in range(feat_n):
 *             f = <LCMSFeature>PyList_GET_ITEM(features, feat_i)
 *             if f is None:             # <<<<<<<<<<<<<<
 *                 missing_features += 1
 *         f = <LCMSFeature>PyList_GET_ITEM(features, 0)
 */
    }
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":390
 *             if f is None:
 *                 missing_features += 1
 *         f = <LCMSFeature>PyList_GET_ITEM(features, 0)             # <<<<<<<<<<<<<<
 *         neutral_mass = calc_neutral_mass(
 *                 f.get_mz(), charge, charge_carrier)
 */
  __pyx_t_17 = PyList_GET_ITEM(__pyx_v_features, 0);
  __pyx_t_1 = ((PyObject *)__pyx_t_17);
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_XDECREF_SET(__pyx_v_f, ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":391
 *                 missing_features += 1
 *         f = <LCMSFeature>PyList_GET_ITEM(features, 0)
 *         neutral_mass = calc_neutral_mass(             # <<<<<<<<<<<<<<
 *                 f.get_mz(), charge, charge_carrier)
 *         scores_array = double_vector_to_ndarray(score_vec)
 */
  __pyx_t_19.__pyx_n = 1;
  __pyx_t_19.charge_carrier = __pyx_v_charge_carrier;
  __pyx_t_18 = __pyx_f_12ms_deisotope_2_c_9averagine_neutral_mass(((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_f->__pyx_base.__pyx_vtab)->__pyx_base.get_mz(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_f)), __pyx_v_charge, &__pyx_t_19); 
  __pyx_v_neutral_mass = __pyx_t_18;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":393
 *         neutral_mass = calc_neutral_mass(
 *                 f.get_mz(), charge, charge_carrier)
 *         scores_array = double_vector_to_ndarray(score_vec)             # <<<<<<<<<<<<<<
 *         times_array = double_vector_to_ndarray(time_vec)
 *         fit = LCMSFeatureSetFit._create(
 */
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_double_vector_to_ndarray(__pyx_v_score_vec); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 393, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 393, __pyx_L1_error)
  __pyx_v_scores_array = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":394
 *                 f.get_mz(), charge, charge_carrier)
 *         scores_array = double_vector_to_ndarray(score_vec)
 *         times_array = double_vector_to_ndarray(time_vec)             # <<<<<<<<<<<<<<
 *         fit = LCMSFeatureSetFit._create(
 *             features, base_tid, final_score, charge, missing_features,
 */
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_double_vector_to_ndarray(__pyx_v_time_vec); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 394, __pyx_L1_error)
  __pyx_v_times_array = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":396
 *         times_array = double_vector_to_ndarray(time_vec)
 *         fit = LCMSFeatureSetFit._create(
 *             features, base_tid, final_score, charge, missing_features,             # <<<<<<<<<<<<<<
 *             [], None, neutral_mass, counter, scores_array, times_array)
 *         free_double_vector(score_vec)
 */
  if (!(likely(((__pyx_v_base_tid) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_base_tid, __pyx_ptype_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern))))) __PYX_ERR(0, 396, __pyx_L1_error)

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":397
 *         fit = LCMSFeatureSetFit._create(
 *             features, base_tid, final_score, charge, missing_features,
 *             [], None, neutral_mass, counter, scores_array, times_array)             # <<<<<<<<<<<<<<
 *         free_double_vector(score_vec)
 *         free_double_vector(time_vec)
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 397, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":395
 *         scores_array = double_vector_to_ndarray(score_vec)
 *         times_array = double_vector_to_ndarray(time_vec)
 *         fit = LCMSFeatureSetFit._create(             # <<<<<<<<<<<<<<
 *             features, base_tid, final_score, charge, missing_features,
 *             [], None, neutral_mass, counter, scores_array, times_array)
 */
  __pyx_t_2 = ((PyObject *)__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit->_create(__pyx_v_features, ((struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *)__pyx_v_base_tid), __pyx_v_final_score, __pyx_v_charge, __pyx_v_missing_features, ((PyObject*)__pyx_t_1), Py_None, __pyx_v_neutral_mass, __pyx_v_counter, __pyx_v_scores_array, __pyx_v_times_array)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 395, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_fit = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":398
 *             features, base_tid, final_score, charge, missing_features,
 *             [], None, neutral_mass, counter, scores_array, times_array)
 *         free_double_vector(score_vec)             # <<<<<<<<<<<<<<
 *         free_double_vector(time_vec)
 *         return fit
 */
  __pyx_f_7brainpy_2_c_13double_vector_free_double_vector(__pyx_v_score_vec);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":399
 *             [], None, neutral_mass, counter, scores_array, times_array)
 *         free_double_vector(score_vec)
 *         free_double_vector(time_vec)             # <<<<<<<<<<<<<<
 *         return fit
 * 
 */
  __pyx_f_7brainpy_2_c_13double_vector_free_double_vector(__pyx_v_time_vec);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":400
 *         free_double_vector(score_vec)
 *         free_double_vector(time_vec)
 *         return fit             # <<<<<<<<<<<<<<
 * 
 *     cdef list _fit_theoretical_distribution_on_features(self, double mz, double error_tolerance, int charge,
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_fit));
  __pyx_r = __pyx_v_fit;
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":346
 *         return acc / count
 * 
 *     cpdef LCMSFeatureSetFit _fit_single_feature_set(self, list features, list base_tid, double error_tolerance,             # <<<<<<<<<<<<<<
 *                                                     int charge, double charge_carrier=PROTON, int max_missed_peaks=1,
 *                                                     double threshold_scale=0.3):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase._fit_single_feature_set", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_conformer);
  __Pyx_XDECREF((PyObject *)__pyx_v_scores_array);
  __Pyx_XDECREF((PyObject *)__pyx_v_times_array);
  __Pyx_XDECREF((PyObject *)__pyx_v_feat_iter);
  __Pyx_XDECREF(__pyx_v_eid);
  __Pyx_XDECREF(__pyx_v_cleaned_eid);
  __Pyx_XDECREF(__pyx_v_tid);
  __Pyx_XDECREF((PyObject *)__pyx_v_f);
  __Pyx_XDECREF((PyObject *)__pyx_v_fit);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_11_fit_single_feature_set(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_10_fit_single_feature_set[] = "LCMSFeatureProcessorBase._fit_single_feature_set(self, list features, list base_tid, double error_tolerance, int charge, double charge_carrier=PROTON, int max_missed_peaks=1, double threshold_scale=0.3) -> LCMSFeatureSetFit";
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_11_fit_single_feature_set(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_features = 0;
  PyObject *__pyx_v_base_tid = 0;
  double __pyx_v_error_tolerance;
  int __pyx_v_charge;
  double __pyx_v_charge_carrier;
  int __pyx_v_max_missed_peaks;
  double __pyx_v_threshold_scale;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_fit_single_feature_set (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_features,&__pyx_n_s_base_tid,&__pyx_n_s_error_tolerance,&__pyx_n_s_charge,&__pyx_n_s_charge_carrier,&__pyx_n_s_max_missed_peaks,&__pyx_n_s_threshold_scale,0};
    PyObject* values[7] = {0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_features)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_base_tid)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_fit_single_feature_set", 0, 4, 7, 1); __PYX_ERR(0, 346, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_error_tolerance)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_fit_single_feature_set", 0, 4, 7, 2); __PYX_ERR(0, 346, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_charge)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_fit_single_feature_set", 0, 4, 7, 3); __PYX_ERR(0, 346, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_charge_carrier);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_max_missed_peaks);
          if (value) { values[5] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_threshold_scale);
          if (value) { values[6] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_fit_single_feature_set") < 0)) __PYX_ERR(0, 346, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_features = ((PyObject*)values[0]);
    __pyx_v_base_tid = ((PyObject*)values[1]);
    __pyx_v_error_tolerance = __pyx_PyFloat_AsDouble(values[2]); if (unlikely((__pyx_v_error_tolerance == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 346, __pyx_L3_error)
    __pyx_v_charge = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_charge == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 347, __pyx_L3_error)
    if (values[4]) {
      __pyx_v_charge_carrier = __pyx_PyFloat_AsDouble(values[4]); if (unlikely((__pyx_v_charge_carrier == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 347, __pyx_L3_error)
    } else {
      __pyx_v_charge_carrier = __pyx_k__4;
    }
    if (values[5]) {
      __pyx_v_max_missed_peaks = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_max_missed_peaks == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 347, __pyx_L3_error)
    } else {
      __pyx_v_max_missed_peaks = ((int)1);
    }
    if (values[6]) {
      __pyx_v_threshold_scale = __pyx_PyFloat_AsDouble(values[6]); if (unlikely((__pyx_v_threshold_scale == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 348, __pyx_L3_error)
    } else {
      __pyx_v_threshold_scale = ((double)0.3);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_fit_single_feature_set", 0, 4, 7, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 346, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase._fit_single_feature_set", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_features), (&PyList_Type), 1, "features", 1))) __PYX_ERR(0, 346, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_base_tid), (&PyList_Type), 1, "base_tid", 1))) __PYX_ERR(0, 346, __pyx_L1_error)
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_10_fit_single_feature_set(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)__pyx_v_self), __pyx_v_features, __pyx_v_base_tid, __pyx_v_error_tolerance, __pyx_v_charge, __pyx_v_charge_carrier, __pyx_v_max_missed_peaks, __pyx_v_threshold_scale);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_10_fit_single_feature_set(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, PyObject *__pyx_v_features, PyObject *__pyx_v_base_tid, double __pyx_v_error_tolerance, int __pyx_v_charge, double __pyx_v_charge_carrier, int __pyx_v_max_missed_peaks, double __pyx_v_threshold_scale) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_single_feature_set __pyx_t_2;
  __Pyx_RefNannySetupContext("_fit_single_feature_set", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 3;
  __pyx_t_2.charge_carrier = __pyx_v_charge_carrier;
  __pyx_t_2.max_missed_peaks = __pyx_v_max_missed_peaks;
  __pyx_t_2.threshold_scale = __pyx_v_threshold_scale;
  __pyx_t_1 = ((PyObject *)__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase->_fit_single_feature_set(__pyx_v_self, __pyx_v_features, __pyx_v_base_tid, __pyx_v_error_tolerance, __pyx_v_charge, 1, &__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 346, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase._fit_single_feature_set", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":402
 *         return fit
 * 
 *     cdef list _fit_theoretical_distribution_on_features(self, double mz, double error_tolerance, int charge,             # <<<<<<<<<<<<<<
 *                                                         TheoreticalIsotopicPattern base_tid,
 *                                                         double charge_carrier=PROTON,
 */

static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_theoretical_distribution_on_features(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, double __pyx_v_mz, double __pyx_v_error_tolerance, int __pyx_v_charge, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *__pyx_v_base_tid, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_theoretical_distribution_on_features *__pyx_optional_args) {
  double __pyx_v_charge_carrier = __pyx_k__5;
  int __pyx_v_max_missed_peaks = ((int)1);
  double __pyx_v_threshold_scale = ((double)0.3);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":406
 *                                                         double charge_carrier=PROTON,
 *                                                         double truncate_after=0.8, int max_missed_peaks=1,
 *                                                         double threshold_scale=0.3, LCMSFeature feature=None):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double score, final_score, score_acc, neutral_mass
 */
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_feature = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)Py_None);
  double __pyx_v_score;
  double __pyx_v_final_score;
  double __pyx_v_score_acc;
  double __pyx_v_neutral_mass;
  PyObject *__pyx_v_feature_groups = 0;
  PyObject *__pyx_v_feature_fits = 0;
  PyObject *__pyx_v_features = 0;
  PyObject *__pyx_v_eid = 0;
  PyObject *__pyx_v_cleaned_eid = 0;
  PyObject *__pyx_v_tid = 0;
  PyArrayObject *__pyx_v_scores_array = 0;
  PyArrayObject *__pyx_v_times_array = 0;
  size_t __pyx_v_combn_i;
  size_t __pyx_v_n_missing;
  size_t __pyx_v_missing_features;
  size_t __pyx_v_counter;
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *__pyx_v_snapped_tid = 0;
  size_t __pyx_v_feat_i;
  size_t __pyx_v_feat_n;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_feat_iter = 0;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_f = 0;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_combn_iter = 0;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *__pyx_v_conformer = 0;
  struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector *__pyx_v_score_vec;
  struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector *__pyx_v_time_vec;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit *__pyx_v_fit = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_match_theoretical_isotopic_distribution __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  double __pyx_t_11;
  PyObject *__pyx_t_12;
  size_t __pyx_t_13;
  size_t __pyx_t_14;
  size_t __pyx_t_15;
  struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_neutral_mass __pyx_t_16;
  int __pyx_t_17;
  __Pyx_RefNannySetupContext("_fit_theoretical_distribution_on_features", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_charge_carrier = __pyx_optional_args->charge_carrier;
      if (__pyx_optional_args->__pyx_n > 2) {
        __pyx_v_max_missed_peaks = __pyx_optional_args->max_missed_peaks;
        if (__pyx_optional_args->__pyx_n > 3) {
          __pyx_v_threshold_scale = __pyx_optional_args->threshold_scale;
          if (__pyx_optional_args->__pyx_n > 4) {
            __pyx_v_feature = __pyx_optional_args->feature;
          }
        }
      }
    }
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":424
 * 
 *         feature_groups = self.match_theoretical_isotopic_distribution(
 *             base_tid.truncated_tid, error_tolerance, interval=feature)             # <<<<<<<<<<<<<<
 *         feature_fits = []
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_base_tid), __pyx_n_s_truncated_tid); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 424, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyList_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 424, __pyx_L1_error)

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":423
 *             dvec* time_vec
 * 
 *         feature_groups = self.match_theoretical_isotopic_distribution(             # <<<<<<<<<<<<<<
 *             base_tid.truncated_tid, error_tolerance, interval=feature)
 *         feature_fits = []
 */
  __pyx_t_3.__pyx_n = 2;
  __pyx_t_3.error_tolerance = __pyx_v_error_tolerance;
  __pyx_t_3.interval = __pyx_v_feature;
  __pyx_t_2 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)__pyx_v_self->__pyx_vtab)->match_theoretical_isotopic_distribution(__pyx_v_self, ((PyObject*)__pyx_t_1), 0, &__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 423, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_feature_groups = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":425
 *         feature_groups = self.match_theoretical_isotopic_distribution(
 *             base_tid.truncated_tid, error_tolerance, interval=feature)
 *         feature_fits = []             # <<<<<<<<<<<<<<
 * 
 *         conformer = envelope_conformer._create()
 */
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 425, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_feature_fits = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":427
 *         feature_fits = []
 * 
 *         conformer = envelope_conformer._create()             # <<<<<<<<<<<<<<
 * 
 *         combn_iter = CartesianProductIterator(feature_groups)
 */
  __pyx_t_2 = ((PyObject *)__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer__create()); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 427, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_conformer = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":429
 *         conformer = envelope_conformer._create()
 * 
 *         combn_iter = CartesianProductIterator(feature_groups)             # <<<<<<<<<<<<<<
 *         combn_i = 0
 *         while combn_iter.has_more():
 */
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator), __pyx_v_feature_groups); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 429, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_combn_iter = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":430
 * 
 *         combn_iter = CartesianProductIterator(feature_groups)
 *         combn_i = 0             # <<<<<<<<<<<<<<
 *         while combn_iter.has_more():
 *             features = combn_iter.get_next_value()
 */
  __pyx_v_combn_i = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":431
 *         combn_iter = CartesianProductIterator(feature_groups)
 *         combn_i = 0
 *         while combn_iter.has_more():             # <<<<<<<<<<<<<<
 *             features = combn_iter.get_next_value()
 *             if features is None:
 */
  while (1) {
    __pyx_t_4 = (((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *)__pyx_v_combn_iter->__pyx_vtab)->has_more(__pyx_v_combn_iter, 0) != 0);
    if (!__pyx_t_4) break;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":432
 *         combn_i = 0
 *         while combn_iter.has_more():
 *             features = combn_iter.get_next_value()             # <<<<<<<<<<<<<<
 *             if features is None:
 *                 break
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *)__pyx_v_combn_iter->__pyx_vtab)->get_next_value(__pyx_v_combn_iter, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 432, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_features, ((PyObject*)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":433
 *         while combn_iter.has_more():
 *             features = combn_iter.get_next_value()
 *             if features is None:             # <<<<<<<<<<<<<<
 *                 break
 *             combn_i += 1
 */
    __pyx_t_4 = (__pyx_v_features == ((PyObject*)Py_None));
    __pyx_t_5 = (__pyx_t_4 != 0);
    if (__pyx_t_5) {

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":434
 *             features = combn_iter.get_next_value()
 *             if features is None:
 *                 break             # <<<<<<<<<<<<<<
 *             combn_i += 1
 *             if combn_i % (100000) == 0:
 */
      goto __pyx_L4_break;

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":433
 *         while combn_iter.has_more():
 *             features = combn_iter.get_next_value()
 *             if features is None:             # <<<<<<<<<<<<<<
 *                 break
 *             combn_i += 1
 */
    }

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":435
 *             if features is None:
 *                 break
 *             combn_i += 1             # <<<<<<<<<<<<<<
 *             if combn_i % (100000) == 0:
 *                 print("\t%d combinations have been considered for m/z %0.3f at charge %d in %s (%0.2f%%)" % (
 */
    __pyx_v_combn_i = (__pyx_v_combn_i + 1);

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":436
 *                 break
 *             combn_i += 1
 *             if combn_i % (100000) == 0:             # <<<<<<<<<<<<<<
 *                 print("\t%d combinations have been considered for m/z %0.3f at charge %d in %s (%0.2f%%)" % (
 *                     combn_i, mz, charge, "-" if feature is None else (
 */
    __pyx_t_5 = (((__pyx_v_combn_i % 0x186A0) == 0) != 0);
    if (__pyx_t_5) {

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":438
 *             if combn_i % (100000) == 0:
 *                 print("\t%d combinations have been considered for m/z %0.3f at charge %d in %s (%0.2f%%)" % (
 *                     combn_i, mz, charge, "-" if feature is None else (             # <<<<<<<<<<<<<<
 *                         "%0.2f-%0.2f" % (feature.start_time, feature.end_time)),
 *                     100. * combn_i / combn_iter.total_combinations))
 */
      __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_combn_i); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 438, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_1 = PyFloat_FromDouble(__pyx_v_mz); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 438, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_charge); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 438, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_5 = (((PyObject *)__pyx_v_feature) == Py_None);
      if ((__pyx_t_5 != 0)) {
        __Pyx_INCREF(__pyx_kp_s__6);
        __pyx_t_7 = __pyx_kp_s__6;
      } else {

        /* "ms_deisotope/_c/feature_map/feature_processor.pyx":439
 *                 print("\t%d combinations have been considered for m/z %0.3f at charge %d in %s (%0.2f%%)" % (
 *                     combn_i, mz, charge, "-" if feature is None else (
 *                         "%0.2f-%0.2f" % (feature.start_time, feature.end_time)),             # <<<<<<<<<<<<<<
 *                     100. * combn_i / combn_iter.total_combinations))
 *             if has_no_valid_features(features):
 */
        __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_feature), __pyx_n_s_start_time); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 439, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_9 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_feature), __pyx_n_s_end_time); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 439, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 439, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_GIVEREF(__pyx_t_8);
        PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_8);
        __Pyx_GIVEREF(__pyx_t_9);
        PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_9);
        __pyx_t_8 = 0;
        __pyx_t_9 = 0;
        __pyx_t_9 = __Pyx_PyString_Format(__pyx_kp_s_0_2f_0_2f, __pyx_t_10); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 439, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __pyx_t_7 = __pyx_t_9;
        __pyx_t_9 = 0;
      }

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":440
 *                     combn_i, mz, charge, "-" if feature is None else (
 *                         "%0.2f-%0.2f" % (feature.start_time, feature.end_time)),
 *                     100. * combn_i / combn_iter.total_combinations))             # <<<<<<<<<<<<<<
 *             if has_no_valid_features(features):
 *                 continue
 */
      __pyx_t_11 = (100. * __pyx_v_combn_i);
      if (unlikely(__pyx_v_combn_iter->total_combinations == 0)) {
        PyErr_SetString(PyExc_ZeroDivisionError, "float division");
        __PYX_ERR(0, 440, __pyx_L1_error)
      }
      __pyx_t_9 = PyFloat_FromDouble((__pyx_t_11 / __pyx_v_combn_iter->total_combinations)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 440, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":438
 *             if combn_i % (100000) == 0:
 *                 print("\t%d combinations have been considered for m/z %0.3f at charge %d in %s (%0.2f%%)" % (
 *                     combn_i, mz, charge, "-" if feature is None else (             # <<<<<<<<<<<<<<
 *                         "%0.2f-%0.2f" % (feature.start_time, feature.end_time)),
 *                     100. * combn_i / combn_iter.total_combinations))
 */
      __pyx_t_10 = PyTuple_New(5); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 438, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_10, 2, __pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_10, 3, __pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_9);
      PyTuple_SET_ITEM(__pyx_t_10, 4, __pyx_t_9);
      __pyx_t_2 = 0;
      __pyx_t_1 = 0;
      __pyx_t_6 = 0;
      __pyx_t_7 = 0;
      __pyx_t_9 = 0;

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":437
 *             combn_i += 1
 *             if combn_i % (100000) == 0:
 *                 print("\t%d combinations have been considered for m/z %0.3f at charge %d in %s (%0.2f%%)" % (             # <<<<<<<<<<<<<<
 *                     combn_i, mz, charge, "-" if feature is None else (
 *                         "%0.2f-%0.2f" % (feature.start_time, feature.end_time)),
 */
      __pyx_t_9 = __Pyx_PyString_Format(__pyx_kp_s_d_combinations_have_been_consid, __pyx_t_10); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 437, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      if (__Pyx_PrintOne(0, __pyx_t_9) < 0) __PYX_ERR(0, 437, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":436
 *                 break
 *             combn_i += 1
 *             if combn_i % (100000) == 0:             # <<<<<<<<<<<<<<
 *                 print("\t%d combinations have been considered for m/z %0.3f at charge %d in %s (%0.2f%%)" % (
 *                     combn_i, mz, charge, "-" if feature is None else (
 */
    }

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":441
 *                         "%0.2f-%0.2f" % (feature.start_time, feature.end_time)),
 *                     100. * combn_i / combn_iter.total_combinations))
 *             if has_no_valid_features(features):             # <<<<<<<<<<<<<<
 *                 continue
 *             # If the monoisotopic feature wasn't actually observed, create a dummy feature
 */
    __pyx_t_5 = (__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_has_no_valid_features(__pyx_v_features) != 0);
    if (__pyx_t_5) {

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":442
 *                     100. * combn_i / combn_iter.total_combinations))
 *             if has_no_valid_features(features):
 *                 continue             # <<<<<<<<<<<<<<
 *             # If the monoisotopic feature wasn't actually observed, create a dummy feature
 *             # since the monoisotopic feature cannot be None
 */
      goto __pyx_L3_continue;

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":441
 *                         "%0.2f-%0.2f" % (feature.start_time, feature.end_time)),
 *                     100. * combn_i / combn_iter.total_combinations))
 *             if has_no_valid_features(features):             # <<<<<<<<<<<<<<
 *                 continue
 *             # If the monoisotopic feature wasn't actually observed, create a dummy feature
 */
    }

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":445
 *             # If the monoisotopic feature wasn't actually observed, create a dummy feature
 *             # since the monoisotopic feature cannot be None
 *             if features[0] is None:             # <<<<<<<<<<<<<<
 *                 features = list(features)
 *                 features[0] = EmptyFeature._create(mz)
 */
    if (unlikely(__pyx_v_features == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 445, __pyx_L1_error)
    }
    __pyx_t_9 = __Pyx_GetItemInt_List(__pyx_v_features, 0, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 445, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_5 = (__pyx_t_9 == Py_None);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_4 = (__pyx_t_5 != 0);
    if (__pyx_t_4) {

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":446
 *             # since the monoisotopic feature cannot be None
 *             if features[0] is None:
 *                 features = list(features)             # <<<<<<<<<<<<<<
 *                 features[0] = EmptyFeature._create(mz)
 *                 snapped_tid = base_tid.clone().shift(mz)
 */
      __pyx_t_9 = PySequence_List(__pyx_v_features); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 446, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF_SET(__pyx_v_features, ((PyObject*)__pyx_t_9));
      __pyx_t_9 = 0;

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":447
 *             if features[0] is None:
 *                 features = list(features)
 *                 features[0] = EmptyFeature._create(mz)             # <<<<<<<<<<<<<<
 *                 snapped_tid = base_tid.clone().shift(mz)
 *             else:
 */
      __pyx_t_9 = ((PyObject *)__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature->_create(__pyx_v_mz)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 447, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      if (unlikely(__Pyx_SetItemInt(__pyx_v_features, 0, __pyx_t_9, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0)) __PYX_ERR(0, 447, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":448
 *                 features = list(features)
 *                 features[0] = EmptyFeature._create(mz)
 *                 snapped_tid = base_tid.clone().shift(mz)             # <<<<<<<<<<<<<<
 *             else:
 *                 f = <LCMSFeature>PyList_GET_ITEM(features, 0)
 */
      __pyx_t_9 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *)__pyx_v_base_tid->__pyx_vtab)->clone(__pyx_v_base_tid, 0)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 448, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *)((struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *)__pyx_t_9)->__pyx_vtab)->shift(((struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *)__pyx_t_9), __pyx_v_mz, 0)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 448, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF_SET(__pyx_v_snapped_tid, ((struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *)__pyx_t_10));
      __pyx_t_10 = 0;

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":445
 *             # If the monoisotopic feature wasn't actually observed, create a dummy feature
 *             # since the monoisotopic feature cannot be None
 *             if features[0] is None:             # <<<<<<<<<<<<<<
 *                 features = list(features)
 *                 features[0] = EmptyFeature._create(mz)
 */
      goto __pyx_L8;
    }

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":450
 *                 snapped_tid = base_tid.clone().shift(mz)
 *             else:
 *                 f = <LCMSFeature>PyList_GET_ITEM(features, 0)             # <<<<<<<<<<<<<<
 *                 snapped_tid = base_tid.clone().shift(f.get_mz())
 *             # feat_iter = FeatureSetIterator._create_with_threshold(features, base_tid, 0.1)
 */
    /*else*/ {
      __pyx_t_12 = PyList_GET_ITEM(__pyx_v_features, 0);
      __pyx_t_10 = ((PyObject *)__pyx_t_12);
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_XDECREF_SET(__pyx_v_f, ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_t_10));
      __pyx_t_10 = 0;

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":451
 *             else:
 *                 f = <LCMSFeature>PyList_GET_ITEM(features, 0)
 *                 snapped_tid = base_tid.clone().shift(f.get_mz())             # <<<<<<<<<<<<<<
 *             # feat_iter = FeatureSetIterator._create_with_threshold(features, base_tid, 0.1)
 *             feat_iter = FeatureSetIterator._create(features)
 */
      __pyx_t_10 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *)__pyx_v_base_tid->__pyx_vtab)->clone(__pyx_v_base_tid, 0)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 451, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_9 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *)((struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *)__pyx_t_10)->__pyx_vtab)->shift(((struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *)__pyx_t_10), ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_f->__pyx_base.__pyx_vtab)->__pyx_base.get_mz(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_f)), 0)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 451, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF_SET(__pyx_v_snapped_tid, ((struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *)__pyx_t_9));
      __pyx_t_9 = 0;
    }
    __pyx_L8:;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":453
 *                 snapped_tid = base_tid.clone().shift(f.get_mz())
 *             # feat_iter = FeatureSetIterator._create_with_threshold(features, base_tid, 0.1)
 *             feat_iter = FeatureSetIterator._create(features)             # <<<<<<<<<<<<<<
 *             counter = 0
 *             score_acc = 0
 */
    __pyx_t_9 = ((PyObject *)__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator->_create(__pyx_v_features)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 453, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_XDECREF_SET(__pyx_v_feat_iter, ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *)__pyx_t_9));
    __pyx_t_9 = 0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":454
 *             # feat_iter = FeatureSetIterator._create_with_threshold(features, base_tid, 0.1)
 *             feat_iter = FeatureSetIterator._create(features)
 *             counter = 0             # <<<<<<<<<<<<<<
 *             score_acc = 0
 * 
 */
    __pyx_v_counter = 0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":455
 *             feat_iter = FeatureSetIterator._create(features)
 *             counter = 0
 *             score_acc = 0             # <<<<<<<<<<<<<<
 * 
 *             score_vec = make_double_vector()
 */
    __pyx_v_score_acc = 0.0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":457
 *             score_acc = 0
 * 
 *             score_vec = make_double_vector()             # <<<<<<<<<<<<<<
 *             time_vec = make_double_vector()
 * 
 */
    __pyx_v_score_vec = __pyx_f_7brainpy_2_c_13double_vector_make_double_vector();

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":458
 * 
 *             score_vec = make_double_vector()
 *             time_vec = make_double_vector()             # <<<<<<<<<<<<<<
 * 
 *             while feat_iter.has_more():
 */
    __pyx_v_time_vec = __pyx_f_7brainpy_2_c_13double_vector_make_double_vector();

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":460
 *             time_vec = make_double_vector()
 * 
 *             while feat_iter.has_more():             # <<<<<<<<<<<<<<
 *                 eid = feat_iter.get_next_value()
 *                 if eid is None:
 */
    while (1) {
      __pyx_t_4 = (((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *)__pyx_v_feat_iter->__pyx_vtab)->has_more(__pyx_v_feat_iter, 0) != 0);
      if (!__pyx_t_4) break;

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":461
 * 
 *             while feat_iter.has_more():
 *                 eid = feat_iter.get_next_value()             # <<<<<<<<<<<<<<
 *                 if eid is None:
 *                     continue
 */
      __pyx_t_9 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *)__pyx_v_feat_iter->__pyx_vtab)->get_next_value(__pyx_v_feat_iter, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 461, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_XDECREF_SET(__pyx_v_eid, ((PyObject*)__pyx_t_9));
      __pyx_t_9 = 0;

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":462
 *             while feat_iter.has_more():
 *                 eid = feat_iter.get_next_value()
 *                 if eid is None:             # <<<<<<<<<<<<<<
 *                     continue
 *                 counter += 1
 */
      __pyx_t_4 = (__pyx_v_eid == ((PyObject*)Py_None));
      __pyx_t_5 = (__pyx_t_4 != 0);
      if (__pyx_t_5) {

        /* "ms_deisotope/_c/feature_map/feature_processor.pyx":463
 *                 eid = feat_iter.get_next_value()
 *                 if eid is None:
 *                     continue             # <<<<<<<<<<<<<<
 *                 counter += 1
 *                 conformer.acquire(eid, snapped_tid.truncated_tid)
 */
        goto __pyx_L9_continue;

        /* "ms_deisotope/_c/feature_map/feature_processor.pyx":462
 *             while feat_iter.has_more():
 *                 eid = feat_iter.get_next_value()
 *                 if eid is None:             # <<<<<<<<<<<<<<
 *                     continue
 *                 counter += 1
 */
      }

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":464
 *                 if eid is None:
 *                     continue
 *                 counter += 1             # <<<<<<<<<<<<<<
 *                 conformer.acquire(eid, snapped_tid.truncated_tid)
 *                 conformer.conform()
 */
      __pyx_v_counter = (__pyx_v_counter + 1);

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":465
 *                     continue
 *                 counter += 1
 *                 conformer.acquire(eid, snapped_tid.truncated_tid)             # <<<<<<<<<<<<<<
 *                 conformer.conform()
 *                 cleaned_eid = conformer.experimental
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_snapped_tid), __pyx_n_s_truncated_tid); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 465, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      if (!(likely(PyList_CheckExact(__pyx_t_9))||((__pyx_t_9) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_9)->tp_name), 0))) __PYX_ERR(0, 465, __pyx_L1_error)
      ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *)__pyx_v_conformer->__pyx_vtab)->acquire(__pyx_v_conformer, __pyx_v_eid, ((PyObject*)__pyx_t_9));
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":466
 *                 counter += 1
 *                 conformer.acquire(eid, snapped_tid.truncated_tid)
 *                 conformer.conform()             # <<<<<<<<<<<<<<
 *                 cleaned_eid = conformer.experimental
 *                 tid = conformer.theoretical
 */
      ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *)__pyx_v_conformer->__pyx_vtab)->conform(__pyx_v_conformer, NULL);

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":467
 *                 conformer.acquire(eid, snapped_tid.truncated_tid)
 *                 conformer.conform()
 *                 cleaned_eid = conformer.experimental             # <<<<<<<<<<<<<<
 *                 tid = conformer.theoretical
 *                 n_missing = conformer.n_missing
 */
      __pyx_t_9 = __pyx_v_conformer->experimental;
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_XDECREF_SET(__pyx_v_cleaned_eid, ((PyObject*)__pyx_t_9));
      __pyx_t_9 = 0;

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":468
 *                 conformer.conform()
 *                 cleaned_eid = conformer.experimental
 *                 tid = conformer.theoretical             # <<<<<<<<<<<<<<
 *                 n_missing = conformer.n_missing
 *                 if n_missing > max_missed_peaks:
 */
      __pyx_t_9 = __pyx_v_conformer->theoretical;
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_XDECREF_SET(__pyx_v_tid, ((PyObject*)__pyx_t_9));
      __pyx_t_9 = 0;

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":469
 *                 cleaned_eid = conformer.experimental
 *                 tid = conformer.theoretical
 *                 n_missing = conformer.n_missing             # <<<<<<<<<<<<<<
 *                 if n_missing > max_missed_peaks:
 *                     continue
 */
      __pyx_t_13 = __pyx_v_conformer->n_missing;
      __pyx_v_n_missing = __pyx_t_13;

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":470
 *                 tid = conformer.theoretical
 *                 n_missing = conformer.n_missing
 *                 if n_missing > max_missed_peaks:             # <<<<<<<<<<<<<<
 *                     continue
 * 
 */
      __pyx_t_5 = ((__pyx_v_n_missing > __pyx_v_max_missed_peaks) != 0);
      if (__pyx_t_5) {

        /* "ms_deisotope/_c/feature_map/feature_processor.pyx":471
 *                 n_missing = conformer.n_missing
 *                 if n_missing > max_missed_peaks:
 *                     continue             # <<<<<<<<<<<<<<
 * 
 *                 score = self.scorer._evaluate(None, cleaned_eid, tid)
 */
        goto __pyx_L9_continue;

        /* "ms_deisotope/_c/feature_map/feature_processor.pyx":470
 *                 tid = conformer.theoretical
 *                 n_missing = conformer.n_missing
 *                 if n_missing > max_missed_peaks:             # <<<<<<<<<<<<<<
 *                     continue
 * 
 */
      }

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":473
 *                     continue
 * 
 *                 score = self.scorer._evaluate(None, cleaned_eid, tid)             # <<<<<<<<<<<<<<
 *                 if npy_isnan(score):
 *                     continue
 */
      __pyx_v_score = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *)__pyx_v_self->scorer->__pyx_vtab)->_evaluate(__pyx_v_self->scorer, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)Py_None), __pyx_v_cleaned_eid, __pyx_v_tid, 0);

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":474
 * 
 *                 score = self.scorer._evaluate(None, cleaned_eid, tid)
 *                 if npy_isnan(score):             # <<<<<<<<<<<<<<
 *                     continue
 *                 score_acc += score
 */
      __pyx_t_5 = (npy_isnan(__pyx_v_score) != 0);
      if (__pyx_t_5) {

        /* "ms_deisotope/_c/feature_map/feature_processor.pyx":475
 *                 score = self.scorer._evaluate(None, cleaned_eid, tid)
 *                 if npy_isnan(score):
 *                     continue             # <<<<<<<<<<<<<<
 *                 score_acc += score
 *                 double_vector_append(score_vec, score)
 */
        goto __pyx_L9_continue;

        /* "ms_deisotope/_c/feature_map/feature_processor.pyx":474
 * 
 *                 score = self.scorer._evaluate(None, cleaned_eid, tid)
 *                 if npy_isnan(score):             # <<<<<<<<<<<<<<
 *                     continue
 *                 score_acc += score
 */
      }

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":476
 *                 if npy_isnan(score):
 *                     continue
 *                 score_acc += score             # <<<<<<<<<<<<<<
 *                 double_vector_append(score_vec, score)
 *                 double_vector_append(time_vec, feat_iter.get_current_time())
 */
      __pyx_v_score_acc = (__pyx_v_score_acc + __pyx_v_score);

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":477
 *                     continue
 *                 score_acc += score
 *                 double_vector_append(score_vec, score)             # <<<<<<<<<<<<<<
 *                 double_vector_append(time_vec, feat_iter.get_current_time())
 * 
 */
      (void)(__pyx_f_7brainpy_2_c_13double_vector_double_vector_append(__pyx_v_score_vec, __pyx_v_score));

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":478
 *                 score_acc += score
 *                 double_vector_append(score_vec, score)
 *                 double_vector_append(time_vec, feat_iter.get_current_time())             # <<<<<<<<<<<<<<
 * 
 *             # Compute feature score from score_vec
 */
      (void)(__pyx_f_7brainpy_2_c_13double_vector_double_vector_append(__pyx_v_time_vec, ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *)__pyx_v_feat_iter->__pyx_vtab)->get_current_time(__pyx_v_feat_iter, 0)));
      __pyx_L9_continue:;
    }

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":481
 * 
 *             # Compute feature score from score_vec
 *             final_score = self._find_thresholded_score(             # <<<<<<<<<<<<<<
 *                 score_vec, threshold_scale)
 *             missing_features = 0
 */
    __pyx_v_final_score = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)__pyx_v_self->__pyx_vtab)->_find_thresholded_score(__pyx_v_self, __pyx_v_score_vec, __pyx_v_threshold_scale);

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":483
 *             final_score = self._find_thresholded_score(
 *                 score_vec, threshold_scale)
 *             missing_features = 0             # <<<<<<<<<<<<<<
 *             feat_n = PyList_GET_SIZE(features)
 *             for feat_i in range(feat_n):
 */
    __pyx_v_missing_features = 0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":484
 *                 score_vec, threshold_scale)
 *             missing_features = 0
 *             feat_n = PyList_GET_SIZE(features)             # <<<<<<<<<<<<<<
 *             for feat_i in range(feat_n):
 *                 f = <LCMSFeature>PyList_GET_ITEM(features, feat_i)
 */
    __pyx_v_feat_n = PyList_GET_SIZE(__pyx_v_features);

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":485
 *             missing_features = 0
 *             feat_n = PyList_GET_SIZE(features)
 *             for feat_i in range(feat_n):             # <<<<<<<<<<<<<<
 *                 f = <LCMSFeature>PyList_GET_ITEM(features, feat_i)
 *                 if f is None:
 */
    __pyx_t_13 = __pyx_v_feat_n;
    __pyx_t_14 = __pyx_t_13;
    for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
      __pyx_v_feat_i = __pyx_t_15;

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":486
 *             feat_n = PyList_GET_SIZE(features)
 *             for feat_i in range(feat_n):
 *                 f = <LCMSFeature>PyList_GET_ITEM(features, feat_i)             # <<<<<<<<<<<<<<
 *                 if f is None:
 *                     missing_features += 1
 */
      __pyx_t_12 = PyList_GET_ITEM(__pyx_v_features, __pyx_v_feat_i);
      __pyx_t_9 = ((PyObject *)__pyx_t_12);
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_XDECREF_SET(__pyx_v_f, ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_t_9));
      __pyx_t_9 = 0;

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":487
 *             for feat_i in range(feat_n):
 *                 f = <LCMSFeature>PyList_GET_ITEM(features, feat_i)
 *                 if f is None:             # <<<<<<<<<<<<<<
 *                     missing_features += 1
 *             f = <LCMSFeature>PyList_GET_ITEM(features, 0)
 */
      __pyx_t_5 = (((PyObject *)__pyx_v_f) == Py_None);
      __pyx_t_4 = (__pyx_t_5 != 0);
      if (__pyx_t_4) {

        /* "ms_deisotope/_c/feature_map/feature_processor.pyx":488
 *                 f = <LCMSFeature>PyList_GET_ITEM(features, feat_i)
 *                 if f is None:
 *                     missing_features += 1             # <<<<<<<<<<<<<<
 *             f = <LCMSFeature>PyList_GET_ITEM(features, 0)
 *             neutral_mass = calc_neutral_mass(
 */
        __pyx_v_missing_features = (__pyx_v_missing_features + 1);

        /* "ms_deisotope/_c/feature_map/feature_processor.pyx":487
 *             for feat_i in range(feat_n):
 *                 f = <LCMSFeature>PyList_GET_ITEM(features, feat_i)
 *                 if f is None:             # <<<<<<<<<<<<<<
 *                     missing_features += 1
 *             f = <LCMSFeature>PyList_GET_ITEM(features, 0)
 */
      }
    }

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":489
 *                 if f is None:
 *                     missing_features += 1
 *             f = <LCMSFeature>PyList_GET_ITEM(features, 0)             # <<<<<<<<<<<<<<
 *             neutral_mass = calc_neutral_mass(
 *                     snapped_tid.get_monoisotopic_mz(), charge, charge_carrier)
 */
    __pyx_t_12 = PyList_GET_ITEM(__pyx_v_features, 0);
    __pyx_t_9 = ((PyObject *)__pyx_t_12);
    __Pyx_INCREF(__pyx_t_9);
    __Pyx_XDECREF_SET(__pyx_v_f, ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_t_9));
    __pyx_t_9 = 0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":490
 *                     missing_features += 1
 *             f = <LCMSFeature>PyList_GET_ITEM(features, 0)
 *             neutral_mass = calc_neutral_mass(             # <<<<<<<<<<<<<<
 *                     snapped_tid.get_monoisotopic_mz(), charge, charge_carrier)
 *             scores_array = double_vector_to_ndarray(score_vec)
 */
    __pyx_t_16.__pyx_n = 1;
    __pyx_t_16.charge_carrier = __pyx_v_charge_carrier;
    __pyx_t_11 = __pyx_f_12ms_deisotope_2_c_9averagine_neutral_mass(((struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *)__pyx_v_snapped_tid->__pyx_vtab)->get_monoisotopic_mz(__pyx_v_snapped_tid), __pyx_v_charge, &__pyx_t_16); 
    __pyx_v_neutral_mass = __pyx_t_11;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":492
 *             neutral_mass = calc_neutral_mass(
 *                     snapped_tid.get_monoisotopic_mz(), charge, charge_carrier)
 *             scores_array = double_vector_to_ndarray(score_vec)             # <<<<<<<<<<<<<<
 *             times_array = double_vector_to_ndarray(time_vec)
 *             fit = LCMSFeatureSetFit._create(
 */
    __pyx_t_9 = __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_double_vector_to_ndarray(__pyx_v_score_vec); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 492, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (!(likely(((__pyx_t_9) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_9, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 492, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_scores_array, ((PyArrayObject *)__pyx_t_9));
    __pyx_t_9 = 0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":493
 *                     snapped_tid.get_monoisotopic_mz(), charge, charge_carrier)
 *             scores_array = double_vector_to_ndarray(score_vec)
 *             times_array = double_vector_to_ndarray(time_vec)             # <<<<<<<<<<<<<<
 *             fit = LCMSFeatureSetFit._create(
 *                 features, snapped_tid,
 */
    __pyx_t_9 = __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_double_vector_to_ndarray(__pyx_v_time_vec); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 493, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (!(likely(((__pyx_t_9) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_9, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 493, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_times_array, ((PyArrayObject *)__pyx_t_9));
    __pyx_t_9 = 0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":497
 *                 features, snapped_tid,
 *                 final_score, charge,
 *                 missing_features, [],             # <<<<<<<<<<<<<<
 *                 None, neutral_mass,
 *                 counter, scores_array,
 */
    __pyx_t_9 = PyList_New(0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 497, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":494
 *             scores_array = double_vector_to_ndarray(score_vec)
 *             times_array = double_vector_to_ndarray(time_vec)
 *             fit = LCMSFeatureSetFit._create(             # <<<<<<<<<<<<<<
 *                 features, snapped_tid,
 *                 final_score, charge,
 */
    __pyx_t_10 = ((PyObject *)__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit->_create(__pyx_v_features, __pyx_v_snapped_tid, __pyx_v_final_score, __pyx_v_charge, __pyx_v_missing_features, ((PyObject*)__pyx_t_9), Py_None, __pyx_v_neutral_mass, __pyx_v_counter, __pyx_v_scores_array, __pyx_v_times_array)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 494, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_XDECREF_SET(__pyx_v_fit, ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit *)__pyx_t_10));
    __pyx_t_10 = 0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":502
 *                 times_array)
 * 
 *             free_double_vector(score_vec)             # <<<<<<<<<<<<<<
 *             free_double_vector(time_vec)
 *             if self.scorer.reject_score(fit.score):
 */
    __pyx_f_7brainpy_2_c_13double_vector_free_double_vector(__pyx_v_score_vec);

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":503
 * 
 *             free_double_vector(score_vec)
 *             free_double_vector(time_vec)             # <<<<<<<<<<<<<<
 *             if self.scorer.reject_score(fit.score):
 *                 continue
 */
    __pyx_f_7brainpy_2_c_13double_vector_free_double_vector(__pyx_v_time_vec);

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":504
 *             free_double_vector(score_vec)
 *             free_double_vector(time_vec)
 *             if self.scorer.reject_score(fit.score):             # <<<<<<<<<<<<<<
 *                 continue
 *             feature_fits.append(fit)
 */
    __pyx_t_4 = (((struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *)__pyx_v_self->scorer->__pyx_vtab)->reject_score(__pyx_v_self->scorer, __pyx_v_fit->score, 0) != 0);
    if (__pyx_t_4) {

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":505
 *             free_double_vector(time_vec)
 *             if self.scorer.reject_score(fit.score):
 *                 continue             # <<<<<<<<<<<<<<
 *             feature_fits.append(fit)
 *         return feature_fits
 */
      goto __pyx_L3_continue;

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":504
 *             free_double_vector(score_vec)
 *             free_double_vector(time_vec)
 *             if self.scorer.reject_score(fit.score):             # <<<<<<<<<<<<<<
 *                 continue
 *             feature_fits.append(fit)
 */
    }

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":506
 *             if self.scorer.reject_score(fit.score):
 *                 continue
 *             feature_fits.append(fit)             # <<<<<<<<<<<<<<
 *         return feature_fits
 * 
 */
    __pyx_t_17 = __Pyx_PyList_Append(__pyx_v_feature_fits, ((PyObject *)__pyx_v_fit)); if (unlikely(__pyx_t_17 == ((int)-1))) __PYX_ERR(0, 506, __pyx_L1_error)
    __pyx_L3_continue:;
  }
  __pyx_L4_break:;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":507
 *                 continue
 *             feature_fits.append(fit)
 *         return feature_fits             # <<<<<<<<<<<<<<
 * 
 *     cpdef list _fit_feature_set(self, double mz, double error_tolerance, int charge, int left_search=1,
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_feature_fits);
  __pyx_r = __pyx_v_feature_fits;
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":402
 *         return fit
 * 
 *     cdef list _fit_theoretical_distribution_on_features(self, double mz, double error_tolerance, int charge,             # <<<<<<<<<<<<<<
 *                                                         TheoreticalIsotopicPattern base_tid,
 *                                                         double charge_carrier=PROTON,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase._fit_theoretical_distribution_on_features", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_feature_groups);
  __Pyx_XDECREF(__pyx_v_feature_fits);
  __Pyx_XDECREF(__pyx_v_features);
  __Pyx_XDECREF(__pyx_v_eid);
  __Pyx_XDECREF(__pyx_v_cleaned_eid);
  __Pyx_XDECREF(__pyx_v_tid);
  __Pyx_XDECREF((PyObject *)__pyx_v_scores_array);
  __Pyx_XDECREF((PyObject *)__pyx_v_times_array);
  __Pyx_XDECREF((PyObject *)__pyx_v_snapped_tid);
  __Pyx_XDECREF((PyObject *)__pyx_v_feat_iter);
  __Pyx_XDECREF((PyObject *)__pyx_v_f);
  __Pyx_XDECREF((PyObject *)__pyx_v_combn_iter);
  __Pyx_XDECREF((PyObject *)__pyx_v_conformer);
  __Pyx_XDECREF((PyObject *)__pyx_v_fit);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":509
 *         return feature_fits
 * 
 *     cpdef list _fit_feature_set(self, double mz, double error_tolerance, int charge, int left_search=1,             # <<<<<<<<<<<<<<
 *                                 int right_search=1, double charge_carrier=PROTON, double truncate_after=0.8,
 *                                 int max_missed_peaks=1, double threshold_scale=0.3, LCMSFeature feature=None):
 */

static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_13_fit_feature_set(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_feature_set(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, double __pyx_v_mz, double __pyx_v_error_tolerance, int __pyx_v_charge, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_feature_set *__pyx_optional_args) {
  int __pyx_v_left_search = ((int)1);
  int __pyx_v_right_search = ((int)1);
  double __pyx_v_charge_carrier = __pyx_k__7;
  double __pyx_v_truncate_after = ((double)0.8);
  int __pyx_v_max_missed_peaks = ((int)1);
  double __pyx_v_threshold_scale = ((double)0.3);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":511
 *     cpdef list _fit_feature_set(self, double mz, double error_tolerance, int charge, int left_search=1,
 *                                 int right_search=1, double charge_carrier=PROTON, double truncate_after=0.8,
 *                                 int max_missed_peaks=1, double threshold_scale=0.3, LCMSFeature feature=None):             # <<<<<<<<<<<<<<
 *         cdef:
 *             list feature_fits
 */
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_feature = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)Py_None);
  PyObject *__pyx_v_feature_fits = 0;
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *__pyx_v_base_tid = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_create_theoretical_distribution __pyx_t_16;
  struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_theoretical_distribution_on_features __pyx_t_17;
  __Pyx_RefNannySetupContext("_fit_feature_set", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_left_search = __pyx_optional_args->left_search;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_right_search = __pyx_optional_args->right_search;
        if (__pyx_optional_args->__pyx_n > 2) {
          __pyx_v_charge_carrier = __pyx_optional_args->charge_carrier;
          if (__pyx_optional_args->__pyx_n > 3) {
            __pyx_v_truncate_after = __pyx_optional_args->truncate_after;
            if (__pyx_optional_args->__pyx_n > 4) {
              __pyx_v_max_missed_peaks = __pyx_optional_args->max_missed_peaks;
              if (__pyx_optional_args->__pyx_n > 5) {
                __pyx_v_threshold_scale = __pyx_optional_args->threshold_scale;
                if (__pyx_optional_args->__pyx_n > 6) {
                  __pyx_v_feature = __pyx_optional_args->feature;
                }
              }
            }
          }
        }
      }
    }
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":509
 *         return feature_fits
 * 
 *     cpdef list _fit_feature_set(self, double mz, double error_tolerance, int charge, int left_search=1,             # <<<<<<<<<<<<<<
 *                                 int right_search=1, double charge_carrier=PROTON, double truncate_after=0.8,
 *                                 int max_missed_peaks=1, double threshold_scale=0.3, LCMSFeature feature=None):
 */
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_fit_feature_set); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 509, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_13_fit_feature_set)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_mz); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 509, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_error_tolerance); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 509, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_charge); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 509, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_left_search); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 509, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_right_search); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 509, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_8 = PyFloat_FromDouble(__pyx_v_charge_carrier); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 509, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_9 = PyFloat_FromDouble(__pyx_v_truncate_after); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 509, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_10 = __Pyx_PyInt_From_int(__pyx_v_max_missed_peaks); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 509, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_11 = PyFloat_FromDouble(__pyx_v_threshold_scale); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 509, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_12 = __pyx_t_1; __pyx_t_13 = NULL;
        __pyx_t_14 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_12))) {
          __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_12);
          if (likely(__pyx_t_13)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
            __Pyx_INCREF(__pyx_t_13);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_12, function);
            __pyx_t_14 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_12)) {
          PyObject *__pyx_temp[11] = {__pyx_t_13, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6, __pyx_t_7, __pyx_t_8, __pyx_t_9, __pyx_t_10, __pyx_t_11, ((PyObject *)__pyx_v_feature)};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_14, 10+__pyx_t_14); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 509, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_12)) {
          PyObject *__pyx_temp[11] = {__pyx_t_13, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6, __pyx_t_7, __pyx_t_8, __pyx_t_9, __pyx_t_10, __pyx_t_11, ((PyObject *)__pyx_v_feature)};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_14, 10+__pyx_t_14); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 509, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        } else
        #endif
        {
          __pyx_t_15 = PyTuple_New(10+__pyx_t_14); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 509, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_15);
          if (__pyx_t_13) {
            __Pyx_GIVEREF(__pyx_t_13); PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_13); __pyx_t_13 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_15, 0+__pyx_t_14, __pyx_t_3);
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_15, 1+__pyx_t_14, __pyx_t_4);
          __Pyx_GIVEREF(__pyx_t_5);
          PyTuple_SET_ITEM(__pyx_t_15, 2+__pyx_t_14, __pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_6);
          PyTuple_SET_ITEM(__pyx_t_15, 3+__pyx_t_14, __pyx_t_6);
          __Pyx_GIVEREF(__pyx_t_7);
          PyTuple_SET_ITEM(__pyx_t_15, 4+__pyx_t_14, __pyx_t_7);
          __Pyx_GIVEREF(__pyx_t_8);
          PyTuple_SET_ITEM(__pyx_t_15, 5+__pyx_t_14, __pyx_t_8);
          __Pyx_GIVEREF(__pyx_t_9);
          PyTuple_SET_ITEM(__pyx_t_15, 6+__pyx_t_14, __pyx_t_9);
          __Pyx_GIVEREF(__pyx_t_10);
          PyTuple_SET_ITEM(__pyx_t_15, 7+__pyx_t_14, __pyx_t_10);
          __Pyx_GIVEREF(__pyx_t_11);
          PyTuple_SET_ITEM(__pyx_t_15, 8+__pyx_t_14, __pyx_t_11);
          __Pyx_INCREF(((PyObject *)__pyx_v_feature));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_feature));
          PyTuple_SET_ITEM(__pyx_t_15, 9+__pyx_t_14, ((PyObject *)__pyx_v_feature));
          __pyx_t_3 = 0;
          __pyx_t_4 = 0;
          __pyx_t_5 = 0;
          __pyx_t_6 = 0;
          __pyx_t_7 = 0;
          __pyx_t_8 = 0;
          __pyx_t_9 = 0;
          __pyx_t_10 = 0;
          __pyx_t_11 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_12, __pyx_t_15, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 509, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        }
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 509, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":515
 *             list feature_fits
 *             TheoreticalIsotopicPattern base_tid
 *         base_tid = self.create_theoretical_distribution(             # <<<<<<<<<<<<<<
 *             mz, charge, charge_carrier, truncate_after)
 *         feature_fits = self._fit_theoretical_distribution_on_features(
 */
  __pyx_t_16.__pyx_n = 2;
  __pyx_t_16.charge_carrier = __pyx_v_charge_carrier;
  __pyx_t_16.truncate_after = __pyx_v_truncate_after;
  __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)__pyx_v_self->__pyx_vtab)->create_theoretical_distribution(__pyx_v_self, __pyx_v_mz, __pyx_v_charge, 0, &__pyx_t_16); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 515, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern))))) __PYX_ERR(0, 515, __pyx_L1_error)
  __pyx_v_base_tid = ((struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":517
 *         base_tid = self.create_theoretical_distribution(
 *             mz, charge, charge_carrier, truncate_after)
 *         feature_fits = self._fit_theoretical_distribution_on_features(             # <<<<<<<<<<<<<<
 *             mz, error_tolerance, charge, base_tid,
 *             charge_carrier, truncate_after, max_missed_peaks, threshold_scale,
 */
  __pyx_t_17.__pyx_n = 5;
  __pyx_t_17.charge_carrier = __pyx_v_charge_carrier;
  __pyx_t_17.truncate_after = __pyx_v_truncate_after;
  __pyx_t_17.max_missed_peaks = __pyx_v_max_missed_peaks;
  __pyx_t_17.threshold_scale = __pyx_v_threshold_scale;
  __pyx_t_17.feature = __pyx_v_feature;
  __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)__pyx_v_self->__pyx_vtab)->_fit_theoretical_distribution_on_features(__pyx_v_self, __pyx_v_mz, __pyx_v_error_tolerance, __pyx_v_charge, __pyx_v_base_tid, &__pyx_t_17); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 517, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_feature_fits = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":521
 *             charge_carrier, truncate_after, max_missed_peaks, threshold_scale,
 *             feature)
 *         return feature_fits             # <<<<<<<<<<<<<<
 * 
 *     cpdef list _fit_composition(self, dict composition, double error_tolerance, int charge, double charge_carrier=PROTON,
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_feature_fits);
  __pyx_r = __pyx_v_feature_fits;
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":509
 *         return feature_fits
 * 
 *     cpdef list _fit_feature_set(self, double mz, double error_tolerance, int charge, int left_search=1,             # <<<<<<<<<<<<<<
 *                                 int right_search=1, double charge_carrier=PROTON, double truncate_after=0.8,
 *                                 int max_missed_peaks=1, double threshold_scale=0.3, LCMSFeature feature=None):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase._fit_feature_set", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_feature_fits);
  __Pyx_XDECREF((PyObject *)__pyx_v_base_tid);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_13_fit_feature_set(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_12_fit_feature_set[] = "LCMSFeatureProcessorBase._fit_feature_set(self, double mz, double error_tolerance, int charge, int left_search=1, int right_search=1, double charge_carrier=PROTON, double truncate_after=0.8, int max_missed_peaks=1, double threshold_scale=0.3, LCMSFeature feature=None) -> list";
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_13_fit_feature_set(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_mz;
  double __pyx_v_error_tolerance;
  int __pyx_v_charge;
  int __pyx_v_left_search;
  int __pyx_v_right_search;
  double __pyx_v_charge_carrier;
  double __pyx_v_truncate_after;
  int __pyx_v_max_missed_peaks;
  double __pyx_v_threshold_scale;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_feature = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_fit_feature_set (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_mz,&__pyx_n_s_error_tolerance,&__pyx_n_s_charge,&__pyx_n_s_left_search,&__pyx_n_s_right_search,&__pyx_n_s_charge_carrier,&__pyx_n_s_truncate_after,&__pyx_n_s_max_missed_peaks,&__pyx_n_s_threshold_scale,&__pyx_n_s_feature,0};
    PyObject* values[10] = {0,0,0,0,0,0,0,0,0,0};

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":511
 *     cpdef list _fit_feature_set(self, double mz, double error_tolerance, int charge, int left_search=1,
 *                                 int right_search=1, double charge_carrier=PROTON, double truncate_after=0.8,
 *                                 int max_missed_peaks=1, double threshold_scale=0.3, LCMSFeature feature=None):             # <<<<<<<<<<<<<<
 *         cdef:
 *             list feature_fits
 */
    values[9] = (PyObject *)((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mz)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_error_tolerance)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_fit_feature_set", 0, 3, 10, 1); __PYX_ERR(0, 509, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_charge)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_fit_feature_set", 0, 3, 10, 2); __PYX_ERR(0, 509, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_left_search);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_right_search);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_charge_carrier);
          if (value) { values[5] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_truncate_after);
          if (value) { values[6] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_max_missed_peaks);
          if (value) { values[7] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_threshold_scale);
          if (value) { values[8] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_feature);
          if (value) { values[9] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_fit_feature_set") < 0)) __PYX_ERR(0, 509, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_mz = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_mz == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 509, __pyx_L3_error)
    __pyx_v_error_tolerance = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_error_tolerance == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 509, __pyx_L3_error)
    __pyx_v_charge = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_charge == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 509, __pyx_L3_error)
    if (values[3]) {
      __pyx_v_left_search = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_left_search == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 509, __pyx_L3_error)
    } else {
      __pyx_v_left_search = ((int)1);
    }
    if (values[4]) {
      __pyx_v_right_search = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_right_search == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 510, __pyx_L3_error)
    } else {
      __pyx_v_right_search = ((int)1);
    }
    if (values[5]) {
      __pyx_v_charge_carrier = __pyx_PyFloat_AsDouble(values[5]); if (unlikely((__pyx_v_charge_carrier == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 510, __pyx_L3_error)
    } else {
      __pyx_v_charge_carrier = __pyx_k__7;
    }
    if (values[6]) {
      __pyx_v_truncate_after = __pyx_PyFloat_AsDouble(values[6]); if (unlikely((__pyx_v_truncate_after == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 510, __pyx_L3_error)
    } else {
      __pyx_v_truncate_after = ((double)0.8);
    }
    if (values[7]) {
      __pyx_v_max_missed_peaks = __Pyx_PyInt_As_int(values[7]); if (unlikely((__pyx_v_max_missed_peaks == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 511, __pyx_L3_error)
    } else {
      __pyx_v_max_missed_peaks = ((int)1);
    }
    if (values[8]) {
      __pyx_v_threshold_scale = __pyx_PyFloat_AsDouble(values[8]); if (unlikely((__pyx_v_threshold_scale == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 511, __pyx_L3_error)
    } else {
      __pyx_v_threshold_scale = ((double)0.3);
    }
    __pyx_v_feature = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)values[9]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_fit_feature_set", 0, 3, 10, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 509, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase._fit_feature_set", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_feature), __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature, 1, "feature", 0))) __PYX_ERR(0, 511, __pyx_L1_error)
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_12_fit_feature_set(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)__pyx_v_self), __pyx_v_mz, __pyx_v_error_tolerance, __pyx_v_charge, __pyx_v_left_search, __pyx_v_right_search, __pyx_v_charge_carrier, __pyx_v_truncate_after, __pyx_v_max_missed_peaks, __pyx_v_threshold_scale, __pyx_v_feature);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":509
 *         return feature_fits
 * 
 *     cpdef list _fit_feature_set(self, double mz, double error_tolerance, int charge, int left_search=1,             # <<<<<<<<<<<<<<
 *                                 int right_search=1, double charge_carrier=PROTON, double truncate_after=0.8,
 *                                 int max_missed_peaks=1, double threshold_scale=0.3, LCMSFeature feature=None):
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_12_fit_feature_set(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, double __pyx_v_mz, double __pyx_v_error_tolerance, int __pyx_v_charge, int __pyx_v_left_search, int __pyx_v_right_search, double __pyx_v_charge_carrier, double __pyx_v_truncate_after, int __pyx_v_max_missed_peaks, double __pyx_v_threshold_scale, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_feature) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_feature_set __pyx_t_2;
  __Pyx_RefNannySetupContext("_fit_feature_set", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 7;
  __pyx_t_2.left_search = __pyx_v_left_search;
  __pyx_t_2.right_search = __pyx_v_right_search;
  __pyx_t_2.charge_carrier = __pyx_v_charge_carrier;
  __pyx_t_2.truncate_after = __pyx_v_truncate_after;
  __pyx_t_2.max_missed_peaks = __pyx_v_max_missed_peaks;
  __pyx_t_2.threshold_scale = __pyx_v_threshold_scale;
  __pyx_t_2.feature = __pyx_v_feature;
  __pyx_t_1 = __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase->_fit_feature_set(__pyx_v_self, __pyx_v_mz, __pyx_v_error_tolerance, __pyx_v_charge, 1, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 509, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase._fit_feature_set", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":523
 *         return feature_fits
 * 
 *     cpdef list _fit_composition(self, dict composition, double error_tolerance, int charge, double charge_carrier=PROTON,             # <<<<<<<<<<<<<<
 *                                 double truncate_after=0.8, double ignore_below=0.05, int max_missed_peaks=1,
 *                                 double threshold_scale=0.3, LCMSFeature feature=None):
 */

static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_15_fit_composition(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_composition(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, PyObject *__pyx_v_composition, double __pyx_v_error_tolerance, int __pyx_v_charge, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_composition *__pyx_optional_args) {
  double __pyx_v_charge_carrier = __pyx_k__8;
  double __pyx_v_truncate_after = ((double)0.8);
  double __pyx_v_ignore_below = ((double)0.05);
  int __pyx_v_max_missed_peaks = ((int)1);
  double __pyx_v_threshold_scale = ((double)0.3);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":525
 *     cpdef list _fit_composition(self, dict composition, double error_tolerance, int charge, double charge_carrier=PROTON,
 *                                 double truncate_after=0.8, double ignore_below=0.05, int max_missed_peaks=1,
 *                                 double threshold_scale=0.3, LCMSFeature feature=None):             # <<<<<<<<<<<<<<
 *         cdef:
 *             list base_tid, kept_tid, feature_fits
 */
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_feature = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)Py_None);
  PyObject *__pyx_v_base_tid = 0;
  PyObject *__pyx_v_kept_tid = 0;
  PyObject *__pyx_v_feature_fits = 0;
  double __pyx_v_total;
  double __pyx_v_cumsum;
  double __pyx_v_mz;
  struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_v_tp = 0;
  size_t __pyx_v_i;
  size_t __pyx_v_n;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution__isotopic_variants __pyx_t_14;
  size_t __pyx_t_15;
  size_t __pyx_t_16;
  size_t __pyx_t_17;
  PyObject *__pyx_t_18;
  int __pyx_t_19;
  int __pyx_t_20;
  double __pyx_t_21;
  struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_theoretical_distribution_on_features __pyx_t_22;
  __Pyx_RefNannySetupContext("_fit_composition", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_charge_carrier = __pyx_optional_args->charge_carrier;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_truncate_after = __pyx_optional_args->truncate_after;
        if (__pyx_optional_args->__pyx_n > 2) {
          __pyx_v_ignore_below = __pyx_optional_args->ignore_below;
          if (__pyx_optional_args->__pyx_n > 3) {
            __pyx_v_max_missed_peaks = __pyx_optional_args->max_missed_peaks;
            if (__pyx_optional_args->__pyx_n > 4) {
              __pyx_v_threshold_scale = __pyx_optional_args->threshold_scale;
              if (__pyx_optional_args->__pyx_n > 5) {
                __pyx_v_feature = __pyx_optional_args->feature;
              }
            }
          }
        }
      }
    }
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":523
 *         return feature_fits
 * 
 *     cpdef list _fit_composition(self, dict composition, double error_tolerance, int charge, double charge_carrier=PROTON,             # <<<<<<<<<<<<<<
 *                                 double truncate_after=0.8, double ignore_below=0.05, int max_missed_peaks=1,
 *                                 double threshold_scale=0.3, LCMSFeature feature=None):
 */
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_fit_composition); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 523, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_15_fit_composition)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_error_tolerance); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 523, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_charge); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 523, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_5 = PyFloat_FromDouble(__pyx_v_charge_carrier); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 523, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_6 = PyFloat_FromDouble(__pyx_v_truncate_after); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 523, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_7 = PyFloat_FromDouble(__pyx_v_ignore_below); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 523, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_max_missed_peaks); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 523, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_9 = PyFloat_FromDouble(__pyx_v_threshold_scale); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 523, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_10 = __pyx_t_1; __pyx_t_11 = NULL;
        __pyx_t_12 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_10))) {
          __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_10);
          if (likely(__pyx_t_11)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
            __Pyx_INCREF(__pyx_t_11);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_10, function);
            __pyx_t_12 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_10)) {
          PyObject *__pyx_temp[10] = {__pyx_t_11, __pyx_v_composition, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6, __pyx_t_7, __pyx_t_8, __pyx_t_9, ((PyObject *)__pyx_v_feature)};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_12, 9+__pyx_t_12); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 523, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_10)) {
          PyObject *__pyx_temp[10] = {__pyx_t_11, __pyx_v_composition, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6, __pyx_t_7, __pyx_t_8, __pyx_t_9, ((PyObject *)__pyx_v_feature)};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_12, 9+__pyx_t_12); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 523, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        } else
        #endif
        {
          __pyx_t_13 = PyTuple_New(9+__pyx_t_12); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 523, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_13);
          if (__pyx_t_11) {
            __Pyx_GIVEREF(__pyx_t_11); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_11); __pyx_t_11 = NULL;
          }
          __Pyx_INCREF(__pyx_v_composition);
          __Pyx_GIVEREF(__pyx_v_composition);
          PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_12, __pyx_v_composition);
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_12, __pyx_t_3);
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_13, 2+__pyx_t_12, __pyx_t_4);
          __Pyx_GIVEREF(__pyx_t_5);
          PyTuple_SET_ITEM(__pyx_t_13, 3+__pyx_t_12, __pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_6);
          PyTuple_SET_ITEM(__pyx_t_13, 4+__pyx_t_12, __pyx_t_6);
          __Pyx_GIVEREF(__pyx_t_7);
          PyTuple_SET_ITEM(__pyx_t_13, 5+__pyx_t_12, __pyx_t_7);
          __Pyx_GIVEREF(__pyx_t_8);
          PyTuple_SET_ITEM(__pyx_t_13, 6+__pyx_t_12, __pyx_t_8);
          __Pyx_GIVEREF(__pyx_t_9);
          PyTuple_SET_ITEM(__pyx_t_13, 7+__pyx_t_12, __pyx_t_9);
          __Pyx_INCREF(((PyObject *)__pyx_v_feature));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_feature));
          PyTuple_SET_ITEM(__pyx_t_13, 8+__pyx_t_12, ((PyObject *)__pyx_v_feature));
          __pyx_t_3 = 0;
          __pyx_t_4 = 0;
          __pyx_t_5 = 0;
          __pyx_t_6 = 0;
          __pyx_t_7 = 0;
          __pyx_t_8 = 0;
          __pyx_t_9 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_t_13, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 523, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        }
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 523, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":534
 *         # Can't refactor shared components with `create_theoretical_isotopic_distribution` because
 *         # this needs to include its own truncate-after component as well as ignore-under
 *         base_tid = _isotopic_variants(composition, npeaks=None, charge=charge, charge_carrier=charge_carrier)             # <<<<<<<<<<<<<<
 *         n = PyList_GET_SIZE(base_tid)
 *         total = 0
 */
  __pyx_t_14.__pyx_n = 3;
  __pyx_t_14.npeaks = Py_None;
  __pyx_t_14.charge = __pyx_v_charge;
  __pyx_t_14.charge_carrier = __pyx_v_charge_carrier;
  __pyx_t_1 = __pyx_f_7brainpy_2_c_21isotopic_distribution__isotopic_variants(__pyx_v_composition, 0, &__pyx_t_14); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_base_tid = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":535
 *         # this needs to include its own truncate-after component as well as ignore-under
 *         base_tid = _isotopic_variants(composition, npeaks=None, charge=charge, charge_carrier=charge_carrier)
 *         n = PyList_GET_SIZE(base_tid)             # <<<<<<<<<<<<<<
 *         total = 0
 *         cumsum = 0
 */
  __pyx_v_n = PyList_GET_SIZE(__pyx_v_base_tid);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":536
 *         base_tid = _isotopic_variants(composition, npeaks=None, charge=charge, charge_carrier=charge_carrier)
 *         n = PyList_GET_SIZE(base_tid)
 *         total = 0             # <<<<<<<<<<<<<<
 *         cumsum = 0
 *         kept_tid = []
 */
  __pyx_v_total = 0.0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":537
 *         n = PyList_GET_SIZE(base_tid)
 *         total = 0
 *         cumsum = 0             # <<<<<<<<<<<<<<
 *         kept_tid = []
 *         for i in range(n):
 */
  __pyx_v_cumsum = 0.0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":538
 *         total = 0
 *         cumsum = 0
 *         kept_tid = []             # <<<<<<<<<<<<<<
 *         for i in range(n):
 *             tp = <TheoreticalPeak>PyList_GET_ITEM(base_tid, i)
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 538, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_kept_tid = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":539
 *         cumsum = 0
 *         kept_tid = []
 *         for i in range(n):             # <<<<<<<<<<<<<<
 *             tp = <TheoreticalPeak>PyList_GET_ITEM(base_tid, i)
 *             cumsum += tp.intensity
 */
  __pyx_t_15 = __pyx_v_n;
  __pyx_t_16 = __pyx_t_15;
  for (__pyx_t_17 = 0; __pyx_t_17 < __pyx_t_16; __pyx_t_17+=1) {
    __pyx_v_i = __pyx_t_17;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":540
 *         kept_tid = []
 *         for i in range(n):
 *             tp = <TheoreticalPeak>PyList_GET_ITEM(base_tid, i)             # <<<<<<<<<<<<<<
 *             cumsum += tp.intensity
 *             if (tp.intensity < ignore_below) and (i > 1):
 */
    __pyx_t_18 = PyList_GET_ITEM(__pyx_v_base_tid, __pyx_v_i);
    __pyx_t_1 = ((PyObject *)__pyx_t_18);
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_tp, ((struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":541
 *         for i in range(n):
 *             tp = <TheoreticalPeak>PyList_GET_ITEM(base_tid, i)
 *             cumsum += tp.intensity             # <<<<<<<<<<<<<<
 *             if (tp.intensity < ignore_below) and (i > 1):
 *                 pass
 */
    __pyx_v_cumsum = (__pyx_v_cumsum + __pyx_v_tp->intensity);

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":542
 *             tp = <TheoreticalPeak>PyList_GET_ITEM(base_tid, i)
 *             cumsum += tp.intensity
 *             if (tp.intensity < ignore_below) and (i > 1):             # <<<<<<<<<<<<<<
 *                 pass
 *             else:
 */
    __pyx_t_20 = ((__pyx_v_tp->intensity < __pyx_v_ignore_below) != 0);
    if (__pyx_t_20) {
    } else {
      __pyx_t_19 = __pyx_t_20;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_20 = ((__pyx_v_i > 1) != 0);
    __pyx_t_19 = __pyx_t_20;
    __pyx_L6_bool_binop_done:;
    if (__pyx_t_19) {
      goto __pyx_L5;
    }

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":545
 *                 pass
 *             else:
 *                 PyList_Append(kept_tid, tp)             # <<<<<<<<<<<<<<
 *                 total += tp.intensity
 *             if cumsum >= truncate_after:
 */
    /*else*/ {
      __pyx_t_12 = PyList_Append(__pyx_v_kept_tid, ((PyObject *)__pyx_v_tp)); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 545, __pyx_L1_error)

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":546
 *             else:
 *                 PyList_Append(kept_tid, tp)
 *                 total += tp.intensity             # <<<<<<<<<<<<<<
 *             if cumsum >= truncate_after:
 *                 break
 */
      __pyx_v_total = (__pyx_v_total + __pyx_v_tp->intensity);
    }
    __pyx_L5:;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":547
 *                 PyList_Append(kept_tid, tp)
 *                 total += tp.intensity
 *             if cumsum >= truncate_after:             # <<<<<<<<<<<<<<
 *                 break
 * 
 */
    __pyx_t_19 = ((__pyx_v_cumsum >= __pyx_v_truncate_after) != 0);
    if (__pyx_t_19) {

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":548
 *                 total += tp.intensity
 *             if cumsum >= truncate_after:
 *                 break             # <<<<<<<<<<<<<<
 * 
 *         n = PyList_GET_SIZE(kept_tid)
 */
      goto __pyx_L4_break;

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":547
 *                 PyList_Append(kept_tid, tp)
 *                 total += tp.intensity
 *             if cumsum >= truncate_after:             # <<<<<<<<<<<<<<
 *                 break
 * 
 */
    }
  }
  __pyx_L4_break:;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":550
 *                 break
 * 
 *         n = PyList_GET_SIZE(kept_tid)             # <<<<<<<<<<<<<<
 *         for i in range(n):
 *             tp = <TheoreticalPeak>PyList_GET_ITEM(kept_tid, i)
 */
  __pyx_v_n = PyList_GET_SIZE(__pyx_v_kept_tid);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":551
 * 
 *         n = PyList_GET_SIZE(kept_tid)
 *         for i in range(n):             # <<<<<<<<<<<<<<
 *             tp = <TheoreticalPeak>PyList_GET_ITEM(kept_tid, i)
 *             tp.intensity /= total
 */
  __pyx_t_15 = __pyx_v_n;
  __pyx_t_16 = __pyx_t_15;
  for (__pyx_t_17 = 0; __pyx_t_17 < __pyx_t_16; __pyx_t_17+=1) {
    __pyx_v_i = __pyx_t_17;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":552
 *         n = PyList_GET_SIZE(kept_tid)
 *         for i in range(n):
 *             tp = <TheoreticalPeak>PyList_GET_ITEM(kept_tid, i)             # <<<<<<<<<<<<<<
 *             tp.intensity /= total
 * 
 */
    __pyx_t_18 = PyList_GET_ITEM(__pyx_v_kept_tid, __pyx_v_i);
    __pyx_t_1 = ((PyObject *)__pyx_t_18);
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_tp, ((struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":553
 *         for i in range(n):
 *             tp = <TheoreticalPeak>PyList_GET_ITEM(kept_tid, i)
 *             tp.intensity /= total             # <<<<<<<<<<<<<<
 * 
 *         if n == 0:
 */
    if (unlikely(__pyx_v_total == 0)) {
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      __PYX_ERR(0, 553, __pyx_L1_error)
    }
    __pyx_v_tp->intensity = (__pyx_v_tp->intensity / __pyx_v_total);
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":555
 *             tp.intensity /= total
 * 
 *         if n == 0:             # <<<<<<<<<<<<<<
 *             return []
 * 
 */
  __pyx_t_19 = ((__pyx_v_n == 0) != 0);
  if (__pyx_t_19) {

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":556
 * 
 *         if n == 0:
 *             return []             # <<<<<<<<<<<<<<
 * 
 *         tp = <TheoreticalPeak>PyList_GET_ITEM(kept_tid, 0)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 556, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":555
 *             tp.intensity /= total
 * 
 *         if n == 0:             # <<<<<<<<<<<<<<
 *             return []
 * 
 */
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":558
 *             return []
 * 
 *         tp = <TheoreticalPeak>PyList_GET_ITEM(kept_tid, 0)             # <<<<<<<<<<<<<<
 *         mz = tp.mz
 * 
 */
  __pyx_t_18 = PyList_GET_ITEM(__pyx_v_kept_tid, 0);
  __pyx_t_1 = ((PyObject *)__pyx_t_18);
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_XDECREF_SET(__pyx_v_tp, ((struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":559
 * 
 *         tp = <TheoreticalPeak>PyList_GET_ITEM(kept_tid, 0)
 *         mz = tp.mz             # <<<<<<<<<<<<<<
 * 
 *         feature_fits = self._fit_theoretical_distribution_on_features(
 */
  __pyx_t_21 = __pyx_v_tp->mz;
  __pyx_v_mz = __pyx_t_21;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":562
 * 
 *         feature_fits = self._fit_theoretical_distribution_on_features(
 *             mz, error_tolerance, charge, kept_tid,             # <<<<<<<<<<<<<<
 *             charge_carrier, truncate_after, max_missed_peaks, threshold_scale,
 *             feature)
 */
  if (!(likely(__Pyx_TypeTest(__pyx_v_kept_tid, __pyx_ptype_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern)))) __PYX_ERR(0, 562, __pyx_L1_error)

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":561
 *         mz = tp.mz
 * 
 *         feature_fits = self._fit_theoretical_distribution_on_features(             # <<<<<<<<<<<<<<
 *             mz, error_tolerance, charge, kept_tid,
 *             charge_carrier, truncate_after, max_missed_peaks, threshold_scale,
 */
  __pyx_t_22.__pyx_n = 5;
  __pyx_t_22.charge_carrier = __pyx_v_charge_carrier;
  __pyx_t_22.truncate_after = __pyx_v_truncate_after;
  __pyx_t_22.max_missed_peaks = __pyx_v_max_missed_peaks;
  __pyx_t_22.threshold_scale = __pyx_v_threshold_scale;
  __pyx_t_22.feature = __pyx_v_feature;
  __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)__pyx_v_self->__pyx_vtab)->_fit_theoretical_distribution_on_features(__pyx_v_self, __pyx_v_mz, __pyx_v_error_tolerance, __pyx_v_charge, ((struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *)__pyx_v_kept_tid), &__pyx_t_22); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 561, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_feature_fits = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":565
 *             charge_carrier, truncate_after, max_missed_peaks, threshold_scale,
 *             feature)
 *         return feature_fits             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_feature_fits);
  __pyx_r = __pyx_v_feature_fits;
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":523
 *         return feature_fits
 * 
 *     cpdef list _fit_composition(self, dict composition, double error_tolerance, int charge, double charge_carrier=PROTON,             # <<<<<<<<<<<<<<
 *                                 double truncate_after=0.8, double ignore_below=0.05, int max_missed_peaks=1,
 *                                 double threshold_scale=0.3, LCMSFeature feature=None):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase._fit_composition", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_base_tid);
  __Pyx_XDECREF(__pyx_v_kept_tid);
  __Pyx_XDECREF(__pyx_v_feature_fits);
  __Pyx_XDECREF((PyObject *)__pyx_v_tp);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_15_fit_composition(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_14_fit_composition[] = "LCMSFeatureProcessorBase._fit_composition(self, dict composition, double error_tolerance, int charge, double charge_carrier=PROTON, double truncate_after=0.8, double ignore_below=0.05, int max_missed_peaks=1, double threshold_scale=0.3, LCMSFeature feature=None) -> list";
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_15_fit_composition(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_composition = 0;
  double __pyx_v_error_tolerance;
  int __pyx_v_charge;
  double __pyx_v_charge_carrier;
  double __pyx_v_truncate_after;
  double __pyx_v_ignore_below;
  int __pyx_v_max_missed_peaks;
  double __pyx_v_threshold_scale;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_feature = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_fit_composition (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_composition,&__pyx_n_s_error_tolerance,&__pyx_n_s_charge,&__pyx_n_s_charge_carrier,&__pyx_n_s_truncate_after,&__pyx_n_s_ignore_below,&__pyx_n_s_max_missed_peaks,&__pyx_n_s_threshold_scale,&__pyx_n_s_feature,0};
    PyObject* values[9] = {0,0,0,0,0,0,0,0,0};

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":525
 *     cpdef list _fit_composition(self, dict composition, double error_tolerance, int charge, double charge_carrier=PROTON,
 *                                 double truncate_after=0.8, double ignore_below=0.05, int max_missed_peaks=1,
 *                                 double threshold_scale=0.3, LCMSFeature feature=None):             # <<<<<<<<<<<<<<
 *         cdef:
 *             list base_tid, kept_tid, feature_fits
 */
    values[8] = (PyObject *)((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_composition)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_error_tolerance)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_fit_composition", 0, 3, 9, 1); __PYX_ERR(0, 523, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_charge)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_fit_composition", 0, 3, 9, 2); __PYX_ERR(0, 523, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_charge_carrier);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_truncate_after);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ignore_below);
          if (value) { values[5] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_max_missed_peaks);
          if (value) { values[6] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_threshold_scale);
          if (value) { values[7] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_feature);
          if (value) { values[8] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_fit_composition") < 0)) __PYX_ERR(0, 523, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_composition = ((PyObject*)values[0]);
    __pyx_v_error_tolerance = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_error_tolerance == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 523, __pyx_L3_error)
    __pyx_v_charge = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_charge == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 523, __pyx_L3_error)
    if (values[3]) {
      __pyx_v_charge_carrier = __pyx_PyFloat_AsDouble(values[3]); if (unlikely((__pyx_v_charge_carrier == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 523, __pyx_L3_error)
    } else {
      __pyx_v_charge_carrier = __pyx_k__8;
    }
    if (values[4]) {
      __pyx_v_truncate_after = __pyx_PyFloat_AsDouble(values[4]); if (unlikely((__pyx_v_truncate_after == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 524, __pyx_L3_error)
    } else {
      __pyx_v_truncate_after = ((double)0.8);
    }
    if (values[5]) {
      __pyx_v_ignore_below = __pyx_PyFloat_AsDouble(values[5]); if (unlikely((__pyx_v_ignore_below == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 524, __pyx_L3_error)
    } else {
      __pyx_v_ignore_below = ((double)0.05);
    }
    if (values[6]) {
      __pyx_v_max_missed_peaks = __Pyx_PyInt_As_int(values[6]); if (unlikely((__pyx_v_max_missed_peaks == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 524, __pyx_L3_error)
    } else {
      __pyx_v_max_missed_peaks = ((int)1);
    }
    if (values[7]) {
      __pyx_v_threshold_scale = __pyx_PyFloat_AsDouble(values[7]); if (unlikely((__pyx_v_threshold_scale == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 525, __pyx_L3_error)
    } else {
      __pyx_v_threshold_scale = ((double)0.3);
    }
    __pyx_v_feature = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)values[8]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_fit_composition", 0, 3, 9, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 523, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase._fit_composition", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_composition), (&PyDict_Type), 1, "composition", 1))) __PYX_ERR(0, 523, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_feature), __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature, 1, "feature", 0))) __PYX_ERR(0, 525, __pyx_L1_error)
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_14_fit_composition(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)__pyx_v_self), __pyx_v_composition, __pyx_v_error_tolerance, __pyx_v_charge, __pyx_v_charge_carrier, __pyx_v_truncate_after, __pyx_v_ignore_below, __pyx_v_max_missed_peaks, __pyx_v_threshold_scale, __pyx_v_feature);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":523
 *         return feature_fits
 * 
 *     cpdef list _fit_composition(self, dict composition, double error_tolerance, int charge, double charge_carrier=PROTON,             # <<<<<<<<<<<<<<
 *                                 double truncate_after=0.8, double ignore_below=0.05, int max_missed_peaks=1,
 *                                 double threshold_scale=0.3, LCMSFeature feature=None):
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_14_fit_composition(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, PyObject *__pyx_v_composition, double __pyx_v_error_tolerance, int __pyx_v_charge, double __pyx_v_charge_carrier, double __pyx_v_truncate_after, double __pyx_v_ignore_below, int __pyx_v_max_missed_peaks, double __pyx_v_threshold_scale, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_feature) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_composition __pyx_t_2;
  __Pyx_RefNannySetupContext("_fit_composition", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 6;
  __pyx_t_2.charge_carrier = __pyx_v_charge_carrier;
  __pyx_t_2.truncate_after = __pyx_v_truncate_after;
  __pyx_t_2.ignore_below = __pyx_v_ignore_below;
  __pyx_t_2.max_missed_peaks = __pyx_v_max_missed_peaks;
  __pyx_t_2.threshold_scale = __pyx_v_threshold_scale;
  __pyx_t_2.feature = __pyx_v_feature;
  __pyx_t_1 = __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase->_fit_composition(__pyx_v_self, __pyx_v_composition, __pyx_v_error_tolerance, __pyx_v_charge, 1, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 523, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase._fit_composition", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":242
 * cdef class LCMSFeatureProcessorBase(object):
 *     cdef:
 *         public LCMSFeatureMap feature_map             # <<<<<<<<<<<<<<
 *         public IsotopicFitterBase scorer
 *         public AveragineCache averagine
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_11feature_map_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_11feature_map_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_11feature_map___get__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_11feature_map___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->feature_map));
  __pyx_r = ((PyObject *)__pyx_v_self->feature_map);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_11feature_map_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_11feature_map_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_11feature_map_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_11feature_map_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap))))) __PYX_ERR(0, 242, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->feature_map);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->feature_map));
  __pyx_v_self->feature_map = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase.feature_map.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_11feature_map_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_11feature_map_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_11feature_map_4__del__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_11feature_map_4__del__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->feature_map);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->feature_map));
  __pyx_v_self->feature_map = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":243
 *     cdef:
 *         public LCMSFeatureMap feature_map
 *         public IsotopicFitterBase scorer             # <<<<<<<<<<<<<<
 *         public AveragineCache averagine
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_6scorer_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_6scorer_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_6scorer___get__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_6scorer___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->scorer));
  __pyx_r = ((PyObject *)__pyx_v_self->scorer);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_6scorer_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_6scorer_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_6scorer_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_6scorer_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitterBase))))) __PYX_ERR(0, 243, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->scorer);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->scorer));
  __pyx_v_self->scorer = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase.scorer.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_6scorer_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_6scorer_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_6scorer_4__del__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_6scorer_4__del__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->scorer);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->scorer));
  __pyx_v_self->scorer = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":244
 *         public LCMSFeatureMap feature_map
 *         public IsotopicFitterBase scorer
 *         public AveragineCache averagine             # <<<<<<<<<<<<<<
 * 
 *     cpdef list create_theoretical_distribution(self, double mz, int charge, double charge_carrier=PROTON, double truncate_after=0.8,
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_9averagine_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_9averagine_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_9averagine___get__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_9averagine___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->averagine));
  __pyx_r = ((PyObject *)__pyx_v_self->averagine);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_9averagine_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_9averagine_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_9averagine_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_9averagine_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_12ms_deisotope_2_c_9averagine_AveragineCache))))) __PYX_ERR(0, 244, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->averagine);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->averagine));
  __pyx_v_self->averagine = ((struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase.averagine.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_9averagine_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_9averagine_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_9averagine_4__del__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_9averagine_4__del__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->averagine);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->averagine));
  __pyx_v_self->averagine = ((struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_17__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_16__reduce_cython__[] = "LCMSFeatureProcessorBase.__reduce_cython__(self)";
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_17__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_16__reduce_cython__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_16__reduce_cython__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.averagine, self.feature_map, self.scorer)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->averagine));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->averagine));
  PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self->averagine));
  __Pyx_INCREF(((PyObject *)__pyx_v_self->feature_map));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->feature_map));
  PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_self->feature_map));
  __Pyx_INCREF(((PyObject *)__pyx_v_self->scorer));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->scorer));
  PyTuple_SET_ITEM(__pyx_t_1, 2, ((PyObject *)__pyx_v_self->scorer));
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.averagine, self.feature_map, self.scorer)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.averagine, self.feature_map, self.scorer)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.averagine is not None or self.feature_map is not None or self.scorer is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.averagine, self.feature_map, self.scorer)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.averagine is not None or self.feature_map is not None or self.scorer is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_LCMSFeatureProcessorBase, (type(self), 0xa9723cd, None), state
 */
  /*else*/ {
    __pyx_t_2 = (((PyObject *)__pyx_v_self->averagine) != Py_None);
    __pyx_t_5 = (__pyx_t_2 != 0);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_3 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (((PyObject *)__pyx_v_self->feature_map) != Py_None);
    __pyx_t_2 = (__pyx_t_5 != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_3 = __pyx_t_2;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_2 = (((PyObject *)__pyx_v_self->scorer) != Py_None);
    __pyx_t_5 = (__pyx_t_2 != 0);
    __pyx_t_3 = __pyx_t_5;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_3;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.averagine is not None or self.feature_map is not None or self.scorer is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_LCMSFeatureProcessorBase, (type(self), 0xa9723cd, None), state
 *     else:
 */
  __pyx_t_3 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":13
 *         use_setstate = self.averagine is not None or self.feature_map is not None or self.scorer is not None
 *     if use_setstate:
 *         return __pyx_unpickle_LCMSFeatureProcessorBase, (type(self), 0xa9723cd, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_LCMSFeatureProcessorBase, (type(self), 0xa9723cd, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_LCMSFeatureProces); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_177677261);
    __Pyx_GIVEREF(__pyx_int_177677261);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_177677261);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_1);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_v_state);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.averagine is not None or self.feature_map is not None or self.scorer is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_LCMSFeatureProcessorBase, (type(self), 0xa9723cd, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_LCMSFeatureProcessorBase, (type(self), 0xa9723cd, None), state
 *     else:
 *         return __pyx_unpickle_LCMSFeatureProcessorBase, (type(self), 0xa9723cd, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_LCMSFeatureProcessorBase__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_pyx_unpickle_LCMSFeatureProces); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_177677261);
    __Pyx_GIVEREF(__pyx_int_177677261);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_177677261);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
    __pyx_t_6 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_LCMSFeatureProcessorBase, (type(self), 0xa9723cd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_LCMSFeatureProcessorBase__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_19__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_18__setstate_cython__[] = "LCMSFeatureProcessorBase.__setstate_cython__(self, __pyx_state)";
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_19__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_18__setstate_cython__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_18__setstate_cython__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_LCMSFeatureProcessorBase, (type(self), 0xa9723cd, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_LCMSFeatureProcessorBase__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor___pyx_unpickle_LCMSFeatureProcessorBase__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_LCMSFeatureProcessorBase, (type(self), 0xa9723cd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_LCMSFeatureProcessorBase__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_envelope_conformer(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_3__pyx_unpickle_envelope_conformer(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_2__pyx_unpickle_envelope_conformer[] = "__pyx_unpickle_envelope_conformer(__pyx_type, long __pyx_checksum, __pyx_state)";
static PyMethodDef __pyx_mdef_12ms_deisotope_2_c_11feature_map_17feature_processor_3__pyx_unpickle_envelope_conformer = {"__pyx_unpickle_envelope_conformer", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_3__pyx_unpickle_envelope_conformer, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_2__pyx_unpickle_envelope_conformer};
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_3__pyx_unpickle_envelope_conformer(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_envelope_conformer (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_envelope_conformer", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_envelope_conformer", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_envelope_conformer") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_envelope_conformer", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.__pyx_unpickle_envelope_conformer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_2__pyx_unpickle_envelope_conformer(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_2__pyx_unpickle_envelope_conformer(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__pyx_unpickle_envelope_conformer", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xba9b7dc:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xba9b7dc = (experimental, n_missing, theoretical))" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0xba9b7dc) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xba9b7dc:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xba9b7dc = (experimental, n_missing, theoretical))" % __pyx_checksum)
 *     __pyx_result = envelope_conformer.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, -1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0xba9b7dc:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xba9b7dc = (experimental, n_missing, theoretical))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = envelope_conformer.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0xba, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xba9b7dc:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xba9b7dc = (experimental, n_missing, theoretical))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xba9b7dc = (experimental, n_missing, theoretical))" % __pyx_checksum)
 *     __pyx_result = envelope_conformer.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_envelope_conformer__set_state(<envelope_conformer> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xba9b7dc = (experimental, n_missing, theoretical))" % __pyx_checksum)
 *     __pyx_result = envelope_conformer.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_envelope_conformer__set_state(<envelope_conformer> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = envelope_conformer.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_envelope_conformer__set_state(<envelope_conformer> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_envelope_conformer__set_state(envelope_conformer __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor___pyx_unpickle_envelope_conformer__set_state(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xba9b7dc = (experimental, n_missing, theoretical))" % __pyx_checksum)
 *     __pyx_result = envelope_conformer.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_envelope_conformer__set_state(<envelope_conformer> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_envelope_conformer__set_state(<envelope_conformer> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_envelope_conformer__set_state(envelope_conformer __pyx_result, tuple __pyx_state):
 *     __pyx_result.experimental = __pyx_state[0]; __pyx_result.n_missing = __pyx_state[1]; __pyx_result.theoretical = __pyx_state[2]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_envelope_conformer(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.__pyx_unpickle_envelope_conformer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_envelope_conformer__set_state(<envelope_conformer> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_envelope_conformer__set_state(envelope_conformer __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.experimental = __pyx_state[0]; __pyx_result.n_missing = __pyx_state[1]; __pyx_result.theoretical = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor___pyx_unpickle_envelope_conformer__set_state(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  __Pyx_RefNannySetupContext("__pyx_unpickle_envelope_conformer__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_envelope_conformer__set_state(envelope_conformer __pyx_result, tuple __pyx_state):
 *     __pyx_result.experimental = __pyx_state[0]; __pyx_result.n_missing = __pyx_state[1]; __pyx_result.theoretical = __pyx_state[2]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyList_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->experimental);
  __Pyx_DECREF(__pyx_v___pyx_result->experimental);
  __pyx_v___pyx_result->experimental = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_size_t(__pyx_t_1); if (unlikely((__pyx_t_2 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->n_missing = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyList_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->theoretical);
  __Pyx_DECREF(__pyx_v___pyx_result->theoretical);
  __pyx_v___pyx_result->theoretical = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_envelope_conformer__set_state(envelope_conformer __pyx_result, tuple __pyx_state):
 *     __pyx_result.experimental = __pyx_state[0]; __pyx_result.n_missing = __pyx_state[1]; __pyx_result.theoretical = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_4 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_5 = ((__pyx_t_4 > 3) != 0);
  if (__pyx_t_5) {
  } else {
    __pyx_t_3 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_5 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_6 = (__pyx_t_5 != 0);
  __pyx_t_3 = __pyx_t_6;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_3) {

    /* "(tree fragment)":14
 *     __pyx_result.experimental = __pyx_state[0]; __pyx_result.n_missing = __pyx_state[1]; __pyx_result.theoretical = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])             # <<<<<<<<<<<<<<
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_update); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_7 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
      }
    }
    __pyx_t_1 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_9, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_7);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_envelope_conformer__set_state(envelope_conformer __pyx_result, tuple __pyx_state):
 *     __pyx_result.experimental = __pyx_state[0]; __pyx_result.n_missing = __pyx_state[1]; __pyx_result.theoretical = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_envelope_conformer__set_state(<envelope_conformer> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_envelope_conformer__set_state(envelope_conformer __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.experimental = __pyx_state[0]; __pyx_result.n_missing = __pyx_state[1]; __pyx_result.theoretical = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.__pyx_unpickle_envelope_conformer__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_LCMSFeatureProcessorBase(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_5__pyx_unpickle_LCMSFeatureProcessorBase(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_4__pyx_unpickle_LCMSFeatureProcessorBase[] = "__pyx_unpickle_LCMSFeatureProcessorBase(__pyx_type, long __pyx_checksum, __pyx_state)";
static PyMethodDef __pyx_mdef_12ms_deisotope_2_c_11feature_map_17feature_processor_5__pyx_unpickle_LCMSFeatureProcessorBase = {"__pyx_unpickle_LCMSFeatureProcessorBase", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_5__pyx_unpickle_LCMSFeatureProcessorBase, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_4__pyx_unpickle_LCMSFeatureProcessorBase};
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_5__pyx_unpickle_LCMSFeatureProcessorBase(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_LCMSFeatureProcessorBase (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_LCMSFeatureProcessorBase", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_LCMSFeatureProcessorBase", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_LCMSFeatureProcessorBase") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_LCMSFeatureProcessorBase", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.__pyx_unpickle_LCMSFeatureProcessorBase", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_4__pyx_unpickle_LCMSFeatureProcessorBase(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_4__pyx_unpickle_LCMSFeatureProcessorBase(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__pyx_unpickle_LCMSFeatureProcessorBase", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xa9723cd:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xa9723cd = (averagine, feature_map, scorer))" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0xa9723cd) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xa9723cd:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xa9723cd = (averagine, feature_map, scorer))" % __pyx_checksum)
 *     __pyx_result = LCMSFeatureProcessorBase.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, -1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0xa9723cd:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xa9723cd = (averagine, feature_map, scorer))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = LCMSFeatureProcessorBase.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0xa9, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xa9723cd:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xa9723cd = (averagine, feature_map, scorer))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xa9723cd = (averagine, feature_map, scorer))" % __pyx_checksum)
 *     __pyx_result = LCMSFeatureProcessorBase.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_LCMSFeatureProcessorBase__set_state(<LCMSFeatureProcessorBase> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xa9723cd = (averagine, feature_map, scorer))" % __pyx_checksum)
 *     __pyx_result = LCMSFeatureProcessorBase.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_LCMSFeatureProcessorBase__set_state(<LCMSFeatureProcessorBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = LCMSFeatureProcessorBase.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_LCMSFeatureProcessorBase__set_state(<LCMSFeatureProcessorBase> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_LCMSFeatureProcessorBase__set_state(LCMSFeatureProcessorBase __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor___pyx_unpickle_LCMSFeatureProcessorBase__set_state(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xa9723cd = (averagine, feature_map, scorer))" % __pyx_checksum)
 *     __pyx_result = LCMSFeatureProcessorBase.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_LCMSFeatureProcessorBase__set_state(<LCMSFeatureProcessorBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_LCMSFeatureProcessorBase__set_state(<LCMSFeatureProcessorBase> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_LCMSFeatureProcessorBase__set_state(LCMSFeatureProcessorBase __pyx_result, tuple __pyx_state):
 *     __pyx_result.averagine = __pyx_state[0]; __pyx_result.feature_map = __pyx_state[1]; __pyx_result.scorer = __pyx_state[2]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_LCMSFeatureProcessorBase(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.__pyx_unpickle_LCMSFeatureProcessorBase", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_LCMSFeatureProcessorBase__set_state(<LCMSFeatureProcessorBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_LCMSFeatureProcessorBase__set_state(LCMSFeatureProcessorBase __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.averagine = __pyx_state[0]; __pyx_result.feature_map = __pyx_state[1]; __pyx_result.scorer = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor___pyx_unpickle_LCMSFeatureProcessorBase__set_state(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("__pyx_unpickle_LCMSFeatureProcessorBase__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_LCMSFeatureProcessorBase__set_state(LCMSFeatureProcessorBase __pyx_result, tuple __pyx_state):
 *     __pyx_result.averagine = __pyx_state[0]; __pyx_result.feature_map = __pyx_state[1]; __pyx_result.scorer = __pyx_state[2]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_12ms_deisotope_2_c_9averagine_AveragineCache))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->averagine);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->averagine));
  __pyx_v___pyx_result->averagine = ((struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->feature_map);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->feature_map));
  __pyx_v___pyx_result->feature_map = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitterBase))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->scorer);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->scorer));
  __pyx_v___pyx_result->scorer = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_LCMSFeatureProcessorBase__set_state(LCMSFeatureProcessorBase __pyx_result, tuple __pyx_state):
 *     __pyx_result.averagine = __pyx_state[0]; __pyx_result.feature_map = __pyx_state[1]; __pyx_result.scorer = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_3 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_4 = ((__pyx_t_3 > 3) != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_4 != 0);
  __pyx_t_2 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.averagine = __pyx_state[0]; __pyx_result.feature_map = __pyx_state[1]; __pyx_result.scorer = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])             # <<<<<<<<<<<<<<
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_LCMSFeatureProcessorBase__set_state(LCMSFeatureProcessorBase __pyx_result, tuple __pyx_state):
 *     __pyx_result.averagine = __pyx_state[0]; __pyx_result.feature_map = __pyx_state[1]; __pyx_result.scorer = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_LCMSFeatureProcessorBase__set_state(<LCMSFeatureProcessorBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_LCMSFeatureProcessorBase__set_state(LCMSFeatureProcessorBase __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.averagine = __pyx_state[0]; __pyx_result.feature_map = __pyx_state[1]; __pyx_result.scorer = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.__pyx_unpickle_LCMSFeatureProcessorBase__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":258
 *         # experimental exception made for __getbuffer__ and __releasebuffer__
 *         # -- the details of this may change.
 *         def __getbuffer__(ndarray self, Py_buffer* info, int flags):             # <<<<<<<<<<<<<<
 *             # This implementation of getbuffer is geared towards Cython
 *             # requirements, and does not yet fulfill the PEP.
 */

/* Python wrapper */
static CYTHON_UNUSED int __pyx_pw_5numpy_7ndarray_1__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
static CYTHON_UNUSED int __pyx_pw_5numpy_7ndarray_1__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5numpy_7ndarray___getbuffer__(((PyArrayObject *)__pyx_v_self), ((Py_buffer *)__pyx_v_info), ((int)__pyx_v_flags));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5numpy_7ndarray___getbuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_v_i;
  int __pyx_v_ndim;
  int __pyx_v_endian_detector;
  int __pyx_v_little_endian;
  int __pyx_v_t;
  char *__pyx_v_f;
  PyArray_Descr *__pyx_v_descr = 0;
  int __pyx_v_offset;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyArray_Descr *__pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  char *__pyx_t_9;
  if (__pyx_v_info == NULL) {
    PyErr_SetString(PyExc_BufferError, "PyObject_GetBuffer: view==NULL argument is obsolete");
    return -1;
  }
  __Pyx_RefNannySetupContext("__getbuffer__", 0);
  __pyx_v_info->obj = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(__pyx_v_info->obj);

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":265
 * 
 *             cdef int i, ndim
 *             cdef int endian_detector = 1             # <<<<<<<<<<<<<<
 *             cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)
 * 
 */
  __pyx_v_endian_detector = 1;

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":266
 *             cdef int i, ndim
 *             cdef int endian_detector = 1
 *             cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)             # <<<<<<<<<<<<<<
 * 
 *             ndim = PyArray_NDIM(self)
 */
  __pyx_v_little_endian = ((((char *)(&__pyx_v_endian_detector))[0]) != 0);

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":268
 *             cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)
 * 
 *             ndim = PyArray_NDIM(self)             # <<<<<<<<<<<<<<
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 */
  __pyx_v_ndim = PyArray_NDIM(__pyx_v_self);

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":270
 *             ndim = PyArray_NDIM(self)
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")
 */
  __pyx_t_2 = (((__pyx_v_flags & PyBUF_C_CONTIGUOUS) == PyBUF_C_CONTIGUOUS) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":271
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):             # <<<<<<<<<<<<<<
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 */
  __pyx_t_2 = ((!(PyArray_CHKFLAGS(__pyx_v_self, NPY_ARRAY_C_CONTIGUOUS) != 0)) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":270
 *             ndim = PyArray_NDIM(self)
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")
 */
  if (unlikely(__pyx_t_1)) {

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":272
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")             # <<<<<<<<<<<<<<
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__9, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 272, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 272, __pyx_L1_error)

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":270
 *             ndim = PyArray_NDIM(self)
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")
 */
  }

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":274
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 */
  __pyx_t_2 = (((__pyx_v_flags & PyBUF_F_CONTIGUOUS) == PyBUF_F_CONTIGUOUS) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L7_bool_binop_done;
  }

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":275
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):             # <<<<<<<<<<<<<<
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 * 
 */
  __pyx_t_2 = ((!(PyArray_CHKFLAGS(__pyx_v_self, NPY_ARRAY_F_CONTIGUOUS) != 0)) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L7_bool_binop_done:;

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":274
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 */
  if (unlikely(__pyx_t_1)) {

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":276
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")             # <<<<<<<<<<<<<<
 * 
 *             info.buf = PyArray_DATA(self)
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__10, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 276, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 276, __pyx_L1_error)

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":274
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 */
  }

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":278
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 * 
 *             info.buf = PyArray_DATA(self)             # <<<<<<<<<<<<<<
 *             info.ndim = ndim
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 */
  __pyx_v_info->buf = PyArray_DATA(__pyx_v_self);

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":279
 * 
 *             info.buf = PyArray_DATA(self)
 *             info.ndim = ndim             # <<<<<<<<<<<<<<
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 *                 # Allocate new buffer for strides and shape info.
 */
  __pyx_v_info->ndim = __pyx_v_ndim;

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":280
 *             info.buf = PyArray_DATA(self)
 *             info.ndim = ndim
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 # Allocate new buffer for strides and shape info.
 *                 # This is allocated as one block, strides first.
 */
  __pyx_t_1 = (((sizeof(npy_intp)) != (sizeof(Py_ssize_t))) != 0);
  if (__pyx_t_1) {

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":283
 *                 # Allocate new buffer for strides and shape info.
 *                 # This is allocated as one block, strides first.
 *                 info.strides = <Py_ssize_t*>PyObject_Malloc(sizeof(Py_ssize_t) * 2 * <size_t>ndim)             # <<<<<<<<<<<<<<
 *                 info.shape = info.strides + ndim
 *                 for i in range(ndim):
 */
    __pyx_v_info->strides = ((Py_ssize_t *)PyObject_Malloc((((sizeof(Py_ssize_t)) * 2) * ((size_t)__pyx_v_ndim))));

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":284
 *                 # This is allocated as one block, strides first.
 *                 info.strides = <Py_ssize_t*>PyObject_Malloc(sizeof(Py_ssize_t) * 2 * <size_t>ndim)
 *                 info.shape = info.strides + ndim             # <<<<<<<<<<<<<<
 *                 for i in range(ndim):
 *                     info.strides[i] = PyArray_STRIDES(self)[i]
 */
    __pyx_v_info->shape = (__pyx_v_info->strides + __pyx_v_ndim);

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":285
 *                 info.strides = <Py_ssize_t*>PyObject_Malloc(sizeof(Py_ssize_t) * 2 * <size_t>ndim)
 *                 info.shape = info.strides + ndim
 *                 for i in range(ndim):             # <<<<<<<<<<<<<<
 *                     info.strides[i] = PyArray_STRIDES(self)[i]
 *                     info.shape[i] = PyArray_DIMS(self)[i]
 */
    __pyx_t_4 = __pyx_v_ndim;
    __pyx_t_5 = __pyx_t_4;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
      __pyx_v_i = __pyx_t_6;

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":286
 *                 info.shape = info.strides + ndim
 *                 for i in range(ndim):
 *                     info.strides[i] = PyArray_STRIDES(self)[i]             # <<<<<<<<<<<<<<
 *                     info.shape[i] = PyArray_DIMS(self)[i]
 *             else:
 */
      (__pyx_v_info->strides[__pyx_v_i]) = (PyArray_STRIDES(__pyx_v_self)[__pyx_v_i]);

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":287
 *                 for i in range(ndim):
 *                     info.strides[i] = PyArray_STRIDES(self)[i]
 *                     info.shape[i] = PyArray_DIMS(self)[i]             # <<<<<<<<<<<<<<
 *             else:
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)
 */
      (__pyx_v_info->shape[__pyx_v_i]) = (PyArray_DIMS(__pyx_v_self)[__pyx_v_i]);
    }

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":280
 *             info.buf = PyArray_DATA(self)
 *             info.ndim = ndim
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 # Allocate new buffer for strides and shape info.
 *                 # This is allocated as one block, strides first.
 */
    goto __pyx_L9;
  }

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":289
 *                     info.shape[i] = PyArray_DIMS(self)[i]
 *             else:
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)             # <<<<<<<<<<<<<<
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)
 *             info.suboffsets = NULL
 */
  /*else*/ {
    __pyx_v_info->strides = ((Py_ssize_t *)PyArray_STRIDES(__pyx_v_self));

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":290
 *             else:
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)             # <<<<<<<<<<<<<<
 *             info.suboffsets = NULL
 *             info.itemsize = PyArray_ITEMSIZE(self)
 */
    __pyx_v_info->shape = ((Py_ssize_t *)PyArray_DIMS(__pyx_v_self));
  }
  __pyx_L9:;

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":291
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)
 *             info.suboffsets = NULL             # <<<<<<<<<<<<<<
 *             info.itemsize = PyArray_ITEMSIZE(self)
 *             info.readonly = not PyArray_ISWRITEABLE(self)
 */
  __pyx_v_info->suboffsets = NULL;

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":292
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)
 *             info.suboffsets = NULL
 *             info.itemsize = PyArray_ITEMSIZE(self)             # <<<<<<<<<<<<<<
 *             info.readonly = not PyArray_ISWRITEABLE(self)
 * 
 */
  __pyx_v_info->itemsize = PyArray_ITEMSIZE(__pyx_v_self);

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":293
 *             info.suboffsets = NULL
 *             info.itemsize = PyArray_ITEMSIZE(self)
 *             info.readonly = not PyArray_ISWRITEABLE(self)             # <<<<<<<<<<<<<<
 * 
 *             cdef int t
 */
  __pyx_v_info->readonly = (!(PyArray_ISWRITEABLE(__pyx_v_self) != 0));

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":296
 * 
 *             cdef int t
 *             cdef char* f = NULL             # <<<<<<<<<<<<<<
 *             cdef dtype descr = <dtype>PyArray_DESCR(self)
 *             cdef int offset
 */
  __pyx_v_f = NULL;

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":297
 *             cdef int t
 *             cdef char* f = NULL
 *             cdef dtype descr = <dtype>PyArray_DESCR(self)             # <<<<<<<<<<<<<<
 *             cdef int offset
 * 
 */
  __pyx_t_7 = PyArray_DESCR(__pyx_v_self);
  __pyx_t_3 = ((PyObject *)__pyx_t_7);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_descr = ((PyArray_Descr *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":300
 *             cdef int offset
 * 
 *             info.obj = self             # <<<<<<<<<<<<<<
 * 
 *             if not PyDataType_HASFIELDS(descr):
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  __Pyx_GOTREF(__pyx_v_info->obj);
  __Pyx_DECREF(__pyx_v_info->obj);
  __pyx_v_info->obj = ((PyObject *)__pyx_v_self);

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":302
 *             info.obj = self
 * 
 *             if not PyDataType_HASFIELDS(descr):             # <<<<<<<<<<<<<<
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or
 */
  __pyx_t_1 = ((!(PyDataType_HASFIELDS(__pyx_v_descr) != 0)) != 0);
  if (__pyx_t_1) {

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":303
 * 
 *             if not PyDataType_HASFIELDS(descr):
 *                 t = descr.type_num             # <<<<<<<<<<<<<<
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):
 */
    __pyx_t_4 = __pyx_v_descr->type_num;
    __pyx_v_t = __pyx_t_4;

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":304
 *             if not PyDataType_HASFIELDS(descr):
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 */
    __pyx_t_2 = ((__pyx_v_descr->byteorder == '>') != 0);
    if (!__pyx_t_2) {
      goto __pyx_L15_next_or;
    } else {
    }
    __pyx_t_2 = (__pyx_v_little_endian != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L14_bool_binop_done;
    }
    __pyx_L15_next_or:;

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":305
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):             # <<<<<<<<<<<<<<
 *                     raise ValueError(u"Non-native byte order not supported")
 *                 if   t == NPY_BYTE:        f = "b"
 */
    __pyx_t_2 = ((__pyx_v_descr->byteorder == '<') != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L14_bool_binop_done;
    }
    __pyx_t_2 = ((!(__pyx_v_little_endian != 0)) != 0);
    __pyx_t_1 = __pyx_t_2;
    __pyx_L14_bool_binop_done:;

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":304
 *             if not PyDataType_HASFIELDS(descr):
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 */
    if (unlikely(__pyx_t_1)) {

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":306
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__11, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 306, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(2, 306, __pyx_L1_error)

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":304
 *             if not PyDataType_HASFIELDS(descr):
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 */
    }

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":307
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 *                 if   t == NPY_BYTE:        f = "b"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_UBYTE:       f = "B"
 *                 elif t == NPY_SHORT:       f = "h"
 */
    switch (__pyx_v_t) {
      case NPY_BYTE:
      __pyx_v_f = ((char *)"b");
      break;
      case NPY_UBYTE:

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":308
 *                     raise ValueError(u"Non-native byte order not supported")
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_SHORT:       f = "h"
 *                 elif t == NPY_USHORT:      f = "H"
 */
      __pyx_v_f = ((char *)"B");
      break;
      case NPY_SHORT:

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":309
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"
 *                 elif t == NPY_SHORT:       f = "h"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_USHORT:      f = "H"
 *                 elif t == NPY_INT:         f = "i"
 */
      __pyx_v_f = ((char *)"h");
      break;
      case NPY_USHORT:

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":310
 *                 elif t == NPY_UBYTE:       f = "B"
 *                 elif t == NPY_SHORT:       f = "h"
 *                 elif t == NPY_USHORT:      f = "H"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_INT:         f = "i"
 *                 elif t == NPY_UINT:        f = "I"
 */
      __pyx_v_f = ((char *)"H");
      break;
      case NPY_INT:

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":311
 *                 elif t == NPY_SHORT:       f = "h"
 *                 elif t == NPY_USHORT:      f = "H"
 *                 elif t == NPY_INT:         f = "i"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_UINT:        f = "I"
 *                 elif t == NPY_LONG:        f = "l"
 */
      __pyx_v_f = ((char *)"i");
      break;
      case NPY_UINT:

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":312
 *                 elif t == NPY_USHORT:      f = "H"
 *                 elif t == NPY_INT:         f = "i"
 *                 elif t == NPY_UINT:        f = "I"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_LONG:        f = "l"
 *                 elif t == NPY_ULONG:       f = "L"
 */
      __pyx_v_f = ((char *)"I");
      break;
      case NPY_LONG:

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":313
 *                 elif t == NPY_INT:         f = "i"
 *                 elif t == NPY_UINT:        f = "I"
 *                 elif t == NPY_LONG:        f = "l"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_ULONG:       f = "L"
 *                 elif t == NPY_LONGLONG:    f = "q"
 */
      __pyx_v_f = ((char *)"l");
      break;
      case NPY_ULONG:

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":314
 *                 elif t == NPY_UINT:        f = "I"
 *                 elif t == NPY_LONG:        f = "l"
 *                 elif t == NPY_ULONG:       f = "L"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_LONGLONG:    f = "q"
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 */
      __pyx_v_f = ((char *)"L");
      break;
      case NPY_LONGLONG:

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":315
 *                 elif t == NPY_LONG:        f = "l"
 *                 elif t == NPY_ULONG:       f = "L"
 *                 elif t == NPY_LONGLONG:    f = "q"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 *                 elif t == NPY_FLOAT:       f = "f"
 */
      __pyx_v_f = ((char *)"q");
      break;
      case NPY_ULONGLONG:

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":316
 *                 elif t == NPY_ULONG:       f = "L"
 *                 elif t == NPY_LONGLONG:    f = "q"
 *                 elif t == NPY_ULONGLONG:   f = "Q"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_FLOAT:       f = "f"
 *                 elif t == NPY_DOUBLE:      f = "d"
 */
      __pyx_v_f = ((char *)"Q");
      break;
      case NPY_FLOAT:

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":317
 *                 elif t == NPY_LONGLONG:    f = "q"
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 *                 elif t == NPY_FLOAT:       f = "f"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_DOUBLE:      f = "d"
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 */
      __pyx_v_f = ((char *)"f");
      break;
      case NPY_DOUBLE:

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":318
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 *                 elif t == NPY_FLOAT:       f = "f"
 *                 elif t == NPY_DOUBLE:      f = "d"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 */
      __pyx_v_f = ((char *)"d");
      break;
      case NPY_LONGDOUBLE:

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":319
 *                 elif t == NPY_FLOAT:       f = "f"
 *                 elif t == NPY_DOUBLE:      f = "d"
 *                 elif t == NPY_LONGDOUBLE:  f = "g"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 */
      __pyx_v_f = ((char *)"g");
      break;
      case NPY_CFLOAT:

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":320
 *                 elif t == NPY_DOUBLE:      f = "d"
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 *                 elif t == NPY_CFLOAT:      f = "Zf"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"
 */
      __pyx_v_f = ((char *)"Zf");
      break;
      case NPY_CDOUBLE:

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":321
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 *                 elif t == NPY_CDOUBLE:     f = "Zd"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"
 *                 elif t == NPY_OBJECT:      f = "O"
 */
      __pyx_v_f = ((char *)"Zd");
      break;
      case NPY_CLONGDOUBLE:

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":322
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_OBJECT:      f = "O"
 *                 else:
 */
      __pyx_v_f = ((char *)"Zg");
      break;
      case NPY_OBJECT:

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":323
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"
 *                 elif t == NPY_OBJECT:      f = "O"             # <<<<<<<<<<<<<<
 *                 else:
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 */
      __pyx_v_f = ((char *)"O");
      break;
      default:

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":325
 *                 elif t == NPY_OBJECT:      f = "O"
 *                 else:
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)             # <<<<<<<<<<<<<<
 *                 info.format = f
 *                 return
 */
      __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_t); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 325, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_8 = PyUnicode_Format(__pyx_kp_u_unknown_dtype_code_in_numpy_pxd, __pyx_t_3); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 325, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 325, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(2, 325, __pyx_L1_error)
      break;
    }

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":326
 *                 else:
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 *                 info.format = f             # <<<<<<<<<<<<<<
 *                 return
 *             else:
 */
    __pyx_v_info->format = __pyx_v_f;

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":327
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 *                 info.format = f
 *                 return             # <<<<<<<<<<<<<<
 *             else:
 *                 info.format = <char*>PyObject_Malloc(_buffer_format_string_len)
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":302
 *             info.obj = self
 * 
 *             if not PyDataType_HASFIELDS(descr):             # <<<<<<<<<<<<<<
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or
 */
  }

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":329
 *                 return
 *             else:
 *                 info.format = <char*>PyObject_Malloc(_buffer_format_string_len)             # <<<<<<<<<<<<<<
 *                 info.format[0] = c'^' # Native data types, manual alignment
 *                 offset = 0
 */
  /*else*/ {
    __pyx_v_info->format = ((char *)PyObject_Malloc(0xFF));

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":330
 *             else:
 *                 info.format = <char*>PyObject_Malloc(_buffer_format_string_len)
 *                 info.format[0] = c'^' # Native data types, manual alignment             # <<<<<<<<<<<<<<
 *                 offset = 0
 *                 f = _util_dtypestring(descr, info.format + 1,
 */
    (__pyx_v_info->format[0]) = '^';

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":331
 *                 info.format = <char*>PyObject_Malloc(_buffer_format_string_len)
 *                 info.format[0] = c'^' # Native data types, manual alignment
 *                 offset = 0             # <<<<<<<<<<<<<<
 *                 f = _util_dtypestring(descr, info.format + 1,
 *                                       info.format + _buffer_format_string_len,
 */
    __pyx_v_offset = 0;

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":332
 *                 info.format[0] = c'^' # Native data types, manual alignment
 *                 offset = 0
 *                 f = _util_dtypestring(descr, info.format + 1,             # <<<<<<<<<<<<<<
 *                                       info.format + _buffer_format_string_len,
 *                                       &offset)
 */
    __pyx_t_9 = __pyx_f_5numpy__util_dtypestring(__pyx_v_descr, (__pyx_v_info->format + 1), (__pyx_v_info->format + 0xFF), (&__pyx_v_offset)); if (unlikely(__pyx_t_9 == ((char *)NULL))) __PYX_ERR(2, 332, __pyx_L1_error)
    __pyx_v_f = __pyx_t_9;

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":335
 *                                       info.format + _buffer_format_string_len,
 *                                       &offset)
 *                 f[0] = c'\0' # Terminate format string             # <<<<<<<<<<<<<<
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 */
    (__pyx_v_f[0]) = '\x00';
  }

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":258
 *         # experimental exception made for __getbuffer__ and __releasebuffer__
 *         # -- the details of this may change.
 *         def __getbuffer__(ndarray self, Py_buffer* info, int flags):             # <<<<<<<<<<<<<<
 *             # This implementation of getbuffer is geared towards Cython
 *             # requirements, and does not yet fulfill the PEP.
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.ndarray.__getbuffer__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  if (__pyx_v_info->obj != NULL) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  goto __pyx_L2;
  __pyx_L0:;
  if (__pyx_v_info->obj == Py_None) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_descr);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":337
 *                 f[0] = c'\0' # Terminate format string
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):             # <<<<<<<<<<<<<<
 *             if PyArray_HASFIELDS(self):
 *                 PyObject_Free(info.format)
 */

/* Python wrapper */
static CYTHON_UNUSED void __pyx_pw_5numpy_7ndarray_3__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info); /*proto*/
static CYTHON_UNUSED void __pyx_pw_5numpy_7ndarray_3__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__releasebuffer__ (wrapper)", 0);
  __pyx_pf_5numpy_7ndarray_2__releasebuffer__(((PyArrayObject *)__pyx_v_self), ((Py_buffer *)__pyx_v_info));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_5numpy_7ndarray_2__releasebuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__releasebuffer__", 0);

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":338
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 *             if PyArray_HASFIELDS(self):             # <<<<<<<<<<<<<<
 *                 PyObject_Free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 */
  __pyx_t_1 = (PyArray_HASFIELDS(__pyx_v_self) != 0);
  if (__pyx_t_1) {

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":339
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 *             if PyArray_HASFIELDS(self):
 *                 PyObject_Free(info.format)             # <<<<<<<<<<<<<<
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 *                 PyObject_Free(info.strides)
 */
    PyObject_Free(__pyx_v_info->format);

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":338
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 *             if PyArray_HASFIELDS(self):             # <<<<<<<<<<<<<<
 *                 PyObject_Free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 */
  }

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":340
 *             if PyArray_HASFIELDS(self):
 *                 PyObject_Free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 PyObject_Free(info.strides)
 *                 # info.shape was stored after info.strides in the same block
 */
  __pyx_t_1 = (((sizeof(npy_intp)) != (sizeof(Py_ssize_t))) != 0);
  if (__pyx_t_1) {

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":341
 *                 PyObject_Free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 *                 PyObject_Free(info.strides)             # <<<<<<<<<<<<<<
 *                 # info.shape was stored after info.strides in the same block
 * 
 */
    PyObject_Free(__pyx_v_info->strides);

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":340
 *             if PyArray_HASFIELDS(self):
 *                 PyObject_Free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 PyObject_Free(info.strides)
 *                 # info.shape was stored after info.strides in the same block
 */
  }

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":337
 *                 f[0] = c'\0' # Terminate format string
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):             # <<<<<<<<<<<<<<
 *             if PyArray_HASFIELDS(self):
 *                 PyObject_Free(info.format)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":821
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew1(PyObject *__pyx_v_a) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew1", 0);

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":822
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 *     return PyArray_MultiIterNew(1, <void*>a)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(1, ((void *)__pyx_v_a)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 822, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":821
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":824
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew2(PyObject *__pyx_v_a, PyObject *__pyx_v_b) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew2", 0);

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":825
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(2, ((void *)__pyx_v_a), ((void *)__pyx_v_b)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 825, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":824
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":827
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew3(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew3", 0);

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":828
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(3, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":827
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew3", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":830
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew4(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew4", 0);

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":831
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(4, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 831, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":830
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":833
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew5(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d, PyObject *__pyx_v_e) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew5", 0);

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":834
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)             # <<<<<<<<<<<<<<
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(5, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d), ((void *)__pyx_v_e)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 834, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":833
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew5", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":836
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyDataType_SHAPE(PyArray_Descr *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("PyDataType_SHAPE", 0);

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":837
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */
  __pyx_t_1 = (PyDataType_HASSUBARRAY(__pyx_v_d) != 0);
  if (__pyx_t_1) {

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":838
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape             # <<<<<<<<<<<<<<
 *     else:
 *         return ()
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject*)__pyx_v_d->subarray->shape));
    __pyx_r = ((PyObject*)__pyx_v_d->subarray->shape);
    goto __pyx_L0;

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":837
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */
  }

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":840
 *         return <tuple>d.subarray.shape
 *     else:
 *         return ()             # <<<<<<<<<<<<<<
 * 
 * cdef inline char* _util_dtypestring(dtype descr, char* f, char* end, int* offset) except NULL:
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_empty_tuple);
    __pyx_r = __pyx_empty_tuple;
    goto __pyx_L0;
  }

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":836
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":842
 *         return ()
 * 
 * cdef inline char* _util_dtypestring(dtype descr, char* f, char* end, int* offset) except NULL:             # <<<<<<<<<<<<<<
 *     # Recursive utility function used in __getbuffer__ to get format
 *     # string. The new location in the format string is returned.
 */

static CYTHON_INLINE char *__pyx_f_5numpy__util_dtypestring(PyArray_Descr *__pyx_v_descr, char *__pyx_v_f, char *__pyx_v_end, int *__pyx_v_offset) {
  PyArray_Descr *__pyx_v_child = 0;
  int __pyx_v_endian_detector;
  int __pyx_v_little_endian;
  PyObject *__pyx_v_fields = 0;
  PyObject *__pyx_v_childname = NULL;
  PyObject *__pyx_v_new_offset = NULL;
  PyObject *__pyx_v_t = NULL;
  char *__pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  long __pyx_t_8;
  char *__pyx_t_9;
  __Pyx_RefNannySetupContext("_util_dtypestring", 0);

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":847
 * 
 *     cdef dtype child
 *     cdef int endian_detector = 1             # <<<<<<<<<<<<<<
 *     cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)
 *     cdef tuple fields
 */
  __pyx_v_endian_detector = 1;

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":848
 *     cdef dtype child
 *     cdef int endian_detector = 1
 *     cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)             # <<<<<<<<<<<<<<
 *     cdef tuple fields
 * 
 */
  __pyx_v_little_endian = ((((char *)(&__pyx_v_endian_detector))[0]) != 0);

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":851
 *     cdef tuple fields
 * 
 *     for childname in descr.names:             # <<<<<<<<<<<<<<
 *         fields = descr.fields[childname]
 *         child, new_offset = fields
 */
  if (unlikely(__pyx_v_descr->names == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(2, 851, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_descr->names; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(2, 851, __pyx_L1_error)
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 851, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_childname, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":852
 * 
 *     for childname in descr.names:
 *         fields = descr.fields[childname]             # <<<<<<<<<<<<<<
 *         child, new_offset = fields
 * 
 */
    if (unlikely(__pyx_v_descr->fields == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 852, __pyx_L1_error)
    }
    __pyx_t_3 = __Pyx_PyDict_GetItem(__pyx_v_descr->fields, __pyx_v_childname); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 852, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(PyTuple_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_t_3)->tp_name), 0))) __PYX_ERR(2, 852, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_fields, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":853
 *     for childname in descr.names:
 *         fields = descr.fields[childname]
 *         child, new_offset = fields             # <<<<<<<<<<<<<<
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:
 */
    if (likely(__pyx_v_fields != Py_None)) {
      PyObject* sequence = __pyx_v_fields;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(2, 853, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      #else
      __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 853, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 853, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
    } else {
      __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(2, 853, __pyx_L1_error)
    }
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_dtype))))) __PYX_ERR(2, 853, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_child, ((PyArray_Descr *)__pyx_t_3));
    __pyx_t_3 = 0;
    __Pyx_XDECREF_SET(__pyx_v_new_offset, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":855
 *         child, new_offset = fields
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:             # <<<<<<<<<<<<<<
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 */
    __pyx_t_4 = __Pyx_PyInt_From_int((__pyx_v_offset[0])); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 855, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = PyNumber_Subtract(__pyx_v_new_offset, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 855, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 855, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_6 = ((((__pyx_v_end - __pyx_v_f) - ((int)__pyx_t_5)) < 15) != 0);
    if (unlikely(__pyx_t_6)) {

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":856
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")             # <<<<<<<<<<<<<<
 * 
 *         if ((child.byteorder == c'>' and little_endian) or
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__12, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 856, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(2, 856, __pyx_L1_error)

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":855
 *         child, new_offset = fields
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:             # <<<<<<<<<<<<<<
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 */
    }

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":858
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 *         if ((child.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")
 */
    __pyx_t_7 = ((__pyx_v_child->byteorder == '>') != 0);
    if (!__pyx_t_7) {
      goto __pyx_L8_next_or;
    } else {
    }
    __pyx_t_7 = (__pyx_v_little_endian != 0);
    if (!__pyx_t_7) {
    } else {
      __pyx_t_6 = __pyx_t_7;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_L8_next_or:;

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":859
 * 
 *         if ((child.byteorder == c'>' and little_endian) or
 *             (child.byteorder == c'<' and not little_endian)):             # <<<<<<<<<<<<<<
 *             raise ValueError(u"Non-native byte order not supported")
 *             # One could encode it in the format string and have Cython
 */
    __pyx_t_7 = ((__pyx_v_child->byteorder == '<') != 0);
    if (__pyx_t_7) {
    } else {
      __pyx_t_6 = __pyx_t_7;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_7 = ((!(__pyx_v_little_endian != 0)) != 0);
    __pyx_t_6 = __pyx_t_7;
    __pyx_L7_bool_binop_done:;

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":858
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 *         if ((child.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")
 */
    if (unlikely(__pyx_t_6)) {

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":860
 *         if ((child.byteorder == c'>' and little_endian) or
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *             # One could encode it in the format string and have Cython
 *             # complain instead, BUT: < and > in format strings also imply
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__11, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 860, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(2, 860, __pyx_L1_error)

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":858
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 *         if ((child.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")
 */
    }

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":870
 * 
 *         # Output padding bytes
 *         while offset[0] < new_offset:             # <<<<<<<<<<<<<<
 *             f[0] = 120 # "x"; pad byte
 *             f += 1
 */
    while (1) {
      __pyx_t_3 = __Pyx_PyInt_From_int((__pyx_v_offset[0])); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 870, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_t_3, __pyx_v_new_offset, Py_LT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 870, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 870, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (!__pyx_t_6) break;

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":871
 *         # Output padding bytes
 *         while offset[0] < new_offset:
 *             f[0] = 120 # "x"; pad byte             # <<<<<<<<<<<<<<
 *             f += 1
 *             offset[0] += 1
 */
      (__pyx_v_f[0]) = 0x78;

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":872
 *         while offset[0] < new_offset:
 *             f[0] = 120 # "x"; pad byte
 *             f += 1             # <<<<<<<<<<<<<<
 *             offset[0] += 1
 * 
 */
      __pyx_v_f = (__pyx_v_f + 1);

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":873
 *             f[0] = 120 # "x"; pad byte
 *             f += 1
 *             offset[0] += 1             # <<<<<<<<<<<<<<
 * 
 *         offset[0] += child.itemsize
 */
      __pyx_t_8 = 0;
      (__pyx_v_offset[__pyx_t_8]) = ((__pyx_v_offset[__pyx_t_8]) + 1);
    }

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":875
 *             offset[0] += 1
 * 
 *         offset[0] += child.itemsize             # <<<<<<<<<<<<<<
 * 
 *         if not PyDataType_HASFIELDS(child):
 */
    __pyx_t_8 = 0;
    (__pyx_v_offset[__pyx_t_8]) = ((__pyx_v_offset[__pyx_t_8]) + __pyx_v_child->elsize);

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":877
 *         offset[0] += child.itemsize
 * 
 *         if not PyDataType_HASFIELDS(child):             # <<<<<<<<<<<<<<
 *             t = child.type_num
 *             if end - f < 5:
 */
    __pyx_t_6 = ((!(PyDataType_HASFIELDS(__pyx_v_child) != 0)) != 0);
    if (__pyx_t_6) {

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":878
 * 
 *         if not PyDataType_HASFIELDS(child):
 *             t = child.type_num             # <<<<<<<<<<<<<<
 *             if end - f < 5:
 *                 raise RuntimeError(u"Format string allocated too short.")
 */
      __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_child->type_num); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 878, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_XDECREF_SET(__pyx_v_t, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":879
 *         if not PyDataType_HASFIELDS(child):
 *             t = child.type_num
 *             if end - f < 5:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(u"Format string allocated too short.")
 * 
 */
      __pyx_t_6 = (((__pyx_v_end - __pyx_v_f) < 5) != 0);
      if (unlikely(__pyx_t_6)) {

        /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":880
 *             t = child.type_num
 *             if end - f < 5:
 *                 raise RuntimeError(u"Format string allocated too short.")             # <<<<<<<<<<<<<<
 * 
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 */
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__13, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 880, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_Raise(__pyx_t_4, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __PYX_ERR(2, 880, __pyx_L1_error)

        /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":879
 *         if not PyDataType_HASFIELDS(child):
 *             t = child.type_num
 *             if end - f < 5:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(u"Format string allocated too short.")
 * 
 */
      }

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":883
 * 
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 *             if   t == NPY_BYTE:        f[0] =  98 #"b"             # <<<<<<<<<<<<<<
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_BYTE); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 883, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 883, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 883, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 98;
        goto __pyx_L15;
      }

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":884
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 *             if   t == NPY_BYTE:        f[0] =  98 #"b"
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"             # <<<<<<<<<<<<<<
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_UBYTE); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 884, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 884, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 884, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 66;
        goto __pyx_L15;
      }

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":885
 *             if   t == NPY_BYTE:        f[0] =  98 #"b"
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"             # <<<<<<<<<<<<<<
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_SHORT); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 885, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 885, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 885, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x68;
        goto __pyx_L15;
      }

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":886
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"             # <<<<<<<<<<<<<<
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_USHORT); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 886, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 886, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 886, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 72;
        goto __pyx_L15;
      }

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":887
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 *             elif t == NPY_INT:         f[0] = 105 #"i"             # <<<<<<<<<<<<<<
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_INT); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 887, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 887, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 887, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x69;
        goto __pyx_L15;
      }

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":888
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 *             elif t == NPY_UINT:        f[0] =  73 #"I"             # <<<<<<<<<<<<<<
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_UINT); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 888, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 888, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 888, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 73;
        goto __pyx_L15;
      }

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":889
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 *             elif t == NPY_LONG:        f[0] = 108 #"l"             # <<<<<<<<<<<<<<
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_LONG); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 889, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 889, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 889, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x6C;
        goto __pyx_L15;
      }

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":890
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"             # <<<<<<<<<<<<<<
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_ULONG); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 890, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 890, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 890, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 76;
        goto __pyx_L15;
      }

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":891
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"             # <<<<<<<<<<<<<<
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_LONGLONG); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 891, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 891, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 891, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x71;
        goto __pyx_L15;
      }

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":892
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"             # <<<<<<<<<<<<<<
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_ULONGLONG); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 892, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 892, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 892, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 81;
        goto __pyx_L15;
      }

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":893
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"             # <<<<<<<<<<<<<<
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_FLOAT); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 893, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 893, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 893, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x66;
        goto __pyx_L15;
      }

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":894
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"             # <<<<<<<<<<<<<<
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_DOUBLE); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 894, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 894, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 894, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x64;
        goto __pyx_L15;
      }

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":895
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"             # <<<<<<<<<<<<<<
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_LONGDOUBLE); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 895, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 895, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 895, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x67;
        goto __pyx_L15;
      }

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":896
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf             # <<<<<<<<<<<<<<
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_CFLOAT); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 896, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 896, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 896, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 90;
        (__pyx_v_f[1]) = 0x66;
        __pyx_v_f = (__pyx_v_f + 1);
        goto __pyx_L15;
      }

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":897
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd             # <<<<<<<<<<<<<<
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_CDOUBLE); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 897, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 897, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 897, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 90;
        (__pyx_v_f[1]) = 0x64;
        __pyx_v_f = (__pyx_v_f + 1);
        goto __pyx_L15;
      }

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":898
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg             # <<<<<<<<<<<<<<
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"
 *             else:
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_CLONGDOUBLE); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 898, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 898, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 898, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 90;
        (__pyx_v_f[1]) = 0x67;
        __pyx_v_f = (__pyx_v_f + 1);
        goto __pyx_L15;
      }

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":899
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"             # <<<<<<<<<<<<<<
 *             else:
 *                 raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_OBJECT); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 899, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 899, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 899, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (likely(__pyx_t_6)) {
        (__pyx_v_f[0]) = 79;
        goto __pyx_L15;
      }

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":901
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"
 *             else:
 *                 raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)             # <<<<<<<<<<<<<<
 *             f += 1
 *         else:
 */
      /*else*/ {
        __pyx_t_3 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_unknown_dtype_code_in_numpy_pxd, __pyx_v_t); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 901, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 901, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_Raise(__pyx_t_4, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __PYX_ERR(2, 901, __pyx_L1_error)
      }
      __pyx_L15:;

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":902
 *             else:
 *                 raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 *             f += 1             # <<<<<<<<<<<<<<
 *         else:
 *             # Cython ignores struct boundary information ("T{...}"),
 */
      __pyx_v_f = (__pyx_v_f + 1);

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":877
 *         offset[0] += child.itemsize
 * 
 *         if not PyDataType_HASFIELDS(child):             # <<<<<<<<<<<<<<
 *             t = child.type_num
 *             if end - f < 5:
 */
      goto __pyx_L13;
    }

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":906
 *             # Cython ignores struct boundary information ("T{...}"),
 *             # so don't output it
 *             f = _util_dtypestring(child, f, end, offset)             # <<<<<<<<<<<<<<
 *     return f
 * 
 */
    /*else*/ {
      __pyx_t_9 = __pyx_f_5numpy__util_dtypestring(__pyx_v_child, __pyx_v_f, __pyx_v_end, __pyx_v_offset); if (unlikely(__pyx_t_9 == ((char *)NULL))) __PYX_ERR(2, 906, __pyx_L1_error)
      __pyx_v_f = __pyx_t_9;
    }
    __pyx_L13:;

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":851
 *     cdef tuple fields
 * 
 *     for childname in descr.names:             # <<<<<<<<<<<<<<
 *         fields = descr.fields[childname]
 *         child, new_offset = fields
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":907
 *             # so don't output it
 *             f = _util_dtypestring(child, f, end, offset)
 *     return f             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_f;
  goto __pyx_L0;

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":842
 *         return ()
 * 
 * cdef inline char* _util_dtypestring(dtype descr, char* f, char* end, int* offset) except NULL:             # <<<<<<<<<<<<<<
 *     # Recursive utility function used in __getbuffer__ to get format
 *     # string. The new location in the format string is returned.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("numpy._util_dtypestring", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_child);
  __Pyx_XDECREF(__pyx_v_fields);
  __Pyx_XDECREF(__pyx_v_childname);
  __Pyx_XDECREF(__pyx_v_new_offset);
  __Pyx_XDECREF(__pyx_v_t);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1022
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

static CYTHON_INLINE void __pyx_f_5numpy_set_array_base(PyArrayObject *__pyx_v_arr, PyObject *__pyx_v_base) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_array_base", 0);

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1023
 * 
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!             # <<<<<<<<<<<<<<
 *     PyArray_SetBaseObject(arr, base)
 * 
 */
  Py_INCREF(__pyx_v_base);

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1024
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object get_array_base(ndarray arr):
 */
  (void)(PyArray_SetBaseObject(__pyx_v_arr, __pyx_v_base));

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1022
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1026
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_get_array_base(PyArrayObject *__pyx_v_arr) {
  PyObject *__pyx_v_base;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("get_array_base", 0);

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1027
 * 
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)             # <<<<<<<<<<<<<<
 *     if base is NULL:
 *         return None
 */
  __pyx_v_base = PyArray_BASE(__pyx_v_arr);

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1028
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */
  __pyx_t_1 = ((__pyx_v_base == NULL) != 0);
  if (__pyx_t_1) {

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1029
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 *         return None             # <<<<<<<<<<<<<<
 *     return <object>base
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1028
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */
  }

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1030
 *     if base is NULL:
 *         return None
 *     return <object>base             # <<<<<<<<<<<<<<
 * 
 * # Versions of the import_* functions which are more suitable for
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_base));
  __pyx_r = ((PyObject *)__pyx_v_base);
  goto __pyx_L0;

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1026
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1034
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_array()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_array(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("import_array", 0);

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1035
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_array()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1036
 * cdef inline int import_array() except -1:
 *     try:
 *         _import_array()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")
 */
      __pyx_t_4 = _import_array(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 1036, __pyx_L3_error)

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1035
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_array()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1037
 *     try:
 *         _import_array()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(2, 1037, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1038
 *         _import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__14, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 1038, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(2, 1038, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1035
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_array()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1034
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_array()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1040
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_umath(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("import_umath", 0);

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1041
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1042
 * cdef inline int import_umath() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 1042, __pyx_L3_error)

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1041
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1043
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(2, 1043, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1044
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__15, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 1044, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(2, 1044, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1041
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1040
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1046
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_ufunc(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("import_ufunc", 0);

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1047
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1048
 * cdef inline int import_ufunc() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 1048, __pyx_L3_error)

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1047
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1049
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(2, 1049, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1050
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__15, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 1050, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(2, 1050, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1047
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1046
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator __pyx_vtable_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator;

static PyObject *__pyx_tp_new_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *)o);
  p->__pyx_vtab = __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator;
  p->collections = ((PyObject*)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator(PyObject *o) {
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *p = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_3__dealloc__(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->collections);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *p = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *)o;
  if (p->collections) {
    e = (*v)(p->collections, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *p = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *)o;
  tmp = ((PyObject*)p->collections);
  p->collections = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_collections(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_11collections_1__get__(o);
}

static int __pyx_setprop_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_collections(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_11collections_3__set__(o, v);
  }
  else {
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_11collections_5__del__(o);
  }
}

static PyObject *__pyx_getprop_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_size(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_4size_1__get__(o);
}

static int __pyx_setprop_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_size(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_4size_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_done(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_4done_1__get__(o);
}

static int __pyx_setprop_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_done(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_4done_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_total_combinations(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_18total_combinations_1__get__(o);
}

static int __pyx_setprop_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_total_combinations(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_18total_combinations_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_specialmethod___pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_19__next__(PyObject *self, CYTHON_UNUSED PyObject *arg) {return __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_19__next__(self);}

static PyMethodDef __pyx_methods_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator[] = {
  {"get_lengths", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_5get_lengths, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_4get_lengths},
  {"get_indices", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_7get_indices, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_6get_indices},
  {"has_more", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_9has_more, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_8has_more},
  {"compose_next_value", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_11compose_next_value, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_10compose_next_value},
  {"advance", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_13advance, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_12advance},
  {"get_next_value", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_15get_next_value, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_14get_next_value},
  {"__next__", (PyCFunction)__pyx_specialmethod___pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_19__next__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_21__reduce_cython__, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_20__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_23__setstate_cython__, METH_O, __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_22__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator[] = {
  {(char *)"collections", __pyx_getprop_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_collections, __pyx_setprop_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_collections, (char *)"collections: list", 0},
  {(char *)"size", __pyx_getprop_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_size, __pyx_setprop_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_size, (char *)"size: 'size_t'", 0},
  {(char *)"done", __pyx_getprop_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_done, __pyx_setprop_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_done, (char *)"done: 'bool'", 0},
  {(char *)"total_combinations", __pyx_getprop_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_total_combinations, __pyx_setprop_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_total_combinations, (char *)"total_combinations: 'size_t'", 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator = {
  PyVarObject_HEAD_INIT(0, 0)
  "ms_deisotope._c.feature_map.feature_processor.CartesianProductIterator", /*tp_name*/
  sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator, /*tp_traverse*/
  __pyx_tp_clear_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_17__iter__, /*tp_iter*/
  __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_19__next__, /*tp_iternext*/
  __pyx_methods_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer __pyx_vtable_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer;

static PyObject *__pyx_tp_new_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *)o);
  p->__pyx_vtab = __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer;
  p->experimental = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->theoretical = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer(PyObject *o) {
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *p = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->experimental);
  Py_CLEAR(p->theoretical);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *p = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *)o;
  if (p->experimental) {
    e = (*v)(p->experimental, a); if (e) return e;
  }
  if (p->theoretical) {
    e = (*v)(p->theoretical, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *p = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *)o;
  tmp = ((PyObject*)p->experimental);
  p->experimental = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->theoretical);
  p->theoretical = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer_1__reduce_cython__, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer___reduce_cython__},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer_3__setstate_cython__, METH_O, __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer_2__setstate_cython__},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer = {
  PyVarObject_HEAD_INIT(0, 0)
  "ms_deisotope._c.feature_map.feature_processor.envelope_conformer", /*tp_name*/
  sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer, /*tp_traverse*/
  __pyx_tp_clear_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase __pyx_vtable_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase;

static PyObject *__pyx_tp_new_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)o);
  p->__pyx_vtab = __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase;
  p->feature_map = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *)Py_None); Py_INCREF(Py_None);
  p->scorer = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *)Py_None); Py_INCREF(Py_None);
  p->averagine = ((struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase(PyObject *o) {
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *p = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->feature_map);
  Py_CLEAR(p->scorer);
  Py_CLEAR(p->averagine);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *p = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)o;
  if (p->feature_map) {
    e = (*v)(((PyObject *)p->feature_map), a); if (e) return e;
  }
  if (p->scorer) {
    e = (*v)(((PyObject *)p->scorer), a); if (e) return e;
  }
  if (p->averagine) {
    e = (*v)(((PyObject *)p->averagine), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *p = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)o;
  tmp = ((PyObject*)p->feature_map);
  p->feature_map = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->scorer);
  p->scorer = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->averagine);
  p->averagine = ((struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_feature_map(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_11feature_map_1__get__(o);
}

static int __pyx_setprop_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_feature_map(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_11feature_map_3__set__(o, v);
  }
  else {
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_11feature_map_5__del__(o);
  }
}

static PyObject *__pyx_getprop_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_scorer(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_6scorer_1__get__(o);
}

static int __pyx_setprop_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_scorer(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_6scorer_3__set__(o, v);
  }
  else {
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_6scorer_5__del__(o);
  }
}

static PyObject *__pyx_getprop_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_averagine(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_9averagine_1__get__(o);
}

static int __pyx_setprop_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_averagine(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_9averagine_3__set__(o, v);
  }
  else {
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_9averagine_5__del__(o);
  }
}

static PyMethodDef __pyx_methods_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase[] = {
  {"create_theoretical_distribution", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_1create_theoretical_distribution, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_create_theoretical_distribution},
  {"find_all_features", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_3find_all_features, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_2find_all_features},
  {"find_features", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_5find_features, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_4find_features},
  {"match_theoretical_isotopic_distribution", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_7match_theoretical_isotopic_distribution, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_6match_theoretical_isotopic_distribution},
  {"conform_envelopes", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_9conform_envelopes, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_8conform_envelopes},
  {"_fit_single_feature_set", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_11_fit_single_feature_set, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_10_fit_single_feature_set},
  {"_fit_feature_set", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_13_fit_feature_set, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_12_fit_feature_set},
  {"_fit_composition", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_15_fit_composition, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_14_fit_composition},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_17__reduce_cython__, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_16__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_19__setstate_cython__, METH_O, __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_18__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase[] = {
  {(char *)"feature_map", __pyx_getprop_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_feature_map, __pyx_setprop_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_feature_map, (char *)"feature_map: ms_deisotope._c.feature_map.feature_map.LCMSFeatureMap", 0},
  {(char *)"scorer", __pyx_getprop_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_scorer, __pyx_setprop_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_scorer, (char *)"scorer: ms_deisotope._c.scoring.IsotopicFitterBase", 0},
  {(char *)"averagine", __pyx_getprop_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_averagine, __pyx_setprop_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_averagine, (char *)"averagine: ms_deisotope._c.averagine.AveragineCache", 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase = {
  PyVarObject_HEAD_INIT(0, 0)
  "ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase", /*tp_name*/
  sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase, /*tp_traverse*/
  __pyx_tp_clear_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {"conform_envelopes", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_1conform_envelopes, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_conform_envelopes},
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_feature_processor(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_feature_processor},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "feature_processor",
    0, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_kp_s_0_2f_0_2f, __pyx_k_0_2f_0_2f, sizeof(__pyx_k_0_2f_0_2f), 0, 0, 1, 0},
  {&__pyx_n_s_CartesianProductIterator, __pyx_k_CartesianProductIterator, sizeof(__pyx_k_CartesianProductIterator), 0, 0, 1, 1},
  {&__pyx_kp_u_Format_string_allocated_too_shor, __pyx_k_Format_string_allocated_too_shor, sizeof(__pyx_k_Format_string_allocated_too_shor), 0, 1, 0, 0},
  {&__pyx_kp_u_Format_string_allocated_too_shor_2, __pyx_k_Format_string_allocated_too_shor_2, sizeof(__pyx_k_Format_string_allocated_too_shor_2), 0, 1, 0, 0},
  {&__pyx_n_s_ImportError, __pyx_k_ImportError, sizeof(__pyx_k_ImportError), 0, 0, 1, 1},
  {&__pyx_kp_s_Incompatible_checksums_s_vs_0xa9, __pyx_k_Incompatible_checksums_s_vs_0xa9, sizeof(__pyx_k_Incompatible_checksums_s_vs_0xa9), 0, 0, 1, 0},
  {&__pyx_kp_s_Incompatible_checksums_s_vs_0xba, __pyx_k_Incompatible_checksums_s_vs_0xba, sizeof(__pyx_k_Incompatible_checksums_s_vs_0xba), 0, 0, 1, 0},
  {&__pyx_n_s_LCMSFeatureProcessorBase, __pyx_k_LCMSFeatureProcessorBase, sizeof(__pyx_k_LCMSFeatureProcessorBase), 0, 0, 1, 1},
  {&__pyx_kp_u_Non_native_byte_order_not_suppor, __pyx_k_Non_native_byte_order_not_suppor, sizeof(__pyx_k_Non_native_byte_order_not_suppor), 0, 1, 0, 0},
  {&__pyx_n_s_PickleError, __pyx_k_PickleError, sizeof(__pyx_k_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_RuntimeError, __pyx_k_RuntimeError, sizeof(__pyx_k_RuntimeError), 0, 0, 1, 1},
  {&__pyx_n_s_StopIteration, __pyx_k_StopIteration, sizeof(__pyx_k_StopIteration), 0, 0, 1, 1},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_kp_s__6, __pyx_k__6, sizeof(__pyx_k__6), 0, 0, 1, 0},
  {&__pyx_n_s_advance, __pyx_k_advance, sizeof(__pyx_k_advance), 0, 0, 1, 1},
  {&__pyx_n_s_base_theoretical, __pyx_k_base_theoretical, sizeof(__pyx_k_base_theoretical), 0, 0, 1, 1},
  {&__pyx_n_s_base_tid, __pyx_k_base_tid, sizeof(__pyx_k_base_tid), 0, 0, 1, 1},
  {&__pyx_n_s_charge, __pyx_k_charge, sizeof(__pyx_k_charge), 0, 0, 1, 1},
  {&__pyx_n_s_charge_carrier, __pyx_k_charge_carrier, sizeof(__pyx_k_charge_carrier), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_collections, __pyx_k_collections, sizeof(__pyx_k_collections), 0, 0, 1, 1},
  {&__pyx_n_s_compose_next_value, __pyx_k_compose_next_value, sizeof(__pyx_k_compose_next_value), 0, 0, 1, 1},
  {&__pyx_n_s_composition, __pyx_k_composition, sizeof(__pyx_k_composition), 0, 0, 1, 1},
  {&__pyx_n_s_conform_envelopes, __pyx_k_conform_envelopes, sizeof(__pyx_k_conform_envelopes), 0, 0, 1, 1},
  {&__pyx_n_s_create_theoretical_distribution, __pyx_k_create_theoretical_distribution, sizeof(__pyx_k_create_theoretical_distribution), 0, 0, 1, 1},
  {&__pyx_kp_s_d_combinations_have_been_consid, __pyx_k_d_combinations_have_been_consid, sizeof(__pyx_k_d_combinations_have_been_consid), 0, 0, 1, 0},
  {&__pyx_n_s_dict, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
  {&__pyx_n_s_end, __pyx_k_end, sizeof(__pyx_k_end), 0, 0, 1, 1},
  {&__pyx_n_s_end_time, __pyx_k_end_time, sizeof(__pyx_k_end_time), 0, 0, 1, 1},
  {&__pyx_n_s_envelope_conformer, __pyx_k_envelope_conformer, sizeof(__pyx_k_envelope_conformer), 0, 0, 1, 1},
  {&__pyx_n_s_error_tolerance, __pyx_k_error_tolerance, sizeof(__pyx_k_error_tolerance), 0, 0, 1, 1},
  {&__pyx_n_s_experimental, __pyx_k_experimental, sizeof(__pyx_k_experimental), 0, 0, 1, 1},
  {&__pyx_n_s_feature, __pyx_k_feature, sizeof(__pyx_k_feature), 0, 0, 1, 1},
  {&__pyx_n_s_features, __pyx_k_features, sizeof(__pyx_k_features), 0, 0, 1, 1},
  {&__pyx_n_s_file, __pyx_k_file, sizeof(__pyx_k_file), 0, 0, 1, 1},
  {&__pyx_n_s_find_all_features, __pyx_k_find_all_features, sizeof(__pyx_k_find_all_features), 0, 0, 1, 1},
  {&__pyx_n_s_find_features, __pyx_k_find_features, sizeof(__pyx_k_find_features), 0, 0, 1, 1},
  {&__pyx_n_s_fit_composition, __pyx_k_fit_composition, sizeof(__pyx_k_fit_composition), 0, 0, 1, 1},
  {&__pyx_n_s_fit_feature_set, __pyx_k_fit_feature_set, sizeof(__pyx_k_fit_feature_set), 0, 0, 1, 1},
  {&__pyx_n_s_fit_single_feature_set, __pyx_k_fit_single_feature_set, sizeof(__pyx_k_fit_single_feature_set), 0, 0, 1, 1},
  {&__pyx_n_s_get_next_value, __pyx_k_get_next_value, sizeof(__pyx_k_get_next_value), 0, 0, 1, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_n_s_has_more, __pyx_k_has_more, sizeof(__pyx_k_has_more), 0, 0, 1, 1},
  {&__pyx_n_s_ignore_below, __pyx_k_ignore_below, sizeof(__pyx_k_ignore_below), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_inf, __pyx_k_inf, sizeof(__pyx_k_inf), 0, 0, 1, 1},
  {&__pyx_n_s_interval, __pyx_k_interval, sizeof(__pyx_k_interval), 0, 0, 1, 1},
  {&__pyx_n_s_left_search, __pyx_k_left_search, sizeof(__pyx_k_left_search), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_map, __pyx_k_map, sizeof(__pyx_k_map), 0, 0, 1, 1},
  {&__pyx_n_s_match_theoretical_isotopic_distr, __pyx_k_match_theoretical_isotopic_distr, sizeof(__pyx_k_match_theoretical_isotopic_distr), 0, 0, 1, 1},
  {&__pyx_n_s_max_missed_peaks, __pyx_k_max_missed_peaks, sizeof(__pyx_k_max_missed_peaks), 0, 0, 1, 1},
  {&__pyx_n_s_minimum_theoretical_abundance, __pyx_k_minimum_theoretical_abundance, sizeof(__pyx_k_minimum_theoretical_abundance), 0, 0, 1, 1},
  {&__pyx_n_s_ms_deisotope__c_feature_map_feat, __pyx_k_ms_deisotope__c_feature_map_feat, sizeof(__pyx_k_ms_deisotope__c_feature_map_feat), 0, 0, 1, 1},
  {&__pyx_n_s_mz, __pyx_k_mz, sizeof(__pyx_k_mz), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_kp_u_ndarray_is_not_C_contiguous, __pyx_k_ndarray_is_not_C_contiguous, sizeof(__pyx_k_ndarray_is_not_C_contiguous), 0, 1, 0, 0},
  {&__pyx_kp_u_ndarray_is_not_Fortran_contiguou, __pyx_k_ndarray_is_not_Fortran_contiguou, sizeof(__pyx_k_ndarray_is_not_Fortran_contiguou), 0, 1, 0, 0},
  {&__pyx_n_s_new, __pyx_k_new, sizeof(__pyx_k_new), 0, 0, 1, 1},
  {&__pyx_kp_s_no_default___reduce___due_to_non, __pyx_k_no_default___reduce___due_to_non, sizeof(__pyx_k_no_default___reduce___due_to_non), 0, 0, 1, 0},
  {&__pyx_n_s_np, __pyx_k_np, sizeof(__pyx_k_np), 0, 0, 1, 1},
  {&__pyx_n_s_numpy, __pyx_k_numpy, sizeof(__pyx_k_numpy), 0, 0, 1, 1},
  {&__pyx_kp_s_numpy_core_multiarray_failed_to, __pyx_k_numpy_core_multiarray_failed_to, sizeof(__pyx_k_numpy_core_multiarray_failed_to), 0, 0, 1, 0},
  {&__pyx_kp_s_numpy_core_umath_failed_to_impor, __pyx_k_numpy_core_umath_failed_to_impor, sizeof(__pyx_k_numpy_core_umath_failed_to_impor), 0, 0, 1, 0},
  {&__pyx_n_s_pickle, __pyx_k_pickle, sizeof(__pyx_k_pickle), 0, 0, 1, 1},
  {&__pyx_n_s_print, __pyx_k_print, sizeof(__pyx_k_print), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_PickleError, __pyx_k_pyx_PickleError, sizeof(__pyx_k_pyx_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_checksum, __pyx_k_pyx_checksum, sizeof(__pyx_k_pyx_checksum), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_result, __pyx_k_pyx_result, sizeof(__pyx_k_pyx_result), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_state, __pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_type, __pyx_k_pyx_type, sizeof(__pyx_k_pyx_type), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_LCMSFeatureProces, __pyx_k_pyx_unpickle_LCMSFeatureProces, sizeof(__pyx_k_pyx_unpickle_LCMSFeatureProces), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_envelope_conforme, __pyx_k_pyx_unpickle_envelope_conforme, sizeof(__pyx_k_pyx_unpickle_envelope_conforme), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {&__pyx_n_s_right_search, __pyx_k_right_search, sizeof(__pyx_k_right_search), 0, 0, 1, 1},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_start_time, __pyx_k_start_time, sizeof(__pyx_k_start_time), 0, 0, 1, 1},
  {&__pyx_kp_s_stringsource, __pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 0, 1, 0},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_theoretical_distribution, __pyx_k_theoretical_distribution, sizeof(__pyx_k_theoretical_distribution), 0, 0, 1, 1},
  {&__pyx_n_s_threshold_scale, __pyx_k_threshold_scale, sizeof(__pyx_k_threshold_scale), 0, 0, 1, 1},
  {&__pyx_n_s_truncate_after, __pyx_k_truncate_after, sizeof(__pyx_k_truncate_after), 0, 0, 1, 1},
  {&__pyx_n_s_truncated_tid, __pyx_k_truncated_tid, sizeof(__pyx_k_truncated_tid), 0, 0, 1, 1},
  {&__pyx_kp_u_unknown_dtype_code_in_numpy_pxd, __pyx_k_unknown_dtype_code_in_numpy_pxd, sizeof(__pyx_k_unknown_dtype_code_in_numpy_pxd), 0, 1, 0, 0},
  {&__pyx_n_s_update, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
  {&__pyx_n_s_zeros, __pyx_k_zeros, sizeof(__pyx_k_zeros), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 55, __pyx_L1_error)
  __pyx_builtin_map = __Pyx_GetBuiltinName(__pyx_n_s_map); if (!__pyx_builtin_map) __PYX_ERR(0, 74, __pyx_L1_error)
  __pyx_builtin_StopIteration = __Pyx_GetBuiltinName(__pyx_n_s_StopIteration); if (!__pyx_builtin_StopIteration) __PYX_ERR(0, 141, __pyx_L1_error)
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(1, 2, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(2, 272, __pyx_L1_error)
  __pyx_builtin_RuntimeError = __Pyx_GetBuiltinName(__pyx_n_s_RuntimeError); if (!__pyx_builtin_RuntimeError) __PYX_ERR(2, 856, __pyx_L1_error)
  __pyx_builtin_ImportError = __Pyx_GetBuiltinName(__pyx_n_s_ImportError); if (!__pyx_builtin_ImportError) __PYX_ERR(2, 1038, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple_ = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple_)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":272
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")             # <<<<<<<<<<<<<<
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 */
  __pyx_tuple__9 = PyTuple_Pack(1, __pyx_kp_u_ndarray_is_not_C_contiguous); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(2, 272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":276
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")             # <<<<<<<<<<<<<<
 * 
 *             info.buf = PyArray_DATA(self)
 */
  __pyx_tuple__10 = PyTuple_Pack(1, __pyx_kp_u_ndarray_is_not_Fortran_contiguou); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(2, 276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":306
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"
 */
  __pyx_tuple__11 = PyTuple_Pack(1, __pyx_kp_u_Non_native_byte_order_not_suppor); if (unlikely(!__pyx_tuple__11)) __PYX_ERR(2, 306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__11);
  __Pyx_GIVEREF(__pyx_tuple__11);

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":856
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")             # <<<<<<<<<<<<<<
 * 
 *         if ((child.byteorder == c'>' and little_endian) or
 */
  __pyx_tuple__12 = PyTuple_Pack(1, __pyx_kp_u_Format_string_allocated_too_shor); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(2, 856, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":880
 *             t = child.type_num
 *             if end - f < 5:
 *                 raise RuntimeError(u"Format string allocated too short.")             # <<<<<<<<<<<<<<
 * 
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 */
  __pyx_tuple__13 = PyTuple_Pack(1, __pyx_kp_u_Format_string_allocated_too_shor_2); if (unlikely(!__pyx_tuple__13)) __PYX_ERR(2, 880, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__13);
  __Pyx_GIVEREF(__pyx_tuple__13);

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1038
 *         _import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
  __pyx_tuple__14 = PyTuple_Pack(1, __pyx_kp_s_numpy_core_multiarray_failed_to); if (unlikely(!__pyx_tuple__14)) __PYX_ERR(2, 1038, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_tuple__14);

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1044
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
  __pyx_tuple__15 = PyTuple_Pack(1, __pyx_kp_s_numpy_core_umath_failed_to_impor); if (unlikely(!__pyx_tuple__15)) __PYX_ERR(2, 1044, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__15);
  __Pyx_GIVEREF(__pyx_tuple__15);

  /* "(tree fragment)":1
 * def __pyx_unpickle_envelope_conformer(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_tuple__16 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__16)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__16);
  __Pyx_GIVEREF(__pyx_tuple__16);
  __pyx_codeobj__17 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__16, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_envelope_conforme, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__17)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_tuple__18 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__18)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__18);
  __Pyx_GIVEREF(__pyx_tuple__18);
  __pyx_codeobj__19 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__18, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_LCMSFeatureProces, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__19)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_int_177677261 = PyInt_FromLong(177677261L); if (unlikely(!__pyx_int_177677261)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_195672028 = PyInt_FromLong(195672028L); if (unlikely(!__pyx_int_195672028)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __pyx_v_12ms_deisotope_2_c_11feature_map_17feature_processor_zeros = Py_None; Py_INCREF(Py_None);
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator = &__pyx_vtable_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator.has_more = (int (*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_has_more;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator.compose_next_value = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_compose_next_value;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator.advance = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_advance;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator.get_next_value = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_get_next_value;
  if (PyType_Ready(&__pyx_type_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator) < 0) __PYX_ERR(0, 60, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator.tp_dictoffset && __pyx_type_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator.tp_dict, __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator) < 0) __PYX_ERR(0, 60, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_CartesianProductIterator, (PyObject *)&__pyx_type_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator) < 0) __PYX_ERR(0, 60, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator) < 0) __PYX_ERR(0, 60, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator = &__pyx_type_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator;
  __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer = &__pyx_vtable_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer._create = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *(*)(void))__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer__create;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer.acquire = (void (*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *, PyObject *, PyObject *))__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer_acquire;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer.conform = (void (*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer_conform *__pyx_optional_args))__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer_conform;
  if (PyType_Ready(&__pyx_type_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer) < 0) __PYX_ERR(0, 166, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer.tp_dictoffset && __pyx_type_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer.tp_dict, __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer) < 0) __PYX_ERR(0, 166, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_envelope_conformer, (PyObject *)&__pyx_type_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer) < 0) __PYX_ERR(0, 166, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer) < 0) __PYX_ERR(0, 166, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer = &__pyx_type_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer;
  __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase = &__pyx_vtable_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase.create_theoretical_distribution = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *, double, int, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_create_theoretical_distribution *__pyx_optional_args))__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_create_theoretical_distribution;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase.find_all_features = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_find_all_features *__pyx_optional_args))__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_find_all_features;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase.find_features = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_find_features *__pyx_optional_args))__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_find_features;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase.match_theoretical_isotopic_distribution = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_match_theoretical_isotopic_distribution *__pyx_optional_args))__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_match_theoretical_isotopic_distribution;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase.conform_envelopes = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *, PyObject *, PyObject *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_conform_envelopes;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase._conform_envelopes = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *, PyObject *, PyObject *, size_t *))__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__conform_envelopes;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase._find_thresholded_score = (double (*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *, struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector *, double))__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__find_thresholded_score;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase._fit_single_feature_set = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit *(*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *, PyObject *, PyObject *, double, int, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_single_feature_set *__pyx_optional_args))__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_single_feature_set;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase._fit_theoretical_distribution_on_features = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *, double, double, int, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_theoretical_distribution_on_features *__pyx_optional_args))__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_theoretical_distribution_on_features;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase._fit_feature_set = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *, double, double, int, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_feature_set *__pyx_optional_args))__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_feature_set;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase._fit_composition = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *, PyObject *, double, int, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_composition *__pyx_optional_args))__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_composition;
  if (PyType_Ready(&__pyx_type_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase) < 0) __PYX_ERR(0, 240, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase.tp_dictoffset && __pyx_type_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase.tp_dict, __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase) < 0) __PYX_ERR(0, 240, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_LCMSFeatureProcessorBase, (PyObject *)&__pyx_type_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase) < 0) __PYX_ERR(0, 240, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase) < 0) __PYX_ERR(0, 240, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase = &__pyx_type_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(3, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_peak_picker._c.peak_set"); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase = __Pyx_ImportType(__pyx_t_1, "ms_peak_picker._c.peak_set", "PeakBase", sizeof(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase) __PYX_ERR(4, 2, __pyx_L1_error)
  __pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakBase = (struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakBase*)__Pyx_GetVtable(__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase->tp_dict); if (unlikely(!__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakBase)) __PYX_ERR(4, 2, __pyx_L1_error)
  __pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak = __Pyx_ImportType(__pyx_t_1, "ms_peak_picker._c.peak_set", "FittedPeak", sizeof(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak) __PYX_ERR(4, 12, __pyx_L1_error)
  __pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_FittedPeak = (struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_FittedPeak*)__Pyx_GetVtable(__pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak->tp_dict); if (unlikely(!__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_FittedPeak)) __PYX_ERR(4, 12, __pyx_L1_error)
  __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet = __Pyx_ImportType(__pyx_t_1, "ms_peak_picker._c.peak_set", "PeakSet", sizeof(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet) __PYX_ERR(4, 29, __pyx_L1_error)
  __pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSet = (struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet*)__Pyx_GetVtable(__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet->tp_dict); if (unlikely(!__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSet)) __PYX_ERR(4, 29, __pyx_L1_error)
  __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed = __Pyx_ImportType(__pyx_t_1, "ms_peak_picker._c.peak_set", "PeakSetIndexed", sizeof(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed) __PYX_ERR(4, 81, __pyx_L1_error)
  __pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed = (struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed*)__Pyx_GetVtable(__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed->tp_dict); if (unlikely(!__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed)) __PYX_ERR(4, 81, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("brainpy._c.composition"); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7brainpy_2_c_11composition_PyComposition = __Pyx_ImportType(__pyx_t_1, "brainpy._c.composition", "PyComposition", sizeof(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7brainpy_2_c_11composition_PyComposition) __PYX_ERR(5, 123, __pyx_L1_error)
  __pyx_vtabptr_7brainpy_2_c_11composition_PyComposition = (struct __pyx_vtabstruct_7brainpy_2_c_11composition_PyComposition*)__Pyx_GetVtable(__pyx_ptype_7brainpy_2_c_11composition_PyComposition->tp_dict); if (unlikely(!__pyx_vtabptr_7brainpy_2_c_11composition_PyComposition)) __PYX_ERR(5, 123, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("brainpy._c.isotopic_distribution"); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak = __Pyx_ImportType(__pyx_t_1, "brainpy._c.isotopic_distribution", "TheoreticalPeak", sizeof(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak) __PYX_ERR(6, 63, __pyx_L1_error)
  __pyx_vtabptr_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak = (struct __pyx_vtabstruct_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak*)__Pyx_GetVtable(__pyx_ptype_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak->tp_dict); if (unlikely(!__pyx_vtabptr_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak)) __PYX_ERR(6, 63, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("numpy"); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5numpy_dtype = __Pyx_ImportType(__pyx_t_1, "numpy", "dtype", sizeof(PyArray_Descr), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_dtype) __PYX_ERR(2, 206, __pyx_L1_error)
  __pyx_ptype_5numpy_flatiter = __Pyx_ImportType(__pyx_t_1, "numpy", "flatiter", sizeof(PyArrayIterObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_flatiter) __PYX_ERR(2, 229, __pyx_L1_error)
  __pyx_ptype_5numpy_broadcast = __Pyx_ImportType(__pyx_t_1, "numpy", "broadcast", sizeof(PyArrayMultiIterObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_broadcast) __PYX_ERR(2, 233, __pyx_L1_error)
  __pyx_ptype_5numpy_ndarray = __Pyx_ImportType(__pyx_t_1, "numpy", "ndarray", sizeof(PyArrayObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_ndarray) __PYX_ERR(2, 242, __pyx_L1_error)
  __pyx_ptype_5numpy_ufunc = __Pyx_ImportType(__pyx_t_1, "numpy", "ufunc", sizeof(PyUFuncObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_ufunc) __PYX_ERR(2, 918, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_peak_picker._c.peak_index"); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_14ms_peak_picker_2_c_10peak_index_PeakIndex = __Pyx_ImportType(__pyx_t_1, "ms_peak_picker._c.peak_index", "PeakIndex", sizeof(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_14ms_peak_picker_2_c_10peak_index_PeakIndex) __PYX_ERR(7, 5, __pyx_L1_error)
  __pyx_vtabptr_14ms_peak_picker_2_c_10peak_index_PeakIndex = (struct __pyx_vtabstruct_14ms_peak_picker_2_c_10peak_index_PeakIndex*)__Pyx_GetVtable(__pyx_ptype_14ms_peak_picker_2_c_10peak_index_PeakIndex->tp_dict); if (unlikely(!__pyx_vtabptr_14ms_peak_picker_2_c_10peak_index_PeakIndex)) __PYX_ERR(7, 5, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_deisotope._c.averagine"); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_12ms_deisotope_2_c_9averagine_Averagine = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.averagine", "Averagine", sizeof(struct __pyx_obj_12ms_deisotope_2_c_9averagine_Averagine), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_9averagine_Averagine) __PYX_ERR(8, 12, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_9averagine_Averagine = (struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_Averagine*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_9averagine_Averagine->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_9averagine_Averagine)) __PYX_ERR(8, 12, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.averagine", "TheoreticalIsotopicPattern", sizeof(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern) __PYX_ERR(8, 23, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern = (struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern)) __PYX_ERR(8, 23, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_9averagine_AveragineCache = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.averagine", "AveragineCache", sizeof(struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_9averagine_AveragineCache) __PYX_ERR(8, 79, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_9averagine_AveragineCache = (struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_AveragineCache*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_9averagine_AveragineCache->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_9averagine_AveragineCache)) __PYX_ERR(8, 79, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_deisotope._c.deconvoluter_base"); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.deconvoluter_base", "DeconvoluterBase", sizeof(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase) __PYX_ERR(9, 25, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase = (struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase)) __PYX_ERR(9, 25, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.deconvoluter_base", "AveragineDeconvoluterBase", sizeof(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase) __PYX_ERR(9, 57, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase = (struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase)) __PYX_ERR(9, 57, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.deconvoluter_base", "MultiAveragineDeconvoluterBase", sizeof(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase) __PYX_ERR(9, 69, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase = (struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase)) __PYX_ERR(9, 69, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.deconvoluter_base", "ChargeIterator", sizeof(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator) __PYX_ERR(9, 91, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator = (struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator)) __PYX_ERR(9, 91, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_deisotope._c.scoring"); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitRecord = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "IsotopicFitRecord", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitRecord) __PYX_ERR(10, 16, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_IsotopicFitRecord = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitRecord*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitRecord->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_IsotopicFitRecord)) __PYX_ERR(10, 16, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_FitSelectorBase = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "FitSelectorBase", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_FitSelectorBase) __PYX_ERR(10, 38, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_FitSelectorBase = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FitSelectorBase*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_FitSelectorBase->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_FitSelectorBase)) __PYX_ERR(10, 38, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitterBase = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "IsotopicFitterBase", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitterBase) __PYX_ERR(10, 50, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_IsotopicFitterBase = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitterBase->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_IsotopicFitterBase)) __PYX_ERR(10, 50, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_LeastSquaresFitter = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "LeastSquaresFitter", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_LeastSquaresFitter), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_LeastSquaresFitter) __PYX_ERR(10, 62, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_LeastSquaresFitter = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_LeastSquaresFitter*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_LeastSquaresFitter->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_LeastSquaresFitter)) __PYX_ERR(10, 62, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_GTestFitter = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "GTestFitter", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_GTestFitter), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_GTestFitter) __PYX_ERR(10, 66, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_GTestFitter = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_GTestFitter*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_GTestFitter->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_GTestFitter)) __PYX_ERR(10, 66, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledGTestFitter = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "ScaledGTestFitter", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledGTestFitter), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledGTestFitter) __PYX_ERR(10, 69, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_ScaledGTestFitter = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_ScaledGTestFitter*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledGTestFitter->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_ScaledGTestFitter)) __PYX_ERR(10, 69, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_MSDeconVFitter = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "MSDeconVFitter", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_MSDeconVFitter) __PYX_ERR(10, 72, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_MSDeconVFitter = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_MSDeconVFitter*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_MSDeconVFitter->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_MSDeconVFitter)) __PYX_ERR(10, 72, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "PenalizedMSDeconVFitter", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter) __PYX_ERR(10, 78, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter)) __PYX_ERR(10, 78, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_FunctionScorer = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "FunctionScorer", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FunctionScorer), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_FunctionScorer) __PYX_ERR(10, 84, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_FunctionScorer = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FunctionScorer*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_FunctionScorer->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_FunctionScorer)) __PYX_ERR(10, 84, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_InterferenceDetection = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "InterferenceDetection", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_InterferenceDetection) __PYX_ERR(10, 89, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_InterferenceDetection = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_InterferenceDetection*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_InterferenceDetection->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_InterferenceDetection)) __PYX_ERR(10, 89, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_DistinctPatternFitter = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "DistinctPatternFitter", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_DistinctPatternFitter) __PYX_ERR(10, 95, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_DistinctPatternFitter = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_DistinctPatternFitter*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_DistinctPatternFitter->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_DistinctPatternFitter)) __PYX_ERR(10, 95, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "ScaledPenalizedMSDeconvFitter", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter) __PYX_ERR(10, 102, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter)) __PYX_ERR(10, 102, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_DotProductFitter = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "DotProductFitter", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_DotProductFitter), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_DotProductFitter) __PYX_ERR(10, 112, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_DotProductFitter = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_DotProductFitter*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_DotProductFitter->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_DotProductFitter)) __PYX_ERR(10, 112, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_deisotope._c.feature_map.lcms_feature"); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.feature_map.lcms_feature", "LCMSFeatureTreeList", sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList) __PYX_ERR(11, 4, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList = (struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList)) __PYX_ERR(11, 4, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.feature_map.lcms_feature", "LCMSFeatureTreeNodeBase", sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase) __PYX_ERR(11, 23, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase = (struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase)) __PYX_ERR(11, 23, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.feature_map.lcms_feature", "LCMSFeatureTreeNode", sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode) __PYX_ERR(11, 44, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode = (struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode)) __PYX_ERR(11, 44, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.feature_map.lcms_feature", "FeatureBase", sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase) __PYX_ERR(11, 56, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase = (struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase)) __PYX_ERR(11, 56, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.feature_map.lcms_feature", "LCMSFeature", sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature) __PYX_ERR(11, 72, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature = (struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature)) __PYX_ERR(11, 72, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.feature_map.lcms_feature", "EmptyFeature", sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature) __PYX_ERR(11, 96, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature = (struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature)) __PYX_ERR(11, 96, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.feature_map.lcms_feature", "FeatureSetIterator", sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator) __PYX_ERR(11, 101, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator = (struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator)) __PYX_ERR(11, 101, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.feature_map.lcms_feature", "RunningWeightedAverage", sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage) __PYX_ERR(11, 129, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage = (struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage)) __PYX_ERR(11, 129, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.feature_map.lcms_feature", "RunningWeightedAverageNeutralMass", sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass) __PYX_ERR(11, 148, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass = (struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass)) __PYX_ERR(11, 148, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_deisotope._c.feature_map.feature_map"); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.feature_map.feature_map", "LCMSFeatureMap", sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap) __PYX_ERR(12, 5, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap = (struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap)) __PYX_ERR(12, 5, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_deisotope._c.peak_set"); if (unlikely(!__pyx_t_1)) __PYX_ERR(13, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_12ms_deisotope_2_c_8peak_set__Index = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.peak_set", "_Index", sizeof(struct __pyx_obj_12ms_deisotope_2_c_8peak_set__Index), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_8peak_set__Index) __PYX_ERR(13, 5, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_8peak_set__Index = (struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set__Index*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_8peak_set__Index->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_8peak_set__Index)) __PYX_ERR(13, 5, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_8peak_set_EnvelopePair = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.peak_set", "EnvelopePair", sizeof(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_EnvelopePair), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_8peak_set_EnvelopePair) __PYX_ERR(13, 14, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_8peak_set_EnvelopePair = (struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_EnvelopePair*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_8peak_set_EnvelopePair->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_EnvelopePair)) __PYX_ERR(13, 14, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_8peak_set_Envelope = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.peak_set", "Envelope", sizeof(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_8peak_set_Envelope) __PYX_ERR(13, 25, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_8peak_set_Envelope = (struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_Envelope*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_8peak_set_Envelope->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_Envelope)) __PYX_ERR(13, 25, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.peak_set", "DeconvolutedPeak", sizeof(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak) __PYX_ERR(13, 40, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak = (struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak)) __PYX_ERR(13, 40, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.peak_set", "DeconvolutedPeakSolution", sizeof(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution) __PYX_ERR(13, 66, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution = (struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution)) __PYX_ERR(13, 66, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.peak_set", "DeconvolutedPeakSet", sizeof(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet) __PYX_ERR(13, 70, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet = (struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet)) __PYX_ERR(13, 70, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.peak_set", "DeconvolutedPeakSetIndexed", sizeof(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed) __PYX_ERR(13, 99, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed = (struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed)) __PYX_ERR(13, 99, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_deisotope._c.feature_map.feature_fit"); if (unlikely(!__pyx_t_1)) __PYX_ERR(14, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_12ms_deisotope_2_c_11feature_map_11feature_fit_map_coord = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.feature_map.feature_fit", "map_coord", sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_map_coord), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_11feature_map_11feature_fit_map_coord) __PYX_ERR(14, 8, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.feature_map.feature_fit", "LCMSFeatureSetFit", sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit) __PYX_ERR(14, 14, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit = (struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit)) __PYX_ERR(14, 14, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __pyx_t_1 = PyImport_ImportModule("ms_peak_picker._c.peak_set"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "INTERVAL_INDEX_SIZE", (void **)&__pyx_vp_14ms_peak_picker_2_c_8peak_set_INTERVAL_INDEX_SIZE, "size_t") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_2 = PyImport_ImportModule("brainpy._c.composition"); if (!__pyx_t_2) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_2, "nist_mass", (void **)&__pyx_vp_7brainpy_2_c_11composition_nist_mass, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_2, "PROTON", (void **)&__pyx_vp_7brainpy_2_c_11composition_PROTON, "double") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_2, "_ElementTable", (void **)&__pyx_vp_7brainpy_2_c_11composition__ElementTable, "struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = PyImport_ImportModule("brainpy._c.isotopic_constants"); if (!__pyx_t_3) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_3, "DEFAULT_ISOTOPIC_CONSTANTS_SIZE", (void **)&__pyx_vp_7brainpy_2_c_18isotopic_constants_DEFAULT_ISOTOPIC_CONSTANTS_SIZE, "size_t") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = PyImport_ImportModule("ms_deisotope._c.averagine"); if (!__pyx_t_4) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_4, "PROTON", (void **)&__pyx_vp_12ms_deisotope_2_c_9averagine_PROTON, "double") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_5 = PyImport_ImportModule("ms_deisotope._c.scoring"); if (!__pyx_t_5) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_5, "INFINITY", (void **)&__pyx_vp_12ms_deisotope_2_c_7scoring_INFINITY, "double") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_6 = PyImport_ImportModule("ms_deisotope._c.peak_set"); if (!__pyx_t_6) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_6, "INTERVAL_INDEX_SIZE", (void **)&__pyx_vp_12ms_deisotope_2_c_8peak_set_INTERVAL_INDEX_SIZE, "size_t") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __pyx_t_1 = PyImport_ImportModule("brainpy._c.double_vector"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "make_double_vector", (void (**)(void))&__pyx_f_7brainpy_2_c_13double_vector_make_double_vector, "struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector *(void)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "double_vector_append", (void (**)(void))&__pyx_f_7brainpy_2_c_13double_vector_double_vector_append, "int (struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector *, double)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "free_double_vector", (void (**)(void))&__pyx_f_7brainpy_2_c_13double_vector_free_double_vector, "void (struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "double_vector_to_list", (void (**)(void))&__pyx_f_7brainpy_2_c_13double_vector_double_vector_to_list, "PyObject *(struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_2 = PyImport_ImportModule("brainpy._c.isotopic_distribution"); if (!__pyx_t_2) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_2, "_isotopic_variants", (void (**)(void))&__pyx_f_7brainpy_2_c_21isotopic_distribution__isotopic_variants, "PyObject *(PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution__isotopic_variants *__pyx_optional_args)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = PyImport_ImportModule("ms_deisotope._c.averagine"); if (!__pyx_t_3) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_3, "neutral_mass", (void (**)(void))&__pyx_f_12ms_deisotope_2_c_9averagine_neutral_mass, "double (double, int, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_neutral_mass *__pyx_optional_args)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_RefNannyFinishContext();
  return -1;
}


#if PY_MAJOR_VERSION < 3
#ifdef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC void
#else
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#endif
#else
#ifdef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initfeature_processor(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initfeature_processor(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_feature_processor(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_feature_processor(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_feature_processor(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  double __pyx_t_4;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'feature_processor' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_feature_processor(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("feature_processor", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_ms_deisotope___c__feature_map__feature_processor) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "ms_deisotope._c.feature_map.feature_processor")) {
      if (unlikely(PyDict_SetItemString(modules, "ms_deisotope._c.feature_map.feature_processor", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) goto __pyx_L1_error;
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) goto __pyx_L1_error;
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely(__Pyx_modinit_type_init_code() != 0)) goto __pyx_L1_error;
  if (unlikely(__Pyx_modinit_type_import_code() != 0)) goto __pyx_L1_error;
  if (unlikely(__Pyx_modinit_variable_import_code() != 0)) goto __pyx_L1_error;
  if (unlikely(__Pyx_modinit_function_import_code() != 0)) goto __pyx_L1_error;
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":34
 * 
 * cimport numpy as np
 * import numpy as np             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_numpy, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_np, __pyx_t_1) < 0) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":37
 * 
 * 
 * np.import_array()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __pyx_f_5numpy_import_array(); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 37, __pyx_L1_error)

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":40
 * 
 * 
 * cdef object zeros = np.zeros             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_XGOTREF(__pyx_v_12ms_deisotope_2_c_11feature_map_17feature_processor_zeros);
  __Pyx_DECREF_SET(__pyx_v_12ms_deisotope_2_c_11feature_map_17feature_processor_zeros, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":43
 * 
 * 
 * cdef double INF = float('inf')             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_4 = __Pyx_PyObject_AsDouble(__pyx_n_s_inf); if (unlikely(__pyx_t_4 == ((double)((double)-1)) && PyErr_Occurred())) __PYX_ERR(0, 43, __pyx_L1_error)
  __pyx_v_12ms_deisotope_2_c_11feature_map_17feature_processor_INF = __pyx_t_4;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":246
 *         public AveragineCache averagine
 * 
 *     cpdef list create_theoretical_distribution(self, double mz, int charge, double charge_carrier=PROTON, double truncate_after=0.8,             # <<<<<<<<<<<<<<
 *                                                double ignore_below=0.05):
 *         cdef:
 */
  __pyx_k__3 = __pyx_v_12ms_deisotope_2_c_9averagine_PROTON;
  __pyx_k__3 = __pyx_v_12ms_deisotope_2_c_9averagine_PROTON;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":347
 * 
 *     cpdef LCMSFeatureSetFit _fit_single_feature_set(self, list features, list base_tid, double error_tolerance,
 *                                                     int charge, double charge_carrier=PROTON, int max_missed_peaks=1,             # <<<<<<<<<<<<<<
 *                                                     double threshold_scale=0.3):
 *         cdef:
 */
  __pyx_k__4 = __pyx_v_12ms_deisotope_2_c_9averagine_PROTON;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":346
 *         return acc / count
 * 
 *     cpdef LCMSFeatureSetFit _fit_single_feature_set(self, list features, list base_tid, double error_tolerance,             # <<<<<<<<<<<<<<
 *                                                     int charge, double charge_carrier=PROTON, int max_missed_peaks=1,
 *                                                     double threshold_scale=0.3):
 */
  __pyx_k__4 = __pyx_v_12ms_deisotope_2_c_9averagine_PROTON;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":404
 *     cdef list _fit_theoretical_distribution_on_features(self, double mz, double error_tolerance, int charge,
 *                                                         TheoreticalIsotopicPattern base_tid,
 *                                                         double charge_carrier=PROTON,             # <<<<<<<<<<<<<<
 *                                                         double truncate_after=0.8, int max_missed_peaks=1,
 *                                                         double threshold_scale=0.3, LCMSFeature feature=None):
 */
  __pyx_k__5 = __pyx_v_12ms_deisotope_2_c_9averagine_PROTON;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":510
 * 
 *     cpdef list _fit_feature_set(self, double mz, double error_tolerance, int charge, int left_search=1,
 *                                 int right_search=1, double charge_carrier=PROTON, double truncate_after=0.8,             # <<<<<<<<<<<<<<
 *                                 int max_missed_peaks=1, double threshold_scale=0.3, LCMSFeature feature=None):
 *         cdef:
 */
  __pyx_k__7 = __pyx_v_12ms_deisotope_2_c_9averagine_PROTON;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":509
 *         return feature_fits
 * 
 *     cpdef list _fit_feature_set(self, double mz, double error_tolerance, int charge, int left_search=1,             # <<<<<<<<<<<<<<
 *                                 int right_search=1, double charge_carrier=PROTON, double truncate_after=0.8,
 *                                 int max_missed_peaks=1, double threshold_scale=0.3, LCMSFeature feature=None):
 */
  __pyx_k__7 = __pyx_v_12ms_deisotope_2_c_9averagine_PROTON;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":523
 *         return feature_fits
 * 
 *     cpdef list _fit_composition(self, dict composition, double error_tolerance, int charge, double charge_carrier=PROTON,             # <<<<<<<<<<<<<<
 *                                 double truncate_after=0.8, double ignore_below=0.05, int max_missed_peaks=1,
 *                                 double threshold_scale=0.3, LCMSFeature feature=None):
 */
  __pyx_k__8 = __pyx_v_12ms_deisotope_2_c_9averagine_PROTON;
  __pyx_k__8 = __pyx_v_12ms_deisotope_2_c_9averagine_PROTON;

  /* "(tree fragment)":1
 * def __pyx_unpickle_envelope_conformer(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_3 = PyCFunction_NewEx(&__pyx_mdef_12ms_deisotope_2_c_11feature_map_17feature_processor_3__pyx_unpickle_envelope_conformer, NULL, __pyx_n_s_ms_deisotope__c_feature_map_feat); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_envelope_conforme, __pyx_t_3) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_envelope_conformer__set_state(<envelope_conformer> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_envelope_conformer__set_state(envelope_conformer __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.experimental = __pyx_state[0]; __pyx_result.n_missing = __pyx_state[1]; __pyx_result.theoretical = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_3 = PyCFunction_NewEx(&__pyx_mdef_12ms_deisotope_2_c_11feature_map_17feature_processor_5__pyx_unpickle_LCMSFeatureProcessorBase, NULL, __pyx_n_s_ms_deisotope__c_feature_map_feat); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_LCMSFeatureProces, __pyx_t_3) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":1
 * # cython: embedsignature=True             # <<<<<<<<<<<<<<
 * #
 * 
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_3) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1046
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init ms_deisotope._c.feature_map.feature_processor", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init ms_deisotope._c.feature_map.feature_processor");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (PyCFunction_GET_FLAGS(func) & METH_FASTCALL) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* IsLittleEndian */
static CYTHON_INLINE int __Pyx_Is_Little_Endian(void)
{
  union {
    uint32_t u32;
    uint8_t u8[4];
  } S;
  S.u32 = 0x01020304;
  return S.u8[0] == 4;
}

/* BufferFormatCheck */
static void __Pyx_BufFmt_Init(__Pyx_BufFmt_Context* ctx,
                              __Pyx_BufFmt_StackElem* stack,
                              __Pyx_TypeInfo* type) {
  stack[0].field = &ctx->root;
  stack[0].parent_offset = 0;
  ctx->root.type = type;
  ctx->root.name = "buffer dtype";
  ctx->root.offset = 0;
  ctx->head = stack;
  ctx->head->field = &ctx->root;
  ctx->fmt_offset = 0;
  ctx->head->parent_offset = 0;
  ctx->new_packmode = '@';
  ctx->enc_packmode = '@';
  ctx->new_count = 1;
  ctx->enc_count = 0;
  ctx->enc_type = 0;
  ctx->is_complex = 0;
  ctx->is_valid_array = 0;
  ctx->struct_alignment = 0;
  while (type->typegroup == 'S') {
    ++ctx->head;
    ctx->head->field = type->fields;
    ctx->head->parent_offset = 0;
    type = type->fields->type;
  }
}
static int __Pyx_BufFmt_ParseNumber(const char** ts) {
    int count;
    const char* t = *ts;
    if (*t < '0' || *t > '9') {
      return -1;
    } else {
        count = *t++ - '0';
        while (*t >= '0' && *t <= '9') {
            count *= 10;
            count += *t++ - '0';
        }
    }
    *ts = t;
    return count;
}
static int __Pyx_BufFmt_ExpectNumber(const char **ts) {
    int number = __Pyx_BufFmt_ParseNumber(ts);
    if (number == -1)
        PyErr_Format(PyExc_ValueError,\
                     "Does not understand character buffer dtype format string ('%c')", **ts);
    return number;
}
static void __Pyx_BufFmt_RaiseUnexpectedChar(char ch) {
  PyErr_Format(PyExc_ValueError,
               "Unexpected format string character: '%c'", ch);
}
static const char* __Pyx_BufFmt_DescribeTypeChar(char ch, int is_complex) {
  switch (ch) {
    case 'c': return "'char'";
    case 'b': return "'signed char'";
    case 'B': return "'unsigned char'";
    case 'h': return "'short'";
    case 'H': return "'unsigned short'";
    case 'i': return "'int'";
    case 'I': return "'unsigned int'";
    case 'l': return "'long'";
    case 'L': return "'unsigned long'";
    case 'q': return "'long long'";
    case 'Q': return "'unsigned long long'";
    case 'f': return (is_complex ? "'complex float'" : "'float'");
    case 'd': return (is_complex ? "'complex double'" : "'double'");
    case 'g': return (is_complex ? "'complex long double'" : "'long double'");
    case 'T': return "a struct";
    case 'O': return "Python object";
    case 'P': return "a pointer";
    case 's': case 'p': return "a string";
    case 0: return "end";
    default: return "unparseable format string";
  }
}
static size_t __Pyx_BufFmt_TypeCharToStandardSize(char ch, int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return 2;
    case 'i': case 'I': case 'l': case 'L': return 4;
    case 'q': case 'Q': return 8;
    case 'f': return (is_complex ? 8 : 4);
    case 'd': return (is_complex ? 16 : 8);
    case 'g': {
      PyErr_SetString(PyExc_ValueError, "Python does not define a standard format string size for long double ('g')..");
      return 0;
    }
    case 'O': case 'P': return sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
static size_t __Pyx_BufFmt_TypeCharToNativeSize(char ch, int is_complex) {
  switch (ch) {
    case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(short);
    case 'i': case 'I': return sizeof(int);
    case 'l': case 'L': return sizeof(long);
    #ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(PY_LONG_LONG);
    #endif
    case 'f': return sizeof(float) * (is_complex ? 2 : 1);
    case 'd': return sizeof(double) * (is_complex ? 2 : 1);
    case 'g': return sizeof(long double) * (is_complex ? 2 : 1);
    case 'O': case 'P': return sizeof(void*);
    default: {
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
  }
}
typedef struct { char c; short x; } __Pyx_st_short;
typedef struct { char c; int x; } __Pyx_st_int;
typedef struct { char c; long x; } __Pyx_st_long;
typedef struct { char c; float x; } __Pyx_st_float;
typedef struct { char c; double x; } __Pyx_st_double;
typedef struct { char c; long double x; } __Pyx_st_longdouble;
typedef struct { char c; void *x; } __Pyx_st_void_p;
#ifdef HAVE_LONG_LONG
typedef struct { char c; PY_LONG_LONG x; } __Pyx_st_longlong;
#endif
static size_t __Pyx_BufFmt_TypeCharToAlignment(char ch, CYTHON_UNUSED int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(__Pyx_st_short) - sizeof(short);
    case 'i': case 'I': return sizeof(__Pyx_st_int) - sizeof(int);
    case 'l': case 'L': return sizeof(__Pyx_st_long) - sizeof(long);
#ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(__Pyx_st_longlong) - sizeof(PY_LONG_LONG);
#endif
    case 'f': return sizeof(__Pyx_st_float) - sizeof(float);
    case 'd': return sizeof(__Pyx_st_double) - sizeof(double);
    case 'g': return sizeof(__Pyx_st_longdouble) - sizeof(long double);
    case 'P': case 'O': return sizeof(__Pyx_st_void_p) - sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
/* These are for computing the padding at the end of the struct to align
   on the first member of the struct. This will probably the same as above,
   but we don't have any guarantees.
 */
typedef struct { short x; char c; } __Pyx_pad_short;
typedef struct { int x; char c; } __Pyx_pad_int;
typedef struct { long x; char c; } __Pyx_pad_long;
typedef struct { float x; char c; } __Pyx_pad_float;
typedef struct { double x; char c; } __Pyx_pad_double;
typedef struct { long double x; char c; } __Pyx_pad_longdouble;
typedef struct { void *x; char c; } __Pyx_pad_void_p;
#ifdef HAVE_LONG_LONG
typedef struct { PY_LONG_LONG x; char c; } __Pyx_pad_longlong;
#endif
static size_t __Pyx_BufFmt_TypeCharToPadding(char ch, CYTHON_UNUSED int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(__Pyx_pad_short) - sizeof(short);
    case 'i': case 'I': return sizeof(__Pyx_pad_int) - sizeof(int);
    case 'l': case 'L': return sizeof(__Pyx_pad_long) - sizeof(long);
#ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(__Pyx_pad_longlong) - sizeof(PY_LONG_LONG);
#endif
    case 'f': return sizeof(__Pyx_pad_float) - sizeof(float);
    case 'd': return sizeof(__Pyx_pad_double) - sizeof(double);
    case 'g': return sizeof(__Pyx_pad_longdouble) - sizeof(long double);
    case 'P': case 'O': return sizeof(__Pyx_pad_void_p) - sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
static char __Pyx_BufFmt_TypeCharToGroup(char ch, int is_complex) {
  switch (ch) {
    case 'c':
        return 'H';
    case 'b': case 'h': case 'i':
    case 'l': case 'q': case 's': case 'p':
        return 'I';
    case 'B': case 'H': case 'I': case 'L': case 'Q':
        return 'U';
    case 'f': case 'd': case 'g':
        return (is_complex ? 'C' : 'R');
    case 'O':
        return 'O';
    case 'P':
        return 'P';
    default: {
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
  }
}
static void __Pyx_BufFmt_RaiseExpected(__Pyx_BufFmt_Context* ctx) {
  if (ctx->head == NULL || ctx->head->field == &ctx->root) {
    const char* expected;
    const char* quote;
    if (ctx->head == NULL) {
      expected = "end";
      quote = "";
    } else {
      expected = ctx->head->field->type->name;
      quote = "'";
    }
    PyErr_Format(PyExc_ValueError,
                 "Buffer dtype mismatch, expected %s%s%s but got %s",
                 quote, expected, quote,
                 __Pyx_BufFmt_DescribeTypeChar(ctx->enc_type, ctx->is_complex));
  } else {
    __Pyx_StructField* field = ctx->head->field;
    __Pyx_StructField* parent = (ctx->head - 1)->field;
    PyErr_Format(PyExc_ValueError,
                 "Buffer dtype mismatch, expected '%s' but got %s in '%s.%s'",
                 field->type->name, __Pyx_BufFmt_DescribeTypeChar(ctx->enc_type, ctx->is_complex),
                 parent->type->name, field->name);
  }
}
static int __Pyx_BufFmt_ProcessTypeChunk(__Pyx_BufFmt_Context* ctx) {
  char group;
  size_t size, offset, arraysize = 1;
  if (ctx->enc_type == 0) return 0;
  if (ctx->head->field->type->arraysize[0]) {
    int i, ndim = 0;
    if (ctx->enc_type == 's' || ctx->enc_type == 'p') {
        ctx->is_valid_array = ctx->head->field->type->ndim == 1;
        ndim = 1;
        if (ctx->enc_count != ctx->head->field->type->arraysize[0]) {
            PyErr_Format(PyExc_ValueError,
                         "Expected a dimension of size %zu, got %zu",
                         ctx->head->field->type->arraysize[0], ctx->enc_count);
            return -1;
        }
    }
    if (!ctx->is_valid_array) {
      PyErr_Format(PyExc_ValueError, "Expected %d dimensions, got %d",
                   ctx->head->field->type->ndim, ndim);
      return -1;
    }
    for (i = 0; i < ctx->head->field->type->ndim; i++) {
      arraysize *= ctx->head->field->type->arraysize[i];
    }
    ctx->is_valid_array = 0;
    ctx->enc_count = 1;
  }
  group = __Pyx_BufFmt_TypeCharToGroup(ctx->enc_type, ctx->is_complex);
  do {
    __Pyx_StructField* field = ctx->head->field;
    __Pyx_TypeInfo* type = field->type;
    if (ctx->enc_packmode == '@' || ctx->enc_packmode == '^') {
      size = __Pyx_BufFmt_TypeCharToNativeSize(ctx->enc_type, ctx->is_complex);
    } else {
      size = __Pyx_BufFmt_TypeCharToStandardSize(ctx->enc_type, ctx->is_complex);
    }
    if (ctx->enc_packmode == '@') {
      size_t align_at = __Pyx_BufFmt_TypeCharToAlignment(ctx->enc_type, ctx->is_complex);
      size_t align_mod_offset;
      if (align_at == 0) return -1;
      align_mod_offset = ctx->fmt_offset % align_at;
      if (align_mod_offset > 0) ctx->fmt_offset += align_at - align_mod_offset;
      if (ctx->struct_alignment == 0)
          ctx->struct_alignment = __Pyx_BufFmt_TypeCharToPadding(ctx->enc_type,
                                                                 ctx->is_complex);
    }
    if (type->size != size || type->typegroup != group) {
      if (type->typegroup == 'C' && type->fields != NULL) {
        size_t parent_offset = ctx->head->parent_offset + field->offset;
        ++ctx->head;
        ctx->head->field = type->fields;
        ctx->head->parent_offset = parent_offset;
        continue;
      }
      if ((type->typegroup == 'H' || group == 'H') && type->size == size) {
      } else {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return -1;
      }
    }
    offset = ctx->head->parent_offset + field->offset;
    if (ctx->fmt_offset != offset) {
      PyErr_Format(PyExc_ValueError,
                   "Buffer dtype mismatch; next field is at offset %" CYTHON_FORMAT_SSIZE_T "d but %" CYTHON_FORMAT_SSIZE_T "d expected",
                   (Py_ssize_t)ctx->fmt_offset, (Py_ssize_t)offset);
      return -1;
    }
    ctx->fmt_offset += size;
    if (arraysize)
      ctx->fmt_offset += (arraysize - 1) * size;
    --ctx->enc_count;
    while (1) {
      if (field == &ctx->root) {
        ctx->head = NULL;
        if (ctx->enc_count != 0) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return -1;
        }
        break;
      }
      ctx->head->field = ++field;
      if (field->type == NULL) {
        --ctx->head;
        field = ctx->head->field;
        continue;
      } else if (field->type->typegroup == 'S') {
        size_t parent_offset = ctx->head->parent_offset + field->offset;
        if (field->type->fields->type == NULL) continue;
        field = field->type->fields;
        ++ctx->head;
        ctx->head->field = field;
        ctx->head->parent_offset = parent_offset;
        break;
      } else {
        break;
      }
    }
  } while (ctx->enc_count);
  ctx->enc_type = 0;
  ctx->is_complex = 0;
  return 0;
}
static PyObject *
__pyx_buffmt_parse_array(__Pyx_BufFmt_Context* ctx, const char** tsp)
{
    const char *ts = *tsp;
    int i = 0, number;
    int ndim = ctx->head->field->type->ndim;
;
    ++ts;
    if (ctx->new_count != 1) {
        PyErr_SetString(PyExc_ValueError,
                        "Cannot handle repeated arrays in format string");
        return NULL;
    }
    if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
    while (*ts && *ts != ')') {
        switch (*ts) {
            case ' ': case '\f': case '\r': case '\n': case '\t': case '\v':  continue;
            default:  break;
        }
        number = __Pyx_BufFmt_ExpectNumber(&ts);
        if (number == -1) return NULL;
        if (i < ndim && (size_t) number != ctx->head->field->type->arraysize[i])
            return PyErr_Format(PyExc_ValueError,
                        "Expected a dimension of size %zu, got %d",
                        ctx->head->field->type->arraysize[i], number);
        if (*ts != ',' && *ts != ')')
            return PyErr_Format(PyExc_ValueError,
                                "Expected a comma in format string, got '%c'", *ts);
        if (*ts == ',') ts++;
        i++;
    }
    if (i != ndim)
        return PyErr_Format(PyExc_ValueError, "Expected %d dimension(s), got %d",
                            ctx->head->field->type->ndim, i);
    if (!*ts) {
        PyErr_SetString(PyExc_ValueError,
                        "Unexpected end of format string, expected ')'");
        return NULL;
    }
    ctx->is_valid_array = 1;
    ctx->new_count = 1;
    *tsp = ++ts;
    return Py_None;
}
static const char* __Pyx_BufFmt_CheckString(__Pyx_BufFmt_Context* ctx, const char* ts) {
  int got_Z = 0;
  while (1) {
    switch(*ts) {
      case 0:
        if (ctx->enc_type != 0 && ctx->head == NULL) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return NULL;
        }
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        if (ctx->head != NULL) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return NULL;
        }
        return ts;
      case ' ':
      case '\r':
      case '\n':
        ++ts;
        break;
      case '<':
        if (!__Pyx_Is_Little_Endian()) {
          PyErr_SetString(PyExc_ValueError, "Little-endian buffer not supported on big-endian compiler");
          return NULL;
        }
        ctx->new_packmode = '=';
        ++ts;
        break;
      case '>':
      case '!':
        if (__Pyx_Is_Little_Endian()) {
          PyErr_SetString(PyExc_ValueError, "Big-endian buffer not supported on little-endian compiler");
          return NULL;
        }
        ctx->new_packmode = '=';
        ++ts;
        break;
      case '=':
      case '@':
      case '^':
        ctx->new_packmode = *ts++;
        break;
      case 'T':
        {
          const char* ts_after_sub;
          size_t i, struct_count = ctx->new_count;
          size_t struct_alignment = ctx->struct_alignment;
          ctx->new_count = 1;
          ++ts;
          if (*ts != '{') {
            PyErr_SetString(PyExc_ValueError, "Buffer acquisition: Expected '{' after 'T'");
            return NULL;
          }
          if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
          ctx->enc_type = 0;
          ctx->enc_count = 0;
          ctx->struct_alignment = 0;
          ++ts;
          ts_after_sub = ts;
          for (i = 0; i != struct_count; ++i) {
            ts_after_sub = __Pyx_BufFmt_CheckString(ctx, ts);
            if (!ts_after_sub) return NULL;
          }
          ts = ts_after_sub;
          if (struct_alignment) ctx->struct_alignment = struct_alignment;
        }
        break;
      case '}':
        {
          size_t alignment = ctx->struct_alignment;
          ++ts;
          if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
          ctx->enc_type = 0;
          if (alignment && ctx->fmt_offset % alignment) {
            ctx->fmt_offset += alignment - (ctx->fmt_offset % alignment);
          }
        }
        return ts;
      case 'x':
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        ctx->fmt_offset += ctx->new_count;
        ctx->new_count = 1;
        ctx->enc_count = 0;
        ctx->enc_type = 0;
        ctx->enc_packmode = ctx->new_packmode;
        ++ts;
        break;
      case 'Z':
        got_Z = 1;
        ++ts;
        if (*ts != 'f' && *ts != 'd' && *ts != 'g') {
          __Pyx_BufFmt_RaiseUnexpectedChar('Z');
          return NULL;
        }
        CYTHON_FALLTHROUGH;
      case 'c': case 'b': case 'B': case 'h': case 'H': case 'i': case 'I':
      case 'l': case 'L': case 'q': case 'Q':
      case 'f': case 'd': case 'g':
      case 'O': case 'p':
        if (ctx->enc_type == *ts && got_Z == ctx->is_complex &&
            ctx->enc_packmode == ctx->new_packmode) {
          ctx->enc_count += ctx->new_count;
          ctx->new_count = 1;
          got_Z = 0;
          ++ts;
          break;
        }
        CYTHON_FALLTHROUGH;
      case 's':
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        ctx->enc_count = ctx->new_count;
        ctx->enc_packmode = ctx->new_packmode;
        ctx->enc_type = *ts;
        ctx->is_complex = got_Z;
        ++ts;
        ctx->new_count = 1;
        got_Z = 0;
        break;
      case ':':
        ++ts;
        while(*ts != ':') ++ts;
        ++ts;
        break;
      case '(':
        if (!__pyx_buffmt_parse_array(ctx, &ts)) return NULL;
        break;
      default:
        {
          int number = __Pyx_BufFmt_ExpectNumber(&ts);
          if (number == -1) return NULL;
          ctx->new_count = (size_t)number;
        }
    }
  }
}

/* BufferGetAndValidate */
  static CYTHON_INLINE void __Pyx_SafeReleaseBuffer(Py_buffer* info) {
  if (unlikely(info->buf == NULL)) return;
  if (info->suboffsets == __Pyx_minusones) info->suboffsets = NULL;
  __Pyx_ReleaseBuffer(info);
}
static void __Pyx_ZeroBuffer(Py_buffer* buf) {
  buf->buf = NULL;
  buf->obj = NULL;
  buf->strides = __Pyx_zeros;
  buf->shape = __Pyx_zeros;
  buf->suboffsets = __Pyx_minusones;
}
static int __Pyx__GetBufferAndValidate(
        Py_buffer* buf, PyObject* obj,  __Pyx_TypeInfo* dtype, int flags,
        int nd, int cast, __Pyx_BufFmt_StackElem* stack)
{
  buf->buf = NULL;
  if (unlikely(__Pyx_GetBuffer(obj, buf, flags) == -1)) {
    __Pyx_ZeroBuffer(buf);
    return -1;
  }
  if (unlikely(buf->ndim != nd)) {
    PyErr_Format(PyExc_ValueError,
                 "Buffer has wrong number of dimensions (expected %d, got %d)",
                 nd, buf->ndim);
    goto fail;
  }
  if (!cast) {
    __Pyx_BufFmt_Context ctx;
    __Pyx_BufFmt_Init(&ctx, stack, dtype);
    if (!__Pyx_BufFmt_CheckString(&ctx, buf->format)) goto fail;
  }
  if (unlikely((size_t)buf->itemsize != dtype->size)) {
    PyErr_Format(PyExc_ValueError,
      "Item size of buffer (%" CYTHON_FORMAT_SSIZE_T "d byte%s) does not match size of '%s' (%" CYTHON_FORMAT_SSIZE_T "d byte%s)",
      buf->itemsize, (buf->itemsize > 1) ? "s" : "",
      dtype->name, (Py_ssize_t)dtype->size, (dtype->size > 1) ? "s" : "");
    goto fail;
  }
  if (buf->suboffsets == NULL) buf->suboffsets = __Pyx_minusones;
  return 0;
fail:;
  __Pyx_SafeReleaseBuffer(buf);
  return -1;
}

/* BufferFallbackError */
  static void __Pyx_RaiseBufferFallbackError(void) {
  PyErr_SetString(PyExc_ValueError,
     "Buffer acquisition failed on assignment; and then reacquiring the old buffer failed too!");
}

/* BufferIndexError */
  static void __Pyx_RaiseBufferIndexError(int axis) {
  PyErr_Format(PyExc_IndexError,
     "Out of bounds on buffer access (axis %d)", axis);
}

/* PyErrFetchRestore */
  #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* RaiseDoubleKeywords */
  static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
  static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_CheckExact(key)) || likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (PyUnicode_GET_SIZE(**name) != PyUnicode_GET_SIZE(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (PyUnicode_GET_SIZE(**argname) != PyUnicode_GET_SIZE(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* RaiseArgTupleInvalid */
  static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* PyDictVersioning */
  #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* PyObjectCallNoArg */
  #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* WriteUnraisableException */
  static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
#ifdef _MSC_VER
    else state = (PyGILState_STATE)-1;
#endif
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* RaiseException */
  #if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* PyErrExceptionMatches */
  #if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* GetAttr */
  static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* GetAttr3 */
  static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r = __Pyx_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
}

/* GetModuleGlobalName */
  #if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* ArgTypeTest */
  static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
    return 0;
}

/* GetItemInt */
  static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* SetItemInt */
  static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v) {
    int r;
    if (!j) return -1;
    r = PyObject_SetItem(o, j, v);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v, int is_list,
                                               CYTHON_NCP_UNUSED int wraparound, CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = (!wraparound) ? i : ((likely(i >= 0)) ? i : i + PyList_GET_SIZE(o));
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o)))) {
            PyObject* old = PyList_GET_ITEM(o, n);
            Py_INCREF(v);
            PyList_SET_ITEM(o, n, v);
            Py_DECREF(old);
            return 1;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_ass_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return -1;
                    PyErr_Clear();
                }
            }
            return m->sq_ass_item(o, i, v);
        }
    }
#else
#if CYTHON_COMPILING_IN_PYPY
    if (is_list || (PySequence_Check(o) && !PyDict_Check(o)))
#else
    if (is_list || PySequence_Check(o))
#endif
    {
        return PySequence_SetItem(o, i, v);
    }
#endif
    return __Pyx_SetItemInt_Generic(o, PyInt_FromSsize_t(i), v);
}

/* Import */
  static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if (strchr(__Pyx_MODULE_NAME, '.')) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
  static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* HasAttr */
  static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!__Pyx_PyBaseString_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_GetAttr(o, n);
    if (unlikely(!r)) {
        PyErr_Clear();
        return 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}

/* DictGetItem */
  #if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            if (unlikely(PyTuple_Check(key))) {
                PyObject* args = PyTuple_Pack(1, key);
                if (likely(args)) {
                    PyErr_SetObject(PyExc_KeyError, args);
                    Py_DECREF(args);
                }
            } else {
                PyErr_SetObject(PyExc_KeyError, key);
            }
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#endif

/* RaiseTooManyValuesToUnpack */
  static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
  static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* RaiseNoneIterError */
  static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* GetTopmostException */
  #if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
  #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* GetException */
  #if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* PyObject_GenericGetAttrNoDict */
  #if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
  #if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetVTable */
  static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* SetupReduce */
  static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStr(meth, __pyx_n_s_name);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    if (_PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate)) goto GOOD;
#else
    if (PyObject_HasAttr(type_obj, __pyx_n_s_getstate)) goto GOOD;
#endif
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_cython); if (unlikely(!reduce_cython)) goto BAD;
            ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto BAD;
            ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto BAD;
            setstate = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate_cython); if (unlikely(!setstate_cython)) goto BAD;
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto BAD;
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto GOOD;
BAD:
    if (!PyErr_Occurred())
        PyErr_Format(PyExc_RuntimeError, "Unable to initialize pickling for %s", ((PyTypeObject*)type_obj)->tp_name);
    ret = -1;
GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* TypeImport */
  #ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* GetVTable */
  static void* __Pyx_GetVtable(PyObject *dict) {
    void* ptr;
    PyObject *ob = PyObject_GetItem(dict, __pyx_n_s_pyx_vtable);
    if (!ob)
        goto bad;
#if PY_VERSION_HEX >= 0x02070000
    ptr = PyCapsule_GetPointer(ob, 0);
#else
    ptr = PyCObject_AsVoidPtr(ob);
#endif
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* pyobject_as_double */
  static double __Pyx__PyObject_AsDouble(PyObject* obj) {
    PyObject* float_value;
#if !CYTHON_USE_TYPE_SLOTS
    float_value = PyNumber_Float(obj);  if ((0)) goto bad;
#else
    PyNumberMethods *nb = Py_TYPE(obj)->tp_as_number;
    if (likely(nb) && likely(nb->nb_float)) {
        float_value = nb->nb_float(obj);
        if (likely(float_value) && unlikely(!PyFloat_Check(float_value))) {
            PyErr_Format(PyExc_TypeError,
                "__float__ returned non-float (type %.200s)",
                Py_TYPE(float_value)->tp_name);
            Py_DECREF(float_value);
            goto bad;
        }
    } else if (PyUnicode_CheckExact(obj) || PyBytes_CheckExact(obj)) {
#if PY_MAJOR_VERSION >= 3
        float_value = PyFloat_FromString(obj);
#else
        float_value = PyFloat_FromString(obj, 0);
#endif
    } else {
        PyObject* args = PyTuple_New(1);
        if (unlikely(!args)) goto bad;
        PyTuple_SET_ITEM(args, 0, obj);
        float_value = PyObject_Call((PyObject*)&PyFloat_Type, args, 0);
        PyTuple_SET_ITEM(args, 0, 0);
        Py_DECREF(args);
    }
#endif
    if (likely(float_value)) {
        double value = PyFloat_AS_DOUBLE(float_value);
        Py_DECREF(float_value);
        return value;
    }
bad:
    return (double)-1;
}

/* CLineInTraceback */
  #ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
  static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, (size_t)new_max*sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
  #include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntFromPyVerify */
  #define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

#if PY_MAJOR_VERSION < 3
static int __Pyx_GetBuffer(PyObject *obj, Py_buffer *view, int flags) {
    if (PyObject_CheckBuffer(obj)) return PyObject_GetBuffer(obj, view, flags);
        if (__Pyx_TypeCheck(obj, __pyx_ptype_5numpy_ndarray)) return __pyx_pw_5numpy_7ndarray_1__getbuffer__(obj, view, flags);
    PyErr_Format(PyExc_TypeError, "'%.200s' does not have the buffer interface", Py_TYPE(obj)->tp_name);
    return -1;
}
static void __Pyx_ReleaseBuffer(Py_buffer *view) {
    PyObject *obj = view->obj;
    if (!obj) return;
    if (PyObject_CheckBuffer(obj)) {
        PyBuffer_Release(view);
        return;
    }
    if ((0)) {}
        else if (__Pyx_TypeCheck(obj, __pyx_ptype_5numpy_ndarray)) __pyx_pw_5numpy_7ndarray_3__releasebuffer__(obj, view);
    view->obj = NULL;
    Py_DECREF(obj);
}
#endif


  /* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_Py_intptr_t(Py_intptr_t value) {
    const Py_intptr_t neg_one = (Py_intptr_t) ((Py_intptr_t) 0 - (Py_intptr_t) 1), const_zero = (Py_intptr_t) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(Py_intptr_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(Py_intptr_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(Py_intptr_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(Py_intptr_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(Py_intptr_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(Py_intptr_t),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* Print */
  #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION < 3
static PyObject *__Pyx_GetStdout(void) {
    PyObject *f = PySys_GetObject((char *)"stdout");
    if (!f) {
        PyErr_SetString(PyExc_RuntimeError, "lost sys.stdout");
    }
    return f;
}
static int __Pyx_Print(PyObject* f, PyObject *arg_tuple, int newline) {
    int i;
    if (!f) {
        if (!(f = __Pyx_GetStdout()))
            return -1;
    }
    Py_INCREF(f);
    for (i=0; i < PyTuple_GET_SIZE(arg_tuple); i++) {
        PyObject* v;
        if (PyFile_SoftSpace(f, 1)) {
            if (PyFile_WriteString(" ", f) < 0)
                goto error;
        }
        v = PyTuple_GET_ITEM(arg_tuple, i);
        if (PyFile_WriteObject(v, f, Py_PRINT_RAW) < 0)
            goto error;
        if (PyString_Check(v)) {
            char *s = PyString_AsString(v);
            Py_ssize_t len = PyString_Size(v);
            if (len > 0) {
                switch (s[len-1]) {
                    case ' ': break;
                    case '\f': case '\r': case '\n': case '\t': case '\v':
                        PyFile_SoftSpace(f, 0);
                        break;
                    default:  break;
                }
            }
        }
    }
    if (newline) {
        if (PyFile_WriteString("\n", f) < 0)
            goto error;
        PyFile_SoftSpace(f, 0);
    }
    Py_DECREF(f);
    return 0;
error:
    Py_DECREF(f);
    return -1;
}
#else
static int __Pyx_Print(PyObject* stream, PyObject *arg_tuple, int newline) {
    PyObject* kwargs = 0;
    PyObject* result = 0;
    PyObject* end_string;
    if (unlikely(!__pyx_print)) {
        __pyx_print = PyObject_GetAttr(__pyx_b, __pyx_n_s_print);
        if (!__pyx_print)
            return -1;
    }
    if (stream) {
        kwargs = PyDict_New();
        if (unlikely(!kwargs))
            return -1;
        if (unlikely(PyDict_SetItem(kwargs, __pyx_n_s_file, stream) < 0))
            goto bad;
        if (!newline) {
            end_string = PyUnicode_FromStringAndSize(" ", 1);
            if (unlikely(!end_string))
                goto bad;
            if (PyDict_SetItem(kwargs, __pyx_n_s_end, end_string) < 0) {
                Py_DECREF(end_string);
                goto bad;
            }
            Py_DECREF(end_string);
        }
    } else if (!newline) {
        if (unlikely(!__pyx_print_kwargs)) {
            __pyx_print_kwargs = PyDict_New();
            if (unlikely(!__pyx_print_kwargs))
                return -1;
            end_string = PyUnicode_FromStringAndSize(" ", 1);
            if (unlikely(!end_string))
                return -1;
            if (PyDict_SetItem(__pyx_print_kwargs, __pyx_n_s_end, end_string) < 0) {
                Py_DECREF(end_string);
                return -1;
            }
            Py_DECREF(end_string);
        }
        kwargs = __pyx_print_kwargs;
    }
    result = PyObject_Call(__pyx_print, arg_tuple, kwargs);
    if (unlikely(kwargs) && (kwargs != __pyx_print_kwargs))
        Py_DECREF(kwargs);
    if (!result)
        return -1;
    Py_DECREF(result);
    return 0;
bad:
    if (kwargs != __pyx_print_kwargs)
        Py_XDECREF(kwargs);
    return -1;
}
#endif

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* Declarations */
  #if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return ::std::complex< float >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return x + y*(__pyx_t_float_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      __pyx_t_float_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
  #if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabsf(b.real) >= fabsf(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                float r = b.imag / b.real;
                float s = (float)(1.0) / (b.real + b.imag * r);
                return __pyx_t_float_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            float r = b.real / b.imag;
            float s = (float)(1.0) / (b.imag + b.real * r);
            return __pyx_t_float_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            float denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_float_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrtf(z.real*z.real + z.imag*z.imag);
          #else
            return hypotf(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
            __pyx_t_float_complex z;
            float r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    float denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(a, a);
                    case 3:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, a);
                    case 4:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if (b.imag == 0) {
                    z.real = powf(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2f(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_float(a);
                theta = atan2f(a.imag, a.real);
            }
            lnr = logf(r);
            z_r = expf(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cosf(z_theta);
            z.imag = z_r * sinf(z_theta);
            return z;
        }
    #endif
#endif

/* Declarations */
  #if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return ::std::complex< double >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return x + y*(__pyx_t_double_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      __pyx_t_double_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
  #if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabs(b.real) >= fabs(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                double r = b.imag / b.real;
                double s = (double)(1.0) / (b.real + b.imag * r);
                return __pyx_t_double_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            double r = b.real / b.imag;
            double s = (double)(1.0) / (b.imag + b.real * r);
            return __pyx_t_double_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            double denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_double_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrt(z.real*z.real + z.imag*z.imag);
          #else
            return hypot(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
            __pyx_t_double_complex z;
            double r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    double denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(a, a);
                    case 3:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, a);
                    case 4:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if (b.imag == 0) {
                    z.real = pow(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_double(a);
                theta = atan2(a.imag, a.real);
            }
            lnr = log(r);
            z_r = exp(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cos(z_theta);
            z.imag = z_r * sin(z_theta);
            return z;
        }
    #endif
#endif

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__NPY_TYPES(enum NPY_TYPES value) {
    const enum NPY_TYPES neg_one = (enum NPY_TYPES) ((enum NPY_TYPES) 0 - (enum NPY_TYPES) 1), const_zero = (enum NPY_TYPES) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum NPY_TYPES) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(enum NPY_TYPES) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum NPY_TYPES) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum NPY_TYPES) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum NPY_TYPES) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(enum NPY_TYPES),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
  static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *x) {
    const size_t neg_one = (size_t) ((size_t) 0 - (size_t) 1), const_zero = (size_t) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(size_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(size_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (size_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(size_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 2 * PyLong_SHIFT) {
                            return (size_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 3 * PyLong_SHIFT) {
                            return (size_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 4 * PyLong_SHIFT) {
                            return (size_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (size_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(size_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(size_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(size_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(size_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) ((((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) ((((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) ((((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(size_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            size_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (size_t) -1;
        }
    } else {
        size_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (size_t) -1;
        val = __Pyx_PyInt_As_size_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to size_t");
    return (size_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to size_t");
    return (size_t) -1;
}

/* CIntFromPy */
  static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
  static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* PrintOne */
  #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION < 3
static int __Pyx_PrintOne(PyObject* f, PyObject *o) {
    if (!f) {
        if (!(f = __Pyx_GetStdout()))
            return -1;
    }
    Py_INCREF(f);
    if (PyFile_SoftSpace(f, 0)) {
        if (PyFile_WriteString(" ", f) < 0)
            goto error;
    }
    if (PyFile_WriteObject(o, f, Py_PRINT_RAW) < 0)
        goto error;
    if (PyFile_WriteString("\n", f) < 0)
        goto error;
    Py_DECREF(f);
    return 0;
error:
    Py_DECREF(f);
    return -1;
    /* the line below is just to avoid C compiler
     * warnings about unused functions */
    return __Pyx_Print(f, NULL, 0);
}
#else
static int __Pyx_PrintOne(PyObject* stream, PyObject *o) {
    int res;
    PyObject* arg_tuple = PyTuple_Pack(1, o);
    if (unlikely(!arg_tuple))
        return -1;
    res = __Pyx_Print(stream, arg_tuple, 1);
    Py_DECREF(arg_tuple);
    return res;
}
#endif

/* FastTypeChecks */
  #if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* CheckBinaryVersion */
  static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* VoidPtrImport */
  #ifndef __PYX_HAVE_RT_ImportVoidPtr
#define __PYX_HAVE_RT_ImportVoidPtr
static int __Pyx_ImportVoidPtr(PyObject *module, const char *name, void **p, const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, name);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C variable %.200s",
                PyModule_GetName(module), name);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C variable %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), name, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    *p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C variable %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), name, sig, desc);
        goto bad;
    }
    *p = PyCObject_AsVoidPtr(cobj);}
#endif
    if (!(*p))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* FunctionImport */
  #ifndef __PYX_HAVE_RT_ImportFunction
#define __PYX_HAVE_RT_ImportFunction
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, funcname);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C function %.200s",
                PyModule_GetName(module), funcname);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    tmp.p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, desc);
        goto bad;
    }
    tmp.p = PyCObject_AsVoidPtr(cobj);}
#endif
    *f = tmp.fp;
    if (!(*f))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* InitStrings */
  static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
