/* Generated by Cython 0.29.6 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [], 
        "include_dirs": [
            "C:\\Users\\Asus\\Miniconda2\\lib\\site-packages\\numpy\\core\\include"
        ], 
        "name": "ms_deisotope._c.peak_dependency_network.intervals", 
        "sources": [
            "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        ]
    }, 
    "module_name": "ms_deisotope._c.peak_dependency_network.intervals"
}
END: Cython Metadata */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_6"
#define CYTHON_HEX_VERSION 0x001D06F0
#define CYTHON_FUTURE_DIVISION 0
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
  #define PyObject_Unicode             PyObject_Str
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? PyMethod_New(func, self) : (Py_INCREF(func), func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif


#define __PYX_ERR(f_index, lineno, Ln_error) \
{ \
  __pyx_filename = __pyx_f[f_index]; __pyx_lineno = lineno; __pyx_clineno = __LINE__; goto Ln_error; \
}

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__ms_deisotope___c__peak_dependency_network__intervals
#define __PYX_HAVE_API__ms_deisotope___c__peak_dependency_network__intervals
/* Early includes */
#include <string.h>
#include <stdio.h>
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "ms_deisotope\\_c\\peak_dependency_network\\intervals.pyx",
  "ms_deisotope\\_c\\peak_dependency_network\\intervals.pxd",
  "stringsource",
  "type.pxd",
};

/*--- Type declarations ---*/
struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin;
struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval;
struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode;

/* "ms_deisotope/_c/peak_dependency_network/intervals.pxd":1
 * cdef class SpanningMixin(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double start
 */
struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin {
  PyObject_HEAD
  struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_vtab;
  double start;
  double end;
};


/* "ms_deisotope/_c/peak_dependency_network/intervals.pxd":19
 * 
 * 
 * cdef class Interval(SpanningMixin):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public object members
 */
struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval {
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin __pyx_base;
  PyObject *members;
  PyObject *data;
};


/* "ms_deisotope/_c/peak_dependency_network/intervals.pxd":25
 * 
 * 
 * cdef class IntervalTreeNode(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double center
 */
struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode {
  PyObject_HEAD
  struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_vtab;
  double center;
  double start;
  double end;
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *left;
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *right;
  PyObject *contained;
  int level;
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *parent;
};



/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":7
 * 
 * 
 * cdef class SpanningMixin(object):             # <<<<<<<<<<<<<<
 *     """Provides methods for checking whether an entity
 *     which has a defined start and end point over a single
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin {
  int (*_contains)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *, double);
  int (*contains)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *, double, int __pyx_skip_dispatch);
  int (*overlaps)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *, int __pyx_skip_dispatch);
  double (*overlap_size)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *, int __pyx_skip_dispatch);
  int (*is_contained_in_interval)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *, int __pyx_skip_dispatch);
  int (*contains_interval)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin;


/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":120
 * 
 * 
 * cdef class Interval(SpanningMixin):             # <<<<<<<<<<<<<<
 *     def __init__(self, start, end, members=None, **kwargs):
 *         if members is None:
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval {
  struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin __pyx_base;
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *__pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval;


/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":158
 * 
 * 
 * cdef class IntervalTreeNode(object):             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, double center, IntervalTreeNode left, list contained, IntervalTreeNode right,
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode {
  PyObject *(*contains_point)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *, double, int __pyx_skip_dispatch);
  PyObject *(*_overlaps_interval)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *, double, double, int __pyx_skip_dispatch);
  PyObject *(*overlaps)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *, double, double, int __pyx_skip_dispatch);
  int (*_eq)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, int nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely((Py_TYPE(obj) == type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
#define __Pyx_GetModuleGlobalNameUncached(var, name)  {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        Py_SIZE(list) = len+1;
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* ListExtend.proto */
static CYTHON_INLINE int __Pyx_PyList_Extend(PyObject* L, PyObject* v) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject* none = _PyList_Extend((PyListObject*)L, v);
    if (unlikely(!none))
        return -1;
    Py_DECREF(none);
    return 0;
#else
    return PyList_SetSlice(L, PY_SSIZE_T_MAX, PY_SSIZE_T_MAX, v);
#endif
}

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod0.proto */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* pop.proto */
static CYTHON_INLINE PyObject* __Pyx__PyObject_Pop(PyObject* L);
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE PyObject* __Pyx_PyList_Pop(PyObject* L);
#define __Pyx_PyObject_Pop(L) (likely(PyList_CheckExact(L)) ?\
    __Pyx_PyList_Pop(L) : __Pyx__PyObject_Pop(L))
#else
#define __Pyx_PyList_Pop(L)  __Pyx__PyObject_Pop(L)
#define __Pyx_PyObject_Pop(L)  __Pyx__PyObject_Pop(L)
#endif

/* UnpackUnboundCMethod.proto */
typedef struct {
    PyObject *type;
    PyObject **method_name;
    PyCFunction func;
    PyObject *method;
    int flag;
} __Pyx_CachedCFunction;

/* CallUnboundCMethod0.proto */
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CallUnboundCMethod0(cfunc, self)\
    (likely((cfunc)->func) ?\
        (likely((cfunc)->flag == METH_NOARGS) ?  (*((cfunc)->func))(self, NULL) :\
         (PY_VERSION_HEX >= 0x030600B1 && likely((cfunc)->flag == METH_FASTCALL) ?\
            (PY_VERSION_HEX >= 0x030700A0 ?\
                (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0) :\
                (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL)) :\
          (PY_VERSION_HEX >= 0x030700A0 && (cfunc)->flag == (METH_FASTCALL | METH_KEYWORDS) ?\
            (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL) :\
            (likely((cfunc)->flag == (METH_VARARGS | METH_KEYWORDS)) ?  ((*(PyCFunctionWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, __pyx_empty_tuple, NULL)) :\
               ((cfunc)->flag == METH_VARARGS ?  (*((cfunc)->func))(self, __pyx_empty_tuple) :\
               __Pyx__CallUnboundCMethod0(cfunc, self)))))) :\
        __Pyx__CallUnboundCMethod0(cfunc, self))
#else
#define __Pyx_CallUnboundCMethod0(cfunc, self)  __Pyx__CallUnboundCMethod0(cfunc, self)
#endif

/* IncludeStringH.proto */
#include <string.h>

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* StrEquals.proto */
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyString_Equals __Pyx_PyUnicode_Equals
#else
#define __Pyx_PyString_Equals __Pyx_PyBytes_Equals
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* append.proto */
static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x);

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* HasAttr.proto */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);

/* CallNextTpTraverse.proto */
static int __Pyx_call_next_tp_traverse(PyObject* obj, visitproc v, void *a, traverseproc current_tp_traverse);

/* CallNextTpClear.proto */
static void __Pyx_call_next_tp_clear(PyObject* obj, inquiry current_tp_dealloc);

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* pyobject_as_double.proto */
static double __Pyx__PyObject_AsDouble(PyObject* obj);
#if CYTHON_COMPILING_IN_PYPY
#define __Pyx_PyObject_AsDouble(obj)\
(likely(PyFloat_CheckExact(obj)) ? PyFloat_AS_DOUBLE(obj) :\
 likely(PyInt_CheckExact(obj)) ?\
 PyFloat_AsDouble(obj) : __Pyx__PyObject_AsDouble(obj))
#else
#define __Pyx_PyObject_AsDouble(obj)\
((likely(PyFloat_CheckExact(obj))) ?\
 PyFloat_AS_DOUBLE(obj) : __Pyx__PyObject_AsDouble(obj))
#endif

/* ClassMethod.proto */
#include "descrobject.h"
static CYTHON_UNUSED PyObject* __Pyx_Method_ClassMethod(PyObject *method);

/* GetNameInClass.proto */
#define __Pyx_GetNameInClass(var, nmspace, name)  (var) = __Pyx__GetNameInClass(nmspace, name)
static PyObject *__Pyx__GetNameInClass(PyObject *nmspace, PyObject *name);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* Print.proto */
static int __Pyx_Print(PyObject*, PyObject *, int);
#if CYTHON_COMPILING_IN_PYPY || PY_MAJOR_VERSION >= 3
static PyObject* __pyx_print = 0;
static PyObject* __pyx_print_kwargs = 0;
#endif

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static int __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin__contains(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, double __pyx_v_i); /* proto*/
static int __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_contains(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, double __pyx_v_i, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_overlaps(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_interval, int __pyx_skip_dispatch); /* proto*/
static double __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_overlap_size(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_interval, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_is_contained_in_interval(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_interval, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_contains_interval(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_interval, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_contains_point(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, double __pyx_v_x, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode__overlaps_interval(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, double __pyx_v_start, double __pyx_v_end, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_overlaps(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, double __pyx_v_start, double __pyx_v_end, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode__eq(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_other, int __pyx_skip_dispatch); /* proto*/

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'cpython' */

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cpython.list' */

/* Module declarations from 'cpython.tuple' */

/* Module declarations from 'ms_deisotope._c.peak_dependency_network.intervals' */
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode = 0;
static double __pyx_v_12ms_deisotope_2_c_23peak_dependency_network_9intervals_INF;
static PyObject *__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_intervals_containing_point(PyObject *, double, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals___pyx_unpickle_SpanningMixin__set_state(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *, PyObject *); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals___pyx_unpickle_Interval__set_state(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *, PyObject *); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals___pyx_unpickle_IntervalTreeNode__set_state(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *, PyObject *); /*proto*/
#define __Pyx_MODULE_NAME "ms_deisotope._c.peak_dependency_network.intervals"
extern int __pyx_module_is_main_ms_deisotope___c__peak_dependency_network__intervals;
int __pyx_module_is_main_ms_deisotope___c__peak_dependency_network__intervals = 0;

/* Implementation of 'ms_deisotope._c.peak_dependency_network.intervals' */
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_AttributeError;
static PyObject *__pyx_builtin_ValueError;
static const char __pyx_k_[] = "!=";
static const char __pyx_k_l[] = "l!=";
static const char __pyx_k_r[] = "r!=";
static const char __pyx_k_x[] = "x";
static const char __pyx_k_eq[] = "_eq";
static const char __pyx_k_end[] = "end";
static const char __pyx_k_inf[] = "inf";
static const char __pyx_k_new[] = "__new__";
static const char __pyx_k_pop[] = "pop";
static const char __pyx_k_dict[] = "__dict__";
static const char __pyx_k_file[] = "file";
static const char __pyx_k_left[] = "left";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_build[] = "build";
static const char __pyx_k_class[] = "__class__";
static const char __pyx_k_deque[] = "deque";
static const char __pyx_k_level[] = "level";
static const char __pyx_k_print[] = "print";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_right[] = "right";
static const char __pyx_k_start[] = "start";
static const char __pyx_k_append[] = "append";
static const char __pyx_k_center[] = "center";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_insert[] = "insert";
static const char __pyx_k_parent[] = "parent";
static const char __pyx_k_pickle[] = "pickle";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_update[] = "update";
static const char __pyx_k_flatten[] = "flatten";
static const char __pyx_k_members[] = "members";
static const char __pyx_k_Interval[] = "Interval";
static const char __pyx_k_contains[] = "contains";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_overlaps[] = "overlaps";
static const char __pyx_k_pyx_type[] = "__pyx_type";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_contained[] = "contained";
static const char __pyx_k_intervals[] = "intervals";
static const char __pyx_k_pyx_state[] = "__pyx_state";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_pyx_result[] = "__pyx_result";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_PickleError[] = "PickleError";
static const char __pyx_k_collections[] = "collections";
static const char __pyx_k_overlap_size[] = "overlap_size";
static const char __pyx_k_pyx_checksum[] = "__pyx_checksum";
static const char __pyx_k_stringsource[] = "stringsource";
static const char __pyx_k_SpanningMixin[] = "SpanningMixin";
static const char __pyx_k_diagnostic_eq[] = "__diagnostic_eq__";
static const char __pyx_k_node_contains[] = "node_contains";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_update_bounds[] = "_update_bounds";
static const char __pyx_k_AttributeError[] = "AttributeError";
static const char __pyx_k_contains_point[] = "contains_point";
static const char __pyx_k_pyx_PickleError[] = "__pyx_PickleError";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_IntervalTreeNode[] = "IntervalTreeNode";
static const char __pyx_k_contains_interval[] = "contains_interval";
static const char __pyx_k_overlaps_interval[] = "_overlaps_interval";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_pyx_unpickle_Interval[] = "__pyx_unpickle_Interval";
static const char __pyx_k_s_start_r_end_r_data_r[] = "%s(start=%r, end=%r, data=%r)";
static const char __pyx_k_is_contained_in_interval[] = "is_contained_in_interval";
static const char __pyx_k_pyx_unpickle_SpanningMixin[] = "__pyx_unpickle_SpanningMixin";
static const char __pyx_k_pyx_unpickle_IntervalTreeNode[] = "__pyx_unpickle_IntervalTreeNode";
static const char __pyx_k_IntervalTreeNode_level_d_center[] = "IntervalTreeNode(level=%d, center=%0.4f, start=%0.4f, end=%0.4f)";
static const char __pyx_k_ms_deisotope__c_peak_dependency[] = "ms_deisotope._c.peak_dependency_network.intervals";
static const char __pyx_k_Incompatible_checksums_s_vs_0x26[] = "Incompatible checksums (%s vs 0x262a9ad = (center, contained, end, left, level, parent, right, start))";
static const char __pyx_k_Incompatible_checksums_s_vs_0xcc[] = "Incompatible checksums (%s vs 0xccc3210 = (end, start))";
static const char __pyx_k_Incompatible_checksums_s_vs_0xea[] = "Incompatible checksums (%s vs 0xeace574 = (data, end, members, start))";
static PyObject *__pyx_kp_s_;
static PyObject *__pyx_n_s_AttributeError;
static PyObject *__pyx_kp_s_Incompatible_checksums_s_vs_0x26;
static PyObject *__pyx_kp_s_Incompatible_checksums_s_vs_0xcc;
static PyObject *__pyx_kp_s_Incompatible_checksums_s_vs_0xea;
static PyObject *__pyx_n_s_Interval;
static PyObject *__pyx_n_s_IntervalTreeNode;
static PyObject *__pyx_kp_s_IntervalTreeNode_level_d_center;
static PyObject *__pyx_n_s_PickleError;
static PyObject *__pyx_n_s_SpanningMixin;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_n_s_append;
static PyObject *__pyx_n_s_build;
static PyObject *__pyx_n_s_center;
static PyObject *__pyx_n_s_class;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_collections;
static PyObject *__pyx_n_s_contained;
static PyObject *__pyx_n_s_contains;
static PyObject *__pyx_n_s_contains_interval;
static PyObject *__pyx_n_s_contains_point;
static PyObject *__pyx_n_s_deque;
static PyObject *__pyx_n_s_diagnostic_eq;
static PyObject *__pyx_n_s_dict;
static PyObject *__pyx_n_s_end;
static PyObject *__pyx_n_s_eq;
static PyObject *__pyx_n_s_file;
static PyObject *__pyx_n_s_flatten;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_inf;
static PyObject *__pyx_n_s_insert;
static PyObject *__pyx_n_s_intervals;
static PyObject *__pyx_n_s_is_contained_in_interval;
static PyObject *__pyx_kp_s_l;
static PyObject *__pyx_n_s_left;
static PyObject *__pyx_n_s_level;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_members;
static PyObject *__pyx_n_s_ms_deisotope__c_peak_dependency;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_new;
static PyObject *__pyx_n_s_node_contains;
static PyObject *__pyx_n_s_overlap_size;
static PyObject *__pyx_n_s_overlaps;
static PyObject *__pyx_n_s_overlaps_interval;
static PyObject *__pyx_n_s_parent;
static PyObject *__pyx_n_s_pickle;
static PyObject *__pyx_n_s_pop;
static PyObject *__pyx_n_s_print;
static PyObject *__pyx_n_s_pyx_PickleError;
static PyObject *__pyx_n_s_pyx_checksum;
static PyObject *__pyx_n_s_pyx_result;
static PyObject *__pyx_n_s_pyx_state;
static PyObject *__pyx_n_s_pyx_type;
static PyObject *__pyx_n_s_pyx_unpickle_Interval;
static PyObject *__pyx_n_s_pyx_unpickle_IntervalTreeNode;
static PyObject *__pyx_n_s_pyx_unpickle_SpanningMixin;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_kp_s_r;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_reduce_cython;
static PyObject *__pyx_n_s_reduce_ex;
static PyObject *__pyx_n_s_right;
static PyObject *__pyx_kp_s_s_start_r_end_r_data_r;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_setstate_cython;
static PyObject *__pyx_n_s_start;
static PyObject *__pyx_kp_s_stringsource;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_update;
static PyObject *__pyx_n_s_update_bounds;
static PyObject *__pyx_n_s_x;
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin___init__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, PyObject *__pyx_v_start, PyObject *__pyx_v_end); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_2__contains__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_4contains(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, double __pyx_v_i); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_6overlaps(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_interval); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_8overlap_size(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_interval); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_10is_contained_in_interval(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_interval); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_12contains_interval(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_interval); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_5start___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_5start_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_3end___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_3end_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_14__reduce_cython__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_16__setstate_cython__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval___init__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *__pyx_v_self, PyObject *__pyx_v_start, PyObject *__pyx_v_end, PyObject *__pyx_v_members, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_2__repr__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_4__getitem__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *__pyx_v_self, PyObject *__pyx_v_k); /* proto */
static Py_ssize_t __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_6__len__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_8__iter__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_7members___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_7members_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_7members_4__del__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_4data___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_4data_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_4data_4__del__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_10__reduce_cython__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_12__setstate_cython__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_intervals_containing_point(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_intervals, double __pyx_v_x); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode___init__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, double __pyx_v_center, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_left, PyObject *__pyx_v_contained, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_right, int __pyx_v_level, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_parent); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_2contains_point(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, double __pyx_v_x); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_4_overlaps_interval(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, double __pyx_v_start, double __pyx_v_end); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6overlaps(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, double __pyx_v_start, double __pyx_v_end); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_8__repr__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_10__richcmp__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_code); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_12_eq(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_other); /* proto */
static Py_hash_t __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_14__hash__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_16__diagnostic_eq__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_18build(PyTypeObject *__pyx_v_cls, PyObject *__pyx_v_intervals); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_20_update_bounds(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_22insert(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, PyObject *__pyx_v_interval); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_24flatten(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_26balance(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_28__iter__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6center___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6center_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5start___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5start_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_3end___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_3end_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_4left___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_4left_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_4left_4__del__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5right___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5right_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5right_4__del__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_9contained___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_9contained_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_9contained_4__del__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5level___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5level_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6parent___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6parent_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6parent_4__del__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_30__reduce_cython__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_32__setstate_cython__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_2__pyx_unpickle_SpanningMixin(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_4__pyx_unpickle_Interval(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_6__pyx_unpickle_IntervalTreeNode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static __Pyx_CachedCFunction __pyx_umethod_PyList_Type_pop = {0, &__pyx_n_s_pop, 0, 0, 0};
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_40020397;
static PyObject *__pyx_int_214708752;
static PyObject *__pyx_int_246211956;
static PyObject *__pyx_int_neg_1;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_codeobj__3;
static PyObject *__pyx_codeobj__5;
static PyObject *__pyx_codeobj__7;
/* Late includes */

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":14
 *     """
 * 
 *     def __init__(self, start, end):             # <<<<<<<<<<<<<<
 *         self.start = start
 *         self.end = end
 */

/* Python wrapper */
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_start = 0;
  PyObject *__pyx_v_end = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_start,&__pyx_n_s_end,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_start)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_end)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, 1); __PYX_ERR(0, 14, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 14, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_start = values[0];
    __pyx_v_end = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 14, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin___init__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *)__pyx_v_self), __pyx_v_start, __pyx_v_end);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin___init__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, PyObject *__pyx_v_start, PyObject *__pyx_v_end) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":15
 * 
 *     def __init__(self, start, end):
 *         self.start = start             # <<<<<<<<<<<<<<
 *         self.end = end
 * 
 */
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_start); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 15, __pyx_L1_error)
  __pyx_v_self->start = __pyx_t_1;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":16
 *     def __init__(self, start, end):
 *         self.start = start
 *         self.end = end             # <<<<<<<<<<<<<<
 * 
 *     def __contains__(self, i):
 */
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_end); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 16, __pyx_L1_error)
  __pyx_v_self->end = __pyx_t_1;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":14
 *     """
 * 
 *     def __init__(self, start, end):             # <<<<<<<<<<<<<<
 *         self.start = start
 *         self.end = end
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":18
 *         self.end = end
 * 
 *     def __contains__(self, i):             # <<<<<<<<<<<<<<
 *         """Tests for point inclusion, `start <= i <= end`
 * 
 */

/* Python wrapper */
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_3__contains__(PyObject *__pyx_v_self, PyObject *__pyx_v_i); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_2__contains__[] = "Tests for point inclusion, `start <= i <= end`\n\n        Parameters\n        ----------\n        i : Number\n            The point to be tested\n\n        Returns\n        -------\n        bool\n        ";
#if CYTHON_COMPILING_IN_CPYTHON
struct wrapperbase __pyx_wrapperbase_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_2__contains__;
#endif
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_3__contains__(PyObject *__pyx_v_self, PyObject *__pyx_v_i) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__contains__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_2__contains__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *)__pyx_v_self), ((PyObject *)__pyx_v_i));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_2__contains__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, PyObject *__pyx_v_i) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  __Pyx_RefNannySetupContext("__contains__", 0);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":30
 *         bool
 *         """
 *         return self._contains(i)             # <<<<<<<<<<<<<<
 * 
 *     cdef bint _contains(self, double i):
 */
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_i); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 30, __pyx_L1_error)
  __pyx_r = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *)__pyx_v_self->__pyx_vtab)->_contains(__pyx_v_self, __pyx_t_1);
  goto __pyx_L0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":18
 *         self.end = end
 * 
 *     def __contains__(self, i):             # <<<<<<<<<<<<<<
 *         """Tests for point inclusion, `start <= i <= end`
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin.__contains__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":32
 *         return self._contains(i)
 * 
 *     cdef bint _contains(self, double i):             # <<<<<<<<<<<<<<
 *         """Tests for point inclusion, `start <= i <= end`
 * 
 */

static int __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin__contains(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, double __pyx_v_i) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("_contains", 0);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":44
 *         bool
 *         """
 *         return self.start <= i <= self.end             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint contains(self, double i):
 */
  __pyx_t_1 = (__pyx_v_self->start <= __pyx_v_i);
  if (__pyx_t_1) {
    __pyx_t_1 = (__pyx_v_i <= __pyx_v_self->end);
  }
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":32
 *         return self._contains(i)
 * 
 *     cdef bint _contains(self, double i):             # <<<<<<<<<<<<<<
 *         """Tests for point inclusion, `start <= i <= end`
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":46
 *         return self.start <= i <= self.end
 * 
 *     cpdef bint contains(self, double i):             # <<<<<<<<<<<<<<
 *         """Tests for point inclusion, `start <= i <= end`
 * 
 */

static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_5contains(PyObject *__pyx_v_self, PyObject *__pyx_arg_i); /*proto*/
static int __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_contains(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, double __pyx_v_i, int __pyx_skip_dispatch) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("contains", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_contains); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 46, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_5contains)) {
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_i); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 46, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 46, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 46, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_6;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":58
 *         bool
 *         """
 *         return self._contains(i)             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint overlaps(self, SpanningMixin interval):
 */
  __pyx_r = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *)__pyx_v_self->__pyx_vtab)->_contains(__pyx_v_self, __pyx_v_i);
  goto __pyx_L0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":46
 *         return self.start <= i <= self.end
 * 
 *     cpdef bint contains(self, double i):             # <<<<<<<<<<<<<<
 *         """Tests for point inclusion, `start <= i <= end`
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_WriteUnraisable("ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin.contains", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_5contains(PyObject *__pyx_v_self, PyObject *__pyx_arg_i); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_4contains[] = "SpanningMixin.contains(self, double i) -> bool\nTests for point inclusion, `start <= i <= end`\n\n        Parameters\n        ----------\n        i : double\n            The point to be tested\n\n        Returns\n        -------\n        bool\n        ";
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_5contains(PyObject *__pyx_v_self, PyObject *__pyx_arg_i) {
  double __pyx_v_i;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("contains (wrapper)", 0);
  assert(__pyx_arg_i); {
    __pyx_v_i = __pyx_PyFloat_AsDouble(__pyx_arg_i); if (unlikely((__pyx_v_i == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 46, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin.contains", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_4contains(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *)__pyx_v_self), ((double)__pyx_v_i));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_4contains(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, double __pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("contains", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_contains(__pyx_v_self, __pyx_v_i, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin.contains", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":60
 *         return self._contains(i)
 * 
 *     cpdef bint overlaps(self, SpanningMixin interval):             # <<<<<<<<<<<<<<
 *         """Tests whether another spanning entity with
 *         a defined start and end point overlaps with
 */

static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_7overlaps(PyObject *__pyx_v_self, PyObject *__pyx_v_interval); /*proto*/
static int __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_overlaps(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_interval, int __pyx_skip_dispatch) {
  int __pyx_v_cond;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("overlaps", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_overlaps); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 60, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_7overlaps)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, ((PyObject *)__pyx_v_interval)) : __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_interval));
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 60, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 60, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_5;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":74
 *         """
 *         cdef bint cond
 *         cond = ((self.start <= interval.start and self.end >= interval.start) or (             # <<<<<<<<<<<<<<
 *             self.start >= interval.start and self.end <= interval.end) or (
 *             self.start >= interval.start and self.end >= interval.end and self.start <= interval.end) or (
 */
  __pyx_t_6 = ((__pyx_v_self->start <= __pyx_v_interval->start) != 0);
  if (!__pyx_t_6) {
    goto __pyx_L4_next_or;
  } else {
  }
  __pyx_t_6 = ((__pyx_v_self->end >= __pyx_v_interval->start) != 0);
  if (!__pyx_t_6) {
  } else {
    __pyx_t_5 = __pyx_t_6;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_L4_next_or:;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":75
 *         cdef bint cond
 *         cond = ((self.start <= interval.start and self.end >= interval.start) or (
 *             self.start >= interval.start and self.end <= interval.end) or (             # <<<<<<<<<<<<<<
 *             self.start >= interval.start and self.end >= interval.end and self.start <= interval.end) or (
 *             self.start <= interval.start and self.end >= interval.start) or (
 */
  __pyx_t_6 = ((__pyx_v_self->start >= __pyx_v_interval->start) != 0);
  if (!__pyx_t_6) {
    goto __pyx_L6_next_or;
  } else {
  }
  __pyx_t_6 = ((__pyx_v_self->end <= __pyx_v_interval->end) != 0);
  if (!__pyx_t_6) {
  } else {
    __pyx_t_5 = __pyx_t_6;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_L6_next_or:;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":76
 *         cond = ((self.start <= interval.start and self.end >= interval.start) or (
 *             self.start >= interval.start and self.end <= interval.end) or (
 *             self.start >= interval.start and self.end >= interval.end and self.start <= interval.end) or (             # <<<<<<<<<<<<<<
 *             self.start <= interval.start and self.end >= interval.start) or (
 *             self.start <= interval.end and self.end >= interval.end))
 */
  __pyx_t_6 = ((__pyx_v_self->start >= __pyx_v_interval->start) != 0);
  if (!__pyx_t_6) {
    goto __pyx_L8_next_or;
  } else {
  }
  __pyx_t_6 = ((__pyx_v_self->end >= __pyx_v_interval->end) != 0);
  if (!__pyx_t_6) {
    goto __pyx_L8_next_or;
  } else {
  }
  __pyx_t_6 = ((__pyx_v_self->start <= __pyx_v_interval->end) != 0);
  if (!__pyx_t_6) {
  } else {
    __pyx_t_5 = __pyx_t_6;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_L8_next_or:;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":77
 *             self.start >= interval.start and self.end <= interval.end) or (
 *             self.start >= interval.start and self.end >= interval.end and self.start <= interval.end) or (
 *             self.start <= interval.start and self.end >= interval.start) or (             # <<<<<<<<<<<<<<
 *             self.start <= interval.end and self.end >= interval.end))
 *         return cond
 */
  __pyx_t_6 = ((__pyx_v_self->start <= __pyx_v_interval->start) != 0);
  if (!__pyx_t_6) {
    goto __pyx_L11_next_or;
  } else {
  }
  __pyx_t_6 = ((__pyx_v_self->end >= __pyx_v_interval->start) != 0);
  if (!__pyx_t_6) {
  } else {
    __pyx_t_5 = __pyx_t_6;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_L11_next_or:;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":78
 *             self.start >= interval.start and self.end >= interval.end and self.start <= interval.end) or (
 *             self.start <= interval.start and self.end >= interval.start) or (
 *             self.start <= interval.end and self.end >= interval.end))             # <<<<<<<<<<<<<<
 *         return cond
 * 
 */
  __pyx_t_6 = ((__pyx_v_self->start <= __pyx_v_interval->end) != 0);
  if (__pyx_t_6) {
  } else {
    __pyx_t_5 = __pyx_t_6;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_6 = ((__pyx_v_self->end >= __pyx_v_interval->end) != 0);
  __pyx_t_5 = __pyx_t_6;
  __pyx_L3_bool_binop_done:;
  __pyx_v_cond = __pyx_t_5;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":79
 *             self.start <= interval.start and self.end >= interval.start) or (
 *             self.start <= interval.end and self.end >= interval.end))
 *         return cond             # <<<<<<<<<<<<<<
 * 
 *     cpdef double overlap_size(self, SpanningMixin interval):
 */
  __pyx_r = __pyx_v_cond;
  goto __pyx_L0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":60
 *         return self._contains(i)
 * 
 *     cpdef bint overlaps(self, SpanningMixin interval):             # <<<<<<<<<<<<<<
 *         """Tests whether another spanning entity with
 *         a defined start and end point overlaps with
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin.overlaps", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_7overlaps(PyObject *__pyx_v_self, PyObject *__pyx_v_interval); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_6overlaps[] = "SpanningMixin.overlaps(self, SpanningMixin interval) -> bool\nTests whether another spanning entity with\n        a defined start and end point overlaps with\n        this spanning entity\n\n        Parameters\n        ----------\n        interval : SpanningMixin\n\n        Returns\n        -------\n        bool\n        ";
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_7overlaps(PyObject *__pyx_v_self, PyObject *__pyx_v_interval) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("overlaps (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_interval), __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin, 1, "interval", 0))) __PYX_ERR(0, 60, __pyx_L1_error)
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_6overlaps(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *)__pyx_v_self), ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *)__pyx_v_interval));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_6overlaps(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_interval) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("overlaps", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_overlaps(__pyx_v_self, __pyx_v_interval, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 60, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin.overlaps", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":81
 *         return cond
 * 
 *     cpdef double overlap_size(self, SpanningMixin interval):             # <<<<<<<<<<<<<<
 *         if self.start <= interval.start <= self.end and self.start <= interval.end <= self.end:
 *             return interval.end - interval.start
 */

static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_9overlap_size(PyObject *__pyx_v_self, PyObject *__pyx_v_interval); /*proto*/
static double __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_overlap_size(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_interval, int __pyx_skip_dispatch) {
  double __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  double __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("overlap_size", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_overlap_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 81, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_9overlap_size)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, ((PyObject *)__pyx_v_interval)) : __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_interval));
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 81, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_5 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_5 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 81, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_5;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":82
 * 
 *     cpdef double overlap_size(self, SpanningMixin interval):
 *         if self.start <= interval.start <= self.end and self.start <= interval.end <= self.end:             # <<<<<<<<<<<<<<
 *             return interval.end - interval.start
 *         elif interval.start <= self.start <= interval.end and interval.start <= self.end <= interval.end:
 */
  __pyx_t_7 = (__pyx_v_self->start <= __pyx_v_interval->start);
  if (__pyx_t_7) {
    __pyx_t_7 = (__pyx_v_interval->start <= __pyx_v_self->end);
  }
  __pyx_t_8 = (__pyx_t_7 != 0);
  if (__pyx_t_8) {
  } else {
    __pyx_t_6 = __pyx_t_8;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_8 = (__pyx_v_self->start <= __pyx_v_interval->end);
  if (__pyx_t_8) {
    __pyx_t_8 = (__pyx_v_interval->end <= __pyx_v_self->end);
  }
  __pyx_t_7 = (__pyx_t_8 != 0);
  __pyx_t_6 = __pyx_t_7;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_6) {

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":83
 *     cpdef double overlap_size(self, SpanningMixin interval):
 *         if self.start <= interval.start <= self.end and self.start <= interval.end <= self.end:
 *             return interval.end - interval.start             # <<<<<<<<<<<<<<
 *         elif interval.start <= self.start <= interval.end and interval.start <= self.end <= interval.end:
 *             return self.end - self.start
 */
    __pyx_r = (__pyx_v_interval->end - __pyx_v_interval->start);
    goto __pyx_L0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":82
 * 
 *     cpdef double overlap_size(self, SpanningMixin interval):
 *         if self.start <= interval.start <= self.end and self.start <= interval.end <= self.end:             # <<<<<<<<<<<<<<
 *             return interval.end - interval.start
 *         elif interval.start <= self.start <= interval.end and interval.start <= self.end <= interval.end:
 */
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":84
 *         if self.start <= interval.start <= self.end and self.start <= interval.end <= self.end:
 *             return interval.end - interval.start
 *         elif interval.start <= self.start <= interval.end and interval.start <= self.end <= interval.end:             # <<<<<<<<<<<<<<
 *             return self.end - self.start
 *         if self.end >= interval.start:
 */
  __pyx_t_7 = (__pyx_v_interval->start <= __pyx_v_self->start);
  if (__pyx_t_7) {
    __pyx_t_7 = (__pyx_v_self->start <= __pyx_v_interval->end);
  }
  __pyx_t_8 = (__pyx_t_7 != 0);
  if (__pyx_t_8) {
  } else {
    __pyx_t_6 = __pyx_t_8;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_8 = (__pyx_v_interval->start <= __pyx_v_self->end);
  if (__pyx_t_8) {
    __pyx_t_8 = (__pyx_v_self->end <= __pyx_v_interval->end);
  }
  __pyx_t_7 = (__pyx_t_8 != 0);
  __pyx_t_6 = __pyx_t_7;
  __pyx_L6_bool_binop_done:;
  if (__pyx_t_6) {

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":85
 *             return interval.end - interval.start
 *         elif interval.start <= self.start <= interval.end and interval.start <= self.end <= interval.end:
 *             return self.end - self.start             # <<<<<<<<<<<<<<
 *         if self.end >= interval.start:
 *             return self.end - interval.start
 */
    __pyx_r = (__pyx_v_self->end - __pyx_v_self->start);
    goto __pyx_L0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":84
 *         if self.start <= interval.start <= self.end and self.start <= interval.end <= self.end:
 *             return interval.end - interval.start
 *         elif interval.start <= self.start <= interval.end and interval.start <= self.end <= interval.end:             # <<<<<<<<<<<<<<
 *             return self.end - self.start
 *         if self.end >= interval.start:
 */
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":86
 *         elif interval.start <= self.start <= interval.end and interval.start <= self.end <= interval.end:
 *             return self.end - self.start
 *         if self.end >= interval.start:             # <<<<<<<<<<<<<<
 *             return self.end - interval.start
 *         elif self.start >= interval.end:
 */
  __pyx_t_6 = ((__pyx_v_self->end >= __pyx_v_interval->start) != 0);
  if (__pyx_t_6) {

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":87
 *             return self.end - self.start
 *         if self.end >= interval.start:
 *             return self.end - interval.start             # <<<<<<<<<<<<<<
 *         elif self.start >= interval.end:
 *             return self.start - interval.end
 */
    __pyx_r = (__pyx_v_self->end - __pyx_v_interval->start);
    goto __pyx_L0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":86
 *         elif interval.start <= self.start <= interval.end and interval.start <= self.end <= interval.end:
 *             return self.end - self.start
 *         if self.end >= interval.start:             # <<<<<<<<<<<<<<
 *             return self.end - interval.start
 *         elif self.start >= interval.end:
 */
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":88
 *         if self.end >= interval.start:
 *             return self.end - interval.start
 *         elif self.start >= interval.end:             # <<<<<<<<<<<<<<
 *             return self.start - interval.end
 * 
 */
  __pyx_t_6 = ((__pyx_v_self->start >= __pyx_v_interval->end) != 0);
  if (__pyx_t_6) {

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":89
 *             return self.end - interval.start
 *         elif self.start >= interval.end:
 *             return self.start - interval.end             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint is_contained_in_interval(self, SpanningMixin interval):
 */
    __pyx_r = (__pyx_v_self->start - __pyx_v_interval->end);
    goto __pyx_L0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":88
 *         if self.end >= interval.start:
 *             return self.end - interval.start
 *         elif self.start >= interval.end:             # <<<<<<<<<<<<<<
 *             return self.start - interval.end
 * 
 */
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":81
 *         return cond
 * 
 *     cpdef double overlap_size(self, SpanningMixin interval):             # <<<<<<<<<<<<<<
 *         if self.start <= interval.start <= self.end and self.start <= interval.end <= self.end:
 *             return interval.end - interval.start
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin.overlap_size", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_9overlap_size(PyObject *__pyx_v_self, PyObject *__pyx_v_interval); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_8overlap_size[] = "SpanningMixin.overlap_size(self, SpanningMixin interval) -> double";
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_9overlap_size(PyObject *__pyx_v_self, PyObject *__pyx_v_interval) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("overlap_size (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_interval), __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin, 1, "interval", 0))) __PYX_ERR(0, 81, __pyx_L1_error)
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_8overlap_size(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *)__pyx_v_self), ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *)__pyx_v_interval));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_8overlap_size(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_interval) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("overlap_size", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_overlap_size(__pyx_v_self, __pyx_v_interval, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 81, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin.overlap_size", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":91
 *             return self.start - interval.end
 * 
 *     cpdef bint is_contained_in_interval(self, SpanningMixin interval):             # <<<<<<<<<<<<<<
 *         """Tests whether this spanning entity is
 *         completely contained inside the other entity
 */

static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_11is_contained_in_interval(PyObject *__pyx_v_self, PyObject *__pyx_v_interval); /*proto*/
static int __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_is_contained_in_interval(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_interval, int __pyx_skip_dispatch) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("is_contained_in_interval", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_is_contained_in_interval); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 91, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_11is_contained_in_interval)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, ((PyObject *)__pyx_v_interval)) : __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_interval));
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 91, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 91, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_5;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":103
 *         bool
 *         """
 *         return self.start >= interval.start and self.end <= interval.end             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint contains_interval(self, SpanningMixin interval):
 */
  __pyx_t_6 = ((__pyx_v_self->start >= __pyx_v_interval->start) != 0);
  if (__pyx_t_6) {
  } else {
    __pyx_t_5 = __pyx_t_6;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_6 = ((__pyx_v_self->end <= __pyx_v_interval->end) != 0);
  __pyx_t_5 = __pyx_t_6;
  __pyx_L3_bool_binop_done:;
  __pyx_r = __pyx_t_5;
  goto __pyx_L0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":91
 *             return self.start - interval.end
 * 
 *     cpdef bint is_contained_in_interval(self, SpanningMixin interval):             # <<<<<<<<<<<<<<
 *         """Tests whether this spanning entity is
 *         completely contained inside the other entity
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin.is_contained_in_interval", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_11is_contained_in_interval(PyObject *__pyx_v_self, PyObject *__pyx_v_interval); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_10is_contained_in_interval[] = "SpanningMixin.is_contained_in_interval(self, SpanningMixin interval) -> bool\nTests whether this spanning entity is\n        completely contained inside the other entity\n\n        Parameters\n        ----------\n        interval : SpanningMixin\n\n        Returns\n        -------\n        bool\n        ";
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_11is_contained_in_interval(PyObject *__pyx_v_self, PyObject *__pyx_v_interval) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_contained_in_interval (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_interval), __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin, 1, "interval", 0))) __PYX_ERR(0, 91, __pyx_L1_error)
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_10is_contained_in_interval(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *)__pyx_v_self), ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *)__pyx_v_interval));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_10is_contained_in_interval(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_interval) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("is_contained_in_interval", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_is_contained_in_interval(__pyx_v_self, __pyx_v_interval, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 91, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin.is_contained_in_interval", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":105
 *         return self.start >= interval.start and self.end <= interval.end
 * 
 *     cpdef bint contains_interval(self, SpanningMixin interval):             # <<<<<<<<<<<<<<
 *         """Tests whether the other spanning entity is
 *         completely contained inside this entity
 */

static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_13contains_interval(PyObject *__pyx_v_self, PyObject *__pyx_v_interval); /*proto*/
static int __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_contains_interval(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_interval, int __pyx_skip_dispatch) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("contains_interval", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_contains_interval); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 105, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_13contains_interval)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, ((PyObject *)__pyx_v_interval)) : __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_interval));
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 105, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 105, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_5;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":117
 *         bool
 *         """
 *         return self.start <= interval.start and self.end >= interval.end             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_6 = ((__pyx_v_self->start <= __pyx_v_interval->start) != 0);
  if (__pyx_t_6) {
  } else {
    __pyx_t_5 = __pyx_t_6;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_6 = ((__pyx_v_self->end >= __pyx_v_interval->end) != 0);
  __pyx_t_5 = __pyx_t_6;
  __pyx_L3_bool_binop_done:;
  __pyx_r = __pyx_t_5;
  goto __pyx_L0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":105
 *         return self.start >= interval.start and self.end <= interval.end
 * 
 *     cpdef bint contains_interval(self, SpanningMixin interval):             # <<<<<<<<<<<<<<
 *         """Tests whether the other spanning entity is
 *         completely contained inside this entity
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin.contains_interval", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_13contains_interval(PyObject *__pyx_v_self, PyObject *__pyx_v_interval); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_12contains_interval[] = "SpanningMixin.contains_interval(self, SpanningMixin interval) -> bool\nTests whether the other spanning entity is\n        completely contained inside this entity\n\n        Parameters\n        ----------\n        interval : SpanningMixin\n\n        Returns\n        -------\n        bool\n        ";
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_13contains_interval(PyObject *__pyx_v_self, PyObject *__pyx_v_interval) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("contains_interval (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_interval), __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin, 1, "interval", 0))) __PYX_ERR(0, 105, __pyx_L1_error)
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_12contains_interval(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *)__pyx_v_self), ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *)__pyx_v_interval));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_12contains_interval(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_interval) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("contains_interval", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_contains_interval(__pyx_v_self, __pyx_v_interval, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin.contains_interval", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pxd":3
 * cdef class SpanningMixin(object):
 *     cdef:
 *         public double start             # <<<<<<<<<<<<<<
 *         public double end
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_5start_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_5start_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_5start___get__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_5start___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->start); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin.start.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_5start_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_5start_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_5start_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_5start_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 3, __pyx_L1_error)
  __pyx_v_self->start = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin.start.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pxd":4
 *     cdef:
 *         public double start
 *         public double end             # <<<<<<<<<<<<<<
 * 
 *     cdef bint _contains(self, double i)
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_3end_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_3end_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_3end___get__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_3end___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->end); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin.end.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_3end_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_3end_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_3end_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_3end_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 4, __pyx_L1_error)
  __pyx_v_self->end = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin.end.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_15__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_14__reduce_cython__[] = "SpanningMixin.__reduce_cython__(self)";
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_15__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_14__reduce_cython__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_14__reduce_cython__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.end, self.start)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->end); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->start); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.end, self.start)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_3 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v__dict = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":7
 *     state = (self.end, self.start)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_4 = (__pyx_v__dict != Py_None);
  __pyx_t_5 = (__pyx_t_4 != 0);
  if (__pyx_t_5) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v__dict);
    __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.end, self.start)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_SpanningMixin, (type(self), 0xccc3210, None), state
 */
  /*else*/ {
    __pyx_v_use_setstate = 0;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_SpanningMixin, (type(self), 0xccc3210, None), state
 *     else:
 */
  __pyx_t_5 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_5) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle_SpanningMixin, (type(self), 0xccc3210, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_SpanningMixin, (type(self), 0xccc3210, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pyx_unpickle_SpanningMixin); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_214708752);
    __Pyx_GIVEREF(__pyx_int_214708752);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_214708752);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_3, 2, Py_None);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);
    __pyx_t_2 = 0;
    __pyx_t_3 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_SpanningMixin, (type(self), 0xccc3210, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_SpanningMixin, (type(self), 0xccc3210, None), state
 *     else:
 *         return __pyx_unpickle_SpanningMixin, (type(self), 0xccc3210, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_SpanningMixin__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pyx_unpickle_SpanningMixin); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_214708752);
    __Pyx_GIVEREF(__pyx_int_214708752);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_214708752);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_state);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3);
    __pyx_t_1 = 0;
    __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_SpanningMixin, (type(self), 0xccc3210, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_SpanningMixin__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_17__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_16__setstate_cython__[] = "SpanningMixin.__setstate_cython__(self, __pyx_state)";
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_17__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_16__setstate_cython__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_16__setstate_cython__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_SpanningMixin, (type(self), 0xccc3210, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_SpanningMixin__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals___pyx_unpickle_SpanningMixin__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_SpanningMixin, (type(self), 0xccc3210, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_SpanningMixin__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":121
 * 
 * cdef class Interval(SpanningMixin):
 *     def __init__(self, start, end, members=None, **kwargs):             # <<<<<<<<<<<<<<
 *         if members is None:
 *             members = []
 */

/* Python wrapper */
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_start = 0;
  PyObject *__pyx_v_end = 0;
  PyObject *__pyx_v_members = 0;
  PyObject *__pyx_v_kwargs = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  __pyx_v_kwargs = PyDict_New(); if (unlikely(!__pyx_v_kwargs)) return -1;
  __Pyx_GOTREF(__pyx_v_kwargs);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_start,&__pyx_n_s_end,&__pyx_n_s_members,0};
    PyObject* values[3] = {0,0,0};
    values[2] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_start)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_end)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 3, 1); __PYX_ERR(0, 121, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_members);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, __pyx_v_kwargs, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 121, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_start = values[0];
    __pyx_v_end = values[1];
    __pyx_v_members = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 121, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.Interval.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval___init__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *)__pyx_v_self), __pyx_v_start, __pyx_v_end, __pyx_v_members, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval___init__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *__pyx_v_self, PyObject *__pyx_v_start, PyObject *__pyx_v_end, PyObject *__pyx_v_members, PyObject *__pyx_v_kwargs) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  double __pyx_t_4;
  __Pyx_RefNannySetupContext("__init__", 0);
  __Pyx_INCREF(__pyx_v_members);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":122
 * cdef class Interval(SpanningMixin):
 *     def __init__(self, start, end, members=None, **kwargs):
 *         if members is None:             # <<<<<<<<<<<<<<
 *             members = []
 *         self.start = start
 */
  __pyx_t_1 = (__pyx_v_members == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":123
 *     def __init__(self, start, end, members=None, **kwargs):
 *         if members is None:
 *             members = []             # <<<<<<<<<<<<<<
 *         self.start = start
 *         self.end = end
 */
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF_SET(__pyx_v_members, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":122
 * cdef class Interval(SpanningMixin):
 *     def __init__(self, start, end, members=None, **kwargs):
 *         if members is None:             # <<<<<<<<<<<<<<
 *             members = []
 *         self.start = start
 */
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":124
 *         if members is None:
 *             members = []
 *         self.start = start             # <<<<<<<<<<<<<<
 *         self.end = end
 *         self.members = members
 */
  __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_v_start); if (unlikely((__pyx_t_4 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 124, __pyx_L1_error)
  __pyx_v_self->__pyx_base.start = __pyx_t_4;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":125
 *             members = []
 *         self.start = start
 *         self.end = end             # <<<<<<<<<<<<<<
 *         self.members = members
 *         self.data = kwargs
 */
  __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_v_end); if (unlikely((__pyx_t_4 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 125, __pyx_L1_error)
  __pyx_v_self->__pyx_base.end = __pyx_t_4;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":126
 *         self.start = start
 *         self.end = end
 *         self.members = members             # <<<<<<<<<<<<<<
 *         self.data = kwargs
 * 
 */
  __Pyx_INCREF(__pyx_v_members);
  __Pyx_GIVEREF(__pyx_v_members);
  __Pyx_GOTREF(__pyx_v_self->members);
  __Pyx_DECREF(__pyx_v_self->members);
  __pyx_v_self->members = __pyx_v_members;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":127
 *         self.end = end
 *         self.members = members
 *         self.data = kwargs             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __Pyx_INCREF(__pyx_v_kwargs);
  __Pyx_GIVEREF(__pyx_v_kwargs);
  __Pyx_GOTREF(__pyx_v_self->data);
  __Pyx_DECREF(__pyx_v_self->data);
  __pyx_v_self->data = __pyx_v_kwargs;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":121
 * 
 * cdef class Interval(SpanningMixin):
 *     def __init__(self, start, end, members=None, **kwargs):             # <<<<<<<<<<<<<<
 *         if members is None:
 *             members = []
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.Interval.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_members);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":129
 *         self.data = kwargs
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "%s(start=%r, end=%r, data=%r)" % (self.__class__.__name__, self.start, self.end, self.data)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_3__repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_2__repr__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_2__repr__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":130
 * 
 *     def __repr__(self):
 *         return "%s(start=%r, end=%r, data=%r)" % (self.__class__.__name__, self.start, self.end, self.data)             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, k):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->__pyx_base.start); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_self->__pyx_base.end); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_3);
  __Pyx_INCREF(__pyx_v_self->data);
  __Pyx_GIVEREF(__pyx_v_self->data);
  PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_v_self->data);
  __pyx_t_2 = 0;
  __pyx_t_1 = 0;
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_s_start_r_end_r_data_r, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":129
 *         self.data = kwargs
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "%s(start=%r, end=%r, data=%r)" % (self.__class__.__name__, self.start, self.end, self.data)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.Interval.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":132
 *         return "%s(start=%r, end=%r, data=%r)" % (self.__class__.__name__, self.start, self.end, self.data)
 * 
 *     def __getitem__(self, k):             # <<<<<<<<<<<<<<
 *         return self.members[k]
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_5__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_k); /*proto*/
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_5__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_k) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_4__getitem__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *)__pyx_v_self), ((PyObject *)__pyx_v_k));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_4__getitem__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *__pyx_v_self, PyObject *__pyx_v_k) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":133
 * 
 *     def __getitem__(self, k):
 *         return self.members[k]             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_self->members, __pyx_v_k); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":132
 *         return "%s(start=%r, end=%r, data=%r)" % (self.__class__.__name__, self.start, self.end, self.data)
 * 
 *     def __getitem__(self, k):             # <<<<<<<<<<<<<<
 *         return self.members[k]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.Interval.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":135
 *         return self.members[k]
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return len(self.members)
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_7__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_7__len__(PyObject *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_6__len__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_6__len__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":136
 * 
 *     def __len__(self):
 *         return len(self.members)             # <<<<<<<<<<<<<<
 * 
 *     def __iter__(self):
 */
  __pyx_t_1 = __pyx_v_self->members;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 136, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  goto __pyx_L0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":135
 *         return self.members[k]
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return len(self.members)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.Interval.__len__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":138
 *         return len(self.members)
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         return iter(self.members)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_9__iter__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_9__iter__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_8__iter__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_8__iter__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__iter__", 0);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":139
 * 
 *     def __iter__(self):
 *         return iter(self.members)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_v_self->members;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":138
 *         return len(self.members)
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         return iter(self.members)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.Interval.__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pxd":21
 * cdef class Interval(SpanningMixin):
 *     cdef:
 *         public object members             # <<<<<<<<<<<<<<
 *         public dict data
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_7members_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_7members_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_7members___get__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_7members___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->members);
  __pyx_r = __pyx_v_self->members;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_7members_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_7members_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_7members_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_7members_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->members);
  __Pyx_DECREF(__pyx_v_self->members);
  __pyx_v_self->members = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_7members_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_7members_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_7members_4__del__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_7members_4__del__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->members);
  __Pyx_DECREF(__pyx_v_self->members);
  __pyx_v_self->members = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pxd":22
 *     cdef:
 *         public object members
 *         public dict data             # <<<<<<<<<<<<<<
 * 
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_4data_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_4data_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_4data___get__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_4data___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->data);
  __pyx_r = __pyx_v_self->data;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_4data_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_4data_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_4data_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_4data_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(PyDict_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(1, 22, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->data);
  __Pyx_DECREF(__pyx_v_self->data);
  __pyx_v_self->data = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.Interval.data.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_4data_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_4data_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_4data_4__del__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_4data_4__del__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->data);
  __Pyx_DECREF(__pyx_v_self->data);
  __pyx_v_self->data = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_11__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_10__reduce_cython__[] = "Interval.__reduce_cython__(self)";
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_11__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_10__reduce_cython__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_10__reduce_cython__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.data, self.end, self.members, self.start)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->__pyx_base.end); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->__pyx_base.start); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(4); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_self->data);
  __Pyx_GIVEREF(__pyx_v_self->data);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_self->data);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->members);
  __Pyx_GIVEREF(__pyx_v_self->members);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_self->members);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.data, self.end, self.members, self.start)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_3 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v__dict = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":7
 *     state = (self.data, self.end, self.members, self.start)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_4 = (__pyx_v__dict != Py_None);
  __pyx_t_5 = (__pyx_t_4 != 0);
  if (__pyx_t_5) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v__dict);
    __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.data is not None or self.members is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.data, self.end, self.members, self.start)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.data is not None or self.members is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Interval, (type(self), 0xeace574, None), state
 */
  /*else*/ {
    __pyx_t_4 = (__pyx_v_self->data != ((PyObject*)Py_None));
    __pyx_t_6 = (__pyx_t_4 != 0);
    if (!__pyx_t_6) {
    } else {
      __pyx_t_5 = __pyx_t_6;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_6 = (__pyx_v_self->members != Py_None);
    __pyx_t_4 = (__pyx_t_6 != 0);
    __pyx_t_5 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_5;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.data is not None or self.members is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Interval, (type(self), 0xeace574, None), state
 *     else:
 */
  __pyx_t_5 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_5) {

    /* "(tree fragment)":13
 *         use_setstate = self.data is not None or self.members is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Interval, (type(self), 0xeace574, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Interval, (type(self), 0xeace574, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pyx_unpickle_Interval); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_246211956);
    __Pyx_GIVEREF(__pyx_int_246211956);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_246211956);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_3, 2, Py_None);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);
    __pyx_t_2 = 0;
    __pyx_t_3 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.data is not None or self.members is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Interval, (type(self), 0xeace574, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Interval, (type(self), 0xeace574, None), state
 *     else:
 *         return __pyx_unpickle_Interval, (type(self), 0xeace574, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Interval__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pyx_unpickle_Interval); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_246211956);
    __Pyx_GIVEREF(__pyx_int_246211956);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_246211956);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_state);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3);
    __pyx_t_1 = 0;
    __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.Interval.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Interval, (type(self), 0xeace574, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Interval__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_13__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_12__setstate_cython__[] = "Interval.__setstate_cython__(self, __pyx_state)";
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_13__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_12__setstate_cython__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_12__setstate_cython__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Interval, (type(self), 0xeace574, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Interval__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals___pyx_unpickle_Interval__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Interval, (type(self), 0xeace574, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Interval__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.Interval.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":142
 * 
 * 
 * cpdef list intervals_containing_point(list intervals, double x):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i
 */

static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_1intervals_containing_point(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_intervals_containing_point(PyObject *__pyx_v_intervals, double __pyx_v_x, CYTHON_UNUSED int __pyx_skip_dispatch) {
  size_t __pyx_v_i;
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_interval = 0;
  PyObject *__pyx_v_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  size_t __pyx_t_4;
  PyObject *__pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("intervals_containing_point", 0);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":147
 *         SpanningMixin interval
 *         list result
 *     result = []             # <<<<<<<<<<<<<<
 *     for i in range(PyList_Size(intervals)):
 *         interval = <SpanningMixin>PyList_GetItem(intervals, i)
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":148
 *         list result
 *     result = []
 *     for i in range(PyList_Size(intervals)):             # <<<<<<<<<<<<<<
 *         interval = <SpanningMixin>PyList_GetItem(intervals, i)
 *         if interval.start <= x <= interval.end:
 */
  __pyx_t_2 = PyList_Size(__pyx_v_intervals); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1L))) __PYX_ERR(0, 148, __pyx_L1_error)
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":149
 *     result = []
 *     for i in range(PyList_Size(intervals)):
 *         interval = <SpanningMixin>PyList_GetItem(intervals, i)             # <<<<<<<<<<<<<<
 *         if interval.start <= x <= interval.end:
 *             result.append(interval)
 */
    __pyx_t_5 = PyList_GetItem(__pyx_v_intervals, __pyx_v_i); if (unlikely(__pyx_t_5 == ((PyObject *)NULL))) __PYX_ERR(0, 149, __pyx_L1_error)
    __pyx_t_1 = ((PyObject *)__pyx_t_5);
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_interval, ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":150
 *     for i in range(PyList_Size(intervals)):
 *         interval = <SpanningMixin>PyList_GetItem(intervals, i)
 *         if interval.start <= x <= interval.end:             # <<<<<<<<<<<<<<
 *             result.append(interval)
 *     return result
 */
    __pyx_t_6 = (__pyx_v_interval->start <= __pyx_v_x);
    if (__pyx_t_6) {
      __pyx_t_6 = (__pyx_v_x <= __pyx_v_interval->end);
    }
    __pyx_t_7 = (__pyx_t_6 != 0);
    if (__pyx_t_7) {

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":151
 *         interval = <SpanningMixin>PyList_GetItem(intervals, i)
 *         if interval.start <= x <= interval.end:
 *             result.append(interval)             # <<<<<<<<<<<<<<
 *     return result
 * 
 */
      __pyx_t_8 = __Pyx_PyList_Append(__pyx_v_result, ((PyObject *)__pyx_v_interval)); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 151, __pyx_L1_error)

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":150
 *     for i in range(PyList_Size(intervals)):
 *         interval = <SpanningMixin>PyList_GetItem(intervals, i)
 *         if interval.start <= x <= interval.end:             # <<<<<<<<<<<<<<
 *             result.append(interval)
 *     return result
 */
    }
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":152
 *         if interval.start <= x <= interval.end:
 *             result.append(interval)
 *     return result             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":142
 * 
 * 
 * cpdef list intervals_containing_point(list intervals, double x):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.intervals_containing_point", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_interval);
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_1intervals_containing_point(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_intervals_containing_point[] = "intervals_containing_point(list intervals, double x) -> list";
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_1intervals_containing_point(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_intervals = 0;
  double __pyx_v_x;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("intervals_containing_point (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_intervals,&__pyx_n_s_x,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_intervals)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("intervals_containing_point", 1, 2, 2, 1); __PYX_ERR(0, 142, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "intervals_containing_point") < 0)) __PYX_ERR(0, 142, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_intervals = ((PyObject*)values[0]);
    __pyx_v_x = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_x == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 142, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("intervals_containing_point", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 142, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.intervals_containing_point", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_intervals), (&PyList_Type), 1, "intervals", 1))) __PYX_ERR(0, 142, __pyx_L1_error)
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_intervals_containing_point(__pyx_self, __pyx_v_intervals, __pyx_v_x);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_intervals_containing_point(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_intervals, double __pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("intervals_containing_point", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_intervals_containing_point(__pyx_v_intervals, __pyx_v_x, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.intervals_containing_point", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":160
 * cdef class IntervalTreeNode(object):
 * 
 *     def __init__(self, double center, IntervalTreeNode left, list contained, IntervalTreeNode right,             # <<<<<<<<<<<<<<
 *                  int level=0, IntervalTreeNode parent=None):
 *         cdef:
 */

/* Python wrapper */
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_center;
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_left = 0;
  PyObject *__pyx_v_contained = 0;
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_right = 0;
  int __pyx_v_level;
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_parent = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_center,&__pyx_n_s_left,&__pyx_n_s_contained,&__pyx_n_s_right,&__pyx_n_s_level,&__pyx_n_s_parent,0};
    PyObject* values[6] = {0,0,0,0,0,0};

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":161
 * 
 *     def __init__(self, double center, IntervalTreeNode left, list contained, IntervalTreeNode right,
 *                  int level=0, IntervalTreeNode parent=None):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double start, end, i_start, i_end
 */
    values[5] = (PyObject *)((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_center)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_left)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 4, 6, 1); __PYX_ERR(0, 160, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_contained)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 4, 6, 2); __PYX_ERR(0, 160, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_right)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 4, 6, 3); __PYX_ERR(0, 160, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_level);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[5] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 160, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_center = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_center == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 160, __pyx_L3_error)
    __pyx_v_left = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)values[1]);
    __pyx_v_contained = ((PyObject*)values[2]);
    __pyx_v_right = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)values[3]);
    if (values[4]) {
      __pyx_v_level = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_level == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 161, __pyx_L3_error)
    } else {
      __pyx_v_level = ((int)0);
    }
    __pyx_v_parent = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)values[5]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 4, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 160, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_left), __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode, 1, "left", 0))) __PYX_ERR(0, 160, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_contained), (&PyList_Type), 1, "contained", 1))) __PYX_ERR(0, 160, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_right), __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode, 1, "right", 0))) __PYX_ERR(0, 160, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_parent), __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode, 1, "parent", 0))) __PYX_ERR(0, 161, __pyx_L1_error)
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode___init__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self), __pyx_v_center, __pyx_v_left, __pyx_v_contained, __pyx_v_right, __pyx_v_level, __pyx_v_parent);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":160
 * cdef class IntervalTreeNode(object):
 * 
 *     def __init__(self, double center, IntervalTreeNode left, list contained, IntervalTreeNode right,             # <<<<<<<<<<<<<<
 *                  int level=0, IntervalTreeNode parent=None):
 *         cdef:
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode___init__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, double __pyx_v_center, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_left, PyObject *__pyx_v_contained, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_right, int __pyx_v_level, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_parent) {
  double __pyx_v_start;
  double __pyx_v_end;
  double __pyx_v_i_start;
  double __pyx_v_i_end;
  size_t __pyx_v_i;
  PyObject *__pyx_v_interval = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  double __pyx_t_4;
  int __pyx_t_5;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":166
 *             size_t i
 * 
 *         self.center = center             # <<<<<<<<<<<<<<
 *         self.left = left
 *         self.contained = contained
 */
  __pyx_v_self->center = __pyx_v_center;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":167
 * 
 *         self.center = center
 *         self.left = left             # <<<<<<<<<<<<<<
 *         self.contained = contained
 *         self.right = right
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_left));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_left));
  __Pyx_GOTREF(__pyx_v_self->left);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->left));
  __pyx_v_self->left = __pyx_v_left;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":168
 *         self.center = center
 *         self.left = left
 *         self.contained = contained             # <<<<<<<<<<<<<<
 *         self.right = right
 *         self.level = level
 */
  __Pyx_INCREF(__pyx_v_contained);
  __Pyx_GIVEREF(__pyx_v_contained);
  __Pyx_GOTREF(__pyx_v_self->contained);
  __Pyx_DECREF(__pyx_v_self->contained);
  __pyx_v_self->contained = __pyx_v_contained;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":169
 *         self.left = left
 *         self.contained = contained
 *         self.right = right             # <<<<<<<<<<<<<<
 *         self.level = level
 *         self.parent = parent
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_right));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_right));
  __Pyx_GOTREF(__pyx_v_self->right);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->right));
  __pyx_v_self->right = __pyx_v_right;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":170
 *         self.contained = contained
 *         self.right = right
 *         self.level = level             # <<<<<<<<<<<<<<
 *         self.parent = parent
 * 
 */
  __pyx_v_self->level = __pyx_v_level;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":171
 *         self.right = right
 *         self.level = level
 *         self.parent = parent             # <<<<<<<<<<<<<<
 * 
 *         start = INF
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_parent));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_parent));
  __Pyx_GOTREF(__pyx_v_self->parent);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->parent));
  __pyx_v_self->parent = __pyx_v_parent;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":173
 *         self.parent = parent
 * 
 *         start = INF             # <<<<<<<<<<<<<<
 *         end = -INF
 *         i = 0
 */
  __pyx_v_start = __pyx_v_12ms_deisotope_2_c_23peak_dependency_network_9intervals_INF;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":174
 * 
 *         start = INF
 *         end = -INF             # <<<<<<<<<<<<<<
 *         i = 0
 * 
 */
  __pyx_v_end = (-__pyx_v_12ms_deisotope_2_c_23peak_dependency_network_9intervals_INF);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":175
 *         start = INF
 *         end = -INF
 *         i = 0             # <<<<<<<<<<<<<<
 * 
 *         for interval in self.contained:
 */
  __pyx_v_i = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":177
 *         i = 0
 * 
 *         for interval in self.contained:             # <<<<<<<<<<<<<<
 *             i += 1
 *             i_start = interval.start
 */
  if (unlikely(__pyx_v_self->contained == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 177, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_self->contained; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 177, __pyx_L1_error)
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 177, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_interval, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":178
 * 
 *         for interval in self.contained:
 *             i += 1             # <<<<<<<<<<<<<<
 *             i_start = interval.start
 *             if i_start < start:
 */
    __pyx_v_i = (__pyx_v_i + 1);

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":179
 *         for interval in self.contained:
 *             i += 1
 *             i_start = interval.start             # <<<<<<<<<<<<<<
 *             if i_start < start:
 *                 start = i_start
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_interval, __pyx_n_s_start); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 179, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_4 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 179, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_i_start = __pyx_t_4;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":180
 *             i += 1
 *             i_start = interval.start
 *             if i_start < start:             # <<<<<<<<<<<<<<
 *                 start = i_start
 *             i_end = interval.end
 */
    __pyx_t_5 = ((__pyx_v_i_start < __pyx_v_start) != 0);
    if (__pyx_t_5) {

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":181
 *             i_start = interval.start
 *             if i_start < start:
 *                 start = i_start             # <<<<<<<<<<<<<<
 *             i_end = interval.end
 *             if i_end > end:
 */
      __pyx_v_start = __pyx_v_i_start;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":180
 *             i += 1
 *             i_start = interval.start
 *             if i_start < start:             # <<<<<<<<<<<<<<
 *                 start = i_start
 *             i_end = interval.end
 */
    }

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":182
 *             if i_start < start:
 *                 start = i_start
 *             i_end = interval.end             # <<<<<<<<<<<<<<
 *             if i_end > end:
 *                 end = i_end
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_interval, __pyx_n_s_end); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 182, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_4 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 182, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_i_end = __pyx_t_4;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":183
 *                 start = i_start
 *             i_end = interval.end
 *             if i_end > end:             # <<<<<<<<<<<<<<
 *                 end = i_end
 * 
 */
    __pyx_t_5 = ((__pyx_v_i_end > __pyx_v_end) != 0);
    if (__pyx_t_5) {

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":184
 *             i_end = interval.end
 *             if i_end > end:
 *                 end = i_end             # <<<<<<<<<<<<<<
 * 
 *         if i > 0:
 */
      __pyx_v_end = __pyx_v_i_end;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":183
 *                 start = i_start
 *             i_end = interval.end
 *             if i_end > end:             # <<<<<<<<<<<<<<
 *                 end = i_end
 * 
 */
    }

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":177
 *         i = 0
 * 
 *         for interval in self.contained:             # <<<<<<<<<<<<<<
 *             i += 1
 *             i_start = interval.start
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":186
 *                 end = i_end
 * 
 *         if i > 0:             # <<<<<<<<<<<<<<
 *             self.start = start
 *             self.end = end
 */
  __pyx_t_5 = ((__pyx_v_i > 0) != 0);
  if (__pyx_t_5) {

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":187
 * 
 *         if i > 0:
 *             self.start = start             # <<<<<<<<<<<<<<
 *             self.end = end
 *         else:
 */
    __pyx_v_self->start = __pyx_v_start;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":188
 *         if i > 0:
 *             self.start = start
 *             self.end = end             # <<<<<<<<<<<<<<
 *         else:
 *             self.start = self.end = center
 */
    __pyx_v_self->end = __pyx_v_end;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":186
 *                 end = i_end
 * 
 *         if i > 0:             # <<<<<<<<<<<<<<
 *             self.start = start
 *             self.end = end
 */
    goto __pyx_L7;
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":190
 *             self.end = end
 *         else:
 *             self.start = self.end = center             # <<<<<<<<<<<<<<
 * 
 *     cpdef list contains_point(self, double x):
 */
  /*else*/ {
    __pyx_v_self->start = __pyx_v_center;
    __pyx_v_self->end = __pyx_v_center;
  }
  __pyx_L7:;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":160
 * cdef class IntervalTreeNode(object):
 * 
 *     def __init__(self, double center, IntervalTreeNode left, list contained, IntervalTreeNode right,             # <<<<<<<<<<<<<<
 *                  int level=0, IntervalTreeNode parent=None):
 *         cdef:
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_interval);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":192
 *             self.start = self.end = center
 * 
 *     cpdef list contains_point(self, double x):             # <<<<<<<<<<<<<<
 *         cdef:
 *             list inner
 */

static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_3contains_point(PyObject *__pyx_v_self, PyObject *__pyx_arg_x); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_contains_point(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, double __pyx_v_x, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_inner = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  __Pyx_RefNannySetupContext("contains_point", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_contains_point); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 192, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_3contains_point)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_x); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 192, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 192, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 192, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":195
 *         cdef:
 *             list inner
 *         inner = intervals_containing_point(self.contained, x)             # <<<<<<<<<<<<<<
 *         if self.left is not None and self.left.start <= x <= self.left.end:
 *             inner.extend(self.left.contains_point(x))
 */
  __pyx_t_1 = __pyx_v_self->contained;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_intervals_containing_point(((PyObject*)__pyx_t_1), __pyx_v_x, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_inner = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":196
 *             list inner
 *         inner = intervals_containing_point(self.contained, x)
 *         if self.left is not None and self.left.start <= x <= self.left.end:             # <<<<<<<<<<<<<<
 *             inner.extend(self.left.contains_point(x))
 *             return inner
 */
  __pyx_t_7 = (((PyObject *)__pyx_v_self->left) != Py_None);
  __pyx_t_8 = (__pyx_t_7 != 0);
  if (__pyx_t_8) {
  } else {
    __pyx_t_6 = __pyx_t_8;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_8 = (__pyx_v_self->left->start <= __pyx_v_x);
  if (__pyx_t_8) {
    __pyx_t_8 = (__pyx_v_x <= __pyx_v_self->left->end);
  }
  __pyx_t_7 = (__pyx_t_8 != 0);
  __pyx_t_6 = __pyx_t_7;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_6) {

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":197
 *         inner = intervals_containing_point(self.contained, x)
 *         if self.left is not None and self.left.start <= x <= self.left.end:
 *             inner.extend(self.left.contains_point(x))             # <<<<<<<<<<<<<<
 *             return inner
 *         if self.right is not None and self.right.start <= x <= self.right.end:
 */
    if (unlikely(__pyx_v_inner == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "extend");
      __PYX_ERR(0, 197, __pyx_L1_error)
    }
    __pyx_t_2 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self->left->__pyx_vtab)->contains_point(__pyx_v_self->left, __pyx_v_x, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 197, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_9 = __Pyx_PyList_Extend(__pyx_v_inner, __pyx_t_2); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 197, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":198
 *         if self.left is not None and self.left.start <= x <= self.left.end:
 *             inner.extend(self.left.contains_point(x))
 *             return inner             # <<<<<<<<<<<<<<
 *         if self.right is not None and self.right.start <= x <= self.right.end:
 *             inner.extend(self.right.contains_point(x))
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_inner);
    __pyx_r = __pyx_v_inner;
    goto __pyx_L0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":196
 *             list inner
 *         inner = intervals_containing_point(self.contained, x)
 *         if self.left is not None and self.left.start <= x <= self.left.end:             # <<<<<<<<<<<<<<
 *             inner.extend(self.left.contains_point(x))
 *             return inner
 */
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":199
 *             inner.extend(self.left.contains_point(x))
 *             return inner
 *         if self.right is not None and self.right.start <= x <= self.right.end:             # <<<<<<<<<<<<<<
 *             inner.extend(self.right.contains_point(x))
 *             return inner
 */
  __pyx_t_7 = (((PyObject *)__pyx_v_self->right) != Py_None);
  __pyx_t_8 = (__pyx_t_7 != 0);
  if (__pyx_t_8) {
  } else {
    __pyx_t_6 = __pyx_t_8;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_8 = (__pyx_v_self->right->start <= __pyx_v_x);
  if (__pyx_t_8) {
    __pyx_t_8 = (__pyx_v_x <= __pyx_v_self->right->end);
  }
  __pyx_t_7 = (__pyx_t_8 != 0);
  __pyx_t_6 = __pyx_t_7;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_6) {

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":200
 *             return inner
 *         if self.right is not None and self.right.start <= x <= self.right.end:
 *             inner.extend(self.right.contains_point(x))             # <<<<<<<<<<<<<<
 *             return inner
 *         else:
 */
    if (unlikely(__pyx_v_inner == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "extend");
      __PYX_ERR(0, 200, __pyx_L1_error)
    }
    __pyx_t_2 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self->right->__pyx_vtab)->contains_point(__pyx_v_self->right, __pyx_v_x, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 200, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_9 = __Pyx_PyList_Extend(__pyx_v_inner, __pyx_t_2); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 200, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":201
 *         if self.right is not None and self.right.start <= x <= self.right.end:
 *             inner.extend(self.right.contains_point(x))
 *             return inner             # <<<<<<<<<<<<<<
 *         else:
 *             return inner
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_inner);
    __pyx_r = __pyx_v_inner;
    goto __pyx_L0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":199
 *             inner.extend(self.left.contains_point(x))
 *             return inner
 *         if self.right is not None and self.right.start <= x <= self.right.end:             # <<<<<<<<<<<<<<
 *             inner.extend(self.right.contains_point(x))
 *             return inner
 */
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":203
 *             return inner
 *         else:
 *             return inner             # <<<<<<<<<<<<<<
 * 
 *     cpdef list _overlaps_interval(self, double start, double end):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_inner);
    __pyx_r = __pyx_v_inner;
    goto __pyx_L0;
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":192
 *             self.start = self.end = center
 * 
 *     cpdef list contains_point(self, double x):             # <<<<<<<<<<<<<<
 *         cdef:
 *             list inner
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.contains_point", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_inner);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_3contains_point(PyObject *__pyx_v_self, PyObject *__pyx_arg_x); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_2contains_point[] = "IntervalTreeNode.contains_point(self, double x) -> list";
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_3contains_point(PyObject *__pyx_v_self, PyObject *__pyx_arg_x) {
  double __pyx_v_x;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("contains_point (wrapper)", 0);
  assert(__pyx_arg_x); {
    __pyx_v_x = __pyx_PyFloat_AsDouble(__pyx_arg_x); if (unlikely((__pyx_v_x == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 192, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.contains_point", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_2contains_point(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self), ((double)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_2contains_point(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, double __pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("contains_point", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_contains_point(__pyx_v_self, __pyx_v_x, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.contains_point", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":205
 *             return inner
 * 
 *     cpdef list _overlaps_interval(self, double start, double end):             # <<<<<<<<<<<<<<
 *         cdef:
 *             SpanningMixin interv
 */

static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5_overlaps_interval(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode__overlaps_interval(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, double __pyx_v_start, double __pyx_v_end, int __pyx_skip_dispatch) {
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_interv = 0;
  size_t __pyx_v_i;
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  Py_ssize_t __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  size_t __pyx_t_11;
  PyObject *__pyx_t_12;
  int __pyx_t_13;
  int __pyx_t_14;
  int __pyx_t_15;
  __Pyx_RefNannySetupContext("_overlaps_interval", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_overlaps_interval); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 205, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5_overlaps_interval)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_start); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 205, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_end); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 205, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_5 = __pyx_t_1; __pyx_t_6 = NULL;
        __pyx_t_7 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_6)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_6);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
            __pyx_t_7 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_5)) {
          PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_3, __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 205, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
          PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_3, __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 205, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 205, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          if (__pyx_t_6) {
            __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_t_3);
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_t_4);
          __pyx_t_3 = 0;
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 205, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        }
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 205, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":210
 *             size_t i
 * 
 *         result = []             # <<<<<<<<<<<<<<
 *         for i in range(PyList_Size(self.contained)):
 *             interv = <SpanningMixin>PyList_GetItem(self.contained, i)
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 210, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":211
 * 
 *         result = []
 *         for i in range(PyList_Size(self.contained)):             # <<<<<<<<<<<<<<
 *             interv = <SpanningMixin>PyList_GetItem(self.contained, i)
 *             if ((start < interv.start and interv.end <= end) or
 */
  __pyx_t_1 = __pyx_v_self->contained;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_9 = PyList_Size(__pyx_t_1); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1L))) __PYX_ERR(0, 211, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_10 = __pyx_t_9;
  for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
    __pyx_v_i = __pyx_t_11;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":212
 *         result = []
 *         for i in range(PyList_Size(self.contained)):
 *             interv = <SpanningMixin>PyList_GetItem(self.contained, i)             # <<<<<<<<<<<<<<
 *             if ((start < interv.start and interv.end <= end) or
 *                     (interv.start <= start and interv.end >= end)):
 */
    __pyx_t_1 = __pyx_v_self->contained;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_12 = PyList_GetItem(__pyx_t_1, __pyx_v_i); if (unlikely(__pyx_t_12 == ((PyObject *)NULL))) __PYX_ERR(0, 212, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = ((PyObject *)__pyx_t_12);
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_interv, ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":213
 *         for i in range(PyList_Size(self.contained)):
 *             interv = <SpanningMixin>PyList_GetItem(self.contained, i)
 *             if ((start < interv.start and interv.end <= end) or             # <<<<<<<<<<<<<<
 *                     (interv.start <= start and interv.end >= end)):
 *                 result.append(interv)
 */
    __pyx_t_14 = ((__pyx_v_start < __pyx_v_interv->start) != 0);
    if (!__pyx_t_14) {
      goto __pyx_L7_next_or;
    } else {
    }
    __pyx_t_14 = ((__pyx_v_interv->end <= __pyx_v_end) != 0);
    if (!__pyx_t_14) {
    } else {
      __pyx_t_13 = __pyx_t_14;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_L7_next_or:;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":214
 *             interv = <SpanningMixin>PyList_GetItem(self.contained, i)
 *             if ((start < interv.start and interv.end <= end) or
 *                     (interv.start <= start and interv.end >= end)):             # <<<<<<<<<<<<<<
 *                 result.append(interv)
 *         return result
 */
    __pyx_t_14 = ((__pyx_v_interv->start <= __pyx_v_start) != 0);
    if (__pyx_t_14) {
    } else {
      __pyx_t_13 = __pyx_t_14;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_14 = ((__pyx_v_interv->end >= __pyx_v_end) != 0);
    __pyx_t_13 = __pyx_t_14;
    __pyx_L6_bool_binop_done:;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":213
 *         for i in range(PyList_Size(self.contained)):
 *             interv = <SpanningMixin>PyList_GetItem(self.contained, i)
 *             if ((start < interv.start and interv.end <= end) or             # <<<<<<<<<<<<<<
 *                     (interv.start <= start and interv.end >= end)):
 *                 result.append(interv)
 */
    if (__pyx_t_13) {

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":215
 *             if ((start < interv.start and interv.end <= end) or
 *                     (interv.start <= start and interv.end >= end)):
 *                 result.append(interv)             # <<<<<<<<<<<<<<
 *         return result
 * 
 */
      __pyx_t_15 = __Pyx_PyList_Append(__pyx_v_result, ((PyObject *)__pyx_v_interv)); if (unlikely(__pyx_t_15 == ((int)-1))) __PYX_ERR(0, 215, __pyx_L1_error)

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":213
 *         for i in range(PyList_Size(self.contained)):
 *             interv = <SpanningMixin>PyList_GetItem(self.contained, i)
 *             if ((start < interv.start and interv.end <= end) or             # <<<<<<<<<<<<<<
 *                     (interv.start <= start and interv.end >= end)):
 *                 result.append(interv)
 */
    }
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":216
 *                     (interv.start <= start and interv.end >= end)):
 *                 result.append(interv)
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     cpdef list overlaps(self, double start, double end):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":205
 *             return inner
 * 
 *     cpdef list _overlaps_interval(self, double start, double end):             # <<<<<<<<<<<<<<
 *         cdef:
 *             SpanningMixin interv
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode._overlaps_interval", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_interv);
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5_overlaps_interval(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_4_overlaps_interval[] = "IntervalTreeNode._overlaps_interval(self, double start, double end) -> list";
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5_overlaps_interval(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_start;
  double __pyx_v_end;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_overlaps_interval (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_start,&__pyx_n_s_end,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_start)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_end)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_overlaps_interval", 1, 2, 2, 1); __PYX_ERR(0, 205, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_overlaps_interval") < 0)) __PYX_ERR(0, 205, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_start = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_start == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 205, __pyx_L3_error)
    __pyx_v_end = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_end == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 205, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_overlaps_interval", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 205, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode._overlaps_interval", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_4_overlaps_interval(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self), __pyx_v_start, __pyx_v_end);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_4_overlaps_interval(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, double __pyx_v_start, double __pyx_v_end) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("_overlaps_interval", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode__overlaps_interval(__pyx_v_self, __pyx_v_start, __pyx_v_end, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 205, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode._overlaps_interval", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":218
 *         return result
 * 
 *     cpdef list overlaps(self, double start, double end):             # <<<<<<<<<<<<<<
 *         """Returns the list of all contained intervals which
 *         overlap with the interval described by `start` and `end`
 */

static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_7overlaps(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_overlaps(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, double __pyx_v_start, double __pyx_v_end, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  __Pyx_RefNannySetupContext("overlaps", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_overlaps); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 218, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_7overlaps)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_start); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 218, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_end); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 218, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_5 = __pyx_t_1; __pyx_t_6 = NULL;
        __pyx_t_7 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_6)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_6);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
            __pyx_t_7 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_5)) {
          PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_3, __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 218, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
          PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_3, __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 218, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 218, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          if (__pyx_t_6) {
            __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_t_3);
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_t_4);
          __pyx_t_3 = 0;
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 218, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        }
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 218, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":235
 *             from all spanning nodes' `contained` list.
 *         """
 *         cdef list result = []             # <<<<<<<<<<<<<<
 *         if start > self.end:
 *             return result
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":236
 *         """
 *         cdef list result = []
 *         if start > self.end:             # <<<<<<<<<<<<<<
 *             return result
 *         if end < self.start:
 */
  __pyx_t_9 = ((__pyx_v_start > __pyx_v_self->end) != 0);
  if (__pyx_t_9) {

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":237
 *         cdef list result = []
 *         if start > self.end:
 *             return result             # <<<<<<<<<<<<<<
 *         if end < self.start:
 *             return result
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_result);
    __pyx_r = __pyx_v_result;
    goto __pyx_L0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":236
 *         """
 *         cdef list result = []
 *         if start > self.end:             # <<<<<<<<<<<<<<
 *             return result
 *         if end < self.start:
 */
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":238
 *         if start > self.end:
 *             return result
 *         if end < self.start:             # <<<<<<<<<<<<<<
 *             return result
 *         elif start <= self.start:
 */
  __pyx_t_9 = ((__pyx_v_end < __pyx_v_self->start) != 0);
  if (__pyx_t_9) {

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":239
 *             return result
 *         if end < self.start:
 *             return result             # <<<<<<<<<<<<<<
 *         elif start <= self.start:
 *             if end < self.start:
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_result);
    __pyx_r = __pyx_v_result;
    goto __pyx_L0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":238
 *         if start > self.end:
 *             return result
 *         if end < self.start:             # <<<<<<<<<<<<<<
 *             return result
 *         elif start <= self.start:
 */
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":240
 *         if end < self.start:
 *             return result
 *         elif start <= self.start:             # <<<<<<<<<<<<<<
 *             if end < self.start:
 *                 return result
 */
  __pyx_t_9 = ((__pyx_v_start <= __pyx_v_self->start) != 0);
  if (__pyx_t_9) {

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":241
 *             return result
 *         elif start <= self.start:
 *             if end < self.start:             # <<<<<<<<<<<<<<
 *                 return result
 *             else:
 */
    __pyx_t_9 = ((__pyx_v_end < __pyx_v_self->start) != 0);
    if (__pyx_t_9) {

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":242
 *         elif start <= self.start:
 *             if end < self.start:
 *                 return result             # <<<<<<<<<<<<<<
 *             else:
 *                 if self.left is not None:
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_result);
      __pyx_r = __pyx_v_result;
      goto __pyx_L0;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":241
 *             return result
 *         elif start <= self.start:
 *             if end < self.start:             # <<<<<<<<<<<<<<
 *                 return result
 *             else:
 */
    }

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":244
 *                 return result
 *             else:
 *                 if self.left is not None:             # <<<<<<<<<<<<<<
 *                     result.extend(self.left.overlaps(start, end))
 *                 result.extend(self._overlaps_interval(start, end))
 */
    /*else*/ {
      __pyx_t_9 = (((PyObject *)__pyx_v_self->left) != Py_None);
      __pyx_t_10 = (__pyx_t_9 != 0);
      if (__pyx_t_10) {

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":245
 *             else:
 *                 if self.left is not None:
 *                     result.extend(self.left.overlaps(start, end))             # <<<<<<<<<<<<<<
 *                 result.extend(self._overlaps_interval(start, end))
 *                 if self.right is not None and end >= self.right.start:
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self->left->__pyx_vtab)->overlaps(__pyx_v_self->left, __pyx_v_start, __pyx_v_end, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 245, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = __Pyx_PyList_Extend(__pyx_v_result, __pyx_t_1); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 245, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":244
 *                 return result
 *             else:
 *                 if self.left is not None:             # <<<<<<<<<<<<<<
 *                     result.extend(self.left.overlaps(start, end))
 *                 result.extend(self._overlaps_interval(start, end))
 */
      }

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":246
 *                 if self.left is not None:
 *                     result.extend(self.left.overlaps(start, end))
 *                 result.extend(self._overlaps_interval(start, end))             # <<<<<<<<<<<<<<
 *                 if self.right is not None and end >= self.right.start:
 *                     result.extend(self.right.overlaps(start, end))
 */
      __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self->__pyx_vtab)->_overlaps_interval(__pyx_v_self, __pyx_v_start, __pyx_v_end, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 246, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_11 = __Pyx_PyList_Extend(__pyx_v_result, __pyx_t_1); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 246, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":247
 *                     result.extend(self.left.overlaps(start, end))
 *                 result.extend(self._overlaps_interval(start, end))
 *                 if self.right is not None and end >= self.right.start:             # <<<<<<<<<<<<<<
 *                     result.extend(self.right.overlaps(start, end))
 *         elif start > self.start:
 */
      __pyx_t_9 = (((PyObject *)__pyx_v_self->right) != Py_None);
      __pyx_t_12 = (__pyx_t_9 != 0);
      if (__pyx_t_12) {
      } else {
        __pyx_t_10 = __pyx_t_12;
        goto __pyx_L8_bool_binop_done;
      }
      __pyx_t_12 = ((__pyx_v_end >= __pyx_v_self->right->start) != 0);
      __pyx_t_10 = __pyx_t_12;
      __pyx_L8_bool_binop_done:;
      if (__pyx_t_10) {

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":248
 *                 result.extend(self._overlaps_interval(start, end))
 *                 if self.right is not None and end >= self.right.start:
 *                     result.extend(self.right.overlaps(start, end))             # <<<<<<<<<<<<<<
 *         elif start > self.start:
 *             if self.left is not None and self.left.end >= start:
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self->right->__pyx_vtab)->overlaps(__pyx_v_self->right, __pyx_v_start, __pyx_v_end, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 248, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = __Pyx_PyList_Extend(__pyx_v_result, __pyx_t_1); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 248, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":247
 *                     result.extend(self.left.overlaps(start, end))
 *                 result.extend(self._overlaps_interval(start, end))
 *                 if self.right is not None and end >= self.right.start:             # <<<<<<<<<<<<<<
 *                     result.extend(self.right.overlaps(start, end))
 *         elif start > self.start:
 */
      }
    }

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":240
 *         if end < self.start:
 *             return result
 *         elif start <= self.start:             # <<<<<<<<<<<<<<
 *             if end < self.start:
 *                 return result
 */
    goto __pyx_L4;
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":249
 *                 if self.right is not None and end >= self.right.start:
 *                     result.extend(self.right.overlaps(start, end))
 *         elif start > self.start:             # <<<<<<<<<<<<<<
 *             if self.left is not None and self.left.end >= start:
 *                 result.extend(self.left.overlaps(start, end))
 */
  __pyx_t_10 = ((__pyx_v_start > __pyx_v_self->start) != 0);
  if (__pyx_t_10) {

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":250
 *                     result.extend(self.right.overlaps(start, end))
 *         elif start > self.start:
 *             if self.left is not None and self.left.end >= start:             # <<<<<<<<<<<<<<
 *                 result.extend(self.left.overlaps(start, end))
 *             result.extend(self._overlaps_interval(start, end))
 */
    __pyx_t_12 = (((PyObject *)__pyx_v_self->left) != Py_None);
    __pyx_t_9 = (__pyx_t_12 != 0);
    if (__pyx_t_9) {
    } else {
      __pyx_t_10 = __pyx_t_9;
      goto __pyx_L11_bool_binop_done;
    }
    __pyx_t_9 = ((__pyx_v_self->left->end >= __pyx_v_start) != 0);
    __pyx_t_10 = __pyx_t_9;
    __pyx_L11_bool_binop_done:;
    if (__pyx_t_10) {

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":251
 *         elif start > self.start:
 *             if self.left is not None and self.left.end >= start:
 *                 result.extend(self.left.overlaps(start, end))             # <<<<<<<<<<<<<<
 *             result.extend(self._overlaps_interval(start, end))
 *             if self.right is not None and end >= self.right.start:
 */
      __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self->left->__pyx_vtab)->overlaps(__pyx_v_self->left, __pyx_v_start, __pyx_v_end, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 251, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_11 = __Pyx_PyList_Extend(__pyx_v_result, __pyx_t_1); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 251, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":250
 *                     result.extend(self.right.overlaps(start, end))
 *         elif start > self.start:
 *             if self.left is not None and self.left.end >= start:             # <<<<<<<<<<<<<<
 *                 result.extend(self.left.overlaps(start, end))
 *             result.extend(self._overlaps_interval(start, end))
 */
    }

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":252
 *             if self.left is not None and self.left.end >= start:
 *                 result.extend(self.left.overlaps(start, end))
 *             result.extend(self._overlaps_interval(start, end))             # <<<<<<<<<<<<<<
 *             if self.right is not None and end >= self.right.start:
 *                 result.extend(self.right.overlaps(start, end))
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self->__pyx_vtab)->_overlaps_interval(__pyx_v_self, __pyx_v_start, __pyx_v_end, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 252, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_11 = __Pyx_PyList_Extend(__pyx_v_result, __pyx_t_1); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 252, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":253
 *                 result.extend(self.left.overlaps(start, end))
 *             result.extend(self._overlaps_interval(start, end))
 *             if self.right is not None and end >= self.right.start:             # <<<<<<<<<<<<<<
 *                 result.extend(self.right.overlaps(start, end))
 *         elif end > self.start:
 */
    __pyx_t_9 = (((PyObject *)__pyx_v_self->right) != Py_None);
    __pyx_t_12 = (__pyx_t_9 != 0);
    if (__pyx_t_12) {
    } else {
      __pyx_t_10 = __pyx_t_12;
      goto __pyx_L14_bool_binop_done;
    }
    __pyx_t_12 = ((__pyx_v_end >= __pyx_v_self->right->start) != 0);
    __pyx_t_10 = __pyx_t_12;
    __pyx_L14_bool_binop_done:;
    if (__pyx_t_10) {

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":254
 *             result.extend(self._overlaps_interval(start, end))
 *             if self.right is not None and end >= self.right.start:
 *                 result.extend(self.right.overlaps(start, end))             # <<<<<<<<<<<<<<
 *         elif end > self.start:
 *             if self.left is not None:
 */
      __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self->right->__pyx_vtab)->overlaps(__pyx_v_self->right, __pyx_v_start, __pyx_v_end, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 254, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_11 = __Pyx_PyList_Extend(__pyx_v_result, __pyx_t_1); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 254, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":253
 *                 result.extend(self.left.overlaps(start, end))
 *             result.extend(self._overlaps_interval(start, end))
 *             if self.right is not None and end >= self.right.start:             # <<<<<<<<<<<<<<
 *                 result.extend(self.right.overlaps(start, end))
 *         elif end > self.start:
 */
    }

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":249
 *                 if self.right is not None and end >= self.right.start:
 *                     result.extend(self.right.overlaps(start, end))
 *         elif start > self.start:             # <<<<<<<<<<<<<<
 *             if self.left is not None and self.left.end >= start:
 *                 result.extend(self.left.overlaps(start, end))
 */
    goto __pyx_L4;
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":255
 *             if self.right is not None and end >= self.right.start:
 *                 result.extend(self.right.overlaps(start, end))
 *         elif end > self.start:             # <<<<<<<<<<<<<<
 *             if self.left is not None:
 *                 result.extend(self.left.overlaps(start, end))
 */
  __pyx_t_10 = ((__pyx_v_end > __pyx_v_self->start) != 0);
  if (__pyx_t_10) {

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":256
 *                 result.extend(self.right.overlaps(start, end))
 *         elif end > self.start:
 *             if self.left is not None:             # <<<<<<<<<<<<<<
 *                 result.extend(self.left.overlaps(start, end))
 *             result.extend(self._overlaps_interval(start, end))
 */
    __pyx_t_10 = (((PyObject *)__pyx_v_self->left) != Py_None);
    __pyx_t_12 = (__pyx_t_10 != 0);
    if (__pyx_t_12) {

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":257
 *         elif end > self.start:
 *             if self.left is not None:
 *                 result.extend(self.left.overlaps(start, end))             # <<<<<<<<<<<<<<
 *             result.extend(self._overlaps_interval(start, end))
 *             if self.right is not None and end >= self.right.start:
 */
      __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self->left->__pyx_vtab)->overlaps(__pyx_v_self->left, __pyx_v_start, __pyx_v_end, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 257, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_11 = __Pyx_PyList_Extend(__pyx_v_result, __pyx_t_1); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 257, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":256
 *                 result.extend(self.right.overlaps(start, end))
 *         elif end > self.start:
 *             if self.left is not None:             # <<<<<<<<<<<<<<
 *                 result.extend(self.left.overlaps(start, end))
 *             result.extend(self._overlaps_interval(start, end))
 */
    }

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":258
 *             if self.left is not None:
 *                 result.extend(self.left.overlaps(start, end))
 *             result.extend(self._overlaps_interval(start, end))             # <<<<<<<<<<<<<<
 *             if self.right is not None and end >= self.right.start:
 *                 result.extend(self.right.overlaps(start, end))
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self->__pyx_vtab)->_overlaps_interval(__pyx_v_self, __pyx_v_start, __pyx_v_end, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 258, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_11 = __Pyx_PyList_Extend(__pyx_v_result, __pyx_t_1); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 258, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":259
 *                 result.extend(self.left.overlaps(start, end))
 *             result.extend(self._overlaps_interval(start, end))
 *             if self.right is not None and end >= self.right.start:             # <<<<<<<<<<<<<<
 *                 result.extend(self.right.overlaps(start, end))
 *         return result
 */
    __pyx_t_10 = (((PyObject *)__pyx_v_self->right) != Py_None);
    __pyx_t_9 = (__pyx_t_10 != 0);
    if (__pyx_t_9) {
    } else {
      __pyx_t_12 = __pyx_t_9;
      goto __pyx_L18_bool_binop_done;
    }
    __pyx_t_9 = ((__pyx_v_end >= __pyx_v_self->right->start) != 0);
    __pyx_t_12 = __pyx_t_9;
    __pyx_L18_bool_binop_done:;
    if (__pyx_t_12) {

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":260
 *             result.extend(self._overlaps_interval(start, end))
 *             if self.right is not None and end >= self.right.start:
 *                 result.extend(self.right.overlaps(start, end))             # <<<<<<<<<<<<<<
 *         return result
 * 
 */
      __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self->right->__pyx_vtab)->overlaps(__pyx_v_self->right, __pyx_v_start, __pyx_v_end, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 260, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_11 = __Pyx_PyList_Extend(__pyx_v_result, __pyx_t_1); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 260, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":259
 *                 result.extend(self.left.overlaps(start, end))
 *             result.extend(self._overlaps_interval(start, end))
 *             if self.right is not None and end >= self.right.start:             # <<<<<<<<<<<<<<
 *                 result.extend(self.right.overlaps(start, end))
 *         return result
 */
    }

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":255
 *             if self.right is not None and end >= self.right.start:
 *                 result.extend(self.right.overlaps(start, end))
 *         elif end > self.start:             # <<<<<<<<<<<<<<
 *             if self.left is not None:
 *                 result.extend(self.left.overlaps(start, end))
 */
  }
  __pyx_L4:;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":261
 *             if self.right is not None and end >= self.right.start:
 *                 result.extend(self.right.overlaps(start, end))
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":218
 *         return result
 * 
 *     cpdef list overlaps(self, double start, double end):             # <<<<<<<<<<<<<<
 *         """Returns the list of all contained intervals which
 *         overlap with the interval described by `start` and `end`
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.overlaps", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_7overlaps(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6overlaps[] = "IntervalTreeNode.overlaps(self, double start, double end) -> list\nReturns the list of all contained intervals which\n        overlap with the interval described by `start` and `end`\n\n        Parameters\n        ----------\n        start : float\n            The start of the query interval\n        end : float\n            The end of the query interval\n\n        Returns\n        -------\n        list\n            A list of all objects which overlap the argument interval\n            from all spanning nodes' `contained` list.\n        ";
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_7overlaps(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_start;
  double __pyx_v_end;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("overlaps (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_start,&__pyx_n_s_end,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_start)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_end)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("overlaps", 1, 2, 2, 1); __PYX_ERR(0, 218, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "overlaps") < 0)) __PYX_ERR(0, 218, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_start = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_start == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 218, __pyx_L3_error)
    __pyx_v_end = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_end == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 218, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("overlaps", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 218, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.overlaps", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6overlaps(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self), __pyx_v_start, __pyx_v_end);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6overlaps(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, double __pyx_v_start, double __pyx_v_end) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("overlaps", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_overlaps(__pyx_v_self, __pyx_v_start, __pyx_v_end, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 218, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.overlaps", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":263
 *         return result
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "IntervalTreeNode(level=%d, center=%0.4f, start=%0.4f, end=%0.4f)" % (
 *             self.level, self.center, self.start, self.end)
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_9__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_9__repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_8__repr__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_8__repr__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":264
 * 
 *     def __repr__(self):
 *         return "IntervalTreeNode(level=%d, center=%0.4f, start=%0.4f, end=%0.4f)" % (             # <<<<<<<<<<<<<<
 *             self.level, self.center, self.start, self.end)
 * 
 */
  __Pyx_XDECREF(__pyx_r);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":265
 *     def __repr__(self):
 *         return "IntervalTreeNode(level=%d, center=%0.4f, start=%0.4f, end=%0.4f)" % (
 *             self.level, self.center, self.start, self.end)             # <<<<<<<<<<<<<<
 * 
 *     def __richcmp__(self, other, int code):
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->level); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->center); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_self->start); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyFloat_FromDouble(__pyx_v_self->end); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_4);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":264
 * 
 *     def __repr__(self):
 *         return "IntervalTreeNode(level=%d, center=%0.4f, start=%0.4f, end=%0.4f)" % (             # <<<<<<<<<<<<<<
 *             self.level, self.center, self.start, self.end)
 * 
 */
  __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_IntervalTreeNode_level_d_center, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":263
 *         return result
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "IntervalTreeNode(level=%d, center=%0.4f, start=%0.4f, end=%0.4f)" % (
 *             self.level, self.center, self.start, self.end)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":267
 *             self.level, self.center, self.start, self.end)
 * 
 *     def __richcmp__(self, other, int code):             # <<<<<<<<<<<<<<
 *         if code == 2:
 *             return self._eq(other)
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_11__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_code); /*proto*/
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_11__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_code) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__richcmp__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_10__richcmp__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self), ((PyObject *)__pyx_v_other), ((int)__pyx_v_code));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_10__richcmp__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_code) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__richcmp__", 0);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":268
 * 
 *     def __richcmp__(self, other, int code):
 *         if code == 2:             # <<<<<<<<<<<<<<
 *             return self._eq(other)
 *         elif code == 3:
 */
  switch (__pyx_v_code) {
    case 2:

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":269
 *     def __richcmp__(self, other, int code):
 *         if code == 2:
 *             return self._eq(other)             # <<<<<<<<<<<<<<
 *         elif code == 3:
 *             return not (self._eq(other))
 */
    __Pyx_XDECREF(__pyx_r);
    if (!(likely(((__pyx_v_other) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_other, __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode))))) __PYX_ERR(0, 269, __pyx_L1_error)
    __pyx_t_1 = __Pyx_PyBool_FromLong(((struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self->__pyx_vtab)->_eq(__pyx_v_self, ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_other), 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 269, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":268
 * 
 *     def __richcmp__(self, other, int code):
 *         if code == 2:             # <<<<<<<<<<<<<<
 *             return self._eq(other)
 *         elif code == 3:
 */
    break;
    case 3:

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":271
 *             return self._eq(other)
 *         elif code == 3:
 *             return not (self._eq(other))             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint _eq(self, IntervalTreeNode other):
 */
    __Pyx_XDECREF(__pyx_r);
    if (!(likely(((__pyx_v_other) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_other, __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode))))) __PYX_ERR(0, 271, __pyx_L1_error)
    __pyx_t_1 = __Pyx_PyBool_FromLong((!(((struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self->__pyx_vtab)->_eq(__pyx_v_self, ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_other), 0) != 0))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 271, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":270
 *         if code == 2:
 *             return self._eq(other)
 *         elif code == 3:             # <<<<<<<<<<<<<<
 *             return not (self._eq(other))
 * 
 */
    break;
    default: break;
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":267
 *             self.level, self.center, self.start, self.end)
 * 
 *     def __richcmp__(self, other, int code):             # <<<<<<<<<<<<<<
 *         if code == 2:
 *             return self._eq(other)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.__richcmp__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":273
 *             return not (self._eq(other))
 * 
 *     cpdef bint _eq(self, IntervalTreeNode other):             # <<<<<<<<<<<<<<
 *         if other is None:
 *             return False
 */

static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_13_eq(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static int __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode__eq(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_other, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_result = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("_eq", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_eq); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 273, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_13_eq)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, ((PyObject *)__pyx_v_other)) : __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_other));
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 273, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 273, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_5;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":274
 * 
 *     cpdef bint _eq(self, IntervalTreeNode other):
 *         if other is None:             # <<<<<<<<<<<<<<
 *             return False
 *         else:
 */
  __pyx_t_5 = (((PyObject *)__pyx_v_other) == Py_None);
  __pyx_t_6 = (__pyx_t_5 != 0);
  if (__pyx_t_6) {

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":275
 *     cpdef bint _eq(self, IntervalTreeNode other):
 *         if other is None:
 *             return False             # <<<<<<<<<<<<<<
 *         else:
 *             result = self.contained == other.contained
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":274
 * 
 *     cpdef bint _eq(self, IntervalTreeNode other):
 *         if other is None:             # <<<<<<<<<<<<<<
 *             return False
 *         else:
 */
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":277
 *             return False
 *         else:
 *             result = self.contained == other.contained             # <<<<<<<<<<<<<<
 *             if result:
 *                 result = self.left == other.left
 */
  /*else*/ {
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_self->contained, __pyx_v_other->contained, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 277, __pyx_L1_error)
    __pyx_v_result = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":278
 *         else:
 *             result = self.contained == other.contained
 *             if result:             # <<<<<<<<<<<<<<
 *                 result = self.left == other.left
 *                 if result:
 */
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_result); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 278, __pyx_L1_error)
    if (__pyx_t_6) {

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":279
 *             result = self.contained == other.contained
 *             if result:
 *                 result = self.left == other.left             # <<<<<<<<<<<<<<
 *                 if result:
 *                     result = self.right == other.right
 */
      __pyx_t_1 = PyObject_RichCompare(((PyObject *)__pyx_v_self->left), ((PyObject *)__pyx_v_other->left), Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 279, __pyx_L1_error)
      __Pyx_DECREF_SET(__pyx_v_result, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":280
 *             if result:
 *                 result = self.left == other.left
 *                 if result:             # <<<<<<<<<<<<<<
 *                     result = self.right == other.right
 *             return result
 */
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_result); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 280, __pyx_L1_error)
      if (__pyx_t_6) {

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":281
 *                 result = self.left == other.left
 *                 if result:
 *                     result = self.right == other.right             # <<<<<<<<<<<<<<
 *             return result
 * 
 */
        __pyx_t_1 = PyObject_RichCompare(((PyObject *)__pyx_v_self->right), ((PyObject *)__pyx_v_other->right), Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 281, __pyx_L1_error)
        __Pyx_DECREF_SET(__pyx_v_result, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":280
 *             if result:
 *                 result = self.left == other.left
 *                 if result:             # <<<<<<<<<<<<<<
 *                     result = self.right == other.right
 *             return result
 */
      }

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":278
 *         else:
 *             result = self.contained == other.contained
 *             if result:             # <<<<<<<<<<<<<<
 *                 result = self.left == other.left
 *                 if result:
 */
    }

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":282
 *                 if result:
 *                     result = self.right == other.right
 *             return result             # <<<<<<<<<<<<<<
 * 
 *     def __hash__(self):
 */
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_result); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 282, __pyx_L1_error)
    __pyx_r = __pyx_t_6;
    goto __pyx_L0;
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":273
 *             return not (self._eq(other))
 * 
 *     cpdef bint _eq(self, IntervalTreeNode other):             # <<<<<<<<<<<<<<
 *         if other is None:
 *             return False
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode._eq", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_13_eq(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_12_eq[] = "IntervalTreeNode._eq(self, IntervalTreeNode other) -> bool";
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_13_eq(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_eq (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode, 1, "other", 0))) __PYX_ERR(0, 273, __pyx_L1_error)
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_12_eq(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self), ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_other));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_12_eq(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("_eq", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode__eq(__pyx_v_self, __pyx_v_other, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode._eq", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":284
 *             return result
 * 
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         return hash((self.start, self.center, self.right, self.level))
 * 
 */

/* Python wrapper */
static Py_hash_t __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_15__hash__(PyObject *__pyx_v_self); /*proto*/
static Py_hash_t __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_15__hash__(PyObject *__pyx_v_self) {
  Py_hash_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__hash__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_14__hash__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_hash_t __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_14__hash__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self) {
  Py_hash_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_hash_t __pyx_t_5;
  __Pyx_RefNannySetupContext("__hash__", 0);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":285
 * 
 *     def __hash__(self):
 *         return hash((self.start, self.center, self.right, self.level))             # <<<<<<<<<<<<<<
 * 
 *     def __diagnostic_eq__(self, other):
 */
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->start); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 285, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->center); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 285, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_self->level); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 285, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 285, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->right));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->right));
  PyTuple_SET_ITEM(__pyx_t_4, 2, ((PyObject *)__pyx_v_self->right));
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_t_3);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_5 = PyObject_Hash(__pyx_t_4); if (unlikely(__pyx_t_5 == ((Py_hash_t)-1))) __PYX_ERR(0, 285, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_5;
  goto __pyx_L0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":284
 *             return result
 * 
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         return hash((self.start, self.center, self.right, self.level))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.__hash__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  if (unlikely(__pyx_r == -1) && !PyErr_Occurred()) __pyx_r = -2;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":287
 *         return hash((self.start, self.center, self.right, self.level))
 * 
 *     def __diagnostic_eq__(self, other):             # <<<<<<<<<<<<<<
 *         if other is None:
 *             return False
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_17__diagnostic_eq__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_16__diagnostic_eq__[] = "IntervalTreeNode.__diagnostic_eq__(self, other)";
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_17__diagnostic_eq__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__diagnostic_eq__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_16__diagnostic_eq__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_16__diagnostic_eq__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  __Pyx_RefNannySetupContext("__diagnostic_eq__", 0);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":288
 * 
 *     def __diagnostic_eq__(self, other):
 *         if other is None:             # <<<<<<<<<<<<<<
 *             return False
 *         else:
 */
  __pyx_t_1 = (__pyx_v_other == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":289
 *     def __diagnostic_eq__(self, other):
 *         if other is None:
 *             return False             # <<<<<<<<<<<<<<
 *         else:
 *             result = self.contained == other.contained
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    goto __pyx_L0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":288
 * 
 *     def __diagnostic_eq__(self, other):
 *         if other is None:             # <<<<<<<<<<<<<<
 *             return False
 *         else:
 */
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":291
 *             return False
 *         else:
 *             result = self.contained == other.contained             # <<<<<<<<<<<<<<
 *             if result:
 *                 try:
 */
  /*else*/ {
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_contained); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 291, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyObject_RichCompare(__pyx_v_self->contained, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 291, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_result = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":292
 *         else:
 *             result = self.contained == other.contained
 *             if result:             # <<<<<<<<<<<<<<
 *                 try:
 *                     result = self.left.__diagnostic_eq__(other.left)
 */
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_result); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 292, __pyx_L1_error)
    if (__pyx_t_2) {

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":293
 *             result = self.contained == other.contained
 *             if result:
 *                 try:             # <<<<<<<<<<<<<<
 *                     result = self.left.__diagnostic_eq__(other.left)
 *                 except AttributeError:
 */
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_5);
        __Pyx_XGOTREF(__pyx_t_6);
        __Pyx_XGOTREF(__pyx_t_7);
        /*try:*/ {

          /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":294
 *             if result:
 *                 try:
 *                     result = self.left.__diagnostic_eq__(other.left)             # <<<<<<<<<<<<<<
 *                 except AttributeError:
 *                     result = self.left == other.left
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->left), __pyx_n_s_diagnostic_eq); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 294, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_left); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 294, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_8);
          __pyx_t_9 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_9)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_9);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
            }
          }
          __pyx_t_4 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_9, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_8);
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 294, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF_SET(__pyx_v_result, __pyx_t_4);
          __pyx_t_4 = 0;

          /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":293
 *             result = self.contained == other.contained
 *             if result:
 *                 try:             # <<<<<<<<<<<<<<
 *                     result = self.left.__diagnostic_eq__(other.left)
 *                 except AttributeError:
 */
        }
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        goto __pyx_L10_try_end;
        __pyx_L5_error:;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":295
 *                 try:
 *                     result = self.left.__diagnostic_eq__(other.left)
 *                 except AttributeError:             # <<<<<<<<<<<<<<
 *                     result = self.left == other.left
 *                 if result:
 */
        __pyx_t_10 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_AttributeError);
        if (__pyx_t_10) {
          __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.__diagnostic_eq__", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_3, &__pyx_t_8) < 0) __PYX_ERR(0, 295, __pyx_L7_except_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_GOTREF(__pyx_t_8);

          /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":296
 *                     result = self.left.__diagnostic_eq__(other.left)
 *                 except AttributeError:
 *                     result = self.left == other.left             # <<<<<<<<<<<<<<
 *                 if result:
 *                     try:
 */
          __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_left); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 296, __pyx_L7_except_error)
          __Pyx_GOTREF(__pyx_t_9);
          __pyx_t_11 = PyObject_RichCompare(((PyObject *)__pyx_v_self->left), __pyx_t_9, Py_EQ); __Pyx_XGOTREF(__pyx_t_11); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 296, __pyx_L7_except_error)
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_DECREF_SET(__pyx_v_result, __pyx_t_11);
          __pyx_t_11 = 0;
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          goto __pyx_L6_exception_handled;
        }
        goto __pyx_L7_except_error;
        __pyx_L7_except_error:;

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":293
 *             result = self.contained == other.contained
 *             if result:
 *                 try:             # <<<<<<<<<<<<<<
 *                     result = self.left.__diagnostic_eq__(other.left)
 *                 except AttributeError:
 */
        __Pyx_XGIVEREF(__pyx_t_5);
        __Pyx_XGIVEREF(__pyx_t_6);
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_6, __pyx_t_7);
        goto __pyx_L1_error;
        __pyx_L6_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_5);
        __Pyx_XGIVEREF(__pyx_t_6);
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_6, __pyx_t_7);
        __pyx_L10_try_end:;
      }

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":297
 *                 except AttributeError:
 *                     result = self.left == other.left
 *                 if result:             # <<<<<<<<<<<<<<
 *                     try:
 *                         result = self.right.__diagnostic_eq__(other.right)
 */
      __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_result); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 297, __pyx_L1_error)
      if (__pyx_t_2) {

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":298
 *                     result = self.left == other.left
 *                 if result:
 *                     try:             # <<<<<<<<<<<<<<
 *                         result = self.right.__diagnostic_eq__(other.right)
 *                     except AttributeError:
 */
        {
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __Pyx_ExceptionSave(&__pyx_t_7, &__pyx_t_6, &__pyx_t_5);
          __Pyx_XGOTREF(__pyx_t_7);
          __Pyx_XGOTREF(__pyx_t_6);
          __Pyx_XGOTREF(__pyx_t_5);
          /*try:*/ {

            /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":299
 *                 if result:
 *                     try:
 *                         result = self.right.__diagnostic_eq__(other.right)             # <<<<<<<<<<<<<<
 *                     except AttributeError:
 *                         result = self.right == other.right
 */
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->right), __pyx_n_s_diagnostic_eq); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 299, __pyx_L14_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_right); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 299, __pyx_L14_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_11 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
              __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_3);
              if (likely(__pyx_t_11)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
                __Pyx_INCREF(__pyx_t_11);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_3, function);
              }
            }
            __pyx_t_8 = (__pyx_t_11) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_11, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4);
            __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 299, __pyx_L14_error)
            __Pyx_GOTREF(__pyx_t_8);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_DECREF_SET(__pyx_v_result, __pyx_t_8);
            __pyx_t_8 = 0;

            /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":298
 *                     result = self.left == other.left
 *                 if result:
 *                     try:             # <<<<<<<<<<<<<<
 *                         result = self.right.__diagnostic_eq__(other.right)
 *                     except AttributeError:
 */
          }
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          goto __pyx_L19_try_end;
          __pyx_L14_error:;
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

          /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":300
 *                     try:
 *                         result = self.right.__diagnostic_eq__(other.right)
 *                     except AttributeError:             # <<<<<<<<<<<<<<
 *                         result = self.right == other.right
 *                     if not result:
 */
          __pyx_t_10 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_AttributeError);
          if (__pyx_t_10) {
            __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.__diagnostic_eq__", __pyx_clineno, __pyx_lineno, __pyx_filename);
            if (__Pyx_GetException(&__pyx_t_8, &__pyx_t_3, &__pyx_t_4) < 0) __PYX_ERR(0, 300, __pyx_L16_except_error)
            __Pyx_GOTREF(__pyx_t_8);
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_GOTREF(__pyx_t_4);

            /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":301
 *                         result = self.right.__diagnostic_eq__(other.right)
 *                     except AttributeError:
 *                         result = self.right == other.right             # <<<<<<<<<<<<<<
 *                     if not result:
 *                         print self.right, "r!=", other.right
 */
            __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_right); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 301, __pyx_L16_except_error)
            __Pyx_GOTREF(__pyx_t_11);
            __pyx_t_9 = PyObject_RichCompare(((PyObject *)__pyx_v_self->right), __pyx_t_11, Py_EQ); __Pyx_XGOTREF(__pyx_t_9); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 301, __pyx_L16_except_error)
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_DECREF_SET(__pyx_v_result, __pyx_t_9);
            __pyx_t_9 = 0;
            __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            goto __pyx_L15_exception_handled;
          }
          goto __pyx_L16_except_error;
          __pyx_L16_except_error:;

          /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":298
 *                     result = self.left == other.left
 *                 if result:
 *                     try:             # <<<<<<<<<<<<<<
 *                         result = self.right.__diagnostic_eq__(other.right)
 *                     except AttributeError:
 */
          __Pyx_XGIVEREF(__pyx_t_7);
          __Pyx_XGIVEREF(__pyx_t_6);
          __Pyx_XGIVEREF(__pyx_t_5);
          __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_6, __pyx_t_5);
          goto __pyx_L1_error;
          __pyx_L15_exception_handled:;
          __Pyx_XGIVEREF(__pyx_t_7);
          __Pyx_XGIVEREF(__pyx_t_6);
          __Pyx_XGIVEREF(__pyx_t_5);
          __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_6, __pyx_t_5);
          __pyx_L19_try_end:;
        }

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":302
 *                     except AttributeError:
 *                         result = self.right == other.right
 *                     if not result:             # <<<<<<<<<<<<<<
 *                         print self.right, "r!=", other.right
 *                 else:
 */
        __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_result); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 302, __pyx_L1_error)
        __pyx_t_1 = ((!__pyx_t_2) != 0);
        if (__pyx_t_1) {

          /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":303
 *                         result = self.right == other.right
 *                     if not result:
 *                         print self.right, "r!=", other.right             # <<<<<<<<<<<<<<
 *                 else:
 *                     print self.left, "l!=", other.left
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_right); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 303, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 303, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_INCREF(((PyObject *)__pyx_v_self->right));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_self->right));
          PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_v_self->right));
          __Pyx_INCREF(__pyx_kp_s_r);
          __Pyx_GIVEREF(__pyx_kp_s_r);
          PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_kp_s_r);
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_4);
          __pyx_t_4 = 0;
          if (__Pyx_Print(0, __pyx_t_3, 1) < 0) __PYX_ERR(0, 303, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":302
 *                     except AttributeError:
 *                         result = self.right == other.right
 *                     if not result:             # <<<<<<<<<<<<<<
 *                         print self.right, "r!=", other.right
 *                 else:
 */
        }

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":297
 *                 except AttributeError:
 *                     result = self.left == other.left
 *                 if result:             # <<<<<<<<<<<<<<
 *                     try:
 *                         result = self.right.__diagnostic_eq__(other.right)
 */
        goto __pyx_L13;
      }

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":305
 *                         print self.right, "r!=", other.right
 *                 else:
 *                     print self.left, "l!=", other.left             # <<<<<<<<<<<<<<
 *             else:
 *                 print self, "!=", other
 */
      /*else*/ {
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_left); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 305, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 305, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(((PyObject *)__pyx_v_self->left));
        __Pyx_GIVEREF(((PyObject *)__pyx_v_self->left));
        PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)__pyx_v_self->left));
        __Pyx_INCREF(__pyx_kp_s_l);
        __Pyx_GIVEREF(__pyx_kp_s_l);
        PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_kp_s_l);
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_3);
        __pyx_t_3 = 0;
        if (__Pyx_Print(0, __pyx_t_4, 1) < 0) __PYX_ERR(0, 305, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __pyx_L13:;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":292
 *         else:
 *             result = self.contained == other.contained
 *             if result:             # <<<<<<<<<<<<<<
 *                 try:
 *                     result = self.left.__diagnostic_eq__(other.left)
 */
      goto __pyx_L4;
    }

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":307
 *                     print self.left, "l!=", other.left
 *             else:
 *                 print self, "!=", other             # <<<<<<<<<<<<<<
 *             return result
 * 
 */
    /*else*/ {
      __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 307, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF(((PyObject *)__pyx_v_self));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
      PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)__pyx_v_self));
      __Pyx_INCREF(__pyx_kp_s_);
      __Pyx_GIVEREF(__pyx_kp_s_);
      PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_kp_s_);
      __Pyx_INCREF(__pyx_v_other);
      __Pyx_GIVEREF(__pyx_v_other);
      PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_other);
      if (__Pyx_Print(0, __pyx_t_4, 1) < 0) __PYX_ERR(0, 307, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __pyx_L4:;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":308
 *             else:
 *                 print self, "!=", other
 *             return result             # <<<<<<<<<<<<<<
 * 
 *     @classmethod
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_result);
    __pyx_r = __pyx_v_result;
    goto __pyx_L0;
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":287
 *         return hash((self.start, self.center, self.right, self.level))
 * 
 *     def __diagnostic_eq__(self, other):             # <<<<<<<<<<<<<<
 *         if other is None:
 *             return False
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.__diagnostic_eq__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":311
 * 
 *     @classmethod
 *     def build(cls, intervals):             # <<<<<<<<<<<<<<
 *         cdef:
 *             IntervalTreeNode root, parent, node, up
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_19build(PyObject *__pyx_v_cls, PyObject *__pyx_v_intervals); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_18build[] = "IntervalTreeNode.build(type cls, intervals)";
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_19build(PyObject *__pyx_v_cls, PyObject *__pyx_v_intervals) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("build (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_18build(((PyTypeObject*)__pyx_v_cls), ((PyObject *)__pyx_v_intervals));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_18build(PyTypeObject *__pyx_v_cls, PyObject *__pyx_v_intervals) {
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_root = 0;
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_parent = 0;
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_node = 0;
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_up = 0;
  PyObject *__pyx_v_stack = 0;
  PyObject *__pyx_v_members = 0;
  PyObject *__pyx_v_left = 0;
  PyObject *__pyx_v_right = 0;
  PyObject *__pyx_v_contained = 0;
  PyObject *__pyx_v_intervals_ = 0;
  PyObject *__pyx_v_bunch = 0;
  PyObject *__pyx_v_side = 0;
  double __pyx_v_center;
  double __pyx_v_accumulator;
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_i = 0;
  size_t __pyx_v_j;
  size_t __pyx_v_n;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6;
  Py_ssize_t __pyx_t_7;
  size_t __pyx_t_8;
  size_t __pyx_t_9;
  size_t __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  double __pyx_t_12;
  double __pyx_t_13;
  double __pyx_t_14;
  __Pyx_RefNannySetupContext("build", 0);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":320
 *             SpanningMixin i
 *             size_t j, n
 *         stack = []             # <<<<<<<<<<<<<<
 *         root = cls(0, None, [], None, -1)
 * 
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 320, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_stack = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":321
 *             size_t j, n
 *         stack = []
 *         root = cls(0, None, [], None, -1)             # <<<<<<<<<<<<<<
 * 
 *         intervals_ = list(intervals)
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_int_0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyTuple_SET_ITEM(__pyx_t_2, 1, Py_None);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyTuple_SET_ITEM(__pyx_t_2, 3, Py_None);
  __Pyx_INCREF(__pyx_int_neg_1);
  __Pyx_GIVEREF(__pyx_int_neg_1);
  PyTuple_SET_ITEM(__pyx_t_2, 4, __pyx_int_neg_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_v_cls), __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode))))) __PYX_ERR(0, 321, __pyx_L1_error)
  __pyx_v_root = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":323
 *         root = cls(0, None, [], None, -1)
 * 
 *         intervals_ = list(intervals)             # <<<<<<<<<<<<<<
 *         if not intervals_:
 *             return root
 */
  __pyx_t_1 = PySequence_List(__pyx_v_intervals); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 323, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_intervals_ = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":324
 * 
 *         intervals_ = list(intervals)
 *         if not intervals_:             # <<<<<<<<<<<<<<
 *             return root
 *         stack.append((root, intervals_, "left"))
 */
  __pyx_t_3 = (PyList_GET_SIZE(__pyx_v_intervals_) != 0);
  __pyx_t_4 = ((!__pyx_t_3) != 0);
  if (__pyx_t_4) {

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":325
 *         intervals_ = list(intervals)
 *         if not intervals_:
 *             return root             # <<<<<<<<<<<<<<
 *         stack.append((root, intervals_, "left"))
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject *)__pyx_v_root));
    __pyx_r = ((PyObject *)__pyx_v_root);
    goto __pyx_L0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":324
 * 
 *         intervals_ = list(intervals)
 *         if not intervals_:             # <<<<<<<<<<<<<<
 *             return root
 *         stack.append((root, intervals_, "left"))
 */
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":326
 *         if not intervals_:
 *             return root
 *         stack.append((root, intervals_, "left"))             # <<<<<<<<<<<<<<
 * 
 *         while stack:
 */
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 326, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(((PyObject *)__pyx_v_root));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_root));
  PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_root));
  __Pyx_INCREF(__pyx_v_intervals_);
  __Pyx_GIVEREF(__pyx_v_intervals_);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_intervals_);
  __Pyx_INCREF(__pyx_n_s_left);
  __Pyx_GIVEREF(__pyx_n_s_left);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_n_s_left);
  __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_stack, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 326, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":328
 *         stack.append((root, intervals_, "left"))
 * 
 *         while stack:             # <<<<<<<<<<<<<<
 *             # parent, members, side = stack.pop()
 *             bunch = <tuple>stack.pop()
 */
  while (1) {
    __pyx_t_4 = (PyList_GET_SIZE(__pyx_v_stack) != 0);
    if (!__pyx_t_4) break;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":330
 *         while stack:
 *             # parent, members, side = stack.pop()
 *             bunch = <tuple>stack.pop()             # <<<<<<<<<<<<<<
 *             parent = <IntervalTreeNode>PyTuple_GetItem(bunch, 0)
 *             members = <list>PyTuple_GetItem(bunch, 1)
 */
    __pyx_t_1 = __Pyx_PyList_Pop(__pyx_v_stack); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 330, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_t_1;
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF_SET(__pyx_v_bunch, ((PyObject*)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":331
 *             # parent, members, side = stack.pop()
 *             bunch = <tuple>stack.pop()
 *             parent = <IntervalTreeNode>PyTuple_GetItem(bunch, 0)             # <<<<<<<<<<<<<<
 *             members = <list>PyTuple_GetItem(bunch, 1)
 *             side = <str>PyTuple_GetItem(bunch, 2)
 */
    __pyx_t_6 = PyTuple_GetItem(__pyx_v_bunch, 0); if (unlikely(__pyx_t_6 == ((PyObject *)NULL))) __PYX_ERR(0, 331, __pyx_L1_error)
    __pyx_t_2 = ((PyObject *)__pyx_t_6);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_parent, ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":332
 *             bunch = <tuple>stack.pop()
 *             parent = <IntervalTreeNode>PyTuple_GetItem(bunch, 0)
 *             members = <list>PyTuple_GetItem(bunch, 1)             # <<<<<<<<<<<<<<
 *             side = <str>PyTuple_GetItem(bunch, 2)
 *             n = PyList_Size(members)
 */
    __pyx_t_6 = PyTuple_GetItem(__pyx_v_bunch, 1); if (unlikely(__pyx_t_6 == ((PyObject *)NULL))) __PYX_ERR(0, 332, __pyx_L1_error)
    __pyx_t_2 = ((PyObject *)__pyx_t_6);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_members, ((PyObject*)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":333
 *             parent = <IntervalTreeNode>PyTuple_GetItem(bunch, 0)
 *             members = <list>PyTuple_GetItem(bunch, 1)
 *             side = <str>PyTuple_GetItem(bunch, 2)             # <<<<<<<<<<<<<<
 *             n = PyList_Size(members)
 *             if n > 0:
 */
    __pyx_t_6 = PyTuple_GetItem(__pyx_v_bunch, 2); if (unlikely(__pyx_t_6 == ((PyObject *)NULL))) __PYX_ERR(0, 333, __pyx_L1_error)
    __pyx_t_2 = ((PyObject *)__pyx_t_6);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_side, ((PyObject*)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":334
 *             members = <list>PyTuple_GetItem(bunch, 1)
 *             side = <str>PyTuple_GetItem(bunch, 2)
 *             n = PyList_Size(members)             # <<<<<<<<<<<<<<
 *             if n > 0:
 *                 accumulator = 0.
 */
    __pyx_t_7 = PyList_Size(__pyx_v_members); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1L))) __PYX_ERR(0, 334, __pyx_L1_error)
    __pyx_v_n = __pyx_t_7;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":335
 *             side = <str>PyTuple_GetItem(bunch, 2)
 *             n = PyList_Size(members)
 *             if n > 0:             # <<<<<<<<<<<<<<
 *                 accumulator = 0.
 *                 j = 0
 */
    __pyx_t_4 = ((__pyx_v_n > 0) != 0);
    if (__pyx_t_4) {

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":336
 *             n = PyList_Size(members)
 *             if n > 0:
 *                 accumulator = 0.             # <<<<<<<<<<<<<<
 *                 j = 0
 *                 for j in range(n):
 */
      __pyx_v_accumulator = 0.;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":337
 *             if n > 0:
 *                 accumulator = 0.
 *                 j = 0             # <<<<<<<<<<<<<<
 *                 for j in range(n):
 *                     i = <SpanningMixin>PyList_GetItem(members, j)
 */
      __pyx_v_j = 0;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":338
 *                 accumulator = 0.
 *                 j = 0
 *                 for j in range(n):             # <<<<<<<<<<<<<<
 *                     i = <SpanningMixin>PyList_GetItem(members, j)
 *                     accumulator += (i.start + i.end) / 2.
 */
      __pyx_t_8 = __pyx_v_n;
      __pyx_t_9 = __pyx_t_8;
      for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_9; __pyx_t_10+=1) {
        __pyx_v_j = __pyx_t_10;

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":339
 *                 j = 0
 *                 for j in range(n):
 *                     i = <SpanningMixin>PyList_GetItem(members, j)             # <<<<<<<<<<<<<<
 *                     accumulator += (i.start + i.end) / 2.
 *                 if j > 0:
 */
        __pyx_t_6 = PyList_GetItem(__pyx_v_members, __pyx_v_j); if (unlikely(__pyx_t_6 == ((PyObject *)NULL))) __PYX_ERR(0, 339, __pyx_L1_error)
        __pyx_t_2 = ((PyObject *)__pyx_t_6);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_XDECREF_SET(__pyx_v_i, ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *)__pyx_t_2));
        __pyx_t_2 = 0;

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":340
 *                 for j in range(n):
 *                     i = <SpanningMixin>PyList_GetItem(members, j)
 *                     accumulator += (i.start + i.end) / 2.             # <<<<<<<<<<<<<<
 *                 if j > 0:
 *                     center = accumulator / (j + 1)
 */
        __pyx_v_accumulator = (__pyx_v_accumulator + ((__pyx_v_i->start + __pyx_v_i->end) / 2.));
      }

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":341
 *                     i = <SpanningMixin>PyList_GetItem(members, j)
 *                     accumulator += (i.start + i.end) / 2.
 *                 if j > 0:             # <<<<<<<<<<<<<<
 *                     center = accumulator / (j + 1)
 *                 else:
 */
      __pyx_t_4 = ((__pyx_v_j > 0) != 0);
      if (__pyx_t_4) {

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":342
 *                     accumulator += (i.start + i.end) / 2.
 *                 if j > 0:
 *                     center = accumulator / (j + 1)             # <<<<<<<<<<<<<<
 *                 else:
 *                     center = 0.0
 */
        __pyx_t_8 = (__pyx_v_j + 1);
        if (unlikely(__pyx_t_8 == 0)) {
          PyErr_SetString(PyExc_ZeroDivisionError, "float division");
          __PYX_ERR(0, 342, __pyx_L1_error)
        }
        __pyx_v_center = (__pyx_v_accumulator / __pyx_t_8);

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":341
 *                     i = <SpanningMixin>PyList_GetItem(members, j)
 *                     accumulator += (i.start + i.end) / 2.
 *                 if j > 0:             # <<<<<<<<<<<<<<
 *                     center = accumulator / (j + 1)
 *                 else:
 */
        goto __pyx_L9;
      }

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":344
 *                     center = accumulator / (j + 1)
 *                 else:
 *                     center = 0.0             # <<<<<<<<<<<<<<
 * 
 *                 left = []
 */
      /*else*/ {
        __pyx_v_center = 0.0;
      }
      __pyx_L9:;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":346
 *                     center = 0.0
 * 
 *                 left = []             # <<<<<<<<<<<<<<
 *                 right = []
 *                 contained = []
 */
      __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 346, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_XDECREF_SET(__pyx_v_left, ((PyObject*)__pyx_t_2));
      __pyx_t_2 = 0;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":347
 * 
 *                 left = []
 *                 right = []             # <<<<<<<<<<<<<<
 *                 contained = []
 * 
 */
      __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 347, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_XDECREF_SET(__pyx_v_right, ((PyObject*)__pyx_t_2));
      __pyx_t_2 = 0;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":348
 *                 left = []
 *                 right = []
 *                 contained = []             # <<<<<<<<<<<<<<
 * 
 *                 if n > 20:
 */
      __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 348, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_XDECREF_SET(__pyx_v_contained, ((PyObject*)__pyx_t_2));
      __pyx_t_2 = 0;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":350
 *                 contained = []
 * 
 *                 if n > 20:             # <<<<<<<<<<<<<<
 *                     j = 0
 *                     for j in range(n):
 */
      __pyx_t_4 = ((__pyx_v_n > 20) != 0);
      if (__pyx_t_4) {

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":351
 * 
 *                 if n > 20:
 *                     j = 0             # <<<<<<<<<<<<<<
 *                     for j in range(n):
 *                         i = <SpanningMixin>PyList_GetItem(members, j)
 */
        __pyx_v_j = 0;

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":352
 *                 if n > 20:
 *                     j = 0
 *                     for j in range(n):             # <<<<<<<<<<<<<<
 *                         i = <SpanningMixin>PyList_GetItem(members, j)
 *                         if abs(i.start - center) < 1e-6 and abs(i.end - center) < 1e-6:
 */
        __pyx_t_8 = __pyx_v_n;
        __pyx_t_9 = __pyx_t_8;
        for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_9; __pyx_t_10+=1) {
          __pyx_v_j = __pyx_t_10;

          /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":353
 *                     j = 0
 *                     for j in range(n):
 *                         i = <SpanningMixin>PyList_GetItem(members, j)             # <<<<<<<<<<<<<<
 *                         if abs(i.start - center) < 1e-6 and abs(i.end - center) < 1e-6:
 *                             contained.append(i)
 */
          __pyx_t_6 = PyList_GetItem(__pyx_v_members, __pyx_v_j); if (unlikely(__pyx_t_6 == ((PyObject *)NULL))) __PYX_ERR(0, 353, __pyx_L1_error)
          __pyx_t_2 = ((PyObject *)__pyx_t_6);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_XDECREF_SET(__pyx_v_i, ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *)__pyx_t_2));
          __pyx_t_2 = 0;

          /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":354
 *                     for j in range(n):
 *                         i = <SpanningMixin>PyList_GetItem(members, j)
 *                         if abs(i.start - center) < 1e-6 and abs(i.end - center) < 1e-6:             # <<<<<<<<<<<<<<
 *                             contained.append(i)
 *                         elif center > i.end:
 */
          __pyx_t_3 = ((fabs((__pyx_v_i->start - __pyx_v_center)) < 1e-6) != 0);
          if (__pyx_t_3) {
          } else {
            __pyx_t_4 = __pyx_t_3;
            goto __pyx_L14_bool_binop_done;
          }
          __pyx_t_3 = ((fabs((__pyx_v_i->end - __pyx_v_center)) < 1e-6) != 0);
          __pyx_t_4 = __pyx_t_3;
          __pyx_L14_bool_binop_done:;
          if (__pyx_t_4) {

            /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":355
 *                         i = <SpanningMixin>PyList_GetItem(members, j)
 *                         if abs(i.start - center) < 1e-6 and abs(i.end - center) < 1e-6:
 *                             contained.append(i)             # <<<<<<<<<<<<<<
 *                         elif center > i.end:
 *                             left.append(i)
 */
            __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_contained, ((PyObject *)__pyx_v_i)); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 355, __pyx_L1_error)

            /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":354
 *                     for j in range(n):
 *                         i = <SpanningMixin>PyList_GetItem(members, j)
 *                         if abs(i.start - center) < 1e-6 and abs(i.end - center) < 1e-6:             # <<<<<<<<<<<<<<
 *                             contained.append(i)
 *                         elif center > i.end:
 */
            goto __pyx_L13;
          }

          /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":356
 *                         if abs(i.start - center) < 1e-6 and abs(i.end - center) < 1e-6:
 *                             contained.append(i)
 *                         elif center > i.end:             # <<<<<<<<<<<<<<
 *                             left.append(i)
 *                         elif center < i.start:
 */
          __pyx_t_4 = ((__pyx_v_center > __pyx_v_i->end) != 0);
          if (__pyx_t_4) {

            /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":357
 *                             contained.append(i)
 *                         elif center > i.end:
 *                             left.append(i)             # <<<<<<<<<<<<<<
 *                         elif center < i.start:
 *                             right.append(i)
 */
            __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_left, ((PyObject *)__pyx_v_i)); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 357, __pyx_L1_error)

            /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":356
 *                         if abs(i.start - center) < 1e-6 and abs(i.end - center) < 1e-6:
 *                             contained.append(i)
 *                         elif center > i.end:             # <<<<<<<<<<<<<<
 *                             left.append(i)
 *                         elif center < i.start:
 */
            goto __pyx_L13;
          }

          /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":358
 *                         elif center > i.end:
 *                             left.append(i)
 *                         elif center < i.start:             # <<<<<<<<<<<<<<
 *                             right.append(i)
 *                         else:
 */
          __pyx_t_4 = ((__pyx_v_center < __pyx_v_i->start) != 0);
          if (__pyx_t_4) {

            /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":359
 *                             left.append(i)
 *                         elif center < i.start:
 *                             right.append(i)             # <<<<<<<<<<<<<<
 *                         else:
 *                             contained.append(i)
 */
            __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_right, ((PyObject *)__pyx_v_i)); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 359, __pyx_L1_error)

            /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":358
 *                         elif center > i.end:
 *                             left.append(i)
 *                         elif center < i.start:             # <<<<<<<<<<<<<<
 *                             right.append(i)
 *                         else:
 */
            goto __pyx_L13;
          }

          /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":361
 *                             right.append(i)
 *                         else:
 *                             contained.append(i)             # <<<<<<<<<<<<<<
 *                 else:
 *                     contained = list(members)
 */
          /*else*/ {
            __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_contained, ((PyObject *)__pyx_v_i)); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 361, __pyx_L1_error)
          }
          __pyx_L13:;
        }

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":350
 *                 contained = []
 * 
 *                 if n > 20:             # <<<<<<<<<<<<<<
 *                     j = 0
 *                     for j in range(n):
 */
        goto __pyx_L10;
      }

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":363
 *                             contained.append(i)
 *                 else:
 *                     contained = list(members)             # <<<<<<<<<<<<<<
 *                 node = cls(center, left=None, contained=contained, right=None, level=parent.level + 1, parent=parent)
 *                 if side == 'left':
 */
      /*else*/ {
        __pyx_t_2 = PySequence_List(__pyx_v_members); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 363, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF_SET(__pyx_v_contained, ((PyObject*)__pyx_t_2));
        __pyx_t_2 = 0;
      }
      __pyx_L10:;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":364
 *                 else:
 *                     contained = list(members)
 *                 node = cls(center, left=None, contained=contained, right=None, level=parent.level + 1, parent=parent)             # <<<<<<<<<<<<<<
 *                 if side == 'left':
 *                     parent.left = node
 */
      __pyx_t_2 = PyFloat_FromDouble(__pyx_v_center); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 364, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 364, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
      __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyDict_NewPresized(5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 364, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_left, Py_None) < 0) __PYX_ERR(0, 364, __pyx_L1_error)
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_contained, __pyx_v_contained) < 0) __PYX_ERR(0, 364, __pyx_L1_error)
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_right, Py_None) < 0) __PYX_ERR(0, 364, __pyx_L1_error)
      __pyx_t_11 = __Pyx_PyInt_From_long((__pyx_v_parent->level + 1)); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 364, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_level, __pyx_t_11) < 0) __PYX_ERR(0, 364, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_parent, ((PyObject *)__pyx_v_parent)) < 0) __PYX_ERR(0, 364, __pyx_L1_error)
      __pyx_t_11 = __Pyx_PyObject_Call(((PyObject *)__pyx_v_cls), __pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 364, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (!(likely(((__pyx_t_11) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_11, __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode))))) __PYX_ERR(0, 364, __pyx_L1_error)
      __Pyx_XDECREF_SET(__pyx_v_node, ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_t_11));
      __pyx_t_11 = 0;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":365
 *                     contained = list(members)
 *                 node = cls(center, left=None, contained=contained, right=None, level=parent.level + 1, parent=parent)
 *                 if side == 'left':             # <<<<<<<<<<<<<<
 *                     parent.left = node
 *                     parent.start = min(node.start, parent.start)
 */
      __pyx_t_4 = (__Pyx_PyString_Equals(__pyx_v_side, __pyx_n_s_left, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 365, __pyx_L1_error)
      __pyx_t_3 = (__pyx_t_4 != 0);
      if (__pyx_t_3) {

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":366
 *                 node = cls(center, left=None, contained=contained, right=None, level=parent.level + 1, parent=parent)
 *                 if side == 'left':
 *                     parent.left = node             # <<<<<<<<<<<<<<
 *                     parent.start = min(node.start, parent.start)
 *                     up = parent.parent
 */
        __Pyx_INCREF(((PyObject *)__pyx_v_node));
        __Pyx_GIVEREF(((PyObject *)__pyx_v_node));
        __Pyx_GOTREF(__pyx_v_parent->left);
        __Pyx_DECREF(((PyObject *)__pyx_v_parent->left));
        __pyx_v_parent->left = __pyx_v_node;

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":367
 *                 if side == 'left':
 *                     parent.left = node
 *                     parent.start = min(node.start, parent.start)             # <<<<<<<<<<<<<<
 *                     up = parent.parent
 *                     while up is not None:
 */
        __pyx_t_12 = __pyx_v_parent->start;
        __pyx_t_13 = __pyx_v_node->start;
        if (((__pyx_t_12 < __pyx_t_13) != 0)) {
          __pyx_t_14 = __pyx_t_12;
        } else {
          __pyx_t_14 = __pyx_t_13;
        }
        __pyx_v_parent->start = __pyx_t_14;

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":368
 *                     parent.left = node
 *                     parent.start = min(node.start, parent.start)
 *                     up = parent.parent             # <<<<<<<<<<<<<<
 *                     while up is not None:
 *                         if up.start > node.start:
 */
        __pyx_t_11 = ((PyObject *)__pyx_v_parent->parent);
        __Pyx_INCREF(__pyx_t_11);
        __Pyx_XDECREF_SET(__pyx_v_up, ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_t_11));
        __pyx_t_11 = 0;

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":369
 *                     parent.start = min(node.start, parent.start)
 *                     up = parent.parent
 *                     while up is not None:             # <<<<<<<<<<<<<<
 *                         if up.start > node.start:
 *                             up.start = node.start
 */
        while (1) {
          __pyx_t_3 = (((PyObject *)__pyx_v_up) != Py_None);
          __pyx_t_4 = (__pyx_t_3 != 0);
          if (!__pyx_t_4) break;

          /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":370
 *                     up = parent.parent
 *                     while up is not None:
 *                         if up.start > node.start:             # <<<<<<<<<<<<<<
 *                             up.start = node.start
 *                         up = up.parent
 */
          __pyx_t_4 = ((__pyx_v_up->start > __pyx_v_node->start) != 0);
          if (__pyx_t_4) {

            /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":371
 *                     while up is not None:
 *                         if up.start > node.start:
 *                             up.start = node.start             # <<<<<<<<<<<<<<
 *                         up = up.parent
 *                 elif side == 'right':
 */
            __pyx_t_14 = __pyx_v_node->start;
            __pyx_v_up->start = __pyx_t_14;

            /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":370
 *                     up = parent.parent
 *                     while up is not None:
 *                         if up.start > node.start:             # <<<<<<<<<<<<<<
 *                             up.start = node.start
 *                         up = up.parent
 */
          }

          /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":372
 *                         if up.start > node.start:
 *                             up.start = node.start
 *                         up = up.parent             # <<<<<<<<<<<<<<
 *                 elif side == 'right':
 *                     parent.right = node
 */
          __pyx_t_11 = ((PyObject *)__pyx_v_up->parent);
          __Pyx_INCREF(__pyx_t_11);
          __Pyx_DECREF_SET(__pyx_v_up, ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_t_11));
          __pyx_t_11 = 0;
        }

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":365
 *                     contained = list(members)
 *                 node = cls(center, left=None, contained=contained, right=None, level=parent.level + 1, parent=parent)
 *                 if side == 'left':             # <<<<<<<<<<<<<<
 *                     parent.left = node
 *                     parent.start = min(node.start, parent.start)
 */
        goto __pyx_L16;
      }

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":373
 *                             up.start = node.start
 *                         up = up.parent
 *                 elif side == 'right':             # <<<<<<<<<<<<<<
 *                     parent.right = node
 *                     parent.end = max(node.end, parent.end)
 */
      __pyx_t_4 = (__Pyx_PyString_Equals(__pyx_v_side, __pyx_n_s_right, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 373, __pyx_L1_error)
      __pyx_t_3 = (__pyx_t_4 != 0);
      if (likely(__pyx_t_3)) {

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":374
 *                         up = up.parent
 *                 elif side == 'right':
 *                     parent.right = node             # <<<<<<<<<<<<<<
 *                     parent.end = max(node.end, parent.end)
 *                     up = parent.parent
 */
        __Pyx_INCREF(((PyObject *)__pyx_v_node));
        __Pyx_GIVEREF(((PyObject *)__pyx_v_node));
        __Pyx_GOTREF(__pyx_v_parent->right);
        __Pyx_DECREF(((PyObject *)__pyx_v_parent->right));
        __pyx_v_parent->right = __pyx_v_node;

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":375
 *                 elif side == 'right':
 *                     parent.right = node
 *                     parent.end = max(node.end, parent.end)             # <<<<<<<<<<<<<<
 *                     up = parent.parent
 *                     while up is not None:
 */
        __pyx_t_14 = __pyx_v_parent->end;
        __pyx_t_12 = __pyx_v_node->end;
        if (((__pyx_t_14 > __pyx_t_12) != 0)) {
          __pyx_t_13 = __pyx_t_14;
        } else {
          __pyx_t_13 = __pyx_t_12;
        }
        __pyx_v_parent->end = __pyx_t_13;

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":376
 *                     parent.right = node
 *                     parent.end = max(node.end, parent.end)
 *                     up = parent.parent             # <<<<<<<<<<<<<<
 *                     while up is not None:
 *                         if up.end < node.end:
 */
        __pyx_t_11 = ((PyObject *)__pyx_v_parent->parent);
        __Pyx_INCREF(__pyx_t_11);
        __Pyx_XDECREF_SET(__pyx_v_up, ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_t_11));
        __pyx_t_11 = 0;

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":377
 *                     parent.end = max(node.end, parent.end)
 *                     up = parent.parent
 *                     while up is not None:             # <<<<<<<<<<<<<<
 *                         if up.end < node.end:
 *                             up.end = node.end
 */
        while (1) {
          __pyx_t_3 = (((PyObject *)__pyx_v_up) != Py_None);
          __pyx_t_4 = (__pyx_t_3 != 0);
          if (!__pyx_t_4) break;

          /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":378
 *                     up = parent.parent
 *                     while up is not None:
 *                         if up.end < node.end:             # <<<<<<<<<<<<<<
 *                             up.end = node.end
 *                         up = up.parent
 */
          __pyx_t_4 = ((__pyx_v_up->end < __pyx_v_node->end) != 0);
          if (__pyx_t_4) {

            /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":379
 *                     while up is not None:
 *                         if up.end < node.end:
 *                             up.end = node.end             # <<<<<<<<<<<<<<
 *                         up = up.parent
 *                 else:
 */
            __pyx_t_13 = __pyx_v_node->end;
            __pyx_v_up->end = __pyx_t_13;

            /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":378
 *                     up = parent.parent
 *                     while up is not None:
 *                         if up.end < node.end:             # <<<<<<<<<<<<<<
 *                             up.end = node.end
 *                         up = up.parent
 */
          }

          /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":380
 *                         if up.end < node.end:
 *                             up.end = node.end
 *                         up = up.parent             # <<<<<<<<<<<<<<
 *                 else:
 *                     raise ValueError(side)
 */
          __pyx_t_11 = ((PyObject *)__pyx_v_up->parent);
          __Pyx_INCREF(__pyx_t_11);
          __Pyx_DECREF_SET(__pyx_v_up, ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_t_11));
          __pyx_t_11 = 0;
        }

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":373
 *                             up.start = node.start
 *                         up = up.parent
 *                 elif side == 'right':             # <<<<<<<<<<<<<<
 *                     parent.right = node
 *                     parent.end = max(node.end, parent.end)
 */
        goto __pyx_L16;
      }

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":382
 *                         up = up.parent
 *                 else:
 *                     raise ValueError(side)             # <<<<<<<<<<<<<<
 *                 stack.append((node, left, "left"))
 *                 stack.append((node, right, "right"))
 */
      /*else*/ {
        __pyx_t_11 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_v_side); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 382, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_Raise(__pyx_t_11, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __PYX_ERR(0, 382, __pyx_L1_error)
      }
      __pyx_L16:;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":383
 *                 else:
 *                     raise ValueError(side)
 *                 stack.append((node, left, "left"))             # <<<<<<<<<<<<<<
 *                 stack.append((node, right, "right"))
 *         return root.left
 */
      __pyx_t_11 = PyTuple_New(3); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 383, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_INCREF(((PyObject *)__pyx_v_node));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_node));
      PyTuple_SET_ITEM(__pyx_t_11, 0, ((PyObject *)__pyx_v_node));
      __Pyx_INCREF(__pyx_v_left);
      __Pyx_GIVEREF(__pyx_v_left);
      PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_v_left);
      __Pyx_INCREF(__pyx_n_s_left);
      __Pyx_GIVEREF(__pyx_n_s_left);
      PyTuple_SET_ITEM(__pyx_t_11, 2, __pyx_n_s_left);
      __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_stack, __pyx_t_11); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 383, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":384
 *                     raise ValueError(side)
 *                 stack.append((node, left, "left"))
 *                 stack.append((node, right, "right"))             # <<<<<<<<<<<<<<
 *         return root.left
 * 
 */
      __pyx_t_11 = PyTuple_New(3); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 384, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_INCREF(((PyObject *)__pyx_v_node));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_node));
      PyTuple_SET_ITEM(__pyx_t_11, 0, ((PyObject *)__pyx_v_node));
      __Pyx_INCREF(__pyx_v_right);
      __Pyx_GIVEREF(__pyx_v_right);
      PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_v_right);
      __Pyx_INCREF(__pyx_n_s_right);
      __Pyx_GIVEREF(__pyx_n_s_right);
      PyTuple_SET_ITEM(__pyx_t_11, 2, __pyx_n_s_right);
      __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_stack, __pyx_t_11); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 384, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":335
 *             side = <str>PyTuple_GetItem(bunch, 2)
 *             n = PyList_Size(members)
 *             if n > 0:             # <<<<<<<<<<<<<<
 *                 accumulator = 0.
 *                 j = 0
 */
    }
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":385
 *                 stack.append((node, left, "left"))
 *                 stack.append((node, right, "right"))
 *         return root.left             # <<<<<<<<<<<<<<
 * 
 *     def _update_bounds(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_root->left));
  __pyx_r = ((PyObject *)__pyx_v_root->left);
  goto __pyx_L0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":311
 * 
 *     @classmethod
 *     def build(cls, intervals):             # <<<<<<<<<<<<<<
 *         cdef:
 *             IntervalTreeNode root, parent, node, up
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.build", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_root);
  __Pyx_XDECREF((PyObject *)__pyx_v_parent);
  __Pyx_XDECREF((PyObject *)__pyx_v_node);
  __Pyx_XDECREF((PyObject *)__pyx_v_up);
  __Pyx_XDECREF(__pyx_v_stack);
  __Pyx_XDECREF(__pyx_v_members);
  __Pyx_XDECREF(__pyx_v_left);
  __Pyx_XDECREF(__pyx_v_right);
  __Pyx_XDECREF(__pyx_v_contained);
  __Pyx_XDECREF(__pyx_v_intervals_);
  __Pyx_XDECREF(__pyx_v_bunch);
  __Pyx_XDECREF(__pyx_v_side);
  __Pyx_XDECREF((PyObject *)__pyx_v_i);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":387
 *         return root.left
 * 
 *     def _update_bounds(self):             # <<<<<<<<<<<<<<
 *         changed = False
 *         if self.left is not None and self.left.start < self.start:
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_21_update_bounds(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_20_update_bounds[] = "IntervalTreeNode._update_bounds(self)";
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_21_update_bounds(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_update_bounds (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_20_update_bounds(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_20_update_bounds(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self) {
  int __pyx_v_changed;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  double __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  __Pyx_RefNannySetupContext("_update_bounds", 0);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":388
 * 
 *     def _update_bounds(self):
 *         changed = False             # <<<<<<<<<<<<<<
 *         if self.left is not None and self.left.start < self.start:
 *             changed = True
 */
  __pyx_v_changed = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":389
 *     def _update_bounds(self):
 *         changed = False
 *         if self.left is not None and self.left.start < self.start:             # <<<<<<<<<<<<<<
 *             changed = True
 *             self.start = self.left.start
 */
  __pyx_t_2 = (((PyObject *)__pyx_v_self->left) != Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = ((__pyx_v_self->left->start < __pyx_v_self->start) != 0);
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":390
 *         changed = False
 *         if self.left is not None and self.left.start < self.start:
 *             changed = True             # <<<<<<<<<<<<<<
 *             self.start = self.left.start
 *         if self.right is not None and self.right.end > self.end:
 */
    __pyx_v_changed = 1;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":391
 *         if self.left is not None and self.left.start < self.start:
 *             changed = True
 *             self.start = self.left.start             # <<<<<<<<<<<<<<
 *         if self.right is not None and self.right.end > self.end:
 *             changed = True
 */
    __pyx_t_4 = __pyx_v_self->left->start;
    __pyx_v_self->start = __pyx_t_4;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":389
 *     def _update_bounds(self):
 *         changed = False
 *         if self.left is not None and self.left.start < self.start:             # <<<<<<<<<<<<<<
 *             changed = True
 *             self.start = self.left.start
 */
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":392
 *             changed = True
 *             self.start = self.left.start
 *         if self.right is not None and self.right.end > self.end:             # <<<<<<<<<<<<<<
 *             changed = True
 *             self.end = self.right.end
 */
  __pyx_t_3 = (((PyObject *)__pyx_v_self->right) != Py_None);
  __pyx_t_2 = (__pyx_t_3 != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_2 = ((__pyx_v_self->right->end > __pyx_v_self->end) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_1) {

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":393
 *             self.start = self.left.start
 *         if self.right is not None and self.right.end > self.end:
 *             changed = True             # <<<<<<<<<<<<<<
 *             self.end = self.right.end
 *         if self.parent is not None:
 */
    __pyx_v_changed = 1;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":394
 *         if self.right is not None and self.right.end > self.end:
 *             changed = True
 *             self.end = self.right.end             # <<<<<<<<<<<<<<
 *         if self.parent is not None:
 *             self.parent._update_bounds()
 */
    __pyx_t_4 = __pyx_v_self->right->end;
    __pyx_v_self->end = __pyx_t_4;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":392
 *             changed = True
 *             self.start = self.left.start
 *         if self.right is not None and self.right.end > self.end:             # <<<<<<<<<<<<<<
 *             changed = True
 *             self.end = self.right.end
 */
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":395
 *             changed = True
 *             self.end = self.right.end
 *         if self.parent is not None:             # <<<<<<<<<<<<<<
 *             self.parent._update_bounds()
 *         return changed
 */
  __pyx_t_1 = (((PyObject *)__pyx_v_self->parent) != Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":396
 *             self.end = self.right.end
 *         if self.parent is not None:
 *             self.parent._update_bounds()             # <<<<<<<<<<<<<<
 *         return changed
 * 
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->parent), __pyx_n_s_update_bounds); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 396, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_5 = (__pyx_t_7) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_7) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 396, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":395
 *             changed = True
 *             self.end = self.right.end
 *         if self.parent is not None:             # <<<<<<<<<<<<<<
 *             self.parent._update_bounds()
 *         return changed
 */
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":397
 *         if self.parent is not None:
 *             self.parent._update_bounds()
 *         return changed             # <<<<<<<<<<<<<<
 * 
 *     def insert(self, interval):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_5 = __Pyx_PyBool_FromLong(__pyx_v_changed); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 397, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":387
 *         return root.left
 * 
 *     def _update_bounds(self):             # <<<<<<<<<<<<<<
 *         changed = False
 *         if self.left is not None and self.left.start < self.start:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode._update_bounds", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":399
 *         return changed
 * 
 *     def insert(self, interval):             # <<<<<<<<<<<<<<
 *         insert_in_self = False
 *         if self.node_contains(interval.start):
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_23insert(PyObject *__pyx_v_self, PyObject *__pyx_v_interval); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_22insert[] = "IntervalTreeNode.insert(self, interval)";
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_23insert(PyObject *__pyx_v_self, PyObject *__pyx_v_interval) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("insert (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_22insert(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self), ((PyObject *)__pyx_v_interval));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_22insert(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, PyObject *__pyx_v_interval) {
  int __pyx_v_insert_in_self;
  int __pyx_v_changed;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  double __pyx_t_9;
  __Pyx_RefNannySetupContext("insert", 0);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":400
 * 
 *     def insert(self, interval):
 *         insert_in_self = False             # <<<<<<<<<<<<<<
 *         if self.node_contains(interval.start):
 *             if self.left is not None and self.left.node_contains(interval.end):
 */
  __pyx_v_insert_in_self = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":401
 *     def insert(self, interval):
 *         insert_in_self = False
 *         if self.node_contains(interval.start):             # <<<<<<<<<<<<<<
 *             if self.left is not None and self.left.node_contains(interval.end):
 *                 return self.left.insert(interval)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_node_contains); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 401, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_interval, __pyx_n_s_start); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 401, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 401, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 401, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_5) {

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":402
 *         insert_in_self = False
 *         if self.node_contains(interval.start):
 *             if self.left is not None and self.left.node_contains(interval.end):             # <<<<<<<<<<<<<<
 *                 return self.left.insert(interval)
 *             elif (self.right is not None and self.right.node_contains(interval.end) and
 */
    __pyx_t_6 = (((PyObject *)__pyx_v_self->left) != Py_None);
    __pyx_t_7 = (__pyx_t_6 != 0);
    if (__pyx_t_7) {
    } else {
      __pyx_t_5 = __pyx_t_7;
      goto __pyx_L5_bool_binop_done;
    }
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->left), __pyx_n_s_node_contains); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 402, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_interval, __pyx_n_s_end); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 402, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 402, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 402, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_5 = __pyx_t_7;
    __pyx_L5_bool_binop_done:;
    if (__pyx_t_5) {

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":403
 *         if self.node_contains(interval.start):
 *             if self.left is not None and self.left.node_contains(interval.end):
 *                 return self.left.insert(interval)             # <<<<<<<<<<<<<<
 *             elif (self.right is not None and self.right.node_contains(interval.end) and
 *                   self.right.node_contains(interval.start)):
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->left), __pyx_n_s_insert); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 403, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
        }
      }
      __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_interval) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_interval);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 403, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_1;
      __pyx_t_1 = 0;
      goto __pyx_L0;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":402
 *         insert_in_self = False
 *         if self.node_contains(interval.start):
 *             if self.left is not None and self.left.node_contains(interval.end):             # <<<<<<<<<<<<<<
 *                 return self.left.insert(interval)
 *             elif (self.right is not None and self.right.node_contains(interval.end) and
 */
    }

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":404
 *             if self.left is not None and self.left.node_contains(interval.end):
 *                 return self.left.insert(interval)
 *             elif (self.right is not None and self.right.node_contains(interval.end) and             # <<<<<<<<<<<<<<
 *                   self.right.node_contains(interval.start)):
 *                 return self.right.insert(interval)
 */
    __pyx_t_7 = (((PyObject *)__pyx_v_self->right) != Py_None);
    __pyx_t_6 = (__pyx_t_7 != 0);
    if (__pyx_t_6) {
    } else {
      __pyx_t_5 = __pyx_t_6;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->right), __pyx_n_s_node_contains); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 404, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_interval, __pyx_n_s_end); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 404, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 404, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 404, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_6) {
    } else {
      __pyx_t_5 = __pyx_t_6;
      goto __pyx_L7_bool_binop_done;
    }

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":405
 *                 return self.left.insert(interval)
 *             elif (self.right is not None and self.right.node_contains(interval.end) and
 *                   self.right.node_contains(interval.start)):             # <<<<<<<<<<<<<<
 *                 return self.right.insert(interval)
 *             else:
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->right), __pyx_n_s_node_contains); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 405, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_interval, __pyx_n_s_start); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 405, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 405, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 405, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_5 = __pyx_t_6;
    __pyx_L7_bool_binop_done:;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":404
 *             if self.left is not None and self.left.node_contains(interval.end):
 *                 return self.left.insert(interval)
 *             elif (self.right is not None and self.right.node_contains(interval.end) and             # <<<<<<<<<<<<<<
 *                   self.right.node_contains(interval.start)):
 *                 return self.right.insert(interval)
 */
    if (__pyx_t_5) {

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":406
 *             elif (self.right is not None and self.right.node_contains(interval.end) and
 *                   self.right.node_contains(interval.start)):
 *                 return self.right.insert(interval)             # <<<<<<<<<<<<<<
 *             else:
 *                 insert_in_self = True
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->right), __pyx_n_s_insert); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 406, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
        }
      }
      __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_interval) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_interval);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 406, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_1;
      __pyx_t_1 = 0;
      goto __pyx_L0;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":404
 *             if self.left is not None and self.left.node_contains(interval.end):
 *                 return self.left.insert(interval)
 *             elif (self.right is not None and self.right.node_contains(interval.end) and             # <<<<<<<<<<<<<<
 *                   self.right.node_contains(interval.start)):
 *                 return self.right.insert(interval)
 */
    }

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":408
 *                 return self.right.insert(interval)
 *             else:
 *                 insert_in_self = True             # <<<<<<<<<<<<<<
 *         elif self.node_contains(interval.end):
 *             if self.right is not None and self.right.node_contains(interval.start):
 */
    /*else*/ {
      __pyx_v_insert_in_self = 1;
    }

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":401
 *     def insert(self, interval):
 *         insert_in_self = False
 *         if self.node_contains(interval.start):             # <<<<<<<<<<<<<<
 *             if self.left is not None and self.left.node_contains(interval.end):
 *                 return self.left.insert(interval)
 */
    goto __pyx_L3;
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":409
 *             else:
 *                 insert_in_self = True
 *         elif self.node_contains(interval.end):             # <<<<<<<<<<<<<<
 *             if self.right is not None and self.right.node_contains(interval.start):
 *                 return self.right.insert(interval)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_node_contains); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 409, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_interval, __pyx_n_s_end); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 409, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 409, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 409, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_5) {

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":410
 *                 insert_in_self = True
 *         elif self.node_contains(interval.end):
 *             if self.right is not None and self.right.node_contains(interval.start):             # <<<<<<<<<<<<<<
 *                 return self.right.insert(interval)
 *             else:
 */
    __pyx_t_6 = (((PyObject *)__pyx_v_self->right) != Py_None);
    __pyx_t_7 = (__pyx_t_6 != 0);
    if (__pyx_t_7) {
    } else {
      __pyx_t_5 = __pyx_t_7;
      goto __pyx_L11_bool_binop_done;
    }
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->right), __pyx_n_s_node_contains); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 410, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_interval, __pyx_n_s_start); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 410, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 410, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 410, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_5 = __pyx_t_7;
    __pyx_L11_bool_binop_done:;
    if (__pyx_t_5) {

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":411
 *         elif self.node_contains(interval.end):
 *             if self.right is not None and self.right.node_contains(interval.start):
 *                 return self.right.insert(interval)             # <<<<<<<<<<<<<<
 *             else:
 *                 insert_in_self = True
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->right), __pyx_n_s_insert); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 411, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
        }
      }
      __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_interval) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_interval);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 411, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_1;
      __pyx_t_1 = 0;
      goto __pyx_L0;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":410
 *                 insert_in_self = True
 *         elif self.node_contains(interval.end):
 *             if self.right is not None and self.right.node_contains(interval.start):             # <<<<<<<<<<<<<<
 *                 return self.right.insert(interval)
 *             else:
 */
    }

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":413
 *                 return self.right.insert(interval)
 *             else:
 *                 insert_in_self = True             # <<<<<<<<<<<<<<
 *         if not insert_in_self and self.parent is None:
 *             insert_in_self = True
 */
    /*else*/ {
      __pyx_v_insert_in_self = 1;
    }

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":409
 *             else:
 *                 insert_in_self = True
 *         elif self.node_contains(interval.end):             # <<<<<<<<<<<<<<
 *             if self.right is not None and self.right.node_contains(interval.start):
 *                 return self.right.insert(interval)
 */
  }
  __pyx_L3:;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":414
 *             else:
 *                 insert_in_self = True
 *         if not insert_in_self and self.parent is None:             # <<<<<<<<<<<<<<
 *             insert_in_self = True
 *         if insert_in_self:
 */
  __pyx_t_7 = ((!(__pyx_v_insert_in_self != 0)) != 0);
  if (__pyx_t_7) {
  } else {
    __pyx_t_5 = __pyx_t_7;
    goto __pyx_L14_bool_binop_done;
  }
  __pyx_t_7 = (((PyObject *)__pyx_v_self->parent) == Py_None);
  __pyx_t_6 = (__pyx_t_7 != 0);
  __pyx_t_5 = __pyx_t_6;
  __pyx_L14_bool_binop_done:;
  if (__pyx_t_5) {

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":415
 *                 insert_in_self = True
 *         if not insert_in_self and self.parent is None:
 *             insert_in_self = True             # <<<<<<<<<<<<<<
 *         if insert_in_self:
 *             self.contained.append(interval)
 */
    __pyx_v_insert_in_self = 1;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":414
 *             else:
 *                 insert_in_self = True
 *         if not insert_in_self and self.parent is None:             # <<<<<<<<<<<<<<
 *             insert_in_self = True
 *         if insert_in_self:
 */
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":416
 *         if not insert_in_self and self.parent is None:
 *             insert_in_self = True
 *         if insert_in_self:             # <<<<<<<<<<<<<<
 *             self.contained.append(interval)
 *             changed = False
 */
  __pyx_t_5 = (__pyx_v_insert_in_self != 0);
  if (__pyx_t_5) {

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":417
 *             insert_in_self = True
 *         if insert_in_self:
 *             self.contained.append(interval)             # <<<<<<<<<<<<<<
 *             changed = False
 *             if interval.start < self.start:
 */
    if (unlikely(__pyx_v_self->contained == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
      __PYX_ERR(0, 417, __pyx_L1_error)
    }
    __pyx_t_8 = __Pyx_PyList_Append(__pyx_v_self->contained, __pyx_v_interval); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 417, __pyx_L1_error)

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":418
 *         if insert_in_self:
 *             self.contained.append(interval)
 *             changed = False             # <<<<<<<<<<<<<<
 *             if interval.start < self.start:
 *                 self.start = interval.start
 */
    __pyx_v_changed = 0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":419
 *             self.contained.append(interval)
 *             changed = False
 *             if interval.start < self.start:             # <<<<<<<<<<<<<<
 *                 self.start = interval.start
 *                 changed = True
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_interval, __pyx_n_s_start); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 419, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->start); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 419, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_LT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 419, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 419, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_5) {

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":420
 *             changed = False
 *             if interval.start < self.start:
 *                 self.start = interval.start             # <<<<<<<<<<<<<<
 *                 changed = True
 *             if interval.end > self.end:
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_interval, __pyx_n_s_start); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 420, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_9 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_9 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 420, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_v_self->start = __pyx_t_9;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":421
 *             if interval.start < self.start:
 *                 self.start = interval.start
 *                 changed = True             # <<<<<<<<<<<<<<
 *             if interval.end > self.end:
 *                 self.end = interval.end
 */
      __pyx_v_changed = 1;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":419
 *             self.contained.append(interval)
 *             changed = False
 *             if interval.start < self.start:             # <<<<<<<<<<<<<<
 *                 self.start = interval.start
 *                 changed = True
 */
    }

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":422
 *                 self.start = interval.start
 *                 changed = True
 *             if interval.end > self.end:             # <<<<<<<<<<<<<<
 *                 self.end = interval.end
 *                 changed = True
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_interval, __pyx_n_s_end); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 422, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->end); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 422, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = PyObject_RichCompare(__pyx_t_3, __pyx_t_2, Py_GT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 422, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 422, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_5) {

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":423
 *                 changed = True
 *             if interval.end > self.end:
 *                 self.end = interval.end             # <<<<<<<<<<<<<<
 *                 changed = True
 *             if changed:
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_interval, __pyx_n_s_end); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 423, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_9 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_9 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 423, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_v_self->end = __pyx_t_9;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":424
 *             if interval.end > self.end:
 *                 self.end = interval.end
 *                 changed = True             # <<<<<<<<<<<<<<
 *             if changed:
 *                 self._update_bounds()
 */
      __pyx_v_changed = 1;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":422
 *                 self.start = interval.start
 *                 changed = True
 *             if interval.end > self.end:             # <<<<<<<<<<<<<<
 *                 self.end = interval.end
 *                 changed = True
 */
    }

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":425
 *                 self.end = interval.end
 *                 changed = True
 *             if changed:             # <<<<<<<<<<<<<<
 *                 self._update_bounds()
 * 
 */
    __pyx_t_5 = (__pyx_v_changed != 0);
    if (__pyx_t_5) {

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":426
 *                 changed = True
 *             if changed:
 *                 self._update_bounds()             # <<<<<<<<<<<<<<
 * 
 *     def flatten(self):
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_update_bounds); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 426, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
        }
      }
      __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 426, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":425
 *                 self.end = interval.end
 *                 changed = True
 *             if changed:             # <<<<<<<<<<<<<<
 *                 self._update_bounds()
 * 
 */
    }

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":416
 *         if not insert_in_self and self.parent is None:
 *             insert_in_self = True
 *         if insert_in_self:             # <<<<<<<<<<<<<<
 *             self.contained.append(interval)
 *             changed = False
 */
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":399
 *         return changed
 * 
 *     def insert(self, interval):             # <<<<<<<<<<<<<<
 *         insert_in_self = False
 *         if self.node_contains(interval.start):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.insert", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":428
 *                 self._update_bounds()
 * 
 *     def flatten(self):             # <<<<<<<<<<<<<<
 *         # perform an infix traversal of the tree and collect
 *         # the elements
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_25flatten(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_24flatten[] = "IntervalTreeNode.flatten(self)";
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_25flatten(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("flatten (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_24flatten(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_24flatten(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self) {
  PyObject *__pyx_v_items = NULL;
  PyObject *__pyx_v_stack = NULL;
  PyObject *__pyx_v_current = NULL;
  int __pyx_v_done;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  __Pyx_RefNannySetupContext("flatten", 0);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":431
 *         # perform an infix traversal of the tree and collect
 *         # the elements
 *         items = []             # <<<<<<<<<<<<<<
 *         stack = deque([])
 *         current = self
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 431, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_items = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":432
 *         # the elements
 *         items = []
 *         stack = deque([])             # <<<<<<<<<<<<<<
 *         current = self
 *         done = False
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_deque); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 432, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 432, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 432, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_stack = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":433
 *         items = []
 *         stack = deque([])
 *         current = self             # <<<<<<<<<<<<<<
 *         done = False
 *         while not done:
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __pyx_v_current = ((PyObject *)__pyx_v_self);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":434
 *         stack = deque([])
 *         current = self
 *         done = False             # <<<<<<<<<<<<<<
 *         while not done:
 *             if current is not None:
 */
  __pyx_v_done = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":435
 *         current = self
 *         done = False
 *         while not done:             # <<<<<<<<<<<<<<
 *             if current is not None:
 *                 stack.append(current)
 */
  while (1) {
    __pyx_t_5 = ((!(__pyx_v_done != 0)) != 0);
    if (!__pyx_t_5) break;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":436
 *         done = False
 *         while not done:
 *             if current is not None:             # <<<<<<<<<<<<<<
 *                 stack.append(current)
 *                 current = current.left
 */
    __pyx_t_5 = (__pyx_v_current != Py_None);
    __pyx_t_6 = (__pyx_t_5 != 0);
    if (__pyx_t_6) {

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":437
 *         while not done:
 *             if current is not None:
 *                 stack.append(current)             # <<<<<<<<<<<<<<
 *                 current = current.left
 *             else:
 */
      __pyx_t_7 = __Pyx_PyObject_Append(__pyx_v_stack, __pyx_v_current); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 437, __pyx_L1_error)

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":438
 *             if current is not None:
 *                 stack.append(current)
 *                 current = current.left             # <<<<<<<<<<<<<<
 *             else:
 *                 if stack:
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_current, __pyx_n_s_left); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 438, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF_SET(__pyx_v_current, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":436
 *         done = False
 *         while not done:
 *             if current is not None:             # <<<<<<<<<<<<<<
 *                 stack.append(current)
 *                 current = current.left
 */
      goto __pyx_L5;
    }

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":440
 *                 current = current.left
 *             else:
 *                 if stack:             # <<<<<<<<<<<<<<
 *                     current = stack.pop()
 *                     items.extend(current.contained)
 */
    /*else*/ {
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_stack); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 440, __pyx_L1_error)
      if (__pyx_t_6) {

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":441
 *             else:
 *                 if stack:
 *                     current = stack.pop()             # <<<<<<<<<<<<<<
 *                     items.extend(current.contained)
 *                     current = current.right
 */
        __pyx_t_1 = __Pyx_PyObject_Pop(__pyx_v_stack); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 441, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF_SET(__pyx_v_current, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":442
 *                 if stack:
 *                     current = stack.pop()
 *                     items.extend(current.contained)             # <<<<<<<<<<<<<<
 *                     current = current.right
 *                 else:
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_current, __pyx_n_s_contained); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 442, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = __Pyx_PyList_Extend(__pyx_v_items, __pyx_t_1); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 442, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":443
 *                     current = stack.pop()
 *                     items.extend(current.contained)
 *                     current = current.right             # <<<<<<<<<<<<<<
 *                 else:
 *                     done = True
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_current, __pyx_n_s_right); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 443, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF_SET(__pyx_v_current, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":440
 *                 current = current.left
 *             else:
 *                 if stack:             # <<<<<<<<<<<<<<
 *                     current = stack.pop()
 *                     items.extend(current.contained)
 */
        goto __pyx_L6;
      }

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":445
 *                     current = current.right
 *                 else:
 *                     done = True             # <<<<<<<<<<<<<<
 *         return items
 * 
 */
      /*else*/ {
        __pyx_v_done = 1;
      }
      __pyx_L6:;
    }
    __pyx_L5:;
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":446
 *                 else:
 *                     done = True
 *         return items             # <<<<<<<<<<<<<<
 * 
 *     def balance(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_items);
  __pyx_r = __pyx_v_items;
  goto __pyx_L0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":428
 *                 self._update_bounds()
 * 
 *     def flatten(self):             # <<<<<<<<<<<<<<
 *         # perform an infix traversal of the tree and collect
 *         # the elements
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.flatten", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_items);
  __Pyx_XDECREF(__pyx_v_stack);
  __Pyx_XDECREF(__pyx_v_current);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":448
 *         return items
 * 
 *     def balance(self):             # <<<<<<<<<<<<<<
 *         items = self.flatten()
 *         tree = self.build(items)
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_27balance(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_26balance[] = "IntervalTreeNode.balance(self)";
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_27balance(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("balance (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_26balance(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_26balance(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self) {
  PyObject *__pyx_v_items = NULL;
  PyObject *__pyx_v_tree = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  double __pyx_t_4;
  __Pyx_RefNannySetupContext("balance", 0);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":449
 * 
 *     def balance(self):
 *         items = self.flatten()             # <<<<<<<<<<<<<<
 *         tree = self.build(items)
 *         self.left = tree.left
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_flatten); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 449, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 449, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_items = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":450
 *     def balance(self):
 *         items = self.flatten()
 *         tree = self.build(items)             # <<<<<<<<<<<<<<
 *         self.left = tree.left
 *         self.right = tree.right
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_build); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 450, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_items) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_items);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 450, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_tree = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":451
 *         items = self.flatten()
 *         tree = self.build(items)
 *         self.left = tree.left             # <<<<<<<<<<<<<<
 *         self.right = tree.right
 *         self.contained = tree.contained
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_tree, __pyx_n_s_left); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode))))) __PYX_ERR(0, 451, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->left);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->left));
  __pyx_v_self->left = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":452
 *         tree = self.build(items)
 *         self.left = tree.left
 *         self.right = tree.right             # <<<<<<<<<<<<<<
 *         self.contained = tree.contained
 *         self.start = tree.start
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_tree, __pyx_n_s_right); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 452, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode))))) __PYX_ERR(0, 452, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->right);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->right));
  __pyx_v_self->right = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":453
 *         self.left = tree.left
 *         self.right = tree.right
 *         self.contained = tree.contained             # <<<<<<<<<<<<<<
 *         self.start = tree.start
 *         self.end = tree.end
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_tree, __pyx_n_s_contained); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 453, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyList_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 453, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->contained);
  __Pyx_DECREF(__pyx_v_self->contained);
  __pyx_v_self->contained = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":454
 *         self.right = tree.right
 *         self.contained = tree.contained
 *         self.start = tree.start             # <<<<<<<<<<<<<<
 *         self.end = tree.end
 *         self.center = tree.center
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_tree, __pyx_n_s_start); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 454, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_4 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 454, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_self->start = __pyx_t_4;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":455
 *         self.contained = tree.contained
 *         self.start = tree.start
 *         self.end = tree.end             # <<<<<<<<<<<<<<
 *         self.center = tree.center
 *         self.left.parent = self
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_tree, __pyx_n_s_end); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 455, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_4 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 455, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_self->end = __pyx_t_4;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":456
 *         self.start = tree.start
 *         self.end = tree.end
 *         self.center = tree.center             # <<<<<<<<<<<<<<
 *         self.left.parent = self
 *         self.right.parent = self
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_tree, __pyx_n_s_center); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 456, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_4 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 456, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_self->center = __pyx_t_4;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":457
 *         self.end = tree.end
 *         self.center = tree.center
 *         self.left.parent = self             # <<<<<<<<<<<<<<
 *         self.right.parent = self
 *         self._update_bounds()
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  __Pyx_GOTREF(__pyx_v_self->left->parent);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->left->parent));
  __pyx_v_self->left->parent = __pyx_v_self;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":458
 *         self.center = tree.center
 *         self.left.parent = self
 *         self.right.parent = self             # <<<<<<<<<<<<<<
 *         self._update_bounds()
 * 
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  __Pyx_GOTREF(__pyx_v_self->right->parent);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->right->parent));
  __pyx_v_self->right->parent = __pyx_v_self;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":459
 *         self.left.parent = self
 *         self.right.parent = self
 *         self._update_bounds()             # <<<<<<<<<<<<<<
 * 
 *     def __iter__(self):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_update_bounds); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 459, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 459, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":448
 *         return items
 * 
 *     def balance(self):             # <<<<<<<<<<<<<<
 *         items = self.flatten()
 *         tree = self.build(items)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.balance", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_items);
  __Pyx_XDECREF(__pyx_v_tree);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":461
 *         self._update_bounds()
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         flat = self.flatten()
 *         return iter(flat)
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_29__iter__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_29__iter__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_28__iter__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_28__iter__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self) {
  PyObject *__pyx_v_flat = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("__iter__", 0);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":462
 * 
 *     def __iter__(self):
 *         flat = self.flatten()             # <<<<<<<<<<<<<<
 *         return iter(flat)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_flatten); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 462, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 462, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_flat = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":463
 *     def __iter__(self):
 *         flat = self.flatten()
 *         return iter(flat)             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyObject_GetIter(__pyx_v_flat); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 463, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":461
 *         self._update_bounds()
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         flat = self.flatten()
 *         return iter(flat)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_flat);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pxd":27
 * cdef class IntervalTreeNode(object):
 *     cdef:
 *         public double center             # <<<<<<<<<<<<<<
 *         public double start
 *         public double end
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6center_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6center_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6center___get__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6center___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->center); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.center.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6center_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6center_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6center_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6center_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 27, __pyx_L1_error)
  __pyx_v_self->center = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.center.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pxd":28
 *     cdef:
 *         public double center
 *         public double start             # <<<<<<<<<<<<<<
 *         public double end
 *         public IntervalTreeNode left
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5start_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5start_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5start___get__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5start___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->start); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.start.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5start_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5start_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5start_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5start_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 28, __pyx_L1_error)
  __pyx_v_self->start = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.start.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pxd":29
 *         public double center
 *         public double start
 *         public double end             # <<<<<<<<<<<<<<
 *         public IntervalTreeNode left
 *         public IntervalTreeNode right
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_3end_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_3end_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_3end___get__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_3end___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->end); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.end.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_3end_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_3end_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_3end_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_3end_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 29, __pyx_L1_error)
  __pyx_v_self->end = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.end.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pxd":30
 *         public double start
 *         public double end
 *         public IntervalTreeNode left             # <<<<<<<<<<<<<<
 *         public IntervalTreeNode right
 *         public list contained
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_4left_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_4left_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_4left___get__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_4left___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->left));
  __pyx_r = ((PyObject *)__pyx_v_self->left);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_4left_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_4left_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_4left_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_4left_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode))))) __PYX_ERR(1, 30, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->left);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->left));
  __pyx_v_self->left = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.left.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_4left_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_4left_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_4left_4__del__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_4left_4__del__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->left);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->left));
  __pyx_v_self->left = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pxd":31
 *         public double end
 *         public IntervalTreeNode left
 *         public IntervalTreeNode right             # <<<<<<<<<<<<<<
 *         public list contained
 *         public int level
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5right_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5right_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5right___get__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5right___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->right));
  __pyx_r = ((PyObject *)__pyx_v_self->right);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5right_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5right_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5right_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5right_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode))))) __PYX_ERR(1, 31, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->right);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->right));
  __pyx_v_self->right = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.right.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5right_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5right_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5right_4__del__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5right_4__del__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->right);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->right));
  __pyx_v_self->right = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pxd":32
 *         public IntervalTreeNode left
 *         public IntervalTreeNode right
 *         public list contained             # <<<<<<<<<<<<<<
 *         public int level
 *         public IntervalTreeNode parent
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_9contained_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_9contained_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_9contained___get__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_9contained___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->contained);
  __pyx_r = __pyx_v_self->contained;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_9contained_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_9contained_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_9contained_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_9contained_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(PyList_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(1, 32, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->contained);
  __Pyx_DECREF(__pyx_v_self->contained);
  __pyx_v_self->contained = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.contained.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_9contained_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_9contained_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_9contained_4__del__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_9contained_4__del__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->contained);
  __Pyx_DECREF(__pyx_v_self->contained);
  __pyx_v_self->contained = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pxd":33
 *         public IntervalTreeNode right
 *         public list contained
 *         public int level             # <<<<<<<<<<<<<<
 *         public IntervalTreeNode parent
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5level_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5level_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5level___get__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5level___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->level); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.level.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5level_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5level_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5level_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5level_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 33, __pyx_L1_error)
  __pyx_v_self->level = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.level.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pxd":34
 *         public list contained
 *         public int level
 *         public IntervalTreeNode parent             # <<<<<<<<<<<<<<
 * 
 *     cpdef list contains_point(self, double x)
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6parent_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6parent_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6parent___get__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6parent___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->parent));
  __pyx_r = ((PyObject *)__pyx_v_self->parent);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6parent_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6parent_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6parent_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6parent_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode))))) __PYX_ERR(1, 34, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->parent);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->parent));
  __pyx_v_self->parent = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.parent.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6parent_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6parent_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6parent_4__del__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6parent_4__del__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->parent);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->parent));
  __pyx_v_self->parent = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_31__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_30__reduce_cython__[] = "IntervalTreeNode.__reduce_cython__(self)";
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_31__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_30__reduce_cython__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_30__reduce_cython__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.center, self.contained, self.end, self.left, self.level, self.parent, self.right, self.start)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->center); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->end); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_self->level); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyFloat_FromDouble(__pyx_v_self->start); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(8); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->contained);
  __Pyx_GIVEREF(__pyx_v_self->contained);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_v_self->contained);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_2);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->left));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->left));
  PyTuple_SET_ITEM(__pyx_t_5, 3, ((PyObject *)__pyx_v_self->left));
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 4, __pyx_t_3);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->parent));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->parent));
  PyTuple_SET_ITEM(__pyx_t_5, 5, ((PyObject *)__pyx_v_self->parent));
  __Pyx_INCREF(((PyObject *)__pyx_v_self->right));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->right));
  PyTuple_SET_ITEM(__pyx_t_5, 6, ((PyObject *)__pyx_v_self->right));
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 7, __pyx_t_4);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.center, self.contained, self.end, self.left, self.level, self.parent, self.right, self.start)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_5 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_v__dict = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "(tree fragment)":7
 *     state = (self.center, self.contained, self.end, self.left, self.level, self.parent, self.right, self.start)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_6 = (__pyx_v__dict != Py_None);
  __pyx_t_7 = (__pyx_t_6 != 0);
  if (__pyx_t_7) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v__dict);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.contained is not None or self.left is not None or self.parent is not None or self.right is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.center, self.contained, self.end, self.left, self.level, self.parent, self.right, self.start)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.contained is not None or self.left is not None or self.parent is not None or self.right is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_IntervalTreeNode, (type(self), 0x262a9ad, None), state
 */
  /*else*/ {
    __pyx_t_6 = (__pyx_v_self->contained != ((PyObject*)Py_None));
    __pyx_t_8 = (__pyx_t_6 != 0);
    if (!__pyx_t_8) {
    } else {
      __pyx_t_7 = __pyx_t_8;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_8 = (((PyObject *)__pyx_v_self->left) != Py_None);
    __pyx_t_6 = (__pyx_t_8 != 0);
    if (!__pyx_t_6) {
    } else {
      __pyx_t_7 = __pyx_t_6;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_6 = (((PyObject *)__pyx_v_self->parent) != Py_None);
    __pyx_t_8 = (__pyx_t_6 != 0);
    if (!__pyx_t_8) {
    } else {
      __pyx_t_7 = __pyx_t_8;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_8 = (((PyObject *)__pyx_v_self->right) != Py_None);
    __pyx_t_6 = (__pyx_t_8 != 0);
    __pyx_t_7 = __pyx_t_6;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_7;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.contained is not None or self.left is not None or self.parent is not None or self.right is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_IntervalTreeNode, (type(self), 0x262a9ad, None), state
 *     else:
 */
  __pyx_t_7 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_7) {

    /* "(tree fragment)":13
 *         use_setstate = self.contained is not None or self.left is not None or self.parent is not None or self.right is not None
 *     if use_setstate:
 *         return __pyx_unpickle_IntervalTreeNode, (type(self), 0x262a9ad, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_IntervalTreeNode, (type(self), 0x262a9ad, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_IntervalTreeNode); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_40020397);
    __Pyx_GIVEREF(__pyx_int_40020397);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_int_40020397);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_5, 2, Py_None);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_5);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_state);
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.contained is not None or self.left is not None or self.parent is not None or self.right is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_IntervalTreeNode, (type(self), 0x262a9ad, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_IntervalTreeNode, (type(self), 0x262a9ad, None), state
 *     else:
 *         return __pyx_unpickle_IntervalTreeNode, (type(self), 0x262a9ad, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_IntervalTreeNode__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_IntervalTreeNode); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_40020397);
    __Pyx_GIVEREF(__pyx_int_40020397);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_int_40020397);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_5);
    __pyx_t_3 = 0;
    __pyx_t_5 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_IntervalTreeNode, (type(self), 0x262a9ad, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_IntervalTreeNode__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_33__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_32__setstate_cython__[] = "IntervalTreeNode.__setstate_cython__(self, __pyx_state)";
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_33__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_32__setstate_cython__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_32__setstate_cython__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_IntervalTreeNode, (type(self), 0x262a9ad, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_IntervalTreeNode__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals___pyx_unpickle_IntervalTreeNode__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_IntervalTreeNode, (type(self), 0x262a9ad, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_IntervalTreeNode__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_SpanningMixin(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_3__pyx_unpickle_SpanningMixin(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_2__pyx_unpickle_SpanningMixin[] = "__pyx_unpickle_SpanningMixin(__pyx_type, long __pyx_checksum, __pyx_state)";
static PyMethodDef __pyx_mdef_12ms_deisotope_2_c_23peak_dependency_network_9intervals_3__pyx_unpickle_SpanningMixin = {"__pyx_unpickle_SpanningMixin", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_3__pyx_unpickle_SpanningMixin, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_2__pyx_unpickle_SpanningMixin};
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_3__pyx_unpickle_SpanningMixin(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_SpanningMixin (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_SpanningMixin", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_SpanningMixin", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_SpanningMixin") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_SpanningMixin", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.__pyx_unpickle_SpanningMixin", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_2__pyx_unpickle_SpanningMixin(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_2__pyx_unpickle_SpanningMixin(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__pyx_unpickle_SpanningMixin", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xccc3210:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xccc3210 = (end, start))" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0xccc3210) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xccc3210:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xccc3210 = (end, start))" % __pyx_checksum)
 *     __pyx_result = SpanningMixin.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, -1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0xccc3210:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xccc3210 = (end, start))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = SpanningMixin.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0xcc, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xccc3210:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xccc3210 = (end, start))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xccc3210 = (end, start))" % __pyx_checksum)
 *     __pyx_result = SpanningMixin.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_SpanningMixin__set_state(<SpanningMixin> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xccc3210 = (end, start))" % __pyx_checksum)
 *     __pyx_result = SpanningMixin.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_SpanningMixin__set_state(<SpanningMixin> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = SpanningMixin.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_SpanningMixin__set_state(<SpanningMixin> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_SpanningMixin__set_state(SpanningMixin __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals___pyx_unpickle_SpanningMixin__set_state(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xccc3210 = (end, start))" % __pyx_checksum)
 *     __pyx_result = SpanningMixin.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_SpanningMixin__set_state(<SpanningMixin> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_SpanningMixin__set_state(<SpanningMixin> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_SpanningMixin__set_state(SpanningMixin __pyx_result, tuple __pyx_state):
 *     __pyx_result.end = __pyx_state[0]; __pyx_result.start = __pyx_state[1]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_SpanningMixin(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.__pyx_unpickle_SpanningMixin", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_SpanningMixin__set_state(<SpanningMixin> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_SpanningMixin__set_state(SpanningMixin __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.end = __pyx_state[0]; __pyx_result.start = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals___pyx_unpickle_SpanningMixin__set_state(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  double __pyx_t_2;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  __Pyx_RefNannySetupContext("__pyx_unpickle_SpanningMixin__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_SpanningMixin__set_state(SpanningMixin __pyx_result, tuple __pyx_state):
 *     __pyx_result.end = __pyx_state[0]; __pyx_result.start = __pyx_state[1]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->end = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->start = __pyx_t_2;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_SpanningMixin__set_state(SpanningMixin __pyx_result, tuple __pyx_state):
 *     __pyx_result.end = __pyx_state[0]; __pyx_result.start = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_4 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_5 = ((__pyx_t_4 > 2) != 0);
  if (__pyx_t_5) {
  } else {
    __pyx_t_3 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_5 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_6 = (__pyx_t_5 != 0);
  __pyx_t_3 = __pyx_t_6;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_3) {

    /* "(tree fragment)":14
 *     __pyx_result.end = __pyx_state[0]; __pyx_result.start = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])             # <<<<<<<<<<<<<<
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_update); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_7 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
      }
    }
    __pyx_t_1 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_9, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_7);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_SpanningMixin__set_state(SpanningMixin __pyx_result, tuple __pyx_state):
 *     __pyx_result.end = __pyx_state[0]; __pyx_result.start = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_SpanningMixin__set_state(<SpanningMixin> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_SpanningMixin__set_state(SpanningMixin __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.end = __pyx_state[0]; __pyx_result.start = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.__pyx_unpickle_SpanningMixin__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Interval(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_5__pyx_unpickle_Interval(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_4__pyx_unpickle_Interval[] = "__pyx_unpickle_Interval(__pyx_type, long __pyx_checksum, __pyx_state)";
static PyMethodDef __pyx_mdef_12ms_deisotope_2_c_23peak_dependency_network_9intervals_5__pyx_unpickle_Interval = {"__pyx_unpickle_Interval", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_5__pyx_unpickle_Interval, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_4__pyx_unpickle_Interval};
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_5__pyx_unpickle_Interval(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Interval (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Interval", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Interval", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_Interval") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Interval", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.__pyx_unpickle_Interval", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_4__pyx_unpickle_Interval(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_4__pyx_unpickle_Interval(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Interval", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xeace574:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xeace574 = (data, end, members, start))" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0xeace574) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xeace574:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xeace574 = (data, end, members, start))" % __pyx_checksum)
 *     __pyx_result = Interval.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, -1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0xeace574:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xeace574 = (data, end, members, start))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = Interval.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0xea, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xeace574:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xeace574 = (data, end, members, start))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xeace574 = (data, end, members, start))" % __pyx_checksum)
 *     __pyx_result = Interval.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Interval__set_state(<Interval> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xeace574 = (data, end, members, start))" % __pyx_checksum)
 *     __pyx_result = Interval.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Interval__set_state(<Interval> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = Interval.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Interval__set_state(<Interval> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Interval__set_state(Interval __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals___pyx_unpickle_Interval__set_state(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xeace574 = (data, end, members, start))" % __pyx_checksum)
 *     __pyx_result = Interval.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Interval__set_state(<Interval> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Interval__set_state(<Interval> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Interval__set_state(Interval __pyx_result, tuple __pyx_state):
 *     __pyx_result.data = __pyx_state[0]; __pyx_result.end = __pyx_state[1]; __pyx_result.members = __pyx_state[2]; __pyx_result.start = __pyx_state[3]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Interval(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.__pyx_unpickle_Interval", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Interval__set_state(<Interval> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Interval__set_state(Interval __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.data = __pyx_state[0]; __pyx_result.end = __pyx_state[1]; __pyx_result.members = __pyx_state[2]; __pyx_result.start = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals___pyx_unpickle_Interval__set_state(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  double __pyx_t_2;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Interval__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Interval__set_state(Interval __pyx_result, tuple __pyx_state):
 *     __pyx_result.data = __pyx_state[0]; __pyx_result.end = __pyx_state[1]; __pyx_result.members = __pyx_state[2]; __pyx_result.start = __pyx_state[3]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[4])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyDict_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->data);
  __Pyx_DECREF(__pyx_v___pyx_result->data);
  __pyx_v___pyx_result->data = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.end = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->members);
  __Pyx_DECREF(__pyx_v___pyx_result->members);
  __pyx_v___pyx_result->members = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.start = __pyx_t_2;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Interval__set_state(Interval __pyx_result, tuple __pyx_state):
 *     __pyx_result.data = __pyx_state[0]; __pyx_result.end = __pyx_state[1]; __pyx_result.members = __pyx_state[2]; __pyx_result.start = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[4])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_4 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_5 = ((__pyx_t_4 > 4) != 0);
  if (__pyx_t_5) {
  } else {
    __pyx_t_3 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_5 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_6 = (__pyx_t_5 != 0);
  __pyx_t_3 = __pyx_t_6;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_3) {

    /* "(tree fragment)":14
 *     __pyx_result.data = __pyx_state[0]; __pyx_result.end = __pyx_state[1]; __pyx_result.members = __pyx_state[2]; __pyx_result.start = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[4])             # <<<<<<<<<<<<<<
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_update); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_7 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
      }
    }
    __pyx_t_1 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_9, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_7);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Interval__set_state(Interval __pyx_result, tuple __pyx_state):
 *     __pyx_result.data = __pyx_state[0]; __pyx_result.end = __pyx_state[1]; __pyx_result.members = __pyx_state[2]; __pyx_result.start = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[4])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Interval__set_state(<Interval> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Interval__set_state(Interval __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.data = __pyx_state[0]; __pyx_result.end = __pyx_state[1]; __pyx_result.members = __pyx_state[2]; __pyx_result.start = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.__pyx_unpickle_Interval__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_IntervalTreeNode(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_7__pyx_unpickle_IntervalTreeNode(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_6__pyx_unpickle_IntervalTreeNode[] = "__pyx_unpickle_IntervalTreeNode(__pyx_type, long __pyx_checksum, __pyx_state)";
static PyMethodDef __pyx_mdef_12ms_deisotope_2_c_23peak_dependency_network_9intervals_7__pyx_unpickle_IntervalTreeNode = {"__pyx_unpickle_IntervalTreeNode", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_7__pyx_unpickle_IntervalTreeNode, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_6__pyx_unpickle_IntervalTreeNode};
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_7__pyx_unpickle_IntervalTreeNode(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_IntervalTreeNode (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_IntervalTreeNode", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_IntervalTreeNode", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_IntervalTreeNode") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_IntervalTreeNode", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.__pyx_unpickle_IntervalTreeNode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_6__pyx_unpickle_IntervalTreeNode(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_6__pyx_unpickle_IntervalTreeNode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__pyx_unpickle_IntervalTreeNode", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x262a9ad:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x262a9ad = (center, contained, end, left, level, parent, right, start))" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0x262a9ad) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x262a9ad:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x262a9ad = (center, contained, end, left, level, parent, right, start))" % __pyx_checksum)
 *     __pyx_result = IntervalTreeNode.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, -1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0x262a9ad:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x262a9ad = (center, contained, end, left, level, parent, right, start))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = IntervalTreeNode.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0x26, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x262a9ad:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x262a9ad = (center, contained, end, left, level, parent, right, start))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x262a9ad = (center, contained, end, left, level, parent, right, start))" % __pyx_checksum)
 *     __pyx_result = IntervalTreeNode.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_IntervalTreeNode__set_state(<IntervalTreeNode> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x262a9ad = (center, contained, end, left, level, parent, right, start))" % __pyx_checksum)
 *     __pyx_result = IntervalTreeNode.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_IntervalTreeNode__set_state(<IntervalTreeNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = IntervalTreeNode.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_IntervalTreeNode__set_state(<IntervalTreeNode> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_IntervalTreeNode__set_state(IntervalTreeNode __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals___pyx_unpickle_IntervalTreeNode__set_state(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x262a9ad = (center, contained, end, left, level, parent, right, start))" % __pyx_checksum)
 *     __pyx_result = IntervalTreeNode.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_IntervalTreeNode__set_state(<IntervalTreeNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_IntervalTreeNode__set_state(<IntervalTreeNode> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_IntervalTreeNode__set_state(IntervalTreeNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center = __pyx_state[0]; __pyx_result.contained = __pyx_state[1]; __pyx_result.end = __pyx_state[2]; __pyx_result.left = __pyx_state[3]; __pyx_result.level = __pyx_state[4]; __pyx_result.parent = __pyx_state[5]; __pyx_result.right = __pyx_state[6]; __pyx_result.start = __pyx_state[7]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_IntervalTreeNode(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.__pyx_unpickle_IntervalTreeNode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_IntervalTreeNode__set_state(<IntervalTreeNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_IntervalTreeNode__set_state(IntervalTreeNode __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.center = __pyx_state[0]; __pyx_result.contained = __pyx_state[1]; __pyx_result.end = __pyx_state[2]; __pyx_result.left = __pyx_state[3]; __pyx_result.level = __pyx_state[4]; __pyx_result.parent = __pyx_state[5]; __pyx_result.right = __pyx_state[6]; __pyx_result.start = __pyx_state[7]
 *     if len(__pyx_state) > 8 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals___pyx_unpickle_IntervalTreeNode__set_state(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  double __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_RefNannySetupContext("__pyx_unpickle_IntervalTreeNode__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_IntervalTreeNode__set_state(IntervalTreeNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center = __pyx_state[0]; __pyx_result.contained = __pyx_state[1]; __pyx_result.end = __pyx_state[2]; __pyx_result.left = __pyx_state[3]; __pyx_result.level = __pyx_state[4]; __pyx_result.parent = __pyx_state[5]; __pyx_result.right = __pyx_state[6]; __pyx_result.start = __pyx_state[7]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 8 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[8])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->center = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyList_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->contained);
  __Pyx_DECREF(__pyx_v___pyx_result->contained);
  __pyx_v___pyx_result->contained = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->end = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->left);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->left));
  __pyx_v___pyx_result->left = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->level = __pyx_t_3;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->parent);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->parent));
  __pyx_v___pyx_result->parent = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->right);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->right));
  __pyx_v___pyx_result->right = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->start = __pyx_t_2;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_IntervalTreeNode__set_state(IntervalTreeNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center = __pyx_state[0]; __pyx_result.contained = __pyx_state[1]; __pyx_result.end = __pyx_state[2]; __pyx_result.left = __pyx_state[3]; __pyx_result.level = __pyx_state[4]; __pyx_result.parent = __pyx_state[5]; __pyx_result.right = __pyx_state[6]; __pyx_result.start = __pyx_state[7]
 *     if len(__pyx_state) > 8 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[8])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_5 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_6 = ((__pyx_t_5 > 8) != 0);
  if (__pyx_t_6) {
  } else {
    __pyx_t_4 = __pyx_t_6;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_6 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_7 = (__pyx_t_6 != 0);
  __pyx_t_4 = __pyx_t_7;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_4) {

    /* "(tree fragment)":14
 *     __pyx_result.center = __pyx_state[0]; __pyx_result.contained = __pyx_state[1]; __pyx_result.end = __pyx_state[2]; __pyx_result.left = __pyx_state[3]; __pyx_result.level = __pyx_state[4]; __pyx_result.parent = __pyx_state[5]; __pyx_result.right = __pyx_state[6]; __pyx_result.start = __pyx_state[7]
 *     if len(__pyx_state) > 8 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[8])             # <<<<<<<<<<<<<<
 */
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_update); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_8 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
      }
    }
    __pyx_t_1 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_10, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_8);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_IntervalTreeNode__set_state(IntervalTreeNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center = __pyx_state[0]; __pyx_result.contained = __pyx_state[1]; __pyx_result.end = __pyx_state[2]; __pyx_result.left = __pyx_state[3]; __pyx_result.level = __pyx_state[4]; __pyx_result.parent = __pyx_state[5]; __pyx_result.right = __pyx_state[6]; __pyx_result.start = __pyx_state[7]
 *     if len(__pyx_state) > 8 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[8])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_IntervalTreeNode__set_state(<IntervalTreeNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_IntervalTreeNode__set_state(IntervalTreeNode __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.center = __pyx_state[0]; __pyx_result.contained = __pyx_state[1]; __pyx_result.end = __pyx_state[2]; __pyx_result.left = __pyx_state[3]; __pyx_result.level = __pyx_state[4]; __pyx_result.parent = __pyx_state[5]; __pyx_result.right = __pyx_state[6]; __pyx_result.start = __pyx_state[7]
 *     if len(__pyx_state) > 8 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.__pyx_unpickle_IntervalTreeNode__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin __pyx_vtable_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin;

static PyObject *__pyx_tp_new_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *)o);
  p->__pyx_vtab = __pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin;
  return o;
}

static void __pyx_tp_dealloc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  (*Py_TYPE(o)->tp_free)(o);
}

static PyObject *__pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_start(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_5start_1__get__(o);
}

static int __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_start(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_5start_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_end(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_3end_1__get__(o);
}

static int __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_end(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_3end_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin[] = {
  {"contains", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_5contains, METH_O, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_4contains},
  {"overlaps", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_7overlaps, METH_O, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_6overlaps},
  {"overlap_size", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_9overlap_size, METH_O, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_8overlap_size},
  {"is_contained_in_interval", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_11is_contained_in_interval, METH_O, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_10is_contained_in_interval},
  {"contains_interval", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_13contains_interval, METH_O, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_12contains_interval},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_15__reduce_cython__, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_14__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_17__setstate_cython__, METH_O, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_16__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin[] = {
  {(char *)"start", __pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_start, __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_start, (char *)"start: 'double'", 0},
  {(char *)"end", __pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_end, __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_end, (char *)"end: 'double'", 0},
  {0, 0, 0, 0, 0}
};

static PySequenceMethods __pyx_tp_as_sequence_SpanningMixin = {
  0, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  0, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_3__contains__, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyTypeObject __pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin = {
  PyVarObject_HEAD_INIT(0, 0)
  "ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin", /*tp_name*/
  sizeof(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_SpanningMixin, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  "SpanningMixin(start, end)\nProvides methods for checking whether an entity\n    which has a defined start and end point over a single\n    dimension contains or overlaps with another entity in\n    that same dimension.\n    ", /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval __pyx_vtable_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval;

static PyObject *__pyx_tp_new_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *p;
  PyObject *o = __pyx_tp_new_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin*)__pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval;
  p->members = Py_None; Py_INCREF(Py_None);
  p->data = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval(PyObject *o) {
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *p = (struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->members);
  Py_CLEAR(p->data);
  #if CYTHON_USE_TYPE_SLOTS
  if (PyType_IS_GC(Py_TYPE(o)->tp_base))
  #endif
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin(o);
}

static int __pyx_tp_traverse_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *p = (struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *)o;
  e = ((likely(__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin)) ? ((__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin->tp_traverse) ? __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval)); if (e) return e;
  if (p->members) {
    e = (*v)(p->members, a); if (e) return e;
  }
  if (p->data) {
    e = (*v)(p->data, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *p = (struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *)o;
  if (likely(__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin)) { if (__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin->tp_clear) __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval);
  tmp = ((PyObject*)p->members);
  p->members = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->data);
  p->data = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}
static PyObject *__pyx_sq_item_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static PyObject *__pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_members(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_7members_1__get__(o);
}

static int __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_members(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_7members_3__set__(o, v);
  }
  else {
    return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_7members_5__del__(o);
  }
}

static PyObject *__pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_data(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_4data_1__get__(o);
}

static int __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_data(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_4data_3__set__(o, v);
  }
  else {
    return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_4data_5__del__(o);
  }
}

static PyMethodDef __pyx_methods_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_11__reduce_cython__, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_10__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_13__setstate_cython__, METH_O, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_12__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval[] = {
  {(char *)"members", __pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_members, __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_members, (char *)"members: object", 0},
  {(char *)"data", __pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_data, __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_data, (char *)"data: dict", 0},
  {0, 0, 0, 0, 0}
};

static PySequenceMethods __pyx_tp_as_sequence_Interval = {
  __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_7__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  #if CYTHON_COMPILING_IN_PYPY
  __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_3__contains__, /*sq_contains*/
  #else
  0, /*sq_contains*/
  #endif
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_Interval = {
  __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_7__len__, /*mp_length*/
  __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_5__getitem__, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval = {
  PyVarObject_HEAD_INIT(0, 0)
  "ms_deisotope._c.peak_dependency_network.intervals.Interval", /*tp_name*/
  sizeof(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_Interval, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_Interval, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "Interval(start, end, members=None, **kwargs)", /*tp_doc*/
  __pyx_tp_traverse_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval, /*tp_traverse*/
  __pyx_tp_clear_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_9__iter__, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode __pyx_vtable_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode;

static PyObject *__pyx_tp_new_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)o);
  p->__pyx_vtab = __pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode;
  p->left = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)Py_None); Py_INCREF(Py_None);
  p->right = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)Py_None); Py_INCREF(Py_None);
  p->contained = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->parent = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode(PyObject *o) {
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *p = (struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->left);
  Py_CLEAR(p->right);
  Py_CLEAR(p->contained);
  Py_CLEAR(p->parent);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *p = (struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)o;
  if (p->left) {
    e = (*v)(((PyObject *)p->left), a); if (e) return e;
  }
  if (p->right) {
    e = (*v)(((PyObject *)p->right), a); if (e) return e;
  }
  if (p->contained) {
    e = (*v)(p->contained, a); if (e) return e;
  }
  if (p->parent) {
    e = (*v)(((PyObject *)p->parent), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *p = (struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)o;
  tmp = ((PyObject*)p->left);
  p->left = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->right);
  p->right = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->contained);
  p->contained = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->parent);
  p->parent = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_center(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6center_1__get__(o);
}

static int __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_center(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6center_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_start(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5start_1__get__(o);
}

static int __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_start(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5start_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_end(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_3end_1__get__(o);
}

static int __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_end(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_3end_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_left(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_4left_1__get__(o);
}

static int __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_left(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_4left_3__set__(o, v);
  }
  else {
    return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_4left_5__del__(o);
  }
}

static PyObject *__pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_right(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5right_1__get__(o);
}

static int __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_right(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5right_3__set__(o, v);
  }
  else {
    return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5right_5__del__(o);
  }
}

static PyObject *__pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_contained(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_9contained_1__get__(o);
}

static int __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_contained(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_9contained_3__set__(o, v);
  }
  else {
    return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_9contained_5__del__(o);
  }
}

static PyObject *__pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_level(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5level_1__get__(o);
}

static int __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_level(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5level_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_parent(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6parent_1__get__(o);
}

static int __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_parent(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6parent_3__set__(o, v);
  }
  else {
    return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6parent_5__del__(o);
  }
}

static PyMethodDef __pyx_methods_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode[] = {
  {"contains_point", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_3contains_point, METH_O, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_2contains_point},
  {"_overlaps_interval", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5_overlaps_interval, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_4_overlaps_interval},
  {"overlaps", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_7overlaps, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6overlaps},
  {"_eq", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_13_eq, METH_O, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_12_eq},
  {"__diagnostic_eq__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_17__diagnostic_eq__, METH_O, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_16__diagnostic_eq__},
  {"build", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_19build, METH_O, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_18build},
  {"_update_bounds", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_21_update_bounds, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_20_update_bounds},
  {"insert", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_23insert, METH_O, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_22insert},
  {"flatten", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_25flatten, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_24flatten},
  {"balance", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_27balance, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_26balance},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_31__reduce_cython__, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_30__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_33__setstate_cython__, METH_O, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_32__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode[] = {
  {(char *)"center", __pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_center, __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_center, (char *)"center: 'double'", 0},
  {(char *)"start", __pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_start, __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_start, (char *)"start: 'double'", 0},
  {(char *)"end", __pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_end, __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_end, (char *)"end: 'double'", 0},
  {(char *)"left", __pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_left, __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_left, (char *)"left: ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode", 0},
  {(char *)"right", __pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_right, __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_right, (char *)"right: ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode", 0},
  {(char *)"contained", __pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_contained, __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_contained, (char *)"contained: list", 0},
  {(char *)"level", __pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_level, __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_level, (char *)"level: 'int'", 0},
  {(char *)"parent", __pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_parent, __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_parent, (char *)"parent: ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode", 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode = {
  PyVarObject_HEAD_INIT(0, 0)
  "ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode", /*tp_name*/
  sizeof(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_9__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_15__hash__, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "IntervalTreeNode(double center, IntervalTreeNode left, list contained, IntervalTreeNode right, int level=0, IntervalTreeNode parent=None)", /*tp_doc*/
  __pyx_tp_traverse_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode, /*tp_traverse*/
  __pyx_tp_clear_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode, /*tp_clear*/
  __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_11__richcmp__, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_29__iter__, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {"intervals_containing_point", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_1intervals_containing_point, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_intervals_containing_point},
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_intervals(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_intervals},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "intervals",
    0, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_kp_s_, __pyx_k_, sizeof(__pyx_k_), 0, 0, 1, 0},
  {&__pyx_n_s_AttributeError, __pyx_k_AttributeError, sizeof(__pyx_k_AttributeError), 0, 0, 1, 1},
  {&__pyx_kp_s_Incompatible_checksums_s_vs_0x26, __pyx_k_Incompatible_checksums_s_vs_0x26, sizeof(__pyx_k_Incompatible_checksums_s_vs_0x26), 0, 0, 1, 0},
  {&__pyx_kp_s_Incompatible_checksums_s_vs_0xcc, __pyx_k_Incompatible_checksums_s_vs_0xcc, sizeof(__pyx_k_Incompatible_checksums_s_vs_0xcc), 0, 0, 1, 0},
  {&__pyx_kp_s_Incompatible_checksums_s_vs_0xea, __pyx_k_Incompatible_checksums_s_vs_0xea, sizeof(__pyx_k_Incompatible_checksums_s_vs_0xea), 0, 0, 1, 0},
  {&__pyx_n_s_Interval, __pyx_k_Interval, sizeof(__pyx_k_Interval), 0, 0, 1, 1},
  {&__pyx_n_s_IntervalTreeNode, __pyx_k_IntervalTreeNode, sizeof(__pyx_k_IntervalTreeNode), 0, 0, 1, 1},
  {&__pyx_kp_s_IntervalTreeNode_level_d_center, __pyx_k_IntervalTreeNode_level_d_center, sizeof(__pyx_k_IntervalTreeNode_level_d_center), 0, 0, 1, 0},
  {&__pyx_n_s_PickleError, __pyx_k_PickleError, sizeof(__pyx_k_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_SpanningMixin, __pyx_k_SpanningMixin, sizeof(__pyx_k_SpanningMixin), 0, 0, 1, 1},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_n_s_append, __pyx_k_append, sizeof(__pyx_k_append), 0, 0, 1, 1},
  {&__pyx_n_s_build, __pyx_k_build, sizeof(__pyx_k_build), 0, 0, 1, 1},
  {&__pyx_n_s_center, __pyx_k_center, sizeof(__pyx_k_center), 0, 0, 1, 1},
  {&__pyx_n_s_class, __pyx_k_class, sizeof(__pyx_k_class), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_collections, __pyx_k_collections, sizeof(__pyx_k_collections), 0, 0, 1, 1},
  {&__pyx_n_s_contained, __pyx_k_contained, sizeof(__pyx_k_contained), 0, 0, 1, 1},
  {&__pyx_n_s_contains, __pyx_k_contains, sizeof(__pyx_k_contains), 0, 0, 1, 1},
  {&__pyx_n_s_contains_interval, __pyx_k_contains_interval, sizeof(__pyx_k_contains_interval), 0, 0, 1, 1},
  {&__pyx_n_s_contains_point, __pyx_k_contains_point, sizeof(__pyx_k_contains_point), 0, 0, 1, 1},
  {&__pyx_n_s_deque, __pyx_k_deque, sizeof(__pyx_k_deque), 0, 0, 1, 1},
  {&__pyx_n_s_diagnostic_eq, __pyx_k_diagnostic_eq, sizeof(__pyx_k_diagnostic_eq), 0, 0, 1, 1},
  {&__pyx_n_s_dict, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
  {&__pyx_n_s_end, __pyx_k_end, sizeof(__pyx_k_end), 0, 0, 1, 1},
  {&__pyx_n_s_eq, __pyx_k_eq, sizeof(__pyx_k_eq), 0, 0, 1, 1},
  {&__pyx_n_s_file, __pyx_k_file, sizeof(__pyx_k_file), 0, 0, 1, 1},
  {&__pyx_n_s_flatten, __pyx_k_flatten, sizeof(__pyx_k_flatten), 0, 0, 1, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_inf, __pyx_k_inf, sizeof(__pyx_k_inf), 0, 0, 1, 1},
  {&__pyx_n_s_insert, __pyx_k_insert, sizeof(__pyx_k_insert), 0, 0, 1, 1},
  {&__pyx_n_s_intervals, __pyx_k_intervals, sizeof(__pyx_k_intervals), 0, 0, 1, 1},
  {&__pyx_n_s_is_contained_in_interval, __pyx_k_is_contained_in_interval, sizeof(__pyx_k_is_contained_in_interval), 0, 0, 1, 1},
  {&__pyx_kp_s_l, __pyx_k_l, sizeof(__pyx_k_l), 0, 0, 1, 0},
  {&__pyx_n_s_left, __pyx_k_left, sizeof(__pyx_k_left), 0, 0, 1, 1},
  {&__pyx_n_s_level, __pyx_k_level, sizeof(__pyx_k_level), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_members, __pyx_k_members, sizeof(__pyx_k_members), 0, 0, 1, 1},
  {&__pyx_n_s_ms_deisotope__c_peak_dependency, __pyx_k_ms_deisotope__c_peak_dependency, sizeof(__pyx_k_ms_deisotope__c_peak_dependency), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_new, __pyx_k_new, sizeof(__pyx_k_new), 0, 0, 1, 1},
  {&__pyx_n_s_node_contains, __pyx_k_node_contains, sizeof(__pyx_k_node_contains), 0, 0, 1, 1},
  {&__pyx_n_s_overlap_size, __pyx_k_overlap_size, sizeof(__pyx_k_overlap_size), 0, 0, 1, 1},
  {&__pyx_n_s_overlaps, __pyx_k_overlaps, sizeof(__pyx_k_overlaps), 0, 0, 1, 1},
  {&__pyx_n_s_overlaps_interval, __pyx_k_overlaps_interval, sizeof(__pyx_k_overlaps_interval), 0, 0, 1, 1},
  {&__pyx_n_s_parent, __pyx_k_parent, sizeof(__pyx_k_parent), 0, 0, 1, 1},
  {&__pyx_n_s_pickle, __pyx_k_pickle, sizeof(__pyx_k_pickle), 0, 0, 1, 1},
  {&__pyx_n_s_pop, __pyx_k_pop, sizeof(__pyx_k_pop), 0, 0, 1, 1},
  {&__pyx_n_s_print, __pyx_k_print, sizeof(__pyx_k_print), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_PickleError, __pyx_k_pyx_PickleError, sizeof(__pyx_k_pyx_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_checksum, __pyx_k_pyx_checksum, sizeof(__pyx_k_pyx_checksum), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_result, __pyx_k_pyx_result, sizeof(__pyx_k_pyx_result), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_state, __pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_type, __pyx_k_pyx_type, sizeof(__pyx_k_pyx_type), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_Interval, __pyx_k_pyx_unpickle_Interval, sizeof(__pyx_k_pyx_unpickle_Interval), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_IntervalTreeNode, __pyx_k_pyx_unpickle_IntervalTreeNode, sizeof(__pyx_k_pyx_unpickle_IntervalTreeNode), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_SpanningMixin, __pyx_k_pyx_unpickle_SpanningMixin, sizeof(__pyx_k_pyx_unpickle_SpanningMixin), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_kp_s_r, __pyx_k_r, sizeof(__pyx_k_r), 0, 0, 1, 0},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {&__pyx_n_s_right, __pyx_k_right, sizeof(__pyx_k_right), 0, 0, 1, 1},
  {&__pyx_kp_s_s_start_r_end_r_data_r, __pyx_k_s_start_r_end_r_data_r, sizeof(__pyx_k_s_start_r_end_r_data_r), 0, 0, 1, 0},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_start, __pyx_k_start, sizeof(__pyx_k_start), 0, 0, 1, 1},
  {&__pyx_kp_s_stringsource, __pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 0, 1, 0},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_update, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
  {&__pyx_n_s_update_bounds, __pyx_k_update_bounds, sizeof(__pyx_k_update_bounds), 0, 0, 1, 1},
  {&__pyx_n_s_x, __pyx_k_x, sizeof(__pyx_k_x), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 148, __pyx_L1_error)
  __pyx_builtin_AttributeError = __Pyx_GetBuiltinName(__pyx_n_s_AttributeError); if (!__pyx_builtin_AttributeError) __PYX_ERR(0, 295, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(0, 382, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "(tree fragment)":1
 * def __pyx_unpickle_SpanningMixin(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_tuple__2 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);
  __pyx_codeobj__3 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_SpanningMixin, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__3)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_tuple__4 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);
  __pyx_codeobj__5 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__4, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Interval, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__5)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_tuple__6 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);
  __pyx_codeobj__7 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_IntervalTreeNode, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__7)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  __pyx_umethod_PyList_Type_pop.type = (PyObject*)&PyList_Type;
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_40020397 = PyInt_FromLong(40020397L); if (unlikely(!__pyx_int_40020397)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_214708752 = PyInt_FromLong(214708752L); if (unlikely(!__pyx_int_214708752)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_246211956 = PyInt_FromLong(246211956L); if (unlikely(!__pyx_int_246211956)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_neg_1 = PyInt_FromLong(-1); if (unlikely(!__pyx_int_neg_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin = &__pyx_vtable_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin;
  __pyx_vtable_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin._contains = (int (*)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *, double))__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin__contains;
  __pyx_vtable_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin.contains = (int (*)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *, double, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_contains;
  __pyx_vtable_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin.overlaps = (int (*)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_overlaps;
  __pyx_vtable_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin.overlap_size = (double (*)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_overlap_size;
  __pyx_vtable_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin.is_contained_in_interval = (int (*)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_is_contained_in_interval;
  __pyx_vtable_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin.contains_interval = (int (*)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_contains_interval;
  if (PyType_Ready(&__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin) < 0) __PYX_ERR(0, 7, __pyx_L1_error)
  __pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin.tp_dictoffset && __pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #if CYTHON_COMPILING_IN_CPYTHON
  {
    PyObject *wrapper = PyObject_GetAttrString((PyObject *)&__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin, "__contains__"); if (unlikely(!wrapper)) __PYX_ERR(0, 7, __pyx_L1_error)
    if (Py_TYPE(wrapper) == &PyWrapperDescr_Type) {
      __pyx_wrapperbase_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_2__contains__ = *((PyWrapperDescrObject *)wrapper)->d_base;
      __pyx_wrapperbase_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_2__contains__.doc = __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_2__contains__;
      ((PyWrapperDescrObject *)wrapper)->d_base = &__pyx_wrapperbase_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_2__contains__;
    }
  }
  #endif
  if (__Pyx_SetVtable(__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin.tp_dict, __pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin) < 0) __PYX_ERR(0, 7, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_SpanningMixin, (PyObject *)&__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin) < 0) __PYX_ERR(0, 7, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin) < 0) __PYX_ERR(0, 7, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin = &__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin;
  __pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval = &__pyx_vtable_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval;
  __pyx_vtable_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval.__pyx_base = *__pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin;
  __pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval.tp_base = __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin;
  if (PyType_Ready(&__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval) < 0) __PYX_ERR(0, 120, __pyx_L1_error)
  __pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval.tp_dictoffset && __pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval.tp_dict, __pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval) < 0) __PYX_ERR(0, 120, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Interval, (PyObject *)&__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval) < 0) __PYX_ERR(0, 120, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval) < 0) __PYX_ERR(0, 120, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval = &__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval;
  __pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode = &__pyx_vtable_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode;
  __pyx_vtable_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode.contains_point = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *, double, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_contains_point;
  __pyx_vtable_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode._overlaps_interval = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *, double, double, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode__overlaps_interval;
  __pyx_vtable_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode.overlaps = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *, double, double, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_overlaps;
  __pyx_vtable_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode._eq = (int (*)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode__eq;
  if (PyType_Ready(&__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode) < 0) __PYX_ERR(0, 158, __pyx_L1_error)
  __pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode.tp_dictoffset && __pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode.tp_dict, __pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode) < 0) __PYX_ERR(0, 158, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_IntervalTreeNode, (PyObject *)&__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode) < 0) __PYX_ERR(0, 158, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode) < 0) __PYX_ERR(0, 158, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode = &__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(3, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#if PY_MAJOR_VERSION < 3
#ifdef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC void
#else
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#endif
#else
#ifdef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initintervals(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initintervals(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_intervals(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_intervals(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_intervals(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  double __pyx_t_3;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'intervals' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_intervals(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("intervals", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  #if CYTHON_COMPILING_IN_PYPY
  Py_INCREF(__pyx_b);
  #endif
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_ms_deisotope___c__peak_dependency_network__intervals) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "ms_deisotope._c.peak_dependency_network.intervals")) {
      if (unlikely(PyDict_SetItemString(modules, "ms_deisotope._c.peak_dependency_network.intervals", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely(__Pyx_modinit_type_init_code() != 0)) goto __pyx_L1_error;
  if (unlikely(__Pyx_modinit_type_import_code() != 0)) goto __pyx_L1_error;
  (void)__Pyx_modinit_variable_import_code();
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":4
 * from cpython.tuple cimport PyTuple_GetItem
 * 
 * from collections import deque             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_deque);
  __Pyx_GIVEREF(__pyx_n_s_deque);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_deque);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_collections, __pyx_t_1, -1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_deque); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_deque, __pyx_t_1) < 0) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":155
 * 
 * 
 * cdef double INF = float('inf')             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_AsDouble(__pyx_n_s_inf); if (unlikely(__pyx_t_3 == ((double)((double)-1)) && PyErr_Occurred())) __PYX_ERR(0, 155, __pyx_L1_error)
  __pyx_v_12ms_deisotope_2_c_23peak_dependency_network_9intervals_INF = __pyx_t_3;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":311
 * 
 *     @classmethod
 *     def build(cls, intervals):             # <<<<<<<<<<<<<<
 *         cdef:
 *             IntervalTreeNode root, parent, node, up
 */
  __Pyx_GetNameInClass(__pyx_t_2, (PyObject *)__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode, __pyx_n_s_build); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":310
 *             return result
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def build(cls, intervals):
 *         cdef:
 */
  __pyx_t_1 = __Pyx_Method_ClassMethod(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 310, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode->tp_dict, __pyx_n_s_build, __pyx_t_1) < 0) __PYX_ERR(0, 311, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  PyType_Modified(__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode);

  /* "(tree fragment)":1
 * def __pyx_unpickle_SpanningMixin(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_12ms_deisotope_2_c_23peak_dependency_network_9intervals_3__pyx_unpickle_SpanningMixin, NULL, __pyx_n_s_ms_deisotope__c_peak_dependency); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_SpanningMixin, __pyx_t_1) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_SpanningMixin__set_state(<SpanningMixin> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_SpanningMixin__set_state(SpanningMixin __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.end = __pyx_state[0]; __pyx_result.start = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_12ms_deisotope_2_c_23peak_dependency_network_9intervals_5__pyx_unpickle_Interval, NULL, __pyx_n_s_ms_deisotope__c_peak_dependency); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Interval, __pyx_t_1) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_IntervalTreeNode(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_12ms_deisotope_2_c_23peak_dependency_network_9intervals_7__pyx_unpickle_IntervalTreeNode, NULL, __pyx_n_s_ms_deisotope__c_peak_dependency); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_IntervalTreeNode, __pyx_t_1) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":1
 * from cpython.list cimport PyList_GetItem, PyList_Size             # <<<<<<<<<<<<<<
 * from cpython.tuple cimport PyTuple_GetItem
 * 
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_1) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init ms_deisotope._c.peak_dependency_network.intervals", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init ms_deisotope._c.peak_dependency_network.intervals");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_CheckExact(key)) || likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (PyUnicode_GET_SIZE(**name) != PyUnicode_GET_SIZE(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (PyUnicode_GET_SIZE(**argname) != PyUnicode_GET_SIZE(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, int nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (PyCFunction_GET_FLAGS(func) & METH_FASTCALL) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* WriteUnraisableException */
static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
#ifdef _MSC_VER
    else state = (PyGILState_STATE)-1;
#endif
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
    return 0;
}

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* GetAttr */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* GetAttr3 */
static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r = __Pyx_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
}

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject* index) {
    PyObject *runerr;
    Py_ssize_t key_value;
    PySequenceMethods *m = Py_TYPE(obj)->tp_as_sequence;
    if (unlikely(!(m && m->sq_item))) {
        PyErr_Format(PyExc_TypeError, "'%.200s' object is not subscriptable", Py_TYPE(obj)->tp_name);
        return NULL;
    }
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError, "cannot fit '%.200s' into an index-sized integer", Py_TYPE(index)->tp_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key) {
    PyMappingMethods *m = Py_TYPE(obj)->tp_as_mapping;
    if (likely(m && m->mp_subscript)) {
        return m->mp_subscript(obj, key);
    }
    return __Pyx_PyObject_GetIndex(obj, key);
}
#endif

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* PyObjectCallNoArg */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* PyObjectGetMethod */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (descr != NULL) {
        *method = descr;
        return 0;
    }
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(name));
#endif
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod0 */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
    PyObject *method = NULL, *result = NULL;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_CallOneArg(method, obj);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) goto bad;
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
}

/* UnpackUnboundCMethod */
static int __Pyx_TryUnpackUnboundCMethod(__Pyx_CachedCFunction* target) {
    PyObject *method;
    method = __Pyx_PyObject_GetAttrStr(target->type, *target->method_name);
    if (unlikely(!method))
        return -1;
    target->method = method;
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION >= 3
    if (likely(__Pyx_TypeCheck(method, &PyMethodDescr_Type)))
    #endif
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject*) method;
        target->func = descr->d_method->ml_meth;
        target->flag = descr->d_method->ml_flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_STACKLESS);
    }
#endif
    return 0;
}

/* CallUnboundCMethod0 */
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self) {
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_ASSUME_SAFE_MACROS
    args = PyTuple_New(1);
    if (unlikely(!args)) goto bad;
    Py_INCREF(self);
    PyTuple_SET_ITEM(args, 0, self);
#else
    args = PyTuple_Pack(1, self);
    if (unlikely(!args)) goto bad;
#endif
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    Py_DECREF(args);
bad:
    return result;
}

/* pop */
static CYTHON_INLINE PyObject* __Pyx__PyObject_Pop(PyObject* L) {
    if (Py_TYPE(L) == &PySet_Type) {
        return PySet_Pop(L);
    }
    return __Pyx_PyObject_CallMethod0(L, __pyx_n_s_pop);
}
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE PyObject* __Pyx_PyList_Pop(PyObject* L) {
    if (likely(PyList_GET_SIZE(L) > (((PyListObject*)L)->allocated >> 1))) {
        Py_SIZE(L) -= 1;
        return PyList_GET_ITEM(L, PyList_GET_SIZE(L));
    }
    return __Pyx_CallUnboundCMethod0(&__pyx_umethod_PyList_Type_pop, L);
}
#endif

/* BytesEquals */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
        #if CYTHON_PEP393_ENABLED
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
        #else
            hash1 = ((PyUnicodeObject*)s1)->hash;
            hash2 = ((PyUnicodeObject*)s2)->hash;
        #endif
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        #if PY_MAJOR_VERSION < 3
        Py_XDECREF(owned_ref);
        #endif
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* PyObjectCallMethod1 */
static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
}

/* append */
static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x) {
    if (likely(PyList_CheckExact(L))) {
        if (unlikely(__Pyx_PyList_Append(L, x) < 0)) return -1;
    } else {
        PyObject* retval = __Pyx_PyObject_CallMethod1(L, __pyx_n_s_append, x);
        if (unlikely(!retval))
            return -1;
        Py_DECREF(retval);
    }
    return 0;
}

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if (strchr(__Pyx_MODULE_NAME, '.')) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* HasAttr */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!__Pyx_PyBaseString_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_GetAttr(o, n);
    if (unlikely(!r)) {
        PyErr_Clear();
        return 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}

/* CallNextTpTraverse */
static int __Pyx_call_next_tp_traverse(PyObject* obj, visitproc v, void *a, traverseproc current_tp_traverse) {
    PyTypeObject* type = Py_TYPE(obj);
    while (type && type->tp_traverse != current_tp_traverse)
        type = type->tp_base;
    while (type && type->tp_traverse == current_tp_traverse)
        type = type->tp_base;
    if (type && type->tp_traverse)
        return type->tp_traverse(obj, v, a);
    return 0;
}

/* CallNextTpClear */
static void __Pyx_call_next_tp_clear(PyObject* obj, inquiry current_tp_clear) {
    PyTypeObject* type = Py_TYPE(obj);
    while (type && type->tp_clear != current_tp_clear)
        type = type->tp_base;
    while (type && type->tp_clear == current_tp_clear)
        type = type->tp_base;
    if (type && type->tp_clear)
        type->tp_clear(obj);
}

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetVTable */
static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* SetupReduce */
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStr(meth, __pyx_n_s_name);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    if (_PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate)) goto GOOD;
#else
    if (PyObject_HasAttr(type_obj, __pyx_n_s_getstate)) goto GOOD;
#endif
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_cython); if (unlikely(!reduce_cython)) goto BAD;
            ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto BAD;
            ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto BAD;
            setstate = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate_cython); if (unlikely(!setstate_cython)) goto BAD;
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto BAD;
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto GOOD;
BAD:
    if (!PyErr_Occurred())
        PyErr_Format(PyExc_RuntimeError, "Unable to initialize pickling for %s", ((PyTypeObject*)type_obj)->tp_name);
    ret = -1;
GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* pyobject_as_double */
static double __Pyx__PyObject_AsDouble(PyObject* obj) {
    PyObject* float_value;
#if !CYTHON_USE_TYPE_SLOTS
    float_value = PyNumber_Float(obj);  if ((0)) goto bad;
#else
    PyNumberMethods *nb = Py_TYPE(obj)->tp_as_number;
    if (likely(nb) && likely(nb->nb_float)) {
        float_value = nb->nb_float(obj);
        if (likely(float_value) && unlikely(!PyFloat_Check(float_value))) {
            PyErr_Format(PyExc_TypeError,
                "__float__ returned non-float (type %.200s)",
                Py_TYPE(float_value)->tp_name);
            Py_DECREF(float_value);
            goto bad;
        }
    } else if (PyUnicode_CheckExact(obj) || PyBytes_CheckExact(obj)) {
#if PY_MAJOR_VERSION >= 3
        float_value = PyFloat_FromString(obj);
#else
        float_value = PyFloat_FromString(obj, 0);
#endif
    } else {
        PyObject* args = PyTuple_New(1);
        if (unlikely(!args)) goto bad;
        PyTuple_SET_ITEM(args, 0, obj);
        float_value = PyObject_Call((PyObject*)&PyFloat_Type, args, 0);
        PyTuple_SET_ITEM(args, 0, 0);
        Py_DECREF(args);
    }
#endif
    if (likely(float_value)) {
        double value = PyFloat_AS_DOUBLE(float_value);
        Py_DECREF(float_value);
        return value;
    }
bad:
    return (double)-1;
}

/* ClassMethod */
static PyObject* __Pyx_Method_ClassMethod(PyObject *method) {
#if CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM <= 0x05080000
    if (PyObject_TypeCheck(method, &PyWrapperDescr_Type)) {
        return PyClassMethod_New(method);
    }
#else
#if CYTHON_COMPILING_IN_PYSTON || CYTHON_COMPILING_IN_PYPY
    if (PyMethodDescr_Check(method))
#else
    static PyTypeObject *methoddescr_type = NULL;
    if (methoddescr_type == NULL) {
       PyObject *meth = PyObject_GetAttrString((PyObject*)&PyList_Type, "append");
       if (!meth) return NULL;
       methoddescr_type = Py_TYPE(meth);
       Py_DECREF(meth);
    }
    if (__Pyx_TypeCheck(method, methoddescr_type))
#endif
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject *)method;
        #if PY_VERSION_HEX < 0x03020000
        PyTypeObject *d_type = descr->d_type;
        #else
        PyTypeObject *d_type = descr->d_common.d_type;
        #endif
        return PyDescr_NewClassMethod(d_type, descr->d_method);
    }
#endif
    else if (PyMethod_Check(method)) {
        return PyClassMethod_New(PyMethod_GET_FUNCTION(method));
    }
    else if (PyCFunction_Check(method)) {
        return PyClassMethod_New(method);
    }
#ifdef __Pyx_CyFunction_USED
    else if (__Pyx_CyFunction_Check(method)) {
        return PyClassMethod_New(method);
    }
#endif
    PyErr_SetString(PyExc_TypeError,
                   "Class-level classmethod() can only be called on "
                   "a method_descriptor or instance method.");
    return NULL;
}

/* GetNameInClass */
static PyObject *__Pyx_GetGlobalNameAfterAttributeLookup(PyObject *name) {
    PyObject *result;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    __Pyx_GetModuleGlobalNameUncached(result, name);
    return result;
}
static PyObject *__Pyx__GetNameInClass(PyObject *nmspace, PyObject *name) {
    PyObject *result;
    result = __Pyx_PyObject_GetAttrStr(nmspace, name);
    if (!result) {
        result = __Pyx_GetGlobalNameAfterAttributeLookup(name);
    }
    return result;
}

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, (size_t)new_max*sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* Print */
#if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION < 3
static PyObject *__Pyx_GetStdout(void) {
    PyObject *f = PySys_GetObject((char *)"stdout");
    if (!f) {
        PyErr_SetString(PyExc_RuntimeError, "lost sys.stdout");
    }
    return f;
}
static int __Pyx_Print(PyObject* f, PyObject *arg_tuple, int newline) {
    int i;
    if (!f) {
        if (!(f = __Pyx_GetStdout()))
            return -1;
    }
    Py_INCREF(f);
    for (i=0; i < PyTuple_GET_SIZE(arg_tuple); i++) {
        PyObject* v;
        if (PyFile_SoftSpace(f, 1)) {
            if (PyFile_WriteString(" ", f) < 0)
                goto error;
        }
        v = PyTuple_GET_ITEM(arg_tuple, i);
        if (PyFile_WriteObject(v, f, Py_PRINT_RAW) < 0)
            goto error;
        if (PyString_Check(v)) {
            char *s = PyString_AsString(v);
            Py_ssize_t len = PyString_Size(v);
            if (len > 0) {
                switch (s[len-1]) {
                    case ' ': break;
                    case '\f': case '\r': case '\n': case '\t': case '\v':
                        PyFile_SoftSpace(f, 0);
                        break;
                    default:  break;
                }
            }
        }
    }
    if (newline) {
        if (PyFile_WriteString("\n", f) < 0)
            goto error;
        PyFile_SoftSpace(f, 0);
    }
    Py_DECREF(f);
    return 0;
error:
    Py_DECREF(f);
    return -1;
}
#else
static int __Pyx_Print(PyObject* stream, PyObject *arg_tuple, int newline) {
    PyObject* kwargs = 0;
    PyObject* result = 0;
    PyObject* end_string;
    if (unlikely(!__pyx_print)) {
        __pyx_print = PyObject_GetAttr(__pyx_b, __pyx_n_s_print);
        if (!__pyx_print)
            return -1;
    }
    if (stream) {
        kwargs = PyDict_New();
        if (unlikely(!kwargs))
            return -1;
        if (unlikely(PyDict_SetItem(kwargs, __pyx_n_s_file, stream) < 0))
            goto bad;
        if (!newline) {
            end_string = PyUnicode_FromStringAndSize(" ", 1);
            if (unlikely(!end_string))
                goto bad;
            if (PyDict_SetItem(kwargs, __pyx_n_s_end, end_string) < 0) {
                Py_DECREF(end_string);
                goto bad;
            }
            Py_DECREF(end_string);
        }
    } else if (!newline) {
        if (unlikely(!__pyx_print_kwargs)) {
            __pyx_print_kwargs = PyDict_New();
            if (unlikely(!__pyx_print_kwargs))
                return -1;
            end_string = PyUnicode_FromStringAndSize(" ", 1);
            if (unlikely(!end_string))
                return -1;
            if (PyDict_SetItem(__pyx_print_kwargs, __pyx_n_s_end, end_string) < 0) {
                Py_DECREF(end_string);
                return -1;
            }
            Py_DECREF(end_string);
        }
        kwargs = __pyx_print_kwargs;
    }
    result = PyObject_Call(__pyx_print, arg_tuple, kwargs);
    if (unlikely(kwargs) && (kwargs != __pyx_print_kwargs))
        Py_DECREF(kwargs);
    if (!result)
        return -1;
    Py_DECREF(result);
    return 0;
bad:
    if (kwargs != __pyx_print_kwargs)
        Py_XDECREF(kwargs);
    return -1;
}
#endif

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntFromPy */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *x) {
    const size_t neg_one = (size_t) ((size_t) 0 - (size_t) 1), const_zero = (size_t) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(size_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(size_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (size_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(size_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 2 * PyLong_SHIFT) {
                            return (size_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 3 * PyLong_SHIFT) {
                            return (size_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 4 * PyLong_SHIFT) {
                            return (size_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (size_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(size_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(size_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(size_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(size_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) ((((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) ((((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) ((((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(size_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            size_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (size_t) -1;
        }
    } else {
        size_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (size_t) -1;
        val = __Pyx_PyInt_As_size_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to size_t");
    return (size_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to size_t");
    return (size_t) -1;
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* InitStrings */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
