/* Generated by Cython 0.29.13 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [
            "C:\\Users\\Asus\\Miniconda2\\envs\\py37\\lib\\site-packages\\numpy\\core\\include\\numpy\\arrayobject.h",
            "C:\\Users\\Asus\\Miniconda2\\envs\\py37\\lib\\site-packages\\numpy\\core\\include\\numpy\\ufuncobject.h"
        ],
        "include_dirs": [
            "C:\\Users\\Asus\\Miniconda2\\envs\\py37\\lib\\site-packages\\numpy\\core\\include",
            "C:\\Users\\Asus\\Miniconda2\\envs\\py37\\lib\\site-packages\\ms_peak_picker\\_c"
        ],
        "name": "ms_deisotope._c.deconvoluter_base",
        "sources": [
            "ms_deisotope/_c/deconvoluter_base.pyx"
        ]
    },
    "module_name": "ms_deisotope._c.deconvoluter_base"
}
END: Cython Metadata */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_13"
#define CYTHON_HEX_VERSION 0x001D0DF0
#define CYTHON_FUTURE_DIVISION 0
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
#if PY_VERSION_HEX >= 0x030800A4 && PY_VERSION_HEX < 0x030800B2
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, 0, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
  #define PyObject_Unicode             PyObject_Str
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? PyMethod_New(func, self) : (Py_INCREF(func), func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif


#define __PYX_ERR(f_index, lineno, Ln_error) \
{ \
  __pyx_filename = __pyx_f[f_index]; __pyx_lineno = lineno; __pyx_clineno = __LINE__; goto Ln_error; \
}

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__ms_deisotope___c__deconvoluter_base
#define __PYX_HAVE_API__ms_deisotope___c__deconvoluter_base
/* Early includes */
#include <string.h>
#include <stdio.h>
#include "numpy/arrayobject.h"
#include "numpy/ufuncobject.h"
#include <stdlib.h>
#include "pythread.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;

/* Header.proto */
#if !defined(CYTHON_CCOMPLEX)
  #if defined(__cplusplus)
    #define CYTHON_CCOMPLEX 1
  #elif defined(_Complex_I)
    #define CYTHON_CCOMPLEX 1
  #else
    #define CYTHON_CCOMPLEX 0
  #endif
#endif
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #include <complex>
  #else
    #include <complex.h>
  #endif
#endif
#if CYTHON_CCOMPLEX && !defined(__cplusplus) && defined(__sun__) && defined(__GNUC__)
  #undef _Complex_I
  #define _Complex_I 1.0fj
#endif


static const char *__pyx_f[] = {
  "ms_deisotope\\_c\\deconvoluter_base.pyx",
  "stringsource",
  "ms_deisotope\\_c\\deconvoluter_base.pxd",
  "__init__.pxd",
  "peak_set.pxd",
  "type.pxd",
  "bool.pxd",
  "complex.pxd",
  "peak_index.pxd",
  "composition.pxd",
  "isotopic_distribution.pxd",
  "ms_deisotope\\_c\\averagine.pxd",
  "ms_deisotope\\_c\\scoring.pxd",
  "ms_deisotope\\_c\\peak_set.pxd",
  "ms_deisotope\\_c\\peak_dependency_network\\intervals.pxd",
  "ms_deisotope\\_c\\peak_dependency_network\\subgraph.pxd",
  "ms_deisotope\\_c\\peak_dependency_network\\peak_network.pxd",
};
/* BufferFormatStructs.proto */
#define IS_UNSIGNED(type) (((type) -1) > 0)
struct __Pyx_StructField_;
#define __PYX_BUF_FLAGS_PACKED_STRUCT (1 << 0)
typedef struct {
  const char* name;
  struct __Pyx_StructField_* fields;
  size_t size;
  size_t arraysize[8];
  int ndim;
  char typegroup;
  char is_unsigned;
  int flags;
} __Pyx_TypeInfo;
typedef struct __Pyx_StructField_ {
  __Pyx_TypeInfo* type;
  const char* name;
  size_t offset;
} __Pyx_StructField;
typedef struct {
  __Pyx_StructField* field;
  size_t parent_offset;
} __Pyx_BufFmt_StackElem;
typedef struct {
  __Pyx_StructField root;
  __Pyx_BufFmt_StackElem* head;
  size_t fmt_offset;
  size_t new_count, enc_count;
  size_t struct_alignment;
  int is_complex;
  char enc_type;
  char new_packmode;
  char enc_packmode;
  char is_valid_array;
} __Pyx_BufFmt_Context;


/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":776
 * # in Cython to enable them only on the right systems.
 * 
 * ctypedef npy_int8       int8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 */
typedef npy_int8 __pyx_t_5numpy_int8_t;

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":777
 * 
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t
 */
typedef npy_int16 __pyx_t_5numpy_int16_t;

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":778
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int64      int64_t
 * #ctypedef npy_int96      int96_t
 */
typedef npy_int32 __pyx_t_5numpy_int32_t;

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":779
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_int96      int96_t
 * #ctypedef npy_int128     int128_t
 */
typedef npy_int64 __pyx_t_5numpy_int64_t;

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":783
 * #ctypedef npy_int128     int128_t
 * 
 * ctypedef npy_uint8      uint8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 */
typedef npy_uint8 __pyx_t_5numpy_uint8_t;

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":784
 * 
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t
 */
typedef npy_uint16 __pyx_t_5numpy_uint16_t;

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":785
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint64     uint64_t
 * #ctypedef npy_uint96     uint96_t
 */
typedef npy_uint32 __pyx_t_5numpy_uint32_t;

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":786
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_uint96     uint96_t
 * #ctypedef npy_uint128    uint128_t
 */
typedef npy_uint64 __pyx_t_5numpy_uint64_t;

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":790
 * #ctypedef npy_uint128    uint128_t
 * 
 * ctypedef npy_float32    float32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_float64    float64_t
 * #ctypedef npy_float80    float80_t
 */
typedef npy_float32 __pyx_t_5numpy_float32_t;

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":791
 * 
 * ctypedef npy_float32    float32_t
 * ctypedef npy_float64    float64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_float80    float80_t
 * #ctypedef npy_float128   float128_t
 */
typedef npy_float64 __pyx_t_5numpy_float64_t;

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":800
 * # The int types are mapped a bit surprising --
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t
 */
typedef npy_long __pyx_t_5numpy_int_t;

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":801
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   longlong_t
 * 
 */
typedef npy_longlong __pyx_t_5numpy_long_t;

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":802
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_ulong      uint_t
 */
typedef npy_longlong __pyx_t_5numpy_longlong_t;

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":804
 * ctypedef npy_longlong   longlong_t
 * 
 * ctypedef npy_ulong      uint_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t
 */
typedef npy_ulong __pyx_t_5numpy_uint_t;

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":805
 * 
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 */
typedef npy_ulonglong __pyx_t_5numpy_ulong_t;

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":806
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_intp       intp_t
 */
typedef npy_ulonglong __pyx_t_5numpy_ulonglong_t;

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":808
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 * ctypedef npy_intp       intp_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uintp      uintp_t
 * 
 */
typedef npy_intp __pyx_t_5numpy_intp_t;

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":809
 * 
 * ctypedef npy_intp       intp_t
 * ctypedef npy_uintp      uintp_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_double     float_t
 */
typedef npy_uintp __pyx_t_5numpy_uintp_t;

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":811
 * ctypedef npy_uintp      uintp_t
 * 
 * ctypedef npy_double     float_t             # <<<<<<<<<<<<<<
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t
 */
typedef npy_double __pyx_t_5numpy_float_t;

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":812
 * 
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longdouble longdouble_t
 * 
 */
typedef npy_double __pyx_t_5numpy_double_t;

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":813
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cfloat      cfloat_t
 */
typedef npy_longdouble __pyx_t_5numpy_longdouble_t;

/* "brainpy/_c/composition.pxd":3
 * from brainpy.mass_dict import nist_mass as __nist_mass
 * 
 * ctypedef long count_type             # <<<<<<<<<<<<<<
 * 
 * cdef dict nist_mass
 */
typedef long __pyx_t_7brainpy_2_c_11composition_count_type;
/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< float > __pyx_t_float_complex;
  #else
    typedef float _Complex __pyx_t_float_complex;
  #endif
#else
    typedef struct { float real, imag; } __pyx_t_float_complex;
#endif
static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float, float);

/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< double > __pyx_t_double_complex;
  #else
    typedef double _Complex __pyx_t_double_complex;
  #endif
#else
    typedef struct { double real, imag; } __pyx_t_double_complex;
#endif
static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double, double);


/*--- Type declarations ---*/
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase;
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak;
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet;
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed;
struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex;
struct __pyx_obj_7brainpy_2_c_11composition_PyComposition;
struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak;
struct __pyx_obj_12ms_deisotope_2_c_9averagine_Averagine;
struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern;
struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_LeastSquaresFitter;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_GTestFitter;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledGTestFitter;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_FunctionScorer;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_DotProductFitter;
struct __pyx_obj_12ms_deisotope_2_c_8peak_set__Index;
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_EnvelopePair;
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope;
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak;
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution;
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet;
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed;
struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin;
struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval;
struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode;
struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_8subgraph_FitNode;
struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_8subgraph_ConnectedSubgraph;
struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_12peak_network_PeakNode;
struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_12peak_network_DependenceCluster;
struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_12peak_network_PeakDependenceGraphBase;
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase;
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase;
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase;
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator;
struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet__has_peak;
struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet_all_peaks_for;
struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_cell;
struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_list;
struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_has_peak_hinted;

/* "ms_peak_picker/_c/peak_set.pxd":40
 *     cdef FittedPeak _get_nearest_peak(self, double mz, double* errout)
 * 
 *     cdef FittedPeak _has_peak(self, double mz, double tolerance=*)             # <<<<<<<<<<<<<<
 * 
 *     cpdef PeakSet clone(self)
 */
struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet__has_peak {
  int __pyx_n;
  double tolerance;
};

/* "ms_peak_picker/_c/peak_set.pxd":48
 *     cdef int _between_bounds(self, double m1, double m2, size_t* startp, size_t* endp)
 * 
 *     cpdef tuple all_peaks_for(self, double mz, double error_tolerance=*)             # <<<<<<<<<<<<<<
 * 
 *     cdef FittedPeak getitem(self, size_t i)
 */
struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet_all_peaks_for {
  int __pyx_n;
  double error_tolerance;
};

/* "ms_peak_picker/_c/peak_set.pxd":62
 * 
 * 
 * cdef struct index_cell:             # <<<<<<<<<<<<<<
 *     double center_value
 *     size_t start
 */
struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_cell {
  double center_value;
  size_t start;
  size_t end;
};

/* "ms_peak_picker/_c/peak_set.pxd":68
 * 
 * 
 * cdef struct index_list:             # <<<<<<<<<<<<<<
 *     index_cell* index
 *     size_t size
 */
struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_list {
  struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_cell *index;
  size_t size;
  double low;
  double high;
};

/* "ms_peak_picker/_c/peak_set.pxd":88
 *     cpdef _allocate_index(self)
 * 
 *     cpdef FittedPeak has_peak_hinted(self, double mz, double tolerance=*, size_t hint_start=*, size_t hint_end=*)             # <<<<<<<<<<<<<<
 * 
 * cdef size_t double_binary_search_ppm(double* array, double value, double tolerance, size_t n)
 */
struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_has_peak_hinted {
  int __pyx_n;
  double tolerance;
  size_t hint_start;
  size_t hint_end;
};

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":815
 * ctypedef npy_longdouble longdouble_t
 * 
 * ctypedef npy_cfloat      cfloat_t             # <<<<<<<<<<<<<<
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t
 */
typedef npy_cfloat __pyx_t_5numpy_cfloat_t;

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":816
 * 
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t             # <<<<<<<<<<<<<<
 * ctypedef npy_clongdouble clongdouble_t
 * 
 */
typedef npy_cdouble __pyx_t_5numpy_cdouble_t;

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":817
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cdouble     complex_t
 */
typedef npy_clongdouble __pyx_t_5numpy_clongdouble_t;

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":819
 * ctypedef npy_clongdouble clongdouble_t
 * 
 * ctypedef npy_cdouble     complex_t             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 */
typedef npy_cdouble __pyx_t_5numpy_complex_t;
struct __pyx_opt_args_14ms_peak_picker_2_c_10peak_index_9PeakIndex__has_peak;
struct __pyx_opt_args_14ms_peak_picker_2_c_10peak_index_9PeakIndex_all_peaks_for;

/* "ms_peak_picker/_c/peak_index.pxd":11
 *         public PeakSet peaks
 * 
 *     cdef FittedPeak _has_peak(self, double mz, double tolerance=*)             # <<<<<<<<<<<<<<
 *     cdef PeakSet _between(self, double start, double stop)
 * 
 */
struct __pyx_opt_args_14ms_peak_picker_2_c_10peak_index_9PeakIndex__has_peak {
  int __pyx_n;
  double tolerance;
};

/* "ms_peak_picker/_c/peak_index.pxd":14
 *     cdef PeakSet _between(self, double start, double stop)
 * 
 *     cpdef tuple all_peaks_for(self, double mz, double error_tolerance=*)             # <<<<<<<<<<<<<<
 * 
 *     cdef size_t get_size(self)
 */
struct __pyx_opt_args_14ms_peak_picker_2_c_10peak_index_9PeakIndex_all_peaks_for {
  int __pyx_n;
  double error_tolerance;
};
struct __pyx_opt_args_7brainpy_2_c_11composition_neutral_mass;
struct __pyx_opt_args_7brainpy_2_c_11composition_mass_charge_ratio;
struct __pyx_t_7brainpy_2_c_11composition_Isotope;
struct __pyx_t_7brainpy_2_c_11composition_IsotopeMap;
struct __pyx_t_7brainpy_2_c_11composition_Element;
struct __pyx_t_7brainpy_2_c_11composition_ElementHashBucket;
struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable;
struct __pyx_t_7brainpy_2_c_11composition_Composition;
struct __pyx_opt_args_7brainpy_2_c_11composition_13PyComposition__set_impl;
struct __pyx_opt_args_7brainpy_2_c_11composition_13PyComposition_pop;

/* "brainpy/_c/composition.pxd":10
 * cdef double PROTON = nist_mass["H+"][0][0]
 * 
 * cdef double neutral_mass(double mz,  int z, double charge_carrier=*) nogil             # <<<<<<<<<<<<<<
 * cdef double mass_charge_ratio(double neutral_mass, int z, double charge_carrier=*) nogil
 * cdef char* _parse_isotope_string(char* label, int* isotope_num, char* element_name) nogil
 */
struct __pyx_opt_args_7brainpy_2_c_11composition_neutral_mass {
  int __pyx_n;
  double charge_carrier;
};

/* "brainpy/_c/composition.pxd":11
 * 
 * cdef double neutral_mass(double mz,  int z, double charge_carrier=*) nogil
 * cdef double mass_charge_ratio(double neutral_mass, int z, double charge_carrier=*) nogil             # <<<<<<<<<<<<<<
 * cdef char* _parse_isotope_string(char* label, int* isotope_num, char* element_name) nogil
 * 
 */
struct __pyx_opt_args_7brainpy_2_c_11composition_mass_charge_ratio {
  int __pyx_n;
  double charge_carrier;
};

/* "brainpy/_c/composition.pxd":17
 * # Isotope and IsotopeMap Declarations
 * 
 * cdef struct Isotope:             # <<<<<<<<<<<<<<
 *     double mass
 *     double abundance
 */
struct __pyx_t_7brainpy_2_c_11composition_Isotope {
  double mass;
  double abundance;
  int neutrons;
  int neutron_shift;
};

/* "brainpy/_c/composition.pxd":23
 *     int neutron_shift
 * 
 * cdef struct IsotopeMap:             # <<<<<<<<<<<<<<
 *     Isotope* bins
 *     size_t size
 */
struct __pyx_t_7brainpy_2_c_11composition_IsotopeMap {
  struct __pyx_t_7brainpy_2_c_11composition_Isotope *bins;
  size_t size;
};

/* "brainpy/_c/composition.pxd":36
 * # Element Declarations
 * 
 * cdef struct Element:             # <<<<<<<<<<<<<<
 *     char* symbol
 *     IsotopeMap* isotopes
 */
struct __pyx_t_7brainpy_2_c_11composition_Element {
  char *symbol;
  struct __pyx_t_7brainpy_2_c_11composition_IsotopeMap *isotopes;
  int monoisotopic_isotope_index;
};

/* "brainpy/_c/composition.pxd":56
 * # ElementHashTable and ElementHashBucket Declarations
 * 
 * cdef struct ElementHashBucket:             # <<<<<<<<<<<<<<
 *     Element** elements
 *     size_t used
 */
struct __pyx_t_7brainpy_2_c_11composition_ElementHashBucket {
  struct __pyx_t_7brainpy_2_c_11composition_Element **elements;
  size_t used;
  size_t size;
};

/* "brainpy/_c/composition.pxd":65
 * 
 * 
 * cdef struct ElementHashTable:             # <<<<<<<<<<<<<<
 *     ElementHashBucket* buckets
 *     size_t size
 */
struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable {
  struct __pyx_t_7brainpy_2_c_11composition_ElementHashBucket *buckets;
  size_t size;
};

/* "brainpy/_c/composition.pxd":94
 * # Composition Declarations
 * 
 * cdef struct Composition:             # <<<<<<<<<<<<<<
 *     char** elements
 *     count_type* counts
 */
struct __pyx_t_7brainpy_2_c_11composition_Composition {
  char **elements;
  __pyx_t_7brainpy_2_c_11composition_count_type *counts;
  size_t size;
  size_t used;
  int max_variants;
};

/* "brainpy/_c/composition.pxd":130
 *     @staticmethod
 *     cdef PyComposition _create(Composition* base)
 *     cdef void _set_impl(self, Composition* composition, bint free_existing=*)             # <<<<<<<<<<<<<<
 * 
 *     cdef void _initialize_from_formula(self, str formula)
 */
struct __pyx_opt_args_7brainpy_2_c_11composition_13PyComposition__set_impl {
  int __pyx_n;
  int free_existing;
};

/* "brainpy/_c/composition.pxd":144
 *     cpdef list items(self)
 * 
 *     cpdef pop(self, str key, object default=*)             # <<<<<<<<<<<<<<
 * 
 *     cdef count_type getitem(self, str key)
 */
struct __pyx_opt_args_7brainpy_2_c_11composition_13PyComposition_pop {
  int __pyx_n;
  PyObject *__pyx_default;
};
struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector;

/* "brainpy/_c/double_vector.pxd":1
 * cdef struct DoubleVector:             # <<<<<<<<<<<<<<
 *     double* v
 *     size_t used
 */
struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector {
  double *v;
  size_t used;
  size_t size;
};
struct __pyx_t_7brainpy_2_c_18isotopic_constants_PolynomialParameters;
struct __pyx_t_7brainpy_2_c_18isotopic_constants_PhiConstants;
struct __pyx_t_7brainpy_2_c_18isotopic_constants_IsotopicConstants;

/* "brainpy/_c/isotopic_constants.pxd":6
 * from brainpy._c.double_vector cimport DoubleVector
 * 
 * ctypedef DoubleVector dvec             # <<<<<<<<<<<<<<
 * 
 * 
 */
typedef struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector __pyx_t_7brainpy_2_c_18isotopic_constants_dvec;

/* "brainpy/_c/isotopic_constants.pxd":9
 * 
 * 
 * cdef struct PolynomialParameters:             # <<<<<<<<<<<<<<
 *     dvec* elementary_symmetric_polynomial
 *     dvec* power_sum
 */
struct __pyx_t_7brainpy_2_c_18isotopic_constants_PolynomialParameters {
  __pyx_t_7brainpy_2_c_18isotopic_constants_dvec *elementary_symmetric_polynomial;
  __pyx_t_7brainpy_2_c_18isotopic_constants_dvec *power_sum;
};

/* "brainpy/_c/isotopic_constants.pxd":14
 * 
 * 
 * cdef struct PhiConstants:             # <<<<<<<<<<<<<<
 *     int order
 *     Element* element
 */
struct __pyx_t_7brainpy_2_c_18isotopic_constants_PhiConstants {
  int order;
  struct __pyx_t_7brainpy_2_c_11composition_Element *element;
  struct __pyx_t_7brainpy_2_c_18isotopic_constants_PolynomialParameters *element_coefficients;
  struct __pyx_t_7brainpy_2_c_18isotopic_constants_PolynomialParameters *mass_coefficients;
};

/* "brainpy/_c/isotopic_constants.pxd":21
 * 
 * 
 * cdef struct IsotopicConstants:             # <<<<<<<<<<<<<<
 *     int order
 *     PhiConstants** constants
 */
struct __pyx_t_7brainpy_2_c_18isotopic_constants_IsotopicConstants {
  int order;
  struct __pyx_t_7brainpy_2_c_18isotopic_constants_PhiConstants **constants;
  size_t size;
  size_t used;
};
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak;
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList;
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache;
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_IsotopicDistribution;
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_make_isotopic_distribution;
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_guess_npeaks;
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution__isotopic_variants;
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_isotopic_variants;

/* "brainpy/_c/isotopic_distribution.pxd":7
 * 
 * 
 * cdef struct Peak:             # <<<<<<<<<<<<<<
 *     double mz
 *     double intensity
 */
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak {
  double mz;
  double intensity;
  int charge;
};

/* "brainpy/_c/isotopic_distribution.pxd":13
 * 
 * 
 * cdef struct PeakList:             # <<<<<<<<<<<<<<
 *     Peak* peaks
 *     size_t used
 */
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList {
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak *peaks;
  size_t used;
  size_t size;
};

/* "brainpy/_c/isotopic_distribution.pxd":35
 * 
 * 
 * cdef struct ElementCache:             # <<<<<<<<<<<<<<
 *     Element** elements
 *     size_t used
 */
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache {
  struct __pyx_t_7brainpy_2_c_11composition_Element **elements;
  size_t used;
  size_t size;
  struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable *source;
};

/* "brainpy/_c/isotopic_distribution.pxd":49
 * 
 * 
 * cdef struct IsotopicDistribution:             # <<<<<<<<<<<<<<
 *     Composition* composition
 *     IsotopicConstants* _isotopic_constants
 */
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_IsotopicDistribution {
  struct __pyx_t_7brainpy_2_c_11composition_Composition *composition;
  struct __pyx_t_7brainpy_2_c_18isotopic_constants_IsotopicConstants *_isotopic_constants;
  int order;
  double average_mass;
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak *monoisotopic_peak;
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache *cache;
};

/* "brainpy/_c/isotopic_distribution.pxd":58
 * 
 * 
 * cdef IsotopicDistribution* make_isotopic_distribution(Composition* composition, int order, ElementCache* cache=*) nogil             # <<<<<<<<<<<<<<
 * cdef void free_isotopic_distribution(IsotopicDistribution* distribution)  nogil
 * 
 */
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_make_isotopic_distribution {
  int __pyx_n;
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache *cache;
};

/* "brainpy/_c/isotopic_distribution.pxd":77
 * 
 * cdef int max_variants(Composition* composition, ElementCache* cache) nogil
 * cdef int guess_npeaks(Composition* composition_struct, size_t max_npeaks, ElementCache* cache=*) nogil             # <<<<<<<<<<<<<<
 * 
 * cpdef list _isotopic_variants(object composition, object npeaks=*, int charge=*, double charge_carrier=*)
 */
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_guess_npeaks {
  int __pyx_n;
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache *cache;
};

/* "brainpy/_c/isotopic_distribution.pxd":79
 * cdef int guess_npeaks(Composition* composition_struct, size_t max_npeaks, ElementCache* cache=*) nogil
 * 
 * cpdef list _isotopic_variants(object composition, object npeaks=*, int charge=*, double charge_carrier=*)             # <<<<<<<<<<<<<<
 * 
 * 
 */
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution__isotopic_variants {
  int __pyx_n;
  PyObject *npeaks;
  int charge;
  double charge_carrier;
};

/* "brainpy/_c/isotopic_distribution.pxd":82
 * 
 * 
 * cdef PeakList* isotopic_variants(Composition* composition, int npeaks, int charge=*, double charge_carrier=*) nogil             # <<<<<<<<<<<<<<
 */
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_isotopic_variants {
  int __pyx_n;
  int charge;
  double charge_carrier;
};
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_mass_charge_ratio;
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_neutral_mass;
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine_scale;
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine_isotopic_cluster;
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine__isotopic_cluster;
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_ignore_below;
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_truncate_after;
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_scale;
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern__scale;
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_14AveragineCache_has_mz_charge_pair;
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_14AveragineCache_isotopic_cluster;
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_isotopic_shift;

/* "ms_deisotope/_c/averagine.pxd":7
 * cdef double PROTON
 * 
 * cdef double mass_charge_ratio(double neutral_mass, int z, double charge_carrier=*)             # <<<<<<<<<<<<<<
 * cdef double neutral_mass(double mz,  int z, double charge_carrier=*)
 * cdef void slide(double mz, list peaklist)
 */
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_mass_charge_ratio {
  int __pyx_n;
  double charge_carrier;
};

/* "ms_deisotope/_c/averagine.pxd":8
 * 
 * cdef double mass_charge_ratio(double neutral_mass, int z, double charge_carrier=*)
 * cdef double neutral_mass(double mz,  int z, double charge_carrier=*)             # <<<<<<<<<<<<<<
 * cdef void slide(double mz, list peaklist)
 * cdef dict scale_dict(dict data, double factor)
 */
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_neutral_mass {
  int __pyx_n;
  double charge_carrier;
};

/* "ms_deisotope/_c/averagine.pxd":17
 *         public dict base_composition
 * 
 *     cpdef dict scale(self, double mz, int charge=*, double charge_carrier=*)             # <<<<<<<<<<<<<<
 *     cpdef TheoreticalIsotopicPattern isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)
 *     cdef TheoreticalIsotopicPattern _isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)
 */
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine_scale {
  int __pyx_n;
  int charge;
  double charge_carrier;
};

/* "ms_deisotope/_c/averagine.pxd":18
 * 
 *     cpdef dict scale(self, double mz, int charge=*, double charge_carrier=*)
 *     cpdef TheoreticalIsotopicPattern isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)             # <<<<<<<<<<<<<<
 *     cdef TheoreticalIsotopicPattern _isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)
 * 
 */
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine_isotopic_cluster {
  int __pyx_n;
  int charge;
  double charge_carrier;
  double truncate_after;
  double ignore_below;
};

/* "ms_deisotope/_c/averagine.pxd":19
 *     cpdef dict scale(self, double mz, int charge=*, double charge_carrier=*)
 *     cpdef TheoreticalIsotopicPattern isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)
 *     cdef TheoreticalIsotopicPattern _isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)             # <<<<<<<<<<<<<<
 * 
 * 
 */
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine__isotopic_cluster {
  int __pyx_n;
  int charge;
  double charge_carrier;
  double truncate_after;
  double ignore_below;
};

/* "ms_deisotope/_c/averagine.pxd":39
 *     cdef bint _eq_inst(self, TheoreticalIsotopicPattern other)
 * 
 *     cpdef TheoreticalIsotopicPattern ignore_below(self, double ignore_below=*)             # <<<<<<<<<<<<<<
 *     cpdef TheoreticalIsotopicPattern truncate_after(self, double truncate_after=*)
 *     cpdef TheoreticalIsotopicPattern shift(self, double mz)
 */
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_ignore_below {
  int __pyx_n;
  double ignore_below;
};

/* "ms_deisotope/_c/averagine.pxd":40
 * 
 *     cpdef TheoreticalIsotopicPattern ignore_below(self, double ignore_below=*)
 *     cpdef TheoreticalIsotopicPattern truncate_after(self, double truncate_after=*)             # <<<<<<<<<<<<<<
 *     cpdef TheoreticalIsotopicPattern shift(self, double mz)
 *     cpdef TheoreticalIsotopicPattern scale(self, list experimental_distribution, str method=*)
 */
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_truncate_after {
  int __pyx_n;
  double truncate_after;
};

/* "ms_deisotope/_c/averagine.pxd":42
 *     cpdef TheoreticalIsotopicPattern truncate_after(self, double truncate_after=*)
 *     cpdef TheoreticalIsotopicPattern shift(self, double mz)
 *     cpdef TheoreticalIsotopicPattern scale(self, list experimental_distribution, str method=*)             # <<<<<<<<<<<<<<
 *     cpdef TheoreticalIsotopicPattern scale_raw(self, double scale_factor)
 *     cpdef double drop_last_peak(self)
 */
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_scale {
  int __pyx_n;
  PyObject *method;
};

/* "ms_deisotope/_c/averagine.pxd":53
 * 
 *     @cython.final
 *     cdef inline TheoreticalIsotopicPattern _scale(self, list experimental_distribution, str method=*)             # <<<<<<<<<<<<<<
 * 
 *     @cython.final
 */
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern__scale {
  int __pyx_n;
  PyObject *method;
};

/* "ms_deisotope/_c/averagine.pxd":88
 *     cdef double _make_cache_key(self, double mz)
 * 
 *     cdef TheoreticalIsotopicPattern has_mz_charge_pair(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)             # <<<<<<<<<<<<<<
 *     cpdef TheoreticalIsotopicPattern isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)
 * 
 */
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_14AveragineCache_has_mz_charge_pair {
  int __pyx_n;
  int charge;
  double charge_carrier;
  double truncate_after;
  double ignore_below;
};

/* "ms_deisotope/_c/averagine.pxd":89
 * 
 *     cdef TheoreticalIsotopicPattern has_mz_charge_pair(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)
 *     cpdef TheoreticalIsotopicPattern isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)             # <<<<<<<<<<<<<<
 * 
 * 
 */
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_14AveragineCache_isotopic_cluster {
  int __pyx_n;
  int charge;
  double charge_carrier;
  double truncate_after;
  double ignore_below;
};

/* "ms_deisotope/_c/averagine.pxd":92
 * 
 * 
 * cpdef double isotopic_shift(int charge=*)             # <<<<<<<<<<<<<<
 */
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_isotopic_shift {
  int __pyx_n;
  int charge;
};
struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord__create;
struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_score_peak;
struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_21InterferenceDetection_detect_interference;

/* "ms_deisotope/_c/scoring.pxd":29
 * 
 *     @staticmethod
 *     cdef IsotopicFitRecord _create(FittedPeak seed_peak, double score, int charge, TheoreticalIsotopicPattern theoretical,             # <<<<<<<<<<<<<<
 *                                    list experimental, object data=*, int missed_peaks=*)
 * 
 */
struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord__create {
  int __pyx_n;
  PyObject *data;
  int missed_peaks;
};

/* "ms_deisotope/_c/scoring.pxd":75
 *     cdef public double mass_error_tolerance
 * 
 *     cdef double score_peak(self, FittedPeak obs, TheoreticalPeak theo, double mass_error_tolerance=*, double minimum_signal_to_noise=*) nogil             # <<<<<<<<<<<<<<
 * 
 * 
 */
struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_score_peak {
  int __pyx_n;
  double mass_error_tolerance;
  double minimum_signal_to_noise;
};

/* "ms_deisotope/_c/scoring.pxd":92
 *     cdef public PeakSet peaklist
 * 
 *     cdef double detect_interference(self, list experimental_peaks, double lower=*, double upper=*)             # <<<<<<<<<<<<<<
 * 
 * 
 */
struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_21InterferenceDetection_detect_interference {
  int __pyx_n;
  double lower;
  double upper;
};
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet__has_peak;
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet_has_peak;
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet_all_peaks_for;
struct __pyx_t_12ms_deisotope_2_c_8peak_set_index_cell;
struct __pyx_t_12ms_deisotope_2_c_8peak_set_index_list;

/* "ms_deisotope/_c/peak_set.pxd":82
 *     cpdef DeconvolutedPeakSet copy(self)
 * 
 *     cdef DeconvolutedPeak _has_peak(self, double neutral_mass, double error_tolerance=*, bint use_mz=*)             # <<<<<<<<<<<<<<
 * 
 *     cpdef DeconvolutedPeak has_peak(self, double neutral_mass, double error_tolerance=*, bint use_mz=*)
 */
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet__has_peak {
  int __pyx_n;
  double error_tolerance;
  int use_mz;
};

/* "ms_deisotope/_c/peak_set.pxd":84
 *     cdef DeconvolutedPeak _has_peak(self, double neutral_mass, double error_tolerance=*, bint use_mz=*)
 * 
 *     cpdef DeconvolutedPeak has_peak(self, double neutral_mass, double error_tolerance=*, bint use_mz=*)             # <<<<<<<<<<<<<<
 * 
 *     cpdef tuple all_peaks_for(self, double neutral_mass, double tolerance=*)
 */
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet_has_peak {
  int __pyx_n;
  double error_tolerance;
  int use_mz;
};

/* "ms_deisotope/_c/peak_set.pxd":86
 *     cpdef DeconvolutedPeak has_peak(self, double neutral_mass, double error_tolerance=*, bint use_mz=*)
 * 
 *     cpdef tuple all_peaks_for(self, double neutral_mass, double tolerance=*)             # <<<<<<<<<<<<<<
 * 
 *     cdef DeconvolutedPeak _get_nearest_peak(self, double neutral_mass, double* errout)
 */
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet_all_peaks_for {
  int __pyx_n;
  double tolerance;
};

/* "ms_deisotope/_c/peak_set.pxd":113
 * cdef size_t INTERVAL_INDEX_SIZE
 * 
 * cdef struct index_cell:             # <<<<<<<<<<<<<<
 *     double center_value
 *     size_t start
 */
struct __pyx_t_12ms_deisotope_2_c_8peak_set_index_cell {
  double center_value;
  size_t start;
  size_t end;
};

/* "ms_deisotope/_c/peak_set.pxd":119
 * 
 * 
 * cdef struct index_list:             # <<<<<<<<<<<<<<
 *     index_cell* index
 *     size_t size
 */
struct __pyx_t_12ms_deisotope_2_c_8peak_set_index_list {
  struct __pyx_t_12ms_deisotope_2_c_8peak_set_index_cell *index;
  size_t size;
  double low;
  double high;
};
struct __pyx_opt_args_12ms_deisotope_2_c_23peak_dependency_network_8subgraph_17ConnectedSubgraph_find_heaviest_path;

/* "ms_deisotope/_c/peak_dependency_network/subgraph.pxd":35
 *     cdef void _init_nodes(self, list fits)
 *     cdef void populate_edges(self)
 *     cpdef find_heaviest_path(self, method=*)             # <<<<<<<<<<<<<<
 */
struct __pyx_opt_args_12ms_deisotope_2_c_23peak_dependency_network_8subgraph_17ConnectedSubgraph_find_heaviest_path {
  int __pyx_n;
  PyObject *method;
};
struct __pyx_opt_args_12ms_deisotope_2_c_23peak_dependency_network_12peak_network_23PeakDependenceGraphBase_nodes_for;

/* "ms_deisotope/_c/peak_dependency_network/peak_network.pxd":81
 *     cpdef _populate_initial_graph(self)
 *     cpdef add_fit_dependence(self, IsotopicFitRecord fit_record)
 *     cpdef list nodes_for(self, IsotopicFitRecord fit_record, dict cache=*)             # <<<<<<<<<<<<<<
 */
struct __pyx_opt_args_12ms_deisotope_2_c_23peak_dependency_network_12peak_network_23PeakDependenceGraphBase_nodes_for {
  int __pyx_n;
  PyObject *cache;
};
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_match_theoretical_isotopic_distribution;
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_subtraction;
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak;
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak;
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak_inplace;
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak_inplace;
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_fit_theoretical_distribution;
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase__fit_peaks_at_charges;
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_fit_theoretical_distribution;
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase__fit_peaks_at_charges;
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base__get_all_peak_charge_pairs;
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base__explore_local;
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_populate_graph;

/* "ms_deisotope/_c/deconvoluter_base.pxd":42
 *     cdef FittedPeak _has_peak(self, double mz, double error_tolerance)
 * 
 *     cpdef list match_theoretical_isotopic_distribution(self, list theoretical_distribution, double error_tolerance=*)             # <<<<<<<<<<<<<<
 *     cpdef scale_theoretical_distribution(self, TheoreticalIsotopicPattern theoretical_distribution, list experimental_distribution)
 *     cpdef IsotopicFitRecord _evaluate_theoretical_distribution(self, list experimental, TheoreticalIsotopicPattern theoretical, FittedPeak peak, int charge)
 */
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_match_theoretical_isotopic_distribution {
  int __pyx_n;
  double error_tolerance;
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":45
 *     cpdef scale_theoretical_distribution(self, TheoreticalIsotopicPattern theoretical_distribution, list experimental_distribution)
 *     cpdef IsotopicFitRecord _evaluate_theoretical_distribution(self, list experimental, TheoreticalIsotopicPattern theoretical, FittedPeak peak, int charge)
 *     cpdef subtraction(self, TheoreticalIsotopicPattern isotopic_cluster, double error_tolerance=*)             # <<<<<<<<<<<<<<
 *     cpdef list fit_incremental_truncation(self, IsotopicFitRecord seed_fit, double lower_bound)
 * 
 */
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_subtraction {
  int __pyx_n;
  double error_tolerance;
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":50
 *     cpdef bint _check_fit(self, IsotopicFitRecord fit)
 * 
 *     cpdef list _find_next_putative_peak(self, double mz, int charge, int step=*, double tolerance=*)             # <<<<<<<<<<<<<<
 *     cpdef list _find_previous_putative_peak(self, double mz, int charge, int step=*, double tolerance=*)
 * 
 */
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak {
  int __pyx_n;
  int step;
  double tolerance;
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":51
 * 
 *     cpdef list _find_next_putative_peak(self, double mz, int charge, int step=*, double tolerance=*)
 *     cpdef list _find_previous_putative_peak(self, double mz, int charge, int step=*, double tolerance=*)             # <<<<<<<<<<<<<<
 * 
 *     cdef int _find_next_putative_peak_inplace(self, double mz, int charge, set result, int step=*, double tolerance=*)
 */
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak {
  int __pyx_n;
  int step;
  double tolerance;
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":53
 *     cpdef list _find_previous_putative_peak(self, double mz, int charge, int step=*, double tolerance=*)
 * 
 *     cdef int _find_next_putative_peak_inplace(self, double mz, int charge, set result, int step=*, double tolerance=*)             # <<<<<<<<<<<<<<
 *     cdef int _find_previous_putative_peak_inplace(self, double mz, int charge, set result, int step=*, double tolerance=*)
 * 
 */
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak_inplace {
  int __pyx_n;
  int step;
  double tolerance;
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":54
 * 
 *     cdef int _find_next_putative_peak_inplace(self, double mz, int charge, set result, int step=*, double tolerance=*)
 *     cdef int _find_previous_putative_peak_inplace(self, double mz, int charge, set result, int step=*, double tolerance=*)             # <<<<<<<<<<<<<<
 * 
 * 
 */
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak_inplace {
  int __pyx_n;
  int step;
  double tolerance;
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":61
 *         public AveragineCache averagine
 * 
 *     cpdef IsotopicFitRecord fit_theoretical_distribution(self, FittedPeak peak, double error_tolerance, int charge,             # <<<<<<<<<<<<<<
 *                                                          double charge_carrier=*, double truncate_after=*,
 *                                                          double ignore_below=*)
 */
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_fit_theoretical_distribution {
  int __pyx_n;
  double charge_carrier;
  double truncate_after;
  double ignore_below;
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":65
 *                                                          double ignore_below=*)
 * 
 *     cpdef set _fit_peaks_at_charges(self, set peak_charge_set, double error_tolerance, double charge_carrier=*, double truncate_after=*,             # <<<<<<<<<<<<<<
 *                                     double ignore_below=*)
 * 
 */
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase__fit_peaks_at_charges {
  int __pyx_n;
  double charge_carrier;
  double truncate_after;
  double ignore_below;
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":73
 *         public list averagines
 * 
 *     cpdef IsotopicFitRecord fit_theoretical_distribution(self, FittedPeak peak, double error_tolerance, int charge,             # <<<<<<<<<<<<<<
 *                                                          AveragineCache  averagine, double charge_carrier=*, double truncate_after=*,
 *                                                          double ignore_below=*)
 */
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_fit_theoretical_distribution {
  int __pyx_n;
  double charge_carrier;
  double truncate_after;
  double ignore_below;
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":76
 *                                                          AveragineCache  averagine, double charge_carrier=*, double truncate_after=*,
 *                                                          double ignore_below=*)
 *     cpdef set _fit_peaks_at_charges(self, set peak_charge_set, double error_tolerance, double charge_carrier=*, double truncate_after=*,             # <<<<<<<<<<<<<<
 *                                     double ignore_below=*)
 * 
 */
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase__fit_peaks_at_charges {
  int __pyx_n;
  double charge_carrier;
  double truncate_after;
  double ignore_below;
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":83
 * 
 * 
 * cpdef set _get_all_peak_charge_pairs(DeconvoluterBase self, FittedPeak peak, double error_tolerance=*,             # <<<<<<<<<<<<<<
 *                                      object charge_range=*,
 *                                      int left_search_limit=*, int right_search_limit=*,
 */
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base__get_all_peak_charge_pairs {
  int __pyx_n;
  double error_tolerance;
  PyObject *charge_range;
  int left_search_limit;
  int right_search_limit;
  int recalculate_starting_peak;
  int use_quick_charge;
};

/* "ms_deisotope/_c/deconvoluter_base.pyx":929
 * 
 * @cython.binding(True)
 * cpdef int _explore_local(DeconvoluterBase self, peak, error_tolerance=ERROR_TOLERANCE,             # <<<<<<<<<<<<<<
 *                          charge_range=(1, 8), left_search_limit=1, right_search_limit=0,
 *                          charge_carrier=PROTON, truncate_after=TRUNCATE_AFTER,
 */
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base__explore_local {
  int __pyx_n;
  PyObject *error_tolerance;
  PyObject *charge_range;
  PyObject *left_search_limit;
  PyObject *right_search_limit;
  PyObject *charge_carrier;
  PyObject *truncate_after;
  PyObject *ignore_below;
};

/* "ms_deisotope/_c/deconvoluter_base.pyx":1009
 * 
 * @cython.binding(True)
 * cpdef populate_graph(DeconvoluterBase self, error_tolerance=ERROR_TOLERANCE, charge_range=(1, 8),             # <<<<<<<<<<<<<<
 *                      left_search_limit=1, right_search_limit=0, charge_carrier=PROTON,
 *                      truncate_after=TRUNCATE_AFTER, ignore_below=IGNORE_BELOW):
 */
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_populate_graph {
  int __pyx_n;
  PyObject *error_tolerance;
  PyObject *charge_range;
  PyObject *left_search_limit;
  PyObject *right_search_limit;
  PyObject *charge_carrier;
  PyObject *truncate_after;
  PyObject *ignore_below;
};

/* "ms_peak_picker/_c/peak_set.pxd":2
 * 
 * cdef class PeakBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mz
 */
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase {
  PyObject_HEAD
  struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_vtab;
  double mz;
  double intensity;
  double area;
};


/* "ms_peak_picker/_c/peak_set.pxd":12
 * 
 * 
 * cdef class FittedPeak(PeakBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double signal_to_noise
 */
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase __pyx_base;
  double signal_to_noise;
  double full_width_at_half_max;
  double left_width;
  double right_width;
  long peak_count;
  long index;
};


/* "ms_peak_picker/_c/peak_set.pxd":29
 *                             double area)
 * 
 * cdef class PeakSet(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public tuple peaks
 */
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet {
  PyObject_HEAD
  struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_vtab;
  PyObject *peaks;
  int indexed;
};


/* "ms_peak_picker/_c/peak_set.pxd":81
 * 
 * 
 * cdef class PeakSetIndexed(PeakSet):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double* mz_index
 */
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet __pyx_base;
  double *mz_index;
  struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_list *interval_index;
};


/* "ms_peak_picker/_c/peak_index.pxd":5
 * 
 * 
 * cdef class PeakIndex(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public np.ndarray mz_array
 */
struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex {
  PyObject_HEAD
  struct __pyx_vtabstruct_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_vtab;
  PyArrayObject *mz_array;
  PyArrayObject *intensity_array;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *peaks;
};


/* "brainpy/_c/composition.pxd":123
 * 
 * 
 * cdef class PyComposition(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         Composition* impl
 */
struct __pyx_obj_7brainpy_2_c_11composition_PyComposition {
  PyObject_HEAD
  struct __pyx_vtabstruct_7brainpy_2_c_11composition_PyComposition *__pyx_vtab;
  struct __pyx_t_7brainpy_2_c_11composition_Composition *impl;
  double cached_mass;
  int _clean;
};


/* "brainpy/_c/isotopic_distribution.pxd":63
 * 
 * @freelist(100000)
 * cdef class TheoreticalPeak(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mz
 */
struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak {
  PyObject_HEAD
  struct __pyx_vtabstruct_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_vtab;
  double mz;
  double intensity;
  int charge;
};


/* "ms_deisotope/_c/averagine.pxd":12
 * cdef dict scale_dict(dict data, double factor)
 * 
 * cdef class Averagine(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double base_mass
 */
struct __pyx_obj_12ms_deisotope_2_c_9averagine_Averagine {
  PyObject_HEAD
  struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_Averagine *__pyx_vtab;
  double base_mass;
  PyObject *base_composition;
};


/* "ms_deisotope/_c/averagine.pxd":23
 * 
 * @cython.final
 * cdef class TheoreticalIsotopicPattern(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list peaklist
 */
struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern {
  PyObject_HEAD
  struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *__pyx_vtab;
  PyObject *peaklist;
  double origin;
  double offset;
};


/* "ms_deisotope/_c/averagine.pxd":79
 * 
 * @cython.final
 * cdef class AveragineCache(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public dict backend
 */
struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache {
  PyObject_HEAD
  struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_AveragineCache *__pyx_vtab;
  PyObject *backend;
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_Averagine *averagine;
  double cache_truncation;
  int enabled;
};


/* "ms_deisotope/_c/scoring.pxd":16
 * 
 * 
 * cdef class IsotopicFitRecord(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public FittedPeak seed_peak
 */
struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord {
  PyObject_HEAD
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_vtab;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *seed_peak;
  double score;
  int charge;
  PyObject *experimental;
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *theoretical;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *monoisotopic_peak;
  int missed_peaks;
  PyObject *data;
  Py_hash_t _hash;
};


/* "ms_deisotope/_c/scoring.pxd":38
 * 
 * 
 * cdef class FitSelectorBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double minimum_score
 */
struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase {
  PyObject_HEAD
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FitSelectorBase *__pyx_vtab;
  double minimum_score;
};


/* "ms_deisotope/_c/scoring.pxd":50
 * 
 * 
 * cdef class IsotopicFitterBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public FitSelectorBase select
 */
struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase {
  PyObject_HEAD
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_vtab;
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *select;
};


/* "ms_deisotope/_c/scoring.pxd":62
 * 
 * 
 * cdef class LeastSquaresFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_12ms_deisotope_2_c_7scoring_LeastSquaresFitter {
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;
};


/* "ms_deisotope/_c/scoring.pxd":66
 * 
 * 
 * cdef class GTestFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_12ms_deisotope_2_c_7scoring_GTestFitter {
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;
};


/* "ms_deisotope/_c/scoring.pxd":69
 *     pass
 * 
 * cdef class ScaledGTestFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledGTestFitter {
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;
};


/* "ms_deisotope/_c/scoring.pxd":72
 *     pass
 * 
 * cdef class MSDeconVFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef public double mass_error_tolerance
 * 
 */
struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter {
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;
  double mass_error_tolerance;
};


/* "ms_deisotope/_c/scoring.pxd":78
 * 
 * 
 * cdef class PenalizedMSDeconVFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mass_error_tolerance
 */
struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter {
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;
  double mass_error_tolerance;
  double penalty_factor;
};


/* "ms_deisotope/_c/scoring.pxd":84
 * 
 * 
 * cdef class FunctionScorer(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         object function
 */
struct __pyx_obj_12ms_deisotope_2_c_7scoring_FunctionScorer {
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;
  PyObject *function;
};


/* "ms_deisotope/_c/scoring.pxd":89
 * 
 * 
 * cdef class InterferenceDetection(object):             # <<<<<<<<<<<<<<
 *     cdef public PeakSet peaklist
 * 
 */
struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection {
  PyObject_HEAD
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_InterferenceDetection *__pyx_vtab;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *peaklist;
};


/* "ms_deisotope/_c/scoring.pxd":95
 * 
 * 
 * cdef class DistinctPatternFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public InterferenceDetection interference_detector
 */
struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter {
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection *interference_detector;
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledGTestFitter *g_test_scaled;
  double peak_count_scale;
  double domain_scale;
};


/* "ms_deisotope/_c/scoring.pxd":102
 *         public double domain_scale
 * 
 * cdef class ScaledPenalizedMSDeconvFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double scale_factor
 */
struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter {
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;
  double scale_factor;
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter *scorer;
};


/* "ms_deisotope/_c/scoring.pxd":112
 * 
 * 
 * cdef class DotProductFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     pass
 */
struct __pyx_obj_12ms_deisotope_2_c_7scoring_DotProductFitter {
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;
};


/* "ms_deisotope/_c/peak_set.pxd":5
 * 
 * 
 * cdef class _Index:             # <<<<<<<<<<<<<<
 *     cdef:
 *         public size_t neutral_mass
 */
struct __pyx_obj_12ms_deisotope_2_c_8peak_set__Index {
  PyObject_HEAD
  struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set__Index *__pyx_vtab;
  size_t neutral_mass;
  size_t mz;
};


/* "ms_deisotope/_c/peak_set.pxd":14
 * 
 * 
 * cdef class EnvelopePair:             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mz
 */
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_EnvelopePair {
  PyObject_HEAD
  struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_EnvelopePair *__pyx_vtab;
  double mz;
  double intensity;
};


/* "ms_deisotope/_c/peak_set.pxd":25
 * 
 * 
 * cdef class Envelope:             # <<<<<<<<<<<<<<
 *     cdef public tuple pairs
 * 
 */
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope {
  PyObject_HEAD
  struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_Envelope *__pyx_vtab;
  PyObject *pairs;
};


/* "ms_deisotope/_c/peak_set.pxd":40
 * 
 * 
 * cdef class DeconvolutedPeak(PeakBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double neutral_mass
 */
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase __pyx_base;
  double neutral_mass;
  double signal_to_noise;
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set__Index *_index;
  double full_width_at_half_max;
  int charge;
  double a_to_a2_ratio;
  double most_abundant_mass;
  double average_mass;
  double score;
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *envelope;
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *fit;
  int chosen_for_msms;
};


/* "ms_deisotope/_c/peak_set.pxd":66
 * 
 * 
 * cdef class DeconvolutedPeakSolution(DeconvolutedPeak):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public object solution
 */
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution {
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak __pyx_base;
  PyObject *solution;
};


/* "ms_deisotope/_c/peak_set.pxd":70
 *         public object solution
 * 
 * cdef class DeconvolutedPeakSet(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public tuple peaks
 */
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet {
  PyObject_HEAD
  struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *__pyx_vtab;
  PyObject *peaks;
  PyObject *_mz_ordered;
  int indexed;
};


/* "ms_deisotope/_c/peak_set.pxd":99
 * 
 * 
 * cdef class DeconvolutedPeakSetIndexed(DeconvolutedPeakSet):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double* neutral_mass_array
 */
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed {
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet __pyx_base;
  double *neutral_mass_array;
  double *mz_array;
  struct __pyx_t_12ms_deisotope_2_c_8peak_set_index_list *interval_index;
  size_t _size;
};


/* "ms_deisotope/_c/peak_dependency_network/intervals.pxd":1
 * cdef class SpanningMixin(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double start
 */
struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin {
  PyObject_HEAD
  struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_vtab;
  double start;
  double end;
};


/* "ms_deisotope/_c/peak_dependency_network/intervals.pxd":19
 * 
 * 
 * cdef class Interval(SpanningMixin):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public object members
 */
struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval {
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin __pyx_base;
  PyObject *members;
  PyObject *data;
};


/* "ms_deisotope/_c/peak_dependency_network/intervals.pxd":25
 * 
 * 
 * cdef class IntervalTreeNode(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double center
 */
struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode {
  PyObject_HEAD
  struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_vtab;
  double center;
  double start;
  double end;
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *left;
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *right;
  PyObject *contained;
  int level;
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *parent;
};


/* "ms_deisotope/_c/peak_dependency_network/subgraph.pxd":8
 * 
 * 
 * cdef class FitNode(SpanningMixin):             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_8subgraph_FitNode {
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin __pyx_base;
  size_t index;
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *fit;
  PyObject *edges;
  PyObject *overlap_edges;
  PyObject *peak_indices;
  long _hash;
  double score;
};


/* "ms_deisotope/_c/peak_dependency_network/subgraph.pxd":28
 * 
 * 
 * cdef class ConnectedSubgraph(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public tuple nodes
 */
struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_8subgraph_ConnectedSubgraph {
  PyObject_HEAD
  struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_8subgraph_ConnectedSubgraph *__pyx_vtab;
  PyObject *nodes;
  int maximize;
};


/* "ms_deisotope/_c/peak_dependency_network/peak_network.pxd":8
 * 
 * 
 * cdef class PeakNode(object):             # <<<<<<<<<<<<<<
 *     """
 *     Represent a single FittedPeak and the junction of multiple
 */
struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_12peak_network_PeakNode {
  PyObject_HEAD
  struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_12peak_network_PeakNode *__pyx_vtab;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *peak;
  PyObject *links;
  long _hash;
};


/* "ms_deisotope/_c/peak_dependency_network/peak_network.pxd":30
 * 
 * 
 * cdef class DependenceCluster(SpanningMixin):             # <<<<<<<<<<<<<<
 *     """
 *     Represent a cluster of isotopic fits which are overlapping
 */
struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_12peak_network_DependenceCluster {
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin __pyx_base;
  PyObject *parent;
  PyObject *dependencies;
  int maximize;
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *best_fit;
};


/* "ms_deisotope/_c/peak_dependency_network/peak_network.pxd":63
 * 
 * 
 * cdef class PeakDependenceGraphBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public dict nodes
 */
struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_12peak_network_PeakDependenceGraphBase {
  PyObject_HEAD
  struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_12peak_network_PeakDependenceGraphBase *__pyx_vtab;
  PyObject *nodes;
  PyObject *dependencies;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *peaklist;
  int max_missed_peaks;
  int use_monoisotopic_superceded_filtering;
  int maximize;
  PyObject *clusters;
  PyObject *_solution_map;
  PyObject *_all_clusters;
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *_interval_tree;
};


/* "ms_deisotope/_c/deconvoluter_base.pxd":25
 * 
 * 
 * cdef class DeconvoluterBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public bint use_subtraction
 */
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase {
  PyObject_HEAD
  struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_vtab;
  int use_subtraction;
  PyObject *scale_method;
  int merge_isobaric_peaks;
  double minimum_intensity;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *peaklist;
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *scorer;
  int verbose;
  PyObject *_slice_cache;
};


/* "ms_deisotope/_c/deconvoluter_base.pxd":57
 * 
 * 
 * cdef class AveragineDeconvoluterBase(DeconvoluterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public AveragineCache averagine
 */
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase {
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase __pyx_base;
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *averagine;
};


/* "ms_deisotope/_c/deconvoluter_base.pxd":69
 * 
 * 
 * cdef class MultiAveragineDeconvoluterBase(DeconvoluterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list averagines
 */
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase {
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase __pyx_base;
  PyObject *averagines;
};


/* "ms_deisotope/_c/deconvoluter_base.pxd":91
 * 
 * 
 * cdef class ChargeIterator(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public int lower
 */
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator {
  PyObject_HEAD
  struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_vtab;
  int lower;
  int upper;
  int sign;
  int *values;
  size_t size;
  size_t index;
};



/* "ms_peak_picker/_c/peak_set.pxd":2
 * 
 * cdef class PeakBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mz
 */

struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakBase {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *(*clone)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *, int __pyx_skip_dispatch);
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *(*copy)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakBase;


/* "ms_peak_picker/_c/peak_set.pxd":12
 * 
 * 
 * cdef class FittedPeak(PeakBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double signal_to_noise
 */

struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_FittedPeak {
  struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakBase __pyx_base;
  int (*_eq)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, int __pyx_skip_dispatch);
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*_create)(double, double, double, double, double, double, long, long, double);
};
static struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_FittedPeak;


/* "ms_peak_picker/_c/peak_set.pxd":29
 *                             double area)
 * 
 * cdef class PeakSet(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public tuple peaks
 */

struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*_create)(PyObject *);
  PyObject *(*reindex)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch);
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*_get_nearest_peak)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, double, double *);
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*_has_peak)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, double, struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet__has_peak *__pyx_optional_args);
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*clone)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch);
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*copy)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch);
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*_between)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, double, double);
  int (*_between_bounds)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, double, double, size_t *, size_t *);
  PyObject *(*all_peaks_for)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet_all_peaks_for *__pyx_optional_args);
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*getitem)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, size_t);
  size_t (*get_size)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *);
  int (*_eq)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSet;


/* "ms_peak_picker/_c/peak_set.pxd":81
 * 
 * 
 * cdef class PeakSetIndexed(PeakSet):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double* mz_index
 */

struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed {
  struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet __pyx_base;
  PyObject *(*_allocate_index)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *, int __pyx_skip_dispatch);
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*has_peak_hinted)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_has_peak_hinted *__pyx_optional_args);
};
static struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed;


/* "ms_peak_picker/_c/peak_index.pxd":5
 * 
 * 
 * cdef class PeakIndex(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public np.ndarray mz_array
 */

struct __pyx_vtabstruct_14ms_peak_picker_2_c_10peak_index_PeakIndex {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*_has_peak)(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *, double, struct __pyx_opt_args_14ms_peak_picker_2_c_10peak_index_9PeakIndex__has_peak *__pyx_optional_args);
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*_between)(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *, double, double);
  PyObject *(*all_peaks_for)(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_14ms_peak_picker_2_c_10peak_index_9PeakIndex_all_peaks_for *__pyx_optional_args);
  size_t (*get_size)(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *);
  struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *(*clone)(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *, int __pyx_skip_dispatch);
  struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *(*copy)(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_vtabptr_14ms_peak_picker_2_c_10peak_index_PeakIndex;


/* "brainpy/_c/composition.pxd":123
 * 
 * 
 * cdef class PyComposition(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         Composition* impl
 */

struct __pyx_vtabstruct_7brainpy_2_c_11composition_PyComposition {
  struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *(*_create)(struct __pyx_t_7brainpy_2_c_11composition_Composition *);
  void (*_set_impl)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, struct __pyx_t_7brainpy_2_c_11composition_Composition *, struct __pyx_opt_args_7brainpy_2_c_11composition_13PyComposition__set_impl *__pyx_optional_args);
  void (*_initialize_from_formula)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, PyObject *);
  double (*mass)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, int __pyx_skip_dispatch);
  int (*__pyx___equality_pycomposition)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, int __pyx_skip_dispatch);
  int (*__pyx___equality_dict)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, PyObject *, int __pyx_skip_dispatch);
  struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *(*copy)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, int __pyx_skip_dispatch);
  PyObject *(*update)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*keys)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, int __pyx_skip_dispatch);
  PyObject *(*values)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, int __pyx_skip_dispatch);
  PyObject *(*items)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, int __pyx_skip_dispatch);
  PyObject *(*pop)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_7brainpy_2_c_11composition_13PyComposition_pop *__pyx_optional_args);
  __pyx_t_7brainpy_2_c_11composition_count_type (*getitem)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, PyObject *);
  void (*setitem)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, PyObject *, __pyx_t_7brainpy_2_c_11composition_count_type);
  void (*increment)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, PyObject *, __pyx_t_7brainpy_2_c_11composition_count_type);
  void (*add_from)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *);
  void (*subtract_from)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *);
  void (*scale_by)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, long);
};
static struct __pyx_vtabstruct_7brainpy_2_c_11composition_PyComposition *__pyx_vtabptr_7brainpy_2_c_11composition_PyComposition;


/* "brainpy/_c/isotopic_distribution.pxd":63
 * 
 * @freelist(100000)
 * cdef class TheoreticalPeak(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mz
 */

struct __pyx_vtabstruct_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak {
  struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *(*_create)(double, double, int);
  int (*_eq)(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *, struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *, int __pyx_skip_dispatch);
  struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *(*clone)(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_vtabptr_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak;


/* "ms_deisotope/_c/averagine.pxd":12
 * cdef dict scale_dict(dict data, double factor)
 * 
 * cdef class Averagine(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double base_mass
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_Averagine {
  PyObject *(*scale)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_Averagine *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine_scale *__pyx_optional_args);
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*isotopic_cluster)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_Averagine *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine_isotopic_cluster *__pyx_optional_args);
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*_isotopic_cluster)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_Averagine *, double, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine__isotopic_cluster *__pyx_optional_args);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_Averagine *__pyx_vtabptr_12ms_deisotope_2_c_9averagine_Averagine;


/* "ms_deisotope/_c/averagine.pxd":23
 * 
 * @cython.final
 * cdef class TheoreticalIsotopicPattern(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list peaklist
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern {
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*_create)(PyObject *, double, double);
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*clone)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch);
  int (*_eq)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, PyObject *, int __pyx_skip_dispatch);
  int (*_eq_inst)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *);
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*ignore_below)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_ignore_below *__pyx_optional_args);
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*truncate_after)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_truncate_after *__pyx_optional_args);
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*shift)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, double, int __pyx_skip_dispatch);
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*scale)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_scale *__pyx_optional_args);
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*scale_raw)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, double, int __pyx_skip_dispatch);
  double (*drop_last_peak)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch);
  double (*total)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch);
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*normalize)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch);
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*_scale)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, PyObject *, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern__scale *__pyx_optional_args);
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*clone_shift)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, double);
  struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *(*get)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, Py_ssize_t);
  size_t (*get_size)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *);
  double (*get_monoisotopic_mz)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *);
  PyObject *(*get_processed_peaks)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *);
  size_t (*_basepeak_index)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch);
  PyObject *(*incremental_truncation)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, double, int __pyx_skip_dispatch);
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*clone_drop_last)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *__pyx_vtabptr_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern;


/* "ms_deisotope/_c/averagine.pxd":79
 * 
 * @cython.final
 * cdef class AveragineCache(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public dict backend
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_AveragineCache {
  double (*_make_cache_key)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *, double);
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*has_mz_charge_pair)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *, double, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_14AveragineCache_has_mz_charge_pair *__pyx_optional_args);
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*isotopic_cluster)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_14AveragineCache_isotopic_cluster *__pyx_optional_args);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_AveragineCache *__pyx_vtabptr_12ms_deisotope_2_c_9averagine_AveragineCache;


/* "ms_deisotope/_c/scoring.pxd":16
 * 
 * 
 * cdef class IsotopicFitRecord(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public FittedPeak seed_peak
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitRecord {
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *(*_create)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, double, int, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, PyObject *, struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord__create *__pyx_optional_args);
  int (*_eq)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch);
  int (*_ne)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch);
  int (*_lt)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch);
  int (*_gt)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_IsotopicFitRecord;


/* "ms_deisotope/_c/scoring.pxd":38
 * 
 * 
 * cdef class FitSelectorBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double minimum_score
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FitSelectorBase {
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *(*best)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *, PyObject *, int __pyx_skip_dispatch);
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *(*_best_from_set)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *, PyObject *);
  int (*reject)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch);
  int (*reject_score)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *, double, int __pyx_skip_dispatch);
  int (*is_maximizing)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FitSelectorBase *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_FitSelectorBase;


/* "ms_deisotope/_c/scoring.pxd":50
 * 
 * 
 * cdef class IsotopicFitterBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public FitSelectorBase select
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase {
  double (*_evaluate)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, PyObject *, PyObject *, int __pyx_skip_dispatch);
  int (*reject)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch);
  int (*reject_score)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *, double, int __pyx_skip_dispatch);
  int (*is_maximizing)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *, int __pyx_skip_dispatch);
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *(*_configure)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *, struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, PyObject *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_IsotopicFitterBase;


/* "ms_deisotope/_c/scoring.pxd":62
 * 
 * 
 * cdef class LeastSquaresFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_LeastSquaresFitter {
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_LeastSquaresFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_LeastSquaresFitter;


/* "ms_deisotope/_c/scoring.pxd":66
 * 
 * 
 * cdef class GTestFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_GTestFitter {
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_GTestFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_GTestFitter;


/* "ms_deisotope/_c/scoring.pxd":69
 *     pass
 * 
 * cdef class ScaledGTestFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_ScaledGTestFitter {
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_ScaledGTestFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_ScaledGTestFitter;


/* "ms_deisotope/_c/scoring.pxd":72
 *     pass
 * 
 * cdef class MSDeconVFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef public double mass_error_tolerance
 * 
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_MSDeconVFitter {
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;
  double (*score_peak)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *, struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_score_peak *__pyx_optional_args);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_MSDeconVFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_MSDeconVFitter;


/* "ms_deisotope/_c/scoring.pxd":78
 * 
 * 
 * cdef class PenalizedMSDeconVFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mass_error_tolerance
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter {
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter;


/* "ms_deisotope/_c/scoring.pxd":84
 * 
 * 
 * cdef class FunctionScorer(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         object function
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FunctionScorer {
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FunctionScorer *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_FunctionScorer;


/* "ms_deisotope/_c/scoring.pxd":89
 * 
 * 
 * cdef class InterferenceDetection(object):             # <<<<<<<<<<<<<<
 *     cdef public PeakSet peaklist
 * 
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_InterferenceDetection {
  double (*detect_interference)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection *, PyObject *, struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_21InterferenceDetection_detect_interference *__pyx_optional_args);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_InterferenceDetection *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_InterferenceDetection;


/* "ms_deisotope/_c/scoring.pxd":95
 * 
 * 
 * cdef class DistinctPatternFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public InterferenceDetection interference_detector
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_DistinctPatternFitter {
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_DistinctPatternFitter;


/* "ms_deisotope/_c/scoring.pxd":102
 *         public double domain_scale
 * 
 * cdef class ScaledPenalizedMSDeconvFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double scale_factor
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter {
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;
  double (*_calculate_scale_factor)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch);
  void (*scale_fitted_peaks)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *, PyObject *, double);
  void (*scale_theoretical_peaks)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *, PyObject *, double);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter;


/* "ms_deisotope/_c/scoring.pxd":112
 * 
 * 
 * cdef class DotProductFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     pass
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_DotProductFitter {
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_DotProductFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_DotProductFitter;


/* "ms_deisotope/_c/peak_set.pxd":5
 * 
 * 
 * cdef class _Index:             # <<<<<<<<<<<<<<
 *     cdef:
 *         public size_t neutral_mass
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set__Index {
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set__Index *(*_create)(size_t, size_t);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set__Index *__pyx_vtabptr_12ms_deisotope_2_c_8peak_set__Index;


/* "ms_deisotope/_c/peak_set.pxd":14
 * 
 * 
 * cdef class EnvelopePair:             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mz
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_EnvelopePair {
  int (*_eq)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_EnvelopePair *, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_EnvelopePair *, int __pyx_skip_dispatch);
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_EnvelopePair *(*_create)(double, double);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_EnvelopePair *__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_EnvelopePair;


/* "ms_deisotope/_c/peak_set.pxd":25
 * 
 * 
 * cdef class Envelope:             # <<<<<<<<<<<<<<
 *     cdef public tuple pairs
 * 
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_Envelope {
  size_t (*get_size)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *);
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_EnvelopePair *(*getitem)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *, size_t);
  int (*_eq)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *, int __pyx_skip_dispatch);
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *(*clone)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *, int __pyx_skip_dispatch);
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *(*_create)(PyObject *);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_Envelope *__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_Envelope;


/* "ms_deisotope/_c/peak_set.pxd":40
 * 
 * 
 * cdef class DeconvolutedPeak(PeakBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double neutral_mass
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak {
  struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakBase __pyx_base;
  int (*_eq)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *, int __pyx_skip_dispatch);
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *(*_create_simple)(double, double, int, double, double, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak;


/* "ms_deisotope/_c/peak_set.pxd":66
 * 
 * 
 * cdef class DeconvolutedPeakSolution(DeconvolutedPeak):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public object solution
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution {
  struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak __pyx_base;
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution *__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution;


/* "ms_deisotope/_c/peak_set.pxd":70
 *         public object solution
 * 
 * cdef class DeconvolutedPeakSet(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public tuple peaks
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet {
  PyObject *(*reindex)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, int __pyx_skip_dispatch);
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *(*clone)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, int __pyx_skip_dispatch);
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *(*copy)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, int __pyx_skip_dispatch);
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *(*_has_peak)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, double, struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet__has_peak *__pyx_optional_args);
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *(*has_peak)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet_has_peak *__pyx_optional_args);
  PyObject *(*all_peaks_for)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet_all_peaks_for *__pyx_optional_args);
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *(*_get_nearest_peak)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, double, double *);
  size_t (*get_size)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *);
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *(*getitem)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, size_t);
  PyObject *(*getslice)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, size_t, size_t);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet;


/* "ms_deisotope/_c/peak_set.pxd":99
 * 
 * 
 * cdef class DeconvolutedPeakSetIndexed(DeconvolutedPeakSet):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double* neutral_mass_array
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed {
  struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet __pyx_base;
  void (*_build_index_arrays)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed *);
  int (*_interval_for)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed *, double, double, size_t *, size_t *);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed *__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed;


/* "ms_deisotope/_c/peak_dependency_network/intervals.pxd":1
 * cdef class SpanningMixin(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double start
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin {
  int (*_contains)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *, double);
  int (*contains)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *, double, int __pyx_skip_dispatch);
  int (*overlaps)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *, int __pyx_skip_dispatch);
  double (*overlap_size)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *, int __pyx_skip_dispatch);
  int (*is_contained_in_interval)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *, int __pyx_skip_dispatch);
  int (*contains_interval)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin;


/* "ms_deisotope/_c/peak_dependency_network/intervals.pxd":19
 * 
 * 
 * cdef class Interval(SpanningMixin):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public object members
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval {
  struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin __pyx_base;
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *__pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval;


/* "ms_deisotope/_c/peak_dependency_network/intervals.pxd":25
 * 
 * 
 * cdef class IntervalTreeNode(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double center
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode {
  PyObject *(*contains_point)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *, double, int __pyx_skip_dispatch);
  PyObject *(*_overlaps_interval)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *, double, double, int __pyx_skip_dispatch);
  PyObject *(*overlaps)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *, double, double, int __pyx_skip_dispatch);
  int (*_eq)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode;


/* "ms_deisotope/_c/peak_dependency_network/subgraph.pxd":8
 * 
 * 
 * cdef class FitNode(SpanningMixin):             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_8subgraph_FitNode {
  struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin __pyx_base;
  int (*isdisjoint)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_8subgraph_FitNode *, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_8subgraph_FitNode *, int __pyx_skip_dispatch);
  PyObject *(*visit)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_8subgraph_FitNode *, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_8subgraph_FitNode *, int __pyx_skip_dispatch);
  void (*_init_fields)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_8subgraph_FitNode *, PyObject *);
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_8subgraph_FitNode *(*_create)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, size_t);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_8subgraph_FitNode *__pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_8subgraph_FitNode;


/* "ms_deisotope/_c/peak_dependency_network/subgraph.pxd":28
 * 
 * 
 * cdef class ConnectedSubgraph(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public tuple nodes
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_8subgraph_ConnectedSubgraph {
  void (*_init_nodes)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_8subgraph_ConnectedSubgraph *, PyObject *);
  void (*populate_edges)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_8subgraph_ConnectedSubgraph *);
  PyObject *(*find_heaviest_path)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_8subgraph_ConnectedSubgraph *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_23peak_dependency_network_8subgraph_17ConnectedSubgraph_find_heaviest_path *__pyx_optional_args);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_8subgraph_ConnectedSubgraph *__pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_8subgraph_ConnectedSubgraph;


/* "ms_deisotope/_c/peak_dependency_network/peak_network.pxd":8
 * 
 * 
 * cdef class PeakNode(object):             # <<<<<<<<<<<<<<
 *     """
 *     Represent a single FittedPeak and the junction of multiple
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_12peak_network_PeakNode {
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_12peak_network_PeakNode *(*_create)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_12peak_network_PeakNode *__pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_12peak_network_PeakNode;


/* "ms_deisotope/_c/peak_dependency_network/peak_network.pxd":30
 * 
 * 
 * cdef class DependenceCluster(SpanningMixin):             # <<<<<<<<<<<<<<
 *     """
 *     Represent a cluster of isotopic fits which are overlapping
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_12peak_network_DependenceCluster {
  struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin __pyx_base;
  PyObject *(*_reset)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_12peak_network_DependenceCluster *, int __pyx_skip_dispatch);
  PyObject *(*add)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_12peak_network_DependenceCluster *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch);
  PyObject *(*disjoint_subset)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_12peak_network_DependenceCluster *, int __pyx_skip_dispatch);
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_8subgraph_ConnectedSubgraph *(*build_graph)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_12peak_network_DependenceCluster *, int __pyx_skip_dispatch);
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *(*_best_fit)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_12peak_network_DependenceCluster *);
  PyObject *(*disjoint_best_fits)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_12peak_network_DependenceCluster *, int __pyx_skip_dispatch);
  double (*_start)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_12peak_network_DependenceCluster *);
  double (*_end)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_12peak_network_DependenceCluster *);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_12peak_network_DependenceCluster *__pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_12peak_network_DependenceCluster;


/* "ms_deisotope/_c/peak_dependency_network/peak_network.pxd":63
 * 
 * 
 * cdef class PeakDependenceGraphBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public dict nodes
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_12peak_network_PeakDependenceGraphBase {
  PyObject *(*reset)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_12peak_network_PeakDependenceGraphBase *, int __pyx_skip_dispatch);
  PyObject *(*_populate_initial_graph)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_12peak_network_PeakDependenceGraphBase *, int __pyx_skip_dispatch);
  PyObject *(*add_fit_dependence)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_12peak_network_PeakDependenceGraphBase *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch);
  PyObject *(*nodes_for)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_12peak_network_PeakDependenceGraphBase *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_23peak_dependency_network_12peak_network_23PeakDependenceGraphBase_nodes_for *__pyx_optional_args);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_12peak_network_PeakDependenceGraphBase *__pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_12peak_network_PeakDependenceGraphBase;


/* "ms_deisotope/_c/deconvoluter_base.pyx":111
 * 
 * 
 * cdef class DeconvoluterBase(object):             # <<<<<<<<<<<<<<
 *     """Base class for all Deconvoluter types. Provides basic configuration for common operations,
 *     regardless of implementation. Because these methods form the backbone of all deconvolution algorithms,
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*between)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, double, int __pyx_skip_dispatch);
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*has_peak)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, double, int __pyx_skip_dispatch);
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*_has_peak)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, double);
  PyObject *(*match_theoretical_isotopic_distribution)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_match_theoretical_isotopic_distribution *__pyx_optional_args);
  PyObject *(*scale_theoretical_distribution)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, PyObject *, int __pyx_skip_dispatch);
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *(*_evaluate_theoretical_distribution)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, PyObject *, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, int, int __pyx_skip_dispatch);
  PyObject *(*subtraction)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_subtraction *__pyx_optional_args);
  PyObject *(*fit_incremental_truncation)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, double, int __pyx_skip_dispatch);
  int (*_check_fit)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch);
  PyObject *(*_find_next_putative_peak)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, int, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak *__pyx_optional_args);
  PyObject *(*_find_previous_putative_peak)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, int, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak *__pyx_optional_args);
  int (*_find_next_putative_peak_inplace)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, int, PyObject *, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak_inplace *__pyx_optional_args);
  int (*_find_previous_putative_peak_inplace)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, int, PyObject *, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak_inplace *__pyx_optional_args);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase;
static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__has_peak(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, double);


/* "ms_deisotope/_c/deconvoluter_base.pyx":440
 * 
 * 
 * cdef class AveragineDeconvoluterBase(DeconvoluterBase):             # <<<<<<<<<<<<<<
 *     """A base class derived from :class:`DeconvoluterBase` which provides some common methods
 *     for fitting isotopic patterns using an Averagine model.
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase {
  struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase __pyx_base;
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *(*fit_theoretical_distribution)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, double, int, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_fit_theoretical_distribution *__pyx_optional_args);
  PyObject *(*_fit_peaks_at_charges)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *, PyObject *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase__fit_peaks_at_charges *__pyx_optional_args);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase;


/* "ms_deisotope/_c/deconvoluter_base.pyx":493
 * 
 * 
 * cdef class MultiAveragineDeconvoluterBase(DeconvoluterBase):             # <<<<<<<<<<<<<<
 *     """A base class derived from :class:`DeconvoluterBase` which provides some common methods
 *     for fitting isotopic patterns using multiple Averagine models.
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase {
  struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase __pyx_base;
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *(*fit_theoretical_distribution)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, double, int, struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_fit_theoretical_distribution *__pyx_optional_args);
  PyObject *(*_fit_peaks_at_charges)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *, PyObject *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase__fit_peaks_at_charges *__pyx_optional_args);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase;


/* "ms_deisotope/_c/deconvoluter_base.pyx":600
 * @cython.final
 * @cython.freelist(10)
 * cdef class ChargeIterator(object):             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, int lo, int hi):
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator {
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *(*_create)(int, int);
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *(*from_tuple)(PyObject *);
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *(*from_quickcharge)(PyObject *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *);
  PyObject *(*sequence_from_quickcharge)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, int __pyx_skip_dispatch);
  PyObject *(*make_sequence)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *, int __pyx_skip_dispatch);
  void (*release_sequence)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *);
  void (*set_bounds)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *, int, int);
  int (*has_more)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *, int __pyx_skip_dispatch);
  PyObject *(*reset)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *, int __pyx_skip_dispatch);
  int (*get_next_value)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *, int __pyx_skip_dispatch);
  size_t (*get_size)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator;
static struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator__create(int, int);
static struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_from_tuple(PyObject *);
static struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_from_quickcharge(PyObject *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *);
static PyObject *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_sequence_from_quickcharge(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, int __pyx_skip_dispatch);
static PyObject *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_make_sequence(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *, int __pyx_skip_dispatch);
static void __pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_release_sequence(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *);
static void __pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_set_bounds(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *, int, int);
static int __pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_has_more(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *, int __pyx_skip_dispatch);
static PyObject *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_reset(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *, int __pyx_skip_dispatch);
static int __pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_get_next_value(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *, int __pyx_skip_dispatch);
static size_t __pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_get_size(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *);

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        Py_SIZE(list) = len+1;
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely((Py_TYPE(obj) == type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
#define __Pyx_GetModuleGlobalNameUncached(var, name)  {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* SliceObject.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(
        PyObject* obj, Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** py_start, PyObject** py_stop, PyObject** py_slice,
        int has_cstart, int has_cstop, int wraparound);

/* py_abs.proto */
#if CYTHON_USE_PYLONG_INTERNALS
static PyObject *__Pyx_PyLong_AbsNeg(PyObject *num);
#define __Pyx_PyNumber_Absolute(x)\
    ((likely(PyLong_CheckExact(x))) ?\
         (likely(Py_SIZE(x) >= 0) ? (Py_INCREF(x), (x)) : __Pyx_PyLong_AbsNeg(x)) :\
         PyNumber_Absolute(x))
#else
#define __Pyx_PyNumber_Absolute(x)  PyNumber_Absolute(x)
#endif

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* ListExtend.proto */
static CYTHON_INLINE int __Pyx_PyList_Extend(PyObject* L, PyObject* v) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject* none = _PyList_Extend((PyListObject*)L, v);
    if (unlikely(!none))
        return -1;
    Py_DECREF(none);
    return 0;
#else
    return PyList_SetSlice(L, PY_SSIZE_T_MAX, PY_SSIZE_T_MAX, v);
#endif
}

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* set_iter.proto */
static CYTHON_INLINE PyObject* __Pyx_set_iterator(PyObject* iterable, int is_set,
                                                  Py_ssize_t* p_orig_length, int* p_source_is_set);
static CYTHON_INLINE int __Pyx_set_iter_next(
        PyObject* iter_obj, Py_ssize_t orig_length,
        Py_ssize_t* ppos, PyObject **value,
        int source_is_set);

/* IsLittleEndian.proto */
static CYTHON_INLINE int __Pyx_Is_Little_Endian(void);

/* BufferFormatCheck.proto */
static const char* __Pyx_BufFmt_CheckString(__Pyx_BufFmt_Context* ctx, const char* ts);
static void __Pyx_BufFmt_Init(__Pyx_BufFmt_Context* ctx,
                              __Pyx_BufFmt_StackElem* stack,
                              __Pyx_TypeInfo* type);

/* BufferGetAndValidate.proto */
#define __Pyx_GetBufferAndValidate(buf, obj, dtype, flags, nd, cast, stack)\
    ((obj == Py_None || obj == NULL) ?\
    (__Pyx_ZeroBuffer(buf), 0) :\
    __Pyx__GetBufferAndValidate(buf, obj, dtype, flags, nd, cast, stack))
static int  __Pyx__GetBufferAndValidate(Py_buffer* buf, PyObject* obj,
    __Pyx_TypeInfo* dtype, int flags, int nd, int cast, __Pyx_BufFmt_StackElem* stack);
static void __Pyx_ZeroBuffer(Py_buffer* buf);
static CYTHON_INLINE void __Pyx_SafeReleaseBuffer(Py_buffer* info);
static Py_ssize_t __Pyx_minusones[] = { -1, -1, -1, -1, -1, -1, -1, -1 };
static Py_ssize_t __Pyx_zeros[] = { 0, 0, 0, 0, 0, 0, 0, 0 };

/* BufferFallbackError.proto */
static void __Pyx_RaiseBufferFallbackError(void);

/* BufferIndexError.proto */
static void __Pyx_RaiseBufferIndexError(int axis);

#define __Pyx_BufPtrStrided1d(type, buf, i0, s0) (type)((char*)buf + i0 * s0)
/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* PyDictContains.proto */
static CYTHON_INLINE int __Pyx_PyDict_ContainsTF(PyObject* item, PyObject* dict, int eq) {
    int result = PyDict_Contains(dict, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* SetItemInt.proto */
#define __Pyx_SetItemInt(o, i, v, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_SetItemInt_Fast(o, (Py_ssize_t)i, v, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list assignment index out of range"), -1) :\
               __Pyx_SetItemInt_Generic(o, to_py_func(i), v)))
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v);
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v,
                                               int is_list, int wraparound, int boundscheck);

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod0.proto */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* RaiseNoneIterError.proto */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* UnpackTupleError.proto */
static void __Pyx_UnpackTupleError(PyObject *, Py_ssize_t index);

/* UnpackTuple2.proto */
#define __Pyx_unpack_tuple2(tuple, value1, value2, is_tuple, has_known_size, decref_tuple)\
    (likely(is_tuple || PyTuple_Check(tuple)) ?\
        (likely(has_known_size || PyTuple_GET_SIZE(tuple) == 2) ?\
            __Pyx_unpack_tuple2_exact(tuple, value1, value2, decref_tuple) :\
            (__Pyx_UnpackTupleError(tuple, 2), -1)) :\
        __Pyx_unpack_tuple2_generic(tuple, value1, value2, has_known_size, decref_tuple))
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
    PyObject* tuple, PyObject** value1, PyObject** value2, int decref_tuple);
static int __Pyx_unpack_tuple2_generic(
    PyObject* tuple, PyObject** value1, PyObject** value2, int has_known_size, int decref_tuple);

/* dict_iter.proto */
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* dict, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_is_dict);
static CYTHON_INLINE int __Pyx_dict_iter_next(PyObject* dict_or_iter, Py_ssize_t orig_length, Py_ssize_t* ppos,
                                              PyObject** pkey, PyObject** pvalue, PyObject** pitem, int is_dict);

/* pyfrozenset_new.proto */
static CYTHON_INLINE PyObject* __Pyx_PyFrozenSet_New(PyObject* it);

/* py_set_discard.proto */
static CYTHON_INLINE int __Pyx_PySet_Discard(PyObject *set, PyObject *key);

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* HasAttr.proto */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* IncludeStringH.proto */
#include <string.h>

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyObject *dict);

/* FetchCommonType.proto */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);

/* CythonFunction.proto */
#define __Pyx_CyFunction_USED 1
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#define __Pyx_CyFunction_GetClassObj(f)\
    (((__pyx_CyFunctionObject *) (f))->func_classobj)
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
    PyCFunctionObject func;
#if PY_VERSION_HEX < 0x030500A0
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
    PyObject *func_classobj;
    void *defaults;
    int defaults_pyobjects;
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
} __pyx_CyFunctionObject;
static PyTypeObject *__pyx_CyFunctionType = 0;
#define __Pyx_CyFunction_Check(obj)  (__Pyx_TypeCheck(obj, __pyx_CyFunctionType))
#define __Pyx_CyFunction_NewEx(ml, flags, qualname, self, module, globals, code)\
    __Pyx_CyFunction_New(__pyx_CyFunctionType, ml, flags, qualname, self, module, globals, code)
static PyObject *__Pyx_CyFunction_New(PyTypeObject *, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *self,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(void);

/* FusedFunction.proto */
typedef struct {
    __pyx_CyFunctionObject func;
    PyObject *__signatures__;
    PyObject *type;
    PyObject *self;
} __pyx_FusedFunctionObject;
#define __pyx_FusedFunction_NewEx(ml, flags, qualname, self, module, globals, code)\
        __pyx_FusedFunction_New(__pyx_FusedFunctionType, ml, flags, qualname, self, module, globals, code)
static PyObject *__pyx_FusedFunction_New(PyTypeObject *type,
                                         PyMethodDef *ml, int flags,
                                         PyObject *qualname, PyObject *self,
                                         PyObject *module, PyObject *globals,
                                         PyObject *code);
static int __pyx_FusedFunction_clear(__pyx_FusedFunctionObject *self);
static PyTypeObject *__pyx_FusedFunctionType = NULL;
static int __pyx_FusedFunction_init(void);
#define __Pyx_FusedFunction_USED

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* BufferStructDeclare.proto */
typedef struct {
  Py_ssize_t shape, strides, suboffsets;
} __Pyx_Buf_DimInfo;
typedef struct {
  size_t refcount;
  Py_buffer pybuffer;
} __Pyx_Buffer;
typedef struct {
  __Pyx_Buffer *rcbuffer;
  char *data;
  __Pyx_Buf_DimInfo diminfo[8];
} __Pyx_LocalBuf_ND;

#if PY_MAJOR_VERSION < 3
    static int __Pyx_GetBuffer(PyObject *obj, Py_buffer *view, int flags);
    static void __Pyx_ReleaseBuffer(Py_buffer *view);
#else
    #define __Pyx_GetBuffer PyObject_GetBuffer
    #define __Pyx_ReleaseBuffer PyBuffer_Release
#endif


/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* RealImag.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #define __Pyx_CREAL(z) ((z).real())
    #define __Pyx_CIMAG(z) ((z).imag())
  #else
    #define __Pyx_CREAL(z) (__real__(z))
    #define __Pyx_CIMAG(z) (__imag__(z))
  #endif
#else
    #define __Pyx_CREAL(z) ((z).real)
    #define __Pyx_CIMAG(z) ((z).imag)
#endif
#if defined(__cplusplus) && CYTHON_CCOMPLEX\
        && (defined(_WIN32) || defined(__clang__) || (defined(__GNUC__) && (__GNUC__ >= 5 || __GNUC__ == 4 && __GNUC_MINOR__ >= 4 )) || __cplusplus >= 201103)
    #define __Pyx_SET_CREAL(z,x) ((z).real(x))
    #define __Pyx_SET_CIMAG(z,y) ((z).imag(y))
#else
    #define __Pyx_SET_CREAL(z,x) __Pyx_CREAL(z) = (x)
    #define __Pyx_SET_CIMAG(z,y) __Pyx_CIMAG(z) = (y)
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_float(a, b)   ((a)==(b))
    #define __Pyx_c_sum_float(a, b)  ((a)+(b))
    #define __Pyx_c_diff_float(a, b) ((a)-(b))
    #define __Pyx_c_prod_float(a, b) ((a)*(b))
    #define __Pyx_c_quot_float(a, b) ((a)/(b))
    #define __Pyx_c_neg_float(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_float(z) ((z)==(float)0)
    #define __Pyx_c_conj_float(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (::std::abs(z))
        #define __Pyx_c_pow_float(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_float(z) ((z)==0)
    #define __Pyx_c_conj_float(z)    (conjf(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (cabsf(z))
        #define __Pyx_c_pow_float(a, b)  (cpowf(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex);
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex);
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex, __pyx_t_float_complex);
    #endif
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_double(a, b)   ((a)==(b))
    #define __Pyx_c_sum_double(a, b)  ((a)+(b))
    #define __Pyx_c_diff_double(a, b) ((a)-(b))
    #define __Pyx_c_prod_double(a, b) ((a)*(b))
    #define __Pyx_c_quot_double(a, b) ((a)/(b))
    #define __Pyx_c_neg_double(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_double(z) ((z)==(double)0)
    #define __Pyx_c_conj_double(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (::std::abs(z))
        #define __Pyx_c_pow_double(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_double(z) ((z)==0)
    #define __Pyx_c_conj_double(z)    (conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (cabs(z))
        #define __Pyx_c_pow_double(a, b)  (cpow(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex);
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex);
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex, __pyx_t_double_complex);
    #endif
#endif

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__NPY_TYPES(enum NPY_TYPES value);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* ImportNumPyArray.proto */
static PyObject *__pyx_numpy_ndarray = NULL;
static PyObject* __Pyx_ImportNumPyArrayTypeIfAvailable(void);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* FunctionExport.proto */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig);

/* VoidPtrImport.proto */
static int __Pyx_ImportVoidPtr(PyObject *module, const char *name, void **p, const char *sig);

/* FunctionImport.proto */
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_between(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, double __pyx_v_m1, double __pyx_v_m2, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_has_peak(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, double __pyx_v_mz, double __pyx_v_error_tolerance, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__has_peak(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, double __pyx_v_mz, double __pyx_v_error_tolerance); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_match_theoretical_isotopic_distribution(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, PyObject *__pyx_v_theoretical_distribution, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_match_theoretical_isotopic_distribution *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_scale_theoretical_distribution(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *__pyx_v_theoretical_distribution, PyObject *__pyx_v_experimental_distribution, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__evaluate_theoretical_distribution(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, PyObject *__pyx_v_experimental, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *__pyx_v_theoretical, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_peak, int __pyx_v_charge, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_fit_incremental_truncation(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_seed_fit, double __pyx_v_lower_bound, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_subtraction(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *__pyx_v_isotopic_cluster, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_subtraction *__pyx_optional_args); /* proto*/
static int __pyx_f_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__check_fit(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_fit, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak_inplace(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, double __pyx_v_mz, int __pyx_v_charge, PyObject *__pyx_v_result, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak_inplace *__pyx_optional_args); /* proto*/
static int __pyx_f_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak_inplace(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, double __pyx_v_mz, int __pyx_v_charge, PyObject *__pyx_v_result, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak_inplace *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, double __pyx_v_mz, int __pyx_v_charge, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, double __pyx_v_mz, int __pyx_v_charge, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak *__pyx_optional_args); /* proto*/
static struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_fit_theoretical_distribution(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_peak, double __pyx_v_error_tolerance, int __pyx_v_charge, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_fit_theoretical_distribution *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase__fit_peaks_at_charges(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *__pyx_v_self, PyObject *__pyx_v_peak_charge_set, double __pyx_v_error_tolerance, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase__fit_peaks_at_charges *__pyx_optional_args); /* proto*/
static struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_fit_theoretical_distribution(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_peak, double __pyx_v_error_tolerance, int __pyx_v_charge, struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *__pyx_v_averagine, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_fit_theoretical_distribution *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase__fit_peaks_at_charges(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *__pyx_v_self, PyObject *__pyx_v_peak_charge_set, double __pyx_v_error_tolerance, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase__fit_peaks_at_charges *__pyx_optional_args); /* proto*/
static struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator__create(int __pyx_v_lo, int __pyx_v_hi); /* proto*/
static struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_from_tuple(PyObject *__pyx_v_charge_range); /* proto*/
static struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_from_quickcharge(PyObject *__pyx_v_charge_range, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaks, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_peak); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_sequence_from_quickcharge(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaks, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_peak, CYTHON_UNUSED int __pyx_skip_dispatch); /* proto*/
static void __pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_set_bounds(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_self, int __pyx_v_lo, int __pyx_v_hi); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_make_sequence(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_self, CYTHON_UNUSED int __pyx_skip_dispatch); /* proto*/
static void __pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_release_sequence(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_reset(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_self, CYTHON_UNUSED int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_has_more(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_self, CYTHON_UNUSED int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_get_next_value(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_self, CYTHON_UNUSED int __pyx_skip_dispatch); /* proto*/
static size_t __pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_get_size(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_self); /* proto*/

/* Module declarations from 'cython' */

/* Module declarations from 'ms_peak_picker._c.peak_set' */
static PyTypeObject *__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase = 0;
static PyTypeObject *__pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak = 0;
static PyTypeObject *__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet = 0;
static PyTypeObject *__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed = 0;
static size_t *__pyx_vp_14ms_peak_picker_2_c_8peak_set_INTERVAL_INDEX_SIZE = 0;
#define __pyx_v_14ms_peak_picker_2_c_8peak_set_INTERVAL_INDEX_SIZE (*__pyx_vp_14ms_peak_picker_2_c_8peak_set_INTERVAL_INDEX_SIZE)

/* Module declarations from 'cpython.buffer' */

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'cpython.version' */

/* Module declarations from 'cpython.exc' */

/* Module declarations from 'cpython.module' */

/* Module declarations from 'cpython.mem' */

/* Module declarations from 'cpython.tuple' */

/* Module declarations from 'cpython.list' */

/* Module declarations from 'cpython.sequence' */

/* Module declarations from 'cpython.mapping' */

/* Module declarations from 'cpython.iterator' */

/* Module declarations from 'cpython.number' */

/* Module declarations from 'cpython.int' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.bool' */
static PyTypeObject *__pyx_ptype_7cpython_4bool_bool = 0;

/* Module declarations from 'cpython.long' */

/* Module declarations from 'cpython.float' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.complex' */
static PyTypeObject *__pyx_ptype_7cpython_7complex_complex = 0;

/* Module declarations from 'cpython.string' */

/* Module declarations from 'cpython.unicode' */

/* Module declarations from 'cpython.dict' */

/* Module declarations from 'cpython.instance' */

/* Module declarations from 'cpython.function' */

/* Module declarations from 'cpython.method' */

/* Module declarations from 'cpython.weakref' */

/* Module declarations from 'cpython.getargs' */

/* Module declarations from 'cpython.pythread' */

/* Module declarations from 'cpython.pystate' */

/* Module declarations from 'cpython.cobject' */

/* Module declarations from 'cpython.oldbuffer' */

/* Module declarations from 'cpython.set' */

/* Module declarations from 'cpython.bytes' */

/* Module declarations from 'cpython.pycapsule' */

/* Module declarations from 'cpython' */

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cpython.ref' */

/* Module declarations from 'numpy' */

/* Module declarations from 'numpy' */
static PyTypeObject *__pyx_ptype_5numpy_dtype = 0;
static PyTypeObject *__pyx_ptype_5numpy_flatiter = 0;
static PyTypeObject *__pyx_ptype_5numpy_broadcast = 0;
static PyTypeObject *__pyx_ptype_5numpy_ndarray = 0;
static PyTypeObject *__pyx_ptype_5numpy_ufunc = 0;
static CYTHON_INLINE char *__pyx_f_5numpy__util_dtypestring(PyArray_Descr *, char *, char *, int *); /*proto*/
static CYTHON_INLINE int __pyx_f_5numpy_import_array(void); /*proto*/

/* Module declarations from 'ms_peak_picker._c.peak_index' */
static PyTypeObject *__pyx_ptype_14ms_peak_picker_2_c_10peak_index_PeakIndex = 0;

/* Module declarations from 'brainpy._c.composition' */
static PyTypeObject *__pyx_ptype_7brainpy_2_c_11composition_PyComposition = 0;
static PyObject **__pyx_vp_7brainpy_2_c_11composition_nist_mass = 0;
#define __pyx_v_7brainpy_2_c_11composition_nist_mass (*__pyx_vp_7brainpy_2_c_11composition_nist_mass)
static double *__pyx_vp_7brainpy_2_c_11composition_PROTON = 0;
#define __pyx_v_7brainpy_2_c_11composition_PROTON (*__pyx_vp_7brainpy_2_c_11composition_PROTON)
static struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable **__pyx_vp_7brainpy_2_c_11composition__ElementTable = 0;
#define __pyx_v_7brainpy_2_c_11composition__ElementTable (*__pyx_vp_7brainpy_2_c_11composition__ElementTable)

/* Module declarations from 'brainpy._c.double_vector' */

/* Module declarations from 'brainpy._c.isotopic_constants' */
static size_t *__pyx_vp_7brainpy_2_c_18isotopic_constants_DEFAULT_ISOTOPIC_CONSTANTS_SIZE = 0;
#define __pyx_v_7brainpy_2_c_18isotopic_constants_DEFAULT_ISOTOPIC_CONSTANTS_SIZE (*__pyx_vp_7brainpy_2_c_18isotopic_constants_DEFAULT_ISOTOPIC_CONSTANTS_SIZE)

/* Module declarations from 'brainpy._c.isotopic_distribution' */
static PyTypeObject *__pyx_ptype_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak = 0;

/* Module declarations from 'ms_deisotope._c.averagine' */
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_9averagine_Averagine = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_9averagine_AveragineCache = 0;
static double *__pyx_vp_12ms_deisotope_2_c_9averagine_PROTON = 0;
#define __pyx_v_12ms_deisotope_2_c_9averagine_PROTON (*__pyx_vp_12ms_deisotope_2_c_9averagine_PROTON)
static double (*__pyx_f_12ms_deisotope_2_c_9averagine_neutral_mass)(double, int, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_neutral_mass *__pyx_optional_args); /*proto*/
static double (*__pyx_f_12ms_deisotope_2_c_9averagine_isotopic_shift)(int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_isotopic_shift *__pyx_optional_args); /*proto*/

/* Module declarations from 'ms_deisotope._c.scoring' */
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitRecord = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_FitSelectorBase = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitterBase = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_LeastSquaresFitter = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_GTestFitter = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledGTestFitter = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_MSDeconVFitter = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_FunctionScorer = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_InterferenceDetection = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_DistinctPatternFitter = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_DotProductFitter = 0;
static double *__pyx_vp_12ms_deisotope_2_c_7scoring_INFINITY = 0;
#define __pyx_v_12ms_deisotope_2_c_7scoring_INFINITY (*__pyx_vp_12ms_deisotope_2_c_7scoring_INFINITY)

/* Module declarations from 'libc.stdlib' */

/* Module declarations from 'ms_deisotope._c.peak_set' */
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_8peak_set__Index = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_8peak_set_EnvelopePair = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_8peak_set_Envelope = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed = 0;
static size_t *__pyx_vp_12ms_deisotope_2_c_8peak_set_INTERVAL_INDEX_SIZE = 0;
#define __pyx_v_12ms_deisotope_2_c_8peak_set_INTERVAL_INDEX_SIZE (*__pyx_vp_12ms_deisotope_2_c_8peak_set_INTERVAL_INDEX_SIZE)

/* Module declarations from 'ms_deisotope._c.peak_dependency_network.intervals' */
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode = 0;

/* Module declarations from 'ms_deisotope._c.peak_dependency_network.subgraph' */
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_8subgraph_FitNode = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_8subgraph_ConnectedSubgraph = 0;

/* Module declarations from 'ms_deisotope._c.peak_dependency_network.peak_network' */
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_12peak_network_PeakNode = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_12peak_network_DependenceCluster = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_12peak_network_PeakDependenceGraphBase = 0;

/* Module declarations from 'ms_deisotope._c.deconvoluter_base' */
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator = 0;
static double __pyx_v_12ms_deisotope_2_c_17deconvoluter_base_ERROR_TOLERANCE;
static double __pyx_v_12ms_deisotope_2_c_17deconvoluter_base_IGNORE_BELOW;
static double __pyx_v_12ms_deisotope_2_c_17deconvoluter_base_TRUNCATE_AFTER;
static int __pyx_f_12ms_deisotope_2_c_17deconvoluter_base_has_multiple_real_peaks(PyObject *); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base__get_all_peak_charge_pairs(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base__get_all_peak_charge_pairs *__pyx_optional_args); /*proto*/
static size_t __pyx_f_12ms_deisotope_2_c_17deconvoluter_base_count_missed_peaks(PyObject *); /*proto*/
static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_make_placeholder_peak(double); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_drop_placeholders(PyObject *); /*proto*/
static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_first_peak(PyObject *); /*proto*/
static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_has_previous_peak_at_charge(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, int, int, double); /*proto*/
static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_has_successor_peak_at_charge(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, int, int, double); /*proto*/
static int __pyx_f_12ms_deisotope_2_c_17deconvoluter_base__explore_local(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base__explore_local *__pyx_optional_args); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_populate_graph(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_populate_graph *__pyx_optional_args); /*proto*/
static struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base__make_deconvoluted_peak(PyObject *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, double, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base___pyx_unpickle_DeconvoluterBase__set_state(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, PyObject *); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base___pyx_unpickle_AveragineDeconvoluterBase__set_state(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *, PyObject *); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base___pyx_unpickle_MultiAveragineDeconvoluterBase__set_state(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *, PyObject *); /*proto*/
static PyArrayObject *__pyx_fuse_0_0__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_quick_charge(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, int, int, int __pyx_skip_dispatch); /*proto*/
static PyArrayObject *__pyx_fuse_0_1__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_quick_charge(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, size_t, int, int, int __pyx_skip_dispatch); /*proto*/
static PyArrayObject *__pyx_fuse_0_2__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_quick_charge(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, long, int, int, int __pyx_skip_dispatch); /*proto*/
static PyArrayObject *__pyx_fuse_1_0__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_quick_charge(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, int, int, int __pyx_skip_dispatch); /*proto*/
static PyArrayObject *__pyx_fuse_1_1__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_quick_charge(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *, size_t, int, int, int __pyx_skip_dispatch); /*proto*/
static PyArrayObject *__pyx_fuse_1_2__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_quick_charge(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *, long, int, int, int __pyx_skip_dispatch); /*proto*/
static __Pyx_TypeInfo __Pyx_TypeInfo_int = { "int", NULL, sizeof(int), { 0 }, 0, IS_UNSIGNED(int) ? 'U' : 'I', IS_UNSIGNED(int), 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t = { "int32_t", NULL, sizeof(__pyx_t_5numpy_int32_t), { 0 }, 0, IS_UNSIGNED(__pyx_t_5numpy_int32_t) ? 'U' : 'I', IS_UNSIGNED(__pyx_t_5numpy_int32_t), 0 };
#define __Pyx_MODULE_NAME "ms_deisotope._c.deconvoluter_base"
extern int __pyx_module_is_main_ms_deisotope___c__deconvoluter_base;
int __pyx_module_is_main_ms_deisotope___c__deconvoluter_base = 0;

/* Implementation of 'ms_deisotope._c.deconvoluter_base' */
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_sorted;
static PyObject *__pyx_builtin_super;
static PyObject *__pyx_builtin_StopIteration;
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin___import__;
static PyObject *__pyx_builtin_IndexError;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_RuntimeError;
static PyObject *__pyx_builtin_ImportError;
static const char __pyx_k_s[] = "s";
static const char __pyx_k__8[] = "";
static const char __pyx_k__9[] = "()";
static const char __pyx_k_hi[] = "hi";
static const char __pyx_k_lo[] = "lo";
static const char __pyx_k_m1[] = "m1";
static const char __pyx_k_m2[] = "m2";
static const char __pyx_k_mz[] = "mz";
static const char __pyx_k_np[] = "np";
static const char __pyx_k__10[] = "|";
static const char __pyx_k_add[] = "add";
static const char __pyx_k_fit[] = "fit";
static const char __pyx_k_key[] = "key";
static const char __pyx_k_new[] = "__new__";
static const char __pyx_k_sum[] = "sum";
static const char __pyx_k_area[] = "area";
static const char __pyx_k_args[] = "args";
static const char __pyx_k_dict[] = "__dict__";
static const char __pyx_k_init[] = "__init__";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_peak[] = "peak";
static const char __pyx_k_self[] = "self";
static const char __pyx_k_step[] = "step";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_class[] = "__class__";
static const char __pyx_k_dtype[] = "dtype";
static const char __pyx_k_empty[] = "empty";
static const char __pyx_k_index[] = "index";
static const char __pyx_k_int32[] = "int32";
static const char __pyx_k_numpy[] = "numpy";
static const char __pyx_k_peaks[] = "peaks";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_score[] = "score";
static const char __pyx_k_split[] = "split";
static const char __pyx_k_strip[] = "strip";
static const char __pyx_k_super[] = "super";
static const char __pyx_k_zeros[] = "zeros";
static const char __pyx_k_charge[] = "charge";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_kwargs[] = "kwargs";
static const char __pyx_k_pickle[] = "pickle";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_size_t[] = "size_t";
static const char __pyx_k_sorted[] = "sorted";
static const char __pyx_k_update[] = "update";
static const char __pyx_k_PeakSet[] = "PeakSet";
static const char __pyx_k_between[] = "between";
static const char __pyx_k_defaults[] = "defaults";
static const char __pyx_k_envelope[] = "envelope";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_has_peak[] = "has_peak";
static const char __pyx_k_operator[] = "operator";
static const char __pyx_k_peak_set[] = "peak_set";
static const char __pyx_k_pyx_type[] = "__pyx_type";
static const char __pyx_k_seed_fit[] = "seed_fit";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_PeakIndex[] = "PeakIndex";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_averagine[] = "averagine";
static const char __pyx_k_check_fit[] = "_check_fit";
static const char __pyx_k_intensity[] = "intensity";
static const char __pyx_k_pyx_state[] = "__pyx_state";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_tolerance[] = "tolerance";
static const char __pyx_k_FittedPeak[] = "FittedPeak";
static const char __pyx_k_IndexError[] = "IndexError";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_attrgetter[] = "attrgetter";
static const char __pyx_k_max_charge[] = "max_charge";
static const char __pyx_k_min_charge[] = "min_charge";
static const char __pyx_k_pyx_result[] = "__pyx_result";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_signatures[] = "signatures";
static const char __pyx_k_ImportError[] = "ImportError";
static const char __pyx_k_PickleError[] = "PickleError";
static const char __pyx_k_lower_bound[] = "lower_bound";
static const char __pyx_k_subtraction[] = "subtraction";
static const char __pyx_k_theoretical[] = "theoretical";
static const char __pyx_k_IGNORE_BELOW[] = "IGNORE_BELOW";
static const char __pyx_k_PeakSet_long[] = "PeakSet|long";
static const char __pyx_k_RuntimeError[] = "RuntimeError";
static const char __pyx_k_average_mass[] = "average_mass";
static const char __pyx_k_charge_range[] = "charge_range";
static const char __pyx_k_experimental[] = "experimental";
static const char __pyx_k_ignore_below[] = "ignore_below";
static const char __pyx_k_neutral_mass[] = "neutral_mass";
static const char __pyx_k_priority_map[] = "_priority_map";
static const char __pyx_k_pyx_checksum[] = "__pyx_checksum";
static const char __pyx_k_quick_charge[] = "quick_charge";
static const char __pyx_k_scale_method[] = "scale_method";
static const char __pyx_k_stringsource[] = "stringsource";
static const char __pyx_k_StopIteration[] = "StopIteration";
static const char __pyx_k_a_to_a2_ratio[] = "a_to_a2_ratio";
static const char __pyx_k_explore_local[] = "_explore_local";
static const char __pyx_k_position_spec[] = "position_spec";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_ChargeIterator[] = "ChargeIterator";
static const char __pyx_k_IGNORE_BELOW_2[] = "_IGNORE_BELOW";
static const char __pyx_k_PeakIndex_long[] = "PeakIndex|long";
static const char __pyx_k_PeakSet_size_t[] = "PeakSet|size_t";
static const char __pyx_k_TRUNCATE_AFTER[] = "TRUNCATE_AFTER";
static const char __pyx_k_charge_carrier[] = "charge_carrier";
static const char __pyx_k_populate_graph[] = "populate_graph";
static const char __pyx_k_truncate_after[] = "truncate_after";
static const char __pyx_k_ERROR_TOLERANCE[] = "ERROR_TOLERANCE";
static const char __pyx_k_error_tolerance[] = "error_tolerance";
static const char __pyx_k_monoisotopic_mz[] = "monoisotopic_mz";
static const char __pyx_k_peak_charge_set[] = "peak_charge_set";
static const char __pyx_k_pyx_PickleError[] = "__pyx_PickleError";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_signal_to_noise[] = "signal_to_noise";
static const char __pyx_k_use_subtraction[] = "use_subtraction";
static const char __pyx_k_DeconvoluterBase[] = "DeconvoluterBase";
static const char __pyx_k_PeakIndex_size_t[] = "PeakIndex|size_t";
static const char __pyx_k_TRUNCATE_AFTER_2[] = "_TRUNCATE_AFTER";
static const char __pyx_k_isotopic_cluster[] = "isotopic_cluster";
static const char __pyx_k_use_quick_charge[] = "use_quick_charge";
static const char __pyx_k_ERROR_TOLERANCE_2[] = "_ERROR_TOLERANCE";
static const char __pyx_k_left_search_limit[] = "left_search_limit";
static const char __pyx_k_minimum_intensity[] = "minimum_intensity";
static const char __pyx_k_PeakSet_FittedPeak[] = "PeakSet|FittedPeak";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_most_abundant_mass[] = "most_abundant_mass";
static const char __pyx_k_right_search_limit[] = "right_search_limit";
static const char __pyx_k_PeakIndex_FittedPeak[] = "PeakIndex|FittedPeak";
static const char __pyx_k_fit_peaks_at_charges[] = "_fit_peaks_at_charges";
static const char __pyx_k_merge_isobaric_peaks[] = "merge_isobaric_peaks";
static const char __pyx_k_fit_all_charge_states[] = "_fit_all_charge_states";
static const char __pyx_k_s_peaklist_s_scorer_s[] = "%s(peaklist=%s, scorer=%s)";
static const char __pyx_k_full_width_at_half_max[] = "full_width_at_half_max";
static const char __pyx_k_make_deconvoluted_peak[] = "_make_deconvoluted_peak";
static const char __pyx_k_ms_deisotope_constants[] = "ms_deisotope.constants";
static const char __pyx_k_find_next_putative_peak[] = "_find_next_putative_peak";
static const char __pyx_k_peak_dependency_network[] = "peak_dependency_network";
static const char __pyx_k_peak_set_cannot_be_empty[] = "peak_set cannot be empty!";
static const char __pyx_k_pyx_fuse_0_0quick_charge[] = "__pyx_fuse_0_0quick_charge";
static const char __pyx_k_pyx_fuse_0_1quick_charge[] = "__pyx_fuse_0_1quick_charge";
static const char __pyx_k_pyx_fuse_0_2quick_charge[] = "__pyx_fuse_0_2quick_charge";
static const char __pyx_k_pyx_fuse_1_0quick_charge[] = "__pyx_fuse_1_0quick_charge";
static const char __pyx_k_pyx_fuse_1_1quick_charge[] = "__pyx_fuse_1_1quick_charge";
static const char __pyx_k_pyx_fuse_1_2quick_charge[] = "__pyx_fuse_1_2quick_charge";
static const char __pyx_k_theoretical_distribution[] = "theoretical_distribution";
static const char __pyx_k_AveragineDeconvoluterBase[] = "AveragineDeconvoluterBase";
static const char __pyx_k_experimental_distribution[] = "experimental_distribution";
static const char __pyx_k_get_all_peak_charge_pairs[] = "_get_all_peak_charge_pairs";
static const char __pyx_k_recalculate_starting_peak[] = "recalculate_starting_peak";
static const char __pyx_k_fit_incremental_truncation[] = "fit_incremental_truncation";
static const char __pyx_k_ms_peak_picker__c_peak_set[] = "ms_peak_picker._c.peak_set";
static const char __pyx_k_No_matching_signature_found[] = "No matching signature found";
static const char __pyx_k_find_previous_putative_peak[] = "_find_previous_putative_peak";
static const char __pyx_k_ndarray_is_not_C_contiguous[] = "ndarray is not C contiguous";
static const char __pyx_k_fit_theoretical_distribution[] = "fit_theoretical_distribution";
static const char __pyx_k_ms_peak_picker__c_peak_index[] = "ms_peak_picker._c.peak_index";
static const char __pyx_k_pyx_unpickle_DeconvoluterBase[] = "__pyx_unpickle_DeconvoluterBase";
static const char __pyx_k_MultiAveragineDeconvoluterBase[] = "MultiAveragineDeconvoluterBase";
static const char __pyx_k_pyx_unpickle_AveragineDeconvol[] = "__pyx_unpickle_AveragineDeconvoluterBase";
static const char __pyx_k_pyx_unpickle_MultiAveragineDec[] = "__pyx_unpickle_MultiAveragineDeconvoluterBase";
static const char __pyx_k_scale_theoretical_distribution[] = "scale_theoretical_distribution";
static const char __pyx_k_d_is_out_of_bounds_for_peak_lis[] = "%d is out of bounds for peak list of size %d in quick_charge";
static const char __pyx_k_evaluate_theoretical_distributi[] = "_evaluate_theoretical_distribution";
static const char __pyx_k_numpy_core_multiarray_failed_to[] = "numpy.core.multiarray failed to import";
static const char __pyx_k_self_values_cannot_be_converted[] = "self.values cannot be converted to a Python object for pickling";
static const char __pyx_k_unknown_dtype_code_in_numpy_pxd[] = "unknown dtype code in numpy.pxd (%d)";
static const char __pyx_k_Expected_at_least_d_argument_s_g[] = "Expected at least %d argument%s, got %d";
static const char __pyx_k_Format_string_allocated_too_shor[] = "Format string allocated too short, see comment in numpy.pxd";
static const char __pyx_k_Function_call_with_ambiguous_arg[] = "Function call with ambiguous argument types";
static const char __pyx_k_Incompatible_checksums_s_vs_0x16[] = "Incompatible checksums (%s vs 0x16fc763 = (_slice_cache, merge_isobaric_peaks, minimum_intensity, peaklist, scale_method, scorer, use_subtraction, verbose))";
static const char __pyx_k_Incompatible_checksums_s_vs_0x9d[] = "Incompatible checksums (%s vs 0x9da1e44 = (_slice_cache, averagine, merge_isobaric_peaks, minimum_intensity, peaklist, scale_method, scorer, use_subtraction, verbose))";
static const char __pyx_k_Incompatible_checksums_s_vs_0xfd[] = "Incompatible checksums (%s vs 0xfdf46c0 = (_slice_cache, averagines, merge_isobaric_peaks, minimum_intensity, peaklist, scale_method, scorer, use_subtraction, verbose))";
static const char __pyx_k_Non_native_byte_order_not_suppor[] = "Non-native byte order not supported";
static const char __pyx_k_match_theoretical_isotopic_distr[] = "match_theoretical_isotopic_distribution";
static const char __pyx_k_ms_deisotope__c_deconvoluter_bas[] = "ms_deisotope\\_c\\deconvoluter_base.pyx";
static const char __pyx_k_ndarray_is_not_Fortran_contiguou[] = "ndarray is not Fortran contiguous";
static const char __pyx_k_numpy_core_umath_failed_to_impor[] = "numpy.core.umath failed to import";
static const char __pyx_k_Format_string_allocated_too_shor_2[] = "Format string allocated too short.";
static const char __pyx_k_ms_deisotope__c_deconvoluter_bas_2[] = "ms_deisotope._c.deconvoluter_base";
static PyObject *__pyx_n_s_AveragineDeconvoluterBase;
static PyObject *__pyx_n_s_ChargeIterator;
static PyObject *__pyx_n_s_DeconvoluterBase;
static PyObject *__pyx_n_s_ERROR_TOLERANCE;
static PyObject *__pyx_n_s_ERROR_TOLERANCE_2;
static PyObject *__pyx_kp_s_Expected_at_least_d_argument_s_g;
static PyObject *__pyx_n_s_FittedPeak;
static PyObject *__pyx_kp_u_Format_string_allocated_too_shor;
static PyObject *__pyx_kp_u_Format_string_allocated_too_shor_2;
static PyObject *__pyx_kp_s_Function_call_with_ambiguous_arg;
static PyObject *__pyx_n_s_IGNORE_BELOW;
static PyObject *__pyx_n_s_IGNORE_BELOW_2;
static PyObject *__pyx_n_s_ImportError;
static PyObject *__pyx_kp_s_Incompatible_checksums_s_vs_0x16;
static PyObject *__pyx_kp_s_Incompatible_checksums_s_vs_0x9d;
static PyObject *__pyx_kp_s_Incompatible_checksums_s_vs_0xfd;
static PyObject *__pyx_n_s_IndexError;
static PyObject *__pyx_n_s_MultiAveragineDeconvoluterBase;
static PyObject *__pyx_kp_s_No_matching_signature_found;
static PyObject *__pyx_kp_u_Non_native_byte_order_not_suppor;
static PyObject *__pyx_n_s_PeakIndex;
static PyObject *__pyx_kp_s_PeakIndex_FittedPeak;
static PyObject *__pyx_kp_s_PeakIndex_long;
static PyObject *__pyx_kp_s_PeakIndex_size_t;
static PyObject *__pyx_n_s_PeakSet;
static PyObject *__pyx_kp_s_PeakSet_FittedPeak;
static PyObject *__pyx_kp_s_PeakSet_long;
static PyObject *__pyx_kp_s_PeakSet_size_t;
static PyObject *__pyx_n_s_PickleError;
static PyObject *__pyx_n_s_RuntimeError;
static PyObject *__pyx_n_s_StopIteration;
static PyObject *__pyx_n_s_TRUNCATE_AFTER;
static PyObject *__pyx_n_s_TRUNCATE_AFTER_2;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_kp_s__10;
static PyObject *__pyx_kp_s__8;
static PyObject *__pyx_kp_s__9;
static PyObject *__pyx_n_s_a_to_a2_ratio;
static PyObject *__pyx_n_s_add;
static PyObject *__pyx_n_s_area;
static PyObject *__pyx_n_s_args;
static PyObject *__pyx_n_s_attrgetter;
static PyObject *__pyx_n_s_average_mass;
static PyObject *__pyx_n_s_averagine;
static PyObject *__pyx_n_s_between;
static PyObject *__pyx_n_s_charge;
static PyObject *__pyx_n_s_charge_carrier;
static PyObject *__pyx_n_s_charge_range;
static PyObject *__pyx_n_s_check_fit;
static PyObject *__pyx_n_s_class;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_kp_s_d_is_out_of_bounds_for_peak_lis;
static PyObject *__pyx_n_s_defaults;
static PyObject *__pyx_n_s_dict;
static PyObject *__pyx_n_s_dtype;
static PyObject *__pyx_n_s_empty;
static PyObject *__pyx_n_s_envelope;
static PyObject *__pyx_n_s_error_tolerance;
static PyObject *__pyx_n_s_evaluate_theoretical_distributi;
static PyObject *__pyx_n_s_experimental;
static PyObject *__pyx_n_s_experimental_distribution;
static PyObject *__pyx_n_s_explore_local;
static PyObject *__pyx_n_s_find_next_putative_peak;
static PyObject *__pyx_n_s_find_previous_putative_peak;
static PyObject *__pyx_n_s_fit;
static PyObject *__pyx_n_s_fit_all_charge_states;
static PyObject *__pyx_n_s_fit_incremental_truncation;
static PyObject *__pyx_n_s_fit_peaks_at_charges;
static PyObject *__pyx_n_s_fit_theoretical_distribution;
static PyObject *__pyx_n_s_full_width_at_half_max;
static PyObject *__pyx_n_s_get_all_peak_charge_pairs;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_n_s_has_peak;
static PyObject *__pyx_n_s_hi;
static PyObject *__pyx_n_s_ignore_below;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_index;
static PyObject *__pyx_n_s_init;
static PyObject *__pyx_n_s_int32;
static PyObject *__pyx_n_s_intensity;
static PyObject *__pyx_n_s_isotopic_cluster;
static PyObject *__pyx_n_s_key;
static PyObject *__pyx_n_s_kwargs;
static PyObject *__pyx_n_s_left_search_limit;
static PyObject *__pyx_n_s_lo;
static PyObject *__pyx_n_s_lower_bound;
static PyObject *__pyx_n_s_m1;
static PyObject *__pyx_n_s_m2;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_make_deconvoluted_peak;
static PyObject *__pyx_n_s_match_theoretical_isotopic_distr;
static PyObject *__pyx_n_s_max_charge;
static PyObject *__pyx_n_s_merge_isobaric_peaks;
static PyObject *__pyx_n_s_min_charge;
static PyObject *__pyx_n_s_minimum_intensity;
static PyObject *__pyx_n_s_monoisotopic_mz;
static PyObject *__pyx_n_s_most_abundant_mass;
static PyObject *__pyx_kp_s_ms_deisotope__c_deconvoluter_bas;
static PyObject *__pyx_n_s_ms_deisotope__c_deconvoluter_bas_2;
static PyObject *__pyx_n_s_ms_deisotope_constants;
static PyObject *__pyx_kp_s_ms_peak_picker__c_peak_index;
static PyObject *__pyx_kp_s_ms_peak_picker__c_peak_set;
static PyObject *__pyx_n_s_mz;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_kp_u_ndarray_is_not_C_contiguous;
static PyObject *__pyx_kp_u_ndarray_is_not_Fortran_contiguou;
static PyObject *__pyx_n_s_neutral_mass;
static PyObject *__pyx_n_s_new;
static PyObject *__pyx_n_s_np;
static PyObject *__pyx_n_s_numpy;
static PyObject *__pyx_kp_s_numpy_core_multiarray_failed_to;
static PyObject *__pyx_kp_s_numpy_core_umath_failed_to_impor;
static PyObject *__pyx_n_s_operator;
static PyObject *__pyx_n_s_peak;
static PyObject *__pyx_n_s_peak_charge_set;
static PyObject *__pyx_n_s_peak_dependency_network;
static PyObject *__pyx_n_s_peak_set;
static PyObject *__pyx_kp_s_peak_set_cannot_be_empty;
static PyObject *__pyx_n_s_peaks;
static PyObject *__pyx_n_s_pickle;
static PyObject *__pyx_n_s_populate_graph;
static PyObject *__pyx_n_s_position_spec;
static PyObject *__pyx_n_s_priority_map;
static PyObject *__pyx_n_s_pyx_PickleError;
static PyObject *__pyx_n_s_pyx_checksum;
static PyObject *__pyx_n_s_pyx_fuse_0_0quick_charge;
static PyObject *__pyx_n_s_pyx_fuse_0_1quick_charge;
static PyObject *__pyx_n_s_pyx_fuse_0_2quick_charge;
static PyObject *__pyx_n_s_pyx_fuse_1_0quick_charge;
static PyObject *__pyx_n_s_pyx_fuse_1_1quick_charge;
static PyObject *__pyx_n_s_pyx_fuse_1_2quick_charge;
static PyObject *__pyx_n_s_pyx_result;
static PyObject *__pyx_n_s_pyx_state;
static PyObject *__pyx_n_s_pyx_type;
static PyObject *__pyx_n_s_pyx_unpickle_AveragineDeconvol;
static PyObject *__pyx_n_s_pyx_unpickle_DeconvoluterBase;
static PyObject *__pyx_n_s_pyx_unpickle_MultiAveragineDec;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_quick_charge;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_recalculate_starting_peak;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_reduce_cython;
static PyObject *__pyx_n_s_reduce_ex;
static PyObject *__pyx_n_s_right_search_limit;
static PyObject *__pyx_n_s_s;
static PyObject *__pyx_kp_s_s_peaklist_s_scorer_s;
static PyObject *__pyx_n_s_scale_method;
static PyObject *__pyx_n_s_scale_theoretical_distribution;
static PyObject *__pyx_n_s_score;
static PyObject *__pyx_n_s_seed_fit;
static PyObject *__pyx_n_s_self;
static PyObject *__pyx_kp_s_self_values_cannot_be_converted;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_setstate_cython;
static PyObject *__pyx_n_s_signal_to_noise;
static PyObject *__pyx_n_s_signatures;
static PyObject *__pyx_n_s_size_t;
static PyObject *__pyx_n_s_sorted;
static PyObject *__pyx_n_s_split;
static PyObject *__pyx_n_s_step;
static PyObject *__pyx_kp_s_stringsource;
static PyObject *__pyx_n_s_strip;
static PyObject *__pyx_n_s_subtraction;
static PyObject *__pyx_n_s_sum;
static PyObject *__pyx_n_s_super;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_theoretical;
static PyObject *__pyx_n_s_theoretical_distribution;
static PyObject *__pyx_n_s_tolerance;
static PyObject *__pyx_n_s_truncate_after;
static PyObject *__pyx_kp_u_unknown_dtype_code_in_numpy_pxd;
static PyObject *__pyx_n_s_update;
static PyObject *__pyx_n_s_use_quick_charge;
static PyObject *__pyx_n_s_use_subtraction;
static PyObject *__pyx_n_s_zeros;
static int __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase___init__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, PyObject *__pyx_v_use_subtraction, PyObject *__pyx_v_scale_method, PyObject *__pyx_v_merge_isobaric_peaks, PyObject *__pyx_v_minimum_intensity, CYTHON_UNUSED PyObject *__pyx_v_args, CYTHON_UNUSED PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_2between(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, double __pyx_v_m1, double __pyx_v_m2); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_4has_peak(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, double __pyx_v_mz, double __pyx_v_error_tolerance); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_6match_theoretical_isotopic_distribution(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, PyObject *__pyx_v_theoretical_distribution, double __pyx_v_error_tolerance); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_8scale_theoretical_distribution(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *__pyx_v_theoretical_distribution, PyObject *__pyx_v_experimental_distribution); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_10_evaluate_theoretical_distribution(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, PyObject *__pyx_v_experimental, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *__pyx_v_theoretical, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_peak, int __pyx_v_charge); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_12fit_incremental_truncation(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_seed_fit, double __pyx_v_lower_bound); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_14subtraction(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *__pyx_v_isotopic_cluster, double __pyx_v_error_tolerance); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_16_check_fit(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_fit); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_18_merge_peaks(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, PyObject *__pyx_v_peak_list); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_20_find_next_putative_peak(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, double __pyx_v_mz, int __pyx_v_charge, int __pyx_v_step, double __pyx_v_tolerance); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_22_find_previous_putative_peak(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, double __pyx_v_mz, int __pyx_v_charge, int __pyx_v_step, double __pyx_v_tolerance); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_24__repr__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_15use_subtraction___get__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_15use_subtraction_2__set__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_12scale_method___get__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_12scale_method_2__set__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_12scale_method_4__del__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_20merge_isobaric_peaks___get__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_20merge_isobaric_peaks_2__set__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_17minimum_intensity___get__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_17minimum_intensity_2__set__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_8peaklist___get__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_8peaklist_2__set__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_8peaklist_4__del__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_6scorer___get__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_6scorer_2__set__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_6scorer_4__del__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_7verbose___get__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_7verbose_2__set__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_12_slice_cache___get__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_12_slice_cache_2__set__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_12_slice_cache_4__del__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_26__reduce_cython__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_28__setstate_cython__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase___init__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *__pyx_v_self, int __pyx_v_use_subtraction, PyObject *__pyx_v_scale_method, int __pyx_v_merge_isobaric_peaks, double __pyx_v_minimum_intensity, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_2fit_theoretical_distribution(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_peak, double __pyx_v_error_tolerance, int __pyx_v_charge, double __pyx_v_charge_carrier, double __pyx_v_truncate_after, double __pyx_v_ignore_below); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_4_fit_peaks_at_charges(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *__pyx_v_self, PyObject *__pyx_v_peak_charge_set, double __pyx_v_error_tolerance, double __pyx_v_charge_carrier, double __pyx_v_truncate_after, double __pyx_v_ignore_below); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_9averagine___get__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_9averagine_2__set__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_9averagine_4__del__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_6__reduce_cython__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_8__setstate_cython__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase___init__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *__pyx_v_self, int __pyx_v_use_subtraction, PyObject *__pyx_v_scale_method, int __pyx_v_merge_isobaric_peaks, double __pyx_v_minimum_intensity, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_2fit_theoretical_distribution(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_peak, double __pyx_v_error_tolerance, int __pyx_v_charge, struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *__pyx_v_averagine, double __pyx_v_charge_carrier, double __pyx_v_truncate_after, double __pyx_v_ignore_below); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_4_fit_peaks_at_charges(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *__pyx_v_self, PyObject *__pyx_v_peak_charge_set, double __pyx_v_error_tolerance, double __pyx_v_charge_carrier, double __pyx_v_truncate_after, double __pyx_v_ignore_below); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_10averagines___get__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_10averagines_2__set__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_10averagines_4__del__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_6__reduce_cython__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_8__setstate_cython__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator___init__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_self, int __pyx_v_lo, int __pyx_v_hi); /* proto */
static void __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_2__dealloc__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_4sequence_from_quickcharge(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaks, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_peak); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_6make_sequence(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_8reset(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_10has_more(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_12get_next_value(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_14__len__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_16__iter__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_18__next__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_5lower___get__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_5lower_2__set__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_5upper___get__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_5upper_2__set__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_4sign___get__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_4sign_2__set__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_4size___get__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_4size_2__set__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_5index___get__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_5index_2__set__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_20__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_22__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_quick_charge(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16__pyx_fuse_0_0quick_charge(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peak_set, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_position_spec, int __pyx_v_min_charge, int __pyx_v_max_charge); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_18__pyx_fuse_0_1quick_charge(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peak_set, size_t __pyx_v_position_spec, int __pyx_v_min_charge, int __pyx_v_max_charge); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_20__pyx_fuse_0_2quick_charge(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peak_set, long __pyx_v_position_spec, int __pyx_v_min_charge, int __pyx_v_max_charge); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_22__pyx_fuse_1_0quick_charge(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_v_peak_set, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_position_spec, int __pyx_v_min_charge, int __pyx_v_max_charge); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_24__pyx_fuse_1_1quick_charge(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_v_peak_set, size_t __pyx_v_position_spec, int __pyx_v_min_charge, int __pyx_v_max_charge); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_26__pyx_fuse_1_2quick_charge(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_v_peak_set, long __pyx_v_position_spec, int __pyx_v_min_charge, int __pyx_v_max_charge); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_2_get_all_peak_charge_pairs(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_peak, double __pyx_v_error_tolerance, PyObject *__pyx_v_charge_range, int __pyx_v_left_search_limit, int __pyx_v_right_search_limit, int __pyx_v_recalculate_starting_peak, int __pyx_v_use_quick_charge); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_4_explore_local(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, PyObject *__pyx_v_peak, PyObject *__pyx_v_error_tolerance, PyObject *__pyx_v_charge_range, PyObject *__pyx_v_left_search_limit, PyObject *__pyx_v_right_search_limit, PyObject *__pyx_v_charge_carrier, PyObject *__pyx_v_truncate_after, PyObject *__pyx_v_ignore_below); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_6populate_graph(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, PyObject *__pyx_v_error_tolerance, PyObject *__pyx_v_charge_range, PyObject *__pyx_v_left_search_limit, PyObject *__pyx_v_right_search_limit, PyObject *__pyx_v_charge_carrier, PyObject *__pyx_v_truncate_after, PyObject *__pyx_v_ignore_below); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_8_make_deconvoluted_peak(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_fit, double __pyx_v_charge_carrier); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_10__pyx_unpickle_DeconvoluterBase(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_12__pyx_unpickle_AveragineDeconvoluterBase(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14__pyx_unpickle_MultiAveragineDeconvoluterBase(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_5numpy_7ndarray___getbuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /* proto */
static void __pyx_pf_5numpy_7ndarray_2__releasebuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info); /* proto */
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_float_5_;
static PyObject *__pyx_float_1eneg_3;
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_4;
static PyObject *__pyx_int_8;
static PyObject *__pyx_int_24102755;
static PyObject *__pyx_int_165289540;
static PyObject *__pyx_int_266290880;
static PyObject *__pyx_int_neg_1;
static double __pyx_k__2;
static double __pyx_k__3;
static double __pyx_k__4;
static double __pyx_k__5;
static double __pyx_k__15;
static PyObject *__pyx_k__17;
static PyObject *__pyx_k__18;
static PyObject *__pyx_k__19;
static PyObject *__pyx_k__20;
static PyObject *__pyx_k__21;
static PyObject *__pyx_k__22;
static PyObject *__pyx_k__23;
static PyObject *__pyx_k__24;
static PyObject *__pyx_slice_;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__7;
static PyObject *__pyx_tuple__11;
static PyObject *__pyx_tuple__12;
static PyObject *__pyx_tuple__13;
static PyObject *__pyx_tuple__14;
static PyObject *__pyx_tuple__16;
static PyObject *__pyx_tuple__25;
static PyObject *__pyx_tuple__26;
static PyObject *__pyx_tuple__27;
static PyObject *__pyx_tuple__28;
static PyObject *__pyx_tuple__29;
static PyObject *__pyx_tuple__30;
static PyObject *__pyx_tuple__31;
static PyObject *__pyx_tuple__32;
static PyObject *__pyx_tuple__34;
static PyObject *__pyx_tuple__36;
static PyObject *__pyx_tuple__38;
static PyObject *__pyx_tuple__40;
static PyObject *__pyx_tuple__42;
static PyObject *__pyx_tuple__44;
static PyObject *__pyx_tuple__46;
static PyObject *__pyx_codeobj__33;
static PyObject *__pyx_codeobj__35;
static PyObject *__pyx_codeobj__37;
static PyObject *__pyx_codeobj__39;
static PyObject *__pyx_codeobj__41;
static PyObject *__pyx_codeobj__43;
static PyObject *__pyx_codeobj__45;
static PyObject *__pyx_codeobj__47;
/* Late includes */

/* "ms_deisotope/_c/deconvoluter_base.pyx":42
 * 
 * 
 * cdef size_t count_missed_peaks(list peaklist):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i
 */

static size_t __pyx_f_12ms_deisotope_2_c_17deconvoluter_base_count_missed_peaks(PyObject *__pyx_v_peaklist) {
  size_t __pyx_v_i;
  int __pyx_v_n;
  int __pyx_v_t;
  void *__pyx_v_peak;
  size_t __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  size_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("count_missed_peaks", 0);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":49
 *         void* peak
 * 
 *     t = n = PyList_GET_SIZE(peaklist)             # <<<<<<<<<<<<<<
 *     for i in range(t):
 *         # peak = <FittedPeak>PyList_GET_ITEM(peaklist, i)
 */
  __pyx_t_1 = PyList_GET_SIZE(__pyx_v_peaklist);
  __pyx_v_t = __pyx_t_1;
  __pyx_v_n = __pyx_t_1;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":50
 * 
 *     t = n = PyList_GET_SIZE(peaklist)
 *     for i in range(t):             # <<<<<<<<<<<<<<
 *         # peak = <FittedPeak>PyList_GET_ITEM(peaklist, i)
 *         peak = PyList_GET_ITEM(peaklist, i)
 */
  __pyx_t_2 = __pyx_v_t;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":52
 *     for i in range(t):
 *         # peak = <FittedPeak>PyList_GET_ITEM(peaklist, i)
 *         peak = PyList_GET_ITEM(peaklist, i)             # <<<<<<<<<<<<<<
 *         if (<FittedPeak>peak).mz > 1 and (<FittedPeak>peak).intensity > 1:
 *             n -= 1
 */
    __pyx_v_peak = PyList_GET_ITEM(__pyx_v_peaklist, __pyx_v_i);

    /* "ms_deisotope/_c/deconvoluter_base.pyx":53
 *         # peak = <FittedPeak>PyList_GET_ITEM(peaklist, i)
 *         peak = PyList_GET_ITEM(peaklist, i)
 *         if (<FittedPeak>peak).mz > 1 and (<FittedPeak>peak).intensity > 1:             # <<<<<<<<<<<<<<
 *             n -= 1
 *     return n
 */
    __pyx_t_6 = ((((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_v_peak)->__pyx_base.mz > 1.0) != 0);
    if (__pyx_t_6) {
    } else {
      __pyx_t_5 = __pyx_t_6;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_6 = ((((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_v_peak)->__pyx_base.intensity > 1.0) != 0);
    __pyx_t_5 = __pyx_t_6;
    __pyx_L6_bool_binop_done:;
    if (__pyx_t_5) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":54
 *         peak = PyList_GET_ITEM(peaklist, i)
 *         if (<FittedPeak>peak).mz > 1 and (<FittedPeak>peak).intensity > 1:
 *             n -= 1             # <<<<<<<<<<<<<<
 *     return n
 * 
 */
      __pyx_v_n = (__pyx_v_n - 1);

      /* "ms_deisotope/_c/deconvoluter_base.pyx":53
 *         # peak = <FittedPeak>PyList_GET_ITEM(peaklist, i)
 *         peak = PyList_GET_ITEM(peaklist, i)
 *         if (<FittedPeak>peak).mz > 1 and (<FittedPeak>peak).intensity > 1:             # <<<<<<<<<<<<<<
 *             n -= 1
 *     return n
 */
    }
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":55
 *         if (<FittedPeak>peak).mz > 1 and (<FittedPeak>peak).intensity > 1:
 *             n -= 1
 *     return n             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_n;
  goto __pyx_L0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":42
 * 
 * 
 * cdef size_t count_missed_peaks(list peaklist):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/deconvoluter_base.pyx":58
 * 
 * 
 * cdef FittedPeak make_placeholder_peak(double mz):             # <<<<<<<<<<<<<<
 *     cdef FittedPeak peak = FittedPeak._create(
 *         mz, intensity=1.0, signal_to_noise=1.0, peak_count=-1, index=0, full_width_at_half_max=0.0,
 */

static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_make_placeholder_peak(double __pyx_v_mz) {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_peak = 0;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("make_placeholder_peak", 0);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":59
 * 
 * cdef FittedPeak make_placeholder_peak(double mz):
 *     cdef FittedPeak peak = FittedPeak._create(             # <<<<<<<<<<<<<<
 *         mz, intensity=1.0, signal_to_noise=1.0, peak_count=-1, index=0, full_width_at_half_max=0.0,
 *         area=1.0, left_width=0.0, right_width=0.0)
 */
  __pyx_t_1 = ((PyObject *)__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_FittedPeak->_create(__pyx_v_mz, 1.0, 1.0, 0.0, 0.0, 0.0, -1L, 0, 1.0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 59, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_peak = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":62
 *         mz, intensity=1.0, signal_to_noise=1.0, peak_count=-1, index=0, full_width_at_half_max=0.0,
 *         area=1.0, left_width=0.0, right_width=0.0)
 *     return peak             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_peak));
  __pyx_r = __pyx_v_peak;
  goto __pyx_L0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":58
 * 
 * 
 * cdef FittedPeak make_placeholder_peak(double mz):             # <<<<<<<<<<<<<<
 *     cdef FittedPeak peak = FittedPeak._create(
 *         mz, intensity=1.0, signal_to_noise=1.0, peak_count=-1, index=0, full_width_at_half_max=0.0,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.make_placeholder_peak", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/deconvoluter_base.pyx":65
 * 
 * 
 * cdef list drop_placeholders(list peaks):             # <<<<<<<<<<<<<<
 *     """Removes all placeholder peaks from an iterable of peaks
 * 
 */

static PyObject *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_drop_placeholders(PyObject *__pyx_v_peaks) {
  size_t __pyx_v_i;
  size_t __pyx_v_n;
  PyObject *__pyx_v_retained = 0;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_peak = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  PyObject *__pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("drop_placeholders", 0);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":80
 *         list retained
 *         FittedPeak peak
 *     retained = []             # <<<<<<<<<<<<<<
 *     n = PyList_GET_SIZE(peaks)
 *     for i in range(n):
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 80, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_retained = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":81
 *         FittedPeak peak
 *     retained = []
 *     n = PyList_GET_SIZE(peaks)             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         peak = <FittedPeak>PyList_GET_ITEM(peaks, i)
 */
  __pyx_v_n = PyList_GET_SIZE(__pyx_v_peaks);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":82
 *     retained = []
 *     n = PyList_GET_SIZE(peaks)
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         peak = <FittedPeak>PyList_GET_ITEM(peaks, i)
 *         if peak.mz > 1 and peak.intensity > 1:
 */
  __pyx_t_2 = __pyx_v_n;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":83
 *     n = PyList_GET_SIZE(peaks)
 *     for i in range(n):
 *         peak = <FittedPeak>PyList_GET_ITEM(peaks, i)             # <<<<<<<<<<<<<<
 *         if peak.mz > 1 and peak.intensity > 1:
 *             retained.append(peak)
 */
    __pyx_t_5 = PyList_GET_ITEM(__pyx_v_peaks, __pyx_v_i);
    __pyx_t_1 = ((PyObject *)__pyx_t_5);
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_peak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":84
 *     for i in range(n):
 *         peak = <FittedPeak>PyList_GET_ITEM(peaks, i)
 *         if peak.mz > 1 and peak.intensity > 1:             # <<<<<<<<<<<<<<
 *             retained.append(peak)
 *     return retained
 */
    __pyx_t_7 = ((__pyx_v_peak->__pyx_base.mz > 1.0) != 0);
    if (__pyx_t_7) {
    } else {
      __pyx_t_6 = __pyx_t_7;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_7 = ((__pyx_v_peak->__pyx_base.intensity > 1.0) != 0);
    __pyx_t_6 = __pyx_t_7;
    __pyx_L6_bool_binop_done:;
    if (__pyx_t_6) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":85
 *         peak = <FittedPeak>PyList_GET_ITEM(peaks, i)
 *         if peak.mz > 1 and peak.intensity > 1:
 *             retained.append(peak)             # <<<<<<<<<<<<<<
 *     return retained
 * 
 */
      __pyx_t_8 = __Pyx_PyList_Append(__pyx_v_retained, ((PyObject *)__pyx_v_peak)); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 85, __pyx_L1_error)

      /* "ms_deisotope/_c/deconvoluter_base.pyx":84
 *     for i in range(n):
 *         peak = <FittedPeak>PyList_GET_ITEM(peaks, i)
 *         if peak.mz > 1 and peak.intensity > 1:             # <<<<<<<<<<<<<<
 *             retained.append(peak)
 *     return retained
 */
    }
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":86
 *         if peak.mz > 1 and peak.intensity > 1:
 *             retained.append(peak)
 *     return retained             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_retained);
  __pyx_r = __pyx_v_retained;
  goto __pyx_L0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":65
 * 
 * 
 * cdef list drop_placeholders(list peaks):             # <<<<<<<<<<<<<<
 *     """Removes all placeholder peaks from an iterable of peaks
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.drop_placeholders", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_retained);
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/deconvoluter_base.pyx":89
 * 
 * 
 * cdef FittedPeak first_peak(peaks):             # <<<<<<<<<<<<<<
 *     """Get the first non-placeholder peak in a list of peaks
 * 
 */

static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_first_peak(PyObject *__pyx_v_peaks) {
  size_t __pyx_v_i;
  size_t __pyx_v_n;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_peak = 0;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  PyObject *__pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  __Pyx_RefNannySetupContext("first_peak", 0);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":103
 *         size_t i, n
 *         FittedPeak peak
 *     n = PyList_GET_SIZE(peaks)             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         peak = <FittedPeak>PyList_GET_ITEM(peaks, i)
 */
  __pyx_v_n = PyList_GET_SIZE(__pyx_v_peaks);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":104
 *         FittedPeak peak
 *     n = PyList_GET_SIZE(peaks)
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         peak = <FittedPeak>PyList_GET_ITEM(peaks, i)
 *         if peak.mz > 1 and peak.intensity > 1:
 */
  __pyx_t_1 = __pyx_v_n;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":105
 *     n = PyList_GET_SIZE(peaks)
 *     for i in range(n):
 *         peak = <FittedPeak>PyList_GET_ITEM(peaks, i)             # <<<<<<<<<<<<<<
 *         if peak.mz > 1 and peak.intensity > 1:
 *             return peak
 */
    __pyx_t_4 = PyList_GET_ITEM(__pyx_v_peaks, __pyx_v_i);
    __pyx_t_5 = ((PyObject *)__pyx_t_4);
    __Pyx_INCREF(__pyx_t_5);
    __Pyx_XDECREF_SET(__pyx_v_peak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_5));
    __pyx_t_5 = 0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":106
 *     for i in range(n):
 *         peak = <FittedPeak>PyList_GET_ITEM(peaks, i)
 *         if peak.mz > 1 and peak.intensity > 1:             # <<<<<<<<<<<<<<
 *             return peak
 *     return None
 */
    __pyx_t_7 = ((__pyx_v_peak->__pyx_base.mz > 1.0) != 0);
    if (__pyx_t_7) {
    } else {
      __pyx_t_6 = __pyx_t_7;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_7 = ((__pyx_v_peak->__pyx_base.intensity > 1.0) != 0);
    __pyx_t_6 = __pyx_t_7;
    __pyx_L6_bool_binop_done:;
    if (__pyx_t_6) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":107
 *         peak = <FittedPeak>PyList_GET_ITEM(peaks, i)
 *         if peak.mz > 1 and peak.intensity > 1:
 *             return peak             # <<<<<<<<<<<<<<
 *     return None
 * 
 */
      __Pyx_XDECREF(((PyObject *)__pyx_r));
      __Pyx_INCREF(((PyObject *)__pyx_v_peak));
      __pyx_r = __pyx_v_peak;
      goto __pyx_L0;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":106
 *     for i in range(n):
 *         peak = <FittedPeak>PyList_GET_ITEM(peaks, i)
 *         if peak.mz > 1 and peak.intensity > 1:             # <<<<<<<<<<<<<<
 *             return peak
 *     return None
 */
    }
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":108
 *         if peak.mz > 1 and peak.intensity > 1:
 *             return peak
 *     return None             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __pyx_r = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)Py_None); __Pyx_INCREF(Py_None);
  goto __pyx_L0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":89
 * 
 * 
 * cdef FittedPeak first_peak(peaks):             # <<<<<<<<<<<<<<
 *     """Get the first non-placeholder peak in a list of peaks
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/deconvoluter_base.pyx":139
 *         Produce extra logging information
 *     """
 *     def __init__(self, use_subtraction=False, scale_method="sum", merge_isobaric_peaks=True,             # <<<<<<<<<<<<<<
 *                   minimum_intensity=5., *args, **kwargs):
 *         self.use_subtraction = use_subtraction
 */

/* Python wrapper */
static int __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_use_subtraction = 0;
  PyObject *__pyx_v_scale_method = 0;
  PyObject *__pyx_v_merge_isobaric_peaks = 0;
  PyObject *__pyx_v_minimum_intensity = 0;
  CYTHON_UNUSED PyObject *__pyx_v_args = 0;
  CYTHON_UNUSED PyObject *__pyx_v_kwargs = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  __pyx_v_kwargs = PyDict_New(); if (unlikely(!__pyx_v_kwargs)) return -1;
  __Pyx_GOTREF(__pyx_v_kwargs);
  if (PyTuple_GET_SIZE(__pyx_args) > 4) {
    __pyx_v_args = PyTuple_GetSlice(__pyx_args, 4, PyTuple_GET_SIZE(__pyx_args));
    if (unlikely(!__pyx_v_args)) {
      __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
      __Pyx_RefNannyFinishContext();
      return -1;
    }
    __Pyx_GOTREF(__pyx_v_args);
  } else {
    __pyx_v_args = __pyx_empty_tuple; __Pyx_INCREF(__pyx_empty_tuple);
  }
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_use_subtraction,&__pyx_n_s_scale_method,&__pyx_n_s_merge_isobaric_peaks,&__pyx_n_s_minimum_intensity,0};
    PyObject* values[4] = {0,0,0,0};
    values[0] = ((PyObject *)Py_False);
    values[1] = ((PyObject *)__pyx_n_s_sum);
    values[2] = ((PyObject *)Py_True);
    values[3] = ((PyObject *)__pyx_float_5_);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        default:
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_use_subtraction);
          if (value) { values[0] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_scale_method);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_merge_isobaric_peaks);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_minimum_intensity);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t used_pos_args = (pos_args < 4) ? pos_args : 4;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, __pyx_v_kwargs, values, used_pos_args, "__init__") < 0)) __PYX_ERR(0, 139, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        default:
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
      }
    }
    __pyx_v_use_subtraction = values[0];
    __pyx_v_scale_method = values[1];
    __pyx_v_merge_isobaric_peaks = values[2];
    __pyx_v_minimum_intensity = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_DECREF(__pyx_v_args); __pyx_v_args = 0;
  __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.DeconvoluterBase.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase___init__(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *)__pyx_v_self), __pyx_v_use_subtraction, __pyx_v_scale_method, __pyx_v_merge_isobaric_peaks, __pyx_v_minimum_intensity, __pyx_v_args, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase___init__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, PyObject *__pyx_v_use_subtraction, PyObject *__pyx_v_scale_method, PyObject *__pyx_v_merge_isobaric_peaks, PyObject *__pyx_v_minimum_intensity, CYTHON_UNUSED PyObject *__pyx_v_args, CYTHON_UNUSED PyObject *__pyx_v_kwargs) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  double __pyx_t_3;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":141
 *     def __init__(self, use_subtraction=False, scale_method="sum", merge_isobaric_peaks=True,
 *                   minimum_intensity=5., *args, **kwargs):
 *         self.use_subtraction = use_subtraction             # <<<<<<<<<<<<<<
 *         self.scale_method = scale_method
 *         self.merge_isobaric_peaks = merge_isobaric_peaks
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_use_subtraction); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 141, __pyx_L1_error)
  __pyx_v_self->use_subtraction = __pyx_t_1;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":142
 *                   minimum_intensity=5., *args, **kwargs):
 *         self.use_subtraction = use_subtraction
 *         self.scale_method = scale_method             # <<<<<<<<<<<<<<
 *         self.merge_isobaric_peaks = merge_isobaric_peaks
 *         self.minimum_intensity = minimum_intensity
 */
  if (!(likely(PyString_CheckExact(__pyx_v_scale_method))||((__pyx_v_scale_method) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_v_scale_method)->tp_name), 0))) __PYX_ERR(0, 142, __pyx_L1_error)
  __pyx_t_2 = __pyx_v_scale_method;
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_self->scale_method);
  __Pyx_DECREF(__pyx_v_self->scale_method);
  __pyx_v_self->scale_method = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":143
 *         self.use_subtraction = use_subtraction
 *         self.scale_method = scale_method
 *         self.merge_isobaric_peaks = merge_isobaric_peaks             # <<<<<<<<<<<<<<
 *         self.minimum_intensity = minimum_intensity
 *         self._slice_cache = {}
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_merge_isobaric_peaks); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 143, __pyx_L1_error)
  __pyx_v_self->merge_isobaric_peaks = __pyx_t_1;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":144
 *         self.scale_method = scale_method
 *         self.merge_isobaric_peaks = merge_isobaric_peaks
 *         self.minimum_intensity = minimum_intensity             # <<<<<<<<<<<<<<
 *         self._slice_cache = {}
 * 
 */
  __pyx_t_3 = __pyx_PyFloat_AsDouble(__pyx_v_minimum_intensity); if (unlikely((__pyx_t_3 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 144, __pyx_L1_error)
  __pyx_v_self->minimum_intensity = __pyx_t_3;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":145
 *         self.merge_isobaric_peaks = merge_isobaric_peaks
 *         self.minimum_intensity = minimum_intensity
 *         self._slice_cache = {}             # <<<<<<<<<<<<<<
 * 
 *     cpdef PeakSet between(self, double m1, double m2):
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_self->_slice_cache);
  __Pyx_DECREF(__pyx_v_self->_slice_cache);
  __pyx_v_self->_slice_cache = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":139
 *         Produce extra logging information
 *     """
 *     def __init__(self, use_subtraction=False, scale_method="sum", merge_isobaric_peaks=True,             # <<<<<<<<<<<<<<
 *                   minimum_intensity=5., *args, **kwargs):
 *         self.use_subtraction = use_subtraction
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.DeconvoluterBase.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/deconvoluter_base.pyx":147
 *         self._slice_cache = {}
 * 
 *     cpdef PeakSet between(self, double m1, double m2):             # <<<<<<<<<<<<<<
 *         cdef:
 *             tuple key
 */

static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_3between(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_between(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, double __pyx_v_m1, double __pyx_v_m2, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_key = 0;
  PyObject *__pyx_v_p;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_region = 0;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  __Pyx_RefNannySetupContext("between", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_between); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 147, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_3between)) {
        __Pyx_XDECREF(((PyObject *)__pyx_r));
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_m1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 147, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_m2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 147, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_5 = __pyx_t_1; __pyx_t_6 = NULL;
        __pyx_t_7 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_6)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_6);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
            __pyx_t_7 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_5)) {
          PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_3, __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 147, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
          PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_3, __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 147, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 147, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          if (__pyx_t_6) {
            __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_t_3);
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_t_4);
          __pyx_t_3 = 0;
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 147, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        }
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet))))) __PYX_ERR(0, 147, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":152
 *             PyObject* p
 *             PeakSet region
 *         key = (m1, m2)             # <<<<<<<<<<<<<<
 *         p = PyDict_GetItem(self._slice_cache, key)
 *         if p == NULL:
 */
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_m1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_m2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_v_key = ((PyObject*)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":153
 *             PeakSet region
 *         key = (m1, m2)
 *         p = PyDict_GetItem(self._slice_cache, key)             # <<<<<<<<<<<<<<
 *         if p == NULL:
 *             region = self.peaklist._between(m1, m2)
 */
  __pyx_t_5 = __pyx_v_self->_slice_cache;
  __Pyx_INCREF(__pyx_t_5);
  __pyx_v_p = PyDict_GetItem(__pyx_t_5, __pyx_v_key);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":154
 *         key = (m1, m2)
 *         p = PyDict_GetItem(self._slice_cache, key)
 *         if p == NULL:             # <<<<<<<<<<<<<<
 *             region = self.peaklist._between(m1, m2)
 *             PyDict_SetItem(self._slice_cache, key, region)
 */
  __pyx_t_9 = ((__pyx_v_p == NULL) != 0);
  if (__pyx_t_9) {

    /* "ms_deisotope/_c/deconvoluter_base.pyx":155
 *         p = PyDict_GetItem(self._slice_cache, key)
 *         if p == NULL:
 *             region = self.peaklist._between(m1, m2)             # <<<<<<<<<<<<<<
 *             PyDict_SetItem(self._slice_cache, key, region)
 *             return region
 */
    __pyx_t_5 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_self->peaklist->__pyx_vtab)->_between(__pyx_v_self->peaklist, __pyx_v_m1, __pyx_v_m2)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 155, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_v_region = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_t_5);
    __pyx_t_5 = 0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":156
 *         if p == NULL:
 *             region = self.peaklist._between(m1, m2)
 *             PyDict_SetItem(self._slice_cache, key, region)             # <<<<<<<<<<<<<<
 *             return region
 *         else:
 */
    __pyx_t_5 = __pyx_v_self->_slice_cache;
    __Pyx_INCREF(__pyx_t_5);
    __pyx_t_7 = PyDict_SetItem(__pyx_t_5, __pyx_v_key, ((PyObject *)__pyx_v_region)); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 156, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":157
 *             region = self.peaklist._between(m1, m2)
 *             PyDict_SetItem(self._slice_cache, key, region)
 *             return region             # <<<<<<<<<<<<<<
 *         else:
 *             region = <PeakSet>p
 */
    __Pyx_XDECREF(((PyObject *)__pyx_r));
    __Pyx_INCREF(((PyObject *)__pyx_v_region));
    __pyx_r = __pyx_v_region;
    goto __pyx_L0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":154
 *         key = (m1, m2)
 *         p = PyDict_GetItem(self._slice_cache, key)
 *         if p == NULL:             # <<<<<<<<<<<<<<
 *             region = self.peaklist._between(m1, m2)
 *             PyDict_SetItem(self._slice_cache, key, region)
 */
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":159
 *             return region
 *         else:
 *             region = <PeakSet>p             # <<<<<<<<<<<<<<
 *             return region
 * 
 */
  /*else*/ {
    __pyx_t_5 = ((PyObject *)__pyx_v_p);
    __Pyx_INCREF(__pyx_t_5);
    __pyx_v_region = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_t_5);
    __pyx_t_5 = 0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":160
 *         else:
 *             region = <PeakSet>p
 *             return region             # <<<<<<<<<<<<<<
 * 
 *     cpdef FittedPeak has_peak(self, double mz, double error_tolerance):
 */
    __Pyx_XDECREF(((PyObject *)__pyx_r));
    __Pyx_INCREF(((PyObject *)__pyx_v_region));
    __pyx_r = __pyx_v_region;
    goto __pyx_L0;
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":147
 *         self._slice_cache = {}
 * 
 *     cpdef PeakSet between(self, double m1, double m2):             # <<<<<<<<<<<<<<
 *         cdef:
 *             tuple key
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.DeconvoluterBase.between", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF((PyObject *)__pyx_v_region);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_3between(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_2between[] = "DeconvoluterBase.between(self, double m1, double m2) -> PeakSet";
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_3between(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_m1;
  double __pyx_v_m2;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("between (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_m1,&__pyx_n_s_m2,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m1)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("between", 1, 2, 2, 1); __PYX_ERR(0, 147, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "between") < 0)) __PYX_ERR(0, 147, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_m1 = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_m1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 147, __pyx_L3_error)
    __pyx_v_m2 = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_m2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 147, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("between", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 147, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.DeconvoluterBase.between", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_2between(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *)__pyx_v_self), __pyx_v_m1, __pyx_v_m2);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_2between(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, double __pyx_v_m1, double __pyx_v_m2) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("between", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_between(__pyx_v_self, __pyx_v_m1, __pyx_v_m2, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.DeconvoluterBase.between", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/deconvoluter_base.pyx":162
 *             return region
 * 
 *     cpdef FittedPeak has_peak(self, double mz, double error_tolerance):             # <<<<<<<<<<<<<<
 *         return self._has_peak(mz, error_tolerance)
 * 
 */

static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_5has_peak(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_has_peak(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, double __pyx_v_mz, double __pyx_v_error_tolerance, int __pyx_skip_dispatch) {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("has_peak", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_has_peak); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 162, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_5has_peak)) {
        __Pyx_XDECREF(((PyObject *)__pyx_r));
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_mz); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 162, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_error_tolerance); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 162, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_5 = __pyx_t_1; __pyx_t_6 = NULL;
        __pyx_t_7 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_6)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_6);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
            __pyx_t_7 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_5)) {
          PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_3, __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 162, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
          PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_3, __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 162, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 162, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          if (__pyx_t_6) {
            __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_t_3);
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_t_4);
          __pyx_t_3 = 0;
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 162, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        }
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak))))) __PYX_ERR(0, 162, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":163
 * 
 *     cpdef FittedPeak has_peak(self, double mz, double error_tolerance):
 *         return self._has_peak(mz, error_tolerance)             # <<<<<<<<<<<<<<
 * 
 *     @cython.final
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __pyx_t_1 = ((PyObject *)__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__has_peak(__pyx_v_self, __pyx_v_mz, __pyx_v_error_tolerance)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":162
 *             return region
 * 
 *     cpdef FittedPeak has_peak(self, double mz, double error_tolerance):             # <<<<<<<<<<<<<<
 *         return self._has_peak(mz, error_tolerance)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.DeconvoluterBase.has_peak", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_5has_peak(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_4has_peak[] = "DeconvoluterBase.has_peak(self, double mz, double error_tolerance) -> FittedPeak";
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_5has_peak(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_mz;
  double __pyx_v_error_tolerance;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("has_peak (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_mz,&__pyx_n_s_error_tolerance,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mz)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_error_tolerance)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("has_peak", 1, 2, 2, 1); __PYX_ERR(0, 162, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "has_peak") < 0)) __PYX_ERR(0, 162, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_mz = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_mz == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 162, __pyx_L3_error)
    __pyx_v_error_tolerance = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_error_tolerance == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 162, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("has_peak", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 162, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.DeconvoluterBase.has_peak", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_4has_peak(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *)__pyx_v_self), __pyx_v_mz, __pyx_v_error_tolerance);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_4has_peak(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, double __pyx_v_mz, double __pyx_v_error_tolerance) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("has_peak", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_has_peak(__pyx_v_self, __pyx_v_mz, __pyx_v_error_tolerance, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.DeconvoluterBase.has_peak", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/deconvoluter_base.pyx":166
 * 
 *     @cython.final
 *     cdef FittedPeak _has_peak(self, double mz, double error_tolerance):             # <<<<<<<<<<<<<<
 *         peak = self.peaklist._has_peak(mz, error_tolerance)
 *         if peak is None or peak.intensity < self.minimum_intensity:
 */

static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__has_peak(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, double __pyx_v_mz, double __pyx_v_error_tolerance) {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_peak = NULL;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet__has_peak __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  __Pyx_RefNannySetupContext("_has_peak", 0);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":167
 *     @cython.final
 *     cdef FittedPeak _has_peak(self, double mz, double error_tolerance):
 *         peak = self.peaklist._has_peak(mz, error_tolerance)             # <<<<<<<<<<<<<<
 *         if peak is None or peak.intensity < self.minimum_intensity:
 *             return make_placeholder_peak(mz)
 */
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.tolerance = __pyx_v_error_tolerance;
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_self->peaklist->__pyx_vtab)->_has_peak(__pyx_v_self->peaklist, __pyx_v_mz, &__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_peak = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":168
 *     cdef FittedPeak _has_peak(self, double mz, double error_tolerance):
 *         peak = self.peaklist._has_peak(mz, error_tolerance)
 *         if peak is None or peak.intensity < self.minimum_intensity:             # <<<<<<<<<<<<<<
 *             return make_placeholder_peak(mz)
 *         return peak
 */
  __pyx_t_4 = (((PyObject *)__pyx_v_peak) == Py_None);
  __pyx_t_5 = (__pyx_t_4 != 0);
  if (!__pyx_t_5) {
  } else {
    __pyx_t_3 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_5 = ((__pyx_v_peak->__pyx_base.intensity < __pyx_v_self->minimum_intensity) != 0);
  __pyx_t_3 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_3) {

    /* "ms_deisotope/_c/deconvoluter_base.pyx":169
 *         peak = self.peaklist._has_peak(mz, error_tolerance)
 *         if peak is None or peak.intensity < self.minimum_intensity:
 *             return make_placeholder_peak(mz)             # <<<<<<<<<<<<<<
 *         return peak
 * 
 */
    __Pyx_XDECREF(((PyObject *)__pyx_r));
    __pyx_t_1 = ((PyObject *)__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_make_placeholder_peak(__pyx_v_mz)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 169, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_1);
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":168
 *     cdef FittedPeak _has_peak(self, double mz, double error_tolerance):
 *         peak = self.peaklist._has_peak(mz, error_tolerance)
 *         if peak is None or peak.intensity < self.minimum_intensity:             # <<<<<<<<<<<<<<
 *             return make_placeholder_peak(mz)
 *         return peak
 */
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":170
 *         if peak is None or peak.intensity < self.minimum_intensity:
 *             return make_placeholder_peak(mz)
 *         return peak             # <<<<<<<<<<<<<<
 * 
 *     cpdef list match_theoretical_isotopic_distribution(self, list theoretical_distribution, double error_tolerance=2e-5):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_peak));
  __pyx_r = __pyx_v_peak;
  goto __pyx_L0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":166
 * 
 *     @cython.final
 *     cdef FittedPeak _has_peak(self, double mz, double error_tolerance):             # <<<<<<<<<<<<<<
 *         peak = self.peaklist._has_peak(mz, error_tolerance)
 *         if peak is None or peak.intensity < self.minimum_intensity:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.DeconvoluterBase._has_peak", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/deconvoluter_base.pyx":172
 *         return peak
 * 
 *     cpdef list match_theoretical_isotopic_distribution(self, list theoretical_distribution, double error_tolerance=2e-5):             # <<<<<<<<<<<<<<
 *         cdef:
 *             list experimental_distribution
 */

static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_7match_theoretical_isotopic_distribution(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_match_theoretical_isotopic_distribution(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, PyObject *__pyx_v_theoretical_distribution, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_match_theoretical_isotopic_distribution *__pyx_optional_args) {
  double __pyx_v_error_tolerance = ((double)2e-5);
  PyObject *__pyx_v_experimental_distribution = 0;
  size_t __pyx_v_i;
  size_t __pyx_v_n;
  double __pyx_v_mz;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_peak = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  size_t __pyx_t_8;
  size_t __pyx_t_9;
  size_t __pyx_t_10;
  PyObject *__pyx_t_11;
  double __pyx_t_12;
  __Pyx_RefNannySetupContext("match_theoretical_isotopic_distribution", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_error_tolerance = __pyx_optional_args->error_tolerance;
    }
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_match_theoretical_isotopic_distr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 172, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_7match_theoretical_isotopic_distribution)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_error_tolerance); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 172, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_theoretical_distribution, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 172, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_theoretical_distribution, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 172, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 172, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(__pyx_v_theoretical_distribution);
          __Pyx_GIVEREF(__pyx_v_theoretical_distribution);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_theoretical_distribution);
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 172, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 172, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":179
 *             FittedPeak peak
 * 
 *         n = PyList_GET_SIZE(theoretical_distribution)             # <<<<<<<<<<<<<<
 *         experimental_distribution = PyList_New(n)
 * 
 */
  __pyx_v_n = PyList_GET_SIZE(__pyx_v_theoretical_distribution);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":180
 * 
 *         n = PyList_GET_SIZE(theoretical_distribution)
 *         experimental_distribution = PyList_New(n)             # <<<<<<<<<<<<<<
 * 
 *         for i in range(n):
 */
  __pyx_t_1 = PyList_New(__pyx_v_n); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_experimental_distribution = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":182
 *         experimental_distribution = PyList_New(n)
 * 
 *         for i in range(n):             # <<<<<<<<<<<<<<
 *             mz = (<TheoreticalPeak>PyList_GET_ITEM(theoretical_distribution, i)).mz
 *             peak = self._has_peak(mz, error_tolerance)
 */
  __pyx_t_8 = __pyx_v_n;
  __pyx_t_9 = __pyx_t_8;
  for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_9; __pyx_t_10+=1) {
    __pyx_v_i = __pyx_t_10;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":183
 * 
 *         for i in range(n):
 *             mz = (<TheoreticalPeak>PyList_GET_ITEM(theoretical_distribution, i)).mz             # <<<<<<<<<<<<<<
 *             peak = self._has_peak(mz, error_tolerance)
 *             Py_INCREF(peak)
 */
    __pyx_t_11 = PyList_GET_ITEM(__pyx_v_theoretical_distribution, __pyx_v_i);
    __pyx_t_12 = ((struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *)__pyx_t_11)->mz;
    __pyx_v_mz = __pyx_t_12;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":184
 *         for i in range(n):
 *             mz = (<TheoreticalPeak>PyList_GET_ITEM(theoretical_distribution, i)).mz
 *             peak = self._has_peak(mz, error_tolerance)             # <<<<<<<<<<<<<<
 *             Py_INCREF(peak)
 *             PyList_SET_ITEM(experimental_distribution, i, peak)
 */
    __pyx_t_1 = ((PyObject *)__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__has_peak(__pyx_v_self, __pyx_v_mz, __pyx_v_error_tolerance)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 184, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_peak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":185
 *             mz = (<TheoreticalPeak>PyList_GET_ITEM(theoretical_distribution, i)).mz
 *             peak = self._has_peak(mz, error_tolerance)
 *             Py_INCREF(peak)             # <<<<<<<<<<<<<<
 *             PyList_SET_ITEM(experimental_distribution, i, peak)
 * 
 */
    Py_INCREF(((PyObject *)__pyx_v_peak));

    /* "ms_deisotope/_c/deconvoluter_base.pyx":186
 *             peak = self._has_peak(mz, error_tolerance)
 *             Py_INCREF(peak)
 *             PyList_SET_ITEM(experimental_distribution, i, peak)             # <<<<<<<<<<<<<<
 * 
 *         return experimental_distribution
 */
    PyList_SET_ITEM(__pyx_v_experimental_distribution, __pyx_v_i, ((PyObject *)__pyx_v_peak));
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":188
 *             PyList_SET_ITEM(experimental_distribution, i, peak)
 * 
 *         return experimental_distribution             # <<<<<<<<<<<<<<
 * 
 *     cpdef scale_theoretical_distribution(self, TheoreticalIsotopicPattern theoretical_distribution,
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_experimental_distribution);
  __pyx_r = __pyx_v_experimental_distribution;
  goto __pyx_L0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":172
 *         return peak
 * 
 *     cpdef list match_theoretical_isotopic_distribution(self, list theoretical_distribution, double error_tolerance=2e-5):             # <<<<<<<<<<<<<<
 *         cdef:
 *             list experimental_distribution
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.DeconvoluterBase.match_theoretical_isotopic_distribution", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_experimental_distribution);
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_7match_theoretical_isotopic_distribution(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_6match_theoretical_isotopic_distribution[] = "DeconvoluterBase.match_theoretical_isotopic_distribution(self, list theoretical_distribution, double error_tolerance=2e-5) -> list";
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_7match_theoretical_isotopic_distribution(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_theoretical_distribution = 0;
  double __pyx_v_error_tolerance;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("match_theoretical_isotopic_distribution (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_theoretical_distribution,&__pyx_n_s_error_tolerance,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_theoretical_distribution)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_error_tolerance);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "match_theoretical_isotopic_distribution") < 0)) __PYX_ERR(0, 172, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_theoretical_distribution = ((PyObject*)values[0]);
    if (values[1]) {
      __pyx_v_error_tolerance = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_error_tolerance == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 172, __pyx_L3_error)
    } else {
      __pyx_v_error_tolerance = ((double)2e-5);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("match_theoretical_isotopic_distribution", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 172, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.DeconvoluterBase.match_theoretical_isotopic_distribution", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_theoretical_distribution), (&PyList_Type), 1, "theoretical_distribution", 1))) __PYX_ERR(0, 172, __pyx_L1_error)
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_6match_theoretical_isotopic_distribution(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *)__pyx_v_self), __pyx_v_theoretical_distribution, __pyx_v_error_tolerance);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_6match_theoretical_isotopic_distribution(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, PyObject *__pyx_v_theoretical_distribution, double __pyx_v_error_tolerance) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_match_theoretical_isotopic_distribution __pyx_t_2;
  __Pyx_RefNannySetupContext("match_theoretical_isotopic_distribution", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.error_tolerance = __pyx_v_error_tolerance;
  __pyx_t_1 = __pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase->match_theoretical_isotopic_distribution(__pyx_v_self, __pyx_v_theoretical_distribution, 1, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.DeconvoluterBase.match_theoretical_isotopic_distribution", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/deconvoluter_base.pyx":190
 *         return experimental_distribution
 * 
 *     cpdef scale_theoretical_distribution(self, TheoreticalIsotopicPattern theoretical_distribution,             # <<<<<<<<<<<<<<
 *                                          list experimental_distribution):
 *         return theoretical_distribution.scale(experimental_distribution, self.scale_method)
 */

static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_9scale_theoretical_distribution(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_scale_theoretical_distribution(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *__pyx_v_theoretical_distribution, PyObject *__pyx_v_experimental_distribution, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_scale __pyx_t_7;
  __Pyx_RefNannySetupContext("scale_theoretical_distribution", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_scale_theoretical_distribution); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 190, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_9scale_theoretical_distribution)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_4, ((PyObject *)__pyx_v_theoretical_distribution), __pyx_v_experimental_distribution};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 190, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_4, ((PyObject *)__pyx_v_theoretical_distribution), __pyx_v_experimental_distribution};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 190, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 190, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          if (__pyx_t_4) {
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_theoretical_distribution));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_theoretical_distribution));
          PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, ((PyObject *)__pyx_v_theoretical_distribution));
          __Pyx_INCREF(__pyx_v_experimental_distribution);
          __Pyx_GIVEREF(__pyx_v_experimental_distribution);
          PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_v_experimental_distribution);
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 190, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":192
 *     cpdef scale_theoretical_distribution(self, TheoreticalIsotopicPattern theoretical_distribution,
 *                                          list experimental_distribution):
 *         return theoretical_distribution.scale(experimental_distribution, self.scale_method)             # <<<<<<<<<<<<<<
 * 
 *     cpdef IsotopicFitRecord _evaluate_theoretical_distribution(self, list experimental,
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_v_self->scale_method;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_7.__pyx_n = 1;
  __pyx_t_7.method = ((PyObject*)__pyx_t_1);
  __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *)__pyx_v_theoretical_distribution->__pyx_vtab)->scale(__pyx_v_theoretical_distribution, __pyx_v_experimental_distribution, 0, &__pyx_t_7)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":190
 *         return experimental_distribution
 * 
 *     cpdef scale_theoretical_distribution(self, TheoreticalIsotopicPattern theoretical_distribution,             # <<<<<<<<<<<<<<
 *                                          list experimental_distribution):
 *         return theoretical_distribution.scale(experimental_distribution, self.scale_method)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.DeconvoluterBase.scale_theoretical_distribution", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_9scale_theoretical_distribution(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_8scale_theoretical_distribution[] = "DeconvoluterBase.scale_theoretical_distribution(self, TheoreticalIsotopicPattern theoretical_distribution, list experimental_distribution)";
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_9scale_theoretical_distribution(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *__pyx_v_theoretical_distribution = 0;
  PyObject *__pyx_v_experimental_distribution = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("scale_theoretical_distribution (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_theoretical_distribution,&__pyx_n_s_experimental_distribution,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_theoretical_distribution)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_experimental_distribution)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scale_theoretical_distribution", 1, 2, 2, 1); __PYX_ERR(0, 190, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "scale_theoretical_distribution") < 0)) __PYX_ERR(0, 190, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_theoretical_distribution = ((struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *)values[0]);
    __pyx_v_experimental_distribution = ((PyObject*)values[1]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("scale_theoretical_distribution", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 190, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.DeconvoluterBase.scale_theoretical_distribution", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_theoretical_distribution), __pyx_ptype_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern, 1, "theoretical_distribution", 0))) __PYX_ERR(0, 190, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_experimental_distribution), (&PyList_Type), 1, "experimental_distribution", 1))) __PYX_ERR(0, 191, __pyx_L1_error)
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_8scale_theoretical_distribution(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *)__pyx_v_self), __pyx_v_theoretical_distribution, __pyx_v_experimental_distribution);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_8scale_theoretical_distribution(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *__pyx_v_theoretical_distribution, PyObject *__pyx_v_experimental_distribution) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("scale_theoretical_distribution", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_scale_theoretical_distribution(__pyx_v_self, __pyx_v_theoretical_distribution, __pyx_v_experimental_distribution, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 190, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.DeconvoluterBase.scale_theoretical_distribution", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/deconvoluter_base.pyx":194
 *         return theoretical_distribution.scale(experimental_distribution, self.scale_method)
 * 
 *     cpdef IsotopicFitRecord _evaluate_theoretical_distribution(self, list experimental,             # <<<<<<<<<<<<<<
 *                                                                TheoreticalIsotopicPattern theoretical,
 *                                                                FittedPeak peak, int charge):
 */

static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_11_evaluate_theoretical_distribution(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__evaluate_theoretical_distribution(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, PyObject *__pyx_v_experimental, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *__pyx_v_theoretical, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_peak, int __pyx_v_charge, int __pyx_skip_dispatch) {
  double __pyx_v_score;
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_fit = 0;
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern__scale __pyx_t_8;
  struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord__create __pyx_t_9;
  __Pyx_RefNannySetupContext("_evaluate_theoretical_distribution", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_evaluate_theoretical_distributi); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 194, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_11_evaluate_theoretical_distribution)) {
        __Pyx_XDECREF(((PyObject *)__pyx_r));
        __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_charge); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 194, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[5] = {__pyx_t_5, __pyx_v_experimental, ((PyObject *)__pyx_v_theoretical), ((PyObject *)__pyx_v_peak), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 4+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 194, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[5] = {__pyx_t_5, __pyx_v_experimental, ((PyObject *)__pyx_v_theoretical), ((PyObject *)__pyx_v_peak), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 4+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 194, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(4+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 194, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(__pyx_v_experimental);
          __Pyx_GIVEREF(__pyx_v_experimental);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_experimental);
          __Pyx_INCREF(((PyObject *)__pyx_v_theoretical));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_theoretical));
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, ((PyObject *)__pyx_v_theoretical));
          __Pyx_INCREF(((PyObject *)__pyx_v_peak));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_peak));
          PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, ((PyObject *)__pyx_v_peak));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 3+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 194, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitRecord))))) __PYX_ERR(0, 194, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":218
 *             double score
 *             IsotopicFitRecord fit
 *         theoretical._scale(experimental, self.scale_method)             # <<<<<<<<<<<<<<
 *         score = self.scorer._evaluate(self.peaklist, experimental, theoretical.peaklist)
 *         fit = IsotopicFitRecord._create(peak, score, charge, theoretical, experimental, None, 0)
 */
  __pyx_t_1 = __pyx_v_self->scale_method;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_8.__pyx_n = 1;
  __pyx_t_8.method = ((PyObject*)__pyx_t_1);
  __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *)__pyx_v_theoretical->__pyx_vtab)->_scale(__pyx_v_theoretical, __pyx_v_experimental, &__pyx_t_8)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 218, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":219
 *             IsotopicFitRecord fit
 *         theoretical._scale(experimental, self.scale_method)
 *         score = self.scorer._evaluate(self.peaklist, experimental, theoretical.peaklist)             # <<<<<<<<<<<<<<
 *         fit = IsotopicFitRecord._create(peak, score, charge, theoretical, experimental, None, 0)
 *         fit.missed_peaks = count_missed_peaks(fit.experimental)
 */
  __pyx_t_2 = ((PyObject *)__pyx_v_self->peaklist);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_1 = __pyx_v_theoretical->peaklist;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_score = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *)__pyx_v_self->scorer->__pyx_vtab)->_evaluate(__pyx_v_self->scorer, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_t_2), __pyx_v_experimental, ((PyObject*)__pyx_t_1), 0);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":220
 *         theoretical._scale(experimental, self.scale_method)
 *         score = self.scorer._evaluate(self.peaklist, experimental, theoretical.peaklist)
 *         fit = IsotopicFitRecord._create(peak, score, charge, theoretical, experimental, None, 0)             # <<<<<<<<<<<<<<
 *         fit.missed_peaks = count_missed_peaks(fit.experimental)
 *         return fit
 */
  __pyx_t_9.__pyx_n = 2;
  __pyx_t_9.data = Py_None;
  __pyx_t_9.missed_peaks = 0;
  __pyx_t_1 = ((PyObject *)__pyx_vtabptr_12ms_deisotope_2_c_7scoring_IsotopicFitRecord->_create(__pyx_v_peak, __pyx_v_score, __pyx_v_charge, __pyx_v_theoretical, __pyx_v_experimental, &__pyx_t_9)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 220, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_fit = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":221
 *         score = self.scorer._evaluate(self.peaklist, experimental, theoretical.peaklist)
 *         fit = IsotopicFitRecord._create(peak, score, charge, theoretical, experimental, None, 0)
 *         fit.missed_peaks = count_missed_peaks(fit.experimental)             # <<<<<<<<<<<<<<
 *         return fit
 * 
 */
  __pyx_t_1 = __pyx_v_fit->experimental;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_fit->missed_peaks = __pyx_f_12ms_deisotope_2_c_17deconvoluter_base_count_missed_peaks(((PyObject*)__pyx_t_1));
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":222
 *         fit = IsotopicFitRecord._create(peak, score, charge, theoretical, experimental, None, 0)
 *         fit.missed_peaks = count_missed_peaks(fit.experimental)
 *         return fit             # <<<<<<<<<<<<<<
 * 
 *     cpdef list fit_incremental_truncation(self, IsotopicFitRecord seed_fit, double lower_bound):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_fit));
  __pyx_r = __pyx_v_fit;
  goto __pyx_L0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":194
 *         return theoretical_distribution.scale(experimental_distribution, self.scale_method)
 * 
 *     cpdef IsotopicFitRecord _evaluate_theoretical_distribution(self, list experimental,             # <<<<<<<<<<<<<<
 *                                                                TheoreticalIsotopicPattern theoretical,
 *                                                                FittedPeak peak, int charge):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.DeconvoluterBase._evaluate_theoretical_distribution", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_fit);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_11_evaluate_theoretical_distribution(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_10_evaluate_theoretical_distribution[] = "DeconvoluterBase._evaluate_theoretical_distribution(self, list experimental, TheoreticalIsotopicPattern theoretical, FittedPeak peak, int charge) -> IsotopicFitRecord\nEvaluate a provided theoretical isotopic pattern fit against a\n        set of matched experimental peaks.\n\n        Parameters\n        ----------\n        experimental : list\n            A list of experimental fitted peaks to test\n        theoretical : :class:`~.TheoreticalIsotopicPattern`\n            A list of theoretical isotopic peaks to test\n        peak : :class:`~.FittedPeak`\n            The seed peak for the isotopic pattern fit\n        charge : int\n            The target charge state\n\n        Returns\n        -------\n        :class:`~.IsotopicFitRecord`\n        ";
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_11_evaluate_theoretical_distribution(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_experimental = 0;
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *__pyx_v_theoretical = 0;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_peak = 0;
  int __pyx_v_charge;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_evaluate_theoretical_distribution (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_experimental,&__pyx_n_s_theoretical,&__pyx_n_s_peak,&__pyx_n_s_charge,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_experimental)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_theoretical)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_evaluate_theoretical_distribution", 1, 4, 4, 1); __PYX_ERR(0, 194, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peak)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_evaluate_theoretical_distribution", 1, 4, 4, 2); __PYX_ERR(0, 194, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_charge)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_evaluate_theoretical_distribution", 1, 4, 4, 3); __PYX_ERR(0, 194, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_evaluate_theoretical_distribution") < 0)) __PYX_ERR(0, 194, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_experimental = ((PyObject*)values[0]);
    __pyx_v_theoretical = ((struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *)values[1]);
    __pyx_v_peak = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)values[2]);
    __pyx_v_charge = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_charge == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 196, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_evaluate_theoretical_distribution", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 194, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.DeconvoluterBase._evaluate_theoretical_distribution", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_experimental), (&PyList_Type), 1, "experimental", 1))) __PYX_ERR(0, 194, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_theoretical), __pyx_ptype_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern, 1, "theoretical", 0))) __PYX_ERR(0, 195, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peak), __pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak, 1, "peak", 0))) __PYX_ERR(0, 196, __pyx_L1_error)
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_10_evaluate_theoretical_distribution(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *)__pyx_v_self), __pyx_v_experimental, __pyx_v_theoretical, __pyx_v_peak, __pyx_v_charge);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_10_evaluate_theoretical_distribution(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, PyObject *__pyx_v_experimental, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *__pyx_v_theoretical, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_peak, int __pyx_v_charge) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("_evaluate_theoretical_distribution", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__evaluate_theoretical_distribution(__pyx_v_self, __pyx_v_experimental, __pyx_v_theoretical, __pyx_v_peak, __pyx_v_charge, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 194, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.DeconvoluterBase._evaluate_theoretical_distribution", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/deconvoluter_base.pyx":224
 *         return fit
 * 
 *     cpdef list fit_incremental_truncation(self, IsotopicFitRecord seed_fit, double lower_bound):             # <<<<<<<<<<<<<<
 *         """Fit incrementally truncated versions of the seed fit to check to see if a narrower
 *         theoretical fit matches the data better.
 */

static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_13fit_incremental_truncation(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_fit_incremental_truncation(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_seed_fit, double __pyx_v_lower_bound, int __pyx_skip_dispatch) {
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_fit = 0;
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *__pyx_v_pattern = 0;
  PyObject *__pyx_v_fits = 0;
  PyObject *__pyx_v_patterns = 0;
  PyObject *__pyx_v_eid = 0;
  size_t __pyx_v_i;
  size_t __pyx_v_n;
  Py_ssize_t __pyx_v_k;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  size_t __pyx_t_8;
  size_t __pyx_t_9;
  size_t __pyx_t_10;
  PyObject *__pyx_t_11;
  __Pyx_RefNannySetupContext("fit_incremental_truncation", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_fit_incremental_truncation); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 224, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_13fit_incremental_truncation)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_lower_bound); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 224, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_seed_fit), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 224, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_seed_fit), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 224, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 224, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_seed_fit));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_seed_fit));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_seed_fit));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 224, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 224, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":246
 *             Py_ssize_t k
 * 
 *         patterns = seed_fit.theoretical.incremental_truncation(lower_bound)             # <<<<<<<<<<<<<<
 *         n = PyList_GET_SIZE(patterns)
 *         fits = PyList_New(n)
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *)__pyx_v_seed_fit->theoretical->__pyx_vtab)->incremental_truncation(__pyx_v_seed_fit->theoretical, __pyx_v_lower_bound, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_patterns = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":247
 * 
 *         patterns = seed_fit.theoretical.incremental_truncation(lower_bound)
 *         n = PyList_GET_SIZE(patterns)             # <<<<<<<<<<<<<<
 *         fits = PyList_New(n)
 *         Py_INCREF(seed_fit)
 */
  __pyx_v_n = PyList_GET_SIZE(__pyx_v_patterns);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":248
 *         patterns = seed_fit.theoretical.incremental_truncation(lower_bound)
 *         n = PyList_GET_SIZE(patterns)
 *         fits = PyList_New(n)             # <<<<<<<<<<<<<<
 *         Py_INCREF(seed_fit)
 *         PyList_SET_ITEM(fits, 0, seed_fit)
 */
  __pyx_t_1 = PyList_New(__pyx_v_n); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 248, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_fits = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":249
 *         n = PyList_GET_SIZE(patterns)
 *         fits = PyList_New(n)
 *         Py_INCREF(seed_fit)             # <<<<<<<<<<<<<<
 *         PyList_SET_ITEM(fits, 0, seed_fit)
 *         for i in range(1, n):
 */
  Py_INCREF(((PyObject *)__pyx_v_seed_fit));

  /* "ms_deisotope/_c/deconvoluter_base.pyx":250
 *         fits = PyList_New(n)
 *         Py_INCREF(seed_fit)
 *         PyList_SET_ITEM(fits, 0, seed_fit)             # <<<<<<<<<<<<<<
 *         for i in range(1, n):
 *             pattern = <TheoreticalIsotopicPattern>PyList_GET_ITEM(patterns, i)
 */
  PyList_SET_ITEM(__pyx_v_fits, 0, ((PyObject *)__pyx_v_seed_fit));

  /* "ms_deisotope/_c/deconvoluter_base.pyx":251
 *         Py_INCREF(seed_fit)
 *         PyList_SET_ITEM(fits, 0, seed_fit)
 *         for i in range(1, n):             # <<<<<<<<<<<<<<
 *             pattern = <TheoreticalIsotopicPattern>PyList_GET_ITEM(patterns, i)
 *             k = pattern.get_size()
 */
  __pyx_t_8 = __pyx_v_n;
  __pyx_t_9 = __pyx_t_8;
  for (__pyx_t_10 = 1; __pyx_t_10 < __pyx_t_9; __pyx_t_10+=1) {
    __pyx_v_i = __pyx_t_10;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":252
 *         PyList_SET_ITEM(fits, 0, seed_fit)
 *         for i in range(1, n):
 *             pattern = <TheoreticalIsotopicPattern>PyList_GET_ITEM(patterns, i)             # <<<<<<<<<<<<<<
 *             k = pattern.get_size()
 *             eid = PyList_GetSlice(seed_fit.experimental, 0, k)
 */
    __pyx_t_11 = PyList_GET_ITEM(__pyx_v_patterns, __pyx_v_i);
    __pyx_t_1 = ((PyObject *)__pyx_t_11);
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_pattern, ((struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":253
 *         for i in range(1, n):
 *             pattern = <TheoreticalIsotopicPattern>PyList_GET_ITEM(patterns, i)
 *             k = pattern.get_size()             # <<<<<<<<<<<<<<
 *             eid = PyList_GetSlice(seed_fit.experimental, 0, k)
 *             fit = self._evaluate_theoretical_distribution(eid, pattern, seed_fit.seed_peak, seed_fit.charge)
 */
    __pyx_v_k = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *)__pyx_v_pattern->__pyx_vtab)->get_size(__pyx_v_pattern);

    /* "ms_deisotope/_c/deconvoluter_base.pyx":254
 *             pattern = <TheoreticalIsotopicPattern>PyList_GET_ITEM(patterns, i)
 *             k = pattern.get_size()
 *             eid = PyList_GetSlice(seed_fit.experimental, 0, k)             # <<<<<<<<<<<<<<
 *             fit = self._evaluate_theoretical_distribution(eid, pattern, seed_fit.seed_peak, seed_fit.charge)
 *             Py_INCREF(fit)
 */
    __pyx_t_1 = __pyx_v_seed_fit->experimental;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_2 = PyList_GetSlice(__pyx_t_1, 0, __pyx_v_k); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 254, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF_SET(__pyx_v_eid, ((PyObject*)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":255
 *             k = pattern.get_size()
 *             eid = PyList_GetSlice(seed_fit.experimental, 0, k)
 *             fit = self._evaluate_theoretical_distribution(eid, pattern, seed_fit.seed_peak, seed_fit.charge)             # <<<<<<<<<<<<<<
 *             Py_INCREF(fit)
 *             PyList_SET_ITEM(fits, i, fit)
 */
    __pyx_t_2 = ((PyObject *)__pyx_v_seed_fit->seed_peak);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *)__pyx_v_self->__pyx_vtab)->_evaluate_theoretical_distribution(__pyx_v_self, __pyx_v_eid, __pyx_v_pattern, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_2), __pyx_v_seed_fit->charge, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 255, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF_SET(__pyx_v_fit, ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":256
 *             eid = PyList_GetSlice(seed_fit.experimental, 0, k)
 *             fit = self._evaluate_theoretical_distribution(eid, pattern, seed_fit.seed_peak, seed_fit.charge)
 *             Py_INCREF(fit)             # <<<<<<<<<<<<<<
 *             PyList_SET_ITEM(fits, i, fit)
 *         return fits
 */
    Py_INCREF(((PyObject *)__pyx_v_fit));

    /* "ms_deisotope/_c/deconvoluter_base.pyx":257
 *             fit = self._evaluate_theoretical_distribution(eid, pattern, seed_fit.seed_peak, seed_fit.charge)
 *             Py_INCREF(fit)
 *             PyList_SET_ITEM(fits, i, fit)             # <<<<<<<<<<<<<<
 *         return fits
 * 
 */
    PyList_SET_ITEM(__pyx_v_fits, __pyx_v_i, ((PyObject *)__pyx_v_fit));
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":258
 *             Py_INCREF(fit)
 *             PyList_SET_ITEM(fits, i, fit)
 *         return fits             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_fits);
  __pyx_r = __pyx_v_fits;
  goto __pyx_L0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":224
 *         return fit
 * 
 *     cpdef list fit_incremental_truncation(self, IsotopicFitRecord seed_fit, double lower_bound):             # <<<<<<<<<<<<<<
 *         """Fit incrementally truncated versions of the seed fit to check to see if a narrower
 *         theoretical fit matches the data better.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.DeconvoluterBase.fit_incremental_truncation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_fit);
  __Pyx_XDECREF((PyObject *)__pyx_v_pattern);
  __Pyx_XDECREF(__pyx_v_fits);
  __Pyx_XDECREF(__pyx_v_patterns);
  __Pyx_XDECREF(__pyx_v_eid);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_13fit_incremental_truncation(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_12fit_incremental_truncation[] = "DeconvoluterBase.fit_incremental_truncation(self, IsotopicFitRecord seed_fit, double lower_bound) -> list\nFit incrementally truncated versions of the seed fit to check to see if a narrower\n        theoretical fit matches the data better.\n\n        Parameters\n        ----------\n        seed_fit : :class:`~.IsotopicFitRecord`\n            The original fit to explore truncations of\n        lower_bound : float\n            The percentage of total signal remaining to stop truncating at.\n\n        Returns\n        -------\n        :class:`list` of :class:`~.IsotopicFitRecord`\n        ";
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_13fit_incremental_truncation(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_seed_fit = 0;
  double __pyx_v_lower_bound;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("fit_incremental_truncation (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_seed_fit,&__pyx_n_s_lower_bound,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_seed_fit)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lower_bound)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("fit_incremental_truncation", 1, 2, 2, 1); __PYX_ERR(0, 224, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "fit_incremental_truncation") < 0)) __PYX_ERR(0, 224, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_seed_fit = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)values[0]);
    __pyx_v_lower_bound = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_lower_bound == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 224, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("fit_incremental_truncation", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 224, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.DeconvoluterBase.fit_incremental_truncation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_seed_fit), __pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitRecord, 1, "seed_fit", 0))) __PYX_ERR(0, 224, __pyx_L1_error)
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_12fit_incremental_truncation(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *)__pyx_v_self), __pyx_v_seed_fit, __pyx_v_lower_bound);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_12fit_incremental_truncation(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_seed_fit, double __pyx_v_lower_bound) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("fit_incremental_truncation", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_fit_incremental_truncation(__pyx_v_self, __pyx_v_seed_fit, __pyx_v_lower_bound, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.DeconvoluterBase.fit_incremental_truncation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/deconvoluter_base.pyx":261
 * 
 * 
 *     cpdef subtraction(self, TheoreticalIsotopicPattern isotopic_cluster, double error_tolerance=2e-5):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i
 */

static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_15subtraction(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_subtraction(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *__pyx_v_isotopic_cluster, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_subtraction *__pyx_optional_args) {
  double __pyx_v_error_tolerance = ((double)2e-5);
  size_t __pyx_v_i;
  double __pyx_v_existing;
  struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_v_peak = 0;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_match = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  size_t __pyx_t_8;
  size_t __pyx_t_9;
  size_t __pyx_t_10;
  struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet__has_peak __pyx_t_11;
  int __pyx_t_12;
  int __pyx_t_13;
  double __pyx_t_14;
  __Pyx_RefNannySetupContext("subtraction", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_error_tolerance = __pyx_optional_args->error_tolerance;
    }
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_subtraction); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 261, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_15subtraction)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_error_tolerance); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 261, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_isotopic_cluster), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 261, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_isotopic_cluster), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 261, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 261, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_isotopic_cluster));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_isotopic_cluster));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_isotopic_cluster));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 261, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":267
 *             TheoreticalPeak peak
 *             FittedPeak match
 *         for i in range(isotopic_cluster.get_size()):             # <<<<<<<<<<<<<<
 *             peak = isotopic_cluster.get(i)
 *             match = self.peaklist._has_peak(peak.mz, error_tolerance)
 */
  __pyx_t_8 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *)__pyx_v_isotopic_cluster->__pyx_vtab)->get_size(__pyx_v_isotopic_cluster);
  __pyx_t_9 = __pyx_t_8;
  for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_9; __pyx_t_10+=1) {
    __pyx_v_i = __pyx_t_10;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":268
 *             FittedPeak match
 *         for i in range(isotopic_cluster.get_size()):
 *             peak = isotopic_cluster.get(i)             # <<<<<<<<<<<<<<
 *             match = self.peaklist._has_peak(peak.mz, error_tolerance)
 *             if match is not None:
 */
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *)__pyx_v_isotopic_cluster->__pyx_vtab)->get(__pyx_v_isotopic_cluster, __pyx_v_i)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 268, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_peak, ((struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":269
 *         for i in range(isotopic_cluster.get_size()):
 *             peak = isotopic_cluster.get(i)
 *             match = self.peaklist._has_peak(peak.mz, error_tolerance)             # <<<<<<<<<<<<<<
 *             if match is not None:
 *                 existing = match.intensity
 */
    __pyx_t_11.__pyx_n = 1;
    __pyx_t_11.tolerance = __pyx_v_error_tolerance;
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_self->peaklist->__pyx_vtab)->_has_peak(__pyx_v_self->peaklist, __pyx_v_peak->mz, &__pyx_t_11)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 269, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_match, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":270
 *             peak = isotopic_cluster.get(i)
 *             match = self.peaklist._has_peak(peak.mz, error_tolerance)
 *             if match is not None:             # <<<<<<<<<<<<<<
 *                 existing = match.intensity
 *                 match.intensity -= peak.intensity
 */
    __pyx_t_12 = (((PyObject *)__pyx_v_match) != Py_None);
    __pyx_t_13 = (__pyx_t_12 != 0);
    if (__pyx_t_13) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":271
 *             match = self.peaklist._has_peak(peak.mz, error_tolerance)
 *             if match is not None:
 *                 existing = match.intensity             # <<<<<<<<<<<<<<
 *                 match.intensity -= peak.intensity
 *                 if (match.intensity < 0) or (peak.intensity > (existing * 0.7)):
 */
      __pyx_t_14 = __pyx_v_match->__pyx_base.intensity;
      __pyx_v_existing = __pyx_t_14;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":272
 *             if match is not None:
 *                 existing = match.intensity
 *                 match.intensity -= peak.intensity             # <<<<<<<<<<<<<<
 *                 if (match.intensity < 0) or (peak.intensity > (existing * 0.7)):
 *                     match.intensity = 1.
 */
      __pyx_v_match->__pyx_base.intensity = (__pyx_v_match->__pyx_base.intensity - __pyx_v_peak->intensity);

      /* "ms_deisotope/_c/deconvoluter_base.pyx":273
 *                 existing = match.intensity
 *                 match.intensity -= peak.intensity
 *                 if (match.intensity < 0) or (peak.intensity > (existing * 0.7)):             # <<<<<<<<<<<<<<
 *                     match.intensity = 1.
 * 
 */
      __pyx_t_12 = ((__pyx_v_match->__pyx_base.intensity < 0.0) != 0);
      if (!__pyx_t_12) {
      } else {
        __pyx_t_13 = __pyx_t_12;
        goto __pyx_L7_bool_binop_done;
      }
      __pyx_t_12 = ((__pyx_v_peak->intensity > (__pyx_v_existing * 0.7)) != 0);
      __pyx_t_13 = __pyx_t_12;
      __pyx_L7_bool_binop_done:;
      if (__pyx_t_13) {

        /* "ms_deisotope/_c/deconvoluter_base.pyx":274
 *                 match.intensity -= peak.intensity
 *                 if (match.intensity < 0) or (peak.intensity > (existing * 0.7)):
 *                     match.intensity = 1.             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint _check_fit(self, IsotopicFitRecord fit):
 */
        __pyx_v_match->__pyx_base.intensity = 1.;

        /* "ms_deisotope/_c/deconvoluter_base.pyx":273
 *                 existing = match.intensity
 *                 match.intensity -= peak.intensity
 *                 if (match.intensity < 0) or (peak.intensity > (existing * 0.7)):             # <<<<<<<<<<<<<<
 *                     match.intensity = 1.
 * 
 */
      }

      /* "ms_deisotope/_c/deconvoluter_base.pyx":270
 *             peak = isotopic_cluster.get(i)
 *             match = self.peaklist._has_peak(peak.mz, error_tolerance)
 *             if match is not None:             # <<<<<<<<<<<<<<
 *                 existing = match.intensity
 *                 match.intensity -= peak.intensity
 */
    }
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":261
 * 
 * 
 *     cpdef subtraction(self, TheoreticalIsotopicPattern isotopic_cluster, double error_tolerance=2e-5):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.DeconvoluterBase.subtraction", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);
  __Pyx_XDECREF((PyObject *)__pyx_v_match);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_15subtraction(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_14subtraction[] = "DeconvoluterBase.subtraction(self, TheoreticalIsotopicPattern isotopic_cluster, double error_tolerance=2e-5)";
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_15subtraction(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *__pyx_v_isotopic_cluster = 0;
  double __pyx_v_error_tolerance;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("subtraction (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_isotopic_cluster,&__pyx_n_s_error_tolerance,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_isotopic_cluster)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_error_tolerance);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "subtraction") < 0)) __PYX_ERR(0, 261, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_isotopic_cluster = ((struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *)values[0]);
    if (values[1]) {
      __pyx_v_error_tolerance = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_error_tolerance == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 261, __pyx_L3_error)
    } else {
      __pyx_v_error_tolerance = ((double)2e-5);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("subtraction", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 261, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.DeconvoluterBase.subtraction", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_isotopic_cluster), __pyx_ptype_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern, 1, "isotopic_cluster", 0))) __PYX_ERR(0, 261, __pyx_L1_error)
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_14subtraction(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *)__pyx_v_self), __pyx_v_isotopic_cluster, __pyx_v_error_tolerance);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_14subtraction(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *__pyx_v_isotopic_cluster, double __pyx_v_error_tolerance) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_subtraction __pyx_t_2;
  __Pyx_RefNannySetupContext("subtraction", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.error_tolerance = __pyx_v_error_tolerance;
  __pyx_t_1 = __pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase->subtraction(__pyx_v_self, __pyx_v_isotopic_cluster, 1, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 261, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.DeconvoluterBase.subtraction", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/deconvoluter_base.pyx":276
 *                     match.intensity = 1.
 * 
 *     cpdef bint _check_fit(self, IsotopicFitRecord fit):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t n
 */

static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_17_check_fit(PyObject *__pyx_v_self, PyObject *__pyx_v_fit); /*proto*/
static int __pyx_f_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__check_fit(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_fit, int __pyx_skip_dispatch) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("_check_fit", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_check_fit); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 276, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_17_check_fit)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, ((PyObject *)__pyx_v_fit)) : __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_fit));
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 276, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 276, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_5;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":279
 *         cdef:
 *             size_t n
 *         if not has_multiple_real_peaks(fit.experimental) and fit.charge > 1:             # <<<<<<<<<<<<<<
 *             return False
 *         if self.scorer.reject(fit):
 */
  __pyx_t_1 = __pyx_v_fit->experimental;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_6 = ((!(__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_has_multiple_real_peaks(((PyObject*)__pyx_t_1)) != 0)) != 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_6) {
  } else {
    __pyx_t_5 = __pyx_t_6;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_6 = ((__pyx_v_fit->charge > 1) != 0);
  __pyx_t_5 = __pyx_t_6;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_5) {

    /* "ms_deisotope/_c/deconvoluter_base.pyx":280
 *             size_t n
 *         if not has_multiple_real_peaks(fit.experimental) and fit.charge > 1:
 *             return False             # <<<<<<<<<<<<<<
 *         if self.scorer.reject(fit):
 *             return False
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":279
 *         cdef:
 *             size_t n
 *         if not has_multiple_real_peaks(fit.experimental) and fit.charge > 1:             # <<<<<<<<<<<<<<
 *             return False
 *         if self.scorer.reject(fit):
 */
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":281
 *         if not has_multiple_real_peaks(fit.experimental) and fit.charge > 1:
 *             return False
 *         if self.scorer.reject(fit):             # <<<<<<<<<<<<<<
 *             return False
 *         return True
 */
  __pyx_t_5 = (((struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *)__pyx_v_self->scorer->__pyx_vtab)->reject(__pyx_v_self->scorer, __pyx_v_fit, 0) != 0);
  if (__pyx_t_5) {

    /* "ms_deisotope/_c/deconvoluter_base.pyx":282
 *             return False
 *         if self.scorer.reject(fit):
 *             return False             # <<<<<<<<<<<<<<
 *         return True
 * 
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":281
 *         if not has_multiple_real_peaks(fit.experimental) and fit.charge > 1:
 *             return False
 *         if self.scorer.reject(fit):             # <<<<<<<<<<<<<<
 *             return False
 *         return True
 */
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":283
 *         if self.scorer.reject(fit):
 *             return False
 *         return True             # <<<<<<<<<<<<<<
 * 
 *     def _merge_peaks(self, peak_list):
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":276
 *                     match.intensity = 1.
 * 
 *     cpdef bint _check_fit(self, IsotopicFitRecord fit):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t n
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("ms_deisotope._c.deconvoluter_base.DeconvoluterBase._check_fit", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_17_check_fit(PyObject *__pyx_v_self, PyObject *__pyx_v_fit); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_16_check_fit[] = "DeconvoluterBase._check_fit(self, IsotopicFitRecord fit) -> bool";
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_17_check_fit(PyObject *__pyx_v_self, PyObject *__pyx_v_fit) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_check_fit (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_fit), __pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitRecord, 1, "fit", 0))) __PYX_ERR(0, 276, __pyx_L1_error)
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_16_check_fit(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *)__pyx_v_self), ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_v_fit));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_16_check_fit(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_fit) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("_check_fit", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__check_fit(__pyx_v_self, __pyx_v_fit, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.DeconvoluterBase._check_fit", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/deconvoluter_base.pyx":285
 *         return True
 * 
 *     def _merge_peaks(self, peak_list):             # <<<<<<<<<<<<<<
 *         peak_list = sorted(peak_list, key=operator.attrgetter("neutral_mass"))
 *         if not peak_list:
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_19_merge_peaks(PyObject *__pyx_v_self, PyObject *__pyx_v_peak_list); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_18_merge_peaks[] = "DeconvoluterBase._merge_peaks(self, peak_list)";
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_19_merge_peaks(PyObject *__pyx_v_self, PyObject *__pyx_v_peak_list) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_merge_peaks (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_18_merge_peaks(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *)__pyx_v_self), ((PyObject *)__pyx_v_peak_list));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_18_merge_peaks(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, PyObject *__pyx_v_peak_list) {
  PyObject *__pyx_v_current_peak = NULL;
  PyObject *__pyx_v_merged_peaks = NULL;
  PyObject *__pyx_v_peak = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  PyObject *(*__pyx_t_9)(PyObject *);
  int __pyx_t_10;
  __Pyx_RefNannySetupContext("_merge_peaks", 0);
  __Pyx_INCREF(__pyx_v_peak_list);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":286
 * 
 *     def _merge_peaks(self, peak_list):
 *         peak_list = sorted(peak_list, key=operator.attrgetter("neutral_mass"))             # <<<<<<<<<<<<<<
 *         if not peak_list:
 *             return []
 */
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_peak_list);
  __Pyx_GIVEREF(__pyx_v_peak_list);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_peak_list);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_operator); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_attrgetter); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_4, __pyx_n_s_neutral_mass) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_n_s_neutral_mass);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_key, __pyx_t_3) < 0) __PYX_ERR(0, 286, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_sorted, __pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF_SET(__pyx_v_peak_list, __pyx_t_3);
  __pyx_t_3 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":287
 *     def _merge_peaks(self, peak_list):
 *         peak_list = sorted(peak_list, key=operator.attrgetter("neutral_mass"))
 *         if not peak_list:             # <<<<<<<<<<<<<<
 *             return []
 *         current_peak = peak_list[0]
 */
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_peak_list); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 287, __pyx_L1_error)
  __pyx_t_7 = ((!__pyx_t_6) != 0);
  if (__pyx_t_7) {

    /* "ms_deisotope/_c/deconvoluter_base.pyx":288
 *         peak_list = sorted(peak_list, key=operator.attrgetter("neutral_mass"))
 *         if not peak_list:
 *             return []             # <<<<<<<<<<<<<<
 *         current_peak = peak_list[0]
 *         merged_peaks = []
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 288, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":287
 *     def _merge_peaks(self, peak_list):
 *         peak_list = sorted(peak_list, key=operator.attrgetter("neutral_mass"))
 *         if not peak_list:             # <<<<<<<<<<<<<<
 *             return []
 *         current_peak = peak_list[0]
 */
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":289
 *         if not peak_list:
 *             return []
 *         current_peak = peak_list[0]             # <<<<<<<<<<<<<<
 *         merged_peaks = []
 *         for peak in peak_list[1:]:
 */
  __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_peak_list, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_current_peak = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":290
 *             return []
 *         current_peak = peak_list[0]
 *         merged_peaks = []             # <<<<<<<<<<<<<<
 *         for peak in peak_list[1:]:
 *             if abs(current_peak.neutral_mass - peak.neutral_mass) < 1e-3 and current_peak.charge == peak.charge:
 */
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 290, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_merged_peaks = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":291
 *         current_peak = peak_list[0]
 *         merged_peaks = []
 *         for peak in peak_list[1:]:             # <<<<<<<<<<<<<<
 *             if abs(current_peak.neutral_mass - peak.neutral_mass) < 1e-3 and current_peak.charge == peak.charge:
 *                 current_peak.intensity += peak.intensity
 */
  __pyx_t_3 = __Pyx_PyObject_GetSlice(__pyx_v_peak_list, 1, 0, NULL, NULL, &__pyx_slice_, 1, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
    __pyx_t_2 = __pyx_t_3; __Pyx_INCREF(__pyx_t_2); __pyx_t_8 = 0;
    __pyx_t_9 = NULL;
  } else {
    __pyx_t_8 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 291, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_9 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 291, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  for (;;) {
    if (likely(!__pyx_t_9)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_8); __Pyx_INCREF(__pyx_t_3); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 291, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 291, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      } else {
        if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_8); __Pyx_INCREF(__pyx_t_3); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 291, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 291, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      }
    } else {
      __pyx_t_3 = __pyx_t_9(__pyx_t_2);
      if (unlikely(!__pyx_t_3)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 291, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_XDECREF_SET(__pyx_v_peak, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":292
 *         merged_peaks = []
 *         for peak in peak_list[1:]:
 *             if abs(current_peak.neutral_mass - peak.neutral_mass) < 1e-3 and current_peak.charge == peak.charge:             # <<<<<<<<<<<<<<
 *                 current_peak.intensity += peak.intensity
 *             else:
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_current_peak, __pyx_n_s_neutral_mass); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 292, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_peak, __pyx_n_s_neutral_mass); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 292, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = PyNumber_Subtract(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 292, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyNumber_Absolute(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 292, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyObject_RichCompare(__pyx_t_1, __pyx_float_1eneg_3, Py_LT); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 292, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 292, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (__pyx_t_6) {
    } else {
      __pyx_t_7 = __pyx_t_6;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_current_peak, __pyx_n_s_charge); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 292, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_peak, __pyx_n_s_charge); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 292, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyObject_RichCompare(__pyx_t_5, __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 292, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 292, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_7 = __pyx_t_6;
    __pyx_L7_bool_binop_done:;
    if (__pyx_t_7) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":293
 *         for peak in peak_list[1:]:
 *             if abs(current_peak.neutral_mass - peak.neutral_mass) < 1e-3 and current_peak.charge == peak.charge:
 *                 current_peak.intensity += peak.intensity             # <<<<<<<<<<<<<<
 *             else:
 *                 merged_peaks.append(current_peak)
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_current_peak, __pyx_n_s_intensity); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 293, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_peak, __pyx_n_s_intensity); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 293, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_5 = PyNumber_InPlaceAdd(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 293, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__Pyx_PyObject_SetAttrStr(__pyx_v_current_peak, __pyx_n_s_intensity, __pyx_t_5) < 0) __PYX_ERR(0, 293, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":292
 *         merged_peaks = []
 *         for peak in peak_list[1:]:
 *             if abs(current_peak.neutral_mass - peak.neutral_mass) < 1e-3 and current_peak.charge == peak.charge:             # <<<<<<<<<<<<<<
 *                 current_peak.intensity += peak.intensity
 *             else:
 */
      goto __pyx_L6;
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":295
 *                 current_peak.intensity += peak.intensity
 *             else:
 *                 merged_peaks.append(current_peak)             # <<<<<<<<<<<<<<
 *                 current_peak = peak
 *         merged_peaks.append(current_peak)
 */
    /*else*/ {
      __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_merged_peaks, __pyx_v_current_peak); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 295, __pyx_L1_error)

      /* "ms_deisotope/_c/deconvoluter_base.pyx":296
 *             else:
 *                 merged_peaks.append(current_peak)
 *                 current_peak = peak             # <<<<<<<<<<<<<<
 *         merged_peaks.append(current_peak)
 *         return merged_peaks
 */
      __Pyx_INCREF(__pyx_v_peak);
      __Pyx_DECREF_SET(__pyx_v_current_peak, __pyx_v_peak);
    }
    __pyx_L6:;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":291
 *         current_peak = peak_list[0]
 *         merged_peaks = []
 *         for peak in peak_list[1:]:             # <<<<<<<<<<<<<<
 *             if abs(current_peak.neutral_mass - peak.neutral_mass) < 1e-3 and current_peak.charge == peak.charge:
 *                 current_peak.intensity += peak.intensity
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":297
 *                 merged_peaks.append(current_peak)
 *                 current_peak = peak
 *         merged_peaks.append(current_peak)             # <<<<<<<<<<<<<<
 *         return merged_peaks
 * 
 */
  __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_merged_peaks, __pyx_v_current_peak); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 297, __pyx_L1_error)

  /* "ms_deisotope/_c/deconvoluter_base.pyx":298
 *                 current_peak = peak
 *         merged_peaks.append(current_peak)
 *         return merged_peaks             # <<<<<<<<<<<<<<
 * 
 *     cdef int _find_next_putative_peak_inplace(self, double mz, int charge, set result, int step=1, double tolerance=2e-5):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_merged_peaks);
  __pyx_r = __pyx_v_merged_peaks;
  goto __pyx_L0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":285
 *         return True
 * 
 *     def _merge_peaks(self, peak_list):             # <<<<<<<<<<<<<<
 *         peak_list = sorted(peak_list, key=operator.attrgetter("neutral_mass"))
 *         if not peak_list:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.DeconvoluterBase._merge_peaks", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_current_peak);
  __Pyx_XDECREF(__pyx_v_merged_peaks);
  __Pyx_XDECREF(__pyx_v_peak);
  __Pyx_XDECREF(__pyx_v_peak_list);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/deconvoluter_base.pyx":300
 *         return merged_peaks
 * 
 *     cdef int _find_next_putative_peak_inplace(self, double mz, int charge, set result, int step=1, double tolerance=2e-5):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double shift, next_peak, prev_peak_mz
 */

static int __pyx_f_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak_inplace(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, double __pyx_v_mz, int __pyx_v_charge, PyObject *__pyx_v_result, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak_inplace *__pyx_optional_args) {
  int __pyx_v_step = ((int)1);
  double __pyx_v_tolerance = ((double)2e-5);
  double __pyx_v_shift;
  double __pyx_v_next_peak;
  double __pyx_v_prev_peak_mz;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_dummy_peak = 0;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_forward = 0;
  size_t __pyx_v_i;
  size_t __pyx_v_start;
  size_t __pyx_v_stop;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_isotopic_shift __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  size_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("_find_next_putative_peak_inplace", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_step = __pyx_optional_args->step;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_tolerance = __pyx_optional_args->tolerance;
      }
    }
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":307
 *             size_t start, stop
 * 
 *         shift = isotopic_shift(charge)             # <<<<<<<<<<<<<<
 *         next_peak = mz + (shift * step)
 *         start = 0
 */
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.charge = __pyx_v_charge;
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_9averagine_isotopic_shift(0, &__pyx_t_2); 
  __pyx_v_shift = __pyx_t_1;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":308
 * 
 *         shift = isotopic_shift(charge)
 *         next_peak = mz + (shift * step)             # <<<<<<<<<<<<<<
 *         start = 0
 *         stop = 0
 */
  __pyx_v_next_peak = (__pyx_v_mz + (__pyx_v_shift * __pyx_v_step));

  /* "ms_deisotope/_c/deconvoluter_base.pyx":309
 *         shift = isotopic_shift(charge)
 *         next_peak = mz + (shift * step)
 *         start = 0             # <<<<<<<<<<<<<<
 *         stop = 0
 *         self.peaklist._between_bounds(
 */
  __pyx_v_start = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":310
 *         next_peak = mz + (shift * step)
 *         start = 0
 *         stop = 0             # <<<<<<<<<<<<<<
 *         self.peaklist._between_bounds(
 *             next_peak - (next_peak * tolerance),
 */
  __pyx_v_stop = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":311
 *         start = 0
 *         stop = 0
 *         self.peaklist._between_bounds(             # <<<<<<<<<<<<<<
 *             next_peak - (next_peak * tolerance),
 *             next_peak + (next_peak * tolerance),
 */
  (void)(((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_self->peaklist->__pyx_vtab)->_between_bounds(__pyx_v_self->peaklist, (__pyx_v_next_peak - (__pyx_v_next_peak * __pyx_v_tolerance)), (__pyx_v_next_peak + (__pyx_v_next_peak * __pyx_v_tolerance)), (&__pyx_v_start), (&__pyx_v_stop)));

  /* "ms_deisotope/_c/deconvoluter_base.pyx":315
 *             next_peak + (next_peak * tolerance),
 *             &start, &stop)
 *         for i in range(start, stop):             # <<<<<<<<<<<<<<
 *             forward = self.peaklist.getitem(i)
 *             prev_peak_mz = forward.mz - (shift * step)
 */
  __pyx_t_3 = __pyx_v_stop;
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_5 = __pyx_v_start; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
    __pyx_v_i = __pyx_t_5;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":316
 *             &start, &stop)
 *         for i in range(start, stop):
 *             forward = self.peaklist.getitem(i)             # <<<<<<<<<<<<<<
 *             prev_peak_mz = forward.mz - (shift * step)
 *             dummy_peak = make_placeholder_peak(prev_peak_mz)
 */
    __pyx_t_6 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_self->peaklist->__pyx_vtab)->getitem(__pyx_v_self->peaklist, __pyx_v_i)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 316, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_XDECREF_SET(__pyx_v_forward, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_6));
    __pyx_t_6 = 0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":317
 *         for i in range(start, stop):
 *             forward = self.peaklist.getitem(i)
 *             prev_peak_mz = forward.mz - (shift * step)             # <<<<<<<<<<<<<<
 *             dummy_peak = make_placeholder_peak(prev_peak_mz)
 *             PySet_Add(result, (dummy_peak, charge))
 */
    __pyx_v_prev_peak_mz = (__pyx_v_forward->__pyx_base.mz - (__pyx_v_shift * __pyx_v_step));

    /* "ms_deisotope/_c/deconvoluter_base.pyx":318
 *             forward = self.peaklist.getitem(i)
 *             prev_peak_mz = forward.mz - (shift * step)
 *             dummy_peak = make_placeholder_peak(prev_peak_mz)             # <<<<<<<<<<<<<<
 *             PySet_Add(result, (dummy_peak, charge))
 *         return stop - start
 */
    __pyx_t_6 = ((PyObject *)__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_make_placeholder_peak(__pyx_v_prev_peak_mz)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 318, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_XDECREF_SET(__pyx_v_dummy_peak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_6));
    __pyx_t_6 = 0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":319
 *             prev_peak_mz = forward.mz - (shift * step)
 *             dummy_peak = make_placeholder_peak(prev_peak_mz)
 *             PySet_Add(result, (dummy_peak, charge))             # <<<<<<<<<<<<<<
 *         return stop - start
 * 
 */
    __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_charge); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 319, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 319, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_INCREF(((PyObject *)__pyx_v_dummy_peak));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_dummy_peak));
    PyTuple_SET_ITEM(__pyx_t_7, 0, ((PyObject *)__pyx_v_dummy_peak));
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_6);
    __pyx_t_6 = 0;
    __pyx_t_8 = PySet_Add(__pyx_v_result, __pyx_t_7); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 319, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":320
 *             dummy_peak = make_placeholder_peak(prev_peak_mz)
 *             PySet_Add(result, (dummy_peak, charge))
 *         return stop - start             # <<<<<<<<<<<<<<
 * 
 *     cdef int _find_previous_putative_peak_inplace(self, double mz, int charge, set result, int step=1, double tolerance=2e-5):
 */
  __pyx_r = (__pyx_v_stop - __pyx_v_start);
  goto __pyx_L0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":300
 *         return merged_peaks
 * 
 *     cdef int _find_next_putative_peak_inplace(self, double mz, int charge, set result, int step=1, double tolerance=2e-5):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double shift, next_peak, prev_peak_mz
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_WriteUnraisable("ms_deisotope._c.deconvoluter_base.DeconvoluterBase._find_next_putative_peak_inplace", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_dummy_peak);
  __Pyx_XDECREF((PyObject *)__pyx_v_forward);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/deconvoluter_base.pyx":322
 *         return stop - start
 * 
 *     cdef int _find_previous_putative_peak_inplace(self, double mz, int charge, set result, int step=1, double tolerance=2e-5):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double shift, prev_peak, prev_peak_mz
 */

static int __pyx_f_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak_inplace(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, double __pyx_v_mz, int __pyx_v_charge, PyObject *__pyx_v_result, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak_inplace *__pyx_optional_args) {
  int __pyx_v_step = ((int)1);
  double __pyx_v_tolerance = ((double)2e-5);
  double __pyx_v_shift;
  double __pyx_v_prev_peak;
  double __pyx_v_prev_peak_mz;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_backward = 0;
  size_t __pyx_v_i;
  size_t __pyx_v_start;
  size_t __pyx_v_stop;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_isotopic_shift __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  size_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak_inplace __pyx_t_8;
  struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak_inplace __pyx_t_9;
  __Pyx_RefNannySetupContext("_find_previous_putative_peak_inplace", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_step = __pyx_optional_args->step;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_tolerance = __pyx_optional_args->tolerance;
      }
    }
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":329
 *             size_t start, stop
 * 
 *         shift = isotopic_shift(charge)             # <<<<<<<<<<<<<<
 *         prev_peak = mz - (shift)
 *         self.peaklist._between_bounds(
 */
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.charge = __pyx_v_charge;
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_9averagine_isotopic_shift(0, &__pyx_t_2); 
  __pyx_v_shift = __pyx_t_1;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":330
 * 
 *         shift = isotopic_shift(charge)
 *         prev_peak = mz - (shift)             # <<<<<<<<<<<<<<
 *         self.peaklist._between_bounds(
 *             prev_peak - (prev_peak * tolerance),
 */
  __pyx_v_prev_peak = (__pyx_v_mz - __pyx_v_shift);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":331
 *         shift = isotopic_shift(charge)
 *         prev_peak = mz - (shift)
 *         self.peaklist._between_bounds(             # <<<<<<<<<<<<<<
 *             prev_peak - (prev_peak * tolerance),
 *             prev_peak + (prev_peak * tolerance),
 */
  (void)(((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_self->peaklist->__pyx_vtab)->_between_bounds(__pyx_v_self->peaklist, (__pyx_v_prev_peak - (__pyx_v_prev_peak * __pyx_v_tolerance)), (__pyx_v_prev_peak + (__pyx_v_prev_peak * __pyx_v_tolerance)), (&__pyx_v_start), (&__pyx_v_stop)));

  /* "ms_deisotope/_c/deconvoluter_base.pyx":336
 *             &start, &stop)
 * 
 *         for i in range(start, stop):             # <<<<<<<<<<<<<<
 *             backward = self.peaklist.getitem(i)
 *             prev_peak_mz = backward.mz
 */
  __pyx_t_3 = __pyx_v_stop;
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_5 = __pyx_v_start; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
    __pyx_v_i = __pyx_t_5;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":337
 * 
 *         for i in range(start, stop):
 *             backward = self.peaklist.getitem(i)             # <<<<<<<<<<<<<<
 *             prev_peak_mz = backward.mz
 *             if step == 1:
 */
    __pyx_t_6 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_self->peaklist->__pyx_vtab)->getitem(__pyx_v_self->peaklist, __pyx_v_i)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 337, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_XDECREF_SET(__pyx_v_backward, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_6));
    __pyx_t_6 = 0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":338
 *         for i in range(start, stop):
 *             backward = self.peaklist.getitem(i)
 *             prev_peak_mz = backward.mz             # <<<<<<<<<<<<<<
 *             if step == 1:
 *                 self._find_next_putative_peak_inplace(prev_peak_mz, charge, result, 1, tolerance)
 */
    __pyx_t_1 = __pyx_v_backward->__pyx_base.mz;
    __pyx_v_prev_peak_mz = __pyx_t_1;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":339
 *             backward = self.peaklist.getitem(i)
 *             prev_peak_mz = backward.mz
 *             if step == 1:             # <<<<<<<<<<<<<<
 *                 self._find_next_putative_peak_inplace(prev_peak_mz, charge, result, 1, tolerance)
 *             else:
 */
    __pyx_t_7 = ((__pyx_v_step == 1) != 0);
    if (__pyx_t_7) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":340
 *             prev_peak_mz = backward.mz
 *             if step == 1:
 *                 self._find_next_putative_peak_inplace(prev_peak_mz, charge, result, 1, tolerance)             # <<<<<<<<<<<<<<
 *             else:
 *                 self._find_previous_putative_peak_inplace(prev_peak_mz, charge, result, step - 1, tolerance)
 */
      __pyx_t_8.__pyx_n = 2;
      __pyx_t_8.step = 1;
      __pyx_t_8.tolerance = __pyx_v_tolerance;
      ((struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *)__pyx_v_self->__pyx_vtab)->_find_next_putative_peak_inplace(__pyx_v_self, __pyx_v_prev_peak_mz, __pyx_v_charge, __pyx_v_result, &__pyx_t_8); 

      /* "ms_deisotope/_c/deconvoluter_base.pyx":339
 *             backward = self.peaklist.getitem(i)
 *             prev_peak_mz = backward.mz
 *             if step == 1:             # <<<<<<<<<<<<<<
 *                 self._find_next_putative_peak_inplace(prev_peak_mz, charge, result, 1, tolerance)
 *             else:
 */
      goto __pyx_L5;
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":342
 *                 self._find_next_putative_peak_inplace(prev_peak_mz, charge, result, 1, tolerance)
 *             else:
 *                 self._find_previous_putative_peak_inplace(prev_peak_mz, charge, result, step - 1, tolerance)             # <<<<<<<<<<<<<<
 *         return stop - start
 * 
 */
    /*else*/ {
      __pyx_t_9.__pyx_n = 2;
      __pyx_t_9.step = (__pyx_v_step - 1);
      __pyx_t_9.tolerance = __pyx_v_tolerance;
      ((struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *)__pyx_v_self->__pyx_vtab)->_find_previous_putative_peak_inplace(__pyx_v_self, __pyx_v_prev_peak_mz, __pyx_v_charge, __pyx_v_result, &__pyx_t_9); 
    }
    __pyx_L5:;
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":343
 *             else:
 *                 self._find_previous_putative_peak_inplace(prev_peak_mz, charge, result, step - 1, tolerance)
 *         return stop - start             # <<<<<<<<<<<<<<
 * 
 *     cpdef list _find_next_putative_peak(self, double mz, int charge, int step=1, double tolerance=2e-5):
 */
  __pyx_r = (__pyx_v_stop - __pyx_v_start);
  goto __pyx_L0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":322
 *         return stop - start
 * 
 *     cdef int _find_previous_putative_peak_inplace(self, double mz, int charge, set result, int step=1, double tolerance=2e-5):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double shift, prev_peak, prev_peak_mz
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_WriteUnraisable("ms_deisotope._c.deconvoluter_base.DeconvoluterBase._find_previous_putative_peak_inplace", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_backward);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/deconvoluter_base.pyx":345
 *         return stop - start
 * 
 *     cpdef list _find_next_putative_peak(self, double mz, int charge, int step=1, double tolerance=2e-5):             # <<<<<<<<<<<<<<
 *         """
 *         Recalibrates the current peak location given the position of the next putative peak
 */

static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_21_find_next_putative_peak(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, double __pyx_v_mz, int __pyx_v_charge, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak *__pyx_optional_args) {
  int __pyx_v_step = ((int)1);
  double __pyx_v_tolerance = ((double)2e-5);
  double __pyx_v_shift;
  double __pyx_v_next_peak;
  double __pyx_v_prev_peak_mz;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist_slice = 0;
  PyObject *__pyx_v_candidates = 0;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_dummy_peak = 0;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_forward = 0;
  size_t __pyx_v_i;
  size_t __pyx_v_n;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  double __pyx_t_11;
  struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_isotopic_shift __pyx_t_12;
  size_t __pyx_t_13;
  size_t __pyx_t_14;
  size_t __pyx_t_15;
  int __pyx_t_16;
  __Pyx_RefNannySetupContext("_find_next_putative_peak", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_step = __pyx_optional_args->step;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_tolerance = __pyx_optional_args->tolerance;
      }
    }
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_find_next_putative_peak); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 345, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_21_find_next_putative_peak)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_mz); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 345, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_charge); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 345, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_step); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 345, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_6 = PyFloat_FromDouble(__pyx_v_tolerance); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 345, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_8 = NULL;
        __pyx_t_9 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_8)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_8);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_9 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[5] = {__pyx_t_8, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_9, 4+__pyx_t_9); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 345, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[5] = {__pyx_t_8, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_9, 4+__pyx_t_9); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 345, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        } else
        #endif
        {
          __pyx_t_10 = PyTuple_New(4+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 345, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          if (__pyx_t_8) {
            __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_8); __pyx_t_8 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_t_3);
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_t_4);
          __Pyx_GIVEREF(__pyx_t_5);
          PyTuple_SET_ITEM(__pyx_t_10, 2+__pyx_t_9, __pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_6);
          PyTuple_SET_ITEM(__pyx_t_10, 3+__pyx_t_9, __pyx_t_6);
          __pyx_t_3 = 0;
          __pyx_t_4 = 0;
          __pyx_t_5 = 0;
          __pyx_t_6 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_10, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 345, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 345, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":380
 *             size_t i, n
 * 
 *         shift = isotopic_shift(charge)             # <<<<<<<<<<<<<<
 *         next_peak = mz + (shift * step)
 *         peaklist_slice = self.between(
 */
  __pyx_t_12.__pyx_n = 1;
  __pyx_t_12.charge = __pyx_v_charge;
  __pyx_t_11 = __pyx_f_12ms_deisotope_2_c_9averagine_isotopic_shift(0, &__pyx_t_12); 
  __pyx_v_shift = __pyx_t_11;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":381
 * 
 *         shift = isotopic_shift(charge)
 *         next_peak = mz + (shift * step)             # <<<<<<<<<<<<<<
 *         peaklist_slice = self.between(
 *             next_peak - (next_peak * tolerance),
 */
  __pyx_v_next_peak = (__pyx_v_mz + (__pyx_v_shift * __pyx_v_step));

  /* "ms_deisotope/_c/deconvoluter_base.pyx":382
 *         shift = isotopic_shift(charge)
 *         next_peak = mz + (shift * step)
 *         peaklist_slice = self.between(             # <<<<<<<<<<<<<<
 *             next_peak - (next_peak * tolerance),
 *             next_peak + (next_peak * tolerance))
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *)__pyx_v_self->__pyx_vtab)->between(__pyx_v_self, (__pyx_v_next_peak - (__pyx_v_next_peak * __pyx_v_tolerance)), (__pyx_v_next_peak + (__pyx_v_next_peak * __pyx_v_tolerance)), 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 382, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_peaklist_slice = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":385
 *             next_peak - (next_peak * tolerance),
 *             next_peak + (next_peak * tolerance))
 *         candidates = []             # <<<<<<<<<<<<<<
 * 
 *         n = peaklist_slice.get_size()
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_candidates = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":387
 *         candidates = []
 * 
 *         n = peaklist_slice.get_size()             # <<<<<<<<<<<<<<
 *         for i in range(n):
 *             forward = peaklist_slice.getitem(i)
 */
  __pyx_v_n = ((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaklist_slice->__pyx_vtab)->get_size(__pyx_v_peaklist_slice);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":388
 * 
 *         n = peaklist_slice.get_size()
 *         for i in range(n):             # <<<<<<<<<<<<<<
 *             forward = peaklist_slice.getitem(i)
 *             prev_peak_mz = forward.mz - (shift * step)
 */
  __pyx_t_13 = __pyx_v_n;
  __pyx_t_14 = __pyx_t_13;
  for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
    __pyx_v_i = __pyx_t_15;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":389
 *         n = peaklist_slice.get_size()
 *         for i in range(n):
 *             forward = peaklist_slice.getitem(i)             # <<<<<<<<<<<<<<
 *             prev_peak_mz = forward.mz - (shift * step)
 *             dummy_peak = make_placeholder_peak(prev_peak_mz)
 */
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaklist_slice->__pyx_vtab)->getitem(__pyx_v_peaklist_slice, __pyx_v_i)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 389, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_forward, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":390
 *         for i in range(n):
 *             forward = peaklist_slice.getitem(i)
 *             prev_peak_mz = forward.mz - (shift * step)             # <<<<<<<<<<<<<<
 *             dummy_peak = make_placeholder_peak(prev_peak_mz)
 *             candidates.append((dummy_peak, charge))
 */
    __pyx_v_prev_peak_mz = (__pyx_v_forward->__pyx_base.mz - (__pyx_v_shift * __pyx_v_step));

    /* "ms_deisotope/_c/deconvoluter_base.pyx":391
 *             forward = peaklist_slice.getitem(i)
 *             prev_peak_mz = forward.mz - (shift * step)
 *             dummy_peak = make_placeholder_peak(prev_peak_mz)             # <<<<<<<<<<<<<<
 *             candidates.append((dummy_peak, charge))
 *         return candidates
 */
    __pyx_t_1 = ((PyObject *)__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_make_placeholder_peak(__pyx_v_prev_peak_mz)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 391, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_dummy_peak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":392
 *             prev_peak_mz = forward.mz - (shift * step)
 *             dummy_peak = make_placeholder_peak(prev_peak_mz)
 *             candidates.append((dummy_peak, charge))             # <<<<<<<<<<<<<<
 *         return candidates
 * 
 */
    __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_charge); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 392, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 392, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)__pyx_v_dummy_peak));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_dummy_peak));
    PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_v_dummy_peak));
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_16 = __Pyx_PyList_Append(__pyx_v_candidates, __pyx_t_2); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(0, 392, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":393
 *             dummy_peak = make_placeholder_peak(prev_peak_mz)
 *             candidates.append((dummy_peak, charge))
 *         return candidates             # <<<<<<<<<<<<<<
 * 
 *     cpdef list _find_previous_putative_peak(self, double mz, int charge, int step=1, double tolerance=2e-5):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_candidates);
  __pyx_r = __pyx_v_candidates;
  goto __pyx_L0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":345
 *         return stop - start
 * 
 *     cpdef list _find_next_putative_peak(self, double mz, int charge, int step=1, double tolerance=2e-5):             # <<<<<<<<<<<<<<
 *         """
 *         Recalibrates the current peak location given the position of the next putative peak
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.DeconvoluterBase._find_next_putative_peak", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_peaklist_slice);
  __Pyx_XDECREF(__pyx_v_candidates);
  __Pyx_XDECREF((PyObject *)__pyx_v_dummy_peak);
  __Pyx_XDECREF((PyObject *)__pyx_v_forward);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_21_find_next_putative_peak(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_20_find_next_putative_peak[] = "DeconvoluterBase._find_next_putative_peak(self, double mz, int charge, int step=1, double tolerance=2e-5) -> list\n\n        Recalibrates the current peak location given the position of the next putative peak\n        in a theoretical isotopic cluster.\n\n        Suppose that the peak at `mz` is roughly in the neighborhood of a real isotopic peak,\n        but the alignment is bad, so it won't make a good starting point for the search for the\n        rest of the peaks in its cluster under a stringent error tolerance.\n\n        However, if we're willing to search for the next putative peak with a more permissive error\n        tolerance, which we expect will be properly aligned with the rest of its isotopic cluster,\n        we can recalibrate the proper starting peak's mz and use that for isotopic cluster fitting.\n\n        Parameters\n        ----------\n        mz : float\n            Starting m/z value to search from\n        charge : int\n            Charge state to use when calculating the step size in m/z\n        step : int, optional\n            The number of steps into the putative isotopic cluster to take. Defaults to 1\n        tolerance : float, optional\n            The error tolerance to accept for finding supporting peaks.\n\n        Returns\n        -------\n        list\n        ";
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_21_find_next_putative_peak(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_mz;
  int __pyx_v_charge;
  int __pyx_v_step;
  double __pyx_v_tolerance;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_find_next_putative_peak (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_mz,&__pyx_n_s_charge,&__pyx_n_s_step,&__pyx_n_s_tolerance,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mz)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_charge)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_find_next_putative_peak", 0, 2, 4, 1); __PYX_ERR(0, 345, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_step);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tolerance);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_find_next_putative_peak") < 0)) __PYX_ERR(0, 345, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_mz = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_mz == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 345, __pyx_L3_error)
    __pyx_v_charge = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_charge == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 345, __pyx_L3_error)
    if (values[2]) {
      __pyx_v_step = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_step == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 345, __pyx_L3_error)
    } else {
      __pyx_v_step = ((int)1);
    }
    if (values[3]) {
      __pyx_v_tolerance = __pyx_PyFloat_AsDouble(values[3]); if (unlikely((__pyx_v_tolerance == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 345, __pyx_L3_error)
    } else {
      __pyx_v_tolerance = ((double)2e-5);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_find_next_putative_peak", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 345, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.DeconvoluterBase._find_next_putative_peak", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_20_find_next_putative_peak(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *)__pyx_v_self), __pyx_v_mz, __pyx_v_charge, __pyx_v_step, __pyx_v_tolerance);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_20_find_next_putative_peak(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, double __pyx_v_mz, int __pyx_v_charge, int __pyx_v_step, double __pyx_v_tolerance) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak __pyx_t_2;
  __Pyx_RefNannySetupContext("_find_next_putative_peak", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 2;
  __pyx_t_2.step = __pyx_v_step;
  __pyx_t_2.tolerance = __pyx_v_tolerance;
  __pyx_t_1 = __pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase->_find_next_putative_peak(__pyx_v_self, __pyx_v_mz, __pyx_v_charge, 1, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 345, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.DeconvoluterBase._find_next_putative_peak", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/deconvoluter_base.pyx":395
 *         return candidates
 * 
 *     cpdef list _find_previous_putative_peak(self, double mz, int charge, int step=1, double tolerance=2e-5):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double shift, prev_peak, prev_peak_mz
 */

static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_23_find_previous_putative_peak(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, double __pyx_v_mz, int __pyx_v_charge, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak *__pyx_optional_args) {
  int __pyx_v_step = ((int)1);
  double __pyx_v_tolerance = ((double)2e-5);
  double __pyx_v_shift;
  double __pyx_v_prev_peak;
  double __pyx_v_prev_peak_mz;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist_slice = 0;
  PyObject *__pyx_v_candidates = 0;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_backward = 0;
  size_t __pyx_v_i;
  size_t __pyx_v_n;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  double __pyx_t_11;
  struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_isotopic_shift __pyx_t_12;
  size_t __pyx_t_13;
  size_t __pyx_t_14;
  size_t __pyx_t_15;
  int __pyx_t_16;
  struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak __pyx_t_17;
  int __pyx_t_18;
  struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak __pyx_t_19;
  __Pyx_RefNannySetupContext("_find_previous_putative_peak", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_step = __pyx_optional_args->step;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_tolerance = __pyx_optional_args->tolerance;
      }
    }
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_find_previous_putative_peak); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 395, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_23_find_previous_putative_peak)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_mz); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 395, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_charge); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 395, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_step); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 395, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_6 = PyFloat_FromDouble(__pyx_v_tolerance); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 395, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_8 = NULL;
        __pyx_t_9 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_8)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_8);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_9 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[5] = {__pyx_t_8, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_9, 4+__pyx_t_9); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 395, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[5] = {__pyx_t_8, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_9, 4+__pyx_t_9); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 395, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        } else
        #endif
        {
          __pyx_t_10 = PyTuple_New(4+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 395, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          if (__pyx_t_8) {
            __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_8); __pyx_t_8 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_t_3);
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_t_4);
          __Pyx_GIVEREF(__pyx_t_5);
          PyTuple_SET_ITEM(__pyx_t_10, 2+__pyx_t_9, __pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_6);
          PyTuple_SET_ITEM(__pyx_t_10, 3+__pyx_t_9, __pyx_t_6);
          __pyx_t_3 = 0;
          __pyx_t_4 = 0;
          __pyx_t_5 = 0;
          __pyx_t_6 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_10, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 395, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 395, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":403
 *             size_t i, n
 * 
 *         shift = isotopic_shift(charge)             # <<<<<<<<<<<<<<
 *         prev_peak = mz - (shift)
 *         peaklist_slice = self.between(
 */
  __pyx_t_12.__pyx_n = 1;
  __pyx_t_12.charge = __pyx_v_charge;
  __pyx_t_11 = __pyx_f_12ms_deisotope_2_c_9averagine_isotopic_shift(0, &__pyx_t_12); 
  __pyx_v_shift = __pyx_t_11;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":404
 * 
 *         shift = isotopic_shift(charge)
 *         prev_peak = mz - (shift)             # <<<<<<<<<<<<<<
 *         peaklist_slice = self.between(
 *             prev_peak - (prev_peak * tolerance),
 */
  __pyx_v_prev_peak = (__pyx_v_mz - __pyx_v_shift);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":405
 *         shift = isotopic_shift(charge)
 *         prev_peak = mz - (shift)
 *         peaklist_slice = self.between(             # <<<<<<<<<<<<<<
 *             prev_peak - (prev_peak * tolerance),
 *             prev_peak + (prev_peak * tolerance))
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *)__pyx_v_self->__pyx_vtab)->between(__pyx_v_self, (__pyx_v_prev_peak - (__pyx_v_prev_peak * __pyx_v_tolerance)), (__pyx_v_prev_peak + (__pyx_v_prev_peak * __pyx_v_tolerance)), 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 405, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_peaklist_slice = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":408
 *             prev_peak - (prev_peak * tolerance),
 *             prev_peak + (prev_peak * tolerance))
 *         candidates = []             # <<<<<<<<<<<<<<
 *         n = peaklist_slice.get_size()
 *         for i in range(n):
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 408, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_candidates = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":409
 *             prev_peak + (prev_peak * tolerance))
 *         candidates = []
 *         n = peaklist_slice.get_size()             # <<<<<<<<<<<<<<
 *         for i in range(n):
 *             backward = peaklist_slice.getitem(i)
 */
  __pyx_v_n = ((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaklist_slice->__pyx_vtab)->get_size(__pyx_v_peaklist_slice);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":410
 *         candidates = []
 *         n = peaklist_slice.get_size()
 *         for i in range(n):             # <<<<<<<<<<<<<<
 *             backward = peaklist_slice.getitem(i)
 *             prev_peak_mz = backward.mz
 */
  __pyx_t_13 = __pyx_v_n;
  __pyx_t_14 = __pyx_t_13;
  for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
    __pyx_v_i = __pyx_t_15;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":411
 *         n = peaklist_slice.get_size()
 *         for i in range(n):
 *             backward = peaklist_slice.getitem(i)             # <<<<<<<<<<<<<<
 *             prev_peak_mz = backward.mz
 *             if step == 1:
 */
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaklist_slice->__pyx_vtab)->getitem(__pyx_v_peaklist_slice, __pyx_v_i)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 411, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_backward, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":412
 *         for i in range(n):
 *             backward = peaklist_slice.getitem(i)
 *             prev_peak_mz = backward.mz             # <<<<<<<<<<<<<<
 *             if step == 1:
 *                 candidates.extend(self._find_next_putative_peak(prev_peak_mz, charge, 1, tolerance))
 */
    __pyx_t_11 = __pyx_v_backward->__pyx_base.mz;
    __pyx_v_prev_peak_mz = __pyx_t_11;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":413
 *             backward = peaklist_slice.getitem(i)
 *             prev_peak_mz = backward.mz
 *             if step == 1:             # <<<<<<<<<<<<<<
 *                 candidates.extend(self._find_next_putative_peak(prev_peak_mz, charge, 1, tolerance))
 *             else:
 */
    __pyx_t_16 = ((__pyx_v_step == 1) != 0);
    if (__pyx_t_16) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":414
 *             prev_peak_mz = backward.mz
 *             if step == 1:
 *                 candidates.extend(self._find_next_putative_peak(prev_peak_mz, charge, 1, tolerance))             # <<<<<<<<<<<<<<
 *             else:
 *                 candidates.extend(
 */
      __pyx_t_17.__pyx_n = 2;
      __pyx_t_17.step = 1;
      __pyx_t_17.tolerance = __pyx_v_tolerance;
      __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *)__pyx_v_self->__pyx_vtab)->_find_next_putative_peak(__pyx_v_self, __pyx_v_prev_peak_mz, __pyx_v_charge, 0, &__pyx_t_17); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 414, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_18 = __Pyx_PyList_Extend(__pyx_v_candidates, __pyx_t_1); if (unlikely(__pyx_t_18 == ((int)-1))) __PYX_ERR(0, 414, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":413
 *             backward = peaklist_slice.getitem(i)
 *             prev_peak_mz = backward.mz
 *             if step == 1:             # <<<<<<<<<<<<<<
 *                 candidates.extend(self._find_next_putative_peak(prev_peak_mz, charge, 1, tolerance))
 *             else:
 */
      goto __pyx_L5;
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":416
 *                 candidates.extend(self._find_next_putative_peak(prev_peak_mz, charge, 1, tolerance))
 *             else:
 *                 candidates.extend(             # <<<<<<<<<<<<<<
 *                     self._find_previous_putative_peak(prev_peak_mz, charge, step - 1, tolerance))
 *         return candidates
 */
    /*else*/ {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":417
 *             else:
 *                 candidates.extend(
 *                     self._find_previous_putative_peak(prev_peak_mz, charge, step - 1, tolerance))             # <<<<<<<<<<<<<<
 *         return candidates
 * 
 */
      __pyx_t_19.__pyx_n = 2;
      __pyx_t_19.step = (__pyx_v_step - 1);
      __pyx_t_19.tolerance = __pyx_v_tolerance;
      __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *)__pyx_v_self->__pyx_vtab)->_find_previous_putative_peak(__pyx_v_self, __pyx_v_prev_peak_mz, __pyx_v_charge, 0, &__pyx_t_19); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 417, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);

      /* "ms_deisotope/_c/deconvoluter_base.pyx":416
 *                 candidates.extend(self._find_next_putative_peak(prev_peak_mz, charge, 1, tolerance))
 *             else:
 *                 candidates.extend(             # <<<<<<<<<<<<<<
 *                     self._find_previous_putative_peak(prev_peak_mz, charge, step - 1, tolerance))
 *         return candidates
 */
      __pyx_t_18 = __Pyx_PyList_Extend(__pyx_v_candidates, __pyx_t_1); if (unlikely(__pyx_t_18 == ((int)-1))) __PYX_ERR(0, 416, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L5:;
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":418
 *                 candidates.extend(
 *                     self._find_previous_putative_peak(prev_peak_mz, charge, step - 1, tolerance))
 *         return candidates             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_candidates);
  __pyx_r = __pyx_v_candidates;
  goto __pyx_L0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":395
 *         return candidates
 * 
 *     cpdef list _find_previous_putative_peak(self, double mz, int charge, int step=1, double tolerance=2e-5):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double shift, prev_peak, prev_peak_mz
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.DeconvoluterBase._find_previous_putative_peak", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_peaklist_slice);
  __Pyx_XDECREF(__pyx_v_candidates);
  __Pyx_XDECREF((PyObject *)__pyx_v_backward);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_23_find_previous_putative_peak(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_22_find_previous_putative_peak[] = "DeconvoluterBase._find_previous_putative_peak(self, double mz, int charge, int step=1, double tolerance=2e-5) -> list";
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_23_find_previous_putative_peak(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_mz;
  int __pyx_v_charge;
  int __pyx_v_step;
  double __pyx_v_tolerance;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_find_previous_putative_peak (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_mz,&__pyx_n_s_charge,&__pyx_n_s_step,&__pyx_n_s_tolerance,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mz)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_charge)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_find_previous_putative_peak", 0, 2, 4, 1); __PYX_ERR(0, 395, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_step);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tolerance);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_find_previous_putative_peak") < 0)) __PYX_ERR(0, 395, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_mz = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_mz == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 395, __pyx_L3_error)
    __pyx_v_charge = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_charge == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 395, __pyx_L3_error)
    if (values[2]) {
      __pyx_v_step = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_step == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 395, __pyx_L3_error)
    } else {
      __pyx_v_step = ((int)1);
    }
    if (values[3]) {
      __pyx_v_tolerance = __pyx_PyFloat_AsDouble(values[3]); if (unlikely((__pyx_v_tolerance == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 395, __pyx_L3_error)
    } else {
      __pyx_v_tolerance = ((double)2e-5);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_find_previous_putative_peak", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 395, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.DeconvoluterBase._find_previous_putative_peak", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_22_find_previous_putative_peak(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *)__pyx_v_self), __pyx_v_mz, __pyx_v_charge, __pyx_v_step, __pyx_v_tolerance);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_22_find_previous_putative_peak(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, double __pyx_v_mz, int __pyx_v_charge, int __pyx_v_step, double __pyx_v_tolerance) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak __pyx_t_2;
  __Pyx_RefNannySetupContext("_find_previous_putative_peak", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 2;
  __pyx_t_2.step = __pyx_v_step;
  __pyx_t_2.tolerance = __pyx_v_tolerance;
  __pyx_t_1 = __pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase->_find_previous_putative_peak(__pyx_v_self, __pyx_v_mz, __pyx_v_charge, 1, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 395, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.DeconvoluterBase._find_previous_putative_peak", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/deconvoluter_base.pyx":420
 *         return candidates
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         type_name = self.__class__.__name__
 *         return "%s(peaklist=%s, scorer=%s)" % (type_name, self.peaklist, self.scorer)
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_25__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_25__repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_24__repr__(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_24__repr__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self) {
  PyObject *__pyx_v_type_name = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":421
 * 
 *     def __repr__(self):
 *         type_name = self.__class__.__name__             # <<<<<<<<<<<<<<
 *         return "%s(peaklist=%s, scorer=%s)" % (type_name, self.peaklist, self.scorer)
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 421, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 421, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_type_name = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":422
 *     def __repr__(self):
 *         type_name = self.__class__.__name__
 *         return "%s(peaklist=%s, scorer=%s)" % (type_name, self.peaklist, self.scorer)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 422, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_type_name);
  __Pyx_GIVEREF(__pyx_v_type_name);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_type_name);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->peaklist));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->peaklist));
  PyTuple_SET_ITEM(__pyx_t_2, 1, ((PyObject *)__pyx_v_self->peaklist));
  __Pyx_INCREF(((PyObject *)__pyx_v_self->scorer));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->scorer));
  PyTuple_SET_ITEM(__pyx_t_2, 2, ((PyObject *)__pyx_v_self->scorer));
  __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_s_peaklist_s_scorer_s, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 422, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":420
 *         return candidates
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         type_name = self.__class__.__name__
 *         return "%s(peaklist=%s, scorer=%s)" % (type_name, self.peaklist, self.scorer)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.DeconvoluterBase.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_type_name);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/deconvoluter_base.pxd":27
 * cdef class DeconvoluterBase(object):
 *     cdef:
 *         public bint use_subtraction             # <<<<<<<<<<<<<<
 *         public str scale_method
 *         public bint merge_isobaric_peaks
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_15use_subtraction_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_15use_subtraction_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_15use_subtraction___get__(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_15use_subtraction___get__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->use_subtraction); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.DeconvoluterBase.use_subtraction.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_15use_subtraction_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_15use_subtraction_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_15use_subtraction_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_15use_subtraction_2__set__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 27, __pyx_L1_error)
  __pyx_v_self->use_subtraction = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.DeconvoluterBase.use_subtraction.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/deconvoluter_base.pxd":28
 *     cdef:
 *         public bint use_subtraction
 *         public str scale_method             # <<<<<<<<<<<<<<
 *         public bint merge_isobaric_peaks
 *         public double minimum_intensity
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_12scale_method_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_12scale_method_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_12scale_method___get__(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_12scale_method___get__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->scale_method);
  __pyx_r = __pyx_v_self->scale_method;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_12scale_method_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_12scale_method_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_12scale_method_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_12scale_method_2__set__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(PyString_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(2, 28, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->scale_method);
  __Pyx_DECREF(__pyx_v_self->scale_method);
  __pyx_v_self->scale_method = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.DeconvoluterBase.scale_method.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_12scale_method_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_12scale_method_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_12scale_method_4__del__(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_12scale_method_4__del__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->scale_method);
  __Pyx_DECREF(__pyx_v_self->scale_method);
  __pyx_v_self->scale_method = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/deconvoluter_base.pxd":29
 *         public bint use_subtraction
 *         public str scale_method
 *         public bint merge_isobaric_peaks             # <<<<<<<<<<<<<<
 *         public double minimum_intensity
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_20merge_isobaric_peaks_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_20merge_isobaric_peaks_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_20merge_isobaric_peaks___get__(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_20merge_isobaric_peaks___get__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->merge_isobaric_peaks); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.DeconvoluterBase.merge_isobaric_peaks.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_20merge_isobaric_peaks_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_20merge_isobaric_peaks_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_20merge_isobaric_peaks_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_20merge_isobaric_peaks_2__set__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 29, __pyx_L1_error)
  __pyx_v_self->merge_isobaric_peaks = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.DeconvoluterBase.merge_isobaric_peaks.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/deconvoluter_base.pxd":30
 *         public str scale_method
 *         public bint merge_isobaric_peaks
 *         public double minimum_intensity             # <<<<<<<<<<<<<<
 * 
 *         public PeakSet peaklist
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_17minimum_intensity_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_17minimum_intensity_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_17minimum_intensity___get__(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_17minimum_intensity___get__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->minimum_intensity); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.DeconvoluterBase.minimum_intensity.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_17minimum_intensity_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_17minimum_intensity_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_17minimum_intensity_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_17minimum_intensity_2__set__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 30, __pyx_L1_error)
  __pyx_v_self->minimum_intensity = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.DeconvoluterBase.minimum_intensity.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/deconvoluter_base.pxd":32
 *         public double minimum_intensity
 * 
 *         public PeakSet peaklist             # <<<<<<<<<<<<<<
 * 
 *         public IsotopicFitterBase scorer
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_8peaklist_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_8peaklist_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_8peaklist___get__(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_8peaklist___get__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->peaklist));
  __pyx_r = ((PyObject *)__pyx_v_self->peaklist);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_8peaklist_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_8peaklist_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_8peaklist_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_8peaklist_2__set__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet))))) __PYX_ERR(2, 32, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->peaklist);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->peaklist));
  __pyx_v_self->peaklist = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.DeconvoluterBase.peaklist.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_8peaklist_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_8peaklist_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_8peaklist_4__del__(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_8peaklist_4__del__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->peaklist);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->peaklist));
  __pyx_v_self->peaklist = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/deconvoluter_base.pxd":34
 *         public PeakSet peaklist
 * 
 *         public IsotopicFitterBase scorer             # <<<<<<<<<<<<<<
 *         public bint verbose
 *         public dict _slice_cache
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_6scorer_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_6scorer_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_6scorer___get__(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_6scorer___get__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->scorer));
  __pyx_r = ((PyObject *)__pyx_v_self->scorer);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_6scorer_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_6scorer_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_6scorer_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_6scorer_2__set__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitterBase))))) __PYX_ERR(2, 34, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->scorer);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->scorer));
  __pyx_v_self->scorer = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.DeconvoluterBase.scorer.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_6scorer_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_6scorer_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_6scorer_4__del__(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_6scorer_4__del__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->scorer);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->scorer));
  __pyx_v_self->scorer = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/deconvoluter_base.pxd":35
 * 
 *         public IsotopicFitterBase scorer
 *         public bint verbose             # <<<<<<<<<<<<<<
 *         public dict _slice_cache
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_7verbose_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_7verbose_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_7verbose___get__(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_7verbose___get__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->verbose); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.DeconvoluterBase.verbose.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_7verbose_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_7verbose_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_7verbose_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_7verbose_2__set__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 35, __pyx_L1_error)
  __pyx_v_self->verbose = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.DeconvoluterBase.verbose.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/deconvoluter_base.pxd":36
 *         public IsotopicFitterBase scorer
 *         public bint verbose
 *         public dict _slice_cache             # <<<<<<<<<<<<<<
 * 
 *     cpdef PeakSet between(self, double m1, double m2)
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_12_slice_cache_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_12_slice_cache_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_12_slice_cache___get__(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_12_slice_cache___get__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_slice_cache);
  __pyx_r = __pyx_v_self->_slice_cache;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_12_slice_cache_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_12_slice_cache_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_12_slice_cache_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_12_slice_cache_2__set__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(PyDict_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(2, 36, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_slice_cache);
  __Pyx_DECREF(__pyx_v_self->_slice_cache);
  __pyx_v_self->_slice_cache = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.DeconvoluterBase._slice_cache.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_12_slice_cache_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_12_slice_cache_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_12_slice_cache_4__del__(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_12_slice_cache_4__del__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_slice_cache);
  __Pyx_DECREF(__pyx_v_self->_slice_cache);
  __pyx_v_self->_slice_cache = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_27__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_26__reduce_cython__[] = "DeconvoluterBase.__reduce_cython__(self)";
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_27__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_26__reduce_cython__(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_26__reduce_cython__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self._slice_cache, self.merge_isobaric_peaks, self.minimum_intensity, self.peaklist, self.scale_method, self.scorer, self.use_subtraction, self.verbose)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->merge_isobaric_peaks); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->minimum_intensity); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_self->use_subtraction); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_v_self->verbose); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(8); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_v_self->_slice_cache);
  __Pyx_GIVEREF(__pyx_v_self->_slice_cache);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_self->_slice_cache);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_2);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->peaklist));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->peaklist));
  PyTuple_SET_ITEM(__pyx_t_5, 3, ((PyObject *)__pyx_v_self->peaklist));
  __Pyx_INCREF(__pyx_v_self->scale_method);
  __Pyx_GIVEREF(__pyx_v_self->scale_method);
  PyTuple_SET_ITEM(__pyx_t_5, 4, __pyx_v_self->scale_method);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->scorer));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->scorer));
  PyTuple_SET_ITEM(__pyx_t_5, 5, ((PyObject *)__pyx_v_self->scorer));
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 6, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 7, __pyx_t_4);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self._slice_cache, self.merge_isobaric_peaks, self.minimum_intensity, self.peaklist, self.scale_method, self.scorer, self.use_subtraction, self.verbose)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_5 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_v__dict = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "(tree fragment)":7
 *     state = (self._slice_cache, self.merge_isobaric_peaks, self.minimum_intensity, self.peaklist, self.scale_method, self.scorer, self.use_subtraction, self.verbose)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_6 = (__pyx_v__dict != Py_None);
  __pyx_t_7 = (__pyx_t_6 != 0);
  if (__pyx_t_7) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v__dict);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self._slice_cache is not None or self.peaklist is not None or self.scale_method is not None or self.scorer is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self._slice_cache, self.merge_isobaric_peaks, self.minimum_intensity, self.peaklist, self.scale_method, self.scorer, self.use_subtraction, self.verbose)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self._slice_cache is not None or self.peaklist is not None or self.scale_method is not None or self.scorer is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_DeconvoluterBase, (type(self), 0x16fc763, None), state
 */
  /*else*/ {
    __pyx_t_6 = (__pyx_v_self->_slice_cache != ((PyObject*)Py_None));
    __pyx_t_8 = (__pyx_t_6 != 0);
    if (!__pyx_t_8) {
    } else {
      __pyx_t_7 = __pyx_t_8;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_8 = (((PyObject *)__pyx_v_self->peaklist) != Py_None);
    __pyx_t_6 = (__pyx_t_8 != 0);
    if (!__pyx_t_6) {
    } else {
      __pyx_t_7 = __pyx_t_6;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_6 = (__pyx_v_self->scale_method != ((PyObject*)Py_None));
    __pyx_t_8 = (__pyx_t_6 != 0);
    if (!__pyx_t_8) {
    } else {
      __pyx_t_7 = __pyx_t_8;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_8 = (((PyObject *)__pyx_v_self->scorer) != Py_None);
    __pyx_t_6 = (__pyx_t_8 != 0);
    __pyx_t_7 = __pyx_t_6;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_7;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._slice_cache is not None or self.peaklist is not None or self.scale_method is not None or self.scorer is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_DeconvoluterBase, (type(self), 0x16fc763, None), state
 *     else:
 */
  __pyx_t_7 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_7) {

    /* "(tree fragment)":13
 *         use_setstate = self._slice_cache is not None or self.peaklist is not None or self.scale_method is not None or self.scorer is not None
 *     if use_setstate:
 *         return __pyx_unpickle_DeconvoluterBase, (type(self), 0x16fc763, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_DeconvoluterBase, (type(self), 0x16fc763, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_DeconvoluterBase); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_24102755);
    __Pyx_GIVEREF(__pyx_int_24102755);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_int_24102755);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_5, 2, Py_None);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_5);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_state);
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._slice_cache is not None or self.peaklist is not None or self.scale_method is not None or self.scorer is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_DeconvoluterBase, (type(self), 0x16fc763, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_DeconvoluterBase, (type(self), 0x16fc763, None), state
 *     else:
 *         return __pyx_unpickle_DeconvoluterBase, (type(self), 0x16fc763, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_DeconvoluterBase__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_DeconvoluterBase); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_24102755);
    __Pyx_GIVEREF(__pyx_int_24102755);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_int_24102755);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_5);
    __pyx_t_3 = 0;
    __pyx_t_5 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.DeconvoluterBase.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_DeconvoluterBase, (type(self), 0x16fc763, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_DeconvoluterBase__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_29__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_28__setstate_cython__[] = "DeconvoluterBase.__setstate_cython__(self, __pyx_state)";
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_29__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_28__setstate_cython__(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_28__setstate_cython__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_DeconvoluterBase, (type(self), 0x16fc763, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_DeconvoluterBase__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_17deconvoluter_base___pyx_unpickle_DeconvoluterBase__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_DeconvoluterBase, (type(self), 0x16fc763, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_DeconvoluterBase__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.DeconvoluterBase.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/deconvoluter_base.pyx":426
 * 
 * 
 * cdef bint has_multiple_real_peaks(list peaklist):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i
 */

static int __pyx_f_12ms_deisotope_2_c_17deconvoluter_base_has_multiple_real_peaks(PyObject *__pyx_v_peaklist) {
  size_t __pyx_v_i;
  int __pyx_v_n;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_peak = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  size_t __pyx_t_3;
  PyObject *__pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  __Pyx_RefNannySetupContext("has_multiple_real_peaks", 0);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":432
 *         FittedPeak peak
 * 
 *     n = 0             # <<<<<<<<<<<<<<
 *     for i in range(PyList_GET_SIZE(peaklist)):
 *         peak = <FittedPeak>PyList_GET_ITEM(peaklist, i)
 */
  __pyx_v_n = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":433
 * 
 *     n = 0
 *     for i in range(PyList_GET_SIZE(peaklist)):             # <<<<<<<<<<<<<<
 *         peak = <FittedPeak>PyList_GET_ITEM(peaklist, i)
 *         if peak.mz > 1 and peak.intensity > 1:
 */
  __pyx_t_1 = PyList_GET_SIZE(__pyx_v_peaklist);
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":434
 *     n = 0
 *     for i in range(PyList_GET_SIZE(peaklist)):
 *         peak = <FittedPeak>PyList_GET_ITEM(peaklist, i)             # <<<<<<<<<<<<<<
 *         if peak.mz > 1 and peak.intensity > 1:
 *             n += 1
 */
    __pyx_t_4 = PyList_GET_ITEM(__pyx_v_peaklist, __pyx_v_i);
    __pyx_t_5 = ((PyObject *)__pyx_t_4);
    __Pyx_INCREF(__pyx_t_5);
    __Pyx_XDECREF_SET(__pyx_v_peak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_5));
    __pyx_t_5 = 0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":435
 *     for i in range(PyList_GET_SIZE(peaklist)):
 *         peak = <FittedPeak>PyList_GET_ITEM(peaklist, i)
 *         if peak.mz > 1 and peak.intensity > 1:             # <<<<<<<<<<<<<<
 *             n += 1
 *     return n > 1
 */
    __pyx_t_7 = ((__pyx_v_peak->__pyx_base.mz > 1.0) != 0);
    if (__pyx_t_7) {
    } else {
      __pyx_t_6 = __pyx_t_7;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_7 = ((__pyx_v_peak->__pyx_base.intensity > 1.0) != 0);
    __pyx_t_6 = __pyx_t_7;
    __pyx_L6_bool_binop_done:;
    if (__pyx_t_6) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":436
 *         peak = <FittedPeak>PyList_GET_ITEM(peaklist, i)
 *         if peak.mz > 1 and peak.intensity > 1:
 *             n += 1             # <<<<<<<<<<<<<<
 *     return n > 1
 * 
 */
      __pyx_v_n = (__pyx_v_n + 1);

      /* "ms_deisotope/_c/deconvoluter_base.pyx":435
 *     for i in range(PyList_GET_SIZE(peaklist)):
 *         peak = <FittedPeak>PyList_GET_ITEM(peaklist, i)
 *         if peak.mz > 1 and peak.intensity > 1:             # <<<<<<<<<<<<<<
 *             n += 1
 *     return n > 1
 */
    }
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":437
 *         if peak.mz > 1 and peak.intensity > 1:
 *             n += 1
 *     return n > 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = (__pyx_v_n > 1);
  goto __pyx_L0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":426
 * 
 * 
 * cdef bint has_multiple_real_peaks(list peaklist):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/deconvoluter_base.pyx":445
 *     """
 * 
 *     def __init__(self, bint use_subtraction=False, str scale_method="sum", bint merge_isobaric_peaks=True,             # <<<<<<<<<<<<<<
 *                  double minimum_intensity=5., *args, **kwargs):
 *         super(AveragineDeconvoluterBase, self).__init__(
 */

/* Python wrapper */
static int __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_v_use_subtraction;
  PyObject *__pyx_v_scale_method = 0;
  int __pyx_v_merge_isobaric_peaks;
  double __pyx_v_minimum_intensity;
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  __pyx_v_kwargs = PyDict_New(); if (unlikely(!__pyx_v_kwargs)) return -1;
  __Pyx_GOTREF(__pyx_v_kwargs);
  if (PyTuple_GET_SIZE(__pyx_args) > 4) {
    __pyx_v_args = PyTuple_GetSlice(__pyx_args, 4, PyTuple_GET_SIZE(__pyx_args));
    if (unlikely(!__pyx_v_args)) {
      __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
      __Pyx_RefNannyFinishContext();
      return -1;
    }
    __Pyx_GOTREF(__pyx_v_args);
  } else {
    __pyx_v_args = __pyx_empty_tuple; __Pyx_INCREF(__pyx_empty_tuple);
  }
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_use_subtraction,&__pyx_n_s_scale_method,&__pyx_n_s_merge_isobaric_peaks,&__pyx_n_s_minimum_intensity,0};
    PyObject* values[4] = {0,0,0,0};
    values[1] = ((PyObject*)__pyx_n_s_sum);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        default:
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_use_subtraction);
          if (value) { values[0] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_scale_method);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_merge_isobaric_peaks);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_minimum_intensity);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t used_pos_args = (pos_args < 4) ? pos_args : 4;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, __pyx_v_kwargs, values, used_pos_args, "__init__") < 0)) __PYX_ERR(0, 445, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        default:
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
      }
    }
    if (values[0]) {
      __pyx_v_use_subtraction = __Pyx_PyObject_IsTrue(values[0]); if (unlikely((__pyx_v_use_subtraction == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 445, __pyx_L3_error)
    } else {
      __pyx_v_use_subtraction = ((int)0);
    }
    __pyx_v_scale_method = ((PyObject*)values[1]);
    if (values[2]) {
      __pyx_v_merge_isobaric_peaks = __Pyx_PyObject_IsTrue(values[2]); if (unlikely((__pyx_v_merge_isobaric_peaks == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 445, __pyx_L3_error)
    } else {
      __pyx_v_merge_isobaric_peaks = ((int)1);
    }
    if (values[3]) {
      __pyx_v_minimum_intensity = __pyx_PyFloat_AsDouble(values[3]); if (unlikely((__pyx_v_minimum_intensity == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 446, __pyx_L3_error)
    } else {
      __pyx_v_minimum_intensity = ((double)5.);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_DECREF(__pyx_v_args); __pyx_v_args = 0;
  __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.AveragineDeconvoluterBase.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_scale_method), (&PyString_Type), 1, "scale_method", 1))) __PYX_ERR(0, 445, __pyx_L1_error)
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase___init__(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *)__pyx_v_self), __pyx_v_use_subtraction, __pyx_v_scale_method, __pyx_v_merge_isobaric_peaks, __pyx_v_minimum_intensity, __pyx_v_args, __pyx_v_kwargs);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase___init__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *__pyx_v_self, int __pyx_v_use_subtraction, PyObject *__pyx_v_scale_method, int __pyx_v_merge_isobaric_peaks, double __pyx_v_minimum_intensity, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":447
 *     def __init__(self, bint use_subtraction=False, str scale_method="sum", bint merge_isobaric_peaks=True,
 *                  double minimum_intensity=5., *args, **kwargs):
 *         super(AveragineDeconvoluterBase, self).__init__(             # <<<<<<<<<<<<<<
 *             use_subtraction, scale_method, merge_isobaric_peaks,
 *             minimum_intensity, *args, **kwargs)
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 447, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(((PyObject *)__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase));
  __Pyx_GIVEREF(((PyObject *)__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase));
  PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase));
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_self));
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 447, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 447, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":448
 *                  double minimum_intensity=5., *args, **kwargs):
 *         super(AveragineDeconvoluterBase, self).__init__(
 *             use_subtraction, scale_method, merge_isobaric_peaks,             # <<<<<<<<<<<<<<
 *             minimum_intensity, *args, **kwargs)
 * 
 */
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_use_subtraction); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 448, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_merge_isobaric_peaks); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 448, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":449
 *         super(AveragineDeconvoluterBase, self).__init__(
 *             use_subtraction, scale_method, merge_isobaric_peaks,
 *             minimum_intensity, *args, **kwargs)             # <<<<<<<<<<<<<<
 * 
 *     cpdef IsotopicFitRecord fit_theoretical_distribution(self, FittedPeak peak, double error_tolerance, int charge,
 */
  __pyx_t_4 = PyFloat_FromDouble(__pyx_v_minimum_intensity); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 449, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":447
 *     def __init__(self, bint use_subtraction=False, str scale_method="sum", bint merge_isobaric_peaks=True,
 *                  double minimum_intensity=5., *args, **kwargs):
 *         super(AveragineDeconvoluterBase, self).__init__(             # <<<<<<<<<<<<<<
 *             use_subtraction, scale_method, merge_isobaric_peaks,
 *             minimum_intensity, *args, **kwargs)
 */
  __pyx_t_5 = PyTuple_New(4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 447, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_scale_method);
  __Pyx_GIVEREF(__pyx_v_scale_method);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_v_scale_method);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_4);
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":449
 *         super(AveragineDeconvoluterBase, self).__init__(
 *             use_subtraction, scale_method, merge_isobaric_peaks,
 *             minimum_intensity, *args, **kwargs)             # <<<<<<<<<<<<<<
 * 
 *     cpdef IsotopicFitRecord fit_theoretical_distribution(self, FittedPeak peak, double error_tolerance, int charge,
 */
  __pyx_t_4 = PyNumber_Add(__pyx_t_5, __pyx_v_args); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 447, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":447
 *     def __init__(self, bint use_subtraction=False, str scale_method="sum", bint merge_isobaric_peaks=True,
 *                  double minimum_intensity=5., *args, **kwargs):
 *         super(AveragineDeconvoluterBase, self).__init__(             # <<<<<<<<<<<<<<
 *             use_subtraction, scale_method, merge_isobaric_peaks,
 *             minimum_intensity, *args, **kwargs)
 */
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_4, __pyx_v_kwargs); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 447, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":445
 *     """
 * 
 *     def __init__(self, bint use_subtraction=False, str scale_method="sum", bint merge_isobaric_peaks=True,             # <<<<<<<<<<<<<<
 *                  double minimum_intensity=5., *args, **kwargs):
 *         super(AveragineDeconvoluterBase, self).__init__(
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.AveragineDeconvoluterBase.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/deconvoluter_base.pyx":451
 *             minimum_intensity, *args, **kwargs)
 * 
 *     cpdef IsotopicFitRecord fit_theoretical_distribution(self, FittedPeak peak, double error_tolerance, int charge,             # <<<<<<<<<<<<<<
 *                                                          double charge_carrier=PROTON, double truncate_after=0.95,
 *                                                          double ignore_below=0):
 */

static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_3fit_theoretical_distribution(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_fit_theoretical_distribution(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_peak, double __pyx_v_error_tolerance, int __pyx_v_charge, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_fit_theoretical_distribution *__pyx_optional_args) {
  double __pyx_v_charge_carrier = __pyx_k__2;
  double __pyx_v_truncate_after = ((double)0.95);
  double __pyx_v_ignore_below = ((double)0.0);
  PyObject *__pyx_v_eid = 0;
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *__pyx_v_tid = 0;
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_14AveragineCache_isotopic_cluster __pyx_t_12;
  struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_match_theoretical_isotopic_distribution __pyx_t_13;
  __Pyx_RefNannySetupContext("fit_theoretical_distribution", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_charge_carrier = __pyx_optional_args->charge_carrier;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_truncate_after = __pyx_optional_args->truncate_after;
        if (__pyx_optional_args->__pyx_n > 2) {
          __pyx_v_ignore_below = __pyx_optional_args->ignore_below;
        }
      }
    }
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_fit_theoretical_distribution); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 451, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_3fit_theoretical_distribution)) {
        __Pyx_XDECREF(((PyObject *)__pyx_r));
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_error_tolerance); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 451, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_charge); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 451, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_5 = PyFloat_FromDouble(__pyx_v_charge_carrier); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 451, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_6 = PyFloat_FromDouble(__pyx_v_truncate_after); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 451, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_7 = PyFloat_FromDouble(__pyx_v_ignore_below); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 451, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_8 = __pyx_t_1; __pyx_t_9 = NULL;
        __pyx_t_10 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
          __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
          if (likely(__pyx_t_9)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
            __Pyx_INCREF(__pyx_t_9);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_8, function);
            __pyx_t_10 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_8)) {
          PyObject *__pyx_temp[7] = {__pyx_t_9, ((PyObject *)__pyx_v_peak), __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6, __pyx_t_7};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_10, 6+__pyx_t_10); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 451, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_8)) {
          PyObject *__pyx_temp[7] = {__pyx_t_9, ((PyObject *)__pyx_v_peak), __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6, __pyx_t_7};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_10, 6+__pyx_t_10); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 451, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        } else
        #endif
        {
          __pyx_t_11 = PyTuple_New(6+__pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 451, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_11);
          if (__pyx_t_9) {
            __Pyx_GIVEREF(__pyx_t_9); PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_9); __pyx_t_9 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_peak));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_peak));
          PyTuple_SET_ITEM(__pyx_t_11, 0+__pyx_t_10, ((PyObject *)__pyx_v_peak));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_11, 1+__pyx_t_10, __pyx_t_3);
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_11, 2+__pyx_t_10, __pyx_t_4);
          __Pyx_GIVEREF(__pyx_t_5);
          PyTuple_SET_ITEM(__pyx_t_11, 3+__pyx_t_10, __pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_6);
          PyTuple_SET_ITEM(__pyx_t_11, 4+__pyx_t_10, __pyx_t_6);
          __Pyx_GIVEREF(__pyx_t_7);
          PyTuple_SET_ITEM(__pyx_t_11, 5+__pyx_t_10, __pyx_t_7);
          __pyx_t_3 = 0;
          __pyx_t_4 = 0;
          __pyx_t_5 = 0;
          __pyx_t_6 = 0;
          __pyx_t_7 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_11, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 451, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        }
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitRecord))))) __PYX_ERR(0, 451, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":458
 *             TheoreticalIsotopicPattern tid
 *             double score
 *         tid = self.averagine.isotopic_cluster(             # <<<<<<<<<<<<<<
 *             peak.mz, charge, charge_carrier=charge_carrier, truncate_after=truncate_after,
 *             ignore_below=ignore_below)
 */
  __pyx_t_12.__pyx_n = 4;
  __pyx_t_12.charge = __pyx_v_charge;
  __pyx_t_12.charge_carrier = __pyx_v_charge_carrier;
  __pyx_t_12.truncate_after = __pyx_v_truncate_after;
  __pyx_t_12.ignore_below = __pyx_v_ignore_below;
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_AveragineCache *)__pyx_v_self->averagine->__pyx_vtab)->isotopic_cluster(__pyx_v_self->averagine, __pyx_v_peak->__pyx_base.mz, 0, &__pyx_t_12)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 458, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_tid = ((struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":461
 *             peak.mz, charge, charge_carrier=charge_carrier, truncate_after=truncate_after,
 *             ignore_below=ignore_below)
 *         eid = self.match_theoretical_isotopic_distribution(tid.peaklist, error_tolerance=error_tolerance)             # <<<<<<<<<<<<<<
 *         return self._evaluate_theoretical_distribution(eid, tid, peak, charge)
 * 
 */
  __pyx_t_1 = __pyx_v_tid->peaklist;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_13.__pyx_n = 1;
  __pyx_t_13.error_tolerance = __pyx_v_error_tolerance;
  __pyx_t_2 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.match_theoretical_isotopic_distribution(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *)__pyx_v_self), ((PyObject*)__pyx_t_1), 0, &__pyx_t_13); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 461, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_eid = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":462
 *             ignore_below=ignore_below)
 *         eid = self.match_theoretical_isotopic_distribution(tid.peaklist, error_tolerance=error_tolerance)
 *         return self._evaluate_theoretical_distribution(eid, tid, peak, charge)             # <<<<<<<<<<<<<<
 * 
 *     cpdef set _fit_peaks_at_charges(self, set peak_charge_set, double error_tolerance, double charge_carrier=PROTON,
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._evaluate_theoretical_distribution(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *)__pyx_v_self), __pyx_v_eid, __pyx_v_tid, __pyx_v_peak, __pyx_v_charge, 0)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 462, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_t_2);
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":451
 *             minimum_intensity, *args, **kwargs)
 * 
 *     cpdef IsotopicFitRecord fit_theoretical_distribution(self, FittedPeak peak, double error_tolerance, int charge,             # <<<<<<<<<<<<<<
 *                                                          double charge_carrier=PROTON, double truncate_after=0.95,
 *                                                          double ignore_below=0):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.AveragineDeconvoluterBase.fit_theoretical_distribution", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_eid);
  __Pyx_XDECREF((PyObject *)__pyx_v_tid);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_3fit_theoretical_distribution(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_2fit_theoretical_distribution[] = "AveragineDeconvoluterBase.fit_theoretical_distribution(self, FittedPeak peak, double error_tolerance, int charge, double charge_carrier=PROTON, double truncate_after=0.95, double ignore_below=0) -> IsotopicFitRecord";
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_3fit_theoretical_distribution(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_peak = 0;
  double __pyx_v_error_tolerance;
  int __pyx_v_charge;
  double __pyx_v_charge_carrier;
  double __pyx_v_truncate_after;
  double __pyx_v_ignore_below;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("fit_theoretical_distribution (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_peak,&__pyx_n_s_error_tolerance,&__pyx_n_s_charge,&__pyx_n_s_charge_carrier,&__pyx_n_s_truncate_after,&__pyx_n_s_ignore_below,0};
    PyObject* values[6] = {0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peak)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_error_tolerance)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("fit_theoretical_distribution", 0, 3, 6, 1); __PYX_ERR(0, 451, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_charge)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("fit_theoretical_distribution", 0, 3, 6, 2); __PYX_ERR(0, 451, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_charge_carrier);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_truncate_after);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ignore_below);
          if (value) { values[5] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "fit_theoretical_distribution") < 0)) __PYX_ERR(0, 451, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_peak = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)values[0]);
    __pyx_v_error_tolerance = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_error_tolerance == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 451, __pyx_L3_error)
    __pyx_v_charge = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_charge == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 451, __pyx_L3_error)
    if (values[3]) {
      __pyx_v_charge_carrier = __pyx_PyFloat_AsDouble(values[3]); if (unlikely((__pyx_v_charge_carrier == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 452, __pyx_L3_error)
    } else {
      __pyx_v_charge_carrier = __pyx_k__2;
    }
    if (values[4]) {
      __pyx_v_truncate_after = __pyx_PyFloat_AsDouble(values[4]); if (unlikely((__pyx_v_truncate_after == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 452, __pyx_L3_error)
    } else {
      __pyx_v_truncate_after = ((double)0.95);
    }
    if (values[5]) {
      __pyx_v_ignore_below = __pyx_PyFloat_AsDouble(values[5]); if (unlikely((__pyx_v_ignore_below == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 453, __pyx_L3_error)
    } else {
      __pyx_v_ignore_below = ((double)0.0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("fit_theoretical_distribution", 0, 3, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 451, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.AveragineDeconvoluterBase.fit_theoretical_distribution", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peak), __pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak, 1, "peak", 0))) __PYX_ERR(0, 451, __pyx_L1_error)
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_2fit_theoretical_distribution(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *)__pyx_v_self), __pyx_v_peak, __pyx_v_error_tolerance, __pyx_v_charge, __pyx_v_charge_carrier, __pyx_v_truncate_after, __pyx_v_ignore_below);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_2fit_theoretical_distribution(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_peak, double __pyx_v_error_tolerance, int __pyx_v_charge, double __pyx_v_charge_carrier, double __pyx_v_truncate_after, double __pyx_v_ignore_below) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_fit_theoretical_distribution __pyx_t_2;
  __Pyx_RefNannySetupContext("fit_theoretical_distribution", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 3;
  __pyx_t_2.charge_carrier = __pyx_v_charge_carrier;
  __pyx_t_2.truncate_after = __pyx_v_truncate_after;
  __pyx_t_2.ignore_below = __pyx_v_ignore_below;
  __pyx_t_1 = ((PyObject *)__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase->fit_theoretical_distribution(__pyx_v_self, __pyx_v_peak, __pyx_v_error_tolerance, __pyx_v_charge, 1, &__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.AveragineDeconvoluterBase.fit_theoretical_distribution", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/deconvoluter_base.pyx":464
 *         return self._evaluate_theoretical_distribution(eid, tid, peak, charge)
 * 
 *     cpdef set _fit_peaks_at_charges(self, set peak_charge_set, double error_tolerance, double charge_carrier=PROTON,             # <<<<<<<<<<<<<<
 *                                     double truncate_after=0.95, double ignore_below=0):
 *         cdef:
 */

static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_5_fit_peaks_at_charges(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase__fit_peaks_at_charges(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *__pyx_v_self, PyObject *__pyx_v_peak_charge_set, double __pyx_v_error_tolerance, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase__fit_peaks_at_charges *__pyx_optional_args) {
  double __pyx_v_charge_carrier = __pyx_k__3;
  double __pyx_v_truncate_after = ((double)0.95);
  double __pyx_v_ignore_below = ((double)0.0);
  PyObject *__pyx_v_results = 0;
  PyObject *__pyx_v_peak_charge = 0;
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_fit = 0;
  int __pyx_v_charge;
  PyObject *__pyx_v_obj = NULL;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_peak = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  int __pyx_t_13;
  PyObject *__pyx_t_14;
  long __pyx_t_15;
  int __pyx_t_16;
  struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_fit_theoretical_distribution __pyx_t_17;
  int __pyx_t_18;
  __Pyx_RefNannySetupContext("_fit_peaks_at_charges", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_charge_carrier = __pyx_optional_args->charge_carrier;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_truncate_after = __pyx_optional_args->truncate_after;
        if (__pyx_optional_args->__pyx_n > 2) {
          __pyx_v_ignore_below = __pyx_optional_args->ignore_below;
        }
      }
    }
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_fit_peaks_at_charges); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 464, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_5_fit_peaks_at_charges)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_error_tolerance); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 464, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_charge_carrier); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 464, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_5 = PyFloat_FromDouble(__pyx_v_truncate_after); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 464, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_6 = PyFloat_FromDouble(__pyx_v_ignore_below); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 464, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_8 = NULL;
        __pyx_t_9 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_8)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_8);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_9 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[6] = {__pyx_t_8, __pyx_v_peak_charge_set, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_9, 5+__pyx_t_9); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 464, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[6] = {__pyx_t_8, __pyx_v_peak_charge_set, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_9, 5+__pyx_t_9); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 464, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        } else
        #endif
        {
          __pyx_t_10 = PyTuple_New(5+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 464, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          if (__pyx_t_8) {
            __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_8); __pyx_t_8 = NULL;
          }
          __Pyx_INCREF(__pyx_v_peak_charge_set);
          __Pyx_GIVEREF(__pyx_v_peak_charge_set);
          PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_v_peak_charge_set);
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_t_3);
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_10, 2+__pyx_t_9, __pyx_t_4);
          __Pyx_GIVEREF(__pyx_t_5);
          PyTuple_SET_ITEM(__pyx_t_10, 3+__pyx_t_9, __pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_6);
          PyTuple_SET_ITEM(__pyx_t_10, 4+__pyx_t_9, __pyx_t_6);
          __pyx_t_3 = 0;
          __pyx_t_4 = 0;
          __pyx_t_5 = 0;
          __pyx_t_6 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_10, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 464, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (!(likely(PySet_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "set", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 464, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":474
 *             int charge
 *             list peak_charge_list
 *         results = set()             # <<<<<<<<<<<<<<
 *         for obj in peak_charge_set:
 *             peak_charge = <tuple>obj
 */
  __pyx_t_1 = PySet_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 474, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_results = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":475
 *             list peak_charge_list
 *         results = set()
 *         for obj in peak_charge_set:             # <<<<<<<<<<<<<<
 *             peak_charge = <tuple>obj
 *             peak = <FittedPeak>PyTuple_GET_ITEM(peak_charge, 0)
 */
  __pyx_t_11 = 0;
  __pyx_t_2 = __Pyx_set_iterator(__pyx_v_peak_charge_set, 1, (&__pyx_t_12), (&__pyx_t_9)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 475, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_1);
  __pyx_t_1 = __pyx_t_2;
  __pyx_t_2 = 0;
  while (1) {
    __pyx_t_13 = __Pyx_set_iter_next(__pyx_t_1, __pyx_t_12, &__pyx_t_11, &__pyx_t_2, __pyx_t_9);
    if (unlikely(__pyx_t_13 == 0)) break;
    if (unlikely(__pyx_t_13 == -1)) __PYX_ERR(0, 475, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_obj, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":476
 *         results = set()
 *         for obj in peak_charge_set:
 *             peak_charge = <tuple>obj             # <<<<<<<<<<<<<<
 *             peak = <FittedPeak>PyTuple_GET_ITEM(peak_charge, 0)
 *             charge = PyInt_AsLong(<object>PyTuple_GET_ITEM(peak_charge, 1))
 */
    __pyx_t_2 = __pyx_v_obj;
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_peak_charge, ((PyObject*)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":477
 *         for obj in peak_charge_set:
 *             peak_charge = <tuple>obj
 *             peak = <FittedPeak>PyTuple_GET_ITEM(peak_charge, 0)             # <<<<<<<<<<<<<<
 *             charge = PyInt_AsLong(<object>PyTuple_GET_ITEM(peak_charge, 1))
 * 
 */
    __pyx_t_14 = PyTuple_GET_ITEM(__pyx_v_peak_charge, 0);
    __pyx_t_2 = ((PyObject *)__pyx_t_14);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_peak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":478
 *             peak_charge = <tuple>obj
 *             peak = <FittedPeak>PyTuple_GET_ITEM(peak_charge, 0)
 *             charge = PyInt_AsLong(<object>PyTuple_GET_ITEM(peak_charge, 1))             # <<<<<<<<<<<<<<
 * 
 *             if peak.mz < 1:
 */
    __pyx_t_14 = PyTuple_GET_ITEM(__pyx_v_peak_charge, 1);
    __pyx_t_15 = PyInt_AsLong(((PyObject *)__pyx_t_14)); if (unlikely(__pyx_t_15 == ((long)-1L) && PyErr_Occurred())) __PYX_ERR(0, 478, __pyx_L1_error)
    __pyx_v_charge = __pyx_t_15;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":480
 *             charge = PyInt_AsLong(<object>PyTuple_GET_ITEM(peak_charge, 1))
 * 
 *             if peak.mz < 1:             # <<<<<<<<<<<<<<
 *                 continue
 * 
 */
    __pyx_t_16 = ((__pyx_v_peak->__pyx_base.mz < 1.0) != 0);
    if (__pyx_t_16) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":481
 * 
 *             if peak.mz < 1:
 *                 continue             # <<<<<<<<<<<<<<
 * 
 *             fit = self.fit_theoretical_distribution(
 */
      goto __pyx_L3_continue;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":480
 *             charge = PyInt_AsLong(<object>PyTuple_GET_ITEM(peak_charge, 1))
 * 
 *             if peak.mz < 1:             # <<<<<<<<<<<<<<
 *                 continue
 * 
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":483
 *                 continue
 * 
 *             fit = self.fit_theoretical_distribution(             # <<<<<<<<<<<<<<
 *                      peak, error_tolerance, charge,
 *                      charge_carrier, truncate_after=truncate_after,
 */
    __pyx_t_17.__pyx_n = 3;
    __pyx_t_17.charge_carrier = __pyx_v_charge_carrier;
    __pyx_t_17.truncate_after = __pyx_v_truncate_after;
    __pyx_t_17.ignore_below = __pyx_v_ignore_below;
    __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *)__pyx_v_self->__pyx_base.__pyx_vtab)->fit_theoretical_distribution(__pyx_v_self, __pyx_v_peak, __pyx_v_error_tolerance, __pyx_v_charge, 0, &__pyx_t_17)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 483, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_fit, ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":487
 *                      charge_carrier, truncate_after=truncate_after,
 *                      ignore_below=ignore_below)
 *             if not self._check_fit(fit):             # <<<<<<<<<<<<<<
 *                 continue
 *             results.add(fit)
 */
    __pyx_t_16 = ((!(((struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._check_fit(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *)__pyx_v_self), __pyx_v_fit, 0) != 0)) != 0);
    if (__pyx_t_16) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":488
 *                      ignore_below=ignore_below)
 *             if not self._check_fit(fit):
 *                 continue             # <<<<<<<<<<<<<<
 *             results.add(fit)
 *         return results
 */
      goto __pyx_L3_continue;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":487
 *                      charge_carrier, truncate_after=truncate_after,
 *                      ignore_below=ignore_below)
 *             if not self._check_fit(fit):             # <<<<<<<<<<<<<<
 *                 continue
 *             results.add(fit)
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":489
 *             if not self._check_fit(fit):
 *                 continue
 *             results.add(fit)             # <<<<<<<<<<<<<<
 *         return results
 * 
 */
    __pyx_t_18 = PySet_Add(__pyx_v_results, ((PyObject *)__pyx_v_fit)); if (unlikely(__pyx_t_18 == ((int)-1))) __PYX_ERR(0, 489, __pyx_L1_error)
    __pyx_L3_continue:;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":490
 *                 continue
 *             results.add(fit)
 *         return results             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_results);
  __pyx_r = __pyx_v_results;
  goto __pyx_L0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":464
 *         return self._evaluate_theoretical_distribution(eid, tid, peak, charge)
 * 
 *     cpdef set _fit_peaks_at_charges(self, set peak_charge_set, double error_tolerance, double charge_carrier=PROTON,             # <<<<<<<<<<<<<<
 *                                     double truncate_after=0.95, double ignore_below=0):
 *         cdef:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.AveragineDeconvoluterBase._fit_peaks_at_charges", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_results);
  __Pyx_XDECREF(__pyx_v_peak_charge);
  __Pyx_XDECREF((PyObject *)__pyx_v_fit);
  __Pyx_XDECREF(__pyx_v_obj);
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_5_fit_peaks_at_charges(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_4_fit_peaks_at_charges[] = "AveragineDeconvoluterBase._fit_peaks_at_charges(self, set peak_charge_set, double error_tolerance, double charge_carrier=PROTON, double truncate_after=0.95, double ignore_below=0) -> set";
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_5_fit_peaks_at_charges(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_peak_charge_set = 0;
  double __pyx_v_error_tolerance;
  double __pyx_v_charge_carrier;
  double __pyx_v_truncate_after;
  double __pyx_v_ignore_below;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_fit_peaks_at_charges (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_peak_charge_set,&__pyx_n_s_error_tolerance,&__pyx_n_s_charge_carrier,&__pyx_n_s_truncate_after,&__pyx_n_s_ignore_below,0};
    PyObject* values[5] = {0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peak_charge_set)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_error_tolerance)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_fit_peaks_at_charges", 0, 2, 5, 1); __PYX_ERR(0, 464, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_charge_carrier);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_truncate_after);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ignore_below);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_fit_peaks_at_charges") < 0)) __PYX_ERR(0, 464, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_peak_charge_set = ((PyObject*)values[0]);
    __pyx_v_error_tolerance = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_error_tolerance == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 464, __pyx_L3_error)
    if (values[2]) {
      __pyx_v_charge_carrier = __pyx_PyFloat_AsDouble(values[2]); if (unlikely((__pyx_v_charge_carrier == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 464, __pyx_L3_error)
    } else {
      __pyx_v_charge_carrier = __pyx_k__3;
    }
    if (values[3]) {
      __pyx_v_truncate_after = __pyx_PyFloat_AsDouble(values[3]); if (unlikely((__pyx_v_truncate_after == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 465, __pyx_L3_error)
    } else {
      __pyx_v_truncate_after = ((double)0.95);
    }
    if (values[4]) {
      __pyx_v_ignore_below = __pyx_PyFloat_AsDouble(values[4]); if (unlikely((__pyx_v_ignore_below == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 465, __pyx_L3_error)
    } else {
      __pyx_v_ignore_below = ((double)0.0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_fit_peaks_at_charges", 0, 2, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 464, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.AveragineDeconvoluterBase._fit_peaks_at_charges", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peak_charge_set), (&PySet_Type), 1, "peak_charge_set", 1))) __PYX_ERR(0, 464, __pyx_L1_error)
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_4_fit_peaks_at_charges(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *)__pyx_v_self), __pyx_v_peak_charge_set, __pyx_v_error_tolerance, __pyx_v_charge_carrier, __pyx_v_truncate_after, __pyx_v_ignore_below);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_4_fit_peaks_at_charges(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *__pyx_v_self, PyObject *__pyx_v_peak_charge_set, double __pyx_v_error_tolerance, double __pyx_v_charge_carrier, double __pyx_v_truncate_after, double __pyx_v_ignore_below) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase__fit_peaks_at_charges __pyx_t_2;
  __Pyx_RefNannySetupContext("_fit_peaks_at_charges", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 3;
  __pyx_t_2.charge_carrier = __pyx_v_charge_carrier;
  __pyx_t_2.truncate_after = __pyx_v_truncate_after;
  __pyx_t_2.ignore_below = __pyx_v_ignore_below;
  __pyx_t_1 = __pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase->_fit_peaks_at_charges(__pyx_v_self, __pyx_v_peak_charge_set, __pyx_v_error_tolerance, 1, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 464, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.AveragineDeconvoluterBase._fit_peaks_at_charges", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/deconvoluter_base.pxd":59
 * cdef class AveragineDeconvoluterBase(DeconvoluterBase):
 *     cdef:
 *         public AveragineCache averagine             # <<<<<<<<<<<<<<
 * 
 *     cpdef IsotopicFitRecord fit_theoretical_distribution(self, FittedPeak peak, double error_tolerance, int charge,
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_9averagine_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_9averagine_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_9averagine___get__(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_9averagine___get__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->averagine));
  __pyx_r = ((PyObject *)__pyx_v_self->averagine);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_9averagine_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_9averagine_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_9averagine_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_9averagine_2__set__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_12ms_deisotope_2_c_9averagine_AveragineCache))))) __PYX_ERR(2, 59, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->averagine);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->averagine));
  __pyx_v_self->averagine = ((struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.AveragineDeconvoluterBase.averagine.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_9averagine_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_9averagine_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_9averagine_4__del__(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_9averagine_4__del__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->averagine);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->averagine));
  __pyx_v_self->averagine = ((struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_7__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_6__reduce_cython__[] = "AveragineDeconvoluterBase.__reduce_cython__(self)";
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_7__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_6__reduce_cython__(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_6__reduce_cython__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self._slice_cache, self.averagine, self.merge_isobaric_peaks, self.minimum_intensity, self.peaklist, self.scale_method, self.scorer, self.use_subtraction, self.verbose)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.merge_isobaric_peaks); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->__pyx_base.minimum_intensity); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.use_subtraction); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.verbose); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(9); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_v_self->__pyx_base._slice_cache);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base._slice_cache);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_self->__pyx_base._slice_cache);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->averagine));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->averagine));
  PyTuple_SET_ITEM(__pyx_t_5, 1, ((PyObject *)__pyx_v_self->averagine));
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_2);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->__pyx_base.peaklist));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->__pyx_base.peaklist));
  PyTuple_SET_ITEM(__pyx_t_5, 4, ((PyObject *)__pyx_v_self->__pyx_base.peaklist));
  __Pyx_INCREF(__pyx_v_self->__pyx_base.scale_method);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.scale_method);
  PyTuple_SET_ITEM(__pyx_t_5, 5, __pyx_v_self->__pyx_base.scale_method);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->__pyx_base.scorer));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->__pyx_base.scorer));
  PyTuple_SET_ITEM(__pyx_t_5, 6, ((PyObject *)__pyx_v_self->__pyx_base.scorer));
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 7, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 8, __pyx_t_4);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self._slice_cache, self.averagine, self.merge_isobaric_peaks, self.minimum_intensity, self.peaklist, self.scale_method, self.scorer, self.use_subtraction, self.verbose)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_5 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_v__dict = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "(tree fragment)":7
 *     state = (self._slice_cache, self.averagine, self.merge_isobaric_peaks, self.minimum_intensity, self.peaklist, self.scale_method, self.scorer, self.use_subtraction, self.verbose)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_6 = (__pyx_v__dict != Py_None);
  __pyx_t_7 = (__pyx_t_6 != 0);
  if (__pyx_t_7) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v__dict);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self._slice_cache is not None or self.averagine is not None or self.peaklist is not None or self.scale_method is not None or self.scorer is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self._slice_cache, self.averagine, self.merge_isobaric_peaks, self.minimum_intensity, self.peaklist, self.scale_method, self.scorer, self.use_subtraction, self.verbose)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self._slice_cache is not None or self.averagine is not None or self.peaklist is not None or self.scale_method is not None or self.scorer is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_AveragineDeconvoluterBase, (type(self), 0x9da1e44, None), state
 */
  /*else*/ {
    __pyx_t_6 = (__pyx_v_self->__pyx_base._slice_cache != ((PyObject*)Py_None));
    __pyx_t_8 = (__pyx_t_6 != 0);
    if (!__pyx_t_8) {
    } else {
      __pyx_t_7 = __pyx_t_8;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_8 = (((PyObject *)__pyx_v_self->averagine) != Py_None);
    __pyx_t_6 = (__pyx_t_8 != 0);
    if (!__pyx_t_6) {
    } else {
      __pyx_t_7 = __pyx_t_6;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_6 = (((PyObject *)__pyx_v_self->__pyx_base.peaklist) != Py_None);
    __pyx_t_8 = (__pyx_t_6 != 0);
    if (!__pyx_t_8) {
    } else {
      __pyx_t_7 = __pyx_t_8;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_8 = (__pyx_v_self->__pyx_base.scale_method != ((PyObject*)Py_None));
    __pyx_t_6 = (__pyx_t_8 != 0);
    if (!__pyx_t_6) {
    } else {
      __pyx_t_7 = __pyx_t_6;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_6 = (((PyObject *)__pyx_v_self->__pyx_base.scorer) != Py_None);
    __pyx_t_8 = (__pyx_t_6 != 0);
    __pyx_t_7 = __pyx_t_8;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_7;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._slice_cache is not None or self.averagine is not None or self.peaklist is not None or self.scale_method is not None or self.scorer is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_AveragineDeconvoluterBase, (type(self), 0x9da1e44, None), state
 *     else:
 */
  __pyx_t_7 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_7) {

    /* "(tree fragment)":13
 *         use_setstate = self._slice_cache is not None or self.averagine is not None or self.peaklist is not None or self.scale_method is not None or self.scorer is not None
 *     if use_setstate:
 *         return __pyx_unpickle_AveragineDeconvoluterBase, (type(self), 0x9da1e44, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_AveragineDeconvoluterBase, (type(self), 0x9da1e44, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_AveragineDeconvol); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_165289540);
    __Pyx_GIVEREF(__pyx_int_165289540);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_int_165289540);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_5, 2, Py_None);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_5);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_state);
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._slice_cache is not None or self.averagine is not None or self.peaklist is not None or self.scale_method is not None or self.scorer is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_AveragineDeconvoluterBase, (type(self), 0x9da1e44, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_AveragineDeconvoluterBase, (type(self), 0x9da1e44, None), state
 *     else:
 *         return __pyx_unpickle_AveragineDeconvoluterBase, (type(self), 0x9da1e44, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_AveragineDeconvoluterBase__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_AveragineDeconvol); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_165289540);
    __Pyx_GIVEREF(__pyx_int_165289540);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_int_165289540);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_5);
    __pyx_t_3 = 0;
    __pyx_t_5 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.AveragineDeconvoluterBase.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_AveragineDeconvoluterBase, (type(self), 0x9da1e44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_AveragineDeconvoluterBase__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_9__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_8__setstate_cython__[] = "AveragineDeconvoluterBase.__setstate_cython__(self, __pyx_state)";
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_9__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_8__setstate_cython__(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_8__setstate_cython__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_AveragineDeconvoluterBase, (type(self), 0x9da1e44, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_AveragineDeconvoluterBase__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_17deconvoluter_base___pyx_unpickle_AveragineDeconvoluterBase__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_AveragineDeconvoluterBase, (type(self), 0x9da1e44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_AveragineDeconvoluterBase__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.AveragineDeconvoluterBase.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/deconvoluter_base.pyx":497
 *     for fitting isotopic patterns using multiple Averagine models.
 *     """
 *     def __init__(self, bint use_subtraction=False, str scale_method="sum", bint merge_isobaric_peaks=True,             # <<<<<<<<<<<<<<
 *                  double minimum_intensity=5., *args, **kwargs):
 *         super(MultiAveragineDeconvoluterBase, self).__init__(
 */

/* Python wrapper */
static int __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_v_use_subtraction;
  PyObject *__pyx_v_scale_method = 0;
  int __pyx_v_merge_isobaric_peaks;
  double __pyx_v_minimum_intensity;
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  __pyx_v_kwargs = PyDict_New(); if (unlikely(!__pyx_v_kwargs)) return -1;
  __Pyx_GOTREF(__pyx_v_kwargs);
  if (PyTuple_GET_SIZE(__pyx_args) > 4) {
    __pyx_v_args = PyTuple_GetSlice(__pyx_args, 4, PyTuple_GET_SIZE(__pyx_args));
    if (unlikely(!__pyx_v_args)) {
      __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
      __Pyx_RefNannyFinishContext();
      return -1;
    }
    __Pyx_GOTREF(__pyx_v_args);
  } else {
    __pyx_v_args = __pyx_empty_tuple; __Pyx_INCREF(__pyx_empty_tuple);
  }
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_use_subtraction,&__pyx_n_s_scale_method,&__pyx_n_s_merge_isobaric_peaks,&__pyx_n_s_minimum_intensity,0};
    PyObject* values[4] = {0,0,0,0};
    values[1] = ((PyObject*)__pyx_n_s_sum);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        default:
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_use_subtraction);
          if (value) { values[0] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_scale_method);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_merge_isobaric_peaks);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_minimum_intensity);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t used_pos_args = (pos_args < 4) ? pos_args : 4;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, __pyx_v_kwargs, values, used_pos_args, "__init__") < 0)) __PYX_ERR(0, 497, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        default:
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
      }
    }
    if (values[0]) {
      __pyx_v_use_subtraction = __Pyx_PyObject_IsTrue(values[0]); if (unlikely((__pyx_v_use_subtraction == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 497, __pyx_L3_error)
    } else {
      __pyx_v_use_subtraction = ((int)0);
    }
    __pyx_v_scale_method = ((PyObject*)values[1]);
    if (values[2]) {
      __pyx_v_merge_isobaric_peaks = __Pyx_PyObject_IsTrue(values[2]); if (unlikely((__pyx_v_merge_isobaric_peaks == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 497, __pyx_L3_error)
    } else {
      __pyx_v_merge_isobaric_peaks = ((int)1);
    }
    if (values[3]) {
      __pyx_v_minimum_intensity = __pyx_PyFloat_AsDouble(values[3]); if (unlikely((__pyx_v_minimum_intensity == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 498, __pyx_L3_error)
    } else {
      __pyx_v_minimum_intensity = ((double)5.);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_DECREF(__pyx_v_args); __pyx_v_args = 0;
  __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.MultiAveragineDeconvoluterBase.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_scale_method), (&PyString_Type), 1, "scale_method", 1))) __PYX_ERR(0, 497, __pyx_L1_error)
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase___init__(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *)__pyx_v_self), __pyx_v_use_subtraction, __pyx_v_scale_method, __pyx_v_merge_isobaric_peaks, __pyx_v_minimum_intensity, __pyx_v_args, __pyx_v_kwargs);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase___init__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *__pyx_v_self, int __pyx_v_use_subtraction, PyObject *__pyx_v_scale_method, int __pyx_v_merge_isobaric_peaks, double __pyx_v_minimum_intensity, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":499
 *     def __init__(self, bint use_subtraction=False, str scale_method="sum", bint merge_isobaric_peaks=True,
 *                  double minimum_intensity=5., *args, **kwargs):
 *         super(MultiAveragineDeconvoluterBase, self).__init__(             # <<<<<<<<<<<<<<
 *             use_subtraction, scale_method, merge_isobaric_peaks,
 *             minimum_intensity, *args, **kwargs)
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(((PyObject *)__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase));
  __Pyx_GIVEREF(((PyObject *)__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase));
  PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase));
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_self));
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":500
 *                  double minimum_intensity=5., *args, **kwargs):
 *         super(MultiAveragineDeconvoluterBase, self).__init__(
 *             use_subtraction, scale_method, merge_isobaric_peaks,             # <<<<<<<<<<<<<<
 *             minimum_intensity, *args, **kwargs)
 * 
 */
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_use_subtraction); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 500, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_merge_isobaric_peaks); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 500, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":501
 *         super(MultiAveragineDeconvoluterBase, self).__init__(
 *             use_subtraction, scale_method, merge_isobaric_peaks,
 *             minimum_intensity, *args, **kwargs)             # <<<<<<<<<<<<<<
 * 
 *     cpdef IsotopicFitRecord fit_theoretical_distribution(self, FittedPeak peak, double error_tolerance, int charge,
 */
  __pyx_t_4 = PyFloat_FromDouble(__pyx_v_minimum_intensity); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 501, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":499
 *     def __init__(self, bint use_subtraction=False, str scale_method="sum", bint merge_isobaric_peaks=True,
 *                  double minimum_intensity=5., *args, **kwargs):
 *         super(MultiAveragineDeconvoluterBase, self).__init__(             # <<<<<<<<<<<<<<
 *             use_subtraction, scale_method, merge_isobaric_peaks,
 *             minimum_intensity, *args, **kwargs)
 */
  __pyx_t_5 = PyTuple_New(4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_scale_method);
  __Pyx_GIVEREF(__pyx_v_scale_method);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_v_scale_method);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_4);
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":501
 *         super(MultiAveragineDeconvoluterBase, self).__init__(
 *             use_subtraction, scale_method, merge_isobaric_peaks,
 *             minimum_intensity, *args, **kwargs)             # <<<<<<<<<<<<<<
 * 
 *     cpdef IsotopicFitRecord fit_theoretical_distribution(self, FittedPeak peak, double error_tolerance, int charge,
 */
  __pyx_t_4 = PyNumber_Add(__pyx_t_5, __pyx_v_args); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":499
 *     def __init__(self, bint use_subtraction=False, str scale_method="sum", bint merge_isobaric_peaks=True,
 *                  double minimum_intensity=5., *args, **kwargs):
 *         super(MultiAveragineDeconvoluterBase, self).__init__(             # <<<<<<<<<<<<<<
 *             use_subtraction, scale_method, merge_isobaric_peaks,
 *             minimum_intensity, *args, **kwargs)
 */
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_4, __pyx_v_kwargs); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":497
 *     for fitting isotopic patterns using multiple Averagine models.
 *     """
 *     def __init__(self, bint use_subtraction=False, str scale_method="sum", bint merge_isobaric_peaks=True,             # <<<<<<<<<<<<<<
 *                  double minimum_intensity=5., *args, **kwargs):
 *         super(MultiAveragineDeconvoluterBase, self).__init__(
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.MultiAveragineDeconvoluterBase.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/deconvoluter_base.pyx":503
 *             minimum_intensity, *args, **kwargs)
 * 
 *     cpdef IsotopicFitRecord fit_theoretical_distribution(self, FittedPeak peak, double error_tolerance, int charge,             # <<<<<<<<<<<<<<
 *                                                          AveragineCache  averagine, double charge_carrier=PROTON,
 *                                                          double truncate_after=0.95, double ignore_below=0):
 */

static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_3fit_theoretical_distribution(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_fit_theoretical_distribution(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_peak, double __pyx_v_error_tolerance, int __pyx_v_charge, struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *__pyx_v_averagine, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_fit_theoretical_distribution *__pyx_optional_args) {
  double __pyx_v_charge_carrier = __pyx_k__4;
  double __pyx_v_truncate_after = ((double)0.95);
  double __pyx_v_ignore_below = ((double)0.0);
  PyObject *__pyx_v_eid = 0;
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *__pyx_v_tid = 0;
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_14AveragineCache_isotopic_cluster __pyx_t_12;
  struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_match_theoretical_isotopic_distribution __pyx_t_13;
  __Pyx_RefNannySetupContext("fit_theoretical_distribution", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_charge_carrier = __pyx_optional_args->charge_carrier;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_truncate_after = __pyx_optional_args->truncate_after;
        if (__pyx_optional_args->__pyx_n > 2) {
          __pyx_v_ignore_below = __pyx_optional_args->ignore_below;
        }
      }
    }
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_fit_theoretical_distribution); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 503, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_3fit_theoretical_distribution)) {
        __Pyx_XDECREF(((PyObject *)__pyx_r));
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_error_tolerance); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 503, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_charge); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 503, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_5 = PyFloat_FromDouble(__pyx_v_charge_carrier); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 503, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_6 = PyFloat_FromDouble(__pyx_v_truncate_after); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 503, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_7 = PyFloat_FromDouble(__pyx_v_ignore_below); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 503, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_8 = __pyx_t_1; __pyx_t_9 = NULL;
        __pyx_t_10 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
          __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
          if (likely(__pyx_t_9)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
            __Pyx_INCREF(__pyx_t_9);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_8, function);
            __pyx_t_10 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_8)) {
          PyObject *__pyx_temp[8] = {__pyx_t_9, ((PyObject *)__pyx_v_peak), __pyx_t_3, __pyx_t_4, ((PyObject *)__pyx_v_averagine), __pyx_t_5, __pyx_t_6, __pyx_t_7};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_10, 7+__pyx_t_10); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 503, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_8)) {
          PyObject *__pyx_temp[8] = {__pyx_t_9, ((PyObject *)__pyx_v_peak), __pyx_t_3, __pyx_t_4, ((PyObject *)__pyx_v_averagine), __pyx_t_5, __pyx_t_6, __pyx_t_7};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_10, 7+__pyx_t_10); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 503, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        } else
        #endif
        {
          __pyx_t_11 = PyTuple_New(7+__pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 503, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_11);
          if (__pyx_t_9) {
            __Pyx_GIVEREF(__pyx_t_9); PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_9); __pyx_t_9 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_peak));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_peak));
          PyTuple_SET_ITEM(__pyx_t_11, 0+__pyx_t_10, ((PyObject *)__pyx_v_peak));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_11, 1+__pyx_t_10, __pyx_t_3);
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_11, 2+__pyx_t_10, __pyx_t_4);
          __Pyx_INCREF(((PyObject *)__pyx_v_averagine));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_averagine));
          PyTuple_SET_ITEM(__pyx_t_11, 3+__pyx_t_10, ((PyObject *)__pyx_v_averagine));
          __Pyx_GIVEREF(__pyx_t_5);
          PyTuple_SET_ITEM(__pyx_t_11, 4+__pyx_t_10, __pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_6);
          PyTuple_SET_ITEM(__pyx_t_11, 5+__pyx_t_10, __pyx_t_6);
          __Pyx_GIVEREF(__pyx_t_7);
          PyTuple_SET_ITEM(__pyx_t_11, 6+__pyx_t_10, __pyx_t_7);
          __pyx_t_3 = 0;
          __pyx_t_4 = 0;
          __pyx_t_5 = 0;
          __pyx_t_6 = 0;
          __pyx_t_7 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_11, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 503, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        }
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitRecord))))) __PYX_ERR(0, 503, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":510
 *             TheoreticalIsotopicPattern tid
 *             double score
 *         tid = averagine.isotopic_cluster(             # <<<<<<<<<<<<<<
 *             peak.mz, charge, charge_carrier=charge_carrier,
 *             truncate_after=truncate_after, ignore_below=ignore_below)
 */
  __pyx_t_12.__pyx_n = 4;
  __pyx_t_12.charge = __pyx_v_charge;
  __pyx_t_12.charge_carrier = __pyx_v_charge_carrier;
  __pyx_t_12.truncate_after = __pyx_v_truncate_after;
  __pyx_t_12.ignore_below = __pyx_v_ignore_below;
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_AveragineCache *)__pyx_v_averagine->__pyx_vtab)->isotopic_cluster(__pyx_v_averagine, __pyx_v_peak->__pyx_base.mz, 0, &__pyx_t_12)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 510, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_tid = ((struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":513
 *             peak.mz, charge, charge_carrier=charge_carrier,
 *             truncate_after=truncate_after, ignore_below=ignore_below)
 *         eid = self.match_theoretical_isotopic_distribution(tid.peaklist, error_tolerance=error_tolerance)             # <<<<<<<<<<<<<<
 *         return self._evaluate_theoretical_distribution(eid, tid, peak, charge)
 * 
 */
  __pyx_t_1 = __pyx_v_tid->peaklist;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_13.__pyx_n = 1;
  __pyx_t_13.error_tolerance = __pyx_v_error_tolerance;
  __pyx_t_2 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.match_theoretical_isotopic_distribution(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *)__pyx_v_self), ((PyObject*)__pyx_t_1), 0, &__pyx_t_13); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 513, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_eid = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":514
 *             truncate_after=truncate_after, ignore_below=ignore_below)
 *         eid = self.match_theoretical_isotopic_distribution(tid.peaklist, error_tolerance=error_tolerance)
 *         return self._evaluate_theoretical_distribution(eid, tid, peak, charge)             # <<<<<<<<<<<<<<
 * 
 *     cpdef set _fit_peaks_at_charges(self, set peak_charge_set, double error_tolerance, double charge_carrier=PROTON,
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._evaluate_theoretical_distribution(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *)__pyx_v_self), __pyx_v_eid, __pyx_v_tid, __pyx_v_peak, __pyx_v_charge, 0)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 514, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_t_2);
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":503
 *             minimum_intensity, *args, **kwargs)
 * 
 *     cpdef IsotopicFitRecord fit_theoretical_distribution(self, FittedPeak peak, double error_tolerance, int charge,             # <<<<<<<<<<<<<<
 *                                                          AveragineCache  averagine, double charge_carrier=PROTON,
 *                                                          double truncate_after=0.95, double ignore_below=0):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.MultiAveragineDeconvoluterBase.fit_theoretical_distribution", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_eid);
  __Pyx_XDECREF((PyObject *)__pyx_v_tid);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_3fit_theoretical_distribution(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_2fit_theoretical_distribution[] = "MultiAveragineDeconvoluterBase.fit_theoretical_distribution(self, FittedPeak peak, double error_tolerance, int charge, AveragineCache averagine, double charge_carrier=PROTON, double truncate_after=0.95, double ignore_below=0) -> IsotopicFitRecord";
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_3fit_theoretical_distribution(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_peak = 0;
  double __pyx_v_error_tolerance;
  int __pyx_v_charge;
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *__pyx_v_averagine = 0;
  double __pyx_v_charge_carrier;
  double __pyx_v_truncate_after;
  double __pyx_v_ignore_below;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("fit_theoretical_distribution (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_peak,&__pyx_n_s_error_tolerance,&__pyx_n_s_charge,&__pyx_n_s_averagine,&__pyx_n_s_charge_carrier,&__pyx_n_s_truncate_after,&__pyx_n_s_ignore_below,0};
    PyObject* values[7] = {0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peak)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_error_tolerance)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("fit_theoretical_distribution", 0, 4, 7, 1); __PYX_ERR(0, 503, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_charge)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("fit_theoretical_distribution", 0, 4, 7, 2); __PYX_ERR(0, 503, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_averagine)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("fit_theoretical_distribution", 0, 4, 7, 3); __PYX_ERR(0, 503, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_charge_carrier);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_truncate_after);
          if (value) { values[5] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ignore_below);
          if (value) { values[6] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "fit_theoretical_distribution") < 0)) __PYX_ERR(0, 503, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_peak = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)values[0]);
    __pyx_v_error_tolerance = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_error_tolerance == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 503, __pyx_L3_error)
    __pyx_v_charge = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_charge == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 503, __pyx_L3_error)
    __pyx_v_averagine = ((struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *)values[3]);
    if (values[4]) {
      __pyx_v_charge_carrier = __pyx_PyFloat_AsDouble(values[4]); if (unlikely((__pyx_v_charge_carrier == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 504, __pyx_L3_error)
    } else {
      __pyx_v_charge_carrier = __pyx_k__4;
    }
    if (values[5]) {
      __pyx_v_truncate_after = __pyx_PyFloat_AsDouble(values[5]); if (unlikely((__pyx_v_truncate_after == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 505, __pyx_L3_error)
    } else {
      __pyx_v_truncate_after = ((double)0.95);
    }
    if (values[6]) {
      __pyx_v_ignore_below = __pyx_PyFloat_AsDouble(values[6]); if (unlikely((__pyx_v_ignore_below == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 505, __pyx_L3_error)
    } else {
      __pyx_v_ignore_below = ((double)0.0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("fit_theoretical_distribution", 0, 4, 7, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 503, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.MultiAveragineDeconvoluterBase.fit_theoretical_distribution", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peak), __pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak, 1, "peak", 0))) __PYX_ERR(0, 503, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_averagine), __pyx_ptype_12ms_deisotope_2_c_9averagine_AveragineCache, 1, "averagine", 0))) __PYX_ERR(0, 504, __pyx_L1_error)
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_2fit_theoretical_distribution(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *)__pyx_v_self), __pyx_v_peak, __pyx_v_error_tolerance, __pyx_v_charge, __pyx_v_averagine, __pyx_v_charge_carrier, __pyx_v_truncate_after, __pyx_v_ignore_below);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_2fit_theoretical_distribution(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_peak, double __pyx_v_error_tolerance, int __pyx_v_charge, struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *__pyx_v_averagine, double __pyx_v_charge_carrier, double __pyx_v_truncate_after, double __pyx_v_ignore_below) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_fit_theoretical_distribution __pyx_t_2;
  __Pyx_RefNannySetupContext("fit_theoretical_distribution", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 3;
  __pyx_t_2.charge_carrier = __pyx_v_charge_carrier;
  __pyx_t_2.truncate_after = __pyx_v_truncate_after;
  __pyx_t_2.ignore_below = __pyx_v_ignore_below;
  __pyx_t_1 = ((PyObject *)__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase->fit_theoretical_distribution(__pyx_v_self, __pyx_v_peak, __pyx_v_error_tolerance, __pyx_v_charge, __pyx_v_averagine, 1, &__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 503, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.MultiAveragineDeconvoluterBase.fit_theoretical_distribution", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/deconvoluter_base.pyx":516
 *         return self._evaluate_theoretical_distribution(eid, tid, peak, charge)
 * 
 *     cpdef set _fit_peaks_at_charges(self, set peak_charge_set, double error_tolerance, double charge_carrier=PROTON,             # <<<<<<<<<<<<<<
 *                                     double truncate_after=0.95, double ignore_below=0):
 *         cdef:
 */

static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_5_fit_peaks_at_charges(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase__fit_peaks_at_charges(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *__pyx_v_self, PyObject *__pyx_v_peak_charge_set, double __pyx_v_error_tolerance, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase__fit_peaks_at_charges *__pyx_optional_args) {
  double __pyx_v_charge_carrier = __pyx_k__5;
  double __pyx_v_truncate_after = ((double)0.95);
  double __pyx_v_ignore_below = ((double)0.0);
  PyObject *__pyx_v_results = 0;
  PyObject *__pyx_v_peak_charge = 0;
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_fit = 0;
  size_t __pyx_v_i;
  size_t __pyx_v_j;
  size_t __pyx_v_n_averagine;
  int __pyx_v_charge;
  PyObject *__pyx_v_peak_charge_list = 0;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_peak = NULL;
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *__pyx_v_averagine = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  size_t __pyx_t_13;
  PyObject *__pyx_t_14;
  long __pyx_t_15;
  int __pyx_t_16;
  size_t __pyx_t_17;
  size_t __pyx_t_18;
  size_t __pyx_t_19;
  struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_fit_theoretical_distribution __pyx_t_20;
  int __pyx_t_21;
  __Pyx_RefNannySetupContext("_fit_peaks_at_charges", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_charge_carrier = __pyx_optional_args->charge_carrier;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_truncate_after = __pyx_optional_args->truncate_after;
        if (__pyx_optional_args->__pyx_n > 2) {
          __pyx_v_ignore_below = __pyx_optional_args->ignore_below;
        }
      }
    }
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_fit_peaks_at_charges); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 516, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_5_fit_peaks_at_charges)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_error_tolerance); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 516, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_charge_carrier); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 516, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_5 = PyFloat_FromDouble(__pyx_v_truncate_after); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 516, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_6 = PyFloat_FromDouble(__pyx_v_ignore_below); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 516, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_8 = NULL;
        __pyx_t_9 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_8)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_8);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_9 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[6] = {__pyx_t_8, __pyx_v_peak_charge_set, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_9, 5+__pyx_t_9); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 516, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[6] = {__pyx_t_8, __pyx_v_peak_charge_set, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_9, 5+__pyx_t_9); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 516, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        } else
        #endif
        {
          __pyx_t_10 = PyTuple_New(5+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 516, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          if (__pyx_t_8) {
            __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_8); __pyx_t_8 = NULL;
          }
          __Pyx_INCREF(__pyx_v_peak_charge_set);
          __Pyx_GIVEREF(__pyx_v_peak_charge_set);
          PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_v_peak_charge_set);
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_t_3);
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_10, 2+__pyx_t_9, __pyx_t_4);
          __Pyx_GIVEREF(__pyx_t_5);
          PyTuple_SET_ITEM(__pyx_t_10, 3+__pyx_t_9, __pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_6);
          PyTuple_SET_ITEM(__pyx_t_10, 4+__pyx_t_9, __pyx_t_6);
          __pyx_t_3 = 0;
          __pyx_t_4 = 0;
          __pyx_t_5 = 0;
          __pyx_t_6 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_10, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 516, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (!(likely(PySet_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "set", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 516, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":525
 *             int charge
 *             list peak_charge_list
 *         results = []             # <<<<<<<<<<<<<<
 *         n_averagine = PyList_GET_SIZE(self.averagines)
 *         peak_charge_list = list(peak_charge_set)
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 525, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_results = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":526
 *             list peak_charge_list
 *         results = []
 *         n_averagine = PyList_GET_SIZE(self.averagines)             # <<<<<<<<<<<<<<
 *         peak_charge_list = list(peak_charge_set)
 *         for i in range(PyList_GET_SIZE(peak_charge_list)):
 */
  __pyx_t_1 = __pyx_v_self->averagines;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_n_averagine = PyList_GET_SIZE(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":527
 *         results = []
 *         n_averagine = PyList_GET_SIZE(self.averagines)
 *         peak_charge_list = list(peak_charge_set)             # <<<<<<<<<<<<<<
 *         for i in range(PyList_GET_SIZE(peak_charge_list)):
 *             peak_charge = <tuple>PyList_GET_ITEM(peak_charge_list, i)
 */
  __pyx_t_1 = PySequence_List(__pyx_v_peak_charge_set); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 527, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_peak_charge_list = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":528
 *         n_averagine = PyList_GET_SIZE(self.averagines)
 *         peak_charge_list = list(peak_charge_set)
 *         for i in range(PyList_GET_SIZE(peak_charge_list)):             # <<<<<<<<<<<<<<
 *             peak_charge = <tuple>PyList_GET_ITEM(peak_charge_list, i)
 *             peak = <FittedPeak>PyTuple_GET_ITEM(peak_charge, 0)
 */
  __pyx_t_11 = PyList_GET_SIZE(__pyx_v_peak_charge_list);
  __pyx_t_12 = __pyx_t_11;
  for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_12; __pyx_t_13+=1) {
    __pyx_v_i = __pyx_t_13;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":529
 *         peak_charge_list = list(peak_charge_set)
 *         for i in range(PyList_GET_SIZE(peak_charge_list)):
 *             peak_charge = <tuple>PyList_GET_ITEM(peak_charge_list, i)             # <<<<<<<<<<<<<<
 *             peak = <FittedPeak>PyTuple_GET_ITEM(peak_charge, 0)
 *             charge = PyInt_AsLong(<object>PyTuple_GET_ITEM(peak_charge, 1))
 */
    __pyx_t_14 = PyList_GET_ITEM(__pyx_v_peak_charge_list, __pyx_v_i);
    __pyx_t_1 = ((PyObject *)__pyx_t_14);
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_peak_charge, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":530
 *         for i in range(PyList_GET_SIZE(peak_charge_list)):
 *             peak_charge = <tuple>PyList_GET_ITEM(peak_charge_list, i)
 *             peak = <FittedPeak>PyTuple_GET_ITEM(peak_charge, 0)             # <<<<<<<<<<<<<<
 *             charge = PyInt_AsLong(<object>PyTuple_GET_ITEM(peak_charge, 1))
 * 
 */
    __pyx_t_14 = PyTuple_GET_ITEM(__pyx_v_peak_charge, 0);
    __pyx_t_1 = ((PyObject *)__pyx_t_14);
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_peak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":531
 *             peak_charge = <tuple>PyList_GET_ITEM(peak_charge_list, i)
 *             peak = <FittedPeak>PyTuple_GET_ITEM(peak_charge, 0)
 *             charge = PyInt_AsLong(<object>PyTuple_GET_ITEM(peak_charge, 1))             # <<<<<<<<<<<<<<
 * 
 *             if peak.mz < 1:
 */
    __pyx_t_14 = PyTuple_GET_ITEM(__pyx_v_peak_charge, 1);
    __pyx_t_15 = PyInt_AsLong(((PyObject *)__pyx_t_14)); if (unlikely(__pyx_t_15 == ((long)-1L) && PyErr_Occurred())) __PYX_ERR(0, 531, __pyx_L1_error)
    __pyx_v_charge = __pyx_t_15;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":533
 *             charge = PyInt_AsLong(<object>PyTuple_GET_ITEM(peak_charge, 1))
 * 
 *             if peak.mz < 1:             # <<<<<<<<<<<<<<
 *                 continue
 *             for j in range(n_averagine):
 */
    __pyx_t_16 = ((__pyx_v_peak->__pyx_base.mz < 1.0) != 0);
    if (__pyx_t_16) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":534
 * 
 *             if peak.mz < 1:
 *                 continue             # <<<<<<<<<<<<<<
 *             for j in range(n_averagine):
 *                 averagine = <AveragineCache>PyList_GET_ITEM(self.averagines, j)
 */
      goto __pyx_L3_continue;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":533
 *             charge = PyInt_AsLong(<object>PyTuple_GET_ITEM(peak_charge, 1))
 * 
 *             if peak.mz < 1:             # <<<<<<<<<<<<<<
 *                 continue
 *             for j in range(n_averagine):
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":535
 *             if peak.mz < 1:
 *                 continue
 *             for j in range(n_averagine):             # <<<<<<<<<<<<<<
 *                 averagine = <AveragineCache>PyList_GET_ITEM(self.averagines, j)
 *                 fit = self.fit_theoretical_distribution(
 */
    __pyx_t_17 = __pyx_v_n_averagine;
    __pyx_t_18 = __pyx_t_17;
    for (__pyx_t_19 = 0; __pyx_t_19 < __pyx_t_18; __pyx_t_19+=1) {
      __pyx_v_j = __pyx_t_19;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":536
 *                 continue
 *             for j in range(n_averagine):
 *                 averagine = <AveragineCache>PyList_GET_ITEM(self.averagines, j)             # <<<<<<<<<<<<<<
 *                 fit = self.fit_theoretical_distribution(
 *                     peak, error_tolerance, charge, averagine, charge_carrier,
 */
      __pyx_t_1 = __pyx_v_self->averagines;
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_14 = PyList_GET_ITEM(__pyx_t_1, __pyx_v_j);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = ((PyObject *)__pyx_t_14);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_averagine, ((struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":537
 *             for j in range(n_averagine):
 *                 averagine = <AveragineCache>PyList_GET_ITEM(self.averagines, j)
 *                 fit = self.fit_theoretical_distribution(             # <<<<<<<<<<<<<<
 *                     peak, error_tolerance, charge, averagine, charge_carrier,
 *                     truncate_after=truncate_after, ignore_below=ignore_below)
 */
      __pyx_t_20.__pyx_n = 3;
      __pyx_t_20.charge_carrier = __pyx_v_charge_carrier;
      __pyx_t_20.truncate_after = __pyx_v_truncate_after;
      __pyx_t_20.ignore_below = __pyx_v_ignore_below;
      __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *)__pyx_v_self->__pyx_base.__pyx_vtab)->fit_theoretical_distribution(__pyx_v_self, __pyx_v_peak, __pyx_v_error_tolerance, __pyx_v_charge, __pyx_v_averagine, 0, &__pyx_t_20)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 537, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_fit, ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":541
 *                     truncate_after=truncate_after, ignore_below=ignore_below)
 *                 # fit.missed_peaks = count_missed_peaks(fit.experimental)
 *                 fit.data = averagine             # <<<<<<<<<<<<<<
 *                 if not self._check_fit(fit):
 *                     continue
 */
      __Pyx_INCREF(((PyObject *)__pyx_v_averagine));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_averagine));
      __Pyx_GOTREF(__pyx_v_fit->data);
      __Pyx_DECREF(__pyx_v_fit->data);
      __pyx_v_fit->data = ((PyObject *)__pyx_v_averagine);

      /* "ms_deisotope/_c/deconvoluter_base.pyx":542
 *                 # fit.missed_peaks = count_missed_peaks(fit.experimental)
 *                 fit.data = averagine
 *                 if not self._check_fit(fit):             # <<<<<<<<<<<<<<
 *                     continue
 *                 # should we track the best fit for each hypothetical peak charge pair
 */
      __pyx_t_16 = ((!(((struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._check_fit(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *)__pyx_v_self), __pyx_v_fit, 0) != 0)) != 0);
      if (__pyx_t_16) {

        /* "ms_deisotope/_c/deconvoluter_base.pyx":543
 *                 fit.data = averagine
 *                 if not self._check_fit(fit):
 *                     continue             # <<<<<<<<<<<<<<
 *                 # should we track the best fit for each hypothetical peak charge pair
 *                 # and only add the best one to the result set? This would save time
 */
        goto __pyx_L6_continue;

        /* "ms_deisotope/_c/deconvoluter_base.pyx":542
 *                 # fit.missed_peaks = count_missed_peaks(fit.experimental)
 *                 fit.data = averagine
 *                 if not self._check_fit(fit):             # <<<<<<<<<<<<<<
 *                     continue
 *                 # should we track the best fit for each hypothetical peak charge pair
 */
      }

      /* "ms_deisotope/_c/deconvoluter_base.pyx":547
 *                 # and only add the best one to the result set? This would save time
 *                 # later.
 *                 results.append(fit)             # <<<<<<<<<<<<<<
 * 
 *         return set(results)
 */
      __pyx_t_21 = __Pyx_PyList_Append(__pyx_v_results, ((PyObject *)__pyx_v_fit)); if (unlikely(__pyx_t_21 == ((int)-1))) __PYX_ERR(0, 547, __pyx_L1_error)
      __pyx_L6_continue:;
    }
    __pyx_L3_continue:;
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":549
 *                 results.append(fit)
 * 
 *         return set(results)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PySet_New(__pyx_v_results); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 549, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":516
 *         return self._evaluate_theoretical_distribution(eid, tid, peak, charge)
 * 
 *     cpdef set _fit_peaks_at_charges(self, set peak_charge_set, double error_tolerance, double charge_carrier=PROTON,             # <<<<<<<<<<<<<<
 *                                     double truncate_after=0.95, double ignore_below=0):
 *         cdef:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.MultiAveragineDeconvoluterBase._fit_peaks_at_charges", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_results);
  __Pyx_XDECREF(__pyx_v_peak_charge);
  __Pyx_XDECREF((PyObject *)__pyx_v_fit);
  __Pyx_XDECREF(__pyx_v_peak_charge_list);
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);
  __Pyx_XDECREF((PyObject *)__pyx_v_averagine);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_5_fit_peaks_at_charges(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_4_fit_peaks_at_charges[] = "MultiAveragineDeconvoluterBase._fit_peaks_at_charges(self, set peak_charge_set, double error_tolerance, double charge_carrier=PROTON, double truncate_after=0.95, double ignore_below=0) -> set";
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_5_fit_peaks_at_charges(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_peak_charge_set = 0;
  double __pyx_v_error_tolerance;
  double __pyx_v_charge_carrier;
  double __pyx_v_truncate_after;
  double __pyx_v_ignore_below;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_fit_peaks_at_charges (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_peak_charge_set,&__pyx_n_s_error_tolerance,&__pyx_n_s_charge_carrier,&__pyx_n_s_truncate_after,&__pyx_n_s_ignore_below,0};
    PyObject* values[5] = {0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peak_charge_set)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_error_tolerance)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_fit_peaks_at_charges", 0, 2, 5, 1); __PYX_ERR(0, 516, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_charge_carrier);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_truncate_after);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ignore_below);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_fit_peaks_at_charges") < 0)) __PYX_ERR(0, 516, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_peak_charge_set = ((PyObject*)values[0]);
    __pyx_v_error_tolerance = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_error_tolerance == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 516, __pyx_L3_error)
    if (values[2]) {
      __pyx_v_charge_carrier = __pyx_PyFloat_AsDouble(values[2]); if (unlikely((__pyx_v_charge_carrier == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 516, __pyx_L3_error)
    } else {
      __pyx_v_charge_carrier = __pyx_k__5;
    }
    if (values[3]) {
      __pyx_v_truncate_after = __pyx_PyFloat_AsDouble(values[3]); if (unlikely((__pyx_v_truncate_after == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 517, __pyx_L3_error)
    } else {
      __pyx_v_truncate_after = ((double)0.95);
    }
    if (values[4]) {
      __pyx_v_ignore_below = __pyx_PyFloat_AsDouble(values[4]); if (unlikely((__pyx_v_ignore_below == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 517, __pyx_L3_error)
    } else {
      __pyx_v_ignore_below = ((double)0.0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_fit_peaks_at_charges", 0, 2, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 516, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.MultiAveragineDeconvoluterBase._fit_peaks_at_charges", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peak_charge_set), (&PySet_Type), 1, "peak_charge_set", 1))) __PYX_ERR(0, 516, __pyx_L1_error)
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_4_fit_peaks_at_charges(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *)__pyx_v_self), __pyx_v_peak_charge_set, __pyx_v_error_tolerance, __pyx_v_charge_carrier, __pyx_v_truncate_after, __pyx_v_ignore_below);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_4_fit_peaks_at_charges(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *__pyx_v_self, PyObject *__pyx_v_peak_charge_set, double __pyx_v_error_tolerance, double __pyx_v_charge_carrier, double __pyx_v_truncate_after, double __pyx_v_ignore_below) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase__fit_peaks_at_charges __pyx_t_2;
  __Pyx_RefNannySetupContext("_fit_peaks_at_charges", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 3;
  __pyx_t_2.charge_carrier = __pyx_v_charge_carrier;
  __pyx_t_2.truncate_after = __pyx_v_truncate_after;
  __pyx_t_2.ignore_below = __pyx_v_ignore_below;
  __pyx_t_1 = __pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase->_fit_peaks_at_charges(__pyx_v_self, __pyx_v_peak_charge_set, __pyx_v_error_tolerance, 1, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 516, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.MultiAveragineDeconvoluterBase._fit_peaks_at_charges", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/deconvoluter_base.pxd":71
 * cdef class MultiAveragineDeconvoluterBase(DeconvoluterBase):
 *     cdef:
 *         public list averagines             # <<<<<<<<<<<<<<
 * 
 *     cpdef IsotopicFitRecord fit_theoretical_distribution(self, FittedPeak peak, double error_tolerance, int charge,
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_10averagines_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_10averagines_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_10averagines___get__(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_10averagines___get__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->averagines);
  __pyx_r = __pyx_v_self->averagines;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_10averagines_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_10averagines_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_10averagines_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_10averagines_2__set__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(PyList_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(2, 71, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->averagines);
  __Pyx_DECREF(__pyx_v_self->averagines);
  __pyx_v_self->averagines = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.MultiAveragineDeconvoluterBase.averagines.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_10averagines_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_10averagines_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_10averagines_4__del__(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_10averagines_4__del__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->averagines);
  __Pyx_DECREF(__pyx_v_self->averagines);
  __pyx_v_self->averagines = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_7__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_6__reduce_cython__[] = "MultiAveragineDeconvoluterBase.__reduce_cython__(self)";
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_7__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_6__reduce_cython__(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_6__reduce_cython__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self._slice_cache, self.averagines, self.merge_isobaric_peaks, self.minimum_intensity, self.peaklist, self.scale_method, self.scorer, self.use_subtraction, self.verbose)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.merge_isobaric_peaks); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->__pyx_base.minimum_intensity); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.use_subtraction); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.verbose); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(9); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_v_self->__pyx_base._slice_cache);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base._slice_cache);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_self->__pyx_base._slice_cache);
  __Pyx_INCREF(__pyx_v_self->averagines);
  __Pyx_GIVEREF(__pyx_v_self->averagines);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_v_self->averagines);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_2);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->__pyx_base.peaklist));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->__pyx_base.peaklist));
  PyTuple_SET_ITEM(__pyx_t_5, 4, ((PyObject *)__pyx_v_self->__pyx_base.peaklist));
  __Pyx_INCREF(__pyx_v_self->__pyx_base.scale_method);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.scale_method);
  PyTuple_SET_ITEM(__pyx_t_5, 5, __pyx_v_self->__pyx_base.scale_method);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->__pyx_base.scorer));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->__pyx_base.scorer));
  PyTuple_SET_ITEM(__pyx_t_5, 6, ((PyObject *)__pyx_v_self->__pyx_base.scorer));
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 7, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 8, __pyx_t_4);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self._slice_cache, self.averagines, self.merge_isobaric_peaks, self.minimum_intensity, self.peaklist, self.scale_method, self.scorer, self.use_subtraction, self.verbose)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_5 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_v__dict = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "(tree fragment)":7
 *     state = (self._slice_cache, self.averagines, self.merge_isobaric_peaks, self.minimum_intensity, self.peaklist, self.scale_method, self.scorer, self.use_subtraction, self.verbose)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_6 = (__pyx_v__dict != Py_None);
  __pyx_t_7 = (__pyx_t_6 != 0);
  if (__pyx_t_7) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v__dict);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self._slice_cache is not None or self.averagines is not None or self.peaklist is not None or self.scale_method is not None or self.scorer is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self._slice_cache, self.averagines, self.merge_isobaric_peaks, self.minimum_intensity, self.peaklist, self.scale_method, self.scorer, self.use_subtraction, self.verbose)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self._slice_cache is not None or self.averagines is not None or self.peaklist is not None or self.scale_method is not None or self.scorer is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_MultiAveragineDeconvoluterBase, (type(self), 0xfdf46c0, None), state
 */
  /*else*/ {
    __pyx_t_6 = (__pyx_v_self->__pyx_base._slice_cache != ((PyObject*)Py_None));
    __pyx_t_8 = (__pyx_t_6 != 0);
    if (!__pyx_t_8) {
    } else {
      __pyx_t_7 = __pyx_t_8;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_8 = (__pyx_v_self->averagines != ((PyObject*)Py_None));
    __pyx_t_6 = (__pyx_t_8 != 0);
    if (!__pyx_t_6) {
    } else {
      __pyx_t_7 = __pyx_t_6;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_6 = (((PyObject *)__pyx_v_self->__pyx_base.peaklist) != Py_None);
    __pyx_t_8 = (__pyx_t_6 != 0);
    if (!__pyx_t_8) {
    } else {
      __pyx_t_7 = __pyx_t_8;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_8 = (__pyx_v_self->__pyx_base.scale_method != ((PyObject*)Py_None));
    __pyx_t_6 = (__pyx_t_8 != 0);
    if (!__pyx_t_6) {
    } else {
      __pyx_t_7 = __pyx_t_6;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_6 = (((PyObject *)__pyx_v_self->__pyx_base.scorer) != Py_None);
    __pyx_t_8 = (__pyx_t_6 != 0);
    __pyx_t_7 = __pyx_t_8;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_7;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._slice_cache is not None or self.averagines is not None or self.peaklist is not None or self.scale_method is not None or self.scorer is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_MultiAveragineDeconvoluterBase, (type(self), 0xfdf46c0, None), state
 *     else:
 */
  __pyx_t_7 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_7) {

    /* "(tree fragment)":13
 *         use_setstate = self._slice_cache is not None or self.averagines is not None or self.peaklist is not None or self.scale_method is not None or self.scorer is not None
 *     if use_setstate:
 *         return __pyx_unpickle_MultiAveragineDeconvoluterBase, (type(self), 0xfdf46c0, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_MultiAveragineDeconvoluterBase, (type(self), 0xfdf46c0, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_MultiAveragineDec); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_266290880);
    __Pyx_GIVEREF(__pyx_int_266290880);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_int_266290880);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_5, 2, Py_None);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_5);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_state);
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._slice_cache is not None or self.averagines is not None or self.peaklist is not None or self.scale_method is not None or self.scorer is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_MultiAveragineDeconvoluterBase, (type(self), 0xfdf46c0, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_MultiAveragineDeconvoluterBase, (type(self), 0xfdf46c0, None), state
 *     else:
 *         return __pyx_unpickle_MultiAveragineDeconvoluterBase, (type(self), 0xfdf46c0, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_MultiAveragineDeconvoluterBase__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_MultiAveragineDec); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_266290880);
    __Pyx_GIVEREF(__pyx_int_266290880);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_int_266290880);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_5);
    __pyx_t_3 = 0;
    __pyx_t_5 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.MultiAveragineDeconvoluterBase.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_MultiAveragineDeconvoluterBase, (type(self), 0xfdf46c0, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_MultiAveragineDeconvoluterBase__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_9__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_8__setstate_cython__[] = "MultiAveragineDeconvoluterBase.__setstate_cython__(self, __pyx_state)";
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_9__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_8__setstate_cython__(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_8__setstate_cython__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_MultiAveragineDeconvoluterBase, (type(self), 0xfdf46c0, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_MultiAveragineDeconvoluterBase__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_17deconvoluter_base___pyx_unpickle_MultiAveragineDeconvoluterBase__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_MultiAveragineDeconvoluterBase, (type(self), 0xfdf46c0, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_MultiAveragineDeconvoluterBase__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.MultiAveragineDeconvoluterBase.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/deconvoluter_base.pyx":552
 * 
 * 
 * cdef FittedPeak has_previous_peak_at_charge(DeconvoluterBase peak_collection, FittedPeak peak, int charge, int step, double error_tolerance):             # <<<<<<<<<<<<<<
 *     """Get the `step`th *preceding* peak from `peak` in a isotopic pattern at
 *     charge state `charge`, or return `None` if it is missing.
 */

static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_has_previous_peak_at_charge(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_peak_collection, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_peak, int __pyx_v_charge, int __pyx_v_step, double __pyx_v_error_tolerance) {
  double __pyx_v_prev;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_isotopic_shift __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("has_previous_peak_at_charge", 0);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":571
 *     FittedPeak
 *     """
 *     prev = peak.mz - isotopic_shift(charge) * step             # <<<<<<<<<<<<<<
 *     return peak_collection._has_peak(prev, error_tolerance)
 * 
 */
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.charge = __pyx_v_charge;
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_9averagine_isotopic_shift(0, &__pyx_t_2); 
  __pyx_v_prev = (__pyx_v_peak->__pyx_base.mz - (__pyx_t_1 * __pyx_v_step));

  /* "ms_deisotope/_c/deconvoluter_base.pyx":572
 *     """
 *     prev = peak.mz - isotopic_shift(charge) * step
 *     return peak_collection._has_peak(prev, error_tolerance)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __pyx_t_3 = ((PyObject *)__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__has_peak(__pyx_v_peak_collection, __pyx_v_prev, __pyx_v_error_tolerance)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 572, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":552
 * 
 * 
 * cdef FittedPeak has_previous_peak_at_charge(DeconvoluterBase peak_collection, FittedPeak peak, int charge, int step, double error_tolerance):             # <<<<<<<<<<<<<<
 *     """Get the `step`th *preceding* peak from `peak` in a isotopic pattern at
 *     charge state `charge`, or return `None` if it is missing.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.has_previous_peak_at_charge", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/deconvoluter_base.pyx":575
 * 
 * 
 * cdef FittedPeak has_successor_peak_at_charge(DeconvoluterBase peak_collection, FittedPeak peak, int charge, int step, double error_tolerance):             # <<<<<<<<<<<<<<
 *     """Get the `step`th *succeeding* peak from `peak` in a isotopic pattern at
 *     charge state `charge`, or return `None` if it is missing.
 */

static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_has_successor_peak_at_charge(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_peak_collection, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_peak, int __pyx_v_charge, int __pyx_v_step, double __pyx_v_error_tolerance) {
  double __pyx_v_nxt;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_isotopic_shift __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("has_successor_peak_at_charge", 0);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":594
 *     FittedPeak
 *     """
 *     nxt = peak.mz + isotopic_shift(charge) * step             # <<<<<<<<<<<<<<
 *     return peak_collection._has_peak(nxt, error_tolerance)
 * 
 */
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.charge = __pyx_v_charge;
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_9averagine_isotopic_shift(0, &__pyx_t_2); 
  __pyx_v_nxt = (__pyx_v_peak->__pyx_base.mz + (__pyx_t_1 * __pyx_v_step));

  /* "ms_deisotope/_c/deconvoluter_base.pyx":595
 *     """
 *     nxt = peak.mz + isotopic_shift(charge) * step
 *     return peak_collection._has_peak(nxt, error_tolerance)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __pyx_t_3 = ((PyObject *)__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__has_peak(__pyx_v_peak_collection, __pyx_v_nxt, __pyx_v_error_tolerance)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 595, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":575
 * 
 * 
 * cdef FittedPeak has_successor_peak_at_charge(DeconvoluterBase peak_collection, FittedPeak peak, int charge, int step, double error_tolerance):             # <<<<<<<<<<<<<<
 *     """Get the `step`th *succeeding* peak from `peak` in a isotopic pattern at
 *     charge state `charge`, or return `None` if it is missing.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.has_successor_peak_at_charge", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/deconvoluter_base.pyx":602
 * cdef class ChargeIterator(object):
 * 
 *     def __init__(self, int lo, int hi):             # <<<<<<<<<<<<<<
 *         self.set_bounds(lo, hi)
 *         self.make_sequence()
 */

/* Python wrapper */
static int __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_v_lo;
  int __pyx_v_hi;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_lo,&__pyx_n_s_hi,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lo)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_hi)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, 1); __PYX_ERR(0, 602, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 602, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_lo = __Pyx_PyInt_As_int(values[0]); if (unlikely((__pyx_v_lo == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 602, __pyx_L3_error)
    __pyx_v_hi = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_hi == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 602, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 602, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.ChargeIterator.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator___init__(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *)__pyx_v_self), __pyx_v_lo, __pyx_v_hi);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator___init__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_self, int __pyx_v_lo, int __pyx_v_hi) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":603
 * 
 *     def __init__(self, int lo, int hi):
 *         self.set_bounds(lo, hi)             # <<<<<<<<<<<<<<
 *         self.make_sequence()
 * 
 */
  __pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_set_bounds(__pyx_v_self, __pyx_v_lo, __pyx_v_hi);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":604
 *     def __init__(self, int lo, int hi):
 *         self.set_bounds(lo, hi)
 *         self.make_sequence()             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_make_sequence(__pyx_v_self, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 604, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":602
 * cdef class ChargeIterator(object):
 * 
 *     def __init__(self, int lo, int hi):             # <<<<<<<<<<<<<<
 *         self.set_bounds(lo, hi)
 *         self.make_sequence()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.ChargeIterator.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/deconvoluter_base.pyx":606
 *         self.make_sequence()
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         free(self.values)
 * 
 */

/* Python wrapper */
static void __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_2__dealloc__(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_2__dealloc__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":607
 * 
 *     def __dealloc__(self):
 *         free(self.values)             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  free(__pyx_v_self->values);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":606
 *         self.make_sequence()
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         free(self.values)
 * 
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "ms_deisotope/_c/deconvoluter_base.pyx":610
 * 
 *     @staticmethod
 *     cdef ChargeIterator _create(int lo, int hi):             # <<<<<<<<<<<<<<
 *         cdef:
 *             ChargeIterator inst
 */

static struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator__create(int __pyx_v_lo, int __pyx_v_hi) {
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_inst = 0;
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("_create", 0);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":613
 *         cdef:
 *             ChargeIterator inst
 *         inst = ChargeIterator.__new__(ChargeIterator)             # <<<<<<<<<<<<<<
 *         inst.set_bounds(lo, hi)
 *         inst.make_sequence()
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator(((PyTypeObject *)__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 613, __pyx_L1_error)
  __Pyx_GOTREF(((PyObject *)__pyx_t_1));
  __pyx_v_inst = ((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":614
 *             ChargeIterator inst
 *         inst = ChargeIterator.__new__(ChargeIterator)
 *         inst.set_bounds(lo, hi)             # <<<<<<<<<<<<<<
 *         inst.make_sequence()
 *         return inst
 */
  __pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_set_bounds(__pyx_v_inst, __pyx_v_lo, __pyx_v_hi);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":615
 *         inst = ChargeIterator.__new__(ChargeIterator)
 *         inst.set_bounds(lo, hi)
 *         inst.make_sequence()             # <<<<<<<<<<<<<<
 *         return inst
 * 
 */
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_make_sequence(__pyx_v_inst, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 615, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":616
 *         inst.set_bounds(lo, hi)
 *         inst.make_sequence()
 *         return inst             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_inst));
  __pyx_r = __pyx_v_inst;
  goto __pyx_L0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":610
 * 
 *     @staticmethod
 *     cdef ChargeIterator _create(int lo, int hi):             # <<<<<<<<<<<<<<
 *         cdef:
 *             ChargeIterator inst
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.ChargeIterator._create", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_inst);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/deconvoluter_base.pyx":619
 * 
 *     @staticmethod
 *     cdef ChargeIterator from_tuple(tuple charge_range):             # <<<<<<<<<<<<<<
 *         cdef:
 *             ChargeIterator inst
 */

static struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_from_tuple(PyObject *__pyx_v_charge_range) {
  int __pyx_v_a;
  int __pyx_v_b;
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1;
  long __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("from_tuple", 0);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":624
 *             int a, b
 * 
 *         a = PyInt_AsLong(<object>PyTuple_GET_ITEM(charge_range, 0))             # <<<<<<<<<<<<<<
 *         b = PyInt_AsLong(<object>PyTuple_GET_ITEM(charge_range, 1))
 *         return ChargeIterator._create(a, b)
 */
  __pyx_t_1 = PyTuple_GET_ITEM(__pyx_v_charge_range, 0);
  __pyx_t_2 = PyInt_AsLong(((PyObject *)__pyx_t_1)); if (unlikely(__pyx_t_2 == ((long)-1L) && PyErr_Occurred())) __PYX_ERR(0, 624, __pyx_L1_error)
  __pyx_v_a = __pyx_t_2;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":625
 * 
 *         a = PyInt_AsLong(<object>PyTuple_GET_ITEM(charge_range, 0))
 *         b = PyInt_AsLong(<object>PyTuple_GET_ITEM(charge_range, 1))             # <<<<<<<<<<<<<<
 *         return ChargeIterator._create(a, b)
 * 
 */
  __pyx_t_1 = PyTuple_GET_ITEM(__pyx_v_charge_range, 1);
  __pyx_t_2 = PyInt_AsLong(((PyObject *)__pyx_t_1)); if (unlikely(__pyx_t_2 == ((long)-1L) && PyErr_Occurred())) __PYX_ERR(0, 625, __pyx_L1_error)
  __pyx_v_b = __pyx_t_2;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":626
 *         a = PyInt_AsLong(<object>PyTuple_GET_ITEM(charge_range, 0))
 *         b = PyInt_AsLong(<object>PyTuple_GET_ITEM(charge_range, 1))
 *         return ChargeIterator._create(a, b)             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __pyx_t_3 = ((PyObject *)__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator__create(__pyx_v_a, __pyx_v_b)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 626, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = ((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":619
 * 
 *     @staticmethod
 *     cdef ChargeIterator from_tuple(tuple charge_range):             # <<<<<<<<<<<<<<
 *         cdef:
 *             ChargeIterator inst
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.ChargeIterator.from_tuple", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/deconvoluter_base.pyx":629
 * 
 *     @staticmethod
 *     cdef ChargeIterator from_quickcharge(tuple charge_range, PeakSet peaks, FittedPeak peak):             # <<<<<<<<<<<<<<
 *         cdef:
 *             ChargeIterator inst
 */

static struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_from_quickcharge(PyObject *__pyx_v_charge_range, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaks, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_peak) {
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_inst = 0;
  int __pyx_v_a;
  int __pyx_v_b;
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1;
  long __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("from_quickcharge", 0);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":633
 *             ChargeIterator inst
 *             int a, b
 *         a = PyInt_AsLong(<object>PyTuple_GET_ITEM(charge_range, 0))             # <<<<<<<<<<<<<<
 *         b = PyInt_AsLong(<object>PyTuple_GET_ITEM(charge_range, 1))
 *         inst = ChargeIterator.__new__(ChargeIterator)
 */
  __pyx_t_1 = PyTuple_GET_ITEM(__pyx_v_charge_range, 0);
  __pyx_t_2 = PyInt_AsLong(((PyObject *)__pyx_t_1)); if (unlikely(__pyx_t_2 == ((long)-1L) && PyErr_Occurred())) __PYX_ERR(0, 633, __pyx_L1_error)
  __pyx_v_a = __pyx_t_2;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":634
 *             int a, b
 *         a = PyInt_AsLong(<object>PyTuple_GET_ITEM(charge_range, 0))
 *         b = PyInt_AsLong(<object>PyTuple_GET_ITEM(charge_range, 1))             # <<<<<<<<<<<<<<
 *         inst = ChargeIterator.__new__(ChargeIterator)
 *         inst.set_bounds(a, b)
 */
  __pyx_t_1 = PyTuple_GET_ITEM(__pyx_v_charge_range, 1);
  __pyx_t_2 = PyInt_AsLong(((PyObject *)__pyx_t_1)); if (unlikely(__pyx_t_2 == ((long)-1L) && PyErr_Occurred())) __PYX_ERR(0, 634, __pyx_L1_error)
  __pyx_v_b = __pyx_t_2;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":635
 *         a = PyInt_AsLong(<object>PyTuple_GET_ITEM(charge_range, 0))
 *         b = PyInt_AsLong(<object>PyTuple_GET_ITEM(charge_range, 1))
 *         inst = ChargeIterator.__new__(ChargeIterator)             # <<<<<<<<<<<<<<
 *         inst.set_bounds(a, b)
 *         inst.sequence_from_quickcharge(peaks, peak)
 */
  __pyx_t_3 = ((PyObject *)__pyx_tp_new_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator(((PyTypeObject *)__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 635, __pyx_L1_error)
  __Pyx_GOTREF(((PyObject *)__pyx_t_3));
  __pyx_v_inst = ((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":636
 *         b = PyInt_AsLong(<object>PyTuple_GET_ITEM(charge_range, 1))
 *         inst = ChargeIterator.__new__(ChargeIterator)
 *         inst.set_bounds(a, b)             # <<<<<<<<<<<<<<
 *         inst.sequence_from_quickcharge(peaks, peak)
 *         return inst
 */
  __pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_set_bounds(__pyx_v_inst, __pyx_v_a, __pyx_v_b);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":637
 *         inst = ChargeIterator.__new__(ChargeIterator)
 *         inst.set_bounds(a, b)
 *         inst.sequence_from_quickcharge(peaks, peak)             # <<<<<<<<<<<<<<
 *         return inst
 * 
 */
  __pyx_t_3 = __pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_sequence_from_quickcharge(__pyx_v_inst, __pyx_v_peaks, __pyx_v_peak, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 637, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":638
 *         inst.set_bounds(a, b)
 *         inst.sequence_from_quickcharge(peaks, peak)
 *         return inst             # <<<<<<<<<<<<<<
 * 
 *     cpdef sequence_from_quickcharge(self, PeakSet peaks, FittedPeak peak):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_inst));
  __pyx_r = __pyx_v_inst;
  goto __pyx_L0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":629
 * 
 *     @staticmethod
 *     cdef ChargeIterator from_quickcharge(tuple charge_range, PeakSet peaks, FittedPeak peak):             # <<<<<<<<<<<<<<
 *         cdef:
 *             ChargeIterator inst
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.ChargeIterator.from_quickcharge", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_inst);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/deconvoluter_base.pyx":640
 *         return inst
 * 
 *     cpdef sequence_from_quickcharge(self, PeakSet peaks, FittedPeak peak):             # <<<<<<<<<<<<<<
 *         cdef:
 *             np.ndarray[int, ndim=1] charges
 */

static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_5sequence_from_quickcharge(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_sequence_from_quickcharge(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaks, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_peak, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyArrayObject *__pyx_v_charges = 0;
  size_t __pyx_v_i;
  size_t __pyx_v_n;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_charges;
  __Pyx_Buffer __pyx_pybuffer_charges;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  size_t __pyx_t_9;
  size_t __pyx_t_10;
  size_t __pyx_t_11;
  size_t __pyx_t_12;
  size_t __pyx_t_13;
  __Pyx_RefNannySetupContext("sequence_from_quickcharge", 0);
  __pyx_pybuffer_charges.pybuffer.buf = NULL;
  __pyx_pybuffer_charges.refcount = 0;
  __pyx_pybuffernd_charges.data = NULL;
  __pyx_pybuffernd_charges.rcbuffer = &__pyx_pybuffer_charges;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":644
 *             np.ndarray[int, ndim=1] charges
 *             size_t i, n
 *         self.release_sequence()             # <<<<<<<<<<<<<<
 *         self.index = 0
 *         charges = quick_charge(peaks, peak.peak_count, abs(self.lower), abs(self.upper))
 */
  __pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_release_sequence(__pyx_v_self);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":645
 *             size_t i, n
 *         self.release_sequence()
 *         self.index = 0             # <<<<<<<<<<<<<<
 *         charges = quick_charge(peaks, peak.peak_count, abs(self.lower), abs(self.upper))
 *         n = charges.shape[0]
 */
  __pyx_v_self->index = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":646
 *         self.release_sequence()
 *         self.index = 0
 *         charges = quick_charge(peaks, peak.peak_count, abs(self.lower), abs(self.upper))             # <<<<<<<<<<<<<<
 *         n = charges.shape[0]
 *         if n == 0:
 */
  __pyx_t_1 = abs(__pyx_v_self->lower); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 646, __pyx_L1_error)
  __pyx_t_2 = abs(__pyx_v_self->upper); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 646, __pyx_L1_error)
  __pyx_t_3 = ((PyObject *)__pyx_fuse_0_2__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_quick_charge(__pyx_v_peaks, __pyx_v_peak->peak_count, __pyx_t_1, __pyx_t_2, 0)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 646, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_charges.rcbuffer->pybuffer);
    __pyx_t_2 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_charges.rcbuffer->pybuffer, (PyObject*)((PyArrayObject *)__pyx_t_3), &__Pyx_TypeInfo_int, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_2 < 0)) {
      PyErr_Fetch(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_charges.rcbuffer->pybuffer, (PyObject*)__pyx_v_charges, &__Pyx_TypeInfo_int, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_4); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_6);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      }
      __pyx_t_4 = __pyx_t_5 = __pyx_t_6 = 0;
    }
    __pyx_pybuffernd_charges.diminfo[0].strides = __pyx_pybuffernd_charges.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_charges.diminfo[0].shape = __pyx_pybuffernd_charges.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 646, __pyx_L1_error)
  }
  __pyx_v_charges = ((PyArrayObject *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":647
 *         self.index = 0
 *         charges = quick_charge(peaks, peak.peak_count, abs(self.lower), abs(self.upper))
 *         n = charges.shape[0]             # <<<<<<<<<<<<<<
 *         if n == 0:
 *             self.size = 1
 */
  __pyx_v_n = (__pyx_v_charges->dimensions[0]);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":648
 *         charges = quick_charge(peaks, peak.peak_count, abs(self.lower), abs(self.upper))
 *         n = charges.shape[0]
 *         if n == 0:             # <<<<<<<<<<<<<<
 *             self.size = 1
 *             self.values = <int*>malloc(sizeof(int) * 1)
 */
  __pyx_t_7 = ((__pyx_v_n == 0) != 0);
  if (__pyx_t_7) {

    /* "ms_deisotope/_c/deconvoluter_base.pyx":649
 *         n = charges.shape[0]
 *         if n == 0:
 *             self.size = 1             # <<<<<<<<<<<<<<
 *             self.values = <int*>malloc(sizeof(int) * 1)
 *             self.values[0] = self.sign
 */
    __pyx_v_self->size = 1;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":650
 *         if n == 0:
 *             self.size = 1
 *             self.values = <int*>malloc(sizeof(int) * 1)             # <<<<<<<<<<<<<<
 *             self.values[0] = self.sign
 *         elif charges[0] != 1:
 */
    __pyx_v_self->values = ((int *)malloc(((sizeof(int)) * 1)));

    /* "ms_deisotope/_c/deconvoluter_base.pyx":651
 *             self.size = 1
 *             self.values = <int*>malloc(sizeof(int) * 1)
 *             self.values[0] = self.sign             # <<<<<<<<<<<<<<
 *         elif charges[0] != 1:
 *             n += 1
 */
    __pyx_t_2 = __pyx_v_self->sign;
    (__pyx_v_self->values[0]) = __pyx_t_2;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":648
 *         charges = quick_charge(peaks, peak.peak_count, abs(self.lower), abs(self.upper))
 *         n = charges.shape[0]
 *         if n == 0:             # <<<<<<<<<<<<<<
 *             self.size = 1
 *             self.values = <int*>malloc(sizeof(int) * 1)
 */
    goto __pyx_L3;
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":652
 *             self.values = <int*>malloc(sizeof(int) * 1)
 *             self.values[0] = self.sign
 *         elif charges[0] != 1:             # <<<<<<<<<<<<<<
 *             n += 1
 *             self.size = n
 */
  __pyx_t_8 = 0;
  __pyx_t_2 = -1;
  if (__pyx_t_8 < 0) {
    __pyx_t_8 += __pyx_pybuffernd_charges.diminfo[0].shape;
    if (unlikely(__pyx_t_8 < 0)) __pyx_t_2 = 0;
  } else if (unlikely(__pyx_t_8 >= __pyx_pybuffernd_charges.diminfo[0].shape)) __pyx_t_2 = 0;
  if (unlikely(__pyx_t_2 != -1)) {
    __Pyx_RaiseBufferIndexError(__pyx_t_2);
    __PYX_ERR(0, 652, __pyx_L1_error)
  }
  __pyx_t_7 = (((*__Pyx_BufPtrStrided1d(int *, __pyx_pybuffernd_charges.rcbuffer->pybuffer.buf, __pyx_t_8, __pyx_pybuffernd_charges.diminfo[0].strides)) != 1) != 0);
  if (__pyx_t_7) {

    /* "ms_deisotope/_c/deconvoluter_base.pyx":653
 *             self.values[0] = self.sign
 *         elif charges[0] != 1:
 *             n += 1             # <<<<<<<<<<<<<<
 *             self.size = n
 *             self.values = <int*>malloc(sizeof(int) * n)
 */
    __pyx_v_n = (__pyx_v_n + 1);

    /* "ms_deisotope/_c/deconvoluter_base.pyx":654
 *         elif charges[0] != 1:
 *             n += 1
 *             self.size = n             # <<<<<<<<<<<<<<
 *             self.values = <int*>malloc(sizeof(int) * n)
 *             self.values[0] = self.sign
 */
    __pyx_v_self->size = __pyx_v_n;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":655
 *             n += 1
 *             self.size = n
 *             self.values = <int*>malloc(sizeof(int) * n)             # <<<<<<<<<<<<<<
 *             self.values[0] = self.sign
 *             for i in range(1, n):
 */
    __pyx_v_self->values = ((int *)malloc(((sizeof(int)) * __pyx_v_n)));

    /* "ms_deisotope/_c/deconvoluter_base.pyx":656
 *             self.size = n
 *             self.values = <int*>malloc(sizeof(int) * n)
 *             self.values[0] = self.sign             # <<<<<<<<<<<<<<
 *             for i in range(1, n):
 *                 self.values[i] = self.sign * charges[i - 1]
 */
    __pyx_t_2 = __pyx_v_self->sign;
    (__pyx_v_self->values[0]) = __pyx_t_2;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":657
 *             self.values = <int*>malloc(sizeof(int) * n)
 *             self.values[0] = self.sign
 *             for i in range(1, n):             # <<<<<<<<<<<<<<
 *                 self.values[i] = self.sign * charges[i - 1]
 *         else:
 */
    __pyx_t_9 = __pyx_v_n;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 1; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":658
 *             self.values[0] = self.sign
 *             for i in range(1, n):
 *                 self.values[i] = self.sign * charges[i - 1]             # <<<<<<<<<<<<<<
 *         else:
 *             self.size = n
 */
      __pyx_t_12 = (__pyx_v_i - 1);
      __pyx_t_2 = -1;
      if (unlikely(__pyx_t_12 >= (size_t)__pyx_pybuffernd_charges.diminfo[0].shape)) __pyx_t_2 = 0;
      if (unlikely(__pyx_t_2 != -1)) {
        __Pyx_RaiseBufferIndexError(__pyx_t_2);
        __PYX_ERR(0, 658, __pyx_L1_error)
      }
      (__pyx_v_self->values[__pyx_v_i]) = (__pyx_v_self->sign * (*__Pyx_BufPtrStrided1d(int *, __pyx_pybuffernd_charges.rcbuffer->pybuffer.buf, __pyx_t_12, __pyx_pybuffernd_charges.diminfo[0].strides)));
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":652
 *             self.values = <int*>malloc(sizeof(int) * 1)
 *             self.values[0] = self.sign
 *         elif charges[0] != 1:             # <<<<<<<<<<<<<<
 *             n += 1
 *             self.size = n
 */
    goto __pyx_L3;
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":660
 *                 self.values[i] = self.sign * charges[i - 1]
 *         else:
 *             self.size = n             # <<<<<<<<<<<<<<
 *             self.values = <int*>malloc(sizeof(int) * n)
 *             for i in range(n):
 */
  /*else*/ {
    __pyx_v_self->size = __pyx_v_n;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":661
 *         else:
 *             self.size = n
 *             self.values = <int*>malloc(sizeof(int) * n)             # <<<<<<<<<<<<<<
 *             for i in range(n):
 *                 self.values[i] = self.sign * charges[i]
 */
    __pyx_v_self->values = ((int *)malloc(((sizeof(int)) * __pyx_v_n)));

    /* "ms_deisotope/_c/deconvoluter_base.pyx":662
 *             self.size = n
 *             self.values = <int*>malloc(sizeof(int) * n)
 *             for i in range(n):             # <<<<<<<<<<<<<<
 *                 self.values[i] = self.sign * charges[i]
 * 
 */
    __pyx_t_9 = __pyx_v_n;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":663
 *             self.values = <int*>malloc(sizeof(int) * n)
 *             for i in range(n):
 *                 self.values[i] = self.sign * charges[i]             # <<<<<<<<<<<<<<
 * 
 *     cdef void set_bounds(self, int lo, int hi):
 */
      __pyx_t_13 = __pyx_v_i;
      __pyx_t_2 = -1;
      if (unlikely(__pyx_t_13 >= (size_t)__pyx_pybuffernd_charges.diminfo[0].shape)) __pyx_t_2 = 0;
      if (unlikely(__pyx_t_2 != -1)) {
        __Pyx_RaiseBufferIndexError(__pyx_t_2);
        __PYX_ERR(0, 663, __pyx_L1_error)
      }
      (__pyx_v_self->values[__pyx_v_i]) = (__pyx_v_self->sign * (*__Pyx_BufPtrStrided1d(int *, __pyx_pybuffernd_charges.rcbuffer->pybuffer.buf, __pyx_t_13, __pyx_pybuffernd_charges.diminfo[0].strides)));
    }
  }
  __pyx_L3:;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":640
 *         return inst
 * 
 *     cpdef sequence_from_quickcharge(self, PeakSet peaks, FittedPeak peak):             # <<<<<<<<<<<<<<
 *         cdef:
 *             np.ndarray[int, ndim=1] charges
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_charges.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.ChargeIterator.sequence_from_quickcharge", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_charges.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_charges);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_5sequence_from_quickcharge(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_4sequence_from_quickcharge[] = "ChargeIterator.sequence_from_quickcharge(self, PeakSet peaks, FittedPeak peak)";
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_5sequence_from_quickcharge(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaks = 0;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_peak = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sequence_from_quickcharge (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_peaks,&__pyx_n_s_peak,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peaks)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peak)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sequence_from_quickcharge", 1, 2, 2, 1); __PYX_ERR(0, 640, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "sequence_from_quickcharge") < 0)) __PYX_ERR(0, 640, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_peaks = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)values[0]);
    __pyx_v_peak = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)values[1]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("sequence_from_quickcharge", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 640, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.ChargeIterator.sequence_from_quickcharge", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peaks), __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet, 1, "peaks", 0))) __PYX_ERR(0, 640, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peak), __pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak, 1, "peak", 0))) __PYX_ERR(0, 640, __pyx_L1_error)
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_4sequence_from_quickcharge(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *)__pyx_v_self), __pyx_v_peaks, __pyx_v_peak);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_4sequence_from_quickcharge(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaks, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_peak) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("sequence_from_quickcharge", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_sequence_from_quickcharge(__pyx_v_self, __pyx_v_peaks, __pyx_v_peak, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 640, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.ChargeIterator.sequence_from_quickcharge", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/deconvoluter_base.pyx":665
 *                 self.values[i] = self.sign * charges[i]
 * 
 *     cdef void set_bounds(self, int lo, int hi):             # <<<<<<<<<<<<<<
 *         cdef:
 *             int abs_lo, abs_hi
 */

static void __pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_set_bounds(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_self, int __pyx_v_lo, int __pyx_v_hi) {
  int __pyx_v_abs_lo;
  int __pyx_v_abs_hi;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  __Pyx_RefNannySetupContext("set_bounds", 0);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":668
 *         cdef:
 *             int abs_lo, abs_hi
 *         self.sign = -1 if lo < 0 else 1             # <<<<<<<<<<<<<<
 *         abs_lo, abs_hi = abs(lo), abs(hi)
 *         if abs_lo < abs_hi:
 */
  if (((__pyx_v_lo < 0) != 0)) {
    __pyx_t_1 = -1;
  } else {
    __pyx_t_1 = 1;
  }
  __pyx_v_self->sign = __pyx_t_1;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":669
 *             int abs_lo, abs_hi
 *         self.sign = -1 if lo < 0 else 1
 *         abs_lo, abs_hi = abs(lo), abs(hi)             # <<<<<<<<<<<<<<
 *         if abs_lo < abs_hi:
 *             self.lower = abs_lo
 */
  __pyx_t_1 = abs(__pyx_v_lo); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 669, __pyx_L1_error)
  __pyx_t_2 = abs(__pyx_v_hi); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 669, __pyx_L1_error)
  __pyx_v_abs_lo = __pyx_t_1;
  __pyx_v_abs_hi = __pyx_t_2;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":670
 *         self.sign = -1 if lo < 0 else 1
 *         abs_lo, abs_hi = abs(lo), abs(hi)
 *         if abs_lo < abs_hi:             # <<<<<<<<<<<<<<
 *             self.lower = abs_lo
 *             self.upper = abs_hi
 */
  __pyx_t_3 = ((__pyx_v_abs_lo < __pyx_v_abs_hi) != 0);
  if (__pyx_t_3) {

    /* "ms_deisotope/_c/deconvoluter_base.pyx":671
 *         abs_lo, abs_hi = abs(lo), abs(hi)
 *         if abs_lo < abs_hi:
 *             self.lower = abs_lo             # <<<<<<<<<<<<<<
 *             self.upper = abs_hi
 *         else:
 */
    __pyx_v_self->lower = __pyx_v_abs_lo;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":672
 *         if abs_lo < abs_hi:
 *             self.lower = abs_lo
 *             self.upper = abs_hi             # <<<<<<<<<<<<<<
 *         else:
 *             self.lower = abs_hi
 */
    __pyx_v_self->upper = __pyx_v_abs_hi;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":670
 *         self.sign = -1 if lo < 0 else 1
 *         abs_lo, abs_hi = abs(lo), abs(hi)
 *         if abs_lo < abs_hi:             # <<<<<<<<<<<<<<
 *             self.lower = abs_lo
 *             self.upper = abs_hi
 */
    goto __pyx_L3;
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":674
 *             self.upper = abs_hi
 *         else:
 *             self.lower = abs_hi             # <<<<<<<<<<<<<<
 *             self.upper = abs_lo
 *         self.size = self.upper - self.lower + 1
 */
  /*else*/ {
    __pyx_v_self->lower = __pyx_v_abs_hi;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":675
 *         else:
 *             self.lower = abs_hi
 *             self.upper = abs_lo             # <<<<<<<<<<<<<<
 *         self.size = self.upper - self.lower + 1
 * 
 */
    __pyx_v_self->upper = __pyx_v_abs_lo;
  }
  __pyx_L3:;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":676
 *             self.lower = abs_hi
 *             self.upper = abs_lo
 *         self.size = self.upper - self.lower + 1             # <<<<<<<<<<<<<<
 * 
 *     cpdef make_sequence(self):
 */
  __pyx_v_self->size = ((__pyx_v_self->upper - __pyx_v_self->lower) + 1);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":665
 *                 self.values[i] = self.sign * charges[i]
 * 
 *     cdef void set_bounds(self, int lo, int hi):             # <<<<<<<<<<<<<<
 *         cdef:
 *             int abs_lo, abs_hi
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("ms_deisotope._c.deconvoluter_base.ChargeIterator.set_bounds", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "ms_deisotope/_c/deconvoluter_base.pyx":678
 *         self.size = self.upper - self.lower + 1
 * 
 *     cpdef make_sequence(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             int v
 */

static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_7make_sequence(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_make_sequence(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_self, CYTHON_UNUSED int __pyx_skip_dispatch) {
  size_t __pyx_v_i;
  size_t __pyx_v_n;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  long __pyx_t_1;
  int __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  size_t __pyx_t_5;
  __Pyx_RefNannySetupContext("make_sequence", 0);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":682
 *             int v
 *             size_t i, n
 *         self.release_sequence()             # <<<<<<<<<<<<<<
 *         n = self.size = self.upper - self.lower + 1
 *         self.index = 0
 */
  __pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_release_sequence(__pyx_v_self);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":683
 *             size_t i, n
 *         self.release_sequence()
 *         n = self.size = self.upper - self.lower + 1             # <<<<<<<<<<<<<<
 *         self.index = 0
 *         if n == 0:
 */
  __pyx_t_1 = ((__pyx_v_self->upper - __pyx_v_self->lower) + 1);
  __pyx_v_n = __pyx_t_1;
  __pyx_v_self->size = __pyx_t_1;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":684
 *         self.release_sequence()
 *         n = self.size = self.upper - self.lower + 1
 *         self.index = 0             # <<<<<<<<<<<<<<
 *         if n == 0:
 *             return
 */
  __pyx_v_self->index = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":685
 *         n = self.size = self.upper - self.lower + 1
 *         self.index = 0
 *         if n == 0:             # <<<<<<<<<<<<<<
 *             return
 *         self.values = <int*>malloc(sizeof(int) * n)
 */
  __pyx_t_2 = ((__pyx_v_n == 0) != 0);
  if (__pyx_t_2) {

    /* "ms_deisotope/_c/deconvoluter_base.pyx":686
 *         self.index = 0
 *         if n == 0:
 *             return             # <<<<<<<<<<<<<<
 *         self.values = <int*>malloc(sizeof(int) * n)
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":685
 *         n = self.size = self.upper - self.lower + 1
 *         self.index = 0
 *         if n == 0:             # <<<<<<<<<<<<<<
 *             return
 *         self.values = <int*>malloc(sizeof(int) * n)
 */
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":687
 *         if n == 0:
 *             return
 *         self.values = <int*>malloc(sizeof(int) * n)             # <<<<<<<<<<<<<<
 * 
 *         for i in range(n):
 */
  __pyx_v_self->values = ((int *)malloc(((sizeof(int)) * __pyx_v_n)));

  /* "ms_deisotope/_c/deconvoluter_base.pyx":689
 *         self.values = <int*>malloc(sizeof(int) * n)
 * 
 *         for i in range(n):             # <<<<<<<<<<<<<<
 *             self.values[i] = (self.upper - i) * self.sign
 * 
 */
  __pyx_t_3 = __pyx_v_n;
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
    __pyx_v_i = __pyx_t_5;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":690
 * 
 *         for i in range(n):
 *             self.values[i] = (self.upper - i) * self.sign             # <<<<<<<<<<<<<<
 * 
 *     cdef void release_sequence(self):
 */
    (__pyx_v_self->values[__pyx_v_i]) = ((__pyx_v_self->upper - __pyx_v_i) * __pyx_v_self->sign);
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":678
 *         self.size = self.upper - self.lower + 1
 * 
 *     cpdef make_sequence(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             int v
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_7make_sequence(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_6make_sequence[] = "ChargeIterator.make_sequence(self)";
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_7make_sequence(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("make_sequence (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_6make_sequence(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_6make_sequence(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("make_sequence", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_make_sequence(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 678, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.ChargeIterator.make_sequence", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/deconvoluter_base.pyx":692
 *             self.values[i] = (self.upper - i) * self.sign
 * 
 *     cdef void release_sequence(self):             # <<<<<<<<<<<<<<
 *         if self.values != NULL:
 *             free(self.values)
 */

static void __pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_release_sequence(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("release_sequence", 0);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":693
 * 
 *     cdef void release_sequence(self):
 *         if self.values != NULL:             # <<<<<<<<<<<<<<
 *             free(self.values)
 *             self.values = NULL
 */
  __pyx_t_1 = ((__pyx_v_self->values != NULL) != 0);
  if (__pyx_t_1) {

    /* "ms_deisotope/_c/deconvoluter_base.pyx":694
 *     cdef void release_sequence(self):
 *         if self.values != NULL:
 *             free(self.values)             # <<<<<<<<<<<<<<
 *             self.values = NULL
 *         self.size = 0
 */
    free(__pyx_v_self->values);

    /* "ms_deisotope/_c/deconvoluter_base.pyx":695
 *         if self.values != NULL:
 *             free(self.values)
 *             self.values = NULL             # <<<<<<<<<<<<<<
 *         self.size = 0
 *         self.reset()
 */
    __pyx_v_self->values = NULL;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":693
 * 
 *     cdef void release_sequence(self):
 *         if self.values != NULL:             # <<<<<<<<<<<<<<
 *             free(self.values)
 *             self.values = NULL
 */
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":696
 *             free(self.values)
 *             self.values = NULL
 *         self.size = 0             # <<<<<<<<<<<<<<
 *         self.reset()
 * 
 */
  __pyx_v_self->size = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":697
 *             self.values = NULL
 *         self.size = 0
 *         self.reset()             # <<<<<<<<<<<<<<
 * 
 *     cpdef reset(self):
 */
  __pyx_t_2 = __pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_reset(__pyx_v_self, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 697, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":692
 *             self.values[i] = (self.upper - i) * self.sign
 * 
 *     cdef void release_sequence(self):             # <<<<<<<<<<<<<<
 *         if self.values != NULL:
 *             free(self.values)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_WriteUnraisable("ms_deisotope._c.deconvoluter_base.ChargeIterator.release_sequence", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "ms_deisotope/_c/deconvoluter_base.pyx":699
 *         self.reset()
 * 
 *     cpdef reset(self):             # <<<<<<<<<<<<<<
 *         self.index = 0
 * 
 */

static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_9reset(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_reset(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_self, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("reset", 0);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":700
 * 
 *     cpdef reset(self):
 *         self.index = 0             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint has_more(self):
 */
  __pyx_v_self->index = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":699
 *         self.reset()
 * 
 *     cpdef reset(self):             # <<<<<<<<<<<<<<
 *         self.index = 0
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_9reset(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_8reset[] = "ChargeIterator.reset(self)";
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_9reset(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("reset (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_8reset(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_8reset(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("reset", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_reset(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 699, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.ChargeIterator.reset", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/deconvoluter_base.pyx":702
 *         self.index = 0
 * 
 *     cpdef bint has_more(self):             # <<<<<<<<<<<<<<
 *         return self.index < self.get_size()
 * 
 */

static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_11has_more(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static int __pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_has_more(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_self, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("has_more", 0);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":703
 * 
 *     cpdef bint has_more(self):
 *         return self.index < self.get_size()             # <<<<<<<<<<<<<<
 * 
 *     cpdef int get_next_value(self):
 */
  __pyx_r = (__pyx_v_self->index < __pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_get_size(__pyx_v_self));
  goto __pyx_L0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":702
 *         self.index = 0
 * 
 *     cpdef bint has_more(self):             # <<<<<<<<<<<<<<
 *         return self.index < self.get_size()
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_11has_more(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_10has_more[] = "ChargeIterator.has_more(self) -> bool";
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_11has_more(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("has_more (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_10has_more(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_10has_more(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("has_more", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_has_more(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 702, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.ChargeIterator.has_more", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/deconvoluter_base.pyx":705
 *         return self.index < self.get_size()
 * 
 *     cpdef int get_next_value(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             int value
 */

static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_13get_next_value(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static int __pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_get_next_value(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_self, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_value;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_next_value", 0);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":708
 *         cdef:
 *             int value
 *         value = self.values[self.index]             # <<<<<<<<<<<<<<
 *         self.index += 1
 *         return value
 */
  __pyx_v_value = (__pyx_v_self->values[__pyx_v_self->index]);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":709
 *             int value
 *         value = self.values[self.index]
 *         self.index += 1             # <<<<<<<<<<<<<<
 *         return value
 * 
 */
  __pyx_v_self->index = (__pyx_v_self->index + 1);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":710
 *         value = self.values[self.index]
 *         self.index += 1
 *         return value             # <<<<<<<<<<<<<<
 * 
 *     cdef size_t get_size(self):
 */
  __pyx_r = __pyx_v_value;
  goto __pyx_L0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":705
 *         return self.index < self.get_size()
 * 
 *     cpdef int get_next_value(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             int value
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_13get_next_value(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_12get_next_value[] = "ChargeIterator.get_next_value(self) -> int";
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_13get_next_value(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_next_value (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_12get_next_value(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_12get_next_value(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("get_next_value", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_get_next_value(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 705, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.ChargeIterator.get_next_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/deconvoluter_base.pyx":712
 *         return value
 * 
 *     cdef size_t get_size(self):             # <<<<<<<<<<<<<<
 *         return self.size
 * 
 */

static size_t __pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_get_size(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_self) {
  size_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_size", 0);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":713
 * 
 *     cdef size_t get_size(self):
 *         return self.size             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  __pyx_r = __pyx_v_self->size;
  goto __pyx_L0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":712
 *         return value
 * 
 *     cdef size_t get_size(self):             # <<<<<<<<<<<<<<
 *         return self.size
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/deconvoluter_base.pyx":715
 *         return self.size
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.size
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_15__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_15__len__(PyObject *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_14__len__(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_14__len__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":716
 * 
 *     def __len__(self):
 *         return self.size             # <<<<<<<<<<<<<<
 * 
 *     def __iter__(self):
 */
  __pyx_r = __pyx_v_self->size;
  goto __pyx_L0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":715
 *         return self.size
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.size
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/deconvoluter_base.pyx":718
 *         return self.size
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         return self
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_17__iter__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_17__iter__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_16__iter__(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_16__iter__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__", 0);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":719
 * 
 *     def __iter__(self):
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def __next__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __pyx_r = ((PyObject *)__pyx_v_self);
  goto __pyx_L0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":718
 *         return self.size
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         return self
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/deconvoluter_base.pyx":721
 *         return self
 * 
 *     def __next__(self):             # <<<<<<<<<<<<<<
 *         if self.index == self.size:
 *             raise StopIteration()
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_19__next__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_19__next__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__next__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_18__next__(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_18__next__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_self) {
  int __pyx_v_value;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__next__", 0);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":722
 * 
 *     def __next__(self):
 *         if self.index == self.size:             # <<<<<<<<<<<<<<
 *             raise StopIteration()
 *         value = self.values[self.index]
 */
  __pyx_t_1 = ((__pyx_v_self->index == __pyx_v_self->size) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "ms_deisotope/_c/deconvoluter_base.pyx":723
 *     def __next__(self):
 *         if self.index == self.size:
 *             raise StopIteration()             # <<<<<<<<<<<<<<
 *         value = self.values[self.index]
 *         self.index += 1
 */
    __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_builtin_StopIteration); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 723, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 723, __pyx_L1_error)

    /* "ms_deisotope/_c/deconvoluter_base.pyx":722
 * 
 *     def __next__(self):
 *         if self.index == self.size:             # <<<<<<<<<<<<<<
 *             raise StopIteration()
 *         value = self.values[self.index]
 */
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":724
 *         if self.index == self.size:
 *             raise StopIteration()
 *         value = self.values[self.index]             # <<<<<<<<<<<<<<
 *         self.index += 1
 *         return value
 */
  __pyx_v_value = (__pyx_v_self->values[__pyx_v_self->index]);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":725
 *             raise StopIteration()
 *         value = self.values[self.index]
 *         self.index += 1             # <<<<<<<<<<<<<<
 *         return value
 * 
 */
  __pyx_v_self->index = (__pyx_v_self->index + 1);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":726
 *         value = self.values[self.index]
 *         self.index += 1
 *         return value             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 726, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":721
 *         return self
 * 
 *     def __next__(self):             # <<<<<<<<<<<<<<
 *         if self.index == self.size:
 *             raise StopIteration()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.ChargeIterator.__next__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/deconvoluter_base.pxd":93
 * cdef class ChargeIterator(object):
 *     cdef:
 *         public int lower             # <<<<<<<<<<<<<<
 *         public int upper
 *         public int sign
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_5lower_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_5lower_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_5lower___get__(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_5lower___get__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->lower); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 93, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.ChargeIterator.lower.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_5lower_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_5lower_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_5lower_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_5lower_2__set__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 93, __pyx_L1_error)
  __pyx_v_self->lower = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.ChargeIterator.lower.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/deconvoluter_base.pxd":94
 *     cdef:
 *         public int lower
 *         public int upper             # <<<<<<<<<<<<<<
 *         public int sign
 *         int* values
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_5upper_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_5upper_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_5upper___get__(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_5upper___get__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->upper); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 94, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.ChargeIterator.upper.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_5upper_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_5upper_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_5upper_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_5upper_2__set__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 94, __pyx_L1_error)
  __pyx_v_self->upper = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.ChargeIterator.upper.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/deconvoluter_base.pxd":95
 *         public int lower
 *         public int upper
 *         public int sign             # <<<<<<<<<<<<<<
 *         int* values
 *         public size_t size
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_4sign_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_4sign_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_4sign___get__(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_4sign___get__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->sign); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 95, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.ChargeIterator.sign.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_4sign_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_4sign_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_4sign_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_4sign_2__set__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 95, __pyx_L1_error)
  __pyx_v_self->sign = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.ChargeIterator.sign.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/deconvoluter_base.pxd":97
 *         public int sign
 *         int* values
 *         public size_t size             # <<<<<<<<<<<<<<
 *         public size_t index
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_4size_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_4size_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_4size___get__(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_4size___get__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->size); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 97, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.ChargeIterator.size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_4size_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_4size_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_4size_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_4size_2__set__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_size_t(__pyx_v_value); if (unlikely((__pyx_t_1 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(2, 97, __pyx_L1_error)
  __pyx_v_self->size = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.ChargeIterator.size.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/deconvoluter_base.pxd":98
 *         int* values
 *         public size_t size
 *         public size_t index             # <<<<<<<<<<<<<<
 * 
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_5index_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_5index_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_5index___get__(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_5index___get__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->index); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 98, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.ChargeIterator.index.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_5index_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_5index_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_5index_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_5index_2__set__(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_size_t(__pyx_v_value); if (unlikely((__pyx_t_1 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(2, 98, __pyx_L1_error)
  __pyx_v_self->index = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.ChargeIterator.index.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.values cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_21__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_20__reduce_cython__[] = "ChargeIterator.__reduce_cython__(self)";
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_21__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_20__reduce_cython__(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_20__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("self.values cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.values cannot be converted to a Python object for pickling")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.values cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.ChargeIterator.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("self.values cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.values cannot be converted to a Python object for pickling")
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_23__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_22__setstate_cython__[] = "ChargeIterator.__setstate_cython__(self, __pyx_state)";
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_23__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_22__setstate_cython__(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_22__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("self.values cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.values cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("self.values cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.values cannot be converted to a Python object for pickling")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.ChargeIterator.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/deconvoluter_base.pyx":730
 * 
 * @cython.cdivision
 * cpdef np.ndarray[np.int32_t, ndim=1] quick_charge(FittedPeakCollection peak_set, FittedPeakOrPosition position_spec, int min_charge, int max_charge):             # <<<<<<<<<<<<<<
 *     """An implementation of Hoopman's QuickCharge [1] algorithm for quickly capping charge
 *     state queries
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_1quick_charge(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_quick_charge[] = "quick_charge(signatures, args, kwargs, defaults)\nAn implementation of Hoopman's QuickCharge [1] algorithm for quickly capping charge\n    state queries\n\n    Parameters\n    ----------\n    peak_set : :class:`ms_peak_picker.PeakSet\n        The centroided peak set to search\n    position_spec : :class:`ms_peak_picker.FittedPeak` or int\n        The position of the peak to start the search from. :attr:`ms_peak_picker.FittedPeak.peak_count`\n        holds the value that will be used if a peak is passed, otherwise the value is assumed to be an\n        index into the `peak_set`.\n    min_charge : int\n        The minimum charge state to consider\n    max_charge : int\n        The maximum charge state to consider\n\n    Returns\n    -------\n    np.ndarray\n        The list of feasible charge states\n\n    References\n    ----------\n    [1] Hoopmann, M. R., Finney, G. L., MacCoss, M. J., Michael R. Hoopmann, Gregory L. Finney,\n        and, MacCoss*, M. J., \342\200\246 MacCoss, M. J. (2007). \"High-speed data reduction, feature detection\n        and MS/MS spectrum quality assessment of shotgun proteomics data sets using high-resolution\n        Mass Spectrometry\". Analytical Chemistry, 79(15), 5620\342\200\2235632. https://doi.org/10.1021/ac0700833\n    ";
static PyMethodDef __pyx_mdef_12ms_deisotope_2_c_17deconvoluter_base_1quick_charge = {"quick_charge", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_1quick_charge, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_quick_charge};
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_1quick_charge(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_signatures = 0;
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED PyObject *__pyx_v_defaults = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fused_cpdef (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_signatures,&__pyx_n_s_args,&__pyx_n_s_kwargs,&__pyx_n_s_defaults,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_signatures)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_args)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 4, 4, 1); __PYX_ERR(0, 730, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_kwargs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 4, 4, 2); __PYX_ERR(0, 730, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_defaults)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 4, 4, 3); __PYX_ERR(0, 730, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fused_cpdef") < 0)) __PYX_ERR(0, 730, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_signatures = values[0];
    __pyx_v_args = values[1];
    __pyx_v_kwargs = values[2];
    __pyx_v_defaults = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 730, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_quick_charge(__pyx_self, __pyx_v_signatures, __pyx_v_args, __pyx_v_kwargs, __pyx_v_defaults);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_quick_charge(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults) {
  PyObject *__pyx_v_dest_sig = NULL;
  Py_ssize_t __pyx_v_i;
  CYTHON_UNUSED PyTypeObject *__pyx_v_ndarray = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_v_itemsize;
  PyObject *__pyx_v_arg = NULL;
  PyObject *__pyx_v_candidates = NULL;
  PyObject *__pyx_v_sig = NULL;
  int __pyx_v_match_found;
  PyObject *__pyx_v_src_sig = NULL;
  PyObject *__pyx_v_dst_type = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  Py_ssize_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  int __pyx_t_16;
  __Pyx_RefNannySetupContext("quick_charge", 0);
  __Pyx_INCREF(__pyx_v_kwargs);
  __pyx_t_1 = PyList_New(1 * 2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 730, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  { Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < 2; __pyx_temp++) {
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      PyList_SET_ITEM(__pyx_t_1, __pyx_temp, Py_None);
    }
  }
  __pyx_v_dest_sig = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_kwargs); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 730, __pyx_L1_error)
  __pyx_t_3 = ((!__pyx_t_4) != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {
    __Pyx_INCREF(Py_None);
    __Pyx_DECREF_SET(__pyx_v_kwargs, Py_None);
  }
  __pyx_t_1 = ((PyObject *)__Pyx_ImportNumPyArrayTypeIfAvailable()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 730, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_ndarray = ((PyTypeObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_v_itemsize = -1L;
  if (unlikely(__pyx_v_args == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 730, __pyx_L1_error)
  }
  __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 730, __pyx_L1_error)
  __pyx_t_2 = ((0 < __pyx_t_5) != 0);
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 730, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_GetItemInt_Tuple(((PyObject*)__pyx_v_args), 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 730, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_arg = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L6;
  }
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L7_bool_binop_done;
  }
  if (unlikely(__pyx_v_kwargs == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 730, __pyx_L1_error)
  }
  __pyx_t_4 = (__Pyx_PyDict_ContainsTF(__pyx_n_s_peak_set, ((PyObject*)__pyx_v_kwargs), Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 730, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_4 != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_kwargs == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 730, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_kwargs), __pyx_n_s_peak_set); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 730, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_arg = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L6;
  }
  /*else*/ {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 730, __pyx_L1_error)
    }
    __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 730, __pyx_L1_error)
    __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 730, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 730, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_int_4);
    __Pyx_GIVEREF(__pyx_int_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_int_4);
    __Pyx_INCREF(__pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_n_s_s);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Expected_at_least_d_argument_s_g, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 730, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 730, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(0, 730, __pyx_L1_error)
  }
  __pyx_L6:;
  while (1) {
    __pyx_t_6 = PyList_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 730, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_kp_s__8);
    __Pyx_GIVEREF(__pyx_kp_s__8);
    PyList_SET_ITEM(__pyx_t_6, 0, __pyx_kp_s__8);
    __pyx_t_1 = PyTuple_New(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 730, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_kp_s_ms_peak_picker__c_peak_set);
    __Pyx_GIVEREF(__pyx_kp_s_ms_peak_picker__c_peak_set);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_s_ms_peak_picker__c_peak_set);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_1, 1, Py_None);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_6);
    __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin___import__, __pyx_t_1, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 730, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_PeakSet); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 730, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_2 = PyObject_IsInstance(__pyx_v_arg, __pyx_t_1); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 730, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_3 = (__pyx_t_2 != 0);
    if (__pyx_t_3) {
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_PeakSet, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0)) __PYX_ERR(0, 730, __pyx_L1_error)
      goto __pyx_L10_break;
    }
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 730, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_kp_s__8);
    __Pyx_GIVEREF(__pyx_kp_s__8);
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_kp_s__8);
    __pyx_t_6 = PyTuple_New(4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 730, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_kp_s_ms_peak_picker__c_peak_index);
    __Pyx_GIVEREF(__pyx_kp_s_ms_peak_picker__c_peak_index);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_kp_s_ms_peak_picker__c_peak_index);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_6, 1, Py_None);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_6, 2, Py_None);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_6, 3, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin___import__, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 730, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_PeakIndex); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 730, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_3 = PyObject_IsInstance(__pyx_v_arg, __pyx_t_6); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 730, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_2 = (__pyx_t_3 != 0);
    if (__pyx_t_2) {
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_PeakIndex, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0)) __PYX_ERR(0, 730, __pyx_L1_error)
      goto __pyx_L10_break;
    }
    if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, Py_None, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0)) __PYX_ERR(0, 730, __pyx_L1_error)
    goto __pyx_L10_break;
  }
  __pyx_L10_break:;
  if (unlikely(__pyx_v_args == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 730, __pyx_L1_error)
  }
  __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 730, __pyx_L1_error)
  __pyx_t_2 = ((1 < __pyx_t_5) != 0);
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 730, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_Tuple(((PyObject*)__pyx_v_args), 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 730, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF_SET(__pyx_v_arg, __pyx_t_6);
    __pyx_t_6 = 0;
    goto __pyx_L13;
  }
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L14_bool_binop_done;
  }
  if (unlikely(__pyx_v_kwargs == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 730, __pyx_L1_error)
  }
  __pyx_t_4 = (__Pyx_PyDict_ContainsTF(__pyx_n_s_position_spec, ((PyObject*)__pyx_v_kwargs), Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 730, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_4 != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L14_bool_binop_done:;
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_kwargs == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 730, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_kwargs), __pyx_n_s_position_spec); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 730, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF_SET(__pyx_v_arg, __pyx_t_6);
    __pyx_t_6 = 0;
    goto __pyx_L13;
  }
  /*else*/ {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 730, __pyx_L1_error)
    }
    __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 730, __pyx_L1_error)
    __pyx_t_6 = PyInt_FromSsize_t(__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 730, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 730, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_int_4);
    __Pyx_GIVEREF(__pyx_int_4);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_int_4);
    __Pyx_INCREF(__pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_n_s_s);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_6);
    __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyString_Format(__pyx_kp_s_Expected_at_least_d_argument_s_g, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 730, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 730, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 730, __pyx_L1_error)
  }
  __pyx_L13:;
  while (1) {
    __pyx_t_3 = PyInt_Check(__pyx_v_arg); 
    __pyx_t_4 = (__pyx_t_3 != 0);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L19_bool_binop_done;
    }
    __pyx_t_4 = PyLong_Check(__pyx_v_arg); 
    __pyx_t_3 = (__pyx_t_4 != 0);
    __pyx_t_2 = __pyx_t_3;
    __pyx_L19_bool_binop_done:;
    __pyx_t_3 = (__pyx_t_2 != 0);
    if (__pyx_t_3) {
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 1, __pyx_n_s_size_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0)) __PYX_ERR(0, 730, __pyx_L1_error)
      goto __pyx_L17_break;
    }
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 730, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_kp_s__8);
    __Pyx_GIVEREF(__pyx_kp_s__8);
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_kp_s__8);
    __pyx_t_6 = PyTuple_New(4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 730, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_kp_s_ms_peak_picker__c_peak_set);
    __Pyx_GIVEREF(__pyx_kp_s_ms_peak_picker__c_peak_set);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_kp_s_ms_peak_picker__c_peak_set);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_6, 1, Py_None);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_6, 2, Py_None);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_6, 3, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin___import__, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 730, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_FittedPeak); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 730, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_3 = PyObject_IsInstance(__pyx_v_arg, __pyx_t_6); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 730, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_2 = (__pyx_t_3 != 0);
    if (__pyx_t_2) {
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 1, __pyx_n_s_FittedPeak, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0)) __PYX_ERR(0, 730, __pyx_L1_error)
      goto __pyx_L17_break;
    }
    if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 1, Py_None, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0)) __PYX_ERR(0, 730, __pyx_L1_error)
    goto __pyx_L17_break;
  }
  __pyx_L17_break:;
  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 730, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_v_candidates = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;
  __pyx_t_5 = 0;
  if (unlikely(__pyx_v_signatures == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 730, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_dict_iterator(((PyObject*)__pyx_v_signatures), 1, ((PyObject *)NULL), (&__pyx_t_7), (&__pyx_t_8)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 730, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __pyx_t_6 = __pyx_t_1;
  __pyx_t_1 = 0;
  while (1) {
    __pyx_t_9 = __Pyx_dict_iter_next(__pyx_t_6, __pyx_t_7, &__pyx_t_5, &__pyx_t_1, NULL, NULL, __pyx_t_8);
    if (unlikely(__pyx_t_9 == 0)) break;
    if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(0, 730, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_sig, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_v_match_found = 0;
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_sig, __pyx_n_s_strip); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 730, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_12 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_10 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_12, __pyx_kp_s__9) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_s__9);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 730, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_split); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 730, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_1 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_10, __pyx_kp_s__10) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_s__10);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 730, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_XDECREF_SET(__pyx_v_src_sig, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_13 = PyList_GET_SIZE(__pyx_v_dest_sig); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(0, 730, __pyx_L1_error)
    __pyx_t_14 = __pyx_t_13;
    for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
      __pyx_v_i = __pyx_t_15;
      __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_v_dest_sig, __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 1, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 730, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_dst_type, __pyx_t_1);
      __pyx_t_1 = 0;
      __pyx_t_2 = (__pyx_v_dst_type != Py_None);
      __pyx_t_3 = (__pyx_t_2 != 0);
      if (__pyx_t_3) {
        __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_src_sig, __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 730, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = PyObject_RichCompare(__pyx_t_1, __pyx_v_dst_type, Py_EQ); __Pyx_XGOTREF(__pyx_t_11); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 730, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_11); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 730, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (__pyx_t_3) {
          __pyx_v_match_found = 1;
          goto __pyx_L27;
        }
        /*else*/ {
          __pyx_v_match_found = 0;
          goto __pyx_L25_break;
        }
        __pyx_L27:;
      }
    }
    __pyx_L25_break:;
    __pyx_t_3 = (__pyx_v_match_found != 0);
    if (__pyx_t_3) {
      __pyx_t_16 = __Pyx_PyList_Append(__pyx_v_candidates, __pyx_v_sig); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(0, 730, __pyx_L1_error)
    }
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_3 = (PyList_GET_SIZE(__pyx_v_candidates) != 0);
  __pyx_t_2 = ((!__pyx_t_3) != 0);
  if (__pyx_t_2) {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__11, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 730, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(0, 730, __pyx_L1_error)
  }
  __pyx_t_7 = PyList_GET_SIZE(__pyx_v_candidates); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 730, __pyx_L1_error)
  __pyx_t_2 = ((__pyx_t_7 > 1) != 0);
  if (__pyx_t_2) {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__12, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 730, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(0, 730, __pyx_L1_error)
  }
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(__pyx_v_signatures == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 730, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_List(__pyx_v_candidates, 0, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 730, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_11 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_signatures), __pyx_t_6); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 730, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_r = __pyx_t_11;
    __pyx_t_11 = 0;
    goto __pyx_L0;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_dest_sig);
  __Pyx_XDECREF(__pyx_v_ndarray);
  __Pyx_XDECREF(__pyx_v_arg);
  __Pyx_XDECREF(__pyx_v_candidates);
  __Pyx_XDECREF(__pyx_v_sig);
  __Pyx_XDECREF(__pyx_v_src_sig);
  __Pyx_XDECREF(__pyx_v_dst_type);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_17__pyx_fuse_0_0quick_charge(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_1quick_charge(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyArrayObject *__pyx_fuse_0_0__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_quick_charge(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peak_set, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_position_spec, int __pyx_v_min_charge, int __pyx_v_max_charge, CYTHON_UNUSED int __pyx_skip_dispatch) {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaks = 0;
  size_t __pyx_v_index;
  size_t __pyx_v_first_index;
  int __pyx_v_charges[0x3E8];
  PyArrayObject *__pyx_v_result = 0;
  double __pyx_v_min_intensity;
  double __pyx_v_diff;
  double __pyx_v_raw_charge;
  double __pyx_v_remain;
  int __pyx_v_charge;
  int __pyx_v_result_size;
  Py_ssize_t __pyx_v_j;
  size_t __pyx_v_i;
  size_t __pyx_v_n;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_result;
  __Pyx_Buffer __pyx_pybuffer_result;
  PyArrayObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  long __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  size_t __pyx_t_7;
  size_t __pyx_t_8;
  size_t __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyArrayObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_0_0quick_charge", 0);
  __pyx_pybuffer_result.pybuffer.buf = NULL;
  __pyx_pybuffer_result.refcount = 0;
  __pyx_pybuffernd_result.data = NULL;
  __pyx_pybuffernd_result.rcbuffer = &__pyx_pybuffer_result;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":771
 *         bint matched
 *     if FittedPeakOrPosition is FittedPeak:
 *         index = position_spec.peak_count             # <<<<<<<<<<<<<<
 *     else:
 *         index = position_spec
 */
  __pyx_t_1 = __pyx_v_position_spec->peak_count;
  __pyx_v_index = __pyx_t_1;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":775
 *         index = position_spec
 *     if FittedPeakCollection is PeakSet:
 *         peaks = peak_set             # <<<<<<<<<<<<<<
 *     else:
 *         peaks = peak_set.peaks
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_peak_set));
  __pyx_v_peaks = __pyx_v_peak_set;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":778
 *     else:
 *         peaks = peak_set.peaks
 *     if index > peaks.get_size():             # <<<<<<<<<<<<<<
 *         if peaks.get_size() == 0:
 *             raise IndexError("peak_set cannot be empty!")
 */
  __pyx_t_2 = ((__pyx_v_index > ((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->get_size(__pyx_v_peaks)) != 0);
  if (__pyx_t_2) {

    /* "ms_deisotope/_c/deconvoluter_base.pyx":779
 *         peaks = peak_set.peaks
 *     if index > peaks.get_size():
 *         if peaks.get_size() == 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("peak_set cannot be empty!")
 *         # Assume we have a slice of the peak list, not the actual complete list, so
 */
    __pyx_t_2 = ((((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->get_size(__pyx_v_peaks) == 0) != 0);
    if (unlikely(__pyx_t_2)) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":780
 *     if index > peaks.get_size():
 *         if peaks.get_size() == 0:
 *             raise IndexError("peak_set cannot be empty!")             # <<<<<<<<<<<<<<
 *         # Assume we have a slice of the peak list, not the actual complete list, so
 *         # adjust the index offset from the first peak in the given list.
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple__13, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 780, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 780, __pyx_L1_error)

      /* "ms_deisotope/_c/deconvoluter_base.pyx":779
 *         peaks = peak_set.peaks
 *     if index > peaks.get_size():
 *         if peaks.get_size() == 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("peak_set cannot be empty!")
 *         # Assume we have a slice of the peak list, not the actual complete list, so
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":783
 *         # Assume we have a slice of the peak list, not the actual complete list, so
 *         # adjust the index offset from the first peak in the given list.
 *         first_index = peaks.getitem(0).peak_count             # <<<<<<<<<<<<<<
 *         if (index - first_index) < peaks.get_size() and (index - first_index) >= 0:
 *             index -= first_index
 */
    __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->getitem(__pyx_v_peaks, 0)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 783, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_3)->peak_count;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_first_index = __pyx_t_1;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":784
 *         # adjust the index offset from the first peak in the given list.
 *         first_index = peaks.getitem(0).peak_count
 *         if (index - first_index) < peaks.get_size() and (index - first_index) >= 0:             # <<<<<<<<<<<<<<
 *             index -= first_index
 *         else:
 */
    __pyx_t_4 = (((__pyx_v_index - __pyx_v_first_index) < ((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->get_size(__pyx_v_peaks)) != 0);
    if (__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_4 = (((__pyx_v_index - __pyx_v_first_index) >= 0) != 0);
    __pyx_t_2 = __pyx_t_4;
    __pyx_L6_bool_binop_done:;
    if (likely(__pyx_t_2)) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":785
 *         first_index = peaks.getitem(0).peak_count
 *         if (index - first_index) < peaks.get_size() and (index - first_index) >= 0:
 *             index -= first_index             # <<<<<<<<<<<<<<
 *         else:
 *             raise IndexError("%d is out of bounds for peak list of size %d in quick_charge" % (index, peaks.get_size()))
 */
      __pyx_v_index = (__pyx_v_index - __pyx_v_first_index);

      /* "ms_deisotope/_c/deconvoluter_base.pyx":784
 *         # adjust the index offset from the first peak in the given list.
 *         first_index = peaks.getitem(0).peak_count
 *         if (index - first_index) < peaks.get_size() and (index - first_index) >= 0:             # <<<<<<<<<<<<<<
 *             index -= first_index
 *         else:
 */
      goto __pyx_L5;
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":787
 *             index -= first_index
 *         else:
 *             raise IndexError("%d is out of bounds for peak list of size %d in quick_charge" % (index, peaks.get_size()))             # <<<<<<<<<<<<<<
 *     n = 1000
 *     result_size = 0
 */
    /*else*/ {
      __pyx_t_3 = __Pyx_PyInt_FromSize_t(__pyx_v_index); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 787, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = __Pyx_PyInt_FromSize_t(((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->get_size(__pyx_v_peaks)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 787, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 787, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_5);
      __pyx_t_3 = 0;
      __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_d_is_out_of_bounds_for_peak_lis, __pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 787, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_IndexError, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 787, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_Raise(__pyx_t_6, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __PYX_ERR(0, 787, __pyx_L1_error)
    }
    __pyx_L5:;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":778
 *     else:
 *         peaks = peak_set.peaks
 *     if index > peaks.get_size():             # <<<<<<<<<<<<<<
 *         if peaks.get_size() == 0:
 *             raise IndexError("peak_set cannot be empty!")
 */
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":788
 *         else:
 *             raise IndexError("%d is out of bounds for peak list of size %d in quick_charge" % (index, peaks.get_size()))
 *     n = 1000             # <<<<<<<<<<<<<<
 *     result_size = 0
 *     min_intensity = peaks.getitem(index).intensity / 4.
 */
  __pyx_v_n = 0x3E8;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":789
 *             raise IndexError("%d is out of bounds for peak list of size %d in quick_charge" % (index, peaks.get_size()))
 *     n = 1000
 *     result_size = 0             # <<<<<<<<<<<<<<
 *     min_intensity = peaks.getitem(index).intensity / 4.
 *     for i in range(n):
 */
  __pyx_v_result_size = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":790
 *     n = 1000
 *     result_size = 0
 *     min_intensity = peaks.getitem(index).intensity / 4.             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         charges[i] = 0
 */
  __pyx_t_6 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->getitem(__pyx_v_peaks, __pyx_v_index)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 790, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_v_min_intensity = (((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_6)->__pyx_base.intensity / 4.);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":791
 *     result_size = 0
 *     min_intensity = peaks.getitem(index).intensity / 4.
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         charges[i] = 0
 *     for j in range(index + 1, (peaks.get_size())):
 */
  __pyx_t_7 = __pyx_v_n;
  __pyx_t_8 = __pyx_t_7;
  for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
    __pyx_v_i = __pyx_t_9;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":792
 *     min_intensity = peaks.getitem(index).intensity / 4.
 *     for i in range(n):
 *         charges[i] = 0             # <<<<<<<<<<<<<<
 *     for j in range(index + 1, (peaks.get_size())):
 *         if peaks.getitem(j).intensity < min_intensity:
 */
    (__pyx_v_charges[__pyx_v_i]) = 0;
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":793
 *     for i in range(n):
 *         charges[i] = 0
 *     for j in range(index + 1, (peaks.get_size())):             # <<<<<<<<<<<<<<
 *         if peaks.getitem(j).intensity < min_intensity:
 *             continue
 */
  __pyx_t_7 = ((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->get_size(__pyx_v_peaks);
  __pyx_t_8 = __pyx_t_7;
  for (__pyx_t_10 = (__pyx_v_index + 1); __pyx_t_10 < __pyx_t_8; __pyx_t_10+=1) {
    __pyx_v_j = __pyx_t_10;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":794
 *         charges[i] = 0
 *     for j in range(index + 1, (peaks.get_size())):
 *         if peaks.getitem(j).intensity < min_intensity:             # <<<<<<<<<<<<<<
 *             continue
 *         diff = peaks.getitem(j).mz - peaks.getitem(index).mz
 */
    __pyx_t_6 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->getitem(__pyx_v_peaks, __pyx_v_j)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 794, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = ((((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_6)->__pyx_base.intensity < __pyx_v_min_intensity) != 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (__pyx_t_2) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":795
 *     for j in range(index + 1, (peaks.get_size())):
 *         if peaks.getitem(j).intensity < min_intensity:
 *             continue             # <<<<<<<<<<<<<<
 *         diff = peaks.getitem(j).mz - peaks.getitem(index).mz
 *         if diff > 1.1:
 */
      goto __pyx_L10_continue;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":794
 *         charges[i] = 0
 *     for j in range(index + 1, (peaks.get_size())):
 *         if peaks.getitem(j).intensity < min_intensity:             # <<<<<<<<<<<<<<
 *             continue
 *         diff = peaks.getitem(j).mz - peaks.getitem(index).mz
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":796
 *         if peaks.getitem(j).intensity < min_intensity:
 *             continue
 *         diff = peaks.getitem(j).mz - peaks.getitem(index).mz             # <<<<<<<<<<<<<<
 *         if diff > 1.1:
 *             break
 */
    __pyx_t_6 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->getitem(__pyx_v_peaks, __pyx_v_j)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 796, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->getitem(__pyx_v_peaks, __pyx_v_index)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 796, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_v_diff = (((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_6)->__pyx_base.mz - ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_5)->__pyx_base.mz);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":797
 *             continue
 *         diff = peaks.getitem(j).mz - peaks.getitem(index).mz
 *         if diff > 1.1:             # <<<<<<<<<<<<<<
 *             break
 *         raw_charge = 1 / diff
 */
    __pyx_t_2 = ((__pyx_v_diff > 1.1) != 0);
    if (__pyx_t_2) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":798
 *         diff = peaks.getitem(j).mz - peaks.getitem(index).mz
 *         if diff > 1.1:
 *             break             # <<<<<<<<<<<<<<
 *         raw_charge = 1 / diff
 *         charge = <int>(raw_charge + 0.5)
 */
      goto __pyx_L11_break;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":797
 *             continue
 *         diff = peaks.getitem(j).mz - peaks.getitem(index).mz
 *         if diff > 1.1:             # <<<<<<<<<<<<<<
 *             break
 *         raw_charge = 1 / diff
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":799
 *         if diff > 1.1:
 *             break
 *         raw_charge = 1 / diff             # <<<<<<<<<<<<<<
 *         charge = <int>(raw_charge + 0.5)
 *         remain = raw_charge - <int>(raw_charge)
 */
    __pyx_v_raw_charge = (1.0 / __pyx_v_diff);

    /* "ms_deisotope/_c/deconvoluter_base.pyx":800
 *             break
 *         raw_charge = 1 / diff
 *         charge = <int>(raw_charge + 0.5)             # <<<<<<<<<<<<<<
 *         remain = raw_charge - <int>(raw_charge)
 *         if 0.2 < remain and remain < 0.8:
 */
    __pyx_v_charge = ((int)(__pyx_v_raw_charge + 0.5));

    /* "ms_deisotope/_c/deconvoluter_base.pyx":801
 *         raw_charge = 1 / diff
 *         charge = <int>(raw_charge + 0.5)
 *         remain = raw_charge - <int>(raw_charge)             # <<<<<<<<<<<<<<
 *         if 0.2 < remain and remain < 0.8:
 *             continue
 */
    __pyx_v_remain = (__pyx_v_raw_charge - ((int)__pyx_v_raw_charge));

    /* "ms_deisotope/_c/deconvoluter_base.pyx":802
 *         charge = <int>(raw_charge + 0.5)
 *         remain = raw_charge - <int>(raw_charge)
 *         if 0.2 < remain and remain < 0.8:             # <<<<<<<<<<<<<<
 *             continue
 *         if (charge < min_charge) or (charge > max_charge):
 */
    __pyx_t_4 = ((0.2 < __pyx_v_remain) != 0);
    if (__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L15_bool_binop_done;
    }
    __pyx_t_4 = ((__pyx_v_remain < 0.8) != 0);
    __pyx_t_2 = __pyx_t_4;
    __pyx_L15_bool_binop_done:;
    if (__pyx_t_2) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":803
 *         remain = raw_charge - <int>(raw_charge)
 *         if 0.2 < remain and remain < 0.8:
 *             continue             # <<<<<<<<<<<<<<
 *         if (charge < min_charge) or (charge > max_charge):
 *             continue
 */
      goto __pyx_L10_continue;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":802
 *         charge = <int>(raw_charge + 0.5)
 *         remain = raw_charge - <int>(raw_charge)
 *         if 0.2 < remain and remain < 0.8:             # <<<<<<<<<<<<<<
 *             continue
 *         if (charge < min_charge) or (charge > max_charge):
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":804
 *         if 0.2 < remain and remain < 0.8:
 *             continue
 *         if (charge < min_charge) or (charge > max_charge):             # <<<<<<<<<<<<<<
 *             continue
 *         if charges[charge] == 0:
 */
    __pyx_t_4 = ((__pyx_v_charge < __pyx_v_min_charge) != 0);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L18_bool_binop_done;
    }
    __pyx_t_4 = ((__pyx_v_charge > __pyx_v_max_charge) != 0);
    __pyx_t_2 = __pyx_t_4;
    __pyx_L18_bool_binop_done:;
    if (__pyx_t_2) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":805
 *             continue
 *         if (charge < min_charge) or (charge > max_charge):
 *             continue             # <<<<<<<<<<<<<<
 *         if charges[charge] == 0:
 *             result_size += 1
 */
      goto __pyx_L10_continue;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":804
 *         if 0.2 < remain and remain < 0.8:
 *             continue
 *         if (charge < min_charge) or (charge > max_charge):             # <<<<<<<<<<<<<<
 *             continue
 *         if charges[charge] == 0:
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":806
 *         if (charge < min_charge) or (charge > max_charge):
 *             continue
 *         if charges[charge] == 0:             # <<<<<<<<<<<<<<
 *             result_size += 1
 *         charges[charge] = 1
 */
    __pyx_t_2 = (((__pyx_v_charges[__pyx_v_charge]) == 0) != 0);
    if (__pyx_t_2) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":807
 *             continue
 *         if charges[charge] == 0:
 *             result_size += 1             # <<<<<<<<<<<<<<
 *         charges[charge] = 1
 *     if result_size == 0:
 */
      __pyx_v_result_size = (__pyx_v_result_size + 1);

      /* "ms_deisotope/_c/deconvoluter_base.pyx":806
 *         if (charge < min_charge) or (charge > max_charge):
 *             continue
 *         if charges[charge] == 0:             # <<<<<<<<<<<<<<
 *             result_size += 1
 *         charges[charge] = 1
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":808
 *         if charges[charge] == 0:
 *             result_size += 1
 *         charges[charge] = 1             # <<<<<<<<<<<<<<
 *     if result_size == 0:
 *         result = np.empty(0, dtype=np.int32)
 */
    (__pyx_v_charges[__pyx_v_charge]) = 1;
    __pyx_L10_continue:;
  }
  __pyx_L11_break:;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":809
 *             result_size += 1
 *         charges[charge] = 1
 *     if result_size == 0:             # <<<<<<<<<<<<<<
 *         result = np.empty(0, dtype=np.int32)
 *         return result
 */
  __pyx_t_2 = ((__pyx_v_result_size == 0) != 0);
  if (__pyx_t_2) {

    /* "ms_deisotope/_c/deconvoluter_base.pyx":810
 *         charges[charge] = 1
 *     if result_size == 0:
 *         result = np.empty(0, dtype=np.int32)             # <<<<<<<<<<<<<<
 *         return result
 *     for j in range(index - 1, -1, -1):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 810, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_empty); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 810, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 810, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 810, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_int32); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 810, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_dtype, __pyx_t_11) < 0) __PYX_ERR(0, 810, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_tuple__14, __pyx_t_5); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 810, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (!(likely(((__pyx_t_11) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_11, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 810, __pyx_L1_error)
    __pyx_t_12 = ((PyArrayObject *)__pyx_t_11);
    {
      __Pyx_BufFmt_StackElem __pyx_stack[1];
      __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_result.rcbuffer->pybuffer);
      __pyx_t_13 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_result.rcbuffer->pybuffer, (PyObject*)__pyx_t_12, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack);
      if (unlikely(__pyx_t_13 < 0)) {
        PyErr_Fetch(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
        if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_result.rcbuffer->pybuffer, (PyObject*)__pyx_v_result, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {
          Py_XDECREF(__pyx_t_14); Py_XDECREF(__pyx_t_15); Py_XDECREF(__pyx_t_16);
          __Pyx_RaiseBufferFallbackError();
        } else {
          PyErr_Restore(__pyx_t_14, __pyx_t_15, __pyx_t_16);
        }
        __pyx_t_14 = __pyx_t_15 = __pyx_t_16 = 0;
      }
      __pyx_pybuffernd_result.diminfo[0].strides = __pyx_pybuffernd_result.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_result.diminfo[0].shape = __pyx_pybuffernd_result.rcbuffer->pybuffer.shape[0];
      if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 810, __pyx_L1_error)
    }
    __pyx_t_12 = 0;
    __pyx_v_result = ((PyArrayObject *)__pyx_t_11);
    __pyx_t_11 = 0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":811
 *     if result_size == 0:
 *         result = np.empty(0, dtype=np.int32)
 *         return result             # <<<<<<<<<<<<<<
 *     for j in range(index - 1, -1, -1):
 *         diff = peaks.getitem(index).mz - peaks.getitem(j).mz
 */
    __Pyx_XDECREF(((PyObject *)__pyx_r));
    __Pyx_INCREF(((PyObject *)__pyx_v_result));
    __pyx_r = ((PyArrayObject *)__pyx_v_result);
    goto __pyx_L0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":809
 *             result_size += 1
 *         charges[charge] = 1
 *     if result_size == 0:             # <<<<<<<<<<<<<<
 *         result = np.empty(0, dtype=np.int32)
 *         return result
 */
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":812
 *         result = np.empty(0, dtype=np.int32)
 *         return result
 *     for j in range(index - 1, -1, -1):             # <<<<<<<<<<<<<<
 *         diff = peaks.getitem(index).mz - peaks.getitem(j).mz
 *         if diff > 1.1:
 */
  for (__pyx_t_10 = (__pyx_v_index - 1); __pyx_t_10 > -1L; __pyx_t_10-=1) {
    __pyx_v_j = __pyx_t_10;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":813
 *         return result
 *     for j in range(index - 1, -1, -1):
 *         diff = peaks.getitem(index).mz - peaks.getitem(j).mz             # <<<<<<<<<<<<<<
 *         if diff > 1.1:
 *             break
 */
    __pyx_t_11 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->getitem(__pyx_v_peaks, __pyx_v_index)); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 813, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_5 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->getitem(__pyx_v_peaks, __pyx_v_j)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 813, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_v_diff = (((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_11)->__pyx_base.mz - ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_5)->__pyx_base.mz);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":814
 *     for j in range(index - 1, -1, -1):
 *         diff = peaks.getitem(index).mz - peaks.getitem(j).mz
 *         if diff > 1.1:             # <<<<<<<<<<<<<<
 *             break
 *         raw_charge = 1 / diff
 */
    __pyx_t_2 = ((__pyx_v_diff > 1.1) != 0);
    if (__pyx_t_2) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":815
 *         diff = peaks.getitem(index).mz - peaks.getitem(j).mz
 *         if diff > 1.1:
 *             break             # <<<<<<<<<<<<<<
 *         raw_charge = 1 / diff
 *         charge = <int>(raw_charge + 0.5)
 */
      goto __pyx_L23_break;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":814
 *     for j in range(index - 1, -1, -1):
 *         diff = peaks.getitem(index).mz - peaks.getitem(j).mz
 *         if diff > 1.1:             # <<<<<<<<<<<<<<
 *             break
 *         raw_charge = 1 / diff
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":816
 *         if diff > 1.1:
 *             break
 *         raw_charge = 1 / diff             # <<<<<<<<<<<<<<
 *         charge = <int>(raw_charge + 0.5)
 *         remain = raw_charge - <int>(raw_charge)
 */
    __pyx_v_raw_charge = (1.0 / __pyx_v_diff);

    /* "ms_deisotope/_c/deconvoluter_base.pyx":817
 *             break
 *         raw_charge = 1 / diff
 *         charge = <int>(raw_charge + 0.5)             # <<<<<<<<<<<<<<
 *         remain = raw_charge - <int>(raw_charge)
 *         if 0.2 < remain and remain < 0.8:
 */
    __pyx_v_charge = ((int)(__pyx_v_raw_charge + 0.5));

    /* "ms_deisotope/_c/deconvoluter_base.pyx":818
 *         raw_charge = 1 / diff
 *         charge = <int>(raw_charge + 0.5)
 *         remain = raw_charge - <int>(raw_charge)             # <<<<<<<<<<<<<<
 *         if 0.2 < remain and remain < 0.8:
 *             continue
 */
    __pyx_v_remain = (__pyx_v_raw_charge - ((int)__pyx_v_raw_charge));

    /* "ms_deisotope/_c/deconvoluter_base.pyx":819
 *         charge = <int>(raw_charge + 0.5)
 *         remain = raw_charge - <int>(raw_charge)
 *         if 0.2 < remain and remain < 0.8:             # <<<<<<<<<<<<<<
 *             continue
 *         if (charge < min_charge) or (charge > max_charge):
 */
    __pyx_t_4 = ((0.2 < __pyx_v_remain) != 0);
    if (__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L26_bool_binop_done;
    }
    __pyx_t_4 = ((__pyx_v_remain < 0.8) != 0);
    __pyx_t_2 = __pyx_t_4;
    __pyx_L26_bool_binop_done:;
    if (__pyx_t_2) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":820
 *         remain = raw_charge - <int>(raw_charge)
 *         if 0.2 < remain and remain < 0.8:
 *             continue             # <<<<<<<<<<<<<<
 *         if (charge < min_charge) or (charge > max_charge):
 *             continue
 */
      goto __pyx_L22_continue;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":819
 *         charge = <int>(raw_charge + 0.5)
 *         remain = raw_charge - <int>(raw_charge)
 *         if 0.2 < remain and remain < 0.8:             # <<<<<<<<<<<<<<
 *             continue
 *         if (charge < min_charge) or (charge > max_charge):
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":821
 *         if 0.2 < remain and remain < 0.8:
 *             continue
 *         if (charge < min_charge) or (charge > max_charge):             # <<<<<<<<<<<<<<
 *             continue
 *         if charges[charge] == 0:
 */
    __pyx_t_4 = ((__pyx_v_charge < __pyx_v_min_charge) != 0);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L29_bool_binop_done;
    }
    __pyx_t_4 = ((__pyx_v_charge > __pyx_v_max_charge) != 0);
    __pyx_t_2 = __pyx_t_4;
    __pyx_L29_bool_binop_done:;
    if (__pyx_t_2) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":822
 *             continue
 *         if (charge < min_charge) or (charge > max_charge):
 *             continue             # <<<<<<<<<<<<<<
 *         if charges[charge] == 0:
 *             result_size += 1
 */
      goto __pyx_L22_continue;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":821
 *         if 0.2 < remain and remain < 0.8:
 *             continue
 *         if (charge < min_charge) or (charge > max_charge):             # <<<<<<<<<<<<<<
 *             continue
 *         if charges[charge] == 0:
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":823
 *         if (charge < min_charge) or (charge > max_charge):
 *             continue
 *         if charges[charge] == 0:             # <<<<<<<<<<<<<<
 *             result_size += 1
 *         charges[charge] = 1
 */
    __pyx_t_2 = (((__pyx_v_charges[__pyx_v_charge]) == 0) != 0);
    if (__pyx_t_2) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":824
 *             continue
 *         if charges[charge] == 0:
 *             result_size += 1             # <<<<<<<<<<<<<<
 *         charges[charge] = 1
 *     if result_size <= 0:
 */
      __pyx_v_result_size = (__pyx_v_result_size + 1);

      /* "ms_deisotope/_c/deconvoluter_base.pyx":823
 *         if (charge < min_charge) or (charge > max_charge):
 *             continue
 *         if charges[charge] == 0:             # <<<<<<<<<<<<<<
 *             result_size += 1
 *         charges[charge] = 1
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":825
 *         if charges[charge] == 0:
 *             result_size += 1
 *         charges[charge] = 1             # <<<<<<<<<<<<<<
 *     if result_size <= 0:
 *         return np.zeros(0, dtype=np.int32)
 */
    (__pyx_v_charges[__pyx_v_charge]) = 1;
    __pyx_L22_continue:;
  }
  __pyx_L23_break:;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":826
 *             result_size += 1
 *         charges[charge] = 1
 *     if result_size <= 0:             # <<<<<<<<<<<<<<
 *         return np.zeros(0, dtype=np.int32)
 *     result = np.zeros(result_size, dtype=np.int32)
 */
  __pyx_t_2 = ((__pyx_v_result_size <= 0) != 0);
  if (__pyx_t_2) {

    /* "ms_deisotope/_c/deconvoluter_base.pyx":827
 *         charges[charge] = 1
 *     if result_size <= 0:
 *         return np.zeros(0, dtype=np.int32)             # <<<<<<<<<<<<<<
 *     result = np.zeros(result_size, dtype=np.int32)
 * 
 */
    __Pyx_XDECREF(((PyObject *)__pyx_r));
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 827, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_zeros); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 827, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 827, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 827, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_int32); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 827, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_dtype, __pyx_t_3) < 0) __PYX_ERR(0, 827, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_tuple__14, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 827, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 827, __pyx_L1_error)
    __pyx_r = ((PyArrayObject *)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":826
 *             result_size += 1
 *         charges[charge] = 1
 *     if result_size <= 0:             # <<<<<<<<<<<<<<
 *         return np.zeros(0, dtype=np.int32)
 *     result = np.zeros(result_size, dtype=np.int32)
 */
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":828
 *     if result_size <= 0:
 *         return np.zeros(0, dtype=np.int32)
 *     result = np.zeros(result_size, dtype=np.int32)             # <<<<<<<<<<<<<<
 * 
 *     i = 0
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_result_size); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_11 = PyTuple_New(1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_int32); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_t_17) < 0) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  __pyx_t_17 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_11, __pyx_t_3); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!(likely(((__pyx_t_17) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_17, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 828, __pyx_L1_error)
  __pyx_t_12 = ((PyArrayObject *)__pyx_t_17);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_result.rcbuffer->pybuffer);
    __pyx_t_13 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_result.rcbuffer->pybuffer, (PyObject*)__pyx_t_12, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_13 < 0)) {
      PyErr_Fetch(&__pyx_t_16, &__pyx_t_15, &__pyx_t_14);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_result.rcbuffer->pybuffer, (PyObject*)__pyx_v_result, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_16); Py_XDECREF(__pyx_t_15); Py_XDECREF(__pyx_t_14);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_16, __pyx_t_15, __pyx_t_14);
      }
      __pyx_t_16 = __pyx_t_15 = __pyx_t_14 = 0;
    }
    __pyx_pybuffernd_result.diminfo[0].strides = __pyx_pybuffernd_result.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_result.diminfo[0].shape = __pyx_pybuffernd_result.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 828, __pyx_L1_error)
  }
  __pyx_t_12 = 0;
  __pyx_v_result = ((PyArrayObject *)__pyx_t_17);
  __pyx_t_17 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":830
 *     result = np.zeros(result_size, dtype=np.int32)
 * 
 *     i = 0             # <<<<<<<<<<<<<<
 *     for j in range(n):
 *         if charges[j] != 0:
 */
  __pyx_v_i = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":831
 * 
 *     i = 0
 *     for j in range(n):             # <<<<<<<<<<<<<<
 *         if charges[j] != 0:
 *             result[i] = j
 */
  __pyx_t_7 = __pyx_v_n;
  __pyx_t_8 = __pyx_t_7;
  for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_8; __pyx_t_10+=1) {
    __pyx_v_j = __pyx_t_10;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":832
 *     i = 0
 *     for j in range(n):
 *         if charges[j] != 0:             # <<<<<<<<<<<<<<
 *             result[i] = j
 *             i += 1
 */
    __pyx_t_2 = (((__pyx_v_charges[__pyx_v_j]) != 0) != 0);
    if (__pyx_t_2) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":833
 *     for j in range(n):
 *         if charges[j] != 0:
 *             result[i] = j             # <<<<<<<<<<<<<<
 *             i += 1
 *     return result
 */
      __pyx_t_9 = __pyx_v_i;
      __pyx_t_13 = -1;
      if (unlikely(__pyx_t_9 >= (size_t)__pyx_pybuffernd_result.diminfo[0].shape)) __pyx_t_13 = 0;
      if (unlikely(__pyx_t_13 != -1)) {
        __Pyx_RaiseBufferIndexError(__pyx_t_13);
        __PYX_ERR(0, 833, __pyx_L1_error)
      }
      *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int32_t *, __pyx_pybuffernd_result.rcbuffer->pybuffer.buf, __pyx_t_9, __pyx_pybuffernd_result.diminfo[0].strides) = __pyx_v_j;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":834
 *         if charges[j] != 0:
 *             result[i] = j
 *             i += 1             # <<<<<<<<<<<<<<
 *     return result
 * 
 */
      __pyx_v_i = (__pyx_v_i + 1);

      /* "ms_deisotope/_c/deconvoluter_base.pyx":832
 *     i = 0
 *     for j in range(n):
 *         if charges[j] != 0:             # <<<<<<<<<<<<<<
 *             result[i] = j
 *             i += 1
 */
    }
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":835
 *             result[i] = j
 *             i += 1
 *     return result             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_result));
  __pyx_r = ((PyArrayObject *)__pyx_v_result);
  goto __pyx_L0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":730
 * 
 * @cython.cdivision
 * cpdef np.ndarray[np.int32_t, ndim=1] quick_charge(FittedPeakCollection peak_set, FittedPeakOrPosition position_spec, int min_charge, int max_charge):             # <<<<<<<<<<<<<<
 *     """An implementation of Hoopman's QuickCharge [1] algorithm for quickly capping charge
 *     state queries
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_17);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_result.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.quick_charge", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_result.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_peaks);
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_17__pyx_fuse_0_0quick_charge(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_0_0__pyx_mdef_12ms_deisotope_2_c_17deconvoluter_base_17__pyx_fuse_0_0quick_charge = {"__pyx_fuse_0_0quick_charge", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_17__pyx_fuse_0_0quick_charge, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_quick_charge};
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_17__pyx_fuse_0_0quick_charge(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peak_set = 0;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_position_spec = 0;
  int __pyx_v_min_charge;
  int __pyx_v_max_charge;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_0_0quick_charge (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_peak_set,&__pyx_n_s_position_spec,&__pyx_n_s_min_charge,&__pyx_n_s_max_charge,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peak_set)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_position_spec)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0_0quick_charge", 1, 4, 4, 1); __PYX_ERR(0, 730, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_min_charge)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0_0quick_charge", 1, 4, 4, 2); __PYX_ERR(0, 730, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_max_charge)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0_0quick_charge", 1, 4, 4, 3); __PYX_ERR(0, 730, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_0_0quick_charge") < 0)) __PYX_ERR(0, 730, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_peak_set = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)values[0]);
    __pyx_v_position_spec = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)values[1]);
    __pyx_v_min_charge = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_min_charge == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 730, __pyx_L3_error)
    __pyx_v_max_charge = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_max_charge == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 730, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0_0quick_charge", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 730, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.__pyx_fuse_0_0quick_charge", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peak_set), __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet, 1, "peak_set", 0))) __PYX_ERR(0, 730, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_position_spec), __pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak, 1, "position_spec", 0))) __PYX_ERR(0, 730, __pyx_L1_error)
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16__pyx_fuse_0_0quick_charge(__pyx_self, __pyx_v_peak_set, __pyx_v_position_spec, __pyx_v_min_charge, __pyx_v_max_charge);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_16__pyx_fuse_0_0quick_charge(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peak_set, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_position_spec, int __pyx_v_min_charge, int __pyx_v_max_charge) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_0_0quick_charge", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_fuse_0_0__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_quick_charge(__pyx_v_peak_set, __pyx_v_position_spec, __pyx_v_min_charge, __pyx_v_max_charge, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 730, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.__pyx_fuse_0_0quick_charge", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_19__pyx_fuse_0_1quick_charge(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_1quick_charge(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyArrayObject *__pyx_fuse_0_1__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_quick_charge(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peak_set, size_t __pyx_v_position_spec, int __pyx_v_min_charge, int __pyx_v_max_charge, CYTHON_UNUSED int __pyx_skip_dispatch) {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaks = 0;
  size_t __pyx_v_index;
  size_t __pyx_v_first_index;
  int __pyx_v_charges[0x3E8];
  PyArrayObject *__pyx_v_result = 0;
  double __pyx_v_min_intensity;
  double __pyx_v_diff;
  double __pyx_v_raw_charge;
  double __pyx_v_remain;
  int __pyx_v_charge;
  int __pyx_v_result_size;
  Py_ssize_t __pyx_v_j;
  size_t __pyx_v_i;
  size_t __pyx_v_n;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_result;
  __Pyx_Buffer __pyx_pybuffer_result;
  PyArrayObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  long __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  size_t __pyx_t_7;
  size_t __pyx_t_8;
  size_t __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyArrayObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_0_1quick_charge", 0);
  __pyx_pybuffer_result.pybuffer.buf = NULL;
  __pyx_pybuffer_result.refcount = 0;
  __pyx_pybuffernd_result.data = NULL;
  __pyx_pybuffernd_result.rcbuffer = &__pyx_pybuffer_result;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":773
 *         index = position_spec.peak_count
 *     else:
 *         index = position_spec             # <<<<<<<<<<<<<<
 *     if FittedPeakCollection is PeakSet:
 *         peaks = peak_set
 */
  __pyx_v_index = __pyx_v_position_spec;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":775
 *         index = position_spec
 *     if FittedPeakCollection is PeakSet:
 *         peaks = peak_set             # <<<<<<<<<<<<<<
 *     else:
 *         peaks = peak_set.peaks
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_peak_set));
  __pyx_v_peaks = __pyx_v_peak_set;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":778
 *     else:
 *         peaks = peak_set.peaks
 *     if index > peaks.get_size():             # <<<<<<<<<<<<<<
 *         if peaks.get_size() == 0:
 *             raise IndexError("peak_set cannot be empty!")
 */
  __pyx_t_1 = ((__pyx_v_index > ((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->get_size(__pyx_v_peaks)) != 0);
  if (__pyx_t_1) {

    /* "ms_deisotope/_c/deconvoluter_base.pyx":779
 *         peaks = peak_set.peaks
 *     if index > peaks.get_size():
 *         if peaks.get_size() == 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("peak_set cannot be empty!")
 *         # Assume we have a slice of the peak list, not the actual complete list, so
 */
    __pyx_t_1 = ((((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->get_size(__pyx_v_peaks) == 0) != 0);
    if (unlikely(__pyx_t_1)) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":780
 *     if index > peaks.get_size():
 *         if peaks.get_size() == 0:
 *             raise IndexError("peak_set cannot be empty!")             # <<<<<<<<<<<<<<
 *         # Assume we have a slice of the peak list, not the actual complete list, so
 *         # adjust the index offset from the first peak in the given list.
 */
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple__13, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 780, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 780, __pyx_L1_error)

      /* "ms_deisotope/_c/deconvoluter_base.pyx":779
 *         peaks = peak_set.peaks
 *     if index > peaks.get_size():
 *         if peaks.get_size() == 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("peak_set cannot be empty!")
 *         # Assume we have a slice of the peak list, not the actual complete list, so
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":783
 *         # Assume we have a slice of the peak list, not the actual complete list, so
 *         # adjust the index offset from the first peak in the given list.
 *         first_index = peaks.getitem(0).peak_count             # <<<<<<<<<<<<<<
 *         if (index - first_index) < peaks.get_size() and (index - first_index) >= 0:
 *             index -= first_index
 */
    __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->getitem(__pyx_v_peaks, 0)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 783, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_2)->peak_count;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_first_index = __pyx_t_3;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":784
 *         # adjust the index offset from the first peak in the given list.
 *         first_index = peaks.getitem(0).peak_count
 *         if (index - first_index) < peaks.get_size() and (index - first_index) >= 0:             # <<<<<<<<<<<<<<
 *             index -= first_index
 *         else:
 */
    __pyx_t_4 = (((__pyx_v_index - __pyx_v_first_index) < ((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->get_size(__pyx_v_peaks)) != 0);
    if (__pyx_t_4) {
    } else {
      __pyx_t_1 = __pyx_t_4;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_4 = (((__pyx_v_index - __pyx_v_first_index) >= 0) != 0);
    __pyx_t_1 = __pyx_t_4;
    __pyx_L6_bool_binop_done:;
    if (likely(__pyx_t_1)) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":785
 *         first_index = peaks.getitem(0).peak_count
 *         if (index - first_index) < peaks.get_size() and (index - first_index) >= 0:
 *             index -= first_index             # <<<<<<<<<<<<<<
 *         else:
 *             raise IndexError("%d is out of bounds for peak list of size %d in quick_charge" % (index, peaks.get_size()))
 */
      __pyx_v_index = (__pyx_v_index - __pyx_v_first_index);

      /* "ms_deisotope/_c/deconvoluter_base.pyx":784
 *         # adjust the index offset from the first peak in the given list.
 *         first_index = peaks.getitem(0).peak_count
 *         if (index - first_index) < peaks.get_size() and (index - first_index) >= 0:             # <<<<<<<<<<<<<<
 *             index -= first_index
 *         else:
 */
      goto __pyx_L5;
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":787
 *             index -= first_index
 *         else:
 *             raise IndexError("%d is out of bounds for peak list of size %d in quick_charge" % (index, peaks.get_size()))             # <<<<<<<<<<<<<<
 *     n = 1000
 *     result_size = 0
 */
    /*else*/ {
      __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_index); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 787, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_5 = __Pyx_PyInt_FromSize_t(((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->get_size(__pyx_v_peaks)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 787, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 787, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_5);
      __pyx_t_2 = 0;
      __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_d_is_out_of_bounds_for_peak_lis, __pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 787, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_IndexError, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 787, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_Raise(__pyx_t_6, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __PYX_ERR(0, 787, __pyx_L1_error)
    }
    __pyx_L5:;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":778
 *     else:
 *         peaks = peak_set.peaks
 *     if index > peaks.get_size():             # <<<<<<<<<<<<<<
 *         if peaks.get_size() == 0:
 *             raise IndexError("peak_set cannot be empty!")
 */
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":788
 *         else:
 *             raise IndexError("%d is out of bounds for peak list of size %d in quick_charge" % (index, peaks.get_size()))
 *     n = 1000             # <<<<<<<<<<<<<<
 *     result_size = 0
 *     min_intensity = peaks.getitem(index).intensity / 4.
 */
  __pyx_v_n = 0x3E8;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":789
 *             raise IndexError("%d is out of bounds for peak list of size %d in quick_charge" % (index, peaks.get_size()))
 *     n = 1000
 *     result_size = 0             # <<<<<<<<<<<<<<
 *     min_intensity = peaks.getitem(index).intensity / 4.
 *     for i in range(n):
 */
  __pyx_v_result_size = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":790
 *     n = 1000
 *     result_size = 0
 *     min_intensity = peaks.getitem(index).intensity / 4.             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         charges[i] = 0
 */
  __pyx_t_6 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->getitem(__pyx_v_peaks, __pyx_v_index)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 790, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_v_min_intensity = (((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_6)->__pyx_base.intensity / 4.);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":791
 *     result_size = 0
 *     min_intensity = peaks.getitem(index).intensity / 4.
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         charges[i] = 0
 *     for j in range(index + 1, (peaks.get_size())):
 */
  __pyx_t_7 = __pyx_v_n;
  __pyx_t_8 = __pyx_t_7;
  for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
    __pyx_v_i = __pyx_t_9;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":792
 *     min_intensity = peaks.getitem(index).intensity / 4.
 *     for i in range(n):
 *         charges[i] = 0             # <<<<<<<<<<<<<<
 *     for j in range(index + 1, (peaks.get_size())):
 *         if peaks.getitem(j).intensity < min_intensity:
 */
    (__pyx_v_charges[__pyx_v_i]) = 0;
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":793
 *     for i in range(n):
 *         charges[i] = 0
 *     for j in range(index + 1, (peaks.get_size())):             # <<<<<<<<<<<<<<
 *         if peaks.getitem(j).intensity < min_intensity:
 *             continue
 */
  __pyx_t_7 = ((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->get_size(__pyx_v_peaks);
  __pyx_t_8 = __pyx_t_7;
  for (__pyx_t_10 = (__pyx_v_index + 1); __pyx_t_10 < __pyx_t_8; __pyx_t_10+=1) {
    __pyx_v_j = __pyx_t_10;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":794
 *         charges[i] = 0
 *     for j in range(index + 1, (peaks.get_size())):
 *         if peaks.getitem(j).intensity < min_intensity:             # <<<<<<<<<<<<<<
 *             continue
 *         diff = peaks.getitem(j).mz - peaks.getitem(index).mz
 */
    __pyx_t_6 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->getitem(__pyx_v_peaks, __pyx_v_j)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 794, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = ((((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_6)->__pyx_base.intensity < __pyx_v_min_intensity) != 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (__pyx_t_1) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":795
 *     for j in range(index + 1, (peaks.get_size())):
 *         if peaks.getitem(j).intensity < min_intensity:
 *             continue             # <<<<<<<<<<<<<<
 *         diff = peaks.getitem(j).mz - peaks.getitem(index).mz
 *         if diff > 1.1:
 */
      goto __pyx_L10_continue;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":794
 *         charges[i] = 0
 *     for j in range(index + 1, (peaks.get_size())):
 *         if peaks.getitem(j).intensity < min_intensity:             # <<<<<<<<<<<<<<
 *             continue
 *         diff = peaks.getitem(j).mz - peaks.getitem(index).mz
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":796
 *         if peaks.getitem(j).intensity < min_intensity:
 *             continue
 *         diff = peaks.getitem(j).mz - peaks.getitem(index).mz             # <<<<<<<<<<<<<<
 *         if diff > 1.1:
 *             break
 */
    __pyx_t_6 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->getitem(__pyx_v_peaks, __pyx_v_j)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 796, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->getitem(__pyx_v_peaks, __pyx_v_index)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 796, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_v_diff = (((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_6)->__pyx_base.mz - ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_5)->__pyx_base.mz);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":797
 *             continue
 *         diff = peaks.getitem(j).mz - peaks.getitem(index).mz
 *         if diff > 1.1:             # <<<<<<<<<<<<<<
 *             break
 *         raw_charge = 1 / diff
 */
    __pyx_t_1 = ((__pyx_v_diff > 1.1) != 0);
    if (__pyx_t_1) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":798
 *         diff = peaks.getitem(j).mz - peaks.getitem(index).mz
 *         if diff > 1.1:
 *             break             # <<<<<<<<<<<<<<
 *         raw_charge = 1 / diff
 *         charge = <int>(raw_charge + 0.5)
 */
      goto __pyx_L11_break;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":797
 *             continue
 *         diff = peaks.getitem(j).mz - peaks.getitem(index).mz
 *         if diff > 1.1:             # <<<<<<<<<<<<<<
 *             break
 *         raw_charge = 1 / diff
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":799
 *         if diff > 1.1:
 *             break
 *         raw_charge = 1 / diff             # <<<<<<<<<<<<<<
 *         charge = <int>(raw_charge + 0.5)
 *         remain = raw_charge - <int>(raw_charge)
 */
    __pyx_v_raw_charge = (1.0 / __pyx_v_diff);

    /* "ms_deisotope/_c/deconvoluter_base.pyx":800
 *             break
 *         raw_charge = 1 / diff
 *         charge = <int>(raw_charge + 0.5)             # <<<<<<<<<<<<<<
 *         remain = raw_charge - <int>(raw_charge)
 *         if 0.2 < remain and remain < 0.8:
 */
    __pyx_v_charge = ((int)(__pyx_v_raw_charge + 0.5));

    /* "ms_deisotope/_c/deconvoluter_base.pyx":801
 *         raw_charge = 1 / diff
 *         charge = <int>(raw_charge + 0.5)
 *         remain = raw_charge - <int>(raw_charge)             # <<<<<<<<<<<<<<
 *         if 0.2 < remain and remain < 0.8:
 *             continue
 */
    __pyx_v_remain = (__pyx_v_raw_charge - ((int)__pyx_v_raw_charge));

    /* "ms_deisotope/_c/deconvoluter_base.pyx":802
 *         charge = <int>(raw_charge + 0.5)
 *         remain = raw_charge - <int>(raw_charge)
 *         if 0.2 < remain and remain < 0.8:             # <<<<<<<<<<<<<<
 *             continue
 *         if (charge < min_charge) or (charge > max_charge):
 */
    __pyx_t_4 = ((0.2 < __pyx_v_remain) != 0);
    if (__pyx_t_4) {
    } else {
      __pyx_t_1 = __pyx_t_4;
      goto __pyx_L15_bool_binop_done;
    }
    __pyx_t_4 = ((__pyx_v_remain < 0.8) != 0);
    __pyx_t_1 = __pyx_t_4;
    __pyx_L15_bool_binop_done:;
    if (__pyx_t_1) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":803
 *         remain = raw_charge - <int>(raw_charge)
 *         if 0.2 < remain and remain < 0.8:
 *             continue             # <<<<<<<<<<<<<<
 *         if (charge < min_charge) or (charge > max_charge):
 *             continue
 */
      goto __pyx_L10_continue;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":802
 *         charge = <int>(raw_charge + 0.5)
 *         remain = raw_charge - <int>(raw_charge)
 *         if 0.2 < remain and remain < 0.8:             # <<<<<<<<<<<<<<
 *             continue
 *         if (charge < min_charge) or (charge > max_charge):
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":804
 *         if 0.2 < remain and remain < 0.8:
 *             continue
 *         if (charge < min_charge) or (charge > max_charge):             # <<<<<<<<<<<<<<
 *             continue
 *         if charges[charge] == 0:
 */
    __pyx_t_4 = ((__pyx_v_charge < __pyx_v_min_charge) != 0);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_1 = __pyx_t_4;
      goto __pyx_L18_bool_binop_done;
    }
    __pyx_t_4 = ((__pyx_v_charge > __pyx_v_max_charge) != 0);
    __pyx_t_1 = __pyx_t_4;
    __pyx_L18_bool_binop_done:;
    if (__pyx_t_1) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":805
 *             continue
 *         if (charge < min_charge) or (charge > max_charge):
 *             continue             # <<<<<<<<<<<<<<
 *         if charges[charge] == 0:
 *             result_size += 1
 */
      goto __pyx_L10_continue;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":804
 *         if 0.2 < remain and remain < 0.8:
 *             continue
 *         if (charge < min_charge) or (charge > max_charge):             # <<<<<<<<<<<<<<
 *             continue
 *         if charges[charge] == 0:
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":806
 *         if (charge < min_charge) or (charge > max_charge):
 *             continue
 *         if charges[charge] == 0:             # <<<<<<<<<<<<<<
 *             result_size += 1
 *         charges[charge] = 1
 */
    __pyx_t_1 = (((__pyx_v_charges[__pyx_v_charge]) == 0) != 0);
    if (__pyx_t_1) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":807
 *             continue
 *         if charges[charge] == 0:
 *             result_size += 1             # <<<<<<<<<<<<<<
 *         charges[charge] = 1
 *     if result_size == 0:
 */
      __pyx_v_result_size = (__pyx_v_result_size + 1);

      /* "ms_deisotope/_c/deconvoluter_base.pyx":806
 *         if (charge < min_charge) or (charge > max_charge):
 *             continue
 *         if charges[charge] == 0:             # <<<<<<<<<<<<<<
 *             result_size += 1
 *         charges[charge] = 1
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":808
 *         if charges[charge] == 0:
 *             result_size += 1
 *         charges[charge] = 1             # <<<<<<<<<<<<<<
 *     if result_size == 0:
 *         result = np.empty(0, dtype=np.int32)
 */
    (__pyx_v_charges[__pyx_v_charge]) = 1;
    __pyx_L10_continue:;
  }
  __pyx_L11_break:;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":809
 *             result_size += 1
 *         charges[charge] = 1
 *     if result_size == 0:             # <<<<<<<<<<<<<<
 *         result = np.empty(0, dtype=np.int32)
 *         return result
 */
  __pyx_t_1 = ((__pyx_v_result_size == 0) != 0);
  if (__pyx_t_1) {

    /* "ms_deisotope/_c/deconvoluter_base.pyx":810
 *         charges[charge] = 1
 *     if result_size == 0:
 *         result = np.empty(0, dtype=np.int32)             # <<<<<<<<<<<<<<
 *         return result
 *     for j in range(index - 1, -1, -1):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 810, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_empty); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 810, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 810, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 810, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_int32); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 810, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_dtype, __pyx_t_11) < 0) __PYX_ERR(0, 810, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_tuple__14, __pyx_t_5); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 810, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (!(likely(((__pyx_t_11) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_11, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 810, __pyx_L1_error)
    __pyx_t_12 = ((PyArrayObject *)__pyx_t_11);
    {
      __Pyx_BufFmt_StackElem __pyx_stack[1];
      __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_result.rcbuffer->pybuffer);
      __pyx_t_13 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_result.rcbuffer->pybuffer, (PyObject*)__pyx_t_12, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack);
      if (unlikely(__pyx_t_13 < 0)) {
        PyErr_Fetch(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
        if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_result.rcbuffer->pybuffer, (PyObject*)__pyx_v_result, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {
          Py_XDECREF(__pyx_t_14); Py_XDECREF(__pyx_t_15); Py_XDECREF(__pyx_t_16);
          __Pyx_RaiseBufferFallbackError();
        } else {
          PyErr_Restore(__pyx_t_14, __pyx_t_15, __pyx_t_16);
        }
        __pyx_t_14 = __pyx_t_15 = __pyx_t_16 = 0;
      }
      __pyx_pybuffernd_result.diminfo[0].strides = __pyx_pybuffernd_result.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_result.diminfo[0].shape = __pyx_pybuffernd_result.rcbuffer->pybuffer.shape[0];
      if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 810, __pyx_L1_error)
    }
    __pyx_t_12 = 0;
    __pyx_v_result = ((PyArrayObject *)__pyx_t_11);
    __pyx_t_11 = 0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":811
 *     if result_size == 0:
 *         result = np.empty(0, dtype=np.int32)
 *         return result             # <<<<<<<<<<<<<<
 *     for j in range(index - 1, -1, -1):
 *         diff = peaks.getitem(index).mz - peaks.getitem(j).mz
 */
    __Pyx_XDECREF(((PyObject *)__pyx_r));
    __Pyx_INCREF(((PyObject *)__pyx_v_result));
    __pyx_r = ((PyArrayObject *)__pyx_v_result);
    goto __pyx_L0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":809
 *             result_size += 1
 *         charges[charge] = 1
 *     if result_size == 0:             # <<<<<<<<<<<<<<
 *         result = np.empty(0, dtype=np.int32)
 *         return result
 */
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":812
 *         result = np.empty(0, dtype=np.int32)
 *         return result
 *     for j in range(index - 1, -1, -1):             # <<<<<<<<<<<<<<
 *         diff = peaks.getitem(index).mz - peaks.getitem(j).mz
 *         if diff > 1.1:
 */
  for (__pyx_t_10 = (__pyx_v_index - 1); __pyx_t_10 > -1L; __pyx_t_10-=1) {
    __pyx_v_j = __pyx_t_10;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":813
 *         return result
 *     for j in range(index - 1, -1, -1):
 *         diff = peaks.getitem(index).mz - peaks.getitem(j).mz             # <<<<<<<<<<<<<<
 *         if diff > 1.1:
 *             break
 */
    __pyx_t_11 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->getitem(__pyx_v_peaks, __pyx_v_index)); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 813, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_5 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->getitem(__pyx_v_peaks, __pyx_v_j)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 813, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_v_diff = (((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_11)->__pyx_base.mz - ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_5)->__pyx_base.mz);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":814
 *     for j in range(index - 1, -1, -1):
 *         diff = peaks.getitem(index).mz - peaks.getitem(j).mz
 *         if diff > 1.1:             # <<<<<<<<<<<<<<
 *             break
 *         raw_charge = 1 / diff
 */
    __pyx_t_1 = ((__pyx_v_diff > 1.1) != 0);
    if (__pyx_t_1) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":815
 *         diff = peaks.getitem(index).mz - peaks.getitem(j).mz
 *         if diff > 1.1:
 *             break             # <<<<<<<<<<<<<<
 *         raw_charge = 1 / diff
 *         charge = <int>(raw_charge + 0.5)
 */
      goto __pyx_L23_break;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":814
 *     for j in range(index - 1, -1, -1):
 *         diff = peaks.getitem(index).mz - peaks.getitem(j).mz
 *         if diff > 1.1:             # <<<<<<<<<<<<<<
 *             break
 *         raw_charge = 1 / diff
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":816
 *         if diff > 1.1:
 *             break
 *         raw_charge = 1 / diff             # <<<<<<<<<<<<<<
 *         charge = <int>(raw_charge + 0.5)
 *         remain = raw_charge - <int>(raw_charge)
 */
    __pyx_v_raw_charge = (1.0 / __pyx_v_diff);

    /* "ms_deisotope/_c/deconvoluter_base.pyx":817
 *             break
 *         raw_charge = 1 / diff
 *         charge = <int>(raw_charge + 0.5)             # <<<<<<<<<<<<<<
 *         remain = raw_charge - <int>(raw_charge)
 *         if 0.2 < remain and remain < 0.8:
 */
    __pyx_v_charge = ((int)(__pyx_v_raw_charge + 0.5));

    /* "ms_deisotope/_c/deconvoluter_base.pyx":818
 *         raw_charge = 1 / diff
 *         charge = <int>(raw_charge + 0.5)
 *         remain = raw_charge - <int>(raw_charge)             # <<<<<<<<<<<<<<
 *         if 0.2 < remain and remain < 0.8:
 *             continue
 */
    __pyx_v_remain = (__pyx_v_raw_charge - ((int)__pyx_v_raw_charge));

    /* "ms_deisotope/_c/deconvoluter_base.pyx":819
 *         charge = <int>(raw_charge + 0.5)
 *         remain = raw_charge - <int>(raw_charge)
 *         if 0.2 < remain and remain < 0.8:             # <<<<<<<<<<<<<<
 *             continue
 *         if (charge < min_charge) or (charge > max_charge):
 */
    __pyx_t_4 = ((0.2 < __pyx_v_remain) != 0);
    if (__pyx_t_4) {
    } else {
      __pyx_t_1 = __pyx_t_4;
      goto __pyx_L26_bool_binop_done;
    }
    __pyx_t_4 = ((__pyx_v_remain < 0.8) != 0);
    __pyx_t_1 = __pyx_t_4;
    __pyx_L26_bool_binop_done:;
    if (__pyx_t_1) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":820
 *         remain = raw_charge - <int>(raw_charge)
 *         if 0.2 < remain and remain < 0.8:
 *             continue             # <<<<<<<<<<<<<<
 *         if (charge < min_charge) or (charge > max_charge):
 *             continue
 */
      goto __pyx_L22_continue;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":819
 *         charge = <int>(raw_charge + 0.5)
 *         remain = raw_charge - <int>(raw_charge)
 *         if 0.2 < remain and remain < 0.8:             # <<<<<<<<<<<<<<
 *             continue
 *         if (charge < min_charge) or (charge > max_charge):
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":821
 *         if 0.2 < remain and remain < 0.8:
 *             continue
 *         if (charge < min_charge) or (charge > max_charge):             # <<<<<<<<<<<<<<
 *             continue
 *         if charges[charge] == 0:
 */
    __pyx_t_4 = ((__pyx_v_charge < __pyx_v_min_charge) != 0);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_1 = __pyx_t_4;
      goto __pyx_L29_bool_binop_done;
    }
    __pyx_t_4 = ((__pyx_v_charge > __pyx_v_max_charge) != 0);
    __pyx_t_1 = __pyx_t_4;
    __pyx_L29_bool_binop_done:;
    if (__pyx_t_1) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":822
 *             continue
 *         if (charge < min_charge) or (charge > max_charge):
 *             continue             # <<<<<<<<<<<<<<
 *         if charges[charge] == 0:
 *             result_size += 1
 */
      goto __pyx_L22_continue;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":821
 *         if 0.2 < remain and remain < 0.8:
 *             continue
 *         if (charge < min_charge) or (charge > max_charge):             # <<<<<<<<<<<<<<
 *             continue
 *         if charges[charge] == 0:
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":823
 *         if (charge < min_charge) or (charge > max_charge):
 *             continue
 *         if charges[charge] == 0:             # <<<<<<<<<<<<<<
 *             result_size += 1
 *         charges[charge] = 1
 */
    __pyx_t_1 = (((__pyx_v_charges[__pyx_v_charge]) == 0) != 0);
    if (__pyx_t_1) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":824
 *             continue
 *         if charges[charge] == 0:
 *             result_size += 1             # <<<<<<<<<<<<<<
 *         charges[charge] = 1
 *     if result_size <= 0:
 */
      __pyx_v_result_size = (__pyx_v_result_size + 1);

      /* "ms_deisotope/_c/deconvoluter_base.pyx":823
 *         if (charge < min_charge) or (charge > max_charge):
 *             continue
 *         if charges[charge] == 0:             # <<<<<<<<<<<<<<
 *             result_size += 1
 *         charges[charge] = 1
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":825
 *         if charges[charge] == 0:
 *             result_size += 1
 *         charges[charge] = 1             # <<<<<<<<<<<<<<
 *     if result_size <= 0:
 *         return np.zeros(0, dtype=np.int32)
 */
    (__pyx_v_charges[__pyx_v_charge]) = 1;
    __pyx_L22_continue:;
  }
  __pyx_L23_break:;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":826
 *             result_size += 1
 *         charges[charge] = 1
 *     if result_size <= 0:             # <<<<<<<<<<<<<<
 *         return np.zeros(0, dtype=np.int32)
 *     result = np.zeros(result_size, dtype=np.int32)
 */
  __pyx_t_1 = ((__pyx_v_result_size <= 0) != 0);
  if (__pyx_t_1) {

    /* "ms_deisotope/_c/deconvoluter_base.pyx":827
 *         charges[charge] = 1
 *     if result_size <= 0:
 *         return np.zeros(0, dtype=np.int32)             # <<<<<<<<<<<<<<
 *     result = np.zeros(result_size, dtype=np.int32)
 * 
 */
    __Pyx_XDECREF(((PyObject *)__pyx_r));
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 827, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_zeros); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 827, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 827, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 827, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_int32); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 827, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_dtype, __pyx_t_2) < 0) __PYX_ERR(0, 827, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_tuple__14, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 827, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 827, __pyx_L1_error)
    __pyx_r = ((PyArrayObject *)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":826
 *             result_size += 1
 *         charges[charge] = 1
 *     if result_size <= 0:             # <<<<<<<<<<<<<<
 *         return np.zeros(0, dtype=np.int32)
 *     result = np.zeros(result_size, dtype=np.int32)
 */
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":828
 *     if result_size <= 0:
 *         return np.zeros(0, dtype=np.int32)
 *     result = np.zeros(result_size, dtype=np.int32)             # <<<<<<<<<<<<<<
 * 
 *     i = 0
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_result_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_11 = PyTuple_New(1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_int32); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_dtype, __pyx_t_17) < 0) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  __pyx_t_17 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_11, __pyx_t_2); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(((__pyx_t_17) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_17, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 828, __pyx_L1_error)
  __pyx_t_12 = ((PyArrayObject *)__pyx_t_17);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_result.rcbuffer->pybuffer);
    __pyx_t_13 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_result.rcbuffer->pybuffer, (PyObject*)__pyx_t_12, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_13 < 0)) {
      PyErr_Fetch(&__pyx_t_16, &__pyx_t_15, &__pyx_t_14);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_result.rcbuffer->pybuffer, (PyObject*)__pyx_v_result, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_16); Py_XDECREF(__pyx_t_15); Py_XDECREF(__pyx_t_14);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_16, __pyx_t_15, __pyx_t_14);
      }
      __pyx_t_16 = __pyx_t_15 = __pyx_t_14 = 0;
    }
    __pyx_pybuffernd_result.diminfo[0].strides = __pyx_pybuffernd_result.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_result.diminfo[0].shape = __pyx_pybuffernd_result.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 828, __pyx_L1_error)
  }
  __pyx_t_12 = 0;
  __pyx_v_result = ((PyArrayObject *)__pyx_t_17);
  __pyx_t_17 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":830
 *     result = np.zeros(result_size, dtype=np.int32)
 * 
 *     i = 0             # <<<<<<<<<<<<<<
 *     for j in range(n):
 *         if charges[j] != 0:
 */
  __pyx_v_i = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":831
 * 
 *     i = 0
 *     for j in range(n):             # <<<<<<<<<<<<<<
 *         if charges[j] != 0:
 *             result[i] = j
 */
  __pyx_t_7 = __pyx_v_n;
  __pyx_t_8 = __pyx_t_7;
  for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_8; __pyx_t_10+=1) {
    __pyx_v_j = __pyx_t_10;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":832
 *     i = 0
 *     for j in range(n):
 *         if charges[j] != 0:             # <<<<<<<<<<<<<<
 *             result[i] = j
 *             i += 1
 */
    __pyx_t_1 = (((__pyx_v_charges[__pyx_v_j]) != 0) != 0);
    if (__pyx_t_1) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":833
 *     for j in range(n):
 *         if charges[j] != 0:
 *             result[i] = j             # <<<<<<<<<<<<<<
 *             i += 1
 *     return result
 */
      __pyx_t_9 = __pyx_v_i;
      __pyx_t_13 = -1;
      if (unlikely(__pyx_t_9 >= (size_t)__pyx_pybuffernd_result.diminfo[0].shape)) __pyx_t_13 = 0;
      if (unlikely(__pyx_t_13 != -1)) {
        __Pyx_RaiseBufferIndexError(__pyx_t_13);
        __PYX_ERR(0, 833, __pyx_L1_error)
      }
      *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int32_t *, __pyx_pybuffernd_result.rcbuffer->pybuffer.buf, __pyx_t_9, __pyx_pybuffernd_result.diminfo[0].strides) = __pyx_v_j;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":834
 *         if charges[j] != 0:
 *             result[i] = j
 *             i += 1             # <<<<<<<<<<<<<<
 *     return result
 * 
 */
      __pyx_v_i = (__pyx_v_i + 1);

      /* "ms_deisotope/_c/deconvoluter_base.pyx":832
 *     i = 0
 *     for j in range(n):
 *         if charges[j] != 0:             # <<<<<<<<<<<<<<
 *             result[i] = j
 *             i += 1
 */
    }
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":835
 *             result[i] = j
 *             i += 1
 *     return result             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_result));
  __pyx_r = ((PyArrayObject *)__pyx_v_result);
  goto __pyx_L0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":730
 * 
 * @cython.cdivision
 * cpdef np.ndarray[np.int32_t, ndim=1] quick_charge(FittedPeakCollection peak_set, FittedPeakOrPosition position_spec, int min_charge, int max_charge):             # <<<<<<<<<<<<<<
 *     """An implementation of Hoopman's QuickCharge [1] algorithm for quickly capping charge
 *     state queries
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_17);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_result.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.quick_charge", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_result.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_peaks);
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_19__pyx_fuse_0_1quick_charge(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_0_1__pyx_mdef_12ms_deisotope_2_c_17deconvoluter_base_19__pyx_fuse_0_1quick_charge = {"__pyx_fuse_0_1quick_charge", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_19__pyx_fuse_0_1quick_charge, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_quick_charge};
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_19__pyx_fuse_0_1quick_charge(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peak_set = 0;
  size_t __pyx_v_position_spec;
  int __pyx_v_min_charge;
  int __pyx_v_max_charge;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_0_1quick_charge (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_peak_set,&__pyx_n_s_position_spec,&__pyx_n_s_min_charge,&__pyx_n_s_max_charge,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peak_set)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_position_spec)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0_1quick_charge", 1, 4, 4, 1); __PYX_ERR(0, 730, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_min_charge)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0_1quick_charge", 1, 4, 4, 2); __PYX_ERR(0, 730, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_max_charge)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0_1quick_charge", 1, 4, 4, 3); __PYX_ERR(0, 730, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_0_1quick_charge") < 0)) __PYX_ERR(0, 730, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_peak_set = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)values[0]);
    __pyx_v_position_spec = __Pyx_PyInt_As_size_t(values[1]); if (unlikely((__pyx_v_position_spec == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 730, __pyx_L3_error)
    __pyx_v_min_charge = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_min_charge == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 730, __pyx_L3_error)
    __pyx_v_max_charge = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_max_charge == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 730, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0_1quick_charge", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 730, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.__pyx_fuse_0_1quick_charge", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peak_set), __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet, 1, "peak_set", 0))) __PYX_ERR(0, 730, __pyx_L1_error)
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_18__pyx_fuse_0_1quick_charge(__pyx_self, __pyx_v_peak_set, __pyx_v_position_spec, __pyx_v_min_charge, __pyx_v_max_charge);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_18__pyx_fuse_0_1quick_charge(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peak_set, size_t __pyx_v_position_spec, int __pyx_v_min_charge, int __pyx_v_max_charge) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_0_1quick_charge", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_fuse_0_1__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_quick_charge(__pyx_v_peak_set, __pyx_v_position_spec, __pyx_v_min_charge, __pyx_v_max_charge, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 730, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.__pyx_fuse_0_1quick_charge", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_21__pyx_fuse_0_2quick_charge(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_1quick_charge(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyArrayObject *__pyx_fuse_0_2__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_quick_charge(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peak_set, long __pyx_v_position_spec, int __pyx_v_min_charge, int __pyx_v_max_charge, CYTHON_UNUSED int __pyx_skip_dispatch) {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaks = 0;
  size_t __pyx_v_index;
  size_t __pyx_v_first_index;
  int __pyx_v_charges[0x3E8];
  PyArrayObject *__pyx_v_result = 0;
  double __pyx_v_min_intensity;
  double __pyx_v_diff;
  double __pyx_v_raw_charge;
  double __pyx_v_remain;
  int __pyx_v_charge;
  int __pyx_v_result_size;
  Py_ssize_t __pyx_v_j;
  size_t __pyx_v_i;
  size_t __pyx_v_n;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_result;
  __Pyx_Buffer __pyx_pybuffer_result;
  PyArrayObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  long __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  size_t __pyx_t_7;
  size_t __pyx_t_8;
  size_t __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyArrayObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_0_2quick_charge", 0);
  __pyx_pybuffer_result.pybuffer.buf = NULL;
  __pyx_pybuffer_result.refcount = 0;
  __pyx_pybuffernd_result.data = NULL;
  __pyx_pybuffernd_result.rcbuffer = &__pyx_pybuffer_result;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":773
 *         index = position_spec.peak_count
 *     else:
 *         index = position_spec             # <<<<<<<<<<<<<<
 *     if FittedPeakCollection is PeakSet:
 *         peaks = peak_set
 */
  __pyx_v_index = __pyx_v_position_spec;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":775
 *         index = position_spec
 *     if FittedPeakCollection is PeakSet:
 *         peaks = peak_set             # <<<<<<<<<<<<<<
 *     else:
 *         peaks = peak_set.peaks
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_peak_set));
  __pyx_v_peaks = __pyx_v_peak_set;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":778
 *     else:
 *         peaks = peak_set.peaks
 *     if index > peaks.get_size():             # <<<<<<<<<<<<<<
 *         if peaks.get_size() == 0:
 *             raise IndexError("peak_set cannot be empty!")
 */
  __pyx_t_1 = ((__pyx_v_index > ((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->get_size(__pyx_v_peaks)) != 0);
  if (__pyx_t_1) {

    /* "ms_deisotope/_c/deconvoluter_base.pyx":779
 *         peaks = peak_set.peaks
 *     if index > peaks.get_size():
 *         if peaks.get_size() == 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("peak_set cannot be empty!")
 *         # Assume we have a slice of the peak list, not the actual complete list, so
 */
    __pyx_t_1 = ((((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->get_size(__pyx_v_peaks) == 0) != 0);
    if (unlikely(__pyx_t_1)) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":780
 *     if index > peaks.get_size():
 *         if peaks.get_size() == 0:
 *             raise IndexError("peak_set cannot be empty!")             # <<<<<<<<<<<<<<
 *         # Assume we have a slice of the peak list, not the actual complete list, so
 *         # adjust the index offset from the first peak in the given list.
 */
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple__13, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 780, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 780, __pyx_L1_error)

      /* "ms_deisotope/_c/deconvoluter_base.pyx":779
 *         peaks = peak_set.peaks
 *     if index > peaks.get_size():
 *         if peaks.get_size() == 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("peak_set cannot be empty!")
 *         # Assume we have a slice of the peak list, not the actual complete list, so
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":783
 *         # Assume we have a slice of the peak list, not the actual complete list, so
 *         # adjust the index offset from the first peak in the given list.
 *         first_index = peaks.getitem(0).peak_count             # <<<<<<<<<<<<<<
 *         if (index - first_index) < peaks.get_size() and (index - first_index) >= 0:
 *             index -= first_index
 */
    __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->getitem(__pyx_v_peaks, 0)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 783, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_2)->peak_count;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_first_index = __pyx_t_3;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":784
 *         # adjust the index offset from the first peak in the given list.
 *         first_index = peaks.getitem(0).peak_count
 *         if (index - first_index) < peaks.get_size() and (index - first_index) >= 0:             # <<<<<<<<<<<<<<
 *             index -= first_index
 *         else:
 */
    __pyx_t_4 = (((__pyx_v_index - __pyx_v_first_index) < ((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->get_size(__pyx_v_peaks)) != 0);
    if (__pyx_t_4) {
    } else {
      __pyx_t_1 = __pyx_t_4;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_4 = (((__pyx_v_index - __pyx_v_first_index) >= 0) != 0);
    __pyx_t_1 = __pyx_t_4;
    __pyx_L6_bool_binop_done:;
    if (likely(__pyx_t_1)) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":785
 *         first_index = peaks.getitem(0).peak_count
 *         if (index - first_index) < peaks.get_size() and (index - first_index) >= 0:
 *             index -= first_index             # <<<<<<<<<<<<<<
 *         else:
 *             raise IndexError("%d is out of bounds for peak list of size %d in quick_charge" % (index, peaks.get_size()))
 */
      __pyx_v_index = (__pyx_v_index - __pyx_v_first_index);

      /* "ms_deisotope/_c/deconvoluter_base.pyx":784
 *         # adjust the index offset from the first peak in the given list.
 *         first_index = peaks.getitem(0).peak_count
 *         if (index - first_index) < peaks.get_size() and (index - first_index) >= 0:             # <<<<<<<<<<<<<<
 *             index -= first_index
 *         else:
 */
      goto __pyx_L5;
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":787
 *             index -= first_index
 *         else:
 *             raise IndexError("%d is out of bounds for peak list of size %d in quick_charge" % (index, peaks.get_size()))             # <<<<<<<<<<<<<<
 *     n = 1000
 *     result_size = 0
 */
    /*else*/ {
      __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_index); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 787, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_5 = __Pyx_PyInt_FromSize_t(((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->get_size(__pyx_v_peaks)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 787, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 787, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_5);
      __pyx_t_2 = 0;
      __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_d_is_out_of_bounds_for_peak_lis, __pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 787, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_IndexError, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 787, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_Raise(__pyx_t_6, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __PYX_ERR(0, 787, __pyx_L1_error)
    }
    __pyx_L5:;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":778
 *     else:
 *         peaks = peak_set.peaks
 *     if index > peaks.get_size():             # <<<<<<<<<<<<<<
 *         if peaks.get_size() == 0:
 *             raise IndexError("peak_set cannot be empty!")
 */
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":788
 *         else:
 *             raise IndexError("%d is out of bounds for peak list of size %d in quick_charge" % (index, peaks.get_size()))
 *     n = 1000             # <<<<<<<<<<<<<<
 *     result_size = 0
 *     min_intensity = peaks.getitem(index).intensity / 4.
 */
  __pyx_v_n = 0x3E8;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":789
 *             raise IndexError("%d is out of bounds for peak list of size %d in quick_charge" % (index, peaks.get_size()))
 *     n = 1000
 *     result_size = 0             # <<<<<<<<<<<<<<
 *     min_intensity = peaks.getitem(index).intensity / 4.
 *     for i in range(n):
 */
  __pyx_v_result_size = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":790
 *     n = 1000
 *     result_size = 0
 *     min_intensity = peaks.getitem(index).intensity / 4.             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         charges[i] = 0
 */
  __pyx_t_6 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->getitem(__pyx_v_peaks, __pyx_v_index)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 790, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_v_min_intensity = (((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_6)->__pyx_base.intensity / 4.);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":791
 *     result_size = 0
 *     min_intensity = peaks.getitem(index).intensity / 4.
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         charges[i] = 0
 *     for j in range(index + 1, (peaks.get_size())):
 */
  __pyx_t_7 = __pyx_v_n;
  __pyx_t_8 = __pyx_t_7;
  for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
    __pyx_v_i = __pyx_t_9;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":792
 *     min_intensity = peaks.getitem(index).intensity / 4.
 *     for i in range(n):
 *         charges[i] = 0             # <<<<<<<<<<<<<<
 *     for j in range(index + 1, (peaks.get_size())):
 *         if peaks.getitem(j).intensity < min_intensity:
 */
    (__pyx_v_charges[__pyx_v_i]) = 0;
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":793
 *     for i in range(n):
 *         charges[i] = 0
 *     for j in range(index + 1, (peaks.get_size())):             # <<<<<<<<<<<<<<
 *         if peaks.getitem(j).intensity < min_intensity:
 *             continue
 */
  __pyx_t_7 = ((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->get_size(__pyx_v_peaks);
  __pyx_t_8 = __pyx_t_7;
  for (__pyx_t_10 = (__pyx_v_index + 1); __pyx_t_10 < __pyx_t_8; __pyx_t_10+=1) {
    __pyx_v_j = __pyx_t_10;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":794
 *         charges[i] = 0
 *     for j in range(index + 1, (peaks.get_size())):
 *         if peaks.getitem(j).intensity < min_intensity:             # <<<<<<<<<<<<<<
 *             continue
 *         diff = peaks.getitem(j).mz - peaks.getitem(index).mz
 */
    __pyx_t_6 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->getitem(__pyx_v_peaks, __pyx_v_j)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 794, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = ((((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_6)->__pyx_base.intensity < __pyx_v_min_intensity) != 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (__pyx_t_1) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":795
 *     for j in range(index + 1, (peaks.get_size())):
 *         if peaks.getitem(j).intensity < min_intensity:
 *             continue             # <<<<<<<<<<<<<<
 *         diff = peaks.getitem(j).mz - peaks.getitem(index).mz
 *         if diff > 1.1:
 */
      goto __pyx_L10_continue;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":794
 *         charges[i] = 0
 *     for j in range(index + 1, (peaks.get_size())):
 *         if peaks.getitem(j).intensity < min_intensity:             # <<<<<<<<<<<<<<
 *             continue
 *         diff = peaks.getitem(j).mz - peaks.getitem(index).mz
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":796
 *         if peaks.getitem(j).intensity < min_intensity:
 *             continue
 *         diff = peaks.getitem(j).mz - peaks.getitem(index).mz             # <<<<<<<<<<<<<<
 *         if diff > 1.1:
 *             break
 */
    __pyx_t_6 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->getitem(__pyx_v_peaks, __pyx_v_j)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 796, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->getitem(__pyx_v_peaks, __pyx_v_index)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 796, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_v_diff = (((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_6)->__pyx_base.mz - ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_5)->__pyx_base.mz);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":797
 *             continue
 *         diff = peaks.getitem(j).mz - peaks.getitem(index).mz
 *         if diff > 1.1:             # <<<<<<<<<<<<<<
 *             break
 *         raw_charge = 1 / diff
 */
    __pyx_t_1 = ((__pyx_v_diff > 1.1) != 0);
    if (__pyx_t_1) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":798
 *         diff = peaks.getitem(j).mz - peaks.getitem(index).mz
 *         if diff > 1.1:
 *             break             # <<<<<<<<<<<<<<
 *         raw_charge = 1 / diff
 *         charge = <int>(raw_charge + 0.5)
 */
      goto __pyx_L11_break;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":797
 *             continue
 *         diff = peaks.getitem(j).mz - peaks.getitem(index).mz
 *         if diff > 1.1:             # <<<<<<<<<<<<<<
 *             break
 *         raw_charge = 1 / diff
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":799
 *         if diff > 1.1:
 *             break
 *         raw_charge = 1 / diff             # <<<<<<<<<<<<<<
 *         charge = <int>(raw_charge + 0.5)
 *         remain = raw_charge - <int>(raw_charge)
 */
    __pyx_v_raw_charge = (1.0 / __pyx_v_diff);

    /* "ms_deisotope/_c/deconvoluter_base.pyx":800
 *             break
 *         raw_charge = 1 / diff
 *         charge = <int>(raw_charge + 0.5)             # <<<<<<<<<<<<<<
 *         remain = raw_charge - <int>(raw_charge)
 *         if 0.2 < remain and remain < 0.8:
 */
    __pyx_v_charge = ((int)(__pyx_v_raw_charge + 0.5));

    /* "ms_deisotope/_c/deconvoluter_base.pyx":801
 *         raw_charge = 1 / diff
 *         charge = <int>(raw_charge + 0.5)
 *         remain = raw_charge - <int>(raw_charge)             # <<<<<<<<<<<<<<
 *         if 0.2 < remain and remain < 0.8:
 *             continue
 */
    __pyx_v_remain = (__pyx_v_raw_charge - ((int)__pyx_v_raw_charge));

    /* "ms_deisotope/_c/deconvoluter_base.pyx":802
 *         charge = <int>(raw_charge + 0.5)
 *         remain = raw_charge - <int>(raw_charge)
 *         if 0.2 < remain and remain < 0.8:             # <<<<<<<<<<<<<<
 *             continue
 *         if (charge < min_charge) or (charge > max_charge):
 */
    __pyx_t_4 = ((0.2 < __pyx_v_remain) != 0);
    if (__pyx_t_4) {
    } else {
      __pyx_t_1 = __pyx_t_4;
      goto __pyx_L15_bool_binop_done;
    }
    __pyx_t_4 = ((__pyx_v_remain < 0.8) != 0);
    __pyx_t_1 = __pyx_t_4;
    __pyx_L15_bool_binop_done:;
    if (__pyx_t_1) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":803
 *         remain = raw_charge - <int>(raw_charge)
 *         if 0.2 < remain and remain < 0.8:
 *             continue             # <<<<<<<<<<<<<<
 *         if (charge < min_charge) or (charge > max_charge):
 *             continue
 */
      goto __pyx_L10_continue;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":802
 *         charge = <int>(raw_charge + 0.5)
 *         remain = raw_charge - <int>(raw_charge)
 *         if 0.2 < remain and remain < 0.8:             # <<<<<<<<<<<<<<
 *             continue
 *         if (charge < min_charge) or (charge > max_charge):
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":804
 *         if 0.2 < remain and remain < 0.8:
 *             continue
 *         if (charge < min_charge) or (charge > max_charge):             # <<<<<<<<<<<<<<
 *             continue
 *         if charges[charge] == 0:
 */
    __pyx_t_4 = ((__pyx_v_charge < __pyx_v_min_charge) != 0);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_1 = __pyx_t_4;
      goto __pyx_L18_bool_binop_done;
    }
    __pyx_t_4 = ((__pyx_v_charge > __pyx_v_max_charge) != 0);
    __pyx_t_1 = __pyx_t_4;
    __pyx_L18_bool_binop_done:;
    if (__pyx_t_1) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":805
 *             continue
 *         if (charge < min_charge) or (charge > max_charge):
 *             continue             # <<<<<<<<<<<<<<
 *         if charges[charge] == 0:
 *             result_size += 1
 */
      goto __pyx_L10_continue;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":804
 *         if 0.2 < remain and remain < 0.8:
 *             continue
 *         if (charge < min_charge) or (charge > max_charge):             # <<<<<<<<<<<<<<
 *             continue
 *         if charges[charge] == 0:
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":806
 *         if (charge < min_charge) or (charge > max_charge):
 *             continue
 *         if charges[charge] == 0:             # <<<<<<<<<<<<<<
 *             result_size += 1
 *         charges[charge] = 1
 */
    __pyx_t_1 = (((__pyx_v_charges[__pyx_v_charge]) == 0) != 0);
    if (__pyx_t_1) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":807
 *             continue
 *         if charges[charge] == 0:
 *             result_size += 1             # <<<<<<<<<<<<<<
 *         charges[charge] = 1
 *     if result_size == 0:
 */
      __pyx_v_result_size = (__pyx_v_result_size + 1);

      /* "ms_deisotope/_c/deconvoluter_base.pyx":806
 *         if (charge < min_charge) or (charge > max_charge):
 *             continue
 *         if charges[charge] == 0:             # <<<<<<<<<<<<<<
 *             result_size += 1
 *         charges[charge] = 1
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":808
 *         if charges[charge] == 0:
 *             result_size += 1
 *         charges[charge] = 1             # <<<<<<<<<<<<<<
 *     if result_size == 0:
 *         result = np.empty(0, dtype=np.int32)
 */
    (__pyx_v_charges[__pyx_v_charge]) = 1;
    __pyx_L10_continue:;
  }
  __pyx_L11_break:;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":809
 *             result_size += 1
 *         charges[charge] = 1
 *     if result_size == 0:             # <<<<<<<<<<<<<<
 *         result = np.empty(0, dtype=np.int32)
 *         return result
 */
  __pyx_t_1 = ((__pyx_v_result_size == 0) != 0);
  if (__pyx_t_1) {

    /* "ms_deisotope/_c/deconvoluter_base.pyx":810
 *         charges[charge] = 1
 *     if result_size == 0:
 *         result = np.empty(0, dtype=np.int32)             # <<<<<<<<<<<<<<
 *         return result
 *     for j in range(index - 1, -1, -1):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 810, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_empty); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 810, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 810, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 810, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_int32); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 810, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_dtype, __pyx_t_11) < 0) __PYX_ERR(0, 810, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_tuple__14, __pyx_t_5); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 810, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (!(likely(((__pyx_t_11) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_11, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 810, __pyx_L1_error)
    __pyx_t_12 = ((PyArrayObject *)__pyx_t_11);
    {
      __Pyx_BufFmt_StackElem __pyx_stack[1];
      __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_result.rcbuffer->pybuffer);
      __pyx_t_13 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_result.rcbuffer->pybuffer, (PyObject*)__pyx_t_12, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack);
      if (unlikely(__pyx_t_13 < 0)) {
        PyErr_Fetch(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
        if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_result.rcbuffer->pybuffer, (PyObject*)__pyx_v_result, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {
          Py_XDECREF(__pyx_t_14); Py_XDECREF(__pyx_t_15); Py_XDECREF(__pyx_t_16);
          __Pyx_RaiseBufferFallbackError();
        } else {
          PyErr_Restore(__pyx_t_14, __pyx_t_15, __pyx_t_16);
        }
        __pyx_t_14 = __pyx_t_15 = __pyx_t_16 = 0;
      }
      __pyx_pybuffernd_result.diminfo[0].strides = __pyx_pybuffernd_result.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_result.diminfo[0].shape = __pyx_pybuffernd_result.rcbuffer->pybuffer.shape[0];
      if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 810, __pyx_L1_error)
    }
    __pyx_t_12 = 0;
    __pyx_v_result = ((PyArrayObject *)__pyx_t_11);
    __pyx_t_11 = 0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":811
 *     if result_size == 0:
 *         result = np.empty(0, dtype=np.int32)
 *         return result             # <<<<<<<<<<<<<<
 *     for j in range(index - 1, -1, -1):
 *         diff = peaks.getitem(index).mz - peaks.getitem(j).mz
 */
    __Pyx_XDECREF(((PyObject *)__pyx_r));
    __Pyx_INCREF(((PyObject *)__pyx_v_result));
    __pyx_r = ((PyArrayObject *)__pyx_v_result);
    goto __pyx_L0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":809
 *             result_size += 1
 *         charges[charge] = 1
 *     if result_size == 0:             # <<<<<<<<<<<<<<
 *         result = np.empty(0, dtype=np.int32)
 *         return result
 */
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":812
 *         result = np.empty(0, dtype=np.int32)
 *         return result
 *     for j in range(index - 1, -1, -1):             # <<<<<<<<<<<<<<
 *         diff = peaks.getitem(index).mz - peaks.getitem(j).mz
 *         if diff > 1.1:
 */
  for (__pyx_t_10 = (__pyx_v_index - 1); __pyx_t_10 > -1L; __pyx_t_10-=1) {
    __pyx_v_j = __pyx_t_10;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":813
 *         return result
 *     for j in range(index - 1, -1, -1):
 *         diff = peaks.getitem(index).mz - peaks.getitem(j).mz             # <<<<<<<<<<<<<<
 *         if diff > 1.1:
 *             break
 */
    __pyx_t_11 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->getitem(__pyx_v_peaks, __pyx_v_index)); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 813, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_5 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->getitem(__pyx_v_peaks, __pyx_v_j)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 813, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_v_diff = (((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_11)->__pyx_base.mz - ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_5)->__pyx_base.mz);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":814
 *     for j in range(index - 1, -1, -1):
 *         diff = peaks.getitem(index).mz - peaks.getitem(j).mz
 *         if diff > 1.1:             # <<<<<<<<<<<<<<
 *             break
 *         raw_charge = 1 / diff
 */
    __pyx_t_1 = ((__pyx_v_diff > 1.1) != 0);
    if (__pyx_t_1) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":815
 *         diff = peaks.getitem(index).mz - peaks.getitem(j).mz
 *         if diff > 1.1:
 *             break             # <<<<<<<<<<<<<<
 *         raw_charge = 1 / diff
 *         charge = <int>(raw_charge + 0.5)
 */
      goto __pyx_L23_break;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":814
 *     for j in range(index - 1, -1, -1):
 *         diff = peaks.getitem(index).mz - peaks.getitem(j).mz
 *         if diff > 1.1:             # <<<<<<<<<<<<<<
 *             break
 *         raw_charge = 1 / diff
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":816
 *         if diff > 1.1:
 *             break
 *         raw_charge = 1 / diff             # <<<<<<<<<<<<<<
 *         charge = <int>(raw_charge + 0.5)
 *         remain = raw_charge - <int>(raw_charge)
 */
    __pyx_v_raw_charge = (1.0 / __pyx_v_diff);

    /* "ms_deisotope/_c/deconvoluter_base.pyx":817
 *             break
 *         raw_charge = 1 / diff
 *         charge = <int>(raw_charge + 0.5)             # <<<<<<<<<<<<<<
 *         remain = raw_charge - <int>(raw_charge)
 *         if 0.2 < remain and remain < 0.8:
 */
    __pyx_v_charge = ((int)(__pyx_v_raw_charge + 0.5));

    /* "ms_deisotope/_c/deconvoluter_base.pyx":818
 *         raw_charge = 1 / diff
 *         charge = <int>(raw_charge + 0.5)
 *         remain = raw_charge - <int>(raw_charge)             # <<<<<<<<<<<<<<
 *         if 0.2 < remain and remain < 0.8:
 *             continue
 */
    __pyx_v_remain = (__pyx_v_raw_charge - ((int)__pyx_v_raw_charge));

    /* "ms_deisotope/_c/deconvoluter_base.pyx":819
 *         charge = <int>(raw_charge + 0.5)
 *         remain = raw_charge - <int>(raw_charge)
 *         if 0.2 < remain and remain < 0.8:             # <<<<<<<<<<<<<<
 *             continue
 *         if (charge < min_charge) or (charge > max_charge):
 */
    __pyx_t_4 = ((0.2 < __pyx_v_remain) != 0);
    if (__pyx_t_4) {
    } else {
      __pyx_t_1 = __pyx_t_4;
      goto __pyx_L26_bool_binop_done;
    }
    __pyx_t_4 = ((__pyx_v_remain < 0.8) != 0);
    __pyx_t_1 = __pyx_t_4;
    __pyx_L26_bool_binop_done:;
    if (__pyx_t_1) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":820
 *         remain = raw_charge - <int>(raw_charge)
 *         if 0.2 < remain and remain < 0.8:
 *             continue             # <<<<<<<<<<<<<<
 *         if (charge < min_charge) or (charge > max_charge):
 *             continue
 */
      goto __pyx_L22_continue;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":819
 *         charge = <int>(raw_charge + 0.5)
 *         remain = raw_charge - <int>(raw_charge)
 *         if 0.2 < remain and remain < 0.8:             # <<<<<<<<<<<<<<
 *             continue
 *         if (charge < min_charge) or (charge > max_charge):
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":821
 *         if 0.2 < remain and remain < 0.8:
 *             continue
 *         if (charge < min_charge) or (charge > max_charge):             # <<<<<<<<<<<<<<
 *             continue
 *         if charges[charge] == 0:
 */
    __pyx_t_4 = ((__pyx_v_charge < __pyx_v_min_charge) != 0);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_1 = __pyx_t_4;
      goto __pyx_L29_bool_binop_done;
    }
    __pyx_t_4 = ((__pyx_v_charge > __pyx_v_max_charge) != 0);
    __pyx_t_1 = __pyx_t_4;
    __pyx_L29_bool_binop_done:;
    if (__pyx_t_1) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":822
 *             continue
 *         if (charge < min_charge) or (charge > max_charge):
 *             continue             # <<<<<<<<<<<<<<
 *         if charges[charge] == 0:
 *             result_size += 1
 */
      goto __pyx_L22_continue;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":821
 *         if 0.2 < remain and remain < 0.8:
 *             continue
 *         if (charge < min_charge) or (charge > max_charge):             # <<<<<<<<<<<<<<
 *             continue
 *         if charges[charge] == 0:
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":823
 *         if (charge < min_charge) or (charge > max_charge):
 *             continue
 *         if charges[charge] == 0:             # <<<<<<<<<<<<<<
 *             result_size += 1
 *         charges[charge] = 1
 */
    __pyx_t_1 = (((__pyx_v_charges[__pyx_v_charge]) == 0) != 0);
    if (__pyx_t_1) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":824
 *             continue
 *         if charges[charge] == 0:
 *             result_size += 1             # <<<<<<<<<<<<<<
 *         charges[charge] = 1
 *     if result_size <= 0:
 */
      __pyx_v_result_size = (__pyx_v_result_size + 1);

      /* "ms_deisotope/_c/deconvoluter_base.pyx":823
 *         if (charge < min_charge) or (charge > max_charge):
 *             continue
 *         if charges[charge] == 0:             # <<<<<<<<<<<<<<
 *             result_size += 1
 *         charges[charge] = 1
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":825
 *         if charges[charge] == 0:
 *             result_size += 1
 *         charges[charge] = 1             # <<<<<<<<<<<<<<
 *     if result_size <= 0:
 *         return np.zeros(0, dtype=np.int32)
 */
    (__pyx_v_charges[__pyx_v_charge]) = 1;
    __pyx_L22_continue:;
  }
  __pyx_L23_break:;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":826
 *             result_size += 1
 *         charges[charge] = 1
 *     if result_size <= 0:             # <<<<<<<<<<<<<<
 *         return np.zeros(0, dtype=np.int32)
 *     result = np.zeros(result_size, dtype=np.int32)
 */
  __pyx_t_1 = ((__pyx_v_result_size <= 0) != 0);
  if (__pyx_t_1) {

    /* "ms_deisotope/_c/deconvoluter_base.pyx":827
 *         charges[charge] = 1
 *     if result_size <= 0:
 *         return np.zeros(0, dtype=np.int32)             # <<<<<<<<<<<<<<
 *     result = np.zeros(result_size, dtype=np.int32)
 * 
 */
    __Pyx_XDECREF(((PyObject *)__pyx_r));
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 827, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_zeros); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 827, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 827, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 827, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_int32); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 827, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_dtype, __pyx_t_2) < 0) __PYX_ERR(0, 827, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_tuple__14, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 827, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 827, __pyx_L1_error)
    __pyx_r = ((PyArrayObject *)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":826
 *             result_size += 1
 *         charges[charge] = 1
 *     if result_size <= 0:             # <<<<<<<<<<<<<<
 *         return np.zeros(0, dtype=np.int32)
 *     result = np.zeros(result_size, dtype=np.int32)
 */
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":828
 *     if result_size <= 0:
 *         return np.zeros(0, dtype=np.int32)
 *     result = np.zeros(result_size, dtype=np.int32)             # <<<<<<<<<<<<<<
 * 
 *     i = 0
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_result_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_11 = PyTuple_New(1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_int32); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_dtype, __pyx_t_17) < 0) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  __pyx_t_17 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_11, __pyx_t_2); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(((__pyx_t_17) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_17, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 828, __pyx_L1_error)
  __pyx_t_12 = ((PyArrayObject *)__pyx_t_17);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_result.rcbuffer->pybuffer);
    __pyx_t_13 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_result.rcbuffer->pybuffer, (PyObject*)__pyx_t_12, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_13 < 0)) {
      PyErr_Fetch(&__pyx_t_16, &__pyx_t_15, &__pyx_t_14);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_result.rcbuffer->pybuffer, (PyObject*)__pyx_v_result, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_16); Py_XDECREF(__pyx_t_15); Py_XDECREF(__pyx_t_14);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_16, __pyx_t_15, __pyx_t_14);
      }
      __pyx_t_16 = __pyx_t_15 = __pyx_t_14 = 0;
    }
    __pyx_pybuffernd_result.diminfo[0].strides = __pyx_pybuffernd_result.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_result.diminfo[0].shape = __pyx_pybuffernd_result.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 828, __pyx_L1_error)
  }
  __pyx_t_12 = 0;
  __pyx_v_result = ((PyArrayObject *)__pyx_t_17);
  __pyx_t_17 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":830
 *     result = np.zeros(result_size, dtype=np.int32)
 * 
 *     i = 0             # <<<<<<<<<<<<<<
 *     for j in range(n):
 *         if charges[j] != 0:
 */
  __pyx_v_i = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":831
 * 
 *     i = 0
 *     for j in range(n):             # <<<<<<<<<<<<<<
 *         if charges[j] != 0:
 *             result[i] = j
 */
  __pyx_t_7 = __pyx_v_n;
  __pyx_t_8 = __pyx_t_7;
  for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_8; __pyx_t_10+=1) {
    __pyx_v_j = __pyx_t_10;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":832
 *     i = 0
 *     for j in range(n):
 *         if charges[j] != 0:             # <<<<<<<<<<<<<<
 *             result[i] = j
 *             i += 1
 */
    __pyx_t_1 = (((__pyx_v_charges[__pyx_v_j]) != 0) != 0);
    if (__pyx_t_1) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":833
 *     for j in range(n):
 *         if charges[j] != 0:
 *             result[i] = j             # <<<<<<<<<<<<<<
 *             i += 1
 *     return result
 */
      __pyx_t_9 = __pyx_v_i;
      __pyx_t_13 = -1;
      if (unlikely(__pyx_t_9 >= (size_t)__pyx_pybuffernd_result.diminfo[0].shape)) __pyx_t_13 = 0;
      if (unlikely(__pyx_t_13 != -1)) {
        __Pyx_RaiseBufferIndexError(__pyx_t_13);
        __PYX_ERR(0, 833, __pyx_L1_error)
      }
      *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int32_t *, __pyx_pybuffernd_result.rcbuffer->pybuffer.buf, __pyx_t_9, __pyx_pybuffernd_result.diminfo[0].strides) = __pyx_v_j;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":834
 *         if charges[j] != 0:
 *             result[i] = j
 *             i += 1             # <<<<<<<<<<<<<<
 *     return result
 * 
 */
      __pyx_v_i = (__pyx_v_i + 1);

      /* "ms_deisotope/_c/deconvoluter_base.pyx":832
 *     i = 0
 *     for j in range(n):
 *         if charges[j] != 0:             # <<<<<<<<<<<<<<
 *             result[i] = j
 *             i += 1
 */
    }
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":835
 *             result[i] = j
 *             i += 1
 *     return result             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_result));
  __pyx_r = ((PyArrayObject *)__pyx_v_result);
  goto __pyx_L0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":730
 * 
 * @cython.cdivision
 * cpdef np.ndarray[np.int32_t, ndim=1] quick_charge(FittedPeakCollection peak_set, FittedPeakOrPosition position_spec, int min_charge, int max_charge):             # <<<<<<<<<<<<<<
 *     """An implementation of Hoopman's QuickCharge [1] algorithm for quickly capping charge
 *     state queries
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_17);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_result.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.quick_charge", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_result.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_peaks);
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_21__pyx_fuse_0_2quick_charge(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_0_2__pyx_mdef_12ms_deisotope_2_c_17deconvoluter_base_21__pyx_fuse_0_2quick_charge = {"__pyx_fuse_0_2quick_charge", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_21__pyx_fuse_0_2quick_charge, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_quick_charge};
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_21__pyx_fuse_0_2quick_charge(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peak_set = 0;
  long __pyx_v_position_spec;
  int __pyx_v_min_charge;
  int __pyx_v_max_charge;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_0_2quick_charge (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_peak_set,&__pyx_n_s_position_spec,&__pyx_n_s_min_charge,&__pyx_n_s_max_charge,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peak_set)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_position_spec)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0_2quick_charge", 1, 4, 4, 1); __PYX_ERR(0, 730, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_min_charge)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0_2quick_charge", 1, 4, 4, 2); __PYX_ERR(0, 730, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_max_charge)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0_2quick_charge", 1, 4, 4, 3); __PYX_ERR(0, 730, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_0_2quick_charge") < 0)) __PYX_ERR(0, 730, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_peak_set = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)values[0]);
    __pyx_v_position_spec = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v_position_spec == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 730, __pyx_L3_error)
    __pyx_v_min_charge = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_min_charge == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 730, __pyx_L3_error)
    __pyx_v_max_charge = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_max_charge == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 730, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0_2quick_charge", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 730, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.__pyx_fuse_0_2quick_charge", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peak_set), __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet, 1, "peak_set", 0))) __PYX_ERR(0, 730, __pyx_L1_error)
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_20__pyx_fuse_0_2quick_charge(__pyx_self, __pyx_v_peak_set, __pyx_v_position_spec, __pyx_v_min_charge, __pyx_v_max_charge);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_20__pyx_fuse_0_2quick_charge(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peak_set, long __pyx_v_position_spec, int __pyx_v_min_charge, int __pyx_v_max_charge) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_0_2quick_charge", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_fuse_0_2__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_quick_charge(__pyx_v_peak_set, __pyx_v_position_spec, __pyx_v_min_charge, __pyx_v_max_charge, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 730, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.__pyx_fuse_0_2quick_charge", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_23__pyx_fuse_1_0quick_charge(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_1quick_charge(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyArrayObject *__pyx_fuse_1_0__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_quick_charge(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_v_peak_set, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_position_spec, int __pyx_v_min_charge, int __pyx_v_max_charge, CYTHON_UNUSED int __pyx_skip_dispatch) {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaks = 0;
  size_t __pyx_v_index;
  size_t __pyx_v_first_index;
  int __pyx_v_charges[0x3E8];
  PyArrayObject *__pyx_v_result = 0;
  double __pyx_v_min_intensity;
  double __pyx_v_diff;
  double __pyx_v_raw_charge;
  double __pyx_v_remain;
  int __pyx_v_charge;
  int __pyx_v_result_size;
  Py_ssize_t __pyx_v_j;
  size_t __pyx_v_i;
  size_t __pyx_v_n;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_result;
  __Pyx_Buffer __pyx_pybuffer_result;
  PyArrayObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  long __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  size_t __pyx_t_7;
  size_t __pyx_t_8;
  size_t __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyArrayObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_1_0quick_charge", 0);
  __pyx_pybuffer_result.pybuffer.buf = NULL;
  __pyx_pybuffer_result.refcount = 0;
  __pyx_pybuffernd_result.data = NULL;
  __pyx_pybuffernd_result.rcbuffer = &__pyx_pybuffer_result;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":771
 *         bint matched
 *     if FittedPeakOrPosition is FittedPeak:
 *         index = position_spec.peak_count             # <<<<<<<<<<<<<<
 *     else:
 *         index = position_spec
 */
  __pyx_t_1 = __pyx_v_position_spec->peak_count;
  __pyx_v_index = __pyx_t_1;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":777
 *         peaks = peak_set
 *     else:
 *         peaks = peak_set.peaks             # <<<<<<<<<<<<<<
 *     if index > peaks.get_size():
 *         if peaks.get_size() == 0:
 */
  __pyx_t_2 = ((PyObject *)__pyx_v_peak_set->peaks);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_peaks = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":778
 *     else:
 *         peaks = peak_set.peaks
 *     if index > peaks.get_size():             # <<<<<<<<<<<<<<
 *         if peaks.get_size() == 0:
 *             raise IndexError("peak_set cannot be empty!")
 */
  __pyx_t_3 = ((__pyx_v_index > ((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->get_size(__pyx_v_peaks)) != 0);
  if (__pyx_t_3) {

    /* "ms_deisotope/_c/deconvoluter_base.pyx":779
 *         peaks = peak_set.peaks
 *     if index > peaks.get_size():
 *         if peaks.get_size() == 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("peak_set cannot be empty!")
 *         # Assume we have a slice of the peak list, not the actual complete list, so
 */
    __pyx_t_3 = ((((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->get_size(__pyx_v_peaks) == 0) != 0);
    if (unlikely(__pyx_t_3)) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":780
 *     if index > peaks.get_size():
 *         if peaks.get_size() == 0:
 *             raise IndexError("peak_set cannot be empty!")             # <<<<<<<<<<<<<<
 *         # Assume we have a slice of the peak list, not the actual complete list, so
 *         # adjust the index offset from the first peak in the given list.
 */
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple__13, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 780, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 780, __pyx_L1_error)

      /* "ms_deisotope/_c/deconvoluter_base.pyx":779
 *         peaks = peak_set.peaks
 *     if index > peaks.get_size():
 *         if peaks.get_size() == 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("peak_set cannot be empty!")
 *         # Assume we have a slice of the peak list, not the actual complete list, so
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":783
 *         # Assume we have a slice of the peak list, not the actual complete list, so
 *         # adjust the index offset from the first peak in the given list.
 *         first_index = peaks.getitem(0).peak_count             # <<<<<<<<<<<<<<
 *         if (index - first_index) < peaks.get_size() and (index - first_index) >= 0:
 *             index -= first_index
 */
    __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->getitem(__pyx_v_peaks, 0)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 783, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_2)->peak_count;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_first_index = __pyx_t_1;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":784
 *         # adjust the index offset from the first peak in the given list.
 *         first_index = peaks.getitem(0).peak_count
 *         if (index - first_index) < peaks.get_size() and (index - first_index) >= 0:             # <<<<<<<<<<<<<<
 *             index -= first_index
 *         else:
 */
    __pyx_t_4 = (((__pyx_v_index - __pyx_v_first_index) < ((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->get_size(__pyx_v_peaks)) != 0);
    if (__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_4 = (((__pyx_v_index - __pyx_v_first_index) >= 0) != 0);
    __pyx_t_3 = __pyx_t_4;
    __pyx_L6_bool_binop_done:;
    if (likely(__pyx_t_3)) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":785
 *         first_index = peaks.getitem(0).peak_count
 *         if (index - first_index) < peaks.get_size() and (index - first_index) >= 0:
 *             index -= first_index             # <<<<<<<<<<<<<<
 *         else:
 *             raise IndexError("%d is out of bounds for peak list of size %d in quick_charge" % (index, peaks.get_size()))
 */
      __pyx_v_index = (__pyx_v_index - __pyx_v_first_index);

      /* "ms_deisotope/_c/deconvoluter_base.pyx":784
 *         # adjust the index offset from the first peak in the given list.
 *         first_index = peaks.getitem(0).peak_count
 *         if (index - first_index) < peaks.get_size() and (index - first_index) >= 0:             # <<<<<<<<<<<<<<
 *             index -= first_index
 *         else:
 */
      goto __pyx_L5;
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":787
 *             index -= first_index
 *         else:
 *             raise IndexError("%d is out of bounds for peak list of size %d in quick_charge" % (index, peaks.get_size()))             # <<<<<<<<<<<<<<
 *     n = 1000
 *     result_size = 0
 */
    /*else*/ {
      __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_index); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 787, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_5 = __Pyx_PyInt_FromSize_t(((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->get_size(__pyx_v_peaks)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 787, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 787, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_5);
      __pyx_t_2 = 0;
      __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_d_is_out_of_bounds_for_peak_lis, __pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 787, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_IndexError, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 787, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_Raise(__pyx_t_6, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __PYX_ERR(0, 787, __pyx_L1_error)
    }
    __pyx_L5:;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":778
 *     else:
 *         peaks = peak_set.peaks
 *     if index > peaks.get_size():             # <<<<<<<<<<<<<<
 *         if peaks.get_size() == 0:
 *             raise IndexError("peak_set cannot be empty!")
 */
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":788
 *         else:
 *             raise IndexError("%d is out of bounds for peak list of size %d in quick_charge" % (index, peaks.get_size()))
 *     n = 1000             # <<<<<<<<<<<<<<
 *     result_size = 0
 *     min_intensity = peaks.getitem(index).intensity / 4.
 */
  __pyx_v_n = 0x3E8;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":789
 *             raise IndexError("%d is out of bounds for peak list of size %d in quick_charge" % (index, peaks.get_size()))
 *     n = 1000
 *     result_size = 0             # <<<<<<<<<<<<<<
 *     min_intensity = peaks.getitem(index).intensity / 4.
 *     for i in range(n):
 */
  __pyx_v_result_size = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":790
 *     n = 1000
 *     result_size = 0
 *     min_intensity = peaks.getitem(index).intensity / 4.             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         charges[i] = 0
 */
  __pyx_t_6 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->getitem(__pyx_v_peaks, __pyx_v_index)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 790, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_v_min_intensity = (((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_6)->__pyx_base.intensity / 4.);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":791
 *     result_size = 0
 *     min_intensity = peaks.getitem(index).intensity / 4.
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         charges[i] = 0
 *     for j in range(index + 1, (peaks.get_size())):
 */
  __pyx_t_7 = __pyx_v_n;
  __pyx_t_8 = __pyx_t_7;
  for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
    __pyx_v_i = __pyx_t_9;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":792
 *     min_intensity = peaks.getitem(index).intensity / 4.
 *     for i in range(n):
 *         charges[i] = 0             # <<<<<<<<<<<<<<
 *     for j in range(index + 1, (peaks.get_size())):
 *         if peaks.getitem(j).intensity < min_intensity:
 */
    (__pyx_v_charges[__pyx_v_i]) = 0;
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":793
 *     for i in range(n):
 *         charges[i] = 0
 *     for j in range(index + 1, (peaks.get_size())):             # <<<<<<<<<<<<<<
 *         if peaks.getitem(j).intensity < min_intensity:
 *             continue
 */
  __pyx_t_7 = ((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->get_size(__pyx_v_peaks);
  __pyx_t_8 = __pyx_t_7;
  for (__pyx_t_10 = (__pyx_v_index + 1); __pyx_t_10 < __pyx_t_8; __pyx_t_10+=1) {
    __pyx_v_j = __pyx_t_10;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":794
 *         charges[i] = 0
 *     for j in range(index + 1, (peaks.get_size())):
 *         if peaks.getitem(j).intensity < min_intensity:             # <<<<<<<<<<<<<<
 *             continue
 *         diff = peaks.getitem(j).mz - peaks.getitem(index).mz
 */
    __pyx_t_6 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->getitem(__pyx_v_peaks, __pyx_v_j)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 794, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_3 = ((((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_6)->__pyx_base.intensity < __pyx_v_min_intensity) != 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (__pyx_t_3) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":795
 *     for j in range(index + 1, (peaks.get_size())):
 *         if peaks.getitem(j).intensity < min_intensity:
 *             continue             # <<<<<<<<<<<<<<
 *         diff = peaks.getitem(j).mz - peaks.getitem(index).mz
 *         if diff > 1.1:
 */
      goto __pyx_L10_continue;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":794
 *         charges[i] = 0
 *     for j in range(index + 1, (peaks.get_size())):
 *         if peaks.getitem(j).intensity < min_intensity:             # <<<<<<<<<<<<<<
 *             continue
 *         diff = peaks.getitem(j).mz - peaks.getitem(index).mz
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":796
 *         if peaks.getitem(j).intensity < min_intensity:
 *             continue
 *         diff = peaks.getitem(j).mz - peaks.getitem(index).mz             # <<<<<<<<<<<<<<
 *         if diff > 1.1:
 *             break
 */
    __pyx_t_6 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->getitem(__pyx_v_peaks, __pyx_v_j)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 796, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->getitem(__pyx_v_peaks, __pyx_v_index)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 796, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_v_diff = (((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_6)->__pyx_base.mz - ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_5)->__pyx_base.mz);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":797
 *             continue
 *         diff = peaks.getitem(j).mz - peaks.getitem(index).mz
 *         if diff > 1.1:             # <<<<<<<<<<<<<<
 *             break
 *         raw_charge = 1 / diff
 */
    __pyx_t_3 = ((__pyx_v_diff > 1.1) != 0);
    if (__pyx_t_3) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":798
 *         diff = peaks.getitem(j).mz - peaks.getitem(index).mz
 *         if diff > 1.1:
 *             break             # <<<<<<<<<<<<<<
 *         raw_charge = 1 / diff
 *         charge = <int>(raw_charge + 0.5)
 */
      goto __pyx_L11_break;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":797
 *             continue
 *         diff = peaks.getitem(j).mz - peaks.getitem(index).mz
 *         if diff > 1.1:             # <<<<<<<<<<<<<<
 *             break
 *         raw_charge = 1 / diff
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":799
 *         if diff > 1.1:
 *             break
 *         raw_charge = 1 / diff             # <<<<<<<<<<<<<<
 *         charge = <int>(raw_charge + 0.5)
 *         remain = raw_charge - <int>(raw_charge)
 */
    __pyx_v_raw_charge = (1.0 / __pyx_v_diff);

    /* "ms_deisotope/_c/deconvoluter_base.pyx":800
 *             break
 *         raw_charge = 1 / diff
 *         charge = <int>(raw_charge + 0.5)             # <<<<<<<<<<<<<<
 *         remain = raw_charge - <int>(raw_charge)
 *         if 0.2 < remain and remain < 0.8:
 */
    __pyx_v_charge = ((int)(__pyx_v_raw_charge + 0.5));

    /* "ms_deisotope/_c/deconvoluter_base.pyx":801
 *         raw_charge = 1 / diff
 *         charge = <int>(raw_charge + 0.5)
 *         remain = raw_charge - <int>(raw_charge)             # <<<<<<<<<<<<<<
 *         if 0.2 < remain and remain < 0.8:
 *             continue
 */
    __pyx_v_remain = (__pyx_v_raw_charge - ((int)__pyx_v_raw_charge));

    /* "ms_deisotope/_c/deconvoluter_base.pyx":802
 *         charge = <int>(raw_charge + 0.5)
 *         remain = raw_charge - <int>(raw_charge)
 *         if 0.2 < remain and remain < 0.8:             # <<<<<<<<<<<<<<
 *             continue
 *         if (charge < min_charge) or (charge > max_charge):
 */
    __pyx_t_4 = ((0.2 < __pyx_v_remain) != 0);
    if (__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L15_bool_binop_done;
    }
    __pyx_t_4 = ((__pyx_v_remain < 0.8) != 0);
    __pyx_t_3 = __pyx_t_4;
    __pyx_L15_bool_binop_done:;
    if (__pyx_t_3) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":803
 *         remain = raw_charge - <int>(raw_charge)
 *         if 0.2 < remain and remain < 0.8:
 *             continue             # <<<<<<<<<<<<<<
 *         if (charge < min_charge) or (charge > max_charge):
 *             continue
 */
      goto __pyx_L10_continue;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":802
 *         charge = <int>(raw_charge + 0.5)
 *         remain = raw_charge - <int>(raw_charge)
 *         if 0.2 < remain and remain < 0.8:             # <<<<<<<<<<<<<<
 *             continue
 *         if (charge < min_charge) or (charge > max_charge):
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":804
 *         if 0.2 < remain and remain < 0.8:
 *             continue
 *         if (charge < min_charge) or (charge > max_charge):             # <<<<<<<<<<<<<<
 *             continue
 *         if charges[charge] == 0:
 */
    __pyx_t_4 = ((__pyx_v_charge < __pyx_v_min_charge) != 0);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L18_bool_binop_done;
    }
    __pyx_t_4 = ((__pyx_v_charge > __pyx_v_max_charge) != 0);
    __pyx_t_3 = __pyx_t_4;
    __pyx_L18_bool_binop_done:;
    if (__pyx_t_3) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":805
 *             continue
 *         if (charge < min_charge) or (charge > max_charge):
 *             continue             # <<<<<<<<<<<<<<
 *         if charges[charge] == 0:
 *             result_size += 1
 */
      goto __pyx_L10_continue;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":804
 *         if 0.2 < remain and remain < 0.8:
 *             continue
 *         if (charge < min_charge) or (charge > max_charge):             # <<<<<<<<<<<<<<
 *             continue
 *         if charges[charge] == 0:
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":806
 *         if (charge < min_charge) or (charge > max_charge):
 *             continue
 *         if charges[charge] == 0:             # <<<<<<<<<<<<<<
 *             result_size += 1
 *         charges[charge] = 1
 */
    __pyx_t_3 = (((__pyx_v_charges[__pyx_v_charge]) == 0) != 0);
    if (__pyx_t_3) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":807
 *             continue
 *         if charges[charge] == 0:
 *             result_size += 1             # <<<<<<<<<<<<<<
 *         charges[charge] = 1
 *     if result_size == 0:
 */
      __pyx_v_result_size = (__pyx_v_result_size + 1);

      /* "ms_deisotope/_c/deconvoluter_base.pyx":806
 *         if (charge < min_charge) or (charge > max_charge):
 *             continue
 *         if charges[charge] == 0:             # <<<<<<<<<<<<<<
 *             result_size += 1
 *         charges[charge] = 1
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":808
 *         if charges[charge] == 0:
 *             result_size += 1
 *         charges[charge] = 1             # <<<<<<<<<<<<<<
 *     if result_size == 0:
 *         result = np.empty(0, dtype=np.int32)
 */
    (__pyx_v_charges[__pyx_v_charge]) = 1;
    __pyx_L10_continue:;
  }
  __pyx_L11_break:;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":809
 *             result_size += 1
 *         charges[charge] = 1
 *     if result_size == 0:             # <<<<<<<<<<<<<<
 *         result = np.empty(0, dtype=np.int32)
 *         return result
 */
  __pyx_t_3 = ((__pyx_v_result_size == 0) != 0);
  if (__pyx_t_3) {

    /* "ms_deisotope/_c/deconvoluter_base.pyx":810
 *         charges[charge] = 1
 *     if result_size == 0:
 *         result = np.empty(0, dtype=np.int32)             # <<<<<<<<<<<<<<
 *         return result
 *     for j in range(index - 1, -1, -1):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 810, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_empty); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 810, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 810, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 810, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_int32); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 810, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_dtype, __pyx_t_11) < 0) __PYX_ERR(0, 810, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_tuple__14, __pyx_t_5); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 810, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (!(likely(((__pyx_t_11) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_11, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 810, __pyx_L1_error)
    __pyx_t_12 = ((PyArrayObject *)__pyx_t_11);
    {
      __Pyx_BufFmt_StackElem __pyx_stack[1];
      __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_result.rcbuffer->pybuffer);
      __pyx_t_13 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_result.rcbuffer->pybuffer, (PyObject*)__pyx_t_12, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack);
      if (unlikely(__pyx_t_13 < 0)) {
        PyErr_Fetch(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
        if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_result.rcbuffer->pybuffer, (PyObject*)__pyx_v_result, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {
          Py_XDECREF(__pyx_t_14); Py_XDECREF(__pyx_t_15); Py_XDECREF(__pyx_t_16);
          __Pyx_RaiseBufferFallbackError();
        } else {
          PyErr_Restore(__pyx_t_14, __pyx_t_15, __pyx_t_16);
        }
        __pyx_t_14 = __pyx_t_15 = __pyx_t_16 = 0;
      }
      __pyx_pybuffernd_result.diminfo[0].strides = __pyx_pybuffernd_result.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_result.diminfo[0].shape = __pyx_pybuffernd_result.rcbuffer->pybuffer.shape[0];
      if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 810, __pyx_L1_error)
    }
    __pyx_t_12 = 0;
    __pyx_v_result = ((PyArrayObject *)__pyx_t_11);
    __pyx_t_11 = 0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":811
 *     if result_size == 0:
 *         result = np.empty(0, dtype=np.int32)
 *         return result             # <<<<<<<<<<<<<<
 *     for j in range(index - 1, -1, -1):
 *         diff = peaks.getitem(index).mz - peaks.getitem(j).mz
 */
    __Pyx_XDECREF(((PyObject *)__pyx_r));
    __Pyx_INCREF(((PyObject *)__pyx_v_result));
    __pyx_r = ((PyArrayObject *)__pyx_v_result);
    goto __pyx_L0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":809
 *             result_size += 1
 *         charges[charge] = 1
 *     if result_size == 0:             # <<<<<<<<<<<<<<
 *         result = np.empty(0, dtype=np.int32)
 *         return result
 */
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":812
 *         result = np.empty(0, dtype=np.int32)
 *         return result
 *     for j in range(index - 1, -1, -1):             # <<<<<<<<<<<<<<
 *         diff = peaks.getitem(index).mz - peaks.getitem(j).mz
 *         if diff > 1.1:
 */
  for (__pyx_t_10 = (__pyx_v_index - 1); __pyx_t_10 > -1L; __pyx_t_10-=1) {
    __pyx_v_j = __pyx_t_10;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":813
 *         return result
 *     for j in range(index - 1, -1, -1):
 *         diff = peaks.getitem(index).mz - peaks.getitem(j).mz             # <<<<<<<<<<<<<<
 *         if diff > 1.1:
 *             break
 */
    __pyx_t_11 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->getitem(__pyx_v_peaks, __pyx_v_index)); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 813, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_5 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->getitem(__pyx_v_peaks, __pyx_v_j)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 813, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_v_diff = (((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_11)->__pyx_base.mz - ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_5)->__pyx_base.mz);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":814
 *     for j in range(index - 1, -1, -1):
 *         diff = peaks.getitem(index).mz - peaks.getitem(j).mz
 *         if diff > 1.1:             # <<<<<<<<<<<<<<
 *             break
 *         raw_charge = 1 / diff
 */
    __pyx_t_3 = ((__pyx_v_diff > 1.1) != 0);
    if (__pyx_t_3) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":815
 *         diff = peaks.getitem(index).mz - peaks.getitem(j).mz
 *         if diff > 1.1:
 *             break             # <<<<<<<<<<<<<<
 *         raw_charge = 1 / diff
 *         charge = <int>(raw_charge + 0.5)
 */
      goto __pyx_L23_break;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":814
 *     for j in range(index - 1, -1, -1):
 *         diff = peaks.getitem(index).mz - peaks.getitem(j).mz
 *         if diff > 1.1:             # <<<<<<<<<<<<<<
 *             break
 *         raw_charge = 1 / diff
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":816
 *         if diff > 1.1:
 *             break
 *         raw_charge = 1 / diff             # <<<<<<<<<<<<<<
 *         charge = <int>(raw_charge + 0.5)
 *         remain = raw_charge - <int>(raw_charge)
 */
    __pyx_v_raw_charge = (1.0 / __pyx_v_diff);

    /* "ms_deisotope/_c/deconvoluter_base.pyx":817
 *             break
 *         raw_charge = 1 / diff
 *         charge = <int>(raw_charge + 0.5)             # <<<<<<<<<<<<<<
 *         remain = raw_charge - <int>(raw_charge)
 *         if 0.2 < remain and remain < 0.8:
 */
    __pyx_v_charge = ((int)(__pyx_v_raw_charge + 0.5));

    /* "ms_deisotope/_c/deconvoluter_base.pyx":818
 *         raw_charge = 1 / diff
 *         charge = <int>(raw_charge + 0.5)
 *         remain = raw_charge - <int>(raw_charge)             # <<<<<<<<<<<<<<
 *         if 0.2 < remain and remain < 0.8:
 *             continue
 */
    __pyx_v_remain = (__pyx_v_raw_charge - ((int)__pyx_v_raw_charge));

    /* "ms_deisotope/_c/deconvoluter_base.pyx":819
 *         charge = <int>(raw_charge + 0.5)
 *         remain = raw_charge - <int>(raw_charge)
 *         if 0.2 < remain and remain < 0.8:             # <<<<<<<<<<<<<<
 *             continue
 *         if (charge < min_charge) or (charge > max_charge):
 */
    __pyx_t_4 = ((0.2 < __pyx_v_remain) != 0);
    if (__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L26_bool_binop_done;
    }
    __pyx_t_4 = ((__pyx_v_remain < 0.8) != 0);
    __pyx_t_3 = __pyx_t_4;
    __pyx_L26_bool_binop_done:;
    if (__pyx_t_3) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":820
 *         remain = raw_charge - <int>(raw_charge)
 *         if 0.2 < remain and remain < 0.8:
 *             continue             # <<<<<<<<<<<<<<
 *         if (charge < min_charge) or (charge > max_charge):
 *             continue
 */
      goto __pyx_L22_continue;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":819
 *         charge = <int>(raw_charge + 0.5)
 *         remain = raw_charge - <int>(raw_charge)
 *         if 0.2 < remain and remain < 0.8:             # <<<<<<<<<<<<<<
 *             continue
 *         if (charge < min_charge) or (charge > max_charge):
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":821
 *         if 0.2 < remain and remain < 0.8:
 *             continue
 *         if (charge < min_charge) or (charge > max_charge):             # <<<<<<<<<<<<<<
 *             continue
 *         if charges[charge] == 0:
 */
    __pyx_t_4 = ((__pyx_v_charge < __pyx_v_min_charge) != 0);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L29_bool_binop_done;
    }
    __pyx_t_4 = ((__pyx_v_charge > __pyx_v_max_charge) != 0);
    __pyx_t_3 = __pyx_t_4;
    __pyx_L29_bool_binop_done:;
    if (__pyx_t_3) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":822
 *             continue
 *         if (charge < min_charge) or (charge > max_charge):
 *             continue             # <<<<<<<<<<<<<<
 *         if charges[charge] == 0:
 *             result_size += 1
 */
      goto __pyx_L22_continue;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":821
 *         if 0.2 < remain and remain < 0.8:
 *             continue
 *         if (charge < min_charge) or (charge > max_charge):             # <<<<<<<<<<<<<<
 *             continue
 *         if charges[charge] == 0:
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":823
 *         if (charge < min_charge) or (charge > max_charge):
 *             continue
 *         if charges[charge] == 0:             # <<<<<<<<<<<<<<
 *             result_size += 1
 *         charges[charge] = 1
 */
    __pyx_t_3 = (((__pyx_v_charges[__pyx_v_charge]) == 0) != 0);
    if (__pyx_t_3) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":824
 *             continue
 *         if charges[charge] == 0:
 *             result_size += 1             # <<<<<<<<<<<<<<
 *         charges[charge] = 1
 *     if result_size <= 0:
 */
      __pyx_v_result_size = (__pyx_v_result_size + 1);

      /* "ms_deisotope/_c/deconvoluter_base.pyx":823
 *         if (charge < min_charge) or (charge > max_charge):
 *             continue
 *         if charges[charge] == 0:             # <<<<<<<<<<<<<<
 *             result_size += 1
 *         charges[charge] = 1
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":825
 *         if charges[charge] == 0:
 *             result_size += 1
 *         charges[charge] = 1             # <<<<<<<<<<<<<<
 *     if result_size <= 0:
 *         return np.zeros(0, dtype=np.int32)
 */
    (__pyx_v_charges[__pyx_v_charge]) = 1;
    __pyx_L22_continue:;
  }
  __pyx_L23_break:;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":826
 *             result_size += 1
 *         charges[charge] = 1
 *     if result_size <= 0:             # <<<<<<<<<<<<<<
 *         return np.zeros(0, dtype=np.int32)
 *     result = np.zeros(result_size, dtype=np.int32)
 */
  __pyx_t_3 = ((__pyx_v_result_size <= 0) != 0);
  if (__pyx_t_3) {

    /* "ms_deisotope/_c/deconvoluter_base.pyx":827
 *         charges[charge] = 1
 *     if result_size <= 0:
 *         return np.zeros(0, dtype=np.int32)             # <<<<<<<<<<<<<<
 *     result = np.zeros(result_size, dtype=np.int32)
 * 
 */
    __Pyx_XDECREF(((PyObject *)__pyx_r));
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 827, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_zeros); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 827, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 827, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 827, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_int32); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 827, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_dtype, __pyx_t_2) < 0) __PYX_ERR(0, 827, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_tuple__14, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 827, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 827, __pyx_L1_error)
    __pyx_r = ((PyArrayObject *)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":826
 *             result_size += 1
 *         charges[charge] = 1
 *     if result_size <= 0:             # <<<<<<<<<<<<<<
 *         return np.zeros(0, dtype=np.int32)
 *     result = np.zeros(result_size, dtype=np.int32)
 */
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":828
 *     if result_size <= 0:
 *         return np.zeros(0, dtype=np.int32)
 *     result = np.zeros(result_size, dtype=np.int32)             # <<<<<<<<<<<<<<
 * 
 *     i = 0
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_result_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_11 = PyTuple_New(1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_int32); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_dtype, __pyx_t_17) < 0) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  __pyx_t_17 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_11, __pyx_t_2); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(((__pyx_t_17) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_17, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 828, __pyx_L1_error)
  __pyx_t_12 = ((PyArrayObject *)__pyx_t_17);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_result.rcbuffer->pybuffer);
    __pyx_t_13 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_result.rcbuffer->pybuffer, (PyObject*)__pyx_t_12, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_13 < 0)) {
      PyErr_Fetch(&__pyx_t_16, &__pyx_t_15, &__pyx_t_14);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_result.rcbuffer->pybuffer, (PyObject*)__pyx_v_result, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_16); Py_XDECREF(__pyx_t_15); Py_XDECREF(__pyx_t_14);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_16, __pyx_t_15, __pyx_t_14);
      }
      __pyx_t_16 = __pyx_t_15 = __pyx_t_14 = 0;
    }
    __pyx_pybuffernd_result.diminfo[0].strides = __pyx_pybuffernd_result.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_result.diminfo[0].shape = __pyx_pybuffernd_result.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 828, __pyx_L1_error)
  }
  __pyx_t_12 = 0;
  __pyx_v_result = ((PyArrayObject *)__pyx_t_17);
  __pyx_t_17 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":830
 *     result = np.zeros(result_size, dtype=np.int32)
 * 
 *     i = 0             # <<<<<<<<<<<<<<
 *     for j in range(n):
 *         if charges[j] != 0:
 */
  __pyx_v_i = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":831
 * 
 *     i = 0
 *     for j in range(n):             # <<<<<<<<<<<<<<
 *         if charges[j] != 0:
 *             result[i] = j
 */
  __pyx_t_7 = __pyx_v_n;
  __pyx_t_8 = __pyx_t_7;
  for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_8; __pyx_t_10+=1) {
    __pyx_v_j = __pyx_t_10;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":832
 *     i = 0
 *     for j in range(n):
 *         if charges[j] != 0:             # <<<<<<<<<<<<<<
 *             result[i] = j
 *             i += 1
 */
    __pyx_t_3 = (((__pyx_v_charges[__pyx_v_j]) != 0) != 0);
    if (__pyx_t_3) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":833
 *     for j in range(n):
 *         if charges[j] != 0:
 *             result[i] = j             # <<<<<<<<<<<<<<
 *             i += 1
 *     return result
 */
      __pyx_t_9 = __pyx_v_i;
      __pyx_t_13 = -1;
      if (unlikely(__pyx_t_9 >= (size_t)__pyx_pybuffernd_result.diminfo[0].shape)) __pyx_t_13 = 0;
      if (unlikely(__pyx_t_13 != -1)) {
        __Pyx_RaiseBufferIndexError(__pyx_t_13);
        __PYX_ERR(0, 833, __pyx_L1_error)
      }
      *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int32_t *, __pyx_pybuffernd_result.rcbuffer->pybuffer.buf, __pyx_t_9, __pyx_pybuffernd_result.diminfo[0].strides) = __pyx_v_j;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":834
 *         if charges[j] != 0:
 *             result[i] = j
 *             i += 1             # <<<<<<<<<<<<<<
 *     return result
 * 
 */
      __pyx_v_i = (__pyx_v_i + 1);

      /* "ms_deisotope/_c/deconvoluter_base.pyx":832
 *     i = 0
 *     for j in range(n):
 *         if charges[j] != 0:             # <<<<<<<<<<<<<<
 *             result[i] = j
 *             i += 1
 */
    }
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":835
 *             result[i] = j
 *             i += 1
 *     return result             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_result));
  __pyx_r = ((PyArrayObject *)__pyx_v_result);
  goto __pyx_L0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":730
 * 
 * @cython.cdivision
 * cpdef np.ndarray[np.int32_t, ndim=1] quick_charge(FittedPeakCollection peak_set, FittedPeakOrPosition position_spec, int min_charge, int max_charge):             # <<<<<<<<<<<<<<
 *     """An implementation of Hoopman's QuickCharge [1] algorithm for quickly capping charge
 *     state queries
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_17);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_result.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.quick_charge", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_result.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_peaks);
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_23__pyx_fuse_1_0quick_charge(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_1_0__pyx_mdef_12ms_deisotope_2_c_17deconvoluter_base_23__pyx_fuse_1_0quick_charge = {"__pyx_fuse_1_0quick_charge", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_23__pyx_fuse_1_0quick_charge, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_quick_charge};
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_23__pyx_fuse_1_0quick_charge(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_v_peak_set = 0;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_position_spec = 0;
  int __pyx_v_min_charge;
  int __pyx_v_max_charge;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_1_0quick_charge (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_peak_set,&__pyx_n_s_position_spec,&__pyx_n_s_min_charge,&__pyx_n_s_max_charge,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peak_set)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_position_spec)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1_0quick_charge", 1, 4, 4, 1); __PYX_ERR(0, 730, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_min_charge)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1_0quick_charge", 1, 4, 4, 2); __PYX_ERR(0, 730, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_max_charge)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1_0quick_charge", 1, 4, 4, 3); __PYX_ERR(0, 730, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_1_0quick_charge") < 0)) __PYX_ERR(0, 730, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_peak_set = ((struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *)values[0]);
    __pyx_v_position_spec = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)values[1]);
    __pyx_v_min_charge = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_min_charge == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 730, __pyx_L3_error)
    __pyx_v_max_charge = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_max_charge == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 730, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1_0quick_charge", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 730, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.__pyx_fuse_1_0quick_charge", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peak_set), __pyx_ptype_14ms_peak_picker_2_c_10peak_index_PeakIndex, 1, "peak_set", 0))) __PYX_ERR(0, 730, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_position_spec), __pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak, 1, "position_spec", 0))) __PYX_ERR(0, 730, __pyx_L1_error)
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_22__pyx_fuse_1_0quick_charge(__pyx_self, __pyx_v_peak_set, __pyx_v_position_spec, __pyx_v_min_charge, __pyx_v_max_charge);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_22__pyx_fuse_1_0quick_charge(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_v_peak_set, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_position_spec, int __pyx_v_min_charge, int __pyx_v_max_charge) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_1_0quick_charge", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_fuse_1_0__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_quick_charge(__pyx_v_peak_set, __pyx_v_position_spec, __pyx_v_min_charge, __pyx_v_max_charge, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 730, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.__pyx_fuse_1_0quick_charge", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_25__pyx_fuse_1_1quick_charge(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_1quick_charge(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyArrayObject *__pyx_fuse_1_1__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_quick_charge(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_v_peak_set, size_t __pyx_v_position_spec, int __pyx_v_min_charge, int __pyx_v_max_charge, CYTHON_UNUSED int __pyx_skip_dispatch) {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaks = 0;
  size_t __pyx_v_index;
  size_t __pyx_v_first_index;
  int __pyx_v_charges[0x3E8];
  PyArrayObject *__pyx_v_result = 0;
  double __pyx_v_min_intensity;
  double __pyx_v_diff;
  double __pyx_v_raw_charge;
  double __pyx_v_remain;
  int __pyx_v_charge;
  int __pyx_v_result_size;
  Py_ssize_t __pyx_v_j;
  size_t __pyx_v_i;
  size_t __pyx_v_n;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_result;
  __Pyx_Buffer __pyx_pybuffer_result;
  PyArrayObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  long __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  size_t __pyx_t_7;
  size_t __pyx_t_8;
  size_t __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyArrayObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_1_1quick_charge", 0);
  __pyx_pybuffer_result.pybuffer.buf = NULL;
  __pyx_pybuffer_result.refcount = 0;
  __pyx_pybuffernd_result.data = NULL;
  __pyx_pybuffernd_result.rcbuffer = &__pyx_pybuffer_result;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":773
 *         index = position_spec.peak_count
 *     else:
 *         index = position_spec             # <<<<<<<<<<<<<<
 *     if FittedPeakCollection is PeakSet:
 *         peaks = peak_set
 */
  __pyx_v_index = __pyx_v_position_spec;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":777
 *         peaks = peak_set
 *     else:
 *         peaks = peak_set.peaks             # <<<<<<<<<<<<<<
 *     if index > peaks.get_size():
 *         if peaks.get_size() == 0:
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_peak_set->peaks);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_peaks = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":778
 *     else:
 *         peaks = peak_set.peaks
 *     if index > peaks.get_size():             # <<<<<<<<<<<<<<
 *         if peaks.get_size() == 0:
 *             raise IndexError("peak_set cannot be empty!")
 */
  __pyx_t_2 = ((__pyx_v_index > ((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->get_size(__pyx_v_peaks)) != 0);
  if (__pyx_t_2) {

    /* "ms_deisotope/_c/deconvoluter_base.pyx":779
 *         peaks = peak_set.peaks
 *     if index > peaks.get_size():
 *         if peaks.get_size() == 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("peak_set cannot be empty!")
 *         # Assume we have a slice of the peak list, not the actual complete list, so
 */
    __pyx_t_2 = ((((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->get_size(__pyx_v_peaks) == 0) != 0);
    if (unlikely(__pyx_t_2)) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":780
 *     if index > peaks.get_size():
 *         if peaks.get_size() == 0:
 *             raise IndexError("peak_set cannot be empty!")             # <<<<<<<<<<<<<<
 *         # Assume we have a slice of the peak list, not the actual complete list, so
 *         # adjust the index offset from the first peak in the given list.
 */
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple__13, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 780, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(0, 780, __pyx_L1_error)

      /* "ms_deisotope/_c/deconvoluter_base.pyx":779
 *         peaks = peak_set.peaks
 *     if index > peaks.get_size():
 *         if peaks.get_size() == 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("peak_set cannot be empty!")
 *         # Assume we have a slice of the peak list, not the actual complete list, so
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":783
 *         # Assume we have a slice of the peak list, not the actual complete list, so
 *         # adjust the index offset from the first peak in the given list.
 *         first_index = peaks.getitem(0).peak_count             # <<<<<<<<<<<<<<
 *         if (index - first_index) < peaks.get_size() and (index - first_index) >= 0:
 *             index -= first_index
 */
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->getitem(__pyx_v_peaks, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 783, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_1)->peak_count;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_first_index = __pyx_t_3;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":784
 *         # adjust the index offset from the first peak in the given list.
 *         first_index = peaks.getitem(0).peak_count
 *         if (index - first_index) < peaks.get_size() and (index - first_index) >= 0:             # <<<<<<<<<<<<<<
 *             index -= first_index
 *         else:
 */
    __pyx_t_4 = (((__pyx_v_index - __pyx_v_first_index) < ((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->get_size(__pyx_v_peaks)) != 0);
    if (__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_4 = (((__pyx_v_index - __pyx_v_first_index) >= 0) != 0);
    __pyx_t_2 = __pyx_t_4;
    __pyx_L6_bool_binop_done:;
    if (likely(__pyx_t_2)) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":785
 *         first_index = peaks.getitem(0).peak_count
 *         if (index - first_index) < peaks.get_size() and (index - first_index) >= 0:
 *             index -= first_index             # <<<<<<<<<<<<<<
 *         else:
 *             raise IndexError("%d is out of bounds for peak list of size %d in quick_charge" % (index, peaks.get_size()))
 */
      __pyx_v_index = (__pyx_v_index - __pyx_v_first_index);

      /* "ms_deisotope/_c/deconvoluter_base.pyx":784
 *         # adjust the index offset from the first peak in the given list.
 *         first_index = peaks.getitem(0).peak_count
 *         if (index - first_index) < peaks.get_size() and (index - first_index) >= 0:             # <<<<<<<<<<<<<<
 *             index -= first_index
 *         else:
 */
      goto __pyx_L5;
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":787
 *             index -= first_index
 *         else:
 *             raise IndexError("%d is out of bounds for peak list of size %d in quick_charge" % (index, peaks.get_size()))             # <<<<<<<<<<<<<<
 *     n = 1000
 *     result_size = 0
 */
    /*else*/ {
      __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_index); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 787, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_5 = __Pyx_PyInt_FromSize_t(((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->get_size(__pyx_v_peaks)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 787, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 787, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_5);
      __pyx_t_1 = 0;
      __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_d_is_out_of_bounds_for_peak_lis, __pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 787, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_IndexError, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 787, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_Raise(__pyx_t_6, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __PYX_ERR(0, 787, __pyx_L1_error)
    }
    __pyx_L5:;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":778
 *     else:
 *         peaks = peak_set.peaks
 *     if index > peaks.get_size():             # <<<<<<<<<<<<<<
 *         if peaks.get_size() == 0:
 *             raise IndexError("peak_set cannot be empty!")
 */
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":788
 *         else:
 *             raise IndexError("%d is out of bounds for peak list of size %d in quick_charge" % (index, peaks.get_size()))
 *     n = 1000             # <<<<<<<<<<<<<<
 *     result_size = 0
 *     min_intensity = peaks.getitem(index).intensity / 4.
 */
  __pyx_v_n = 0x3E8;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":789
 *             raise IndexError("%d is out of bounds for peak list of size %d in quick_charge" % (index, peaks.get_size()))
 *     n = 1000
 *     result_size = 0             # <<<<<<<<<<<<<<
 *     min_intensity = peaks.getitem(index).intensity / 4.
 *     for i in range(n):
 */
  __pyx_v_result_size = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":790
 *     n = 1000
 *     result_size = 0
 *     min_intensity = peaks.getitem(index).intensity / 4.             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         charges[i] = 0
 */
  __pyx_t_6 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->getitem(__pyx_v_peaks, __pyx_v_index)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 790, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_v_min_intensity = (((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_6)->__pyx_base.intensity / 4.);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":791
 *     result_size = 0
 *     min_intensity = peaks.getitem(index).intensity / 4.
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         charges[i] = 0
 *     for j in range(index + 1, (peaks.get_size())):
 */
  __pyx_t_7 = __pyx_v_n;
  __pyx_t_8 = __pyx_t_7;
  for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
    __pyx_v_i = __pyx_t_9;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":792
 *     min_intensity = peaks.getitem(index).intensity / 4.
 *     for i in range(n):
 *         charges[i] = 0             # <<<<<<<<<<<<<<
 *     for j in range(index + 1, (peaks.get_size())):
 *         if peaks.getitem(j).intensity < min_intensity:
 */
    (__pyx_v_charges[__pyx_v_i]) = 0;
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":793
 *     for i in range(n):
 *         charges[i] = 0
 *     for j in range(index + 1, (peaks.get_size())):             # <<<<<<<<<<<<<<
 *         if peaks.getitem(j).intensity < min_intensity:
 *             continue
 */
  __pyx_t_7 = ((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->get_size(__pyx_v_peaks);
  __pyx_t_8 = __pyx_t_7;
  for (__pyx_t_10 = (__pyx_v_index + 1); __pyx_t_10 < __pyx_t_8; __pyx_t_10+=1) {
    __pyx_v_j = __pyx_t_10;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":794
 *         charges[i] = 0
 *     for j in range(index + 1, (peaks.get_size())):
 *         if peaks.getitem(j).intensity < min_intensity:             # <<<<<<<<<<<<<<
 *             continue
 *         diff = peaks.getitem(j).mz - peaks.getitem(index).mz
 */
    __pyx_t_6 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->getitem(__pyx_v_peaks, __pyx_v_j)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 794, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = ((((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_6)->__pyx_base.intensity < __pyx_v_min_intensity) != 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (__pyx_t_2) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":795
 *     for j in range(index + 1, (peaks.get_size())):
 *         if peaks.getitem(j).intensity < min_intensity:
 *             continue             # <<<<<<<<<<<<<<
 *         diff = peaks.getitem(j).mz - peaks.getitem(index).mz
 *         if diff > 1.1:
 */
      goto __pyx_L10_continue;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":794
 *         charges[i] = 0
 *     for j in range(index + 1, (peaks.get_size())):
 *         if peaks.getitem(j).intensity < min_intensity:             # <<<<<<<<<<<<<<
 *             continue
 *         diff = peaks.getitem(j).mz - peaks.getitem(index).mz
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":796
 *         if peaks.getitem(j).intensity < min_intensity:
 *             continue
 *         diff = peaks.getitem(j).mz - peaks.getitem(index).mz             # <<<<<<<<<<<<<<
 *         if diff > 1.1:
 *             break
 */
    __pyx_t_6 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->getitem(__pyx_v_peaks, __pyx_v_j)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 796, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->getitem(__pyx_v_peaks, __pyx_v_index)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 796, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_v_diff = (((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_6)->__pyx_base.mz - ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_5)->__pyx_base.mz);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":797
 *             continue
 *         diff = peaks.getitem(j).mz - peaks.getitem(index).mz
 *         if diff > 1.1:             # <<<<<<<<<<<<<<
 *             break
 *         raw_charge = 1 / diff
 */
    __pyx_t_2 = ((__pyx_v_diff > 1.1) != 0);
    if (__pyx_t_2) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":798
 *         diff = peaks.getitem(j).mz - peaks.getitem(index).mz
 *         if diff > 1.1:
 *             break             # <<<<<<<<<<<<<<
 *         raw_charge = 1 / diff
 *         charge = <int>(raw_charge + 0.5)
 */
      goto __pyx_L11_break;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":797
 *             continue
 *         diff = peaks.getitem(j).mz - peaks.getitem(index).mz
 *         if diff > 1.1:             # <<<<<<<<<<<<<<
 *             break
 *         raw_charge = 1 / diff
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":799
 *         if diff > 1.1:
 *             break
 *         raw_charge = 1 / diff             # <<<<<<<<<<<<<<
 *         charge = <int>(raw_charge + 0.5)
 *         remain = raw_charge - <int>(raw_charge)
 */
    __pyx_v_raw_charge = (1.0 / __pyx_v_diff);

    /* "ms_deisotope/_c/deconvoluter_base.pyx":800
 *             break
 *         raw_charge = 1 / diff
 *         charge = <int>(raw_charge + 0.5)             # <<<<<<<<<<<<<<
 *         remain = raw_charge - <int>(raw_charge)
 *         if 0.2 < remain and remain < 0.8:
 */
    __pyx_v_charge = ((int)(__pyx_v_raw_charge + 0.5));

    /* "ms_deisotope/_c/deconvoluter_base.pyx":801
 *         raw_charge = 1 / diff
 *         charge = <int>(raw_charge + 0.5)
 *         remain = raw_charge - <int>(raw_charge)             # <<<<<<<<<<<<<<
 *         if 0.2 < remain and remain < 0.8:
 *             continue
 */
    __pyx_v_remain = (__pyx_v_raw_charge - ((int)__pyx_v_raw_charge));

    /* "ms_deisotope/_c/deconvoluter_base.pyx":802
 *         charge = <int>(raw_charge + 0.5)
 *         remain = raw_charge - <int>(raw_charge)
 *         if 0.2 < remain and remain < 0.8:             # <<<<<<<<<<<<<<
 *             continue
 *         if (charge < min_charge) or (charge > max_charge):
 */
    __pyx_t_4 = ((0.2 < __pyx_v_remain) != 0);
    if (__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L15_bool_binop_done;
    }
    __pyx_t_4 = ((__pyx_v_remain < 0.8) != 0);
    __pyx_t_2 = __pyx_t_4;
    __pyx_L15_bool_binop_done:;
    if (__pyx_t_2) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":803
 *         remain = raw_charge - <int>(raw_charge)
 *         if 0.2 < remain and remain < 0.8:
 *             continue             # <<<<<<<<<<<<<<
 *         if (charge < min_charge) or (charge > max_charge):
 *             continue
 */
      goto __pyx_L10_continue;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":802
 *         charge = <int>(raw_charge + 0.5)
 *         remain = raw_charge - <int>(raw_charge)
 *         if 0.2 < remain and remain < 0.8:             # <<<<<<<<<<<<<<
 *             continue
 *         if (charge < min_charge) or (charge > max_charge):
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":804
 *         if 0.2 < remain and remain < 0.8:
 *             continue
 *         if (charge < min_charge) or (charge > max_charge):             # <<<<<<<<<<<<<<
 *             continue
 *         if charges[charge] == 0:
 */
    __pyx_t_4 = ((__pyx_v_charge < __pyx_v_min_charge) != 0);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L18_bool_binop_done;
    }
    __pyx_t_4 = ((__pyx_v_charge > __pyx_v_max_charge) != 0);
    __pyx_t_2 = __pyx_t_4;
    __pyx_L18_bool_binop_done:;
    if (__pyx_t_2) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":805
 *             continue
 *         if (charge < min_charge) or (charge > max_charge):
 *             continue             # <<<<<<<<<<<<<<
 *         if charges[charge] == 0:
 *             result_size += 1
 */
      goto __pyx_L10_continue;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":804
 *         if 0.2 < remain and remain < 0.8:
 *             continue
 *         if (charge < min_charge) or (charge > max_charge):             # <<<<<<<<<<<<<<
 *             continue
 *         if charges[charge] == 0:
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":806
 *         if (charge < min_charge) or (charge > max_charge):
 *             continue
 *         if charges[charge] == 0:             # <<<<<<<<<<<<<<
 *             result_size += 1
 *         charges[charge] = 1
 */
    __pyx_t_2 = (((__pyx_v_charges[__pyx_v_charge]) == 0) != 0);
    if (__pyx_t_2) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":807
 *             continue
 *         if charges[charge] == 0:
 *             result_size += 1             # <<<<<<<<<<<<<<
 *         charges[charge] = 1
 *     if result_size == 0:
 */
      __pyx_v_result_size = (__pyx_v_result_size + 1);

      /* "ms_deisotope/_c/deconvoluter_base.pyx":806
 *         if (charge < min_charge) or (charge > max_charge):
 *             continue
 *         if charges[charge] == 0:             # <<<<<<<<<<<<<<
 *             result_size += 1
 *         charges[charge] = 1
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":808
 *         if charges[charge] == 0:
 *             result_size += 1
 *         charges[charge] = 1             # <<<<<<<<<<<<<<
 *     if result_size == 0:
 *         result = np.empty(0, dtype=np.int32)
 */
    (__pyx_v_charges[__pyx_v_charge]) = 1;
    __pyx_L10_continue:;
  }
  __pyx_L11_break:;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":809
 *             result_size += 1
 *         charges[charge] = 1
 *     if result_size == 0:             # <<<<<<<<<<<<<<
 *         result = np.empty(0, dtype=np.int32)
 *         return result
 */
  __pyx_t_2 = ((__pyx_v_result_size == 0) != 0);
  if (__pyx_t_2) {

    /* "ms_deisotope/_c/deconvoluter_base.pyx":810
 *         charges[charge] = 1
 *     if result_size == 0:
 *         result = np.empty(0, dtype=np.int32)             # <<<<<<<<<<<<<<
 *         return result
 *     for j in range(index - 1, -1, -1):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 810, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_empty); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 810, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 810, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 810, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_int32); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 810, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_dtype, __pyx_t_11) < 0) __PYX_ERR(0, 810, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_tuple__14, __pyx_t_5); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 810, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (!(likely(((__pyx_t_11) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_11, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 810, __pyx_L1_error)
    __pyx_t_12 = ((PyArrayObject *)__pyx_t_11);
    {
      __Pyx_BufFmt_StackElem __pyx_stack[1];
      __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_result.rcbuffer->pybuffer);
      __pyx_t_13 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_result.rcbuffer->pybuffer, (PyObject*)__pyx_t_12, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack);
      if (unlikely(__pyx_t_13 < 0)) {
        PyErr_Fetch(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
        if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_result.rcbuffer->pybuffer, (PyObject*)__pyx_v_result, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {
          Py_XDECREF(__pyx_t_14); Py_XDECREF(__pyx_t_15); Py_XDECREF(__pyx_t_16);
          __Pyx_RaiseBufferFallbackError();
        } else {
          PyErr_Restore(__pyx_t_14, __pyx_t_15, __pyx_t_16);
        }
        __pyx_t_14 = __pyx_t_15 = __pyx_t_16 = 0;
      }
      __pyx_pybuffernd_result.diminfo[0].strides = __pyx_pybuffernd_result.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_result.diminfo[0].shape = __pyx_pybuffernd_result.rcbuffer->pybuffer.shape[0];
      if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 810, __pyx_L1_error)
    }
    __pyx_t_12 = 0;
    __pyx_v_result = ((PyArrayObject *)__pyx_t_11);
    __pyx_t_11 = 0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":811
 *     if result_size == 0:
 *         result = np.empty(0, dtype=np.int32)
 *         return result             # <<<<<<<<<<<<<<
 *     for j in range(index - 1, -1, -1):
 *         diff = peaks.getitem(index).mz - peaks.getitem(j).mz
 */
    __Pyx_XDECREF(((PyObject *)__pyx_r));
    __Pyx_INCREF(((PyObject *)__pyx_v_result));
    __pyx_r = ((PyArrayObject *)__pyx_v_result);
    goto __pyx_L0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":809
 *             result_size += 1
 *         charges[charge] = 1
 *     if result_size == 0:             # <<<<<<<<<<<<<<
 *         result = np.empty(0, dtype=np.int32)
 *         return result
 */
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":812
 *         result = np.empty(0, dtype=np.int32)
 *         return result
 *     for j in range(index - 1, -1, -1):             # <<<<<<<<<<<<<<
 *         diff = peaks.getitem(index).mz - peaks.getitem(j).mz
 *         if diff > 1.1:
 */
  for (__pyx_t_10 = (__pyx_v_index - 1); __pyx_t_10 > -1L; __pyx_t_10-=1) {
    __pyx_v_j = __pyx_t_10;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":813
 *         return result
 *     for j in range(index - 1, -1, -1):
 *         diff = peaks.getitem(index).mz - peaks.getitem(j).mz             # <<<<<<<<<<<<<<
 *         if diff > 1.1:
 *             break
 */
    __pyx_t_11 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->getitem(__pyx_v_peaks, __pyx_v_index)); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 813, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_5 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->getitem(__pyx_v_peaks, __pyx_v_j)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 813, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_v_diff = (((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_11)->__pyx_base.mz - ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_5)->__pyx_base.mz);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":814
 *     for j in range(index - 1, -1, -1):
 *         diff = peaks.getitem(index).mz - peaks.getitem(j).mz
 *         if diff > 1.1:             # <<<<<<<<<<<<<<
 *             break
 *         raw_charge = 1 / diff
 */
    __pyx_t_2 = ((__pyx_v_diff > 1.1) != 0);
    if (__pyx_t_2) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":815
 *         diff = peaks.getitem(index).mz - peaks.getitem(j).mz
 *         if diff > 1.1:
 *             break             # <<<<<<<<<<<<<<
 *         raw_charge = 1 / diff
 *         charge = <int>(raw_charge + 0.5)
 */
      goto __pyx_L23_break;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":814
 *     for j in range(index - 1, -1, -1):
 *         diff = peaks.getitem(index).mz - peaks.getitem(j).mz
 *         if diff > 1.1:             # <<<<<<<<<<<<<<
 *             break
 *         raw_charge = 1 / diff
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":816
 *         if diff > 1.1:
 *             break
 *         raw_charge = 1 / diff             # <<<<<<<<<<<<<<
 *         charge = <int>(raw_charge + 0.5)
 *         remain = raw_charge - <int>(raw_charge)
 */
    __pyx_v_raw_charge = (1.0 / __pyx_v_diff);

    /* "ms_deisotope/_c/deconvoluter_base.pyx":817
 *             break
 *         raw_charge = 1 / diff
 *         charge = <int>(raw_charge + 0.5)             # <<<<<<<<<<<<<<
 *         remain = raw_charge - <int>(raw_charge)
 *         if 0.2 < remain and remain < 0.8:
 */
    __pyx_v_charge = ((int)(__pyx_v_raw_charge + 0.5));

    /* "ms_deisotope/_c/deconvoluter_base.pyx":818
 *         raw_charge = 1 / diff
 *         charge = <int>(raw_charge + 0.5)
 *         remain = raw_charge - <int>(raw_charge)             # <<<<<<<<<<<<<<
 *         if 0.2 < remain and remain < 0.8:
 *             continue
 */
    __pyx_v_remain = (__pyx_v_raw_charge - ((int)__pyx_v_raw_charge));

    /* "ms_deisotope/_c/deconvoluter_base.pyx":819
 *         charge = <int>(raw_charge + 0.5)
 *         remain = raw_charge - <int>(raw_charge)
 *         if 0.2 < remain and remain < 0.8:             # <<<<<<<<<<<<<<
 *             continue
 *         if (charge < min_charge) or (charge > max_charge):
 */
    __pyx_t_4 = ((0.2 < __pyx_v_remain) != 0);
    if (__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L26_bool_binop_done;
    }
    __pyx_t_4 = ((__pyx_v_remain < 0.8) != 0);
    __pyx_t_2 = __pyx_t_4;
    __pyx_L26_bool_binop_done:;
    if (__pyx_t_2) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":820
 *         remain = raw_charge - <int>(raw_charge)
 *         if 0.2 < remain and remain < 0.8:
 *             continue             # <<<<<<<<<<<<<<
 *         if (charge < min_charge) or (charge > max_charge):
 *             continue
 */
      goto __pyx_L22_continue;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":819
 *         charge = <int>(raw_charge + 0.5)
 *         remain = raw_charge - <int>(raw_charge)
 *         if 0.2 < remain and remain < 0.8:             # <<<<<<<<<<<<<<
 *             continue
 *         if (charge < min_charge) or (charge > max_charge):
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":821
 *         if 0.2 < remain and remain < 0.8:
 *             continue
 *         if (charge < min_charge) or (charge > max_charge):             # <<<<<<<<<<<<<<
 *             continue
 *         if charges[charge] == 0:
 */
    __pyx_t_4 = ((__pyx_v_charge < __pyx_v_min_charge) != 0);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L29_bool_binop_done;
    }
    __pyx_t_4 = ((__pyx_v_charge > __pyx_v_max_charge) != 0);
    __pyx_t_2 = __pyx_t_4;
    __pyx_L29_bool_binop_done:;
    if (__pyx_t_2) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":822
 *             continue
 *         if (charge < min_charge) or (charge > max_charge):
 *             continue             # <<<<<<<<<<<<<<
 *         if charges[charge] == 0:
 *             result_size += 1
 */
      goto __pyx_L22_continue;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":821
 *         if 0.2 < remain and remain < 0.8:
 *             continue
 *         if (charge < min_charge) or (charge > max_charge):             # <<<<<<<<<<<<<<
 *             continue
 *         if charges[charge] == 0:
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":823
 *         if (charge < min_charge) or (charge > max_charge):
 *             continue
 *         if charges[charge] == 0:             # <<<<<<<<<<<<<<
 *             result_size += 1
 *         charges[charge] = 1
 */
    __pyx_t_2 = (((__pyx_v_charges[__pyx_v_charge]) == 0) != 0);
    if (__pyx_t_2) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":824
 *             continue
 *         if charges[charge] == 0:
 *             result_size += 1             # <<<<<<<<<<<<<<
 *         charges[charge] = 1
 *     if result_size <= 0:
 */
      __pyx_v_result_size = (__pyx_v_result_size + 1);

      /* "ms_deisotope/_c/deconvoluter_base.pyx":823
 *         if (charge < min_charge) or (charge > max_charge):
 *             continue
 *         if charges[charge] == 0:             # <<<<<<<<<<<<<<
 *             result_size += 1
 *         charges[charge] = 1
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":825
 *         if charges[charge] == 0:
 *             result_size += 1
 *         charges[charge] = 1             # <<<<<<<<<<<<<<
 *     if result_size <= 0:
 *         return np.zeros(0, dtype=np.int32)
 */
    (__pyx_v_charges[__pyx_v_charge]) = 1;
    __pyx_L22_continue:;
  }
  __pyx_L23_break:;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":826
 *             result_size += 1
 *         charges[charge] = 1
 *     if result_size <= 0:             # <<<<<<<<<<<<<<
 *         return np.zeros(0, dtype=np.int32)
 *     result = np.zeros(result_size, dtype=np.int32)
 */
  __pyx_t_2 = ((__pyx_v_result_size <= 0) != 0);
  if (__pyx_t_2) {

    /* "ms_deisotope/_c/deconvoluter_base.pyx":827
 *         charges[charge] = 1
 *     if result_size <= 0:
 *         return np.zeros(0, dtype=np.int32)             # <<<<<<<<<<<<<<
 *     result = np.zeros(result_size, dtype=np.int32)
 * 
 */
    __Pyx_XDECREF(((PyObject *)__pyx_r));
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 827, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_zeros); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 827, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 827, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 827, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_int32); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 827, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_dtype, __pyx_t_1) < 0) __PYX_ERR(0, 827, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_tuple__14, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 827, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 827, __pyx_L1_error)
    __pyx_r = ((PyArrayObject *)__pyx_t_1);
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":826
 *             result_size += 1
 *         charges[charge] = 1
 *     if result_size <= 0:             # <<<<<<<<<<<<<<
 *         return np.zeros(0, dtype=np.int32)
 *     result = np.zeros(result_size, dtype=np.int32)
 */
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":828
 *     if result_size <= 0:
 *         return np.zeros(0, dtype=np.int32)
 *     result = np.zeros(result_size, dtype=np.int32)             # <<<<<<<<<<<<<<
 * 
 *     i = 0
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_result_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_11 = PyTuple_New(1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_int32); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_17) < 0) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  __pyx_t_17 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_11, __pyx_t_1); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_17) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_17, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 828, __pyx_L1_error)
  __pyx_t_12 = ((PyArrayObject *)__pyx_t_17);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_result.rcbuffer->pybuffer);
    __pyx_t_13 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_result.rcbuffer->pybuffer, (PyObject*)__pyx_t_12, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_13 < 0)) {
      PyErr_Fetch(&__pyx_t_16, &__pyx_t_15, &__pyx_t_14);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_result.rcbuffer->pybuffer, (PyObject*)__pyx_v_result, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_16); Py_XDECREF(__pyx_t_15); Py_XDECREF(__pyx_t_14);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_16, __pyx_t_15, __pyx_t_14);
      }
      __pyx_t_16 = __pyx_t_15 = __pyx_t_14 = 0;
    }
    __pyx_pybuffernd_result.diminfo[0].strides = __pyx_pybuffernd_result.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_result.diminfo[0].shape = __pyx_pybuffernd_result.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 828, __pyx_L1_error)
  }
  __pyx_t_12 = 0;
  __pyx_v_result = ((PyArrayObject *)__pyx_t_17);
  __pyx_t_17 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":830
 *     result = np.zeros(result_size, dtype=np.int32)
 * 
 *     i = 0             # <<<<<<<<<<<<<<
 *     for j in range(n):
 *         if charges[j] != 0:
 */
  __pyx_v_i = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":831
 * 
 *     i = 0
 *     for j in range(n):             # <<<<<<<<<<<<<<
 *         if charges[j] != 0:
 *             result[i] = j
 */
  __pyx_t_7 = __pyx_v_n;
  __pyx_t_8 = __pyx_t_7;
  for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_8; __pyx_t_10+=1) {
    __pyx_v_j = __pyx_t_10;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":832
 *     i = 0
 *     for j in range(n):
 *         if charges[j] != 0:             # <<<<<<<<<<<<<<
 *             result[i] = j
 *             i += 1
 */
    __pyx_t_2 = (((__pyx_v_charges[__pyx_v_j]) != 0) != 0);
    if (__pyx_t_2) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":833
 *     for j in range(n):
 *         if charges[j] != 0:
 *             result[i] = j             # <<<<<<<<<<<<<<
 *             i += 1
 *     return result
 */
      __pyx_t_9 = __pyx_v_i;
      __pyx_t_13 = -1;
      if (unlikely(__pyx_t_9 >= (size_t)__pyx_pybuffernd_result.diminfo[0].shape)) __pyx_t_13 = 0;
      if (unlikely(__pyx_t_13 != -1)) {
        __Pyx_RaiseBufferIndexError(__pyx_t_13);
        __PYX_ERR(0, 833, __pyx_L1_error)
      }
      *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int32_t *, __pyx_pybuffernd_result.rcbuffer->pybuffer.buf, __pyx_t_9, __pyx_pybuffernd_result.diminfo[0].strides) = __pyx_v_j;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":834
 *         if charges[j] != 0:
 *             result[i] = j
 *             i += 1             # <<<<<<<<<<<<<<
 *     return result
 * 
 */
      __pyx_v_i = (__pyx_v_i + 1);

      /* "ms_deisotope/_c/deconvoluter_base.pyx":832
 *     i = 0
 *     for j in range(n):
 *         if charges[j] != 0:             # <<<<<<<<<<<<<<
 *             result[i] = j
 *             i += 1
 */
    }
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":835
 *             result[i] = j
 *             i += 1
 *     return result             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_result));
  __pyx_r = ((PyArrayObject *)__pyx_v_result);
  goto __pyx_L0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":730
 * 
 * @cython.cdivision
 * cpdef np.ndarray[np.int32_t, ndim=1] quick_charge(FittedPeakCollection peak_set, FittedPeakOrPosition position_spec, int min_charge, int max_charge):             # <<<<<<<<<<<<<<
 *     """An implementation of Hoopman's QuickCharge [1] algorithm for quickly capping charge
 *     state queries
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_17);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_result.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.quick_charge", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_result.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_peaks);
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_25__pyx_fuse_1_1quick_charge(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_1_1__pyx_mdef_12ms_deisotope_2_c_17deconvoluter_base_25__pyx_fuse_1_1quick_charge = {"__pyx_fuse_1_1quick_charge", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_25__pyx_fuse_1_1quick_charge, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_quick_charge};
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_25__pyx_fuse_1_1quick_charge(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_v_peak_set = 0;
  size_t __pyx_v_position_spec;
  int __pyx_v_min_charge;
  int __pyx_v_max_charge;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_1_1quick_charge (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_peak_set,&__pyx_n_s_position_spec,&__pyx_n_s_min_charge,&__pyx_n_s_max_charge,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peak_set)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_position_spec)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1_1quick_charge", 1, 4, 4, 1); __PYX_ERR(0, 730, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_min_charge)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1_1quick_charge", 1, 4, 4, 2); __PYX_ERR(0, 730, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_max_charge)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1_1quick_charge", 1, 4, 4, 3); __PYX_ERR(0, 730, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_1_1quick_charge") < 0)) __PYX_ERR(0, 730, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_peak_set = ((struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *)values[0]);
    __pyx_v_position_spec = __Pyx_PyInt_As_size_t(values[1]); if (unlikely((__pyx_v_position_spec == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 730, __pyx_L3_error)
    __pyx_v_min_charge = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_min_charge == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 730, __pyx_L3_error)
    __pyx_v_max_charge = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_max_charge == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 730, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1_1quick_charge", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 730, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.__pyx_fuse_1_1quick_charge", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peak_set), __pyx_ptype_14ms_peak_picker_2_c_10peak_index_PeakIndex, 1, "peak_set", 0))) __PYX_ERR(0, 730, __pyx_L1_error)
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_24__pyx_fuse_1_1quick_charge(__pyx_self, __pyx_v_peak_set, __pyx_v_position_spec, __pyx_v_min_charge, __pyx_v_max_charge);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_24__pyx_fuse_1_1quick_charge(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_v_peak_set, size_t __pyx_v_position_spec, int __pyx_v_min_charge, int __pyx_v_max_charge) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_1_1quick_charge", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_fuse_1_1__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_quick_charge(__pyx_v_peak_set, __pyx_v_position_spec, __pyx_v_min_charge, __pyx_v_max_charge, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 730, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.__pyx_fuse_1_1quick_charge", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_27__pyx_fuse_1_2quick_charge(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_1quick_charge(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyArrayObject *__pyx_fuse_1_2__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_quick_charge(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_v_peak_set, long __pyx_v_position_spec, int __pyx_v_min_charge, int __pyx_v_max_charge, CYTHON_UNUSED int __pyx_skip_dispatch) {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaks = 0;
  size_t __pyx_v_index;
  size_t __pyx_v_first_index;
  int __pyx_v_charges[0x3E8];
  PyArrayObject *__pyx_v_result = 0;
  double __pyx_v_min_intensity;
  double __pyx_v_diff;
  double __pyx_v_raw_charge;
  double __pyx_v_remain;
  int __pyx_v_charge;
  int __pyx_v_result_size;
  Py_ssize_t __pyx_v_j;
  size_t __pyx_v_i;
  size_t __pyx_v_n;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_result;
  __Pyx_Buffer __pyx_pybuffer_result;
  PyArrayObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  long __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  size_t __pyx_t_7;
  size_t __pyx_t_8;
  size_t __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyArrayObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_1_2quick_charge", 0);
  __pyx_pybuffer_result.pybuffer.buf = NULL;
  __pyx_pybuffer_result.refcount = 0;
  __pyx_pybuffernd_result.data = NULL;
  __pyx_pybuffernd_result.rcbuffer = &__pyx_pybuffer_result;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":773
 *         index = position_spec.peak_count
 *     else:
 *         index = position_spec             # <<<<<<<<<<<<<<
 *     if FittedPeakCollection is PeakSet:
 *         peaks = peak_set
 */
  __pyx_v_index = __pyx_v_position_spec;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":777
 *         peaks = peak_set
 *     else:
 *         peaks = peak_set.peaks             # <<<<<<<<<<<<<<
 *     if index > peaks.get_size():
 *         if peaks.get_size() == 0:
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_peak_set->peaks);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_peaks = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":778
 *     else:
 *         peaks = peak_set.peaks
 *     if index > peaks.get_size():             # <<<<<<<<<<<<<<
 *         if peaks.get_size() == 0:
 *             raise IndexError("peak_set cannot be empty!")
 */
  __pyx_t_2 = ((__pyx_v_index > ((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->get_size(__pyx_v_peaks)) != 0);
  if (__pyx_t_2) {

    /* "ms_deisotope/_c/deconvoluter_base.pyx":779
 *         peaks = peak_set.peaks
 *     if index > peaks.get_size():
 *         if peaks.get_size() == 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("peak_set cannot be empty!")
 *         # Assume we have a slice of the peak list, not the actual complete list, so
 */
    __pyx_t_2 = ((((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->get_size(__pyx_v_peaks) == 0) != 0);
    if (unlikely(__pyx_t_2)) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":780
 *     if index > peaks.get_size():
 *         if peaks.get_size() == 0:
 *             raise IndexError("peak_set cannot be empty!")             # <<<<<<<<<<<<<<
 *         # Assume we have a slice of the peak list, not the actual complete list, so
 *         # adjust the index offset from the first peak in the given list.
 */
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple__13, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 780, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(0, 780, __pyx_L1_error)

      /* "ms_deisotope/_c/deconvoluter_base.pyx":779
 *         peaks = peak_set.peaks
 *     if index > peaks.get_size():
 *         if peaks.get_size() == 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("peak_set cannot be empty!")
 *         # Assume we have a slice of the peak list, not the actual complete list, so
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":783
 *         # Assume we have a slice of the peak list, not the actual complete list, so
 *         # adjust the index offset from the first peak in the given list.
 *         first_index = peaks.getitem(0).peak_count             # <<<<<<<<<<<<<<
 *         if (index - first_index) < peaks.get_size() and (index - first_index) >= 0:
 *             index -= first_index
 */
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->getitem(__pyx_v_peaks, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 783, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_1)->peak_count;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_first_index = __pyx_t_3;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":784
 *         # adjust the index offset from the first peak in the given list.
 *         first_index = peaks.getitem(0).peak_count
 *         if (index - first_index) < peaks.get_size() and (index - first_index) >= 0:             # <<<<<<<<<<<<<<
 *             index -= first_index
 *         else:
 */
    __pyx_t_4 = (((__pyx_v_index - __pyx_v_first_index) < ((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->get_size(__pyx_v_peaks)) != 0);
    if (__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_4 = (((__pyx_v_index - __pyx_v_first_index) >= 0) != 0);
    __pyx_t_2 = __pyx_t_4;
    __pyx_L6_bool_binop_done:;
    if (likely(__pyx_t_2)) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":785
 *         first_index = peaks.getitem(0).peak_count
 *         if (index - first_index) < peaks.get_size() and (index - first_index) >= 0:
 *             index -= first_index             # <<<<<<<<<<<<<<
 *         else:
 *             raise IndexError("%d is out of bounds for peak list of size %d in quick_charge" % (index, peaks.get_size()))
 */
      __pyx_v_index = (__pyx_v_index - __pyx_v_first_index);

      /* "ms_deisotope/_c/deconvoluter_base.pyx":784
 *         # adjust the index offset from the first peak in the given list.
 *         first_index = peaks.getitem(0).peak_count
 *         if (index - first_index) < peaks.get_size() and (index - first_index) >= 0:             # <<<<<<<<<<<<<<
 *             index -= first_index
 *         else:
 */
      goto __pyx_L5;
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":787
 *             index -= first_index
 *         else:
 *             raise IndexError("%d is out of bounds for peak list of size %d in quick_charge" % (index, peaks.get_size()))             # <<<<<<<<<<<<<<
 *     n = 1000
 *     result_size = 0
 */
    /*else*/ {
      __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_index); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 787, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_5 = __Pyx_PyInt_FromSize_t(((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->get_size(__pyx_v_peaks)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 787, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 787, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_5);
      __pyx_t_1 = 0;
      __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_d_is_out_of_bounds_for_peak_lis, __pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 787, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_IndexError, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 787, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_Raise(__pyx_t_6, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __PYX_ERR(0, 787, __pyx_L1_error)
    }
    __pyx_L5:;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":778
 *     else:
 *         peaks = peak_set.peaks
 *     if index > peaks.get_size():             # <<<<<<<<<<<<<<
 *         if peaks.get_size() == 0:
 *             raise IndexError("peak_set cannot be empty!")
 */
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":788
 *         else:
 *             raise IndexError("%d is out of bounds for peak list of size %d in quick_charge" % (index, peaks.get_size()))
 *     n = 1000             # <<<<<<<<<<<<<<
 *     result_size = 0
 *     min_intensity = peaks.getitem(index).intensity / 4.
 */
  __pyx_v_n = 0x3E8;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":789
 *             raise IndexError("%d is out of bounds for peak list of size %d in quick_charge" % (index, peaks.get_size()))
 *     n = 1000
 *     result_size = 0             # <<<<<<<<<<<<<<
 *     min_intensity = peaks.getitem(index).intensity / 4.
 *     for i in range(n):
 */
  __pyx_v_result_size = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":790
 *     n = 1000
 *     result_size = 0
 *     min_intensity = peaks.getitem(index).intensity / 4.             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         charges[i] = 0
 */
  __pyx_t_6 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->getitem(__pyx_v_peaks, __pyx_v_index)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 790, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_v_min_intensity = (((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_6)->__pyx_base.intensity / 4.);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":791
 *     result_size = 0
 *     min_intensity = peaks.getitem(index).intensity / 4.
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         charges[i] = 0
 *     for j in range(index + 1, (peaks.get_size())):
 */
  __pyx_t_7 = __pyx_v_n;
  __pyx_t_8 = __pyx_t_7;
  for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
    __pyx_v_i = __pyx_t_9;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":792
 *     min_intensity = peaks.getitem(index).intensity / 4.
 *     for i in range(n):
 *         charges[i] = 0             # <<<<<<<<<<<<<<
 *     for j in range(index + 1, (peaks.get_size())):
 *         if peaks.getitem(j).intensity < min_intensity:
 */
    (__pyx_v_charges[__pyx_v_i]) = 0;
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":793
 *     for i in range(n):
 *         charges[i] = 0
 *     for j in range(index + 1, (peaks.get_size())):             # <<<<<<<<<<<<<<
 *         if peaks.getitem(j).intensity < min_intensity:
 *             continue
 */
  __pyx_t_7 = ((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->get_size(__pyx_v_peaks);
  __pyx_t_8 = __pyx_t_7;
  for (__pyx_t_10 = (__pyx_v_index + 1); __pyx_t_10 < __pyx_t_8; __pyx_t_10+=1) {
    __pyx_v_j = __pyx_t_10;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":794
 *         charges[i] = 0
 *     for j in range(index + 1, (peaks.get_size())):
 *         if peaks.getitem(j).intensity < min_intensity:             # <<<<<<<<<<<<<<
 *             continue
 *         diff = peaks.getitem(j).mz - peaks.getitem(index).mz
 */
    __pyx_t_6 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->getitem(__pyx_v_peaks, __pyx_v_j)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 794, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = ((((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_6)->__pyx_base.intensity < __pyx_v_min_intensity) != 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (__pyx_t_2) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":795
 *     for j in range(index + 1, (peaks.get_size())):
 *         if peaks.getitem(j).intensity < min_intensity:
 *             continue             # <<<<<<<<<<<<<<
 *         diff = peaks.getitem(j).mz - peaks.getitem(index).mz
 *         if diff > 1.1:
 */
      goto __pyx_L10_continue;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":794
 *         charges[i] = 0
 *     for j in range(index + 1, (peaks.get_size())):
 *         if peaks.getitem(j).intensity < min_intensity:             # <<<<<<<<<<<<<<
 *             continue
 *         diff = peaks.getitem(j).mz - peaks.getitem(index).mz
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":796
 *         if peaks.getitem(j).intensity < min_intensity:
 *             continue
 *         diff = peaks.getitem(j).mz - peaks.getitem(index).mz             # <<<<<<<<<<<<<<
 *         if diff > 1.1:
 *             break
 */
    __pyx_t_6 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->getitem(__pyx_v_peaks, __pyx_v_j)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 796, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->getitem(__pyx_v_peaks, __pyx_v_index)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 796, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_v_diff = (((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_6)->__pyx_base.mz - ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_5)->__pyx_base.mz);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":797
 *             continue
 *         diff = peaks.getitem(j).mz - peaks.getitem(index).mz
 *         if diff > 1.1:             # <<<<<<<<<<<<<<
 *             break
 *         raw_charge = 1 / diff
 */
    __pyx_t_2 = ((__pyx_v_diff > 1.1) != 0);
    if (__pyx_t_2) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":798
 *         diff = peaks.getitem(j).mz - peaks.getitem(index).mz
 *         if diff > 1.1:
 *             break             # <<<<<<<<<<<<<<
 *         raw_charge = 1 / diff
 *         charge = <int>(raw_charge + 0.5)
 */
      goto __pyx_L11_break;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":797
 *             continue
 *         diff = peaks.getitem(j).mz - peaks.getitem(index).mz
 *         if diff > 1.1:             # <<<<<<<<<<<<<<
 *             break
 *         raw_charge = 1 / diff
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":799
 *         if diff > 1.1:
 *             break
 *         raw_charge = 1 / diff             # <<<<<<<<<<<<<<
 *         charge = <int>(raw_charge + 0.5)
 *         remain = raw_charge - <int>(raw_charge)
 */
    __pyx_v_raw_charge = (1.0 / __pyx_v_diff);

    /* "ms_deisotope/_c/deconvoluter_base.pyx":800
 *             break
 *         raw_charge = 1 / diff
 *         charge = <int>(raw_charge + 0.5)             # <<<<<<<<<<<<<<
 *         remain = raw_charge - <int>(raw_charge)
 *         if 0.2 < remain and remain < 0.8:
 */
    __pyx_v_charge = ((int)(__pyx_v_raw_charge + 0.5));

    /* "ms_deisotope/_c/deconvoluter_base.pyx":801
 *         raw_charge = 1 / diff
 *         charge = <int>(raw_charge + 0.5)
 *         remain = raw_charge - <int>(raw_charge)             # <<<<<<<<<<<<<<
 *         if 0.2 < remain and remain < 0.8:
 *             continue
 */
    __pyx_v_remain = (__pyx_v_raw_charge - ((int)__pyx_v_raw_charge));

    /* "ms_deisotope/_c/deconvoluter_base.pyx":802
 *         charge = <int>(raw_charge + 0.5)
 *         remain = raw_charge - <int>(raw_charge)
 *         if 0.2 < remain and remain < 0.8:             # <<<<<<<<<<<<<<
 *             continue
 *         if (charge < min_charge) or (charge > max_charge):
 */
    __pyx_t_4 = ((0.2 < __pyx_v_remain) != 0);
    if (__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L15_bool_binop_done;
    }
    __pyx_t_4 = ((__pyx_v_remain < 0.8) != 0);
    __pyx_t_2 = __pyx_t_4;
    __pyx_L15_bool_binop_done:;
    if (__pyx_t_2) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":803
 *         remain = raw_charge - <int>(raw_charge)
 *         if 0.2 < remain and remain < 0.8:
 *             continue             # <<<<<<<<<<<<<<
 *         if (charge < min_charge) or (charge > max_charge):
 *             continue
 */
      goto __pyx_L10_continue;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":802
 *         charge = <int>(raw_charge + 0.5)
 *         remain = raw_charge - <int>(raw_charge)
 *         if 0.2 < remain and remain < 0.8:             # <<<<<<<<<<<<<<
 *             continue
 *         if (charge < min_charge) or (charge > max_charge):
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":804
 *         if 0.2 < remain and remain < 0.8:
 *             continue
 *         if (charge < min_charge) or (charge > max_charge):             # <<<<<<<<<<<<<<
 *             continue
 *         if charges[charge] == 0:
 */
    __pyx_t_4 = ((__pyx_v_charge < __pyx_v_min_charge) != 0);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L18_bool_binop_done;
    }
    __pyx_t_4 = ((__pyx_v_charge > __pyx_v_max_charge) != 0);
    __pyx_t_2 = __pyx_t_4;
    __pyx_L18_bool_binop_done:;
    if (__pyx_t_2) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":805
 *             continue
 *         if (charge < min_charge) or (charge > max_charge):
 *             continue             # <<<<<<<<<<<<<<
 *         if charges[charge] == 0:
 *             result_size += 1
 */
      goto __pyx_L10_continue;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":804
 *         if 0.2 < remain and remain < 0.8:
 *             continue
 *         if (charge < min_charge) or (charge > max_charge):             # <<<<<<<<<<<<<<
 *             continue
 *         if charges[charge] == 0:
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":806
 *         if (charge < min_charge) or (charge > max_charge):
 *             continue
 *         if charges[charge] == 0:             # <<<<<<<<<<<<<<
 *             result_size += 1
 *         charges[charge] = 1
 */
    __pyx_t_2 = (((__pyx_v_charges[__pyx_v_charge]) == 0) != 0);
    if (__pyx_t_2) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":807
 *             continue
 *         if charges[charge] == 0:
 *             result_size += 1             # <<<<<<<<<<<<<<
 *         charges[charge] = 1
 *     if result_size == 0:
 */
      __pyx_v_result_size = (__pyx_v_result_size + 1);

      /* "ms_deisotope/_c/deconvoluter_base.pyx":806
 *         if (charge < min_charge) or (charge > max_charge):
 *             continue
 *         if charges[charge] == 0:             # <<<<<<<<<<<<<<
 *             result_size += 1
 *         charges[charge] = 1
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":808
 *         if charges[charge] == 0:
 *             result_size += 1
 *         charges[charge] = 1             # <<<<<<<<<<<<<<
 *     if result_size == 0:
 *         result = np.empty(0, dtype=np.int32)
 */
    (__pyx_v_charges[__pyx_v_charge]) = 1;
    __pyx_L10_continue:;
  }
  __pyx_L11_break:;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":809
 *             result_size += 1
 *         charges[charge] = 1
 *     if result_size == 0:             # <<<<<<<<<<<<<<
 *         result = np.empty(0, dtype=np.int32)
 *         return result
 */
  __pyx_t_2 = ((__pyx_v_result_size == 0) != 0);
  if (__pyx_t_2) {

    /* "ms_deisotope/_c/deconvoluter_base.pyx":810
 *         charges[charge] = 1
 *     if result_size == 0:
 *         result = np.empty(0, dtype=np.int32)             # <<<<<<<<<<<<<<
 *         return result
 *     for j in range(index - 1, -1, -1):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 810, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_empty); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 810, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 810, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 810, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_int32); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 810, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_dtype, __pyx_t_11) < 0) __PYX_ERR(0, 810, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_tuple__14, __pyx_t_5); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 810, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (!(likely(((__pyx_t_11) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_11, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 810, __pyx_L1_error)
    __pyx_t_12 = ((PyArrayObject *)__pyx_t_11);
    {
      __Pyx_BufFmt_StackElem __pyx_stack[1];
      __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_result.rcbuffer->pybuffer);
      __pyx_t_13 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_result.rcbuffer->pybuffer, (PyObject*)__pyx_t_12, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack);
      if (unlikely(__pyx_t_13 < 0)) {
        PyErr_Fetch(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
        if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_result.rcbuffer->pybuffer, (PyObject*)__pyx_v_result, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {
          Py_XDECREF(__pyx_t_14); Py_XDECREF(__pyx_t_15); Py_XDECREF(__pyx_t_16);
          __Pyx_RaiseBufferFallbackError();
        } else {
          PyErr_Restore(__pyx_t_14, __pyx_t_15, __pyx_t_16);
        }
        __pyx_t_14 = __pyx_t_15 = __pyx_t_16 = 0;
      }
      __pyx_pybuffernd_result.diminfo[0].strides = __pyx_pybuffernd_result.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_result.diminfo[0].shape = __pyx_pybuffernd_result.rcbuffer->pybuffer.shape[0];
      if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 810, __pyx_L1_error)
    }
    __pyx_t_12 = 0;
    __pyx_v_result = ((PyArrayObject *)__pyx_t_11);
    __pyx_t_11 = 0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":811
 *     if result_size == 0:
 *         result = np.empty(0, dtype=np.int32)
 *         return result             # <<<<<<<<<<<<<<
 *     for j in range(index - 1, -1, -1):
 *         diff = peaks.getitem(index).mz - peaks.getitem(j).mz
 */
    __Pyx_XDECREF(((PyObject *)__pyx_r));
    __Pyx_INCREF(((PyObject *)__pyx_v_result));
    __pyx_r = ((PyArrayObject *)__pyx_v_result);
    goto __pyx_L0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":809
 *             result_size += 1
 *         charges[charge] = 1
 *     if result_size == 0:             # <<<<<<<<<<<<<<
 *         result = np.empty(0, dtype=np.int32)
 *         return result
 */
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":812
 *         result = np.empty(0, dtype=np.int32)
 *         return result
 *     for j in range(index - 1, -1, -1):             # <<<<<<<<<<<<<<
 *         diff = peaks.getitem(index).mz - peaks.getitem(j).mz
 *         if diff > 1.1:
 */
  for (__pyx_t_10 = (__pyx_v_index - 1); __pyx_t_10 > -1L; __pyx_t_10-=1) {
    __pyx_v_j = __pyx_t_10;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":813
 *         return result
 *     for j in range(index - 1, -1, -1):
 *         diff = peaks.getitem(index).mz - peaks.getitem(j).mz             # <<<<<<<<<<<<<<
 *         if diff > 1.1:
 *             break
 */
    __pyx_t_11 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->getitem(__pyx_v_peaks, __pyx_v_index)); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 813, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_5 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->getitem(__pyx_v_peaks, __pyx_v_j)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 813, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_v_diff = (((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_11)->__pyx_base.mz - ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_5)->__pyx_base.mz);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":814
 *     for j in range(index - 1, -1, -1):
 *         diff = peaks.getitem(index).mz - peaks.getitem(j).mz
 *         if diff > 1.1:             # <<<<<<<<<<<<<<
 *             break
 *         raw_charge = 1 / diff
 */
    __pyx_t_2 = ((__pyx_v_diff > 1.1) != 0);
    if (__pyx_t_2) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":815
 *         diff = peaks.getitem(index).mz - peaks.getitem(j).mz
 *         if diff > 1.1:
 *             break             # <<<<<<<<<<<<<<
 *         raw_charge = 1 / diff
 *         charge = <int>(raw_charge + 0.5)
 */
      goto __pyx_L23_break;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":814
 *     for j in range(index - 1, -1, -1):
 *         diff = peaks.getitem(index).mz - peaks.getitem(j).mz
 *         if diff > 1.1:             # <<<<<<<<<<<<<<
 *             break
 *         raw_charge = 1 / diff
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":816
 *         if diff > 1.1:
 *             break
 *         raw_charge = 1 / diff             # <<<<<<<<<<<<<<
 *         charge = <int>(raw_charge + 0.5)
 *         remain = raw_charge - <int>(raw_charge)
 */
    __pyx_v_raw_charge = (1.0 / __pyx_v_diff);

    /* "ms_deisotope/_c/deconvoluter_base.pyx":817
 *             break
 *         raw_charge = 1 / diff
 *         charge = <int>(raw_charge + 0.5)             # <<<<<<<<<<<<<<
 *         remain = raw_charge - <int>(raw_charge)
 *         if 0.2 < remain and remain < 0.8:
 */
    __pyx_v_charge = ((int)(__pyx_v_raw_charge + 0.5));

    /* "ms_deisotope/_c/deconvoluter_base.pyx":818
 *         raw_charge = 1 / diff
 *         charge = <int>(raw_charge + 0.5)
 *         remain = raw_charge - <int>(raw_charge)             # <<<<<<<<<<<<<<
 *         if 0.2 < remain and remain < 0.8:
 *             continue
 */
    __pyx_v_remain = (__pyx_v_raw_charge - ((int)__pyx_v_raw_charge));

    /* "ms_deisotope/_c/deconvoluter_base.pyx":819
 *         charge = <int>(raw_charge + 0.5)
 *         remain = raw_charge - <int>(raw_charge)
 *         if 0.2 < remain and remain < 0.8:             # <<<<<<<<<<<<<<
 *             continue
 *         if (charge < min_charge) or (charge > max_charge):
 */
    __pyx_t_4 = ((0.2 < __pyx_v_remain) != 0);
    if (__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L26_bool_binop_done;
    }
    __pyx_t_4 = ((__pyx_v_remain < 0.8) != 0);
    __pyx_t_2 = __pyx_t_4;
    __pyx_L26_bool_binop_done:;
    if (__pyx_t_2) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":820
 *         remain = raw_charge - <int>(raw_charge)
 *         if 0.2 < remain and remain < 0.8:
 *             continue             # <<<<<<<<<<<<<<
 *         if (charge < min_charge) or (charge > max_charge):
 *             continue
 */
      goto __pyx_L22_continue;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":819
 *         charge = <int>(raw_charge + 0.5)
 *         remain = raw_charge - <int>(raw_charge)
 *         if 0.2 < remain and remain < 0.8:             # <<<<<<<<<<<<<<
 *             continue
 *         if (charge < min_charge) or (charge > max_charge):
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":821
 *         if 0.2 < remain and remain < 0.8:
 *             continue
 *         if (charge < min_charge) or (charge > max_charge):             # <<<<<<<<<<<<<<
 *             continue
 *         if charges[charge] == 0:
 */
    __pyx_t_4 = ((__pyx_v_charge < __pyx_v_min_charge) != 0);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L29_bool_binop_done;
    }
    __pyx_t_4 = ((__pyx_v_charge > __pyx_v_max_charge) != 0);
    __pyx_t_2 = __pyx_t_4;
    __pyx_L29_bool_binop_done:;
    if (__pyx_t_2) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":822
 *             continue
 *         if (charge < min_charge) or (charge > max_charge):
 *             continue             # <<<<<<<<<<<<<<
 *         if charges[charge] == 0:
 *             result_size += 1
 */
      goto __pyx_L22_continue;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":821
 *         if 0.2 < remain and remain < 0.8:
 *             continue
 *         if (charge < min_charge) or (charge > max_charge):             # <<<<<<<<<<<<<<
 *             continue
 *         if charges[charge] == 0:
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":823
 *         if (charge < min_charge) or (charge > max_charge):
 *             continue
 *         if charges[charge] == 0:             # <<<<<<<<<<<<<<
 *             result_size += 1
 *         charges[charge] = 1
 */
    __pyx_t_2 = (((__pyx_v_charges[__pyx_v_charge]) == 0) != 0);
    if (__pyx_t_2) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":824
 *             continue
 *         if charges[charge] == 0:
 *             result_size += 1             # <<<<<<<<<<<<<<
 *         charges[charge] = 1
 *     if result_size <= 0:
 */
      __pyx_v_result_size = (__pyx_v_result_size + 1);

      /* "ms_deisotope/_c/deconvoluter_base.pyx":823
 *         if (charge < min_charge) or (charge > max_charge):
 *             continue
 *         if charges[charge] == 0:             # <<<<<<<<<<<<<<
 *             result_size += 1
 *         charges[charge] = 1
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":825
 *         if charges[charge] == 0:
 *             result_size += 1
 *         charges[charge] = 1             # <<<<<<<<<<<<<<
 *     if result_size <= 0:
 *         return np.zeros(0, dtype=np.int32)
 */
    (__pyx_v_charges[__pyx_v_charge]) = 1;
    __pyx_L22_continue:;
  }
  __pyx_L23_break:;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":826
 *             result_size += 1
 *         charges[charge] = 1
 *     if result_size <= 0:             # <<<<<<<<<<<<<<
 *         return np.zeros(0, dtype=np.int32)
 *     result = np.zeros(result_size, dtype=np.int32)
 */
  __pyx_t_2 = ((__pyx_v_result_size <= 0) != 0);
  if (__pyx_t_2) {

    /* "ms_deisotope/_c/deconvoluter_base.pyx":827
 *         charges[charge] = 1
 *     if result_size <= 0:
 *         return np.zeros(0, dtype=np.int32)             # <<<<<<<<<<<<<<
 *     result = np.zeros(result_size, dtype=np.int32)
 * 
 */
    __Pyx_XDECREF(((PyObject *)__pyx_r));
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 827, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_zeros); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 827, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 827, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 827, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_int32); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 827, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_dtype, __pyx_t_1) < 0) __PYX_ERR(0, 827, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_tuple__14, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 827, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 827, __pyx_L1_error)
    __pyx_r = ((PyArrayObject *)__pyx_t_1);
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":826
 *             result_size += 1
 *         charges[charge] = 1
 *     if result_size <= 0:             # <<<<<<<<<<<<<<
 *         return np.zeros(0, dtype=np.int32)
 *     result = np.zeros(result_size, dtype=np.int32)
 */
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":828
 *     if result_size <= 0:
 *         return np.zeros(0, dtype=np.int32)
 *     result = np.zeros(result_size, dtype=np.int32)             # <<<<<<<<<<<<<<
 * 
 *     i = 0
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_result_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_11 = PyTuple_New(1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_int32); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_17) < 0) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  __pyx_t_17 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_11, __pyx_t_1); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_17) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_17, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 828, __pyx_L1_error)
  __pyx_t_12 = ((PyArrayObject *)__pyx_t_17);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_result.rcbuffer->pybuffer);
    __pyx_t_13 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_result.rcbuffer->pybuffer, (PyObject*)__pyx_t_12, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_13 < 0)) {
      PyErr_Fetch(&__pyx_t_16, &__pyx_t_15, &__pyx_t_14);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_result.rcbuffer->pybuffer, (PyObject*)__pyx_v_result, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_16); Py_XDECREF(__pyx_t_15); Py_XDECREF(__pyx_t_14);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_16, __pyx_t_15, __pyx_t_14);
      }
      __pyx_t_16 = __pyx_t_15 = __pyx_t_14 = 0;
    }
    __pyx_pybuffernd_result.diminfo[0].strides = __pyx_pybuffernd_result.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_result.diminfo[0].shape = __pyx_pybuffernd_result.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 828, __pyx_L1_error)
  }
  __pyx_t_12 = 0;
  __pyx_v_result = ((PyArrayObject *)__pyx_t_17);
  __pyx_t_17 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":830
 *     result = np.zeros(result_size, dtype=np.int32)
 * 
 *     i = 0             # <<<<<<<<<<<<<<
 *     for j in range(n):
 *         if charges[j] != 0:
 */
  __pyx_v_i = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":831
 * 
 *     i = 0
 *     for j in range(n):             # <<<<<<<<<<<<<<
 *         if charges[j] != 0:
 *             result[i] = j
 */
  __pyx_t_7 = __pyx_v_n;
  __pyx_t_8 = __pyx_t_7;
  for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_8; __pyx_t_10+=1) {
    __pyx_v_j = __pyx_t_10;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":832
 *     i = 0
 *     for j in range(n):
 *         if charges[j] != 0:             # <<<<<<<<<<<<<<
 *             result[i] = j
 *             i += 1
 */
    __pyx_t_2 = (((__pyx_v_charges[__pyx_v_j]) != 0) != 0);
    if (__pyx_t_2) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":833
 *     for j in range(n):
 *         if charges[j] != 0:
 *             result[i] = j             # <<<<<<<<<<<<<<
 *             i += 1
 *     return result
 */
      __pyx_t_9 = __pyx_v_i;
      __pyx_t_13 = -1;
      if (unlikely(__pyx_t_9 >= (size_t)__pyx_pybuffernd_result.diminfo[0].shape)) __pyx_t_13 = 0;
      if (unlikely(__pyx_t_13 != -1)) {
        __Pyx_RaiseBufferIndexError(__pyx_t_13);
        __PYX_ERR(0, 833, __pyx_L1_error)
      }
      *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int32_t *, __pyx_pybuffernd_result.rcbuffer->pybuffer.buf, __pyx_t_9, __pyx_pybuffernd_result.diminfo[0].strides) = __pyx_v_j;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":834
 *         if charges[j] != 0:
 *             result[i] = j
 *             i += 1             # <<<<<<<<<<<<<<
 *     return result
 * 
 */
      __pyx_v_i = (__pyx_v_i + 1);

      /* "ms_deisotope/_c/deconvoluter_base.pyx":832
 *     i = 0
 *     for j in range(n):
 *         if charges[j] != 0:             # <<<<<<<<<<<<<<
 *             result[i] = j
 *             i += 1
 */
    }
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":835
 *             result[i] = j
 *             i += 1
 *     return result             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_result));
  __pyx_r = ((PyArrayObject *)__pyx_v_result);
  goto __pyx_L0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":730
 * 
 * @cython.cdivision
 * cpdef np.ndarray[np.int32_t, ndim=1] quick_charge(FittedPeakCollection peak_set, FittedPeakOrPosition position_spec, int min_charge, int max_charge):             # <<<<<<<<<<<<<<
 *     """An implementation of Hoopman's QuickCharge [1] algorithm for quickly capping charge
 *     state queries
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_17);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_result.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.quick_charge", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_result.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_peaks);
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_27__pyx_fuse_1_2quick_charge(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_1_2__pyx_mdef_12ms_deisotope_2_c_17deconvoluter_base_27__pyx_fuse_1_2quick_charge = {"__pyx_fuse_1_2quick_charge", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_27__pyx_fuse_1_2quick_charge, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_quick_charge};
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_27__pyx_fuse_1_2quick_charge(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_v_peak_set = 0;
  long __pyx_v_position_spec;
  int __pyx_v_min_charge;
  int __pyx_v_max_charge;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_1_2quick_charge (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_peak_set,&__pyx_n_s_position_spec,&__pyx_n_s_min_charge,&__pyx_n_s_max_charge,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peak_set)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_position_spec)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1_2quick_charge", 1, 4, 4, 1); __PYX_ERR(0, 730, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_min_charge)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1_2quick_charge", 1, 4, 4, 2); __PYX_ERR(0, 730, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_max_charge)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1_2quick_charge", 1, 4, 4, 3); __PYX_ERR(0, 730, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_1_2quick_charge") < 0)) __PYX_ERR(0, 730, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_peak_set = ((struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *)values[0]);
    __pyx_v_position_spec = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v_position_spec == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 730, __pyx_L3_error)
    __pyx_v_min_charge = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_min_charge == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 730, __pyx_L3_error)
    __pyx_v_max_charge = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_max_charge == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 730, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1_2quick_charge", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 730, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.__pyx_fuse_1_2quick_charge", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peak_set), __pyx_ptype_14ms_peak_picker_2_c_10peak_index_PeakIndex, 1, "peak_set", 0))) __PYX_ERR(0, 730, __pyx_L1_error)
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_26__pyx_fuse_1_2quick_charge(__pyx_self, __pyx_v_peak_set, __pyx_v_position_spec, __pyx_v_min_charge, __pyx_v_max_charge);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_26__pyx_fuse_1_2quick_charge(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_v_peak_set, long __pyx_v_position_spec, int __pyx_v_min_charge, int __pyx_v_max_charge) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_1_2quick_charge", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_fuse_1_2__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_quick_charge(__pyx_v_peak_set, __pyx_v_position_spec, __pyx_v_min_charge, __pyx_v_max_charge, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 730, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.__pyx_fuse_1_2quick_charge", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/deconvoluter_base.pyx":839
 * 
 * @cython.binding(True)
 * cpdef set _get_all_peak_charge_pairs(DeconvoluterBase self, FittedPeak peak, double error_tolerance=ERROR_TOLERANCE,             # <<<<<<<<<<<<<<
 *                                      object charge_range=(1, 8), int left_search_limit=3, int right_search_limit=3,
 *                                      bint recalculate_starting_peak=True, bint use_quick_charge=False):
 */

static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_3_get_all_peak_charge_pairs(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base__get_all_peak_charge_pairs(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_peak, CYTHON_UNUSED int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base__get_all_peak_charge_pairs *__pyx_optional_args) {
  double __pyx_v_error_tolerance = __pyx_k__15;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":840
 * @cython.binding(True)
 * cpdef set _get_all_peak_charge_pairs(DeconvoluterBase self, FittedPeak peak, double error_tolerance=ERROR_TOLERANCE,
 *                                      object charge_range=(1, 8), int left_search_limit=3, int right_search_limit=3,             # <<<<<<<<<<<<<<
 *                                      bint recalculate_starting_peak=True, bint use_quick_charge=False):
 *         """Construct the set of all unique candidate (monoisotopic peak, charge state) pairs using
 */
  PyObject *__pyx_v_charge_range = ((PyObject *)__pyx_tuple__16);
  int __pyx_v_left_search_limit = ((int)3);
  int __pyx_v_right_search_limit = ((int)3);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":841
 * cpdef set _get_all_peak_charge_pairs(DeconvoluterBase self, FittedPeak peak, double error_tolerance=ERROR_TOLERANCE,
 *                                      object charge_range=(1, 8), int left_search_limit=3, int right_search_limit=3,
 *                                      bint recalculate_starting_peak=True, bint use_quick_charge=False):             # <<<<<<<<<<<<<<
 *         """Construct the set of all unique candidate (monoisotopic peak, charge state) pairs using
 *         the provided search parameters.
 */
  int __pyx_v_recalculate_starting_peak = ((int)1);
  int __pyx_v_use_quick_charge = ((int)0);
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_v_charge_iterator = 0;
  int __pyx_v_charge;
  size_t __pyx_v_i;
  PyObject *__pyx_v_target_peaks = 0;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_prev_peak = 0;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_nxt_peak = 0;
  PyObject *__pyx_v_add_ = 0;
  CYTHON_UNUSED PyObject *__pyx_v_update_ = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  size_t __pyx_t_8;
  int __pyx_t_9;
  struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak_inplace __pyx_t_10;
  struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak_inplace __pyx_t_11;
  long __pyx_t_12;
  long __pyx_t_13;
  __Pyx_RefNannySetupContext("_get_all_peak_charge_pairs", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_error_tolerance = __pyx_optional_args->error_tolerance;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_charge_range = __pyx_optional_args->charge_range;
        if (__pyx_optional_args->__pyx_n > 2) {
          __pyx_v_left_search_limit = __pyx_optional_args->left_search_limit;
          if (__pyx_optional_args->__pyx_n > 3) {
            __pyx_v_right_search_limit = __pyx_optional_args->right_search_limit;
            if (__pyx_optional_args->__pyx_n > 4) {
              __pyx_v_recalculate_starting_peak = __pyx_optional_args->recalculate_starting_peak;
              if (__pyx_optional_args->__pyx_n > 5) {
                __pyx_v_use_quick_charge = __pyx_optional_args->use_quick_charge;
              }
            }
          }
        }
      }
    }
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":877
 *             object add_, update_
 * 
 *         if use_quick_charge:             # <<<<<<<<<<<<<<
 *             charge_iterator = ChargeIterator.from_quickcharge(tuple(charge_range), self.peaklist, peak)
 *         else:
 */
  __pyx_t_1 = (__pyx_v_use_quick_charge != 0);
  if (__pyx_t_1) {

    /* "ms_deisotope/_c/deconvoluter_base.pyx":878
 * 
 *         if use_quick_charge:
 *             charge_iterator = ChargeIterator.from_quickcharge(tuple(charge_range), self.peaklist, peak)             # <<<<<<<<<<<<<<
 *         else:
 *             charge_iterator = ChargeIterator.from_tuple(tuple(charge_range))
 */
    __pyx_t_2 = __Pyx_PySequence_Tuple(__pyx_v_charge_range); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 878, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = ((PyObject *)__pyx_v_self->peaklist);
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_4 = ((PyObject *)__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_from_quickcharge(((PyObject*)__pyx_t_2), ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_t_3), __pyx_v_peak)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 878, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_charge_iterator = ((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *)__pyx_t_4);
    __pyx_t_4 = 0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":877
 *             object add_, update_
 * 
 *         if use_quick_charge:             # <<<<<<<<<<<<<<
 *             charge_iterator = ChargeIterator.from_quickcharge(tuple(charge_range), self.peaklist, peak)
 *         else:
 */
    goto __pyx_L3;
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":880
 *             charge_iterator = ChargeIterator.from_quickcharge(tuple(charge_range), self.peaklist, peak)
 *         else:
 *             charge_iterator = ChargeIterator.from_tuple(tuple(charge_range))             # <<<<<<<<<<<<<<
 * 
 *         target_peaks = set()
 */
  /*else*/ {
    __pyx_t_4 = __Pyx_PySequence_Tuple(__pyx_v_charge_range); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 880, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = ((PyObject *)__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_from_tuple(((PyObject*)__pyx_t_4))); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 880, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_charge_iterator = ((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *)__pyx_t_3);
    __pyx_t_3 = 0;
  }
  __pyx_L3:;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":882
 *             charge_iterator = ChargeIterator.from_tuple(tuple(charge_range))
 * 
 *         target_peaks = set()             # <<<<<<<<<<<<<<
 *         add_ = target_peaks.add
 *         update_ = target_peaks.update
 */
  __pyx_t_3 = PySet_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 882, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_target_peaks = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":883
 * 
 *         target_peaks = set()
 *         add_ = target_peaks.add             # <<<<<<<<<<<<<<
 *         update_ = target_peaks.update
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_target_peaks, __pyx_n_s_add); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 883, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_add_ = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":884
 *         target_peaks = set()
 *         add_ = target_peaks.add
 *         update_ = target_peaks.update             # <<<<<<<<<<<<<<
 * 
 *         while charge_iterator.has_more():
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_target_peaks, __pyx_n_s_update); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 884, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_update_ = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":886
 *         update_ = target_peaks.update
 * 
 *         while charge_iterator.has_more():             # <<<<<<<<<<<<<<
 *             charge = charge_iterator.get_next_value()
 *             add_((peak, charge))
 */
  while (1) {
    __pyx_t_1 = (__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_has_more(__pyx_v_charge_iterator, 0) != 0);
    if (!__pyx_t_1) break;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":887
 * 
 *         while charge_iterator.has_more():
 *             charge = charge_iterator.get_next_value()             # <<<<<<<<<<<<<<
 *             add_((peak, charge))
 * 
 */
    __pyx_v_charge = __pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_get_next_value(__pyx_v_charge_iterator, 0);

    /* "ms_deisotope/_c/deconvoluter_base.pyx":888
 *         while charge_iterator.has_more():
 *             charge = charge_iterator.get_next_value()
 *             add_((peak, charge))             # <<<<<<<<<<<<<<
 * 
 *             # Look Left
 */
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_charge); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 888, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 888, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(((PyObject *)__pyx_v_peak));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_peak));
    PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)__pyx_v_peak));
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_5 = PySet_Add(__pyx_v_target_peaks, __pyx_t_4); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 888, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":891
 * 
 *             # Look Left
 *             for i in range(1, left_search_limit):             # <<<<<<<<<<<<<<
 *                 prev_peak = has_previous_peak_at_charge(
 *                     self, peak, charge, i, error_tolerance)
 */
    __pyx_t_6 = __pyx_v_left_search_limit;
    __pyx_t_7 = __pyx_t_6;
    for (__pyx_t_8 = 1; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {
      __pyx_v_i = __pyx_t_8;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":892
 *             # Look Left
 *             for i in range(1, left_search_limit):
 *                 prev_peak = has_previous_peak_at_charge(             # <<<<<<<<<<<<<<
 *                     self, peak, charge, i, error_tolerance)
 *                 if prev_peak is None:
 */
      __pyx_t_4 = ((PyObject *)__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_has_previous_peak_at_charge(__pyx_v_self, __pyx_v_peak, __pyx_v_charge, __pyx_v_i, __pyx_v_error_tolerance)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 892, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_XDECREF_SET(__pyx_v_prev_peak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_4));
      __pyx_t_4 = 0;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":894
 *                 prev_peak = has_previous_peak_at_charge(
 *                     self, peak, charge, i, error_tolerance)
 *                 if prev_peak is None:             # <<<<<<<<<<<<<<
 *                     continue
 *                 add_((prev_peak, charge))
 */
      __pyx_t_1 = (((PyObject *)__pyx_v_prev_peak) == Py_None);
      __pyx_t_9 = (__pyx_t_1 != 0);
      if (__pyx_t_9) {

        /* "ms_deisotope/_c/deconvoluter_base.pyx":895
 *                     self, peak, charge, i, error_tolerance)
 *                 if prev_peak is None:
 *                     continue             # <<<<<<<<<<<<<<
 *                 add_((prev_peak, charge))
 * 
 */
        goto __pyx_L6_continue;

        /* "ms_deisotope/_c/deconvoluter_base.pyx":894
 *                 prev_peak = has_previous_peak_at_charge(
 *                     self, peak, charge, i, error_tolerance)
 *                 if prev_peak is None:             # <<<<<<<<<<<<<<
 *                     continue
 *                 add_((prev_peak, charge))
 */
      }

      /* "ms_deisotope/_c/deconvoluter_base.pyx":896
 *                 if prev_peak is None:
 *                     continue
 *                 add_((prev_peak, charge))             # <<<<<<<<<<<<<<
 * 
 *                 if recalculate_starting_peak:
 */
      __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_charge); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 896, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 896, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(((PyObject *)__pyx_v_prev_peak));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_prev_peak));
      PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_v_prev_peak));
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_4);
      __pyx_t_4 = 0;
      __pyx_t_5 = PySet_Add(__pyx_v_target_peaks, __pyx_t_3); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 896, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":898
 *                 add_((prev_peak, charge))
 * 
 *                 if recalculate_starting_peak:             # <<<<<<<<<<<<<<
 *                     # update_(self._find_previous_putative_peak(
 *                     #     peak.mz, charge, i, 2 * error_tolerance))
 */
      __pyx_t_9 = (__pyx_v_recalculate_starting_peak != 0);
      if (__pyx_t_9) {

        /* "ms_deisotope/_c/deconvoluter_base.pyx":901
 *                     # update_(self._find_previous_putative_peak(
 *                     #     peak.mz, charge, i, 2 * error_tolerance))
 *                     self._find_previous_putative_peak_inplace(             # <<<<<<<<<<<<<<
 *                         peak.mz, charge, target_peaks, i, 2 * error_tolerance)
 * 
 */
        __pyx_t_10.__pyx_n = 2;
        __pyx_t_10.step = __pyx_v_i;
        __pyx_t_10.tolerance = (2.0 * __pyx_v_error_tolerance);
        ((struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *)__pyx_v_self->__pyx_vtab)->_find_previous_putative_peak_inplace(__pyx_v_self, __pyx_v_peak->__pyx_base.mz, __pyx_v_charge, __pyx_v_target_peaks, &__pyx_t_10); 

        /* "ms_deisotope/_c/deconvoluter_base.pyx":898
 *                 add_((prev_peak, charge))
 * 
 *                 if recalculate_starting_peak:             # <<<<<<<<<<<<<<
 *                     # update_(self._find_previous_putative_peak(
 *                     #     peak.mz, charge, i, 2 * error_tolerance))
 */
      }
      __pyx_L6_continue:;
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":905
 * 
 *             # Look Right
 *             for i in range(1, right_search_limit):             # <<<<<<<<<<<<<<
 *                 nxt_peak = has_successor_peak_at_charge(
 *                     self, peak, charge, i, error_tolerance)
 */
    __pyx_t_6 = __pyx_v_right_search_limit;
    __pyx_t_7 = __pyx_t_6;
    for (__pyx_t_8 = 1; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {
      __pyx_v_i = __pyx_t_8;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":906
 *             # Look Right
 *             for i in range(1, right_search_limit):
 *                 nxt_peak = has_successor_peak_at_charge(             # <<<<<<<<<<<<<<
 *                     self, peak, charge, i, error_tolerance)
 *                 if nxt_peak is None:
 */
      __pyx_t_3 = ((PyObject *)__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_has_successor_peak_at_charge(__pyx_v_self, __pyx_v_peak, __pyx_v_charge, __pyx_v_i, __pyx_v_error_tolerance)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 906, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_XDECREF_SET(__pyx_v_nxt_peak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_3));
      __pyx_t_3 = 0;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":908
 *                 nxt_peak = has_successor_peak_at_charge(
 *                     self, peak, charge, i, error_tolerance)
 *                 if nxt_peak is None:             # <<<<<<<<<<<<<<
 *                     continue
 *                 add_((nxt_peak, charge))
 */
      __pyx_t_9 = (((PyObject *)__pyx_v_nxt_peak) == Py_None);
      __pyx_t_1 = (__pyx_t_9 != 0);
      if (__pyx_t_1) {

        /* "ms_deisotope/_c/deconvoluter_base.pyx":909
 *                     self, peak, charge, i, error_tolerance)
 *                 if nxt_peak is None:
 *                     continue             # <<<<<<<<<<<<<<
 *                 add_((nxt_peak, charge))
 * 
 */
        goto __pyx_L10_continue;

        /* "ms_deisotope/_c/deconvoluter_base.pyx":908
 *                 nxt_peak = has_successor_peak_at_charge(
 *                     self, peak, charge, i, error_tolerance)
 *                 if nxt_peak is None:             # <<<<<<<<<<<<<<
 *                     continue
 *                 add_((nxt_peak, charge))
 */
      }

      /* "ms_deisotope/_c/deconvoluter_base.pyx":910
 *                 if nxt_peak is None:
 *                     continue
 *                 add_((nxt_peak, charge))             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_charge); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 910, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 910, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF(((PyObject *)__pyx_v_nxt_peak));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_nxt_peak));
      PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)__pyx_v_nxt_peak));
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3);
      __pyx_t_3 = 0;
      __pyx_t_5 = PySet_Add(__pyx_v_target_peaks, __pyx_t_4); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 910, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":913
 * 
 * 
 *                 if recalculate_starting_peak:             # <<<<<<<<<<<<<<
 *                     # update_(self._find_next_putative_peak(
 *                     #     peak.mz, charge, i, 2 * error_tolerance))
 */
      __pyx_t_1 = (__pyx_v_recalculate_starting_peak != 0);
      if (__pyx_t_1) {

        /* "ms_deisotope/_c/deconvoluter_base.pyx":916
 *                     # update_(self._find_next_putative_peak(
 *                     #     peak.mz, charge, i, 2 * error_tolerance))
 *                     self._find_next_putative_peak_inplace(             # <<<<<<<<<<<<<<
 *                         peak.mz, charge, target_peaks, i, 2 * error_tolerance)
 * 
 */
        __pyx_t_11.__pyx_n = 2;
        __pyx_t_11.step = __pyx_v_i;
        __pyx_t_11.tolerance = (2.0 * __pyx_v_error_tolerance);
        ((struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *)__pyx_v_self->__pyx_vtab)->_find_next_putative_peak_inplace(__pyx_v_self, __pyx_v_peak->__pyx_base.mz, __pyx_v_charge, __pyx_v_target_peaks, &__pyx_t_11); 

        /* "ms_deisotope/_c/deconvoluter_base.pyx":913
 * 
 * 
 *                 if recalculate_starting_peak:             # <<<<<<<<<<<<<<
 *                     # update_(self._find_next_putative_peak(
 *                     #     peak.mz, charge, i, 2 * error_tolerance))
 */
      }
      __pyx_L10_continue:;
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":919
 *                         peak.mz, charge, target_peaks, i, 2 * error_tolerance)
 * 
 *             if recalculate_starting_peak:             # <<<<<<<<<<<<<<
 *                 for i in range(min(left_search_limit, 2)):
 *                     # update_(self._find_next_putative_peak(
 */
    __pyx_t_1 = (__pyx_v_recalculate_starting_peak != 0);
    if (__pyx_t_1) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":920
 * 
 *             if recalculate_starting_peak:
 *                 for i in range(min(left_search_limit, 2)):             # <<<<<<<<<<<<<<
 *                     # update_(self._find_next_putative_peak(
 *                     #     peak.mz, charge, step=i, tolerance=2 * error_tolerance))
 */
      __pyx_t_12 = 2;
      __pyx_t_6 = __pyx_v_left_search_limit;
      if (((__pyx_t_12 < __pyx_t_6) != 0)) {
        __pyx_t_13 = __pyx_t_12;
      } else {
        __pyx_t_13 = __pyx_t_6;
      }
      __pyx_t_12 = __pyx_t_13;
      __pyx_t_13 = __pyx_t_12;
      for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_13; __pyx_t_8+=1) {
        __pyx_v_i = __pyx_t_8;

        /* "ms_deisotope/_c/deconvoluter_base.pyx":923
 *                     # update_(self._find_next_putative_peak(
 *                     #     peak.mz, charge, step=i, tolerance=2 * error_tolerance))
 *                     self._find_next_putative_peak_inplace(peak.mz, charge, target_peaks, step=i, tolerance=2 * error_tolerance)             # <<<<<<<<<<<<<<
 * 
 *         return target_peaks
 */
        __pyx_t_11.__pyx_n = 2;
        __pyx_t_11.step = __pyx_v_i;
        __pyx_t_11.tolerance = (2.0 * __pyx_v_error_tolerance);
        ((struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *)__pyx_v_self->__pyx_vtab)->_find_next_putative_peak_inplace(__pyx_v_self, __pyx_v_peak->__pyx_base.mz, __pyx_v_charge, __pyx_v_target_peaks, &__pyx_t_11); 
      }

      /* "ms_deisotope/_c/deconvoluter_base.pyx":919
 *                         peak.mz, charge, target_peaks, i, 2 * error_tolerance)
 * 
 *             if recalculate_starting_peak:             # <<<<<<<<<<<<<<
 *                 for i in range(min(left_search_limit, 2)):
 *                     # update_(self._find_next_putative_peak(
 */
    }
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":925
 *                     self._find_next_putative_peak_inplace(peak.mz, charge, target_peaks, step=i, tolerance=2 * error_tolerance)
 * 
 *         return target_peaks             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_target_peaks);
  __pyx_r = __pyx_v_target_peaks;
  goto __pyx_L0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":839
 * 
 * @cython.binding(True)
 * cpdef set _get_all_peak_charge_pairs(DeconvoluterBase self, FittedPeak peak, double error_tolerance=ERROR_TOLERANCE,             # <<<<<<<<<<<<<<
 *                                      object charge_range=(1, 8), int left_search_limit=3, int right_search_limit=3,
 *                                      bint recalculate_starting_peak=True, bint use_quick_charge=False):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base._get_all_peak_charge_pairs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_charge_iterator);
  __Pyx_XDECREF(__pyx_v_target_peaks);
  __Pyx_XDECREF((PyObject *)__pyx_v_prev_peak);
  __Pyx_XDECREF((PyObject *)__pyx_v_nxt_peak);
  __Pyx_XDECREF(__pyx_v_add_);
  __Pyx_XDECREF(__pyx_v_update_);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_3_get_all_peak_charge_pairs(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_2_get_all_peak_charge_pairs[] = "_get_all_peak_charge_pairs(DeconvoluterBase self, FittedPeak peak, double error_tolerance=ERROR_TOLERANCE, charge_range=(1, 8), int left_search_limit=3, int right_search_limit=3, bool recalculate_starting_peak=True, bool use_quick_charge=False) -> set\nConstruct the set of all unique candidate (monoisotopic peak, charge state) pairs using\n        the provided search parameters.\n\n        The search is performed using :func:`has_previous_peak_at_charge`, :func:`has_successor_peak_at_charge`,\n        :meth:`_find_previous_putative_peak`, and :meth:`_find_next_putative_peak`.\n\n        Parameters\n        ----------\n        peak : FittedPeak\n            The peak to start the search from\n        error_tolerance : float, optional\n            The parts-per-million error tolerance in m/z to search with. Defaults to ERROR_TOLERANCE\n        charge_range : tuple, optional\n            The range of charge states to consider. Defaults to (1, 8)\n        left_search_limit : int, optional\n            The number of steps to search to the left of `peak`. Defaults to 3\n        right_search_limit : int, optional\n            The number of steps to search to the right of `peak`. Defaults to 3\n        recalculate_starting_peak : bool, optional\n            Whether or not to re-calculate the putative starting peak m/z based upon nearby\n            peaks close to where isotopic peaks for `peak` should be. Defaults to True\n\n        Returns\n        -------\n        set\n            The set of all unique candidate (monoisotopic peak, charge state)\n        ";
static PyMethodDef __pyx_mdef_12ms_deisotope_2_c_17deconvoluter_base_3_get_all_peak_charge_pairs = {"_get_all_peak_charge_pairs", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_3_get_all_peak_charge_pairs, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_2_get_all_peak_charge_pairs};
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_3_get_all_peak_charge_pairs(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self = 0;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_peak = 0;
  double __pyx_v_error_tolerance;
  PyObject *__pyx_v_charge_range = 0;
  int __pyx_v_left_search_limit;
  int __pyx_v_right_search_limit;
  int __pyx_v_recalculate_starting_peak;
  int __pyx_v_use_quick_charge;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_all_peak_charge_pairs (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_peak,&__pyx_n_s_error_tolerance,&__pyx_n_s_charge_range,&__pyx_n_s_left_search_limit,&__pyx_n_s_right_search_limit,&__pyx_n_s_recalculate_starting_peak,&__pyx_n_s_use_quick_charge,0};
    PyObject* values[8] = {0,0,0,0,0,0,0,0};

    /* "ms_deisotope/_c/deconvoluter_base.pyx":840
 * @cython.binding(True)
 * cpdef set _get_all_peak_charge_pairs(DeconvoluterBase self, FittedPeak peak, double error_tolerance=ERROR_TOLERANCE,
 *                                      object charge_range=(1, 8), int left_search_limit=3, int right_search_limit=3,             # <<<<<<<<<<<<<<
 *                                      bint recalculate_starting_peak=True, bint use_quick_charge=False):
 *         """Construct the set of all unique candidate (monoisotopic peak, charge state) pairs using
 */
    values[3] = ((PyObject *)__pyx_tuple__16);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peak)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_get_all_peak_charge_pairs", 0, 2, 8, 1); __PYX_ERR(0, 839, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_error_tolerance);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_charge_range);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_left_search_limit);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_right_search_limit);
          if (value) { values[5] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_recalculate_starting_peak);
          if (value) { values[6] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_use_quick_charge);
          if (value) { values[7] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_get_all_peak_charge_pairs") < 0)) __PYX_ERR(0, 839, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = ((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *)values[0]);
    __pyx_v_peak = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)values[1]);
    if (values[2]) {
      __pyx_v_error_tolerance = __pyx_PyFloat_AsDouble(values[2]); if (unlikely((__pyx_v_error_tolerance == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 839, __pyx_L3_error)
    } else {
      __pyx_v_error_tolerance = __pyx_k__15;
    }
    __pyx_v_charge_range = values[3];
    if (values[4]) {
      __pyx_v_left_search_limit = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_left_search_limit == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 840, __pyx_L3_error)
    } else {
      __pyx_v_left_search_limit = ((int)3);
    }
    if (values[5]) {
      __pyx_v_right_search_limit = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_right_search_limit == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 840, __pyx_L3_error)
    } else {
      __pyx_v_right_search_limit = ((int)3);
    }
    if (values[6]) {
      __pyx_v_recalculate_starting_peak = __Pyx_PyObject_IsTrue(values[6]); if (unlikely((__pyx_v_recalculate_starting_peak == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 841, __pyx_L3_error)
    } else {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":841
 * cpdef set _get_all_peak_charge_pairs(DeconvoluterBase self, FittedPeak peak, double error_tolerance=ERROR_TOLERANCE,
 *                                      object charge_range=(1, 8), int left_search_limit=3, int right_search_limit=3,
 *                                      bint recalculate_starting_peak=True, bint use_quick_charge=False):             # <<<<<<<<<<<<<<
 *         """Construct the set of all unique candidate (monoisotopic peak, charge state) pairs using
 *         the provided search parameters.
 */
      __pyx_v_recalculate_starting_peak = ((int)1);
    }
    if (values[7]) {
      __pyx_v_use_quick_charge = __Pyx_PyObject_IsTrue(values[7]); if (unlikely((__pyx_v_use_quick_charge == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 841, __pyx_L3_error)
    } else {
      __pyx_v_use_quick_charge = ((int)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_get_all_peak_charge_pairs", 0, 2, 8, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 839, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base._get_all_peak_charge_pairs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_self), __pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase, 1, "self", 0))) __PYX_ERR(0, 839, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peak), __pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak, 1, "peak", 0))) __PYX_ERR(0, 839, __pyx_L1_error)
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_2_get_all_peak_charge_pairs(__pyx_self, __pyx_v_self, __pyx_v_peak, __pyx_v_error_tolerance, __pyx_v_charge_range, __pyx_v_left_search_limit, __pyx_v_right_search_limit, __pyx_v_recalculate_starting_peak, __pyx_v_use_quick_charge);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":839
 * 
 * @cython.binding(True)
 * cpdef set _get_all_peak_charge_pairs(DeconvoluterBase self, FittedPeak peak, double error_tolerance=ERROR_TOLERANCE,             # <<<<<<<<<<<<<<
 *                                      object charge_range=(1, 8), int left_search_limit=3, int right_search_limit=3,
 *                                      bint recalculate_starting_peak=True, bint use_quick_charge=False):
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_2_get_all_peak_charge_pairs(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_peak, double __pyx_v_error_tolerance, PyObject *__pyx_v_charge_range, int __pyx_v_left_search_limit, int __pyx_v_right_search_limit, int __pyx_v_recalculate_starting_peak, int __pyx_v_use_quick_charge) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base__get_all_peak_charge_pairs __pyx_t_2;
  __Pyx_RefNannySetupContext("_get_all_peak_charge_pairs", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 6;
  __pyx_t_2.error_tolerance = __pyx_v_error_tolerance;
  __pyx_t_2.charge_range = __pyx_v_charge_range;
  __pyx_t_2.left_search_limit = __pyx_v_left_search_limit;
  __pyx_t_2.right_search_limit = __pyx_v_right_search_limit;
  __pyx_t_2.recalculate_starting_peak = __pyx_v_recalculate_starting_peak;
  __pyx_t_2.use_quick_charge = __pyx_v_use_quick_charge;
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_17deconvoluter_base__get_all_peak_charge_pairs(__pyx_v_self, __pyx_v_peak, 0, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 839, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base._get_all_peak_charge_pairs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/deconvoluter_base.pyx":929
 * 
 * @cython.binding(True)
 * cpdef int _explore_local(DeconvoluterBase self, peak, error_tolerance=ERROR_TOLERANCE,             # <<<<<<<<<<<<<<
 *                          charge_range=(1, 8), left_search_limit=1, right_search_limit=0,
 *                          charge_carrier=PROTON, truncate_after=TRUNCATE_AFTER,
 */

static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_5_explore_local(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_12ms_deisotope_2_c_17deconvoluter_base__explore_local(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, PyObject *__pyx_v_peak, CYTHON_UNUSED int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base__explore_local *__pyx_optional_args) {
  PyObject *__pyx_v_error_tolerance = __pyx_k__17;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":930
 * @cython.binding(True)
 * cpdef int _explore_local(DeconvoluterBase self, peak, error_tolerance=ERROR_TOLERANCE,
 *                          charge_range=(1, 8), left_search_limit=1, right_search_limit=0,             # <<<<<<<<<<<<<<
 *                          charge_carrier=PROTON, truncate_after=TRUNCATE_AFTER,
 *                          ignore_below=IGNORE_BELOW):
 */
  PyObject *__pyx_v_charge_range = ((PyObject *)__pyx_tuple__16);
  PyObject *__pyx_v_left_search_limit = ((PyObject *)__pyx_int_1);
  PyObject *__pyx_v_charge_carrier = __pyx_k__18;
  PyObject *__pyx_v_truncate_after = __pyx_k__19;
  PyObject *__pyx_v_ignore_below = __pyx_k__20;
  PyObject *__pyx_v_hold = 0;
  PyObject *__pyx_v_results = 0;
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_fit = 0;
  size_t __pyx_v_n;
  size_t __pyx_v_i;
  size_t __pyx_v_stop;
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_12peak_network_PeakDependenceGraphBase *__pyx_v_peak_graph = 0;
  PyObject *__pyx_v_obj = NULL;
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_candidate = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  long __pyx_t_12;
  long __pyx_t_13;
  size_t __pyx_t_14;
  size_t __pyx_t_15;
  size_t __pyx_t_16;
  __Pyx_RefNannySetupContext("_explore_local", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_error_tolerance = __pyx_optional_args->error_tolerance;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_charge_range = __pyx_optional_args->charge_range;
        if (__pyx_optional_args->__pyx_n > 2) {
          __pyx_v_left_search_limit = __pyx_optional_args->left_search_limit;
          if (__pyx_optional_args->__pyx_n > 4) {
            __pyx_v_charge_carrier = __pyx_optional_args->charge_carrier;
            if (__pyx_optional_args->__pyx_n > 5) {
              __pyx_v_truncate_after = __pyx_optional_args->truncate_after;
              if (__pyx_optional_args->__pyx_n > 6) {
                __pyx_v_ignore_below = __pyx_optional_args->ignore_below;
              }
            }
          }
        }
      }
    }
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":976
 *         PeakDependenceGraphBase peak_graph
 * 
 *     results = self._fit_all_charge_states(             # <<<<<<<<<<<<<<
 *         peak, error_tolerance=error_tolerance, charge_range=charge_range, left_search_limit=left_search_limit,
 *         charge_carrier=charge_carrier, truncate_after=truncate_after, ignore_below=ignore_below)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_fit_all_charge_states); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 976, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":977
 * 
 *     results = self._fit_all_charge_states(
 *         peak, error_tolerance=error_tolerance, charge_range=charge_range, left_search_limit=left_search_limit,             # <<<<<<<<<<<<<<
 *         charge_carrier=charge_carrier, truncate_after=truncate_after, ignore_below=ignore_below)
 * 
 */
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 976, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_peak);
  __Pyx_GIVEREF(__pyx_v_peak);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_peak);
  __pyx_t_3 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 977, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_error_tolerance, __pyx_v_error_tolerance) < 0) __PYX_ERR(0, 977, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_charge_range, __pyx_v_charge_range) < 0) __PYX_ERR(0, 977, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_left_search_limit, __pyx_v_left_search_limit) < 0) __PYX_ERR(0, 977, __pyx_L1_error)

  /* "ms_deisotope/_c/deconvoluter_base.pyx":978
 *     results = self._fit_all_charge_states(
 *         peak, error_tolerance=error_tolerance, charge_range=charge_range, left_search_limit=left_search_limit,
 *         charge_carrier=charge_carrier, truncate_after=truncate_after, ignore_below=ignore_below)             # <<<<<<<<<<<<<<
 * 
 *     peak_graph = <PeakDependenceGraphBase>self.peak_dependency_network
 */
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_charge_carrier, __pyx_v_charge_carrier) < 0) __PYX_ERR(0, 977, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_truncate_after, __pyx_v_truncate_after) < 0) __PYX_ERR(0, 977, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_ignore_below, __pyx_v_ignore_below) < 0) __PYX_ERR(0, 977, __pyx_L1_error)

  /* "ms_deisotope/_c/deconvoluter_base.pyx":976
 *         PeakDependenceGraphBase peak_graph
 * 
 *     results = self._fit_all_charge_states(             # <<<<<<<<<<<<<<
 *         peak, error_tolerance=error_tolerance, charge_range=charge_range, left_search_limit=left_search_limit,
 *         charge_carrier=charge_carrier, truncate_after=truncate_after, ignore_below=ignore_below)
 */
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 976, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!(likely(PySet_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "set", Py_TYPE(__pyx_t_4)->tp_name), 0))) __PYX_ERR(0, 976, __pyx_L1_error)
  __pyx_v_results = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":980
 *         charge_carrier=charge_carrier, truncate_after=truncate_after, ignore_below=ignore_below)
 * 
 *     peak_graph = <PeakDependenceGraphBase>self.peak_dependency_network             # <<<<<<<<<<<<<<
 * 
 *     hold = set()
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_peak_dependency_network); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 980, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __pyx_t_4;
  __Pyx_INCREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_peak_graph = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_12peak_network_PeakDependenceGraphBase *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":982
 *     peak_graph = <PeakDependenceGraphBase>self.peak_dependency_network
 * 
 *     hold = set()             # <<<<<<<<<<<<<<
 *     for obj in results:
 *         fit = <IsotopicFitRecord>obj
 */
  __pyx_t_3 = PySet_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 982, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_hold = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":983
 * 
 *     hold = set()
 *     for obj in results:             # <<<<<<<<<<<<<<
 *         fit = <IsotopicFitRecord>obj
 *         if fit.charge > 1 and PyList_GET_SIZE(drop_placeholders(fit.experimental)) == 1:
 */
  __pyx_t_5 = 0;
  __pyx_t_4 = __Pyx_set_iterator(__pyx_v_results, 1, (&__pyx_t_6), (&__pyx_t_7)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 983, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_3);
  __pyx_t_3 = __pyx_t_4;
  __pyx_t_4 = 0;
  while (1) {
    __pyx_t_8 = __Pyx_set_iter_next(__pyx_t_3, __pyx_t_6, &__pyx_t_5, &__pyx_t_4, __pyx_t_7);
    if (unlikely(__pyx_t_8 == 0)) break;
    if (unlikely(__pyx_t_8 == -1)) __PYX_ERR(0, 983, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_XDECREF_SET(__pyx_v_obj, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":984
 *     hold = set()
 *     for obj in results:
 *         fit = <IsotopicFitRecord>obj             # <<<<<<<<<<<<<<
 *         if fit.charge > 1 and PyList_GET_SIZE(drop_placeholders(fit.experimental)) == 1:
 *             continue
 */
    __pyx_t_4 = __pyx_v_obj;
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_XDECREF_SET(__pyx_v_fit, ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":985
 *     for obj in results:
 *         fit = <IsotopicFitRecord>obj
 *         if fit.charge > 1 and PyList_GET_SIZE(drop_placeholders(fit.experimental)) == 1:             # <<<<<<<<<<<<<<
 *             continue
 *         hold.add(fit)
 */
    __pyx_t_10 = ((__pyx_v_fit->charge > 1) != 0);
    if (__pyx_t_10) {
    } else {
      __pyx_t_9 = __pyx_t_10;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_4 = __pyx_v_fit->experimental;
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_2 = __pyx_f_12ms_deisotope_2_c_17deconvoluter_base_drop_placeholders(((PyObject*)__pyx_t_4)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 985, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_10 = ((PyList_GET_SIZE(__pyx_t_2) == 1) != 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_9 = __pyx_t_10;
    __pyx_L6_bool_binop_done:;
    if (__pyx_t_9) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":986
 *         fit = <IsotopicFitRecord>obj
 *         if fit.charge > 1 and PyList_GET_SIZE(drop_placeholders(fit.experimental)) == 1:
 *             continue             # <<<<<<<<<<<<<<
 *         hold.add(fit)
 * 
 */
      goto __pyx_L3_continue;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":985
 *     for obj in results:
 *         fit = <IsotopicFitRecord>obj
 *         if fit.charge > 1 and PyList_GET_SIZE(drop_placeholders(fit.experimental)) == 1:             # <<<<<<<<<<<<<<
 *             continue
 *         hold.add(fit)
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":987
 *         if fit.charge > 1 and PyList_GET_SIZE(drop_placeholders(fit.experimental)) == 1:
 *             continue
 *         hold.add(fit)             # <<<<<<<<<<<<<<
 * 
 *     results = hold
 */
    __pyx_t_11 = PySet_Add(__pyx_v_hold, ((PyObject *)__pyx_v_fit)); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 987, __pyx_L1_error)
    __pyx_L3_continue:;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":989
 *         hold.add(fit)
 * 
 *     results = hold             # <<<<<<<<<<<<<<
 *     n = len(results)
 * 
 */
  __Pyx_INCREF(__pyx_v_hold);
  __Pyx_DECREF_SET(__pyx_v_results, __pyx_v_hold);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":990
 * 
 *     results = hold
 *     n = len(results)             # <<<<<<<<<<<<<<
 * 
 *     stop = max(min(n // 2, 100), 10)
 */
  __pyx_t_6 = PySet_GET_SIZE(__pyx_v_results); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 990, __pyx_L1_error)
  __pyx_v_n = __pyx_t_6;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":992
 *     n = len(results)
 * 
 *     stop = max(min(n // 2, 100), 10)             # <<<<<<<<<<<<<<
 *     if n == 0:
 *         return 0
 */
  __pyx_t_12 = 10;
  __pyx_t_13 = 0x64;
  __pyx_t_14 = (__pyx_v_n / 2);
  if (((__pyx_t_13 < __pyx_t_14) != 0)) {
    __pyx_t_15 = __pyx_t_13;
  } else {
    __pyx_t_15 = __pyx_t_14;
  }
  __pyx_t_14 = __pyx_t_15;
  if (((__pyx_t_12 > __pyx_t_14) != 0)) {
    __pyx_t_15 = __pyx_t_12;
  } else {
    __pyx_t_15 = __pyx_t_14;
  }
  __pyx_v_stop = __pyx_t_15;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":993
 * 
 *     stop = max(min(n // 2, 100), 10)
 *     if n == 0:             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  __pyx_t_9 = ((__pyx_v_n == 0) != 0);
  if (__pyx_t_9) {

    /* "ms_deisotope/_c/deconvoluter_base.pyx":994
 *     stop = max(min(n // 2, 100), 10)
 *     if n == 0:
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     for i in range(stop):
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":993
 * 
 *     stop = max(min(n // 2, 100), 10)
 *     if n == 0:             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":996
 *         return 0
 * 
 *     for i in range(stop):             # <<<<<<<<<<<<<<
 *         if len(results) == 0:
 *             break
 */
  __pyx_t_15 = __pyx_v_stop;
  __pyx_t_14 = __pyx_t_15;
  for (__pyx_t_16 = 0; __pyx_t_16 < __pyx_t_14; __pyx_t_16+=1) {
    __pyx_v_i = __pyx_t_16;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":997
 * 
 *     for i in range(stop):
 *         if len(results) == 0:             # <<<<<<<<<<<<<<
 *             break
 *         candidate = self.scorer.select.best(results)
 */
    __pyx_t_6 = PySet_GET_SIZE(__pyx_v_results); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 997, __pyx_L1_error)
    __pyx_t_9 = ((__pyx_t_6 == 0) != 0);
    if (__pyx_t_9) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":998
 *     for i in range(stop):
 *         if len(results) == 0:
 *             break             # <<<<<<<<<<<<<<
 *         candidate = self.scorer.select.best(results)
 *         if candidate is None:
 */
      goto __pyx_L10_break;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":997
 * 
 *     for i in range(stop):
 *         if len(results) == 0:             # <<<<<<<<<<<<<<
 *             break
 *         candidate = self.scorer.select.best(results)
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":999
 *         if len(results) == 0:
 *             break
 *         candidate = self.scorer.select.best(results)             # <<<<<<<<<<<<<<
 *         if candidate is None:
 *                 break
 */
    __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FitSelectorBase *)__pyx_v_self->scorer->select->__pyx_vtab)->best(__pyx_v_self->scorer->select, __pyx_v_results, 0)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 999, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_XDECREF_SET(__pyx_v_candidate, ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":1000
 *             break
 *         candidate = self.scorer.select.best(results)
 *         if candidate is None:             # <<<<<<<<<<<<<<
 *                 break
 *         peak_graph.add_fit_dependence(candidate)
 */
    __pyx_t_9 = (((PyObject *)__pyx_v_candidate) == Py_None);
    __pyx_t_10 = (__pyx_t_9 != 0);
    if (__pyx_t_10) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":1001
 *         candidate = self.scorer.select.best(results)
 *         if candidate is None:
 *                 break             # <<<<<<<<<<<<<<
 *         peak_graph.add_fit_dependence(candidate)
 *         results.discard(candidate)
 */
      goto __pyx_L10_break;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":1000
 *             break
 *         candidate = self.scorer.select.best(results)
 *         if candidate is None:             # <<<<<<<<<<<<<<
 *                 break
 *         peak_graph.add_fit_dependence(candidate)
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":1002
 *         if candidate is None:
 *                 break
 *         peak_graph.add_fit_dependence(candidate)             # <<<<<<<<<<<<<<
 *         results.discard(candidate)
 * 
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_12peak_network_PeakDependenceGraphBase *)__pyx_v_peak_graph->__pyx_vtab)->add_fit_dependence(__pyx_v_peak_graph, __pyx_v_candidate, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1002, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":1003
 *                 break
 *         peak_graph.add_fit_dependence(candidate)
 *         results.discard(candidate)             # <<<<<<<<<<<<<<
 * 
 *     return i
 */
    __pyx_t_11 = __Pyx_PySet_Discard(__pyx_v_results, ((PyObject *)__pyx_v_candidate)); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 1003, __pyx_L1_error)
  }
  __pyx_L10_break:;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":1005
 *         results.discard(candidate)
 * 
 *     return i             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_i;
  goto __pyx_L0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":929
 * 
 * @cython.binding(True)
 * cpdef int _explore_local(DeconvoluterBase self, peak, error_tolerance=ERROR_TOLERANCE,             # <<<<<<<<<<<<<<
 *                          charge_range=(1, 8), left_search_limit=1, right_search_limit=0,
 *                          charge_carrier=PROTON, truncate_after=TRUNCATE_AFTER,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("ms_deisotope._c.deconvoluter_base._explore_local", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_hold);
  __Pyx_XDECREF(__pyx_v_results);
  __Pyx_XDECREF((PyObject *)__pyx_v_fit);
  __Pyx_XDECREF((PyObject *)__pyx_v_peak_graph);
  __Pyx_XDECREF(__pyx_v_obj);
  __Pyx_XDECREF((PyObject *)__pyx_v_candidate);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_5_explore_local(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_4_explore_local[] = "_explore_local(DeconvoluterBase self, peak, error_tolerance=ERROR_TOLERANCE, charge_range=(1, 8), left_search_limit=1, right_search_limit=0, charge_carrier=PROTON, truncate_after=TRUNCATE_AFTER, ignore_below=IGNORE_BELOW) -> int\nGiven a peak, explore the local neighborhood for candidate isotopic fits and add each\n    fit above a threshold to the peak dependence graph.\n\n    The threshold assumes that a single peak's neighborhood will contain many, many fits, but\n    that only the top `n` scoring fits are worth considering. For now, `n` is fixed at 100 or\n    the half number of fits returned, whichever is larger. This is to prevent the fit graph\n    from growing out of control and wasting time storing impractical fits. Any fit added to\n    the graph will have to pass :attr:`scorer.select` as well, so weak fits will never be added,\n    regardless of how many fits are allowed to be inserted.\n\n    Parameters\n    ----------\n    peak : :class:`~.FittedPeak`\n        The peak to start the search from\n    error_tolerance : float, optional\n        The parts-per-million error tolerance in m/z to search with. Defaults to |ERROR_TOLERANCE|\n    charge_range : tuple, optional\n        The range of charge states to consider. Defaults to (1, 8)\n    left_search_limit : int, optional\n        The number of steps to search to the left of `peak`. Defaults to 1\n    right_search_limit : int, optional\n        The number of steps to search to the right of `peak`. Defaults to 0\n    charge_carrier : float, optional\n        The mass of the charge carrier. Defaults to |PROTON|\n    truncate_after : float, optional\n        The percent of intensity to ensure is included in a theoretical isotopic pattern\n        starting from the monoisotopic peak. This will cause theoretical isotopic patterns\n        to be truncated, excluding trailing peaks which do not contribute substantially to\n        the overall shape of the isotopic pattern.\n\n    Returns\n    -------\n    int""\n        The number of fits added to the graph\n    ";
static PyMethodDef __pyx_mdef_12ms_deisotope_2_c_17deconvoluter_base_5_explore_local = {"_explore_local", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_5_explore_local, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_4_explore_local};
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_5_explore_local(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self = 0;
  PyObject *__pyx_v_peak = 0;
  PyObject *__pyx_v_error_tolerance = 0;
  PyObject *__pyx_v_charge_range = 0;
  PyObject *__pyx_v_left_search_limit = 0;
  PyObject *__pyx_v_right_search_limit = 0;
  PyObject *__pyx_v_charge_carrier = 0;
  PyObject *__pyx_v_truncate_after = 0;
  PyObject *__pyx_v_ignore_below = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_explore_local (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_peak,&__pyx_n_s_error_tolerance,&__pyx_n_s_charge_range,&__pyx_n_s_left_search_limit,&__pyx_n_s_right_search_limit,&__pyx_n_s_charge_carrier,&__pyx_n_s_truncate_after,&__pyx_n_s_ignore_below,0};
    PyObject* values[9] = {0,0,0,0,0,0,0,0,0};
    values[2] = __pyx_k__17;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":930
 * @cython.binding(True)
 * cpdef int _explore_local(DeconvoluterBase self, peak, error_tolerance=ERROR_TOLERANCE,
 *                          charge_range=(1, 8), left_search_limit=1, right_search_limit=0,             # <<<<<<<<<<<<<<
 *                          charge_carrier=PROTON, truncate_after=TRUNCATE_AFTER,
 *                          ignore_below=IGNORE_BELOW):
 */
    values[3] = ((PyObject *)__pyx_tuple__16);
    values[4] = ((PyObject *)__pyx_int_1);
    values[5] = ((PyObject *)__pyx_int_0);
    values[6] = __pyx_k__18;
    values[7] = __pyx_k__19;
    values[8] = __pyx_k__20;
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peak)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_explore_local", 0, 2, 9, 1); __PYX_ERR(0, 929, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_error_tolerance);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_charge_range);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_left_search_limit);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_right_search_limit);
          if (value) { values[5] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_charge_carrier);
          if (value) { values[6] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_truncate_after);
          if (value) { values[7] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ignore_below);
          if (value) { values[8] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_explore_local") < 0)) __PYX_ERR(0, 929, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = ((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *)values[0]);
    __pyx_v_peak = values[1];
    __pyx_v_error_tolerance = values[2];
    __pyx_v_charge_range = values[3];
    __pyx_v_left_search_limit = values[4];
    __pyx_v_right_search_limit = values[5];
    __pyx_v_charge_carrier = values[6];
    __pyx_v_truncate_after = values[7];
    __pyx_v_ignore_below = values[8];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_explore_local", 0, 2, 9, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 929, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base._explore_local", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_self), __pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase, 1, "self", 0))) __PYX_ERR(0, 929, __pyx_L1_error)
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_4_explore_local(__pyx_self, __pyx_v_self, __pyx_v_peak, __pyx_v_error_tolerance, __pyx_v_charge_range, __pyx_v_left_search_limit, __pyx_v_right_search_limit, __pyx_v_charge_carrier, __pyx_v_truncate_after, __pyx_v_ignore_below);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":929
 * 
 * @cython.binding(True)
 * cpdef int _explore_local(DeconvoluterBase self, peak, error_tolerance=ERROR_TOLERANCE,             # <<<<<<<<<<<<<<
 *                          charge_range=(1, 8), left_search_limit=1, right_search_limit=0,
 *                          charge_carrier=PROTON, truncate_after=TRUNCATE_AFTER,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_4_explore_local(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, PyObject *__pyx_v_peak, PyObject *__pyx_v_error_tolerance, PyObject *__pyx_v_charge_range, PyObject *__pyx_v_left_search_limit, PyObject *__pyx_v_right_search_limit, PyObject *__pyx_v_charge_carrier, PyObject *__pyx_v_truncate_after, PyObject *__pyx_v_ignore_below) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base__explore_local __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("_explore_local", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 7;
  __pyx_t_2.error_tolerance = __pyx_v_error_tolerance;
  __pyx_t_2.charge_range = __pyx_v_charge_range;
  __pyx_t_2.left_search_limit = __pyx_v_left_search_limit;
  __pyx_t_2.right_search_limit = __pyx_v_right_search_limit;
  __pyx_t_2.charge_carrier = __pyx_v_charge_carrier;
  __pyx_t_2.truncate_after = __pyx_v_truncate_after;
  __pyx_t_2.ignore_below = __pyx_v_ignore_below;
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_17deconvoluter_base__explore_local(__pyx_v_self, __pyx_v_peak, 0, &__pyx_t_2); 
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 929, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base._explore_local", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/deconvoluter_base.pyx":1009
 * 
 * @cython.binding(True)
 * cpdef populate_graph(DeconvoluterBase self, error_tolerance=ERROR_TOLERANCE, charge_range=(1, 8),             # <<<<<<<<<<<<<<
 *                      left_search_limit=1, right_search_limit=0, charge_carrier=PROTON,
 *                      truncate_after=TRUNCATE_AFTER, ignore_below=IGNORE_BELOW):
 */

static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_7populate_graph(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_populate_graph(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, CYTHON_UNUSED int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_populate_graph *__pyx_optional_args) {
  PyObject *__pyx_v_error_tolerance = __pyx_k__21;
  PyObject *__pyx_v_charge_range = ((PyObject *)__pyx_tuple__16);
  PyObject *__pyx_v_left_search_limit = ((PyObject *)__pyx_int_1);
  PyObject *__pyx_v_right_search_limit = ((PyObject *)__pyx_int_0);
  PyObject *__pyx_v_charge_carrier = __pyx_k__22;
  PyObject *__pyx_v_truncate_after = __pyx_k__23;
  PyObject *__pyx_v_ignore_below = __pyx_k__24;
  size_t __pyx_v_i;
  size_t __pyx_v_n;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_peak = 0;
  PyObject *__pyx_v__priority_map = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base__explore_local __pyx_t_9;
  __Pyx_RefNannySetupContext("populate_graph", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_error_tolerance = __pyx_optional_args->error_tolerance;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_charge_range = __pyx_optional_args->charge_range;
        if (__pyx_optional_args->__pyx_n > 2) {
          __pyx_v_left_search_limit = __pyx_optional_args->left_search_limit;
          if (__pyx_optional_args->__pyx_n > 3) {
            __pyx_v_right_search_limit = __pyx_optional_args->right_search_limit;
            if (__pyx_optional_args->__pyx_n > 4) {
              __pyx_v_charge_carrier = __pyx_optional_args->charge_carrier;
              if (__pyx_optional_args->__pyx_n > 5) {
                __pyx_v_truncate_after = __pyx_optional_args->truncate_after;
                if (__pyx_optional_args->__pyx_n > 6) {
                  __pyx_v_ignore_below = __pyx_optional_args->ignore_below;
                }
              }
            }
          }
        }
      }
    }
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":1038
 *         FittedPeak peak
 *         dict _priority_map
 *     _priority_map = self._priority_map             # <<<<<<<<<<<<<<
 *     n = self.peaklist.get_size()
 *     for i in range(n):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_priority_map); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1038, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyDict_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 1038, __pyx_L1_error)
  __pyx_v__priority_map = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":1039
 *         dict _priority_map
 *     _priority_map = self._priority_map
 *     n = self.peaklist.get_size()             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         peak = self.peaklist.getitem(i)
 */
  __pyx_v_n = ((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_self->peaklist->__pyx_vtab)->get_size(__pyx_v_self->peaklist);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":1040
 *     _priority_map = self._priority_map
 *     n = self.peaklist.get_size()
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         peak = self.peaklist.getitem(i)
 *         if peak in _priority_map or peak.intensity < self.minimum_intensity:
 */
  __pyx_t_2 = __pyx_v_n;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":1041
 *     n = self.peaklist.get_size()
 *     for i in range(n):
 *         peak = self.peaklist.getitem(i)             # <<<<<<<<<<<<<<
 *         if peak in _priority_map or peak.intensity < self.minimum_intensity:
 *             continue
 */
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_self->peaklist->__pyx_vtab)->getitem(__pyx_v_self->peaklist, __pyx_v_i)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1041, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_peak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":1042
 *     for i in range(n):
 *         peak = self.peaklist.getitem(i)
 *         if peak in _priority_map or peak.intensity < self.minimum_intensity:             # <<<<<<<<<<<<<<
 *             continue
 *         n = _explore_local(self,
 */
    if (unlikely(__pyx_v__priority_map == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 1042, __pyx_L1_error)
    }
    __pyx_t_6 = (__Pyx_PyDict_ContainsTF(((PyObject *)__pyx_v_peak), __pyx_v__priority_map, Py_EQ)); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 1042, __pyx_L1_error)
    __pyx_t_7 = (__pyx_t_6 != 0);
    if (!__pyx_t_7) {
    } else {
      __pyx_t_5 = __pyx_t_7;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_7 = ((__pyx_v_peak->__pyx_base.intensity < __pyx_v_self->minimum_intensity) != 0);
    __pyx_t_5 = __pyx_t_7;
    __pyx_L6_bool_binop_done:;
    if (__pyx_t_5) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":1043
 *         peak = self.peaklist.getitem(i)
 *         if peak in _priority_map or peak.intensity < self.minimum_intensity:
 *             continue             # <<<<<<<<<<<<<<
 *         n = _explore_local(self,
 *             peak, error_tolerance=error_tolerance, charge_range=charge_range,
 */
      goto __pyx_L3_continue;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":1042
 *     for i in range(n):
 *         peak = self.peaklist.getitem(i)
 *         if peak in _priority_map or peak.intensity < self.minimum_intensity:             # <<<<<<<<<<<<<<
 *             continue
 *         n = _explore_local(self,
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":1044
 *         if peak in _priority_map or peak.intensity < self.minimum_intensity:
 *             continue
 *         n = _explore_local(self,             # <<<<<<<<<<<<<<
 *             peak, error_tolerance=error_tolerance, charge_range=charge_range,
 *             left_search_limit=left_search_limit, right_search_limit=right_search_limit,
 */
    __pyx_t_9.__pyx_n = 7;
    __pyx_t_9.error_tolerance = __pyx_v_error_tolerance;
    __pyx_t_9.charge_range = __pyx_v_charge_range;
    __pyx_t_9.left_search_limit = __pyx_v_left_search_limit;
    __pyx_t_9.right_search_limit = __pyx_v_right_search_limit;
    __pyx_t_9.charge_carrier = __pyx_v_charge_carrier;
    __pyx_t_9.truncate_after = __pyx_v_truncate_after;
    __pyx_t_9.ignore_below = __pyx_v_ignore_below;
    __pyx_t_8 = __pyx_f_12ms_deisotope_2_c_17deconvoluter_base__explore_local(__pyx_v_self, ((PyObject *)__pyx_v_peak), 0, &__pyx_t_9); 
    __pyx_v_n = __pyx_t_8;
    __pyx_L3_continue:;
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":1009
 * 
 * @cython.binding(True)
 * cpdef populate_graph(DeconvoluterBase self, error_tolerance=ERROR_TOLERANCE, charge_range=(1, 8),             # <<<<<<<<<<<<<<
 *                      left_search_limit=1, right_search_limit=0, charge_carrier=PROTON,
 *                      truncate_after=TRUNCATE_AFTER, ignore_below=IGNORE_BELOW):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.populate_graph", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);
  __Pyx_XDECREF(__pyx_v__priority_map);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_7populate_graph(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_6populate_graph[] = "populate_graph(DeconvoluterBase self, error_tolerance=ERROR_TOLERANCE, charge_range=(1, 8), left_search_limit=1, right_search_limit=0, charge_carrier=PROTON, truncate_after=TRUNCATE_AFTER, ignore_below=IGNORE_BELOW)\nVisit each experimental peak and execute :meth:`_explore_local` on it with the provided\n    parameters, populating the peak dependence graph with all viable candidates.\n\n    Parameters\n    ----------\n    peak : :class:`~.FittedPeak`\n    error_tolerance : float, optional\n        The parts-per-million error tolerance in m/z to search with. Defaults to |ERROR_TOLERANCE|\n    charge_range : tuple, optional\n        The range of charge states to consider. Defaults to (1, 8)\n    left_search_limit : int, optional\n        The number of steps to search to the left of `peak`. Defaults to 1\n    right_search_limit : int, optional\n        The number of steps to search to the right of `peak`. Defaults to 0\n    charge_carrier : float, optional\n        The mass of the charge carrier. Defaults to |PROTON|\n    truncate_after : float, optional\n        The percent of intensity to ensure is included in a theoretical isotopic pattern\n        starting from the monoisotopic peak. This will cause theoretical isotopic patterns\n        to be truncated, excluding trailing peaks which do not contribute substantially to\n        the overall shape of the isotopic pattern.\n    ";
static PyMethodDef __pyx_mdef_12ms_deisotope_2_c_17deconvoluter_base_7populate_graph = {"populate_graph", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_7populate_graph, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_6populate_graph};
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_7populate_graph(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self = 0;
  PyObject *__pyx_v_error_tolerance = 0;
  PyObject *__pyx_v_charge_range = 0;
  PyObject *__pyx_v_left_search_limit = 0;
  PyObject *__pyx_v_right_search_limit = 0;
  PyObject *__pyx_v_charge_carrier = 0;
  PyObject *__pyx_v_truncate_after = 0;
  PyObject *__pyx_v_ignore_below = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("populate_graph (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_error_tolerance,&__pyx_n_s_charge_range,&__pyx_n_s_left_search_limit,&__pyx_n_s_right_search_limit,&__pyx_n_s_charge_carrier,&__pyx_n_s_truncate_after,&__pyx_n_s_ignore_below,0};
    PyObject* values[8] = {0,0,0,0,0,0,0,0};
    values[1] = __pyx_k__21;
    values[2] = ((PyObject *)__pyx_tuple__16);
    values[3] = ((PyObject *)__pyx_int_1);
    values[4] = ((PyObject *)__pyx_int_0);
    values[5] = __pyx_k__22;
    values[6] = __pyx_k__23;
    values[7] = __pyx_k__24;
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_error_tolerance);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_charge_range);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_left_search_limit);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_right_search_limit);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_charge_carrier);
          if (value) { values[5] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_truncate_after);
          if (value) { values[6] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ignore_below);
          if (value) { values[7] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "populate_graph") < 0)) __PYX_ERR(0, 1009, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = ((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *)values[0]);
    __pyx_v_error_tolerance = values[1];
    __pyx_v_charge_range = values[2];
    __pyx_v_left_search_limit = values[3];
    __pyx_v_right_search_limit = values[4];
    __pyx_v_charge_carrier = values[5];
    __pyx_v_truncate_after = values[6];
    __pyx_v_ignore_below = values[7];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("populate_graph", 0, 1, 8, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1009, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.populate_graph", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_self), __pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase, 1, "self", 0))) __PYX_ERR(0, 1009, __pyx_L1_error)
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_6populate_graph(__pyx_self, __pyx_v_self, __pyx_v_error_tolerance, __pyx_v_charge_range, __pyx_v_left_search_limit, __pyx_v_right_search_limit, __pyx_v_charge_carrier, __pyx_v_truncate_after, __pyx_v_ignore_below);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_6populate_graph(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_self, PyObject *__pyx_v_error_tolerance, PyObject *__pyx_v_charge_range, PyObject *__pyx_v_left_search_limit, PyObject *__pyx_v_right_search_limit, PyObject *__pyx_v_charge_carrier, PyObject *__pyx_v_truncate_after, PyObject *__pyx_v_ignore_below) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_populate_graph __pyx_t_2;
  __Pyx_RefNannySetupContext("populate_graph", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 7;
  __pyx_t_2.error_tolerance = __pyx_v_error_tolerance;
  __pyx_t_2.charge_range = __pyx_v_charge_range;
  __pyx_t_2.left_search_limit = __pyx_v_left_search_limit;
  __pyx_t_2.right_search_limit = __pyx_v_right_search_limit;
  __pyx_t_2.charge_carrier = __pyx_v_charge_carrier;
  __pyx_t_2.truncate_after = __pyx_v_truncate_after;
  __pyx_t_2.ignore_below = __pyx_v_ignore_below;
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_17deconvoluter_base_populate_graph(__pyx_v_self, 0, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1009, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.populate_graph", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_deisotope/_c/deconvoluter_base.pyx":1052
 * 
 * @cython.binding(True)
 * cpdef DeconvolutedPeak _make_deconvoluted_peak(self, IsotopicFitRecord fit, double charge_carrier):             # <<<<<<<<<<<<<<
 *     '''Helper method to package a finished :class:`~.IsotopicFitRecord` into a :class:`~.DeconvolutedPeak`
 * 
 */

static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_9_make_deconvoluted_peak(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base__make_deconvoluted_peak(CYTHON_UNUSED PyObject *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_fit, double __pyx_v_charge_carrier, CYTHON_UNUSED int __pyx_skip_dispatch) {
  double __pyx_v_score;
  double __pyx_v_total_abundance;
  double __pyx_v_signal_to_noise;
  double __pyx_v_full_width_at_half_max;
  double __pyx_v_most_abundant_mz;
  double __pyx_v_most_abundant_mz_intensity;
  double __pyx_v_average_mz;
  double __pyx_v_area;
  double __pyx_v_a_to_a2_ratio;
  int __pyx_v_charge;
  PyObject *__pyx_v_eid = 0;
  PyObject *__pyx_v_rep_eid = 0;
  PyObject *__pyx_v_envelope = 0;
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *__pyx_v_tid = 0;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_fpeak = 0;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_reference_peak = 0;
  size_t __pyx_v_i;
  size_t __pyx_v_n;
  double __pyx_v_monoisotopic_mass;
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *__pyx_v_dpeak = NULL;
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  size_t __pyx_t_5;
  size_t __pyx_t_6;
  PyObject *__pyx_t_7;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  double __pyx_t_12;
  struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_neutral_mass __pyx_t_13;
  __Pyx_RefNannySetupContext("_make_deconvoluted_peak", 0);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":1076
 *         size_t i, n
 * 
 *     score = fit.score             # <<<<<<<<<<<<<<
 *     charge = fit.charge
 *     eid = fit.experimental
 */
  __pyx_t_1 = __pyx_v_fit->score;
  __pyx_v_score = __pyx_t_1;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":1077
 * 
 *     score = fit.score
 *     charge = fit.charge             # <<<<<<<<<<<<<<
 *     eid = fit.experimental
 *     tid = fit.theoretical
 */
  __pyx_t_2 = __pyx_v_fit->charge;
  __pyx_v_charge = __pyx_t_2;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":1078
 *     score = fit.score
 *     charge = fit.charge
 *     eid = fit.experimental             # <<<<<<<<<<<<<<
 *     tid = fit.theoretical
 *     rep_eid = drop_placeholders(eid)
 */
  __pyx_t_3 = __pyx_v_fit->experimental;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_eid = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":1079
 *     charge = fit.charge
 *     eid = fit.experimental
 *     tid = fit.theoretical             # <<<<<<<<<<<<<<
 *     rep_eid = drop_placeholders(eid)
 * 
 */
  __pyx_t_3 = ((PyObject *)__pyx_v_fit->theoretical);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_tid = ((struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":1080
 *     eid = fit.experimental
 *     tid = fit.theoretical
 *     rep_eid = drop_placeholders(eid)             # <<<<<<<<<<<<<<
 * 
 *     total_abundance = 0
 */
  __pyx_t_3 = __pyx_f_12ms_deisotope_2_c_17deconvoluter_base_drop_placeholders(__pyx_v_eid); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1080, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_rep_eid = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":1082
 *     rep_eid = drop_placeholders(eid)
 * 
 *     total_abundance = 0             # <<<<<<<<<<<<<<
 *     signal_to_noise = 0
 *     full_width_at_half_max = 0
 */
  __pyx_v_total_abundance = 0.0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":1083
 * 
 *     total_abundance = 0
 *     signal_to_noise = 0             # <<<<<<<<<<<<<<
 *     full_width_at_half_max = 0
 *     most_abundant_mz = 0
 */
  __pyx_v_signal_to_noise = 0.0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":1084
 *     total_abundance = 0
 *     signal_to_noise = 0
 *     full_width_at_half_max = 0             # <<<<<<<<<<<<<<
 *     most_abundant_mz = 0
 *     most_abundant_mz_intensity = 0
 */
  __pyx_v_full_width_at_half_max = 0.0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":1085
 *     signal_to_noise = 0
 *     full_width_at_half_max = 0
 *     most_abundant_mz = 0             # <<<<<<<<<<<<<<
 *     most_abundant_mz_intensity = 0
 *     average_mz = 0
 */
  __pyx_v_most_abundant_mz = 0.0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":1086
 *     full_width_at_half_max = 0
 *     most_abundant_mz = 0
 *     most_abundant_mz_intensity = 0             # <<<<<<<<<<<<<<
 *     average_mz = 0
 *     area = 0
 */
  __pyx_v_most_abundant_mz_intensity = 0.0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":1087
 *     most_abundant_mz = 0
 *     most_abundant_mz_intensity = 0
 *     average_mz = 0             # <<<<<<<<<<<<<<
 *     area = 0
 *     envelope = []
 */
  __pyx_v_average_mz = 0.0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":1088
 *     most_abundant_mz_intensity = 0
 *     average_mz = 0
 *     area = 0             # <<<<<<<<<<<<<<
 *     envelope = []
 *     n = PyList_GET_SIZE(rep_eid)
 */
  __pyx_v_area = 0.0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":1089
 *     average_mz = 0
 *     area = 0
 *     envelope = []             # <<<<<<<<<<<<<<
 *     n = PyList_GET_SIZE(rep_eid)
 *     for i in range(n):
 */
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1089, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_envelope = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":1090
 *     area = 0
 *     envelope = []
 *     n = PyList_GET_SIZE(rep_eid)             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         fpeak = <FittedPeak>PyList_GET_ITEM(rep_eid, i)
 */
  __pyx_v_n = PyList_GET_SIZE(__pyx_v_rep_eid);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":1091
 *     envelope = []
 *     n = PyList_GET_SIZE(rep_eid)
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         fpeak = <FittedPeak>PyList_GET_ITEM(rep_eid, i)
 *         if fpeak.intensity > most_abundant_mz_intensity:
 */
  __pyx_t_4 = __pyx_v_n;
  __pyx_t_5 = __pyx_t_4;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_i = __pyx_t_6;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":1092
 *     n = PyList_GET_SIZE(rep_eid)
 *     for i in range(n):
 *         fpeak = <FittedPeak>PyList_GET_ITEM(rep_eid, i)             # <<<<<<<<<<<<<<
 *         if fpeak.intensity > most_abundant_mz_intensity:
 *             most_abundant_mz_intensity = fpeak.intensity
 */
    __pyx_t_7 = PyList_GET_ITEM(__pyx_v_rep_eid, __pyx_v_i);
    __pyx_t_3 = ((PyObject *)__pyx_t_7);
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_XDECREF_SET(__pyx_v_fpeak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":1093
 *     for i in range(n):
 *         fpeak = <FittedPeak>PyList_GET_ITEM(rep_eid, i)
 *         if fpeak.intensity > most_abundant_mz_intensity:             # <<<<<<<<<<<<<<
 *             most_abundant_mz_intensity = fpeak.intensity
 *             most_abundant_mz = fpeak.mz
 */
    __pyx_t_8 = ((__pyx_v_fpeak->__pyx_base.intensity > __pyx_v_most_abundant_mz_intensity) != 0);
    if (__pyx_t_8) {

      /* "ms_deisotope/_c/deconvoluter_base.pyx":1094
 *         fpeak = <FittedPeak>PyList_GET_ITEM(rep_eid, i)
 *         if fpeak.intensity > most_abundant_mz_intensity:
 *             most_abundant_mz_intensity = fpeak.intensity             # <<<<<<<<<<<<<<
 *             most_abundant_mz = fpeak.mz
 *         total_abundance += fpeak.intensity
 */
      __pyx_t_1 = __pyx_v_fpeak->__pyx_base.intensity;
      __pyx_v_most_abundant_mz_intensity = __pyx_t_1;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":1095
 *         if fpeak.intensity > most_abundant_mz_intensity:
 *             most_abundant_mz_intensity = fpeak.intensity
 *             most_abundant_mz = fpeak.mz             # <<<<<<<<<<<<<<
 *         total_abundance += fpeak.intensity
 *         average_mz += fpeak.mz
 */
      __pyx_t_1 = __pyx_v_fpeak->__pyx_base.mz;
      __pyx_v_most_abundant_mz = __pyx_t_1;

      /* "ms_deisotope/_c/deconvoluter_base.pyx":1093
 *     for i in range(n):
 *         fpeak = <FittedPeak>PyList_GET_ITEM(rep_eid, i)
 *         if fpeak.intensity > most_abundant_mz_intensity:             # <<<<<<<<<<<<<<
 *             most_abundant_mz_intensity = fpeak.intensity
 *             most_abundant_mz = fpeak.mz
 */
    }

    /* "ms_deisotope/_c/deconvoluter_base.pyx":1096
 *             most_abundant_mz_intensity = fpeak.intensity
 *             most_abundant_mz = fpeak.mz
 *         total_abundance += fpeak.intensity             # <<<<<<<<<<<<<<
 *         average_mz += fpeak.mz
 *         area += fpeak.area
 */
    __pyx_v_total_abundance = (__pyx_v_total_abundance + __pyx_v_fpeak->__pyx_base.intensity);

    /* "ms_deisotope/_c/deconvoluter_base.pyx":1097
 *             most_abundant_mz = fpeak.mz
 *         total_abundance += fpeak.intensity
 *         average_mz += fpeak.mz             # <<<<<<<<<<<<<<
 *         area += fpeak.area
 *         signal_to_noise += fpeak.signal_to_noise
 */
    __pyx_v_average_mz = (__pyx_v_average_mz + __pyx_v_fpeak->__pyx_base.mz);

    /* "ms_deisotope/_c/deconvoluter_base.pyx":1098
 *         total_abundance += fpeak.intensity
 *         average_mz += fpeak.mz
 *         area += fpeak.area             # <<<<<<<<<<<<<<
 *         signal_to_noise += fpeak.signal_to_noise
 *         full_width_at_half_max += fpeak.full_width_at_half_max
 */
    __pyx_v_area = (__pyx_v_area + __pyx_v_fpeak->__pyx_base.area);

    /* "ms_deisotope/_c/deconvoluter_base.pyx":1099
 *         average_mz += fpeak.mz
 *         area += fpeak.area
 *         signal_to_noise += fpeak.signal_to_noise             # <<<<<<<<<<<<<<
 *         full_width_at_half_max += fpeak.full_width_at_half_max
 *     full_width_at_half_max /= n
 */
    __pyx_v_signal_to_noise = (__pyx_v_signal_to_noise + __pyx_v_fpeak->signal_to_noise);

    /* "ms_deisotope/_c/deconvoluter_base.pyx":1100
 *         area += fpeak.area
 *         signal_to_noise += fpeak.signal_to_noise
 *         full_width_at_half_max += fpeak.full_width_at_half_max             # <<<<<<<<<<<<<<
 *     full_width_at_half_max /= n
 *     signal_to_noise /= n
 */
    __pyx_v_full_width_at_half_max = (__pyx_v_full_width_at_half_max + __pyx_v_fpeak->full_width_at_half_max);
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":1101
 *         signal_to_noise += fpeak.signal_to_noise
 *         full_width_at_half_max += fpeak.full_width_at_half_max
 *     full_width_at_half_max /= n             # <<<<<<<<<<<<<<
 *     signal_to_noise /= n
 *     average_mz /= n
 */
  if (unlikely(__pyx_v_n == 0)) {
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    __PYX_ERR(0, 1101, __pyx_L1_error)
  }
  __pyx_v_full_width_at_half_max = (__pyx_v_full_width_at_half_max / __pyx_v_n);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":1102
 *         full_width_at_half_max += fpeak.full_width_at_half_max
 *     full_width_at_half_max /= n
 *     signal_to_noise /= n             # <<<<<<<<<<<<<<
 *     average_mz /= n
 * 
 */
  if (unlikely(__pyx_v_n == 0)) {
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    __PYX_ERR(0, 1102, __pyx_L1_error)
  }
  __pyx_v_signal_to_noise = (__pyx_v_signal_to_noise / __pyx_v_n);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":1103
 *     full_width_at_half_max /= n
 *     signal_to_noise /= n
 *     average_mz /= n             # <<<<<<<<<<<<<<
 * 
 *     n = PyList_GET_SIZE(eid)
 */
  if (unlikely(__pyx_v_n == 0)) {
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    __PYX_ERR(0, 1103, __pyx_L1_error)
  }
  __pyx_v_average_mz = (__pyx_v_average_mz / __pyx_v_n);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":1105
 *     average_mz /= n
 * 
 *     n = PyList_GET_SIZE(eid)             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         fpeak = <FittedPeak>PyList_GET_ITEM(eid, i)
 */
  __pyx_v_n = PyList_GET_SIZE(__pyx_v_eid);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":1106
 * 
 *     n = PyList_GET_SIZE(eid)
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         fpeak = <FittedPeak>PyList_GET_ITEM(eid, i)
 *         envelope.append((fpeak.mz, fpeak.intensity))
 */
  __pyx_t_4 = __pyx_v_n;
  __pyx_t_5 = __pyx_t_4;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_i = __pyx_t_6;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":1107
 *     n = PyList_GET_SIZE(eid)
 *     for i in range(n):
 *         fpeak = <FittedPeak>PyList_GET_ITEM(eid, i)             # <<<<<<<<<<<<<<
 *         envelope.append((fpeak.mz, fpeak.intensity))
 * 
 */
    __pyx_t_7 = PyList_GET_ITEM(__pyx_v_eid, __pyx_v_i);
    __pyx_t_3 = ((PyObject *)__pyx_t_7);
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_XDECREF_SET(__pyx_v_fpeak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "ms_deisotope/_c/deconvoluter_base.pyx":1108
 *     for i in range(n):
 *         fpeak = <FittedPeak>PyList_GET_ITEM(eid, i)
 *         envelope.append((fpeak.mz, fpeak.intensity))             # <<<<<<<<<<<<<<
 * 
 *     a_to_a2_ratio = 0
 */
    __pyx_t_3 = PyFloat_FromDouble(__pyx_v_fpeak->__pyx_base.mz); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1108, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_9 = PyFloat_FromDouble(__pyx_v_fpeak->__pyx_base.intensity); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1108, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1108, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_9);
    __pyx_t_3 = 0;
    __pyx_t_9 = 0;
    __pyx_t_11 = __Pyx_PyList_Append(__pyx_v_envelope, __pyx_t_10); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 1108, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":1110
 *         envelope.append((fpeak.mz, fpeak.intensity))
 * 
 *     a_to_a2_ratio = 0             # <<<<<<<<<<<<<<
 *     if tid.get_size() > 2:
 *         a_to_a2_ratio = tid.get(0).intensity / tid.get(2).intensity
 */
  __pyx_v_a_to_a2_ratio = 0.0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":1111
 * 
 *     a_to_a2_ratio = 0
 *     if tid.get_size() > 2:             # <<<<<<<<<<<<<<
 *         a_to_a2_ratio = tid.get(0).intensity / tid.get(2).intensity
 * 
 */
  __pyx_t_8 = ((((struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *)__pyx_v_tid->__pyx_vtab)->get_size(__pyx_v_tid) > 2) != 0);
  if (__pyx_t_8) {

    /* "ms_deisotope/_c/deconvoluter_base.pyx":1112
 *     a_to_a2_ratio = 0
 *     if tid.get_size() > 2:
 *         a_to_a2_ratio = tid.get(0).intensity / tid.get(2).intensity             # <<<<<<<<<<<<<<
 * 
 *     monoisotopic_mass = neutral_mass(
 */
    __pyx_t_10 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *)__pyx_v_tid->__pyx_vtab)->get(__pyx_v_tid, 0)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1112, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_1 = ((struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *)__pyx_t_10)->intensity;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *)__pyx_v_tid->__pyx_vtab)->get(__pyx_v_tid, 2)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1112, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_12 = ((struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *)__pyx_t_10)->intensity;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(__pyx_t_12 == 0)) {
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      __PYX_ERR(0, 1112, __pyx_L1_error)
    }
    __pyx_v_a_to_a2_ratio = (__pyx_t_1 / __pyx_t_12);

    /* "ms_deisotope/_c/deconvoluter_base.pyx":1111
 * 
 *     a_to_a2_ratio = 0
 *     if tid.get_size() > 2:             # <<<<<<<<<<<<<<
 *         a_to_a2_ratio = tid.get(0).intensity / tid.get(2).intensity
 * 
 */
  }

  /* "ms_deisotope/_c/deconvoluter_base.pyx":1114
 *         a_to_a2_ratio = tid.get(0).intensity / tid.get(2).intensity
 * 
 *     monoisotopic_mass = neutral_mass(             # <<<<<<<<<<<<<<
 *         tid.get_monoisotopic_mz(), charge, charge_carrier)
 *     reference_peak = first_peak(eid)
 */
  __pyx_t_13.__pyx_n = 1;
  __pyx_t_13.charge_carrier = __pyx_v_charge_carrier;
  __pyx_t_12 = __pyx_f_12ms_deisotope_2_c_9averagine_neutral_mass(((struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *)__pyx_v_tid->__pyx_vtab)->get_monoisotopic_mz(__pyx_v_tid), __pyx_v_charge, &__pyx_t_13); 
  __pyx_v_monoisotopic_mass = __pyx_t_12;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":1116
 *     monoisotopic_mass = neutral_mass(
 *         tid.get_monoisotopic_mz(), charge, charge_carrier)
 *     reference_peak = first_peak(eid)             # <<<<<<<<<<<<<<
 * 
 *     dpeak = DeconvolutedPeak(
 */
  __pyx_t_10 = ((PyObject *)__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_first_peak(__pyx_v_eid)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_v_reference_peak = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_10);
  __pyx_t_10 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":1119
 * 
 *     dpeak = DeconvolutedPeak(
 *         neutral_mass=monoisotopic_mass, intensity=total_abundance, charge=charge,             # <<<<<<<<<<<<<<
 *         signal_to_noise=signal_to_noise,
 *         index=reference_peak.index if reference_peak is not None else -1,
 */
  __pyx_t_10 = __Pyx_PyDict_NewPresized(14); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_9 = PyFloat_FromDouble(__pyx_v_monoisotopic_mass); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_t_10, __pyx_n_s_neutral_mass, __pyx_t_9) < 0) __PYX_ERR(0, 1119, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = PyFloat_FromDouble(__pyx_v_total_abundance); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_t_10, __pyx_n_s_intensity, __pyx_t_9) < 0) __PYX_ERR(0, 1119, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_PyInt_From_int(__pyx_v_charge); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_t_10, __pyx_n_s_charge, __pyx_t_9) < 0) __PYX_ERR(0, 1119, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":1120
 *     dpeak = DeconvolutedPeak(
 *         neutral_mass=monoisotopic_mass, intensity=total_abundance, charge=charge,
 *         signal_to_noise=signal_to_noise,             # <<<<<<<<<<<<<<
 *         index=reference_peak.index if reference_peak is not None else -1,
 *         full_width_at_half_max=full_width_at_half_max,
 */
  __pyx_t_9 = PyFloat_FromDouble(__pyx_v_signal_to_noise); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_t_10, __pyx_n_s_signal_to_noise, __pyx_t_9) < 0) __PYX_ERR(0, 1119, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":1121
 *         neutral_mass=monoisotopic_mass, intensity=total_abundance, charge=charge,
 *         signal_to_noise=signal_to_noise,
 *         index=reference_peak.index if reference_peak is not None else -1,             # <<<<<<<<<<<<<<
 *         full_width_at_half_max=full_width_at_half_max,
 *         a_to_a2_ratio=a_to_a2_ratio,
 */
  __pyx_t_8 = (((PyObject *)__pyx_v_reference_peak) != Py_None);
  if ((__pyx_t_8 != 0)) {
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v_reference_peak->index); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1121, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_9 = __pyx_t_3;
    __pyx_t_3 = 0;
  } else {
    __Pyx_INCREF(__pyx_int_neg_1);
    __pyx_t_9 = __pyx_int_neg_1;
  }
  if (PyDict_SetItem(__pyx_t_10, __pyx_n_s_index, __pyx_t_9) < 0) __PYX_ERR(0, 1119, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":1122
 *         signal_to_noise=signal_to_noise,
 *         index=reference_peak.index if reference_peak is not None else -1,
 *         full_width_at_half_max=full_width_at_half_max,             # <<<<<<<<<<<<<<
 *         a_to_a2_ratio=a_to_a2_ratio,
 *         most_abundant_mass=neutral_mass(most_abundant_mz, charge),
 */
  __pyx_t_9 = PyFloat_FromDouble(__pyx_v_full_width_at_half_max); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_t_10, __pyx_n_s_full_width_at_half_max, __pyx_t_9) < 0) __PYX_ERR(0, 1119, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":1123
 *         index=reference_peak.index if reference_peak is not None else -1,
 *         full_width_at_half_max=full_width_at_half_max,
 *         a_to_a2_ratio=a_to_a2_ratio,             # <<<<<<<<<<<<<<
 *         most_abundant_mass=neutral_mass(most_abundant_mz, charge),
 *         average_mass=neutral_mass(average_mz, charge),
 */
  __pyx_t_9 = PyFloat_FromDouble(__pyx_v_a_to_a2_ratio); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_t_10, __pyx_n_s_a_to_a2_ratio, __pyx_t_9) < 0) __PYX_ERR(0, 1119, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":1124
 *         full_width_at_half_max=full_width_at_half_max,
 *         a_to_a2_ratio=a_to_a2_ratio,
 *         most_abundant_mass=neutral_mass(most_abundant_mz, charge),             # <<<<<<<<<<<<<<
 *         average_mass=neutral_mass(average_mz, charge),
 *         score=score,
 */
  __pyx_t_9 = PyFloat_FromDouble(__pyx_f_12ms_deisotope_2_c_9averagine_neutral_mass(__pyx_v_most_abundant_mz, __pyx_v_charge, NULL)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_t_10, __pyx_n_s_most_abundant_mass, __pyx_t_9) < 0) __PYX_ERR(0, 1119, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":1125
 *         a_to_a2_ratio=a_to_a2_ratio,
 *         most_abundant_mass=neutral_mass(most_abundant_mz, charge),
 *         average_mass=neutral_mass(average_mz, charge),             # <<<<<<<<<<<<<<
 *         score=score,
 *         envelope=envelope,
 */
  __pyx_t_9 = PyFloat_FromDouble(__pyx_f_12ms_deisotope_2_c_9averagine_neutral_mass(__pyx_v_average_mz, __pyx_v_charge, NULL)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_t_10, __pyx_n_s_average_mass, __pyx_t_9) < 0) __PYX_ERR(0, 1119, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":1126
 *         most_abundant_mass=neutral_mass(most_abundant_mz, charge),
 *         average_mass=neutral_mass(average_mz, charge),
 *         score=score,             # <<<<<<<<<<<<<<
 *         envelope=envelope,
 *         mz=tid.monoisotopic_mz, fit=fit,
 */
  __pyx_t_9 = PyFloat_FromDouble(__pyx_v_score); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_t_10, __pyx_n_s_score, __pyx_t_9) < 0) __PYX_ERR(0, 1119, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":1127
 *         average_mass=neutral_mass(average_mz, charge),
 *         score=score,
 *         envelope=envelope,             # <<<<<<<<<<<<<<
 *         mz=tid.monoisotopic_mz, fit=fit,
 *         area=area)
 */
  if (PyDict_SetItem(__pyx_t_10, __pyx_n_s_envelope, __pyx_v_envelope) < 0) __PYX_ERR(0, 1119, __pyx_L1_error)

  /* "ms_deisotope/_c/deconvoluter_base.pyx":1128
 *         score=score,
 *         envelope=envelope,
 *         mz=tid.monoisotopic_mz, fit=fit,             # <<<<<<<<<<<<<<
 *         area=area)
 *     return dpeak
 */
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_tid), __pyx_n_s_monoisotopic_mz); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_t_10, __pyx_n_s_mz, __pyx_t_9) < 0) __PYX_ERR(0, 1119, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_10, __pyx_n_s_fit, ((PyObject *)__pyx_v_fit)) < 0) __PYX_ERR(0, 1119, __pyx_L1_error)

  /* "ms_deisotope/_c/deconvoluter_base.pyx":1129
 *         envelope=envelope,
 *         mz=tid.monoisotopic_mz, fit=fit,
 *         area=area)             # <<<<<<<<<<<<<<
 *     return dpeak
 */
  __pyx_t_9 = PyFloat_FromDouble(__pyx_v_area); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_t_10, __pyx_n_s_area, __pyx_t_9) < 0) __PYX_ERR(0, 1119, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":1118
 *     reference_peak = first_peak(eid)
 * 
 *     dpeak = DeconvolutedPeak(             # <<<<<<<<<<<<<<
 *         neutral_mass=monoisotopic_mass, intensity=total_abundance, charge=charge,
 *         signal_to_noise=signal_to_noise,
 */
  __pyx_t_9 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak), __pyx_empty_tuple, __pyx_t_10); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_v_dpeak = ((struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *)__pyx_t_9);
  __pyx_t_9 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":1130
 *         mz=tid.monoisotopic_mz, fit=fit,
 *         area=area)
 *     return dpeak             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_dpeak));
  __pyx_r = __pyx_v_dpeak;
  goto __pyx_L0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":1052
 * 
 * @cython.binding(True)
 * cpdef DeconvolutedPeak _make_deconvoluted_peak(self, IsotopicFitRecord fit, double charge_carrier):             # <<<<<<<<<<<<<<
 *     '''Helper method to package a finished :class:`~.IsotopicFitRecord` into a :class:`~.DeconvolutedPeak`
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base._make_deconvoluted_peak", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_eid);
  __Pyx_XDECREF(__pyx_v_rep_eid);
  __Pyx_XDECREF(__pyx_v_envelope);
  __Pyx_XDECREF((PyObject *)__pyx_v_tid);
  __Pyx_XDECREF((PyObject *)__pyx_v_fpeak);
  __Pyx_XDECREF((PyObject *)__pyx_v_reference_peak);
  __Pyx_XDECREF((PyObject *)__pyx_v_dpeak);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_9_make_deconvoluted_peak(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_8_make_deconvoluted_peak[] = "_make_deconvoluted_peak(self, IsotopicFitRecord fit, double charge_carrier) -> DeconvolutedPeak\nHelper method to package a finished :class:`~.IsotopicFitRecord` into a :class:`~.DeconvolutedPeak`\n\n    Arguments\n    ---------\n    fit: :class:`~.IsotopicFitRecord`\n        The fit to package\n    charge_carrier: :class:`float`\n        The mass of the charge carrier\n\n    Returns\n    -------\n    :class:`~.DeconvolutedPeak`\n    ";
static PyMethodDef __pyx_mdef_12ms_deisotope_2_c_17deconvoluter_base_9_make_deconvoluted_peak = {"_make_deconvoluted_peak", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_9_make_deconvoluted_peak, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_8_make_deconvoluted_peak};
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_9_make_deconvoluted_peak(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_fit = 0;
  double __pyx_v_charge_carrier;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_make_deconvoluted_peak (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_fit,&__pyx_n_s_charge_carrier,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fit)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_make_deconvoluted_peak", 1, 3, 3, 1); __PYX_ERR(0, 1052, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_charge_carrier)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_make_deconvoluted_peak", 1, 3, 3, 2); __PYX_ERR(0, 1052, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_make_deconvoluted_peak") < 0)) __PYX_ERR(0, 1052, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_fit = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)values[1]);
    __pyx_v_charge_carrier = __pyx_PyFloat_AsDouble(values[2]); if (unlikely((__pyx_v_charge_carrier == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1052, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_make_deconvoluted_peak", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1052, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base._make_deconvoluted_peak", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_fit), __pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitRecord, 1, "fit", 0))) __PYX_ERR(0, 1052, __pyx_L1_error)
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_8_make_deconvoluted_peak(__pyx_self, __pyx_v_self, __pyx_v_fit, __pyx_v_charge_carrier);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_8_make_deconvoluted_peak(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_fit, double __pyx_v_charge_carrier) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("_make_deconvoluted_peak", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_12ms_deisotope_2_c_17deconvoluter_base__make_deconvoluted_peak(__pyx_v_self, __pyx_v_fit, __pyx_v_charge_carrier, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1052, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base._make_deconvoluted_peak", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_DeconvoluterBase(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_11__pyx_unpickle_DeconvoluterBase(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_10__pyx_unpickle_DeconvoluterBase[] = "__pyx_unpickle_DeconvoluterBase(__pyx_type, long __pyx_checksum, __pyx_state)";
static PyMethodDef __pyx_mdef_12ms_deisotope_2_c_17deconvoluter_base_11__pyx_unpickle_DeconvoluterBase = {"__pyx_unpickle_DeconvoluterBase", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_11__pyx_unpickle_DeconvoluterBase, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_10__pyx_unpickle_DeconvoluterBase};
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_11__pyx_unpickle_DeconvoluterBase(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_DeconvoluterBase (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_DeconvoluterBase", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_DeconvoluterBase", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_DeconvoluterBase") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_DeconvoluterBase", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.__pyx_unpickle_DeconvoluterBase", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_10__pyx_unpickle_DeconvoluterBase(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_10__pyx_unpickle_DeconvoluterBase(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__pyx_unpickle_DeconvoluterBase", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x16fc763:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x16fc763 = (_slice_cache, merge_isobaric_peaks, minimum_intensity, peaklist, scale_method, scorer, use_subtraction, verbose))" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0x16fc763) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x16fc763:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x16fc763 = (_slice_cache, merge_isobaric_peaks, minimum_intensity, peaklist, scale_method, scorer, use_subtraction, verbose))" % __pyx_checksum)
 *     __pyx_result = DeconvoluterBase.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, -1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0x16fc763:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x16fc763 = (_slice_cache, merge_isobaric_peaks, minimum_intensity, peaklist, scale_method, scorer, use_subtraction, verbose))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = DeconvoluterBase.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0x16, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x16fc763:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x16fc763 = (_slice_cache, merge_isobaric_peaks, minimum_intensity, peaklist, scale_method, scorer, use_subtraction, verbose))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x16fc763 = (_slice_cache, merge_isobaric_peaks, minimum_intensity, peaklist, scale_method, scorer, use_subtraction, verbose))" % __pyx_checksum)
 *     __pyx_result = DeconvoluterBase.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_DeconvoluterBase__set_state(<DeconvoluterBase> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x16fc763 = (_slice_cache, merge_isobaric_peaks, minimum_intensity, peaklist, scale_method, scorer, use_subtraction, verbose))" % __pyx_checksum)
 *     __pyx_result = DeconvoluterBase.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_DeconvoluterBase__set_state(<DeconvoluterBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = DeconvoluterBase.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_DeconvoluterBase__set_state(<DeconvoluterBase> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_DeconvoluterBase__set_state(DeconvoluterBase __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_12ms_deisotope_2_c_17deconvoluter_base___pyx_unpickle_DeconvoluterBase__set_state(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x16fc763 = (_slice_cache, merge_isobaric_peaks, minimum_intensity, peaklist, scale_method, scorer, use_subtraction, verbose))" % __pyx_checksum)
 *     __pyx_result = DeconvoluterBase.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_DeconvoluterBase__set_state(<DeconvoluterBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_DeconvoluterBase__set_state(<DeconvoluterBase> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_DeconvoluterBase__set_state(DeconvoluterBase __pyx_result, tuple __pyx_state):
 *     __pyx_result._slice_cache = __pyx_state[0]; __pyx_result.merge_isobaric_peaks = __pyx_state[1]; __pyx_result.minimum_intensity = __pyx_state[2]; __pyx_result.peaklist = __pyx_state[3]; __pyx_result.scale_method = __pyx_state[4]; __pyx_result.scorer = __pyx_state[5]; __pyx_result.use_subtraction = __pyx_state[6]; __pyx_result.verbose = __pyx_state[7]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_DeconvoluterBase(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.__pyx_unpickle_DeconvoluterBase", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_DeconvoluterBase__set_state(<DeconvoluterBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_DeconvoluterBase__set_state(DeconvoluterBase __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._slice_cache = __pyx_state[0]; __pyx_result.merge_isobaric_peaks = __pyx_state[1]; __pyx_result.minimum_intensity = __pyx_state[2]; __pyx_result.peaklist = __pyx_state[3]; __pyx_result.scale_method = __pyx_state[4]; __pyx_result.scorer = __pyx_state[5]; __pyx_result.use_subtraction = __pyx_state[6]; __pyx_result.verbose = __pyx_state[7]
 *     if len(__pyx_state) > 8 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base___pyx_unpickle_DeconvoluterBase__set_state(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  double __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  __Pyx_RefNannySetupContext("__pyx_unpickle_DeconvoluterBase__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_DeconvoluterBase__set_state(DeconvoluterBase __pyx_result, tuple __pyx_state):
 *     __pyx_result._slice_cache = __pyx_state[0]; __pyx_result.merge_isobaric_peaks = __pyx_state[1]; __pyx_result.minimum_intensity = __pyx_state[2]; __pyx_result.peaklist = __pyx_state[3]; __pyx_result.scale_method = __pyx_state[4]; __pyx_result.scorer = __pyx_state[5]; __pyx_result.use_subtraction = __pyx_state[6]; __pyx_result.verbose = __pyx_state[7]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 8 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[8])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyDict_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->_slice_cache);
  __Pyx_DECREF(__pyx_v___pyx_result->_slice_cache);
  __pyx_v___pyx_result->_slice_cache = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->merge_isobaric_peaks = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_3 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->minimum_intensity = __pyx_t_3;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->peaklist);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->peaklist));
  __pyx_v___pyx_result->peaklist = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyString_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->scale_method);
  __Pyx_DECREF(__pyx_v___pyx_result->scale_method);
  __pyx_v___pyx_result->scale_method = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitterBase))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->scorer);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->scorer));
  __pyx_v___pyx_result->scorer = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->use_subtraction = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->verbose = __pyx_t_2;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_DeconvoluterBase__set_state(DeconvoluterBase __pyx_result, tuple __pyx_state):
 *     __pyx_result._slice_cache = __pyx_state[0]; __pyx_result.merge_isobaric_peaks = __pyx_state[1]; __pyx_result.minimum_intensity = __pyx_state[2]; __pyx_result.peaklist = __pyx_state[3]; __pyx_result.scale_method = __pyx_state[4]; __pyx_result.scorer = __pyx_state[5]; __pyx_result.use_subtraction = __pyx_state[6]; __pyx_result.verbose = __pyx_state[7]
 *     if len(__pyx_state) > 8 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[8])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_4 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_5 = ((__pyx_t_4 > 8) != 0);
  if (__pyx_t_5) {
  } else {
    __pyx_t_2 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_5 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_6 = (__pyx_t_5 != 0);
  __pyx_t_2 = __pyx_t_6;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result._slice_cache = __pyx_state[0]; __pyx_result.merge_isobaric_peaks = __pyx_state[1]; __pyx_result.minimum_intensity = __pyx_state[2]; __pyx_result.peaklist = __pyx_state[3]; __pyx_result.scale_method = __pyx_state[4]; __pyx_result.scorer = __pyx_state[5]; __pyx_result.use_subtraction = __pyx_state[6]; __pyx_result.verbose = __pyx_state[7]
 *     if len(__pyx_state) > 8 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[8])             # <<<<<<<<<<<<<<
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_update); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_7 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
      }
    }
    __pyx_t_1 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_9, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_7);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_DeconvoluterBase__set_state(DeconvoluterBase __pyx_result, tuple __pyx_state):
 *     __pyx_result._slice_cache = __pyx_state[0]; __pyx_result.merge_isobaric_peaks = __pyx_state[1]; __pyx_result.minimum_intensity = __pyx_state[2]; __pyx_result.peaklist = __pyx_state[3]; __pyx_result.scale_method = __pyx_state[4]; __pyx_result.scorer = __pyx_state[5]; __pyx_result.use_subtraction = __pyx_state[6]; __pyx_result.verbose = __pyx_state[7]
 *     if len(__pyx_state) > 8 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[8])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_DeconvoluterBase__set_state(<DeconvoluterBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_DeconvoluterBase__set_state(DeconvoluterBase __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._slice_cache = __pyx_state[0]; __pyx_result.merge_isobaric_peaks = __pyx_state[1]; __pyx_result.minimum_intensity = __pyx_state[2]; __pyx_result.peaklist = __pyx_state[3]; __pyx_result.scale_method = __pyx_state[4]; __pyx_result.scorer = __pyx_state[5]; __pyx_result.use_subtraction = __pyx_state[6]; __pyx_result.verbose = __pyx_state[7]
 *     if len(__pyx_state) > 8 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.__pyx_unpickle_DeconvoluterBase__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_AveragineDeconvoluterBase(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_13__pyx_unpickle_AveragineDeconvoluterBase(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_12__pyx_unpickle_AveragineDeconvoluterBase[] = "__pyx_unpickle_AveragineDeconvoluterBase(__pyx_type, long __pyx_checksum, __pyx_state)";
static PyMethodDef __pyx_mdef_12ms_deisotope_2_c_17deconvoluter_base_13__pyx_unpickle_AveragineDeconvoluterBase = {"__pyx_unpickle_AveragineDeconvoluterBase", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_13__pyx_unpickle_AveragineDeconvoluterBase, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_12__pyx_unpickle_AveragineDeconvoluterBase};
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_13__pyx_unpickle_AveragineDeconvoluterBase(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_AveragineDeconvoluterBase (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_AveragineDeconvoluterBase", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_AveragineDeconvoluterBase", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_AveragineDeconvoluterBase") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_AveragineDeconvoluterBase", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.__pyx_unpickle_AveragineDeconvoluterBase", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_12__pyx_unpickle_AveragineDeconvoluterBase(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_12__pyx_unpickle_AveragineDeconvoluterBase(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__pyx_unpickle_AveragineDeconvoluterBase", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x9da1e44:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x9da1e44 = (_slice_cache, averagine, merge_isobaric_peaks, minimum_intensity, peaklist, scale_method, scorer, use_subtraction, verbose))" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0x9da1e44) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x9da1e44:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x9da1e44 = (_slice_cache, averagine, merge_isobaric_peaks, minimum_intensity, peaklist, scale_method, scorer, use_subtraction, verbose))" % __pyx_checksum)
 *     __pyx_result = AveragineDeconvoluterBase.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, -1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0x9da1e44:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x9da1e44 = (_slice_cache, averagine, merge_isobaric_peaks, minimum_intensity, peaklist, scale_method, scorer, use_subtraction, verbose))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = AveragineDeconvoluterBase.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0x9d, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x9da1e44:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x9da1e44 = (_slice_cache, averagine, merge_isobaric_peaks, minimum_intensity, peaklist, scale_method, scorer, use_subtraction, verbose))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x9da1e44 = (_slice_cache, averagine, merge_isobaric_peaks, minimum_intensity, peaklist, scale_method, scorer, use_subtraction, verbose))" % __pyx_checksum)
 *     __pyx_result = AveragineDeconvoluterBase.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_AveragineDeconvoluterBase__set_state(<AveragineDeconvoluterBase> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x9da1e44 = (_slice_cache, averagine, merge_isobaric_peaks, minimum_intensity, peaklist, scale_method, scorer, use_subtraction, verbose))" % __pyx_checksum)
 *     __pyx_result = AveragineDeconvoluterBase.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_AveragineDeconvoluterBase__set_state(<AveragineDeconvoluterBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = AveragineDeconvoluterBase.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_AveragineDeconvoluterBase__set_state(<AveragineDeconvoluterBase> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_AveragineDeconvoluterBase__set_state(AveragineDeconvoluterBase __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_12ms_deisotope_2_c_17deconvoluter_base___pyx_unpickle_AveragineDeconvoluterBase__set_state(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x9da1e44 = (_slice_cache, averagine, merge_isobaric_peaks, minimum_intensity, peaklist, scale_method, scorer, use_subtraction, verbose))" % __pyx_checksum)
 *     __pyx_result = AveragineDeconvoluterBase.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_AveragineDeconvoluterBase__set_state(<AveragineDeconvoluterBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_AveragineDeconvoluterBase__set_state(<AveragineDeconvoluterBase> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_AveragineDeconvoluterBase__set_state(AveragineDeconvoluterBase __pyx_result, tuple __pyx_state):
 *     __pyx_result._slice_cache = __pyx_state[0]; __pyx_result.averagine = __pyx_state[1]; __pyx_result.merge_isobaric_peaks = __pyx_state[2]; __pyx_result.minimum_intensity = __pyx_state[3]; __pyx_result.peaklist = __pyx_state[4]; __pyx_result.scale_method = __pyx_state[5]; __pyx_result.scorer = __pyx_state[6]; __pyx_result.use_subtraction = __pyx_state[7]; __pyx_result.verbose = __pyx_state[8]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_AveragineDeconvoluterBase(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.__pyx_unpickle_AveragineDeconvoluterBase", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_AveragineDeconvoluterBase__set_state(<AveragineDeconvoluterBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_AveragineDeconvoluterBase__set_state(AveragineDeconvoluterBase __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._slice_cache = __pyx_state[0]; __pyx_result.averagine = __pyx_state[1]; __pyx_result.merge_isobaric_peaks = __pyx_state[2]; __pyx_result.minimum_intensity = __pyx_state[3]; __pyx_result.peaklist = __pyx_state[4]; __pyx_result.scale_method = __pyx_state[5]; __pyx_result.scorer = __pyx_state[6]; __pyx_result.use_subtraction = __pyx_state[7]; __pyx_result.verbose = __pyx_state[8]
 *     if len(__pyx_state) > 9 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base___pyx_unpickle_AveragineDeconvoluterBase__set_state(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  double __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  __Pyx_RefNannySetupContext("__pyx_unpickle_AveragineDeconvoluterBase__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_AveragineDeconvoluterBase__set_state(AveragineDeconvoluterBase __pyx_result, tuple __pyx_state):
 *     __pyx_result._slice_cache = __pyx_state[0]; __pyx_result.averagine = __pyx_state[1]; __pyx_result.merge_isobaric_peaks = __pyx_state[2]; __pyx_result.minimum_intensity = __pyx_state[3]; __pyx_result.peaklist = __pyx_state[4]; __pyx_result.scale_method = __pyx_state[5]; __pyx_result.scorer = __pyx_state[6]; __pyx_result.use_subtraction = __pyx_state[7]; __pyx_result.verbose = __pyx_state[8]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 9 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[9])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyDict_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base._slice_cache);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base._slice_cache);
  __pyx_v___pyx_result->__pyx_base._slice_cache = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_12ms_deisotope_2_c_9averagine_AveragineCache))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->averagine);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->averagine));
  __pyx_v___pyx_result->averagine = ((struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.merge_isobaric_peaks = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_3 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.minimum_intensity = __pyx_t_3;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.peaklist);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->__pyx_base.peaklist));
  __pyx_v___pyx_result->__pyx_base.peaklist = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyString_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.scale_method);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.scale_method);
  __pyx_v___pyx_result->__pyx_base.scale_method = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitterBase))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.scorer);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->__pyx_base.scorer));
  __pyx_v___pyx_result->__pyx_base.scorer = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.use_subtraction = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.verbose = __pyx_t_2;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_AveragineDeconvoluterBase__set_state(AveragineDeconvoluterBase __pyx_result, tuple __pyx_state):
 *     __pyx_result._slice_cache = __pyx_state[0]; __pyx_result.averagine = __pyx_state[1]; __pyx_result.merge_isobaric_peaks = __pyx_state[2]; __pyx_result.minimum_intensity = __pyx_state[3]; __pyx_result.peaklist = __pyx_state[4]; __pyx_result.scale_method = __pyx_state[5]; __pyx_result.scorer = __pyx_state[6]; __pyx_result.use_subtraction = __pyx_state[7]; __pyx_result.verbose = __pyx_state[8]
 *     if len(__pyx_state) > 9 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[9])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_4 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_5 = ((__pyx_t_4 > 9) != 0);
  if (__pyx_t_5) {
  } else {
    __pyx_t_2 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_5 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_6 = (__pyx_t_5 != 0);
  __pyx_t_2 = __pyx_t_6;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result._slice_cache = __pyx_state[0]; __pyx_result.averagine = __pyx_state[1]; __pyx_result.merge_isobaric_peaks = __pyx_state[2]; __pyx_result.minimum_intensity = __pyx_state[3]; __pyx_result.peaklist = __pyx_state[4]; __pyx_result.scale_method = __pyx_state[5]; __pyx_result.scorer = __pyx_state[6]; __pyx_result.use_subtraction = __pyx_state[7]; __pyx_result.verbose = __pyx_state[8]
 *     if len(__pyx_state) > 9 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[9])             # <<<<<<<<<<<<<<
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_update); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_7 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 9, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
      }
    }
    __pyx_t_1 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_9, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_7);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_AveragineDeconvoluterBase__set_state(AveragineDeconvoluterBase __pyx_result, tuple __pyx_state):
 *     __pyx_result._slice_cache = __pyx_state[0]; __pyx_result.averagine = __pyx_state[1]; __pyx_result.merge_isobaric_peaks = __pyx_state[2]; __pyx_result.minimum_intensity = __pyx_state[3]; __pyx_result.peaklist = __pyx_state[4]; __pyx_result.scale_method = __pyx_state[5]; __pyx_result.scorer = __pyx_state[6]; __pyx_result.use_subtraction = __pyx_state[7]; __pyx_result.verbose = __pyx_state[8]
 *     if len(__pyx_state) > 9 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[9])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_AveragineDeconvoluterBase__set_state(<AveragineDeconvoluterBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_AveragineDeconvoluterBase__set_state(AveragineDeconvoluterBase __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._slice_cache = __pyx_state[0]; __pyx_result.averagine = __pyx_state[1]; __pyx_result.merge_isobaric_peaks = __pyx_state[2]; __pyx_result.minimum_intensity = __pyx_state[3]; __pyx_result.peaklist = __pyx_state[4]; __pyx_result.scale_method = __pyx_state[5]; __pyx_result.scorer = __pyx_state[6]; __pyx_result.use_subtraction = __pyx_state[7]; __pyx_result.verbose = __pyx_state[8]
 *     if len(__pyx_state) > 9 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.__pyx_unpickle_AveragineDeconvoluterBase__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_MultiAveragineDeconvoluterBase(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_15__pyx_unpickle_MultiAveragineDeconvoluterBase(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_14__pyx_unpickle_MultiAveragineDeconvoluterBase[] = "__pyx_unpickle_MultiAveragineDeconvoluterBase(__pyx_type, long __pyx_checksum, __pyx_state)";
static PyMethodDef __pyx_mdef_12ms_deisotope_2_c_17deconvoluter_base_15__pyx_unpickle_MultiAveragineDeconvoluterBase = {"__pyx_unpickle_MultiAveragineDeconvoluterBase", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_15__pyx_unpickle_MultiAveragineDeconvoluterBase, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_14__pyx_unpickle_MultiAveragineDeconvoluterBase};
static PyObject *__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_15__pyx_unpickle_MultiAveragineDeconvoluterBase(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_MultiAveragineDeconvoluterBase (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_MultiAveragineDeconvoluterBase", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_MultiAveragineDeconvoluterBase", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_MultiAveragineDeconvoluterBase") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_MultiAveragineDeconvoluterBase", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.__pyx_unpickle_MultiAveragineDeconvoluterBase", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14__pyx_unpickle_MultiAveragineDeconvoluterBase(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12ms_deisotope_2_c_17deconvoluter_base_14__pyx_unpickle_MultiAveragineDeconvoluterBase(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__pyx_unpickle_MultiAveragineDeconvoluterBase", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xfdf46c0:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xfdf46c0 = (_slice_cache, averagines, merge_isobaric_peaks, minimum_intensity, peaklist, scale_method, scorer, use_subtraction, verbose))" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0xfdf46c0) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xfdf46c0:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xfdf46c0 = (_slice_cache, averagines, merge_isobaric_peaks, minimum_intensity, peaklist, scale_method, scorer, use_subtraction, verbose))" % __pyx_checksum)
 *     __pyx_result = MultiAveragineDeconvoluterBase.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, -1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0xfdf46c0:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xfdf46c0 = (_slice_cache, averagines, merge_isobaric_peaks, minimum_intensity, peaklist, scale_method, scorer, use_subtraction, verbose))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = MultiAveragineDeconvoluterBase.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0xfd, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xfdf46c0:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xfdf46c0 = (_slice_cache, averagines, merge_isobaric_peaks, minimum_intensity, peaklist, scale_method, scorer, use_subtraction, verbose))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xfdf46c0 = (_slice_cache, averagines, merge_isobaric_peaks, minimum_intensity, peaklist, scale_method, scorer, use_subtraction, verbose))" % __pyx_checksum)
 *     __pyx_result = MultiAveragineDeconvoluterBase.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_MultiAveragineDeconvoluterBase__set_state(<MultiAveragineDeconvoluterBase> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xfdf46c0 = (_slice_cache, averagines, merge_isobaric_peaks, minimum_intensity, peaklist, scale_method, scorer, use_subtraction, verbose))" % __pyx_checksum)
 *     __pyx_result = MultiAveragineDeconvoluterBase.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_MultiAveragineDeconvoluterBase__set_state(<MultiAveragineDeconvoluterBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = MultiAveragineDeconvoluterBase.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_MultiAveragineDeconvoluterBase__set_state(<MultiAveragineDeconvoluterBase> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_MultiAveragineDeconvoluterBase__set_state(MultiAveragineDeconvoluterBase __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_12ms_deisotope_2_c_17deconvoluter_base___pyx_unpickle_MultiAveragineDeconvoluterBase__set_state(((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xfdf46c0 = (_slice_cache, averagines, merge_isobaric_peaks, minimum_intensity, peaklist, scale_method, scorer, use_subtraction, verbose))" % __pyx_checksum)
 *     __pyx_result = MultiAveragineDeconvoluterBase.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_MultiAveragineDeconvoluterBase__set_state(<MultiAveragineDeconvoluterBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_MultiAveragineDeconvoluterBase__set_state(<MultiAveragineDeconvoluterBase> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_MultiAveragineDeconvoluterBase__set_state(MultiAveragineDeconvoluterBase __pyx_result, tuple __pyx_state):
 *     __pyx_result._slice_cache = __pyx_state[0]; __pyx_result.averagines = __pyx_state[1]; __pyx_result.merge_isobaric_peaks = __pyx_state[2]; __pyx_result.minimum_intensity = __pyx_state[3]; __pyx_result.peaklist = __pyx_state[4]; __pyx_result.scale_method = __pyx_state[5]; __pyx_result.scorer = __pyx_state[6]; __pyx_result.use_subtraction = __pyx_state[7]; __pyx_result.verbose = __pyx_state[8]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_MultiAveragineDeconvoluterBase(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.__pyx_unpickle_MultiAveragineDeconvoluterBase", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_MultiAveragineDeconvoluterBase__set_state(<MultiAveragineDeconvoluterBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_MultiAveragineDeconvoluterBase__set_state(MultiAveragineDeconvoluterBase __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._slice_cache = __pyx_state[0]; __pyx_result.averagines = __pyx_state[1]; __pyx_result.merge_isobaric_peaks = __pyx_state[2]; __pyx_result.minimum_intensity = __pyx_state[3]; __pyx_result.peaklist = __pyx_state[4]; __pyx_result.scale_method = __pyx_state[5]; __pyx_result.scorer = __pyx_state[6]; __pyx_result.use_subtraction = __pyx_state[7]; __pyx_result.verbose = __pyx_state[8]
 *     if len(__pyx_state) > 9 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_12ms_deisotope_2_c_17deconvoluter_base___pyx_unpickle_MultiAveragineDeconvoluterBase__set_state(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  double __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  __Pyx_RefNannySetupContext("__pyx_unpickle_MultiAveragineDeconvoluterBase__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_MultiAveragineDeconvoluterBase__set_state(MultiAveragineDeconvoluterBase __pyx_result, tuple __pyx_state):
 *     __pyx_result._slice_cache = __pyx_state[0]; __pyx_result.averagines = __pyx_state[1]; __pyx_result.merge_isobaric_peaks = __pyx_state[2]; __pyx_result.minimum_intensity = __pyx_state[3]; __pyx_result.peaklist = __pyx_state[4]; __pyx_result.scale_method = __pyx_state[5]; __pyx_result.scorer = __pyx_state[6]; __pyx_result.use_subtraction = __pyx_state[7]; __pyx_result.verbose = __pyx_state[8]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 9 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[9])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyDict_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base._slice_cache);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base._slice_cache);
  __pyx_v___pyx_result->__pyx_base._slice_cache = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyList_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->averagines);
  __Pyx_DECREF(__pyx_v___pyx_result->averagines);
  __pyx_v___pyx_result->averagines = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.merge_isobaric_peaks = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_3 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.minimum_intensity = __pyx_t_3;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.peaklist);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->__pyx_base.peaklist));
  __pyx_v___pyx_result->__pyx_base.peaklist = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyString_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.scale_method);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.scale_method);
  __pyx_v___pyx_result->__pyx_base.scale_method = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitterBase))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.scorer);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->__pyx_base.scorer));
  __pyx_v___pyx_result->__pyx_base.scorer = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.use_subtraction = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.verbose = __pyx_t_2;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_MultiAveragineDeconvoluterBase__set_state(MultiAveragineDeconvoluterBase __pyx_result, tuple __pyx_state):
 *     __pyx_result._slice_cache = __pyx_state[0]; __pyx_result.averagines = __pyx_state[1]; __pyx_result.merge_isobaric_peaks = __pyx_state[2]; __pyx_result.minimum_intensity = __pyx_state[3]; __pyx_result.peaklist = __pyx_state[4]; __pyx_result.scale_method = __pyx_state[5]; __pyx_result.scorer = __pyx_state[6]; __pyx_result.use_subtraction = __pyx_state[7]; __pyx_result.verbose = __pyx_state[8]
 *     if len(__pyx_state) > 9 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[9])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_4 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_5 = ((__pyx_t_4 > 9) != 0);
  if (__pyx_t_5) {
  } else {
    __pyx_t_2 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_5 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_6 = (__pyx_t_5 != 0);
  __pyx_t_2 = __pyx_t_6;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result._slice_cache = __pyx_state[0]; __pyx_result.averagines = __pyx_state[1]; __pyx_result.merge_isobaric_peaks = __pyx_state[2]; __pyx_result.minimum_intensity = __pyx_state[3]; __pyx_result.peaklist = __pyx_state[4]; __pyx_result.scale_method = __pyx_state[5]; __pyx_result.scorer = __pyx_state[6]; __pyx_result.use_subtraction = __pyx_state[7]; __pyx_result.verbose = __pyx_state[8]
 *     if len(__pyx_state) > 9 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[9])             # <<<<<<<<<<<<<<
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_update); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_7 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 9, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
      }
    }
    __pyx_t_1 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_9, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_7);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_MultiAveragineDeconvoluterBase__set_state(MultiAveragineDeconvoluterBase __pyx_result, tuple __pyx_state):
 *     __pyx_result._slice_cache = __pyx_state[0]; __pyx_result.averagines = __pyx_state[1]; __pyx_result.merge_isobaric_peaks = __pyx_state[2]; __pyx_result.minimum_intensity = __pyx_state[3]; __pyx_result.peaklist = __pyx_state[4]; __pyx_result.scale_method = __pyx_state[5]; __pyx_result.scorer = __pyx_state[6]; __pyx_result.use_subtraction = __pyx_state[7]; __pyx_result.verbose = __pyx_state[8]
 *     if len(__pyx_state) > 9 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[9])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_MultiAveragineDeconvoluterBase__set_state(<MultiAveragineDeconvoluterBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_MultiAveragineDeconvoluterBase__set_state(MultiAveragineDeconvoluterBase __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._slice_cache = __pyx_state[0]; __pyx_result.averagines = __pyx_state[1]; __pyx_result.merge_isobaric_peaks = __pyx_state[2]; __pyx_result.minimum_intensity = __pyx_state[3]; __pyx_result.peaklist = __pyx_state[4]; __pyx_result.scale_method = __pyx_state[5]; __pyx_result.scorer = __pyx_state[6]; __pyx_result.use_subtraction = __pyx_state[7]; __pyx_result.verbose = __pyx_state[8]
 *     if len(__pyx_state) > 9 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("ms_deisotope._c.deconvoluter_base.__pyx_unpickle_MultiAveragineDeconvoluterBase__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":258
 *         # experimental exception made for __getbuffer__ and __releasebuffer__
 *         # -- the details of this may change.
 *         def __getbuffer__(ndarray self, Py_buffer* info, int flags):             # <<<<<<<<<<<<<<
 *             # This implementation of getbuffer is geared towards Cython
 *             # requirements, and does not yet fulfill the PEP.
 */

/* Python wrapper */
static CYTHON_UNUSED int __pyx_pw_5numpy_7ndarray_1__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
static CYTHON_UNUSED int __pyx_pw_5numpy_7ndarray_1__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5numpy_7ndarray___getbuffer__(((PyArrayObject *)__pyx_v_self), ((Py_buffer *)__pyx_v_info), ((int)__pyx_v_flags));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5numpy_7ndarray___getbuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_v_i;
  int __pyx_v_ndim;
  int __pyx_v_endian_detector;
  int __pyx_v_little_endian;
  int __pyx_v_t;
  char *__pyx_v_f;
  PyArray_Descr *__pyx_v_descr = 0;
  int __pyx_v_offset;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyArray_Descr *__pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  char *__pyx_t_9;
  if (__pyx_v_info == NULL) {
    PyErr_SetString(PyExc_BufferError, "PyObject_GetBuffer: view==NULL argument is obsolete");
    return -1;
  }
  __Pyx_RefNannySetupContext("__getbuffer__", 0);
  __pyx_v_info->obj = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(__pyx_v_info->obj);

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":265
 * 
 *             cdef int i, ndim
 *             cdef int endian_detector = 1             # <<<<<<<<<<<<<<
 *             cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)
 * 
 */
  __pyx_v_endian_detector = 1;

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":266
 *             cdef int i, ndim
 *             cdef int endian_detector = 1
 *             cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)             # <<<<<<<<<<<<<<
 * 
 *             ndim = PyArray_NDIM(self)
 */
  __pyx_v_little_endian = ((((char *)(&__pyx_v_endian_detector))[0]) != 0);

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":268
 *             cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)
 * 
 *             ndim = PyArray_NDIM(self)             # <<<<<<<<<<<<<<
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 */
  __pyx_v_ndim = PyArray_NDIM(__pyx_v_self);

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":270
 *             ndim = PyArray_NDIM(self)
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")
 */
  __pyx_t_2 = (((__pyx_v_flags & PyBUF_C_CONTIGUOUS) == PyBUF_C_CONTIGUOUS) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":271
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):             # <<<<<<<<<<<<<<
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 */
  __pyx_t_2 = ((!(PyArray_CHKFLAGS(__pyx_v_self, NPY_ARRAY_C_CONTIGUOUS) != 0)) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":270
 *             ndim = PyArray_NDIM(self)
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")
 */
  if (unlikely(__pyx_t_1)) {

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":272
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")             # <<<<<<<<<<<<<<
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__25, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 272, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(3, 272, __pyx_L1_error)

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":270
 *             ndim = PyArray_NDIM(self)
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")
 */
  }

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":274
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 */
  __pyx_t_2 = (((__pyx_v_flags & PyBUF_F_CONTIGUOUS) == PyBUF_F_CONTIGUOUS) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L7_bool_binop_done;
  }

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":275
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):             # <<<<<<<<<<<<<<
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 * 
 */
  __pyx_t_2 = ((!(PyArray_CHKFLAGS(__pyx_v_self, NPY_ARRAY_F_CONTIGUOUS) != 0)) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L7_bool_binop_done:;

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":274
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 */
  if (unlikely(__pyx_t_1)) {

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":276
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")             # <<<<<<<<<<<<<<
 * 
 *             info.buf = PyArray_DATA(self)
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__26, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 276, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(3, 276, __pyx_L1_error)

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":274
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 */
  }

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":278
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 * 
 *             info.buf = PyArray_DATA(self)             # <<<<<<<<<<<<<<
 *             info.ndim = ndim
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 */
  __pyx_v_info->buf = PyArray_DATA(__pyx_v_self);

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":279
 * 
 *             info.buf = PyArray_DATA(self)
 *             info.ndim = ndim             # <<<<<<<<<<<<<<
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 *                 # Allocate new buffer for strides and shape info.
 */
  __pyx_v_info->ndim = __pyx_v_ndim;

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":280
 *             info.buf = PyArray_DATA(self)
 *             info.ndim = ndim
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 # Allocate new buffer for strides and shape info.
 *                 # This is allocated as one block, strides first.
 */
  __pyx_t_1 = (((sizeof(npy_intp)) != (sizeof(Py_ssize_t))) != 0);
  if (__pyx_t_1) {

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":283
 *                 # Allocate new buffer for strides and shape info.
 *                 # This is allocated as one block, strides first.
 *                 info.strides = <Py_ssize_t*>PyObject_Malloc(sizeof(Py_ssize_t) * 2 * <size_t>ndim)             # <<<<<<<<<<<<<<
 *                 info.shape = info.strides + ndim
 *                 for i in range(ndim):
 */
    __pyx_v_info->strides = ((Py_ssize_t *)PyObject_Malloc((((sizeof(Py_ssize_t)) * 2) * ((size_t)__pyx_v_ndim))));

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":284
 *                 # This is allocated as one block, strides first.
 *                 info.strides = <Py_ssize_t*>PyObject_Malloc(sizeof(Py_ssize_t) * 2 * <size_t>ndim)
 *                 info.shape = info.strides + ndim             # <<<<<<<<<<<<<<
 *                 for i in range(ndim):
 *                     info.strides[i] = PyArray_STRIDES(self)[i]
 */
    __pyx_v_info->shape = (__pyx_v_info->strides + __pyx_v_ndim);

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":285
 *                 info.strides = <Py_ssize_t*>PyObject_Malloc(sizeof(Py_ssize_t) * 2 * <size_t>ndim)
 *                 info.shape = info.strides + ndim
 *                 for i in range(ndim):             # <<<<<<<<<<<<<<
 *                     info.strides[i] = PyArray_STRIDES(self)[i]
 *                     info.shape[i] = PyArray_DIMS(self)[i]
 */
    __pyx_t_4 = __pyx_v_ndim;
    __pyx_t_5 = __pyx_t_4;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
      __pyx_v_i = __pyx_t_6;

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":286
 *                 info.shape = info.strides + ndim
 *                 for i in range(ndim):
 *                     info.strides[i] = PyArray_STRIDES(self)[i]             # <<<<<<<<<<<<<<
 *                     info.shape[i] = PyArray_DIMS(self)[i]
 *             else:
 */
      (__pyx_v_info->strides[__pyx_v_i]) = (PyArray_STRIDES(__pyx_v_self)[__pyx_v_i]);

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":287
 *                 for i in range(ndim):
 *                     info.strides[i] = PyArray_STRIDES(self)[i]
 *                     info.shape[i] = PyArray_DIMS(self)[i]             # <<<<<<<<<<<<<<
 *             else:
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)
 */
      (__pyx_v_info->shape[__pyx_v_i]) = (PyArray_DIMS(__pyx_v_self)[__pyx_v_i]);
    }

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":280
 *             info.buf = PyArray_DATA(self)
 *             info.ndim = ndim
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 # Allocate new buffer for strides and shape info.
 *                 # This is allocated as one block, strides first.
 */
    goto __pyx_L9;
  }

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":289
 *                     info.shape[i] = PyArray_DIMS(self)[i]
 *             else:
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)             # <<<<<<<<<<<<<<
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)
 *             info.suboffsets = NULL
 */
  /*else*/ {
    __pyx_v_info->strides = ((Py_ssize_t *)PyArray_STRIDES(__pyx_v_self));

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":290
 *             else:
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)             # <<<<<<<<<<<<<<
 *             info.suboffsets = NULL
 *             info.itemsize = PyArray_ITEMSIZE(self)
 */
    __pyx_v_info->shape = ((Py_ssize_t *)PyArray_DIMS(__pyx_v_self));
  }
  __pyx_L9:;

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":291
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)
 *             info.suboffsets = NULL             # <<<<<<<<<<<<<<
 *             info.itemsize = PyArray_ITEMSIZE(self)
 *             info.readonly = not PyArray_ISWRITEABLE(self)
 */
  __pyx_v_info->suboffsets = NULL;

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":292
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)
 *             info.suboffsets = NULL
 *             info.itemsize = PyArray_ITEMSIZE(self)             # <<<<<<<<<<<<<<
 *             info.readonly = not PyArray_ISWRITEABLE(self)
 * 
 */
  __pyx_v_info->itemsize = PyArray_ITEMSIZE(__pyx_v_self);

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":293
 *             info.suboffsets = NULL
 *             info.itemsize = PyArray_ITEMSIZE(self)
 *             info.readonly = not PyArray_ISWRITEABLE(self)             # <<<<<<<<<<<<<<
 * 
 *             cdef int t
 */
  __pyx_v_info->readonly = (!(PyArray_ISWRITEABLE(__pyx_v_self) != 0));

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":296
 * 
 *             cdef int t
 *             cdef char* f = NULL             # <<<<<<<<<<<<<<
 *             cdef dtype descr = <dtype>PyArray_DESCR(self)
 *             cdef int offset
 */
  __pyx_v_f = NULL;

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":297
 *             cdef int t
 *             cdef char* f = NULL
 *             cdef dtype descr = <dtype>PyArray_DESCR(self)             # <<<<<<<<<<<<<<
 *             cdef int offset
 * 
 */
  __pyx_t_7 = PyArray_DESCR(__pyx_v_self);
  __pyx_t_3 = ((PyObject *)__pyx_t_7);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_descr = ((PyArray_Descr *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":300
 *             cdef int offset
 * 
 *             info.obj = self             # <<<<<<<<<<<<<<
 * 
 *             if not PyDataType_HASFIELDS(descr):
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  __Pyx_GOTREF(__pyx_v_info->obj);
  __Pyx_DECREF(__pyx_v_info->obj);
  __pyx_v_info->obj = ((PyObject *)__pyx_v_self);

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":302
 *             info.obj = self
 * 
 *             if not PyDataType_HASFIELDS(descr):             # <<<<<<<<<<<<<<
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or
 */
  __pyx_t_1 = ((!(PyDataType_HASFIELDS(__pyx_v_descr) != 0)) != 0);
  if (__pyx_t_1) {

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":303
 * 
 *             if not PyDataType_HASFIELDS(descr):
 *                 t = descr.type_num             # <<<<<<<<<<<<<<
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):
 */
    __pyx_t_4 = __pyx_v_descr->type_num;
    __pyx_v_t = __pyx_t_4;

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":304
 *             if not PyDataType_HASFIELDS(descr):
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 */
    __pyx_t_2 = ((__pyx_v_descr->byteorder == '>') != 0);
    if (!__pyx_t_2) {
      goto __pyx_L15_next_or;
    } else {
    }
    __pyx_t_2 = (__pyx_v_little_endian != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L14_bool_binop_done;
    }
    __pyx_L15_next_or:;

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":305
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):             # <<<<<<<<<<<<<<
 *                     raise ValueError(u"Non-native byte order not supported")
 *                 if   t == NPY_BYTE:        f = "b"
 */
    __pyx_t_2 = ((__pyx_v_descr->byteorder == '<') != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L14_bool_binop_done;
    }
    __pyx_t_2 = ((!(__pyx_v_little_endian != 0)) != 0);
    __pyx_t_1 = __pyx_t_2;
    __pyx_L14_bool_binop_done:;

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":304
 *             if not PyDataType_HASFIELDS(descr):
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 */
    if (unlikely(__pyx_t_1)) {

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":306
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__27, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 306, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(3, 306, __pyx_L1_error)

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":304
 *             if not PyDataType_HASFIELDS(descr):
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 */
    }

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":307
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 *                 if   t == NPY_BYTE:        f = "b"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_UBYTE:       f = "B"
 *                 elif t == NPY_SHORT:       f = "h"
 */
    switch (__pyx_v_t) {
      case NPY_BYTE:
      __pyx_v_f = ((char *)"b");
      break;
      case NPY_UBYTE:

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":308
 *                     raise ValueError(u"Non-native byte order not supported")
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_SHORT:       f = "h"
 *                 elif t == NPY_USHORT:      f = "H"
 */
      __pyx_v_f = ((char *)"B");
      break;
      case NPY_SHORT:

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":309
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"
 *                 elif t == NPY_SHORT:       f = "h"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_USHORT:      f = "H"
 *                 elif t == NPY_INT:         f = "i"
 */
      __pyx_v_f = ((char *)"h");
      break;
      case NPY_USHORT:

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":310
 *                 elif t == NPY_UBYTE:       f = "B"
 *                 elif t == NPY_SHORT:       f = "h"
 *                 elif t == NPY_USHORT:      f = "H"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_INT:         f = "i"
 *                 elif t == NPY_UINT:        f = "I"
 */
      __pyx_v_f = ((char *)"H");
      break;
      case NPY_INT:

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":311
 *                 elif t == NPY_SHORT:       f = "h"
 *                 elif t == NPY_USHORT:      f = "H"
 *                 elif t == NPY_INT:         f = "i"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_UINT:        f = "I"
 *                 elif t == NPY_LONG:        f = "l"
 */
      __pyx_v_f = ((char *)"i");
      break;
      case NPY_UINT:

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":312
 *                 elif t == NPY_USHORT:      f = "H"
 *                 elif t == NPY_INT:         f = "i"
 *                 elif t == NPY_UINT:        f = "I"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_LONG:        f = "l"
 *                 elif t == NPY_ULONG:       f = "L"
 */
      __pyx_v_f = ((char *)"I");
      break;
      case NPY_LONG:

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":313
 *                 elif t == NPY_INT:         f = "i"
 *                 elif t == NPY_UINT:        f = "I"
 *                 elif t == NPY_LONG:        f = "l"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_ULONG:       f = "L"
 *                 elif t == NPY_LONGLONG:    f = "q"
 */
      __pyx_v_f = ((char *)"l");
      break;
      case NPY_ULONG:

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":314
 *                 elif t == NPY_UINT:        f = "I"
 *                 elif t == NPY_LONG:        f = "l"
 *                 elif t == NPY_ULONG:       f = "L"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_LONGLONG:    f = "q"
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 */
      __pyx_v_f = ((char *)"L");
      break;
      case NPY_LONGLONG:

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":315
 *                 elif t == NPY_LONG:        f = "l"
 *                 elif t == NPY_ULONG:       f = "L"
 *                 elif t == NPY_LONGLONG:    f = "q"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 *                 elif t == NPY_FLOAT:       f = "f"
 */
      __pyx_v_f = ((char *)"q");
      break;
      case NPY_ULONGLONG:

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":316
 *                 elif t == NPY_ULONG:       f = "L"
 *                 elif t == NPY_LONGLONG:    f = "q"
 *                 elif t == NPY_ULONGLONG:   f = "Q"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_FLOAT:       f = "f"
 *                 elif t == NPY_DOUBLE:      f = "d"
 */
      __pyx_v_f = ((char *)"Q");
      break;
      case NPY_FLOAT:

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":317
 *                 elif t == NPY_LONGLONG:    f = "q"
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 *                 elif t == NPY_FLOAT:       f = "f"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_DOUBLE:      f = "d"
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 */
      __pyx_v_f = ((char *)"f");
      break;
      case NPY_DOUBLE:

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":318
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 *                 elif t == NPY_FLOAT:       f = "f"
 *                 elif t == NPY_DOUBLE:      f = "d"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 */
      __pyx_v_f = ((char *)"d");
      break;
      case NPY_LONGDOUBLE:

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":319
 *                 elif t == NPY_FLOAT:       f = "f"
 *                 elif t == NPY_DOUBLE:      f = "d"
 *                 elif t == NPY_LONGDOUBLE:  f = "g"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 */
      __pyx_v_f = ((char *)"g");
      break;
      case NPY_CFLOAT:

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":320
 *                 elif t == NPY_DOUBLE:      f = "d"
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 *                 elif t == NPY_CFLOAT:      f = "Zf"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"
 */
      __pyx_v_f = ((char *)"Zf");
      break;
      case NPY_CDOUBLE:

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":321
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 *                 elif t == NPY_CDOUBLE:     f = "Zd"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"
 *                 elif t == NPY_OBJECT:      f = "O"
 */
      __pyx_v_f = ((char *)"Zd");
      break;
      case NPY_CLONGDOUBLE:

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":322
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_OBJECT:      f = "O"
 *                 else:
 */
      __pyx_v_f = ((char *)"Zg");
      break;
      case NPY_OBJECT:

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":323
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"
 *                 elif t == NPY_OBJECT:      f = "O"             # <<<<<<<<<<<<<<
 *                 else:
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 */
      __pyx_v_f = ((char *)"O");
      break;
      default:

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":325
 *                 elif t == NPY_OBJECT:      f = "O"
 *                 else:
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)             # <<<<<<<<<<<<<<
 *                 info.format = f
 *                 return
 */
      __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_t); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 325, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_8 = PyUnicode_Format(__pyx_kp_u_unknown_dtype_code_in_numpy_pxd, __pyx_t_3); if (unlikely(!__pyx_t_8)) __PYX_ERR(3, 325, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 325, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(3, 325, __pyx_L1_error)
      break;
    }

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":326
 *                 else:
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 *                 info.format = f             # <<<<<<<<<<<<<<
 *                 return
 *             else:
 */
    __pyx_v_info->format = __pyx_v_f;

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":327
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 *                 info.format = f
 *                 return             # <<<<<<<<<<<<<<
 *             else:
 *                 info.format = <char*>PyObject_Malloc(_buffer_format_string_len)
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":302
 *             info.obj = self
 * 
 *             if not PyDataType_HASFIELDS(descr):             # <<<<<<<<<<<<<<
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or
 */
  }

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":329
 *                 return
 *             else:
 *                 info.format = <char*>PyObject_Malloc(_buffer_format_string_len)             # <<<<<<<<<<<<<<
 *                 info.format[0] = c'^' # Native data types, manual alignment
 *                 offset = 0
 */
  /*else*/ {
    __pyx_v_info->format = ((char *)PyObject_Malloc(0xFF));

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":330
 *             else:
 *                 info.format = <char*>PyObject_Malloc(_buffer_format_string_len)
 *                 info.format[0] = c'^' # Native data types, manual alignment             # <<<<<<<<<<<<<<
 *                 offset = 0
 *                 f = _util_dtypestring(descr, info.format + 1,
 */
    (__pyx_v_info->format[0]) = '^';

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":331
 *                 info.format = <char*>PyObject_Malloc(_buffer_format_string_len)
 *                 info.format[0] = c'^' # Native data types, manual alignment
 *                 offset = 0             # <<<<<<<<<<<<<<
 *                 f = _util_dtypestring(descr, info.format + 1,
 *                                       info.format + _buffer_format_string_len,
 */
    __pyx_v_offset = 0;

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":332
 *                 info.format[0] = c'^' # Native data types, manual alignment
 *                 offset = 0
 *                 f = _util_dtypestring(descr, info.format + 1,             # <<<<<<<<<<<<<<
 *                                       info.format + _buffer_format_string_len,
 *                                       &offset)
 */
    __pyx_t_9 = __pyx_f_5numpy__util_dtypestring(__pyx_v_descr, (__pyx_v_info->format + 1), (__pyx_v_info->format + 0xFF), (&__pyx_v_offset)); if (unlikely(__pyx_t_9 == ((char *)NULL))) __PYX_ERR(3, 332, __pyx_L1_error)
    __pyx_v_f = __pyx_t_9;

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":335
 *                                       info.format + _buffer_format_string_len,
 *                                       &offset)
 *                 f[0] = c'\0' # Terminate format string             # <<<<<<<<<<<<<<
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 */
    (__pyx_v_f[0]) = '\x00';
  }

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":258
 *         # experimental exception made for __getbuffer__ and __releasebuffer__
 *         # -- the details of this may change.
 *         def __getbuffer__(ndarray self, Py_buffer* info, int flags):             # <<<<<<<<<<<<<<
 *             # This implementation of getbuffer is geared towards Cython
 *             # requirements, and does not yet fulfill the PEP.
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.ndarray.__getbuffer__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  if (__pyx_v_info->obj != NULL) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  goto __pyx_L2;
  __pyx_L0:;
  if (__pyx_v_info->obj == Py_None) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_descr);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":337
 *                 f[0] = c'\0' # Terminate format string
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):             # <<<<<<<<<<<<<<
 *             if PyArray_HASFIELDS(self):
 *                 PyObject_Free(info.format)
 */

/* Python wrapper */
static CYTHON_UNUSED void __pyx_pw_5numpy_7ndarray_3__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info); /*proto*/
static CYTHON_UNUSED void __pyx_pw_5numpy_7ndarray_3__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__releasebuffer__ (wrapper)", 0);
  __pyx_pf_5numpy_7ndarray_2__releasebuffer__(((PyArrayObject *)__pyx_v_self), ((Py_buffer *)__pyx_v_info));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_5numpy_7ndarray_2__releasebuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__releasebuffer__", 0);

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":338
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 *             if PyArray_HASFIELDS(self):             # <<<<<<<<<<<<<<
 *                 PyObject_Free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 */
  __pyx_t_1 = (PyArray_HASFIELDS(__pyx_v_self) != 0);
  if (__pyx_t_1) {

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":339
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 *             if PyArray_HASFIELDS(self):
 *                 PyObject_Free(info.format)             # <<<<<<<<<<<<<<
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 *                 PyObject_Free(info.strides)
 */
    PyObject_Free(__pyx_v_info->format);

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":338
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 *             if PyArray_HASFIELDS(self):             # <<<<<<<<<<<<<<
 *                 PyObject_Free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 */
  }

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":340
 *             if PyArray_HASFIELDS(self):
 *                 PyObject_Free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 PyObject_Free(info.strides)
 *                 # info.shape was stored after info.strides in the same block
 */
  __pyx_t_1 = (((sizeof(npy_intp)) != (sizeof(Py_ssize_t))) != 0);
  if (__pyx_t_1) {

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":341
 *                 PyObject_Free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 *                 PyObject_Free(info.strides)             # <<<<<<<<<<<<<<
 *                 # info.shape was stored after info.strides in the same block
 * 
 */
    PyObject_Free(__pyx_v_info->strides);

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":340
 *             if PyArray_HASFIELDS(self):
 *                 PyObject_Free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 PyObject_Free(info.strides)
 *                 # info.shape was stored after info.strides in the same block
 */
  }

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":337
 *                 f[0] = c'\0' # Terminate format string
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):             # <<<<<<<<<<<<<<
 *             if PyArray_HASFIELDS(self):
 *                 PyObject_Free(info.format)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":821
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew1(PyObject *__pyx_v_a) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew1", 0);

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":822
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 *     return PyArray_MultiIterNew(1, <void*>a)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(1, ((void *)__pyx_v_a)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 822, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":821
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":824
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew2(PyObject *__pyx_v_a, PyObject *__pyx_v_b) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew2", 0);

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":825
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(2, ((void *)__pyx_v_a), ((void *)__pyx_v_b)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 825, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":824
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":827
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew3(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew3", 0);

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":828
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(3, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":827
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew3", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":830
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew4(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew4", 0);

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":831
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(4, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 831, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":830
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":833
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew5(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d, PyObject *__pyx_v_e) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew5", 0);

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":834
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)             # <<<<<<<<<<<<<<
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(5, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d), ((void *)__pyx_v_e)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 834, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":833
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew5", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":836
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyDataType_SHAPE(PyArray_Descr *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("PyDataType_SHAPE", 0);

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":837
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */
  __pyx_t_1 = (PyDataType_HASSUBARRAY(__pyx_v_d) != 0);
  if (__pyx_t_1) {

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":838
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape             # <<<<<<<<<<<<<<
 *     else:
 *         return ()
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject*)__pyx_v_d->subarray->shape));
    __pyx_r = ((PyObject*)__pyx_v_d->subarray->shape);
    goto __pyx_L0;

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":837
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */
  }

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":840
 *         return <tuple>d.subarray.shape
 *     else:
 *         return ()             # <<<<<<<<<<<<<<
 * 
 * cdef inline char* _util_dtypestring(dtype descr, char* f, char* end, int* offset) except NULL:
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_empty_tuple);
    __pyx_r = __pyx_empty_tuple;
    goto __pyx_L0;
  }

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":836
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":842
 *         return ()
 * 
 * cdef inline char* _util_dtypestring(dtype descr, char* f, char* end, int* offset) except NULL:             # <<<<<<<<<<<<<<
 *     # Recursive utility function used in __getbuffer__ to get format
 *     # string. The new location in the format string is returned.
 */

static CYTHON_INLINE char *__pyx_f_5numpy__util_dtypestring(PyArray_Descr *__pyx_v_descr, char *__pyx_v_f, char *__pyx_v_end, int *__pyx_v_offset) {
  PyArray_Descr *__pyx_v_child = 0;
  int __pyx_v_endian_detector;
  int __pyx_v_little_endian;
  PyObject *__pyx_v_fields = 0;
  PyObject *__pyx_v_childname = NULL;
  PyObject *__pyx_v_new_offset = NULL;
  PyObject *__pyx_v_t = NULL;
  char *__pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  long __pyx_t_8;
  char *__pyx_t_9;
  __Pyx_RefNannySetupContext("_util_dtypestring", 0);

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":847
 * 
 *     cdef dtype child
 *     cdef int endian_detector = 1             # <<<<<<<<<<<<<<
 *     cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)
 *     cdef tuple fields
 */
  __pyx_v_endian_detector = 1;

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":848
 *     cdef dtype child
 *     cdef int endian_detector = 1
 *     cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)             # <<<<<<<<<<<<<<
 *     cdef tuple fields
 * 
 */
  __pyx_v_little_endian = ((((char *)(&__pyx_v_endian_detector))[0]) != 0);

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":851
 *     cdef tuple fields
 * 
 *     for childname in descr.names:             # <<<<<<<<<<<<<<
 *         fields = descr.fields[childname]
 *         child, new_offset = fields
 */
  if (unlikely(__pyx_v_descr->names == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(3, 851, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_descr->names; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(3, 851, __pyx_L1_error)
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 851, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_childname, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":852
 * 
 *     for childname in descr.names:
 *         fields = descr.fields[childname]             # <<<<<<<<<<<<<<
 *         child, new_offset = fields
 * 
 */
    if (unlikely(__pyx_v_descr->fields == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(3, 852, __pyx_L1_error)
    }
    __pyx_t_3 = __Pyx_PyDict_GetItem(__pyx_v_descr->fields, __pyx_v_childname); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 852, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(PyTuple_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_t_3)->tp_name), 0))) __PYX_ERR(3, 852, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_fields, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":853
 *     for childname in descr.names:
 *         fields = descr.fields[childname]
 *         child, new_offset = fields             # <<<<<<<<<<<<<<
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:
 */
    if (likely(__pyx_v_fields != Py_None)) {
      PyObject* sequence = __pyx_v_fields;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(3, 853, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      #else
      __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 853, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 853, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
    } else {
      __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(3, 853, __pyx_L1_error)
    }
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_dtype))))) __PYX_ERR(3, 853, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_child, ((PyArray_Descr *)__pyx_t_3));
    __pyx_t_3 = 0;
    __Pyx_XDECREF_SET(__pyx_v_new_offset, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":855
 *         child, new_offset = fields
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:             # <<<<<<<<<<<<<<
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 */
    __pyx_t_4 = __Pyx_PyInt_From_int((__pyx_v_offset[0])); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 855, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = PyNumber_Subtract(__pyx_v_new_offset, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 855, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(3, 855, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_6 = ((((__pyx_v_end - __pyx_v_f) - ((int)__pyx_t_5)) < 15) != 0);
    if (unlikely(__pyx_t_6)) {

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":856
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")             # <<<<<<<<<<<<<<
 * 
 *         if ((child.byteorder == c'>' and little_endian) or
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__28, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 856, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(3, 856, __pyx_L1_error)

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":855
 *         child, new_offset = fields
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:             # <<<<<<<<<<<<<<
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 */
    }

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":858
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 *         if ((child.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")
 */
    __pyx_t_7 = ((__pyx_v_child->byteorder == '>') != 0);
    if (!__pyx_t_7) {
      goto __pyx_L8_next_or;
    } else {
    }
    __pyx_t_7 = (__pyx_v_little_endian != 0);
    if (!__pyx_t_7) {
    } else {
      __pyx_t_6 = __pyx_t_7;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_L8_next_or:;

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":859
 * 
 *         if ((child.byteorder == c'>' and little_endian) or
 *             (child.byteorder == c'<' and not little_endian)):             # <<<<<<<<<<<<<<
 *             raise ValueError(u"Non-native byte order not supported")
 *             # One could encode it in the format string and have Cython
 */
    __pyx_t_7 = ((__pyx_v_child->byteorder == '<') != 0);
    if (__pyx_t_7) {
    } else {
      __pyx_t_6 = __pyx_t_7;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_7 = ((!(__pyx_v_little_endian != 0)) != 0);
    __pyx_t_6 = __pyx_t_7;
    __pyx_L7_bool_binop_done:;

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":858
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 *         if ((child.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")
 */
    if (unlikely(__pyx_t_6)) {

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":860
 *         if ((child.byteorder == c'>' and little_endian) or
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *             # One could encode it in the format string and have Cython
 *             # complain instead, BUT: < and > in format strings also imply
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__27, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 860, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(3, 860, __pyx_L1_error)

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":858
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 *         if ((child.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")
 */
    }

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":870
 * 
 *         # Output padding bytes
 *         while offset[0] < new_offset:             # <<<<<<<<<<<<<<
 *             f[0] = 120 # "x"; pad byte
 *             f += 1
 */
    while (1) {
      __pyx_t_3 = __Pyx_PyInt_From_int((__pyx_v_offset[0])); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 870, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_t_3, __pyx_v_new_offset, Py_LT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 870, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 870, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (!__pyx_t_6) break;

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":871
 *         # Output padding bytes
 *         while offset[0] < new_offset:
 *             f[0] = 120 # "x"; pad byte             # <<<<<<<<<<<<<<
 *             f += 1
 *             offset[0] += 1
 */
      (__pyx_v_f[0]) = 0x78;

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":872
 *         while offset[0] < new_offset:
 *             f[0] = 120 # "x"; pad byte
 *             f += 1             # <<<<<<<<<<<<<<
 *             offset[0] += 1
 * 
 */
      __pyx_v_f = (__pyx_v_f + 1);

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":873
 *             f[0] = 120 # "x"; pad byte
 *             f += 1
 *             offset[0] += 1             # <<<<<<<<<<<<<<
 * 
 *         offset[0] += child.itemsize
 */
      __pyx_t_8 = 0;
      (__pyx_v_offset[__pyx_t_8]) = ((__pyx_v_offset[__pyx_t_8]) + 1);
    }

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":875
 *             offset[0] += 1
 * 
 *         offset[0] += child.itemsize             # <<<<<<<<<<<<<<
 * 
 *         if not PyDataType_HASFIELDS(child):
 */
    __pyx_t_8 = 0;
    (__pyx_v_offset[__pyx_t_8]) = ((__pyx_v_offset[__pyx_t_8]) + __pyx_v_child->elsize);

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":877
 *         offset[0] += child.itemsize
 * 
 *         if not PyDataType_HASFIELDS(child):             # <<<<<<<<<<<<<<
 *             t = child.type_num
 *             if end - f < 5:
 */
    __pyx_t_6 = ((!(PyDataType_HASFIELDS(__pyx_v_child) != 0)) != 0);
    if (__pyx_t_6) {

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":878
 * 
 *         if not PyDataType_HASFIELDS(child):
 *             t = child.type_num             # <<<<<<<<<<<<<<
 *             if end - f < 5:
 *                 raise RuntimeError(u"Format string allocated too short.")
 */
      __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_child->type_num); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 878, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_XDECREF_SET(__pyx_v_t, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":879
 *         if not PyDataType_HASFIELDS(child):
 *             t = child.type_num
 *             if end - f < 5:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(u"Format string allocated too short.")
 * 
 */
      __pyx_t_6 = (((__pyx_v_end - __pyx_v_f) < 5) != 0);
      if (unlikely(__pyx_t_6)) {

        /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":880
 *             t = child.type_num
 *             if end - f < 5:
 *                 raise RuntimeError(u"Format string allocated too short.")             # <<<<<<<<<<<<<<
 * 
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 */
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__29, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 880, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_Raise(__pyx_t_4, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __PYX_ERR(3, 880, __pyx_L1_error)

        /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":879
 *         if not PyDataType_HASFIELDS(child):
 *             t = child.type_num
 *             if end - f < 5:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(u"Format string allocated too short.")
 * 
 */
      }

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":883
 * 
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 *             if   t == NPY_BYTE:        f[0] =  98 #"b"             # <<<<<<<<<<<<<<
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_BYTE); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 883, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 883, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 883, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 98;
        goto __pyx_L15;
      }

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":884
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 *             if   t == NPY_BYTE:        f[0] =  98 #"b"
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"             # <<<<<<<<<<<<<<
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_UBYTE); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 884, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 884, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 884, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 66;
        goto __pyx_L15;
      }

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":885
 *             if   t == NPY_BYTE:        f[0] =  98 #"b"
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"             # <<<<<<<<<<<<<<
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_SHORT); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 885, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 885, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 885, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x68;
        goto __pyx_L15;
      }

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":886
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"             # <<<<<<<<<<<<<<
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_USHORT); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 886, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 886, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 886, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 72;
        goto __pyx_L15;
      }

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":887
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 *             elif t == NPY_INT:         f[0] = 105 #"i"             # <<<<<<<<<<<<<<
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_INT); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 887, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 887, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 887, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x69;
        goto __pyx_L15;
      }

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":888
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 *             elif t == NPY_UINT:        f[0] =  73 #"I"             # <<<<<<<<<<<<<<
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_UINT); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 888, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 888, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 888, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 73;
        goto __pyx_L15;
      }

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":889
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 *             elif t == NPY_LONG:        f[0] = 108 #"l"             # <<<<<<<<<<<<<<
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_LONG); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 889, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 889, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 889, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x6C;
        goto __pyx_L15;
      }

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":890
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"             # <<<<<<<<<<<<<<
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_ULONG); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 890, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 890, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 890, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 76;
        goto __pyx_L15;
      }

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":891
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"             # <<<<<<<<<<<<<<
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_LONGLONG); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 891, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 891, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 891, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x71;
        goto __pyx_L15;
      }

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":892
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"             # <<<<<<<<<<<<<<
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_ULONGLONG); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 892, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 892, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 892, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 81;
        goto __pyx_L15;
      }

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":893
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"             # <<<<<<<<<<<<<<
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_FLOAT); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 893, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 893, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 893, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x66;
        goto __pyx_L15;
      }

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":894
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"             # <<<<<<<<<<<<<<
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_DOUBLE); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 894, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 894, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 894, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x64;
        goto __pyx_L15;
      }

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":895
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"             # <<<<<<<<<<<<<<
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_LONGDOUBLE); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 895, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 895, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 895, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x67;
        goto __pyx_L15;
      }

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":896
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf             # <<<<<<<<<<<<<<
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_CFLOAT); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 896, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 896, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 896, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 90;
        (__pyx_v_f[1]) = 0x66;
        __pyx_v_f = (__pyx_v_f + 1);
        goto __pyx_L15;
      }

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":897
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd             # <<<<<<<<<<<<<<
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_CDOUBLE); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 897, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 897, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 897, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 90;
        (__pyx_v_f[1]) = 0x64;
        __pyx_v_f = (__pyx_v_f + 1);
        goto __pyx_L15;
      }

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":898
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg             # <<<<<<<<<<<<<<
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"
 *             else:
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_CLONGDOUBLE); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 898, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 898, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 898, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 90;
        (__pyx_v_f[1]) = 0x67;
        __pyx_v_f = (__pyx_v_f + 1);
        goto __pyx_L15;
      }

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":899
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"             # <<<<<<<<<<<<<<
 *             else:
 *                 raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_OBJECT); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 899, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 899, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 899, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (likely(__pyx_t_6)) {
        (__pyx_v_f[0]) = 79;
        goto __pyx_L15;
      }

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":901
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"
 *             else:
 *                 raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)             # <<<<<<<<<<<<<<
 *             f += 1
 *         else:
 */
      /*else*/ {
        __pyx_t_3 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_unknown_dtype_code_in_numpy_pxd, __pyx_v_t); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 901, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 901, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_Raise(__pyx_t_4, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __PYX_ERR(3, 901, __pyx_L1_error)
      }
      __pyx_L15:;

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":902
 *             else:
 *                 raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 *             f += 1             # <<<<<<<<<<<<<<
 *         else:
 *             # Cython ignores struct boundary information ("T{...}"),
 */
      __pyx_v_f = (__pyx_v_f + 1);

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":877
 *         offset[0] += child.itemsize
 * 
 *         if not PyDataType_HASFIELDS(child):             # <<<<<<<<<<<<<<
 *             t = child.type_num
 *             if end - f < 5:
 */
      goto __pyx_L13;
    }

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":906
 *             # Cython ignores struct boundary information ("T{...}"),
 *             # so don't output it
 *             f = _util_dtypestring(child, f, end, offset)             # <<<<<<<<<<<<<<
 *     return f
 * 
 */
    /*else*/ {
      __pyx_t_9 = __pyx_f_5numpy__util_dtypestring(__pyx_v_child, __pyx_v_f, __pyx_v_end, __pyx_v_offset); if (unlikely(__pyx_t_9 == ((char *)NULL))) __PYX_ERR(3, 906, __pyx_L1_error)
      __pyx_v_f = __pyx_t_9;
    }
    __pyx_L13:;

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":851
 *     cdef tuple fields
 * 
 *     for childname in descr.names:             # <<<<<<<<<<<<<<
 *         fields = descr.fields[childname]
 *         child, new_offset = fields
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":907
 *             # so don't output it
 *             f = _util_dtypestring(child, f, end, offset)
 *     return f             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_f;
  goto __pyx_L0;

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":842
 *         return ()
 * 
 * cdef inline char* _util_dtypestring(dtype descr, char* f, char* end, int* offset) except NULL:             # <<<<<<<<<<<<<<
 *     # Recursive utility function used in __getbuffer__ to get format
 *     # string. The new location in the format string is returned.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("numpy._util_dtypestring", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_child);
  __Pyx_XDECREF(__pyx_v_fields);
  __Pyx_XDECREF(__pyx_v_childname);
  __Pyx_XDECREF(__pyx_v_new_offset);
  __Pyx_XDECREF(__pyx_v_t);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1022
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

static CYTHON_INLINE void __pyx_f_5numpy_set_array_base(PyArrayObject *__pyx_v_arr, PyObject *__pyx_v_base) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_array_base", 0);

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1023
 * 
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!             # <<<<<<<<<<<<<<
 *     PyArray_SetBaseObject(arr, base)
 * 
 */
  Py_INCREF(__pyx_v_base);

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1024
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object get_array_base(ndarray arr):
 */
  (void)(PyArray_SetBaseObject(__pyx_v_arr, __pyx_v_base));

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1022
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1026
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_get_array_base(PyArrayObject *__pyx_v_arr) {
  PyObject *__pyx_v_base;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("get_array_base", 0);

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1027
 * 
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)             # <<<<<<<<<<<<<<
 *     if base is NULL:
 *         return None
 */
  __pyx_v_base = PyArray_BASE(__pyx_v_arr);

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1028
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */
  __pyx_t_1 = ((__pyx_v_base == NULL) != 0);
  if (__pyx_t_1) {

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1029
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 *         return None             # <<<<<<<<<<<<<<
 *     return <object>base
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1028
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */
  }

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1030
 *     if base is NULL:
 *         return None
 *     return <object>base             # <<<<<<<<<<<<<<
 * 
 * # Versions of the import_* functions which are more suitable for
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_base));
  __pyx_r = ((PyObject *)__pyx_v_base);
  goto __pyx_L0;

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1026
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1034
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_array()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_array(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("import_array", 0);

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1035
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_array()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1036
 * cdef inline int import_array() except -1:
 *     try:
 *         _import_array()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")
 */
      __pyx_t_4 = _import_array(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(3, 1036, __pyx_L3_error)

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1035
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_array()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1037
 *     try:
 *         _import_array()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(3, 1037, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1038
 *         _import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__30, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(3, 1038, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(3, 1038, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1035
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_array()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1034
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_array()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1040
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_umath(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("import_umath", 0);

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1041
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1042
 * cdef inline int import_umath() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(3, 1042, __pyx_L3_error)

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1041
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1043
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(3, 1043, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1044
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__31, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(3, 1044, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(3, 1044, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1041
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1040
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1046
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_ufunc(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("import_ufunc", 0);

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1047
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1048
 * cdef inline int import_ufunc() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(3, 1048, __pyx_L3_error)

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1047
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1049
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(3, 1049, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1050
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__31, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(3, 1050, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(3, 1050, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1047
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1046
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase __pyx_vtable_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase;

static PyObject *__pyx_tp_new_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *)o);
  p->__pyx_vtab = __pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase;
  p->scale_method = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->peaklist = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)Py_None); Py_INCREF(Py_None);
  p->scorer = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *)Py_None); Py_INCREF(Py_None);
  p->_slice_cache = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase(PyObject *o) {
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *p = (struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->scale_method);
  Py_CLEAR(p->peaklist);
  Py_CLEAR(p->scorer);
  Py_CLEAR(p->_slice_cache);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *p = (struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *)o;
  if (p->peaklist) {
    e = (*v)(((PyObject *)p->peaklist), a); if (e) return e;
  }
  if (p->scorer) {
    e = (*v)(((PyObject *)p->scorer), a); if (e) return e;
  }
  if (p->_slice_cache) {
    e = (*v)(p->_slice_cache, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *p = (struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *)o;
  tmp = ((PyObject*)p->peaklist);
  p->peaklist = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->scorer);
  p->scorer = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_slice_cache);
  p->_slice_cache = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_use_subtraction(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_15use_subtraction_1__get__(o);
}

static int __pyx_setprop_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_use_subtraction(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_15use_subtraction_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_scale_method(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_12scale_method_1__get__(o);
}

static int __pyx_setprop_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_scale_method(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_12scale_method_3__set__(o, v);
  }
  else {
    return __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_12scale_method_5__del__(o);
  }
}

static PyObject *__pyx_getprop_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_merge_isobaric_peaks(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_20merge_isobaric_peaks_1__get__(o);
}

static int __pyx_setprop_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_merge_isobaric_peaks(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_20merge_isobaric_peaks_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_minimum_intensity(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_17minimum_intensity_1__get__(o);
}

static int __pyx_setprop_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_minimum_intensity(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_17minimum_intensity_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_peaklist(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_8peaklist_1__get__(o);
}

static int __pyx_setprop_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_peaklist(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_8peaklist_3__set__(o, v);
  }
  else {
    return __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_8peaklist_5__del__(o);
  }
}

static PyObject *__pyx_getprop_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_scorer(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_6scorer_1__get__(o);
}

static int __pyx_setprop_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_scorer(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_6scorer_3__set__(o, v);
  }
  else {
    return __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_6scorer_5__del__(o);
  }
}

static PyObject *__pyx_getprop_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_verbose(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_7verbose_1__get__(o);
}

static int __pyx_setprop_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_verbose(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_7verbose_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__slice_cache(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_12_slice_cache_1__get__(o);
}

static int __pyx_setprop_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__slice_cache(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_12_slice_cache_3__set__(o, v);
  }
  else {
    return __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_12_slice_cache_5__del__(o);
  }
}

static PyMethodDef __pyx_methods_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase[] = {
  {"between", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_3between, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_2between},
  {"has_peak", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_5has_peak, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_4has_peak},
  {"match_theoretical_isotopic_distribution", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_7match_theoretical_isotopic_distribution, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_6match_theoretical_isotopic_distribution},
  {"scale_theoretical_distribution", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_9scale_theoretical_distribution, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_8scale_theoretical_distribution},
  {"_evaluate_theoretical_distribution", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_11_evaluate_theoretical_distribution, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_10_evaluate_theoretical_distribution},
  {"fit_incremental_truncation", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_13fit_incremental_truncation, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_12fit_incremental_truncation},
  {"subtraction", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_15subtraction, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_14subtraction},
  {"_check_fit", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_17_check_fit, METH_O, __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_16_check_fit},
  {"_merge_peaks", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_19_merge_peaks, METH_O, __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_18_merge_peaks},
  {"_find_next_putative_peak", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_21_find_next_putative_peak, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_20_find_next_putative_peak},
  {"_find_previous_putative_peak", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_23_find_previous_putative_peak, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_22_find_previous_putative_peak},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_27__reduce_cython__, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_26__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_29__setstate_cython__, METH_O, __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_28__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase[] = {
  {(char *)"use_subtraction", __pyx_getprop_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_use_subtraction, __pyx_setprop_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_use_subtraction, (char *)"use_subtraction: 'bool'", 0},
  {(char *)"scale_method", __pyx_getprop_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_scale_method, __pyx_setprop_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_scale_method, (char *)"scale_method: str", 0},
  {(char *)"merge_isobaric_peaks", __pyx_getprop_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_merge_isobaric_peaks, __pyx_setprop_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_merge_isobaric_peaks, (char *)"merge_isobaric_peaks: 'bool'", 0},
  {(char *)"minimum_intensity", __pyx_getprop_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_minimum_intensity, __pyx_setprop_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_minimum_intensity, (char *)"minimum_intensity: 'double'", 0},
  {(char *)"peaklist", __pyx_getprop_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_peaklist, __pyx_setprop_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_peaklist, (char *)"peaklist: ms_peak_picker._c.peak_set.PeakSet", 0},
  {(char *)"scorer", __pyx_getprop_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_scorer, __pyx_setprop_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_scorer, (char *)"scorer: ms_deisotope._c.scoring.IsotopicFitterBase", 0},
  {(char *)"verbose", __pyx_getprop_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_verbose, __pyx_setprop_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_verbose, (char *)"verbose: 'bool'", 0},
  {(char *)"_slice_cache", __pyx_getprop_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__slice_cache, __pyx_setprop_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__slice_cache, (char *)"_slice_cache: dict", 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase = {
  PyVarObject_HEAD_INIT(0, 0)
  "ms_deisotope._c.deconvoluter_base.DeconvoluterBase", /*tp_name*/
  sizeof(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_25__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "DeconvoluterBase(use_subtraction=False, scale_method='sum', merge_isobaric_peaks=True, minimum_intensity=5., *args, **kwargs)\nBase class for all Deconvoluter types. Provides basic configuration for common operations,\n    regardless of implementation. Because these methods form the backbone of all deconvolution algorithms,\n    this class has a C-extension implementation as well.\n\n    Attributes\n    ----------\n    peaklist : ms_peak_picker.PeakSet\n        The centroided mass spectrum to deconvolute\n    scorer : IsotopicFitterBase\n        The criterion for evaluating individual isotopic pattern fits\n    merge_isobaric_peaks : bool\n        If multiple passes produce peaks with identical mass values,\n        should those peaks be summed\n    minimum_intensity : float\n        Experimental peaks whose intensity is below this level will be ignored\n        by peak querying methods\n    scale_method : str\n        The name of the method to use to scale theoretical isotopic pattern intensities\n        to match the experimental isotopic pattern. For a description of options, see\n        :meth:`~.TheoreticalIsotopicPattern.scale`.\n    use_subtraction : bool\n        Whether or not to apply a subtraction procedure to experimental peaks after they\n        have been fitted. This is only necessary if the same signal may be examined multiple\n        times as in a multi-pass method or when peak dependence is not considered\n    verbose : bool\n        Produce extra logging information\n    ", /*tp_doc*/
  __pyx_tp_traverse_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase, /*tp_traverse*/
  __pyx_tp_clear_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase __pyx_vtable_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase;

static PyObject *__pyx_tp_new_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *p;
  PyObject *o = __pyx_tp_new_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase*)__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase;
  p->averagine = ((struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase(PyObject *o) {
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *p = (struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->averagine);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase(o);
}

static int __pyx_tp_traverse_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *p = (struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *)o;
  e = __pyx_tp_traverse_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase(o, v, a); if (e) return e;
  if (p->averagine) {
    e = (*v)(((PyObject *)p->averagine), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *p = (struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *)o;
  __pyx_tp_clear_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase(o);
  tmp = ((PyObject*)p->averagine);
  p->averagine = ((struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_averagine(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_9averagine_1__get__(o);
}

static int __pyx_setprop_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_averagine(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_9averagine_3__set__(o, v);
  }
  else {
    return __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_9averagine_5__del__(o);
  }
}

static PyMethodDef __pyx_methods_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase[] = {
  {"fit_theoretical_distribution", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_3fit_theoretical_distribution, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_2fit_theoretical_distribution},
  {"_fit_peaks_at_charges", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_5_fit_peaks_at_charges, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_4_fit_peaks_at_charges},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_7__reduce_cython__, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_6__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_9__setstate_cython__, METH_O, __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_8__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase[] = {
  {(char *)"averagine", __pyx_getprop_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_averagine, __pyx_setprop_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_averagine, (char *)"averagine: ms_deisotope._c.averagine.AveragineCache", 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase = {
  PyVarObject_HEAD_INIT(0, 0)
  "ms_deisotope._c.deconvoluter_base.AveragineDeconvoluterBase", /*tp_name*/
  sizeof(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY
  __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_25__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "AveragineDeconvoluterBase(bool use_subtraction=False, str scale_method='sum', bool merge_isobaric_peaks=True, double minimum_intensity=5., *args, **kwargs)\nA base class derived from :class:`DeconvoluterBase` which provides some common methods\n    for fitting isotopic patterns using an Averagine model.\n    ", /*tp_doc*/
  __pyx_tp_traverse_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase, /*tp_traverse*/
  __pyx_tp_clear_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase __pyx_vtable_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase;

static PyObject *__pyx_tp_new_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *p;
  PyObject *o = __pyx_tp_new_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase*)__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase;
  p->averagines = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase(PyObject *o) {
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *p = (struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->averagines);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase(o);
}

static int __pyx_tp_traverse_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *p = (struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *)o;
  e = __pyx_tp_traverse_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase(o, v, a); if (e) return e;
  if (p->averagines) {
    e = (*v)(p->averagines, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *p = (struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *)o;
  __pyx_tp_clear_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase(o);
  tmp = ((PyObject*)p->averagines);
  p->averagines = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_averagines(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_10averagines_1__get__(o);
}

static int __pyx_setprop_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_averagines(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_10averagines_3__set__(o, v);
  }
  else {
    return __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_10averagines_5__del__(o);
  }
}

static PyMethodDef __pyx_methods_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase[] = {
  {"fit_theoretical_distribution", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_3fit_theoretical_distribution, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_2fit_theoretical_distribution},
  {"_fit_peaks_at_charges", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_5_fit_peaks_at_charges, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_4_fit_peaks_at_charges},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_7__reduce_cython__, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_6__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_9__setstate_cython__, METH_O, __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_8__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase[] = {
  {(char *)"averagines", __pyx_getprop_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_averagines, __pyx_setprop_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_averagines, (char *)"averagines: list", 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase = {
  PyVarObject_HEAD_INIT(0, 0)
  "ms_deisotope._c.deconvoluter_base.MultiAveragineDeconvoluterBase", /*tp_name*/
  sizeof(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY
  __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_25__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "MultiAveragineDeconvoluterBase(bool use_subtraction=False, str scale_method='sum', bool merge_isobaric_peaks=True, double minimum_intensity=5., *args, **kwargs)\nA base class derived from :class:`DeconvoluterBase` which provides some common methods\n    for fitting isotopic patterns using multiple Averagine models.\n    ", /*tp_doc*/
  __pyx_tp_traverse_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase, /*tp_traverse*/
  __pyx_tp_clear_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator __pyx_vtable_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator;

static struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_freelist_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator[10];
static int __pyx_freecount_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator = 0;

static PyObject *__pyx_tp_new_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *p;
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator)))) {
    o = (PyObject*)__pyx_freelist_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator[--__pyx_freecount_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator];
    memset(o, 0, sizeof(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator));
    (void) PyObject_INIT(o, t);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  p = ((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *)o);
  p->__pyx_vtab = __pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator;
  return o;
}

static void __pyx_tp_dealloc_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator(PyObject *o) {
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_3__dealloc__(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator < 10) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator)))) {
    __pyx_freelist_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator[__pyx_freecount_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator++] = ((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static PyObject *__pyx_getprop_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_lower(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_5lower_1__get__(o);
}

static int __pyx_setprop_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_lower(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_5lower_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_upper(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_5upper_1__get__(o);
}

static int __pyx_setprop_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_upper(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_5upper_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_sign(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_4sign_1__get__(o);
}

static int __pyx_setprop_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_sign(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_4sign_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_size(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_4size_1__get__(o);
}

static int __pyx_setprop_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_size(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_4size_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_index(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_5index_1__get__(o);
}

static int __pyx_setprop_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_index(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_5index_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_specialmethod___pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_19__next__(PyObject *self, CYTHON_UNUSED PyObject *arg) {return __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_19__next__(self);}

static PyMethodDef __pyx_methods_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator[] = {
  {"sequence_from_quickcharge", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_5sequence_from_quickcharge, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_4sequence_from_quickcharge},
  {"make_sequence", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_7make_sequence, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_6make_sequence},
  {"reset", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_9reset, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_8reset},
  {"has_more", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_11has_more, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_10has_more},
  {"get_next_value", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_13get_next_value, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_12get_next_value},
  {"__next__", (PyCFunction)__pyx_specialmethod___pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_19__next__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_21__reduce_cython__, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_20__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_23__setstate_cython__, METH_O, __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_22__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator[] = {
  {(char *)"lower", __pyx_getprop_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_lower, __pyx_setprop_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_lower, (char *)"lower: 'int'", 0},
  {(char *)"upper", __pyx_getprop_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_upper, __pyx_setprop_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_upper, (char *)"upper: 'int'", 0},
  {(char *)"sign", __pyx_getprop_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_sign, __pyx_setprop_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_sign, (char *)"sign: 'int'", 0},
  {(char *)"size", __pyx_getprop_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_size, __pyx_setprop_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_size, (char *)"size: 'size_t'", 0},
  {(char *)"index", __pyx_getprop_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_index, __pyx_setprop_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_index, (char *)"index: 'size_t'", 0},
  {0, 0, 0, 0, 0}
};

static PySequenceMethods __pyx_tp_as_sequence_ChargeIterator = {
  __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_15__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  0, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_ChargeIterator = {
  __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_15__len__, /*mp_length*/
  0, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator = {
  PyVarObject_HEAD_INIT(0, 0)
  "ms_deisotope._c.deconvoluter_base.ChargeIterator", /*tp_name*/
  sizeof(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_ChargeIterator, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_ChargeIterator, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER, /*tp_flags*/
  "ChargeIterator(int lo, int hi)", /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_17__iter__, /*tp_iter*/
  __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_19__next__, /*tp_iternext*/
  __pyx_methods_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {"_get_all_peak_charge_pairs", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_3_get_all_peak_charge_pairs, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_2_get_all_peak_charge_pairs},
  {"_explore_local", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_5_explore_local, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_4_explore_local},
  {"populate_graph", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_7populate_graph, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_6populate_graph},
  {"_make_deconvoluted_peak", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_17deconvoluter_base_9_make_deconvoluted_peak, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_17deconvoluter_base_8_make_deconvoluted_peak},
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_deconvoluter_base(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_deconvoluter_base},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "deconvoluter_base",
    0, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_n_s_AveragineDeconvoluterBase, __pyx_k_AveragineDeconvoluterBase, sizeof(__pyx_k_AveragineDeconvoluterBase), 0, 0, 1, 1},
  {&__pyx_n_s_ChargeIterator, __pyx_k_ChargeIterator, sizeof(__pyx_k_ChargeIterator), 0, 0, 1, 1},
  {&__pyx_n_s_DeconvoluterBase, __pyx_k_DeconvoluterBase, sizeof(__pyx_k_DeconvoluterBase), 0, 0, 1, 1},
  {&__pyx_n_s_ERROR_TOLERANCE, __pyx_k_ERROR_TOLERANCE, sizeof(__pyx_k_ERROR_TOLERANCE), 0, 0, 1, 1},
  {&__pyx_n_s_ERROR_TOLERANCE_2, __pyx_k_ERROR_TOLERANCE_2, sizeof(__pyx_k_ERROR_TOLERANCE_2), 0, 0, 1, 1},
  {&__pyx_kp_s_Expected_at_least_d_argument_s_g, __pyx_k_Expected_at_least_d_argument_s_g, sizeof(__pyx_k_Expected_at_least_d_argument_s_g), 0, 0, 1, 0},
  {&__pyx_n_s_FittedPeak, __pyx_k_FittedPeak, sizeof(__pyx_k_FittedPeak), 0, 0, 1, 1},
  {&__pyx_kp_u_Format_string_allocated_too_shor, __pyx_k_Format_string_allocated_too_shor, sizeof(__pyx_k_Format_string_allocated_too_shor), 0, 1, 0, 0},
  {&__pyx_kp_u_Format_string_allocated_too_shor_2, __pyx_k_Format_string_allocated_too_shor_2, sizeof(__pyx_k_Format_string_allocated_too_shor_2), 0, 1, 0, 0},
  {&__pyx_kp_s_Function_call_with_ambiguous_arg, __pyx_k_Function_call_with_ambiguous_arg, sizeof(__pyx_k_Function_call_with_ambiguous_arg), 0, 0, 1, 0},
  {&__pyx_n_s_IGNORE_BELOW, __pyx_k_IGNORE_BELOW, sizeof(__pyx_k_IGNORE_BELOW), 0, 0, 1, 1},
  {&__pyx_n_s_IGNORE_BELOW_2, __pyx_k_IGNORE_BELOW_2, sizeof(__pyx_k_IGNORE_BELOW_2), 0, 0, 1, 1},
  {&__pyx_n_s_ImportError, __pyx_k_ImportError, sizeof(__pyx_k_ImportError), 0, 0, 1, 1},
  {&__pyx_kp_s_Incompatible_checksums_s_vs_0x16, __pyx_k_Incompatible_checksums_s_vs_0x16, sizeof(__pyx_k_Incompatible_checksums_s_vs_0x16), 0, 0, 1, 0},
  {&__pyx_kp_s_Incompatible_checksums_s_vs_0x9d, __pyx_k_Incompatible_checksums_s_vs_0x9d, sizeof(__pyx_k_Incompatible_checksums_s_vs_0x9d), 0, 0, 1, 0},
  {&__pyx_kp_s_Incompatible_checksums_s_vs_0xfd, __pyx_k_Incompatible_checksums_s_vs_0xfd, sizeof(__pyx_k_Incompatible_checksums_s_vs_0xfd), 0, 0, 1, 0},
  {&__pyx_n_s_IndexError, __pyx_k_IndexError, sizeof(__pyx_k_IndexError), 0, 0, 1, 1},
  {&__pyx_n_s_MultiAveragineDeconvoluterBase, __pyx_k_MultiAveragineDeconvoluterBase, sizeof(__pyx_k_MultiAveragineDeconvoluterBase), 0, 0, 1, 1},
  {&__pyx_kp_s_No_matching_signature_found, __pyx_k_No_matching_signature_found, sizeof(__pyx_k_No_matching_signature_found), 0, 0, 1, 0},
  {&__pyx_kp_u_Non_native_byte_order_not_suppor, __pyx_k_Non_native_byte_order_not_suppor, sizeof(__pyx_k_Non_native_byte_order_not_suppor), 0, 1, 0, 0},
  {&__pyx_n_s_PeakIndex, __pyx_k_PeakIndex, sizeof(__pyx_k_PeakIndex), 0, 0, 1, 1},
  {&__pyx_kp_s_PeakIndex_FittedPeak, __pyx_k_PeakIndex_FittedPeak, sizeof(__pyx_k_PeakIndex_FittedPeak), 0, 0, 1, 0},
  {&__pyx_kp_s_PeakIndex_long, __pyx_k_PeakIndex_long, sizeof(__pyx_k_PeakIndex_long), 0, 0, 1, 0},
  {&__pyx_kp_s_PeakIndex_size_t, __pyx_k_PeakIndex_size_t, sizeof(__pyx_k_PeakIndex_size_t), 0, 0, 1, 0},
  {&__pyx_n_s_PeakSet, __pyx_k_PeakSet, sizeof(__pyx_k_PeakSet), 0, 0, 1, 1},
  {&__pyx_kp_s_PeakSet_FittedPeak, __pyx_k_PeakSet_FittedPeak, sizeof(__pyx_k_PeakSet_FittedPeak), 0, 0, 1, 0},
  {&__pyx_kp_s_PeakSet_long, __pyx_k_PeakSet_long, sizeof(__pyx_k_PeakSet_long), 0, 0, 1, 0},
  {&__pyx_kp_s_PeakSet_size_t, __pyx_k_PeakSet_size_t, sizeof(__pyx_k_PeakSet_size_t), 0, 0, 1, 0},
  {&__pyx_n_s_PickleError, __pyx_k_PickleError, sizeof(__pyx_k_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_RuntimeError, __pyx_k_RuntimeError, sizeof(__pyx_k_RuntimeError), 0, 0, 1, 1},
  {&__pyx_n_s_StopIteration, __pyx_k_StopIteration, sizeof(__pyx_k_StopIteration), 0, 0, 1, 1},
  {&__pyx_n_s_TRUNCATE_AFTER, __pyx_k_TRUNCATE_AFTER, sizeof(__pyx_k_TRUNCATE_AFTER), 0, 0, 1, 1},
  {&__pyx_n_s_TRUNCATE_AFTER_2, __pyx_k_TRUNCATE_AFTER_2, sizeof(__pyx_k_TRUNCATE_AFTER_2), 0, 0, 1, 1},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_kp_s__10, __pyx_k__10, sizeof(__pyx_k__10), 0, 0, 1, 0},
  {&__pyx_kp_s__8, __pyx_k__8, sizeof(__pyx_k__8), 0, 0, 1, 0},
  {&__pyx_kp_s__9, __pyx_k__9, sizeof(__pyx_k__9), 0, 0, 1, 0},
  {&__pyx_n_s_a_to_a2_ratio, __pyx_k_a_to_a2_ratio, sizeof(__pyx_k_a_to_a2_ratio), 0, 0, 1, 1},
  {&__pyx_n_s_add, __pyx_k_add, sizeof(__pyx_k_add), 0, 0, 1, 1},
  {&__pyx_n_s_area, __pyx_k_area, sizeof(__pyx_k_area), 0, 0, 1, 1},
  {&__pyx_n_s_args, __pyx_k_args, sizeof(__pyx_k_args), 0, 0, 1, 1},
  {&__pyx_n_s_attrgetter, __pyx_k_attrgetter, sizeof(__pyx_k_attrgetter), 0, 0, 1, 1},
  {&__pyx_n_s_average_mass, __pyx_k_average_mass, sizeof(__pyx_k_average_mass), 0, 0, 1, 1},
  {&__pyx_n_s_averagine, __pyx_k_averagine, sizeof(__pyx_k_averagine), 0, 0, 1, 1},
  {&__pyx_n_s_between, __pyx_k_between, sizeof(__pyx_k_between), 0, 0, 1, 1},
  {&__pyx_n_s_charge, __pyx_k_charge, sizeof(__pyx_k_charge), 0, 0, 1, 1},
  {&__pyx_n_s_charge_carrier, __pyx_k_charge_carrier, sizeof(__pyx_k_charge_carrier), 0, 0, 1, 1},
  {&__pyx_n_s_charge_range, __pyx_k_charge_range, sizeof(__pyx_k_charge_range), 0, 0, 1, 1},
  {&__pyx_n_s_check_fit, __pyx_k_check_fit, sizeof(__pyx_k_check_fit), 0, 0, 1, 1},
  {&__pyx_n_s_class, __pyx_k_class, sizeof(__pyx_k_class), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_kp_s_d_is_out_of_bounds_for_peak_lis, __pyx_k_d_is_out_of_bounds_for_peak_lis, sizeof(__pyx_k_d_is_out_of_bounds_for_peak_lis), 0, 0, 1, 0},
  {&__pyx_n_s_defaults, __pyx_k_defaults, sizeof(__pyx_k_defaults), 0, 0, 1, 1},
  {&__pyx_n_s_dict, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
  {&__pyx_n_s_dtype, __pyx_k_dtype, sizeof(__pyx_k_dtype), 0, 0, 1, 1},
  {&__pyx_n_s_empty, __pyx_k_empty, sizeof(__pyx_k_empty), 0, 0, 1, 1},
  {&__pyx_n_s_envelope, __pyx_k_envelope, sizeof(__pyx_k_envelope), 0, 0, 1, 1},
  {&__pyx_n_s_error_tolerance, __pyx_k_error_tolerance, sizeof(__pyx_k_error_tolerance), 0, 0, 1, 1},
  {&__pyx_n_s_evaluate_theoretical_distributi, __pyx_k_evaluate_theoretical_distributi, sizeof(__pyx_k_evaluate_theoretical_distributi), 0, 0, 1, 1},
  {&__pyx_n_s_experimental, __pyx_k_experimental, sizeof(__pyx_k_experimental), 0, 0, 1, 1},
  {&__pyx_n_s_experimental_distribution, __pyx_k_experimental_distribution, sizeof(__pyx_k_experimental_distribution), 0, 0, 1, 1},
  {&__pyx_n_s_explore_local, __pyx_k_explore_local, sizeof(__pyx_k_explore_local), 0, 0, 1, 1},
  {&__pyx_n_s_find_next_putative_peak, __pyx_k_find_next_putative_peak, sizeof(__pyx_k_find_next_putative_peak), 0, 0, 1, 1},
  {&__pyx_n_s_find_previous_putative_peak, __pyx_k_find_previous_putative_peak, sizeof(__pyx_k_find_previous_putative_peak), 0, 0, 1, 1},
  {&__pyx_n_s_fit, __pyx_k_fit, sizeof(__pyx_k_fit), 0, 0, 1, 1},
  {&__pyx_n_s_fit_all_charge_states, __pyx_k_fit_all_charge_states, sizeof(__pyx_k_fit_all_charge_states), 0, 0, 1, 1},
  {&__pyx_n_s_fit_incremental_truncation, __pyx_k_fit_incremental_truncation, sizeof(__pyx_k_fit_incremental_truncation), 0, 0, 1, 1},
  {&__pyx_n_s_fit_peaks_at_charges, __pyx_k_fit_peaks_at_charges, sizeof(__pyx_k_fit_peaks_at_charges), 0, 0, 1, 1},
  {&__pyx_n_s_fit_theoretical_distribution, __pyx_k_fit_theoretical_distribution, sizeof(__pyx_k_fit_theoretical_distribution), 0, 0, 1, 1},
  {&__pyx_n_s_full_width_at_half_max, __pyx_k_full_width_at_half_max, sizeof(__pyx_k_full_width_at_half_max), 0, 0, 1, 1},
  {&__pyx_n_s_get_all_peak_charge_pairs, __pyx_k_get_all_peak_charge_pairs, sizeof(__pyx_k_get_all_peak_charge_pairs), 0, 0, 1, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_n_s_has_peak, __pyx_k_has_peak, sizeof(__pyx_k_has_peak), 0, 0, 1, 1},
  {&__pyx_n_s_hi, __pyx_k_hi, sizeof(__pyx_k_hi), 0, 0, 1, 1},
  {&__pyx_n_s_ignore_below, __pyx_k_ignore_below, sizeof(__pyx_k_ignore_below), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_index, __pyx_k_index, sizeof(__pyx_k_index), 0, 0, 1, 1},
  {&__pyx_n_s_init, __pyx_k_init, sizeof(__pyx_k_init), 0, 0, 1, 1},
  {&__pyx_n_s_int32, __pyx_k_int32, sizeof(__pyx_k_int32), 0, 0, 1, 1},
  {&__pyx_n_s_intensity, __pyx_k_intensity, sizeof(__pyx_k_intensity), 0, 0, 1, 1},
  {&__pyx_n_s_isotopic_cluster, __pyx_k_isotopic_cluster, sizeof(__pyx_k_isotopic_cluster), 0, 0, 1, 1},
  {&__pyx_n_s_key, __pyx_k_key, sizeof(__pyx_k_key), 0, 0, 1, 1},
  {&__pyx_n_s_kwargs, __pyx_k_kwargs, sizeof(__pyx_k_kwargs), 0, 0, 1, 1},
  {&__pyx_n_s_left_search_limit, __pyx_k_left_search_limit, sizeof(__pyx_k_left_search_limit), 0, 0, 1, 1},
  {&__pyx_n_s_lo, __pyx_k_lo, sizeof(__pyx_k_lo), 0, 0, 1, 1},
  {&__pyx_n_s_lower_bound, __pyx_k_lower_bound, sizeof(__pyx_k_lower_bound), 0, 0, 1, 1},
  {&__pyx_n_s_m1, __pyx_k_m1, sizeof(__pyx_k_m1), 0, 0, 1, 1},
  {&__pyx_n_s_m2, __pyx_k_m2, sizeof(__pyx_k_m2), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_make_deconvoluted_peak, __pyx_k_make_deconvoluted_peak, sizeof(__pyx_k_make_deconvoluted_peak), 0, 0, 1, 1},
  {&__pyx_n_s_match_theoretical_isotopic_distr, __pyx_k_match_theoretical_isotopic_distr, sizeof(__pyx_k_match_theoretical_isotopic_distr), 0, 0, 1, 1},
  {&__pyx_n_s_max_charge, __pyx_k_max_charge, sizeof(__pyx_k_max_charge), 0, 0, 1, 1},
  {&__pyx_n_s_merge_isobaric_peaks, __pyx_k_merge_isobaric_peaks, sizeof(__pyx_k_merge_isobaric_peaks), 0, 0, 1, 1},
  {&__pyx_n_s_min_charge, __pyx_k_min_charge, sizeof(__pyx_k_min_charge), 0, 0, 1, 1},
  {&__pyx_n_s_minimum_intensity, __pyx_k_minimum_intensity, sizeof(__pyx_k_minimum_intensity), 0, 0, 1, 1},
  {&__pyx_n_s_monoisotopic_mz, __pyx_k_monoisotopic_mz, sizeof(__pyx_k_monoisotopic_mz), 0, 0, 1, 1},
  {&__pyx_n_s_most_abundant_mass, __pyx_k_most_abundant_mass, sizeof(__pyx_k_most_abundant_mass), 0, 0, 1, 1},
  {&__pyx_kp_s_ms_deisotope__c_deconvoluter_bas, __pyx_k_ms_deisotope__c_deconvoluter_bas, sizeof(__pyx_k_ms_deisotope__c_deconvoluter_bas), 0, 0, 1, 0},
  {&__pyx_n_s_ms_deisotope__c_deconvoluter_bas_2, __pyx_k_ms_deisotope__c_deconvoluter_bas_2, sizeof(__pyx_k_ms_deisotope__c_deconvoluter_bas_2), 0, 0, 1, 1},
  {&__pyx_n_s_ms_deisotope_constants, __pyx_k_ms_deisotope_constants, sizeof(__pyx_k_ms_deisotope_constants), 0, 0, 1, 1},
  {&__pyx_kp_s_ms_peak_picker__c_peak_index, __pyx_k_ms_peak_picker__c_peak_index, sizeof(__pyx_k_ms_peak_picker__c_peak_index), 0, 0, 1, 0},
  {&__pyx_kp_s_ms_peak_picker__c_peak_set, __pyx_k_ms_peak_picker__c_peak_set, sizeof(__pyx_k_ms_peak_picker__c_peak_set), 0, 0, 1, 0},
  {&__pyx_n_s_mz, __pyx_k_mz, sizeof(__pyx_k_mz), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_kp_u_ndarray_is_not_C_contiguous, __pyx_k_ndarray_is_not_C_contiguous, sizeof(__pyx_k_ndarray_is_not_C_contiguous), 0, 1, 0, 0},
  {&__pyx_kp_u_ndarray_is_not_Fortran_contiguou, __pyx_k_ndarray_is_not_Fortran_contiguou, sizeof(__pyx_k_ndarray_is_not_Fortran_contiguou), 0, 1, 0, 0},
  {&__pyx_n_s_neutral_mass, __pyx_k_neutral_mass, sizeof(__pyx_k_neutral_mass), 0, 0, 1, 1},
  {&__pyx_n_s_new, __pyx_k_new, sizeof(__pyx_k_new), 0, 0, 1, 1},
  {&__pyx_n_s_np, __pyx_k_np, sizeof(__pyx_k_np), 0, 0, 1, 1},
  {&__pyx_n_s_numpy, __pyx_k_numpy, sizeof(__pyx_k_numpy), 0, 0, 1, 1},
  {&__pyx_kp_s_numpy_core_multiarray_failed_to, __pyx_k_numpy_core_multiarray_failed_to, sizeof(__pyx_k_numpy_core_multiarray_failed_to), 0, 0, 1, 0},
  {&__pyx_kp_s_numpy_core_umath_failed_to_impor, __pyx_k_numpy_core_umath_failed_to_impor, sizeof(__pyx_k_numpy_core_umath_failed_to_impor), 0, 0, 1, 0},
  {&__pyx_n_s_operator, __pyx_k_operator, sizeof(__pyx_k_operator), 0, 0, 1, 1},
  {&__pyx_n_s_peak, __pyx_k_peak, sizeof(__pyx_k_peak), 0, 0, 1, 1},
  {&__pyx_n_s_peak_charge_set, __pyx_k_peak_charge_set, sizeof(__pyx_k_peak_charge_set), 0, 0, 1, 1},
  {&__pyx_n_s_peak_dependency_network, __pyx_k_peak_dependency_network, sizeof(__pyx_k_peak_dependency_network), 0, 0, 1, 1},
  {&__pyx_n_s_peak_set, __pyx_k_peak_set, sizeof(__pyx_k_peak_set), 0, 0, 1, 1},
  {&__pyx_kp_s_peak_set_cannot_be_empty, __pyx_k_peak_set_cannot_be_empty, sizeof(__pyx_k_peak_set_cannot_be_empty), 0, 0, 1, 0},
  {&__pyx_n_s_peaks, __pyx_k_peaks, sizeof(__pyx_k_peaks), 0, 0, 1, 1},
  {&__pyx_n_s_pickle, __pyx_k_pickle, sizeof(__pyx_k_pickle), 0, 0, 1, 1},
  {&__pyx_n_s_populate_graph, __pyx_k_populate_graph, sizeof(__pyx_k_populate_graph), 0, 0, 1, 1},
  {&__pyx_n_s_position_spec, __pyx_k_position_spec, sizeof(__pyx_k_position_spec), 0, 0, 1, 1},
  {&__pyx_n_s_priority_map, __pyx_k_priority_map, sizeof(__pyx_k_priority_map), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_PickleError, __pyx_k_pyx_PickleError, sizeof(__pyx_k_pyx_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_checksum, __pyx_k_pyx_checksum, sizeof(__pyx_k_pyx_checksum), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_fuse_0_0quick_charge, __pyx_k_pyx_fuse_0_0quick_charge, sizeof(__pyx_k_pyx_fuse_0_0quick_charge), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_fuse_0_1quick_charge, __pyx_k_pyx_fuse_0_1quick_charge, sizeof(__pyx_k_pyx_fuse_0_1quick_charge), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_fuse_0_2quick_charge, __pyx_k_pyx_fuse_0_2quick_charge, sizeof(__pyx_k_pyx_fuse_0_2quick_charge), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_fuse_1_0quick_charge, __pyx_k_pyx_fuse_1_0quick_charge, sizeof(__pyx_k_pyx_fuse_1_0quick_charge), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_fuse_1_1quick_charge, __pyx_k_pyx_fuse_1_1quick_charge, sizeof(__pyx_k_pyx_fuse_1_1quick_charge), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_fuse_1_2quick_charge, __pyx_k_pyx_fuse_1_2quick_charge, sizeof(__pyx_k_pyx_fuse_1_2quick_charge), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_result, __pyx_k_pyx_result, sizeof(__pyx_k_pyx_result), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_state, __pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_type, __pyx_k_pyx_type, sizeof(__pyx_k_pyx_type), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_AveragineDeconvol, __pyx_k_pyx_unpickle_AveragineDeconvol, sizeof(__pyx_k_pyx_unpickle_AveragineDeconvol), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_DeconvoluterBase, __pyx_k_pyx_unpickle_DeconvoluterBase, sizeof(__pyx_k_pyx_unpickle_DeconvoluterBase), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_MultiAveragineDec, __pyx_k_pyx_unpickle_MultiAveragineDec, sizeof(__pyx_k_pyx_unpickle_MultiAveragineDec), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_quick_charge, __pyx_k_quick_charge, sizeof(__pyx_k_quick_charge), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_recalculate_starting_peak, __pyx_k_recalculate_starting_peak, sizeof(__pyx_k_recalculate_starting_peak), 0, 0, 1, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {&__pyx_n_s_right_search_limit, __pyx_k_right_search_limit, sizeof(__pyx_k_right_search_limit), 0, 0, 1, 1},
  {&__pyx_n_s_s, __pyx_k_s, sizeof(__pyx_k_s), 0, 0, 1, 1},
  {&__pyx_kp_s_s_peaklist_s_scorer_s, __pyx_k_s_peaklist_s_scorer_s, sizeof(__pyx_k_s_peaklist_s_scorer_s), 0, 0, 1, 0},
  {&__pyx_n_s_scale_method, __pyx_k_scale_method, sizeof(__pyx_k_scale_method), 0, 0, 1, 1},
  {&__pyx_n_s_scale_theoretical_distribution, __pyx_k_scale_theoretical_distribution, sizeof(__pyx_k_scale_theoretical_distribution), 0, 0, 1, 1},
  {&__pyx_n_s_score, __pyx_k_score, sizeof(__pyx_k_score), 0, 0, 1, 1},
  {&__pyx_n_s_seed_fit, __pyx_k_seed_fit, sizeof(__pyx_k_seed_fit), 0, 0, 1, 1},
  {&__pyx_n_s_self, __pyx_k_self, sizeof(__pyx_k_self), 0, 0, 1, 1},
  {&__pyx_kp_s_self_values_cannot_be_converted, __pyx_k_self_values_cannot_be_converted, sizeof(__pyx_k_self_values_cannot_be_converted), 0, 0, 1, 0},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_signal_to_noise, __pyx_k_signal_to_noise, sizeof(__pyx_k_signal_to_noise), 0, 0, 1, 1},
  {&__pyx_n_s_signatures, __pyx_k_signatures, sizeof(__pyx_k_signatures), 0, 0, 1, 1},
  {&__pyx_n_s_size_t, __pyx_k_size_t, sizeof(__pyx_k_size_t), 0, 0, 1, 1},
  {&__pyx_n_s_sorted, __pyx_k_sorted, sizeof(__pyx_k_sorted), 0, 0, 1, 1},
  {&__pyx_n_s_split, __pyx_k_split, sizeof(__pyx_k_split), 0, 0, 1, 1},
  {&__pyx_n_s_step, __pyx_k_step, sizeof(__pyx_k_step), 0, 0, 1, 1},
  {&__pyx_kp_s_stringsource, __pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 0, 1, 0},
  {&__pyx_n_s_strip, __pyx_k_strip, sizeof(__pyx_k_strip), 0, 0, 1, 1},
  {&__pyx_n_s_subtraction, __pyx_k_subtraction, sizeof(__pyx_k_subtraction), 0, 0, 1, 1},
  {&__pyx_n_s_sum, __pyx_k_sum, sizeof(__pyx_k_sum), 0, 0, 1, 1},
  {&__pyx_n_s_super, __pyx_k_super, sizeof(__pyx_k_super), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_theoretical, __pyx_k_theoretical, sizeof(__pyx_k_theoretical), 0, 0, 1, 1},
  {&__pyx_n_s_theoretical_distribution, __pyx_k_theoretical_distribution, sizeof(__pyx_k_theoretical_distribution), 0, 0, 1, 1},
  {&__pyx_n_s_tolerance, __pyx_k_tolerance, sizeof(__pyx_k_tolerance), 0, 0, 1, 1},
  {&__pyx_n_s_truncate_after, __pyx_k_truncate_after, sizeof(__pyx_k_truncate_after), 0, 0, 1, 1},
  {&__pyx_kp_u_unknown_dtype_code_in_numpy_pxd, __pyx_k_unknown_dtype_code_in_numpy_pxd, sizeof(__pyx_k_unknown_dtype_code_in_numpy_pxd), 0, 1, 0, 0},
  {&__pyx_n_s_update, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
  {&__pyx_n_s_use_quick_charge, __pyx_k_use_quick_charge, sizeof(__pyx_k_use_quick_charge), 0, 0, 1, 1},
  {&__pyx_n_s_use_subtraction, __pyx_k_use_subtraction, sizeof(__pyx_k_use_subtraction), 0, 0, 1, 1},
  {&__pyx_n_s_zeros, __pyx_k_zeros, sizeof(__pyx_k_zeros), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 50, __pyx_L1_error)
  __pyx_builtin_sorted = __Pyx_GetBuiltinName(__pyx_n_s_sorted); if (!__pyx_builtin_sorted) __PYX_ERR(0, 286, __pyx_L1_error)
  __pyx_builtin_super = __Pyx_GetBuiltinName(__pyx_n_s_super); if (!__pyx_builtin_super) __PYX_ERR(0, 447, __pyx_L1_error)
  __pyx_builtin_StopIteration = __Pyx_GetBuiltinName(__pyx_n_s_StopIteration); if (!__pyx_builtin_StopIteration) __PYX_ERR(0, 723, __pyx_L1_error)
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(1, 2, __pyx_L1_error)
  __pyx_builtin___import__ = __Pyx_GetBuiltinName(__pyx_n_s_import); if (!__pyx_builtin___import__) __PYX_ERR(0, 730, __pyx_L1_error)
  __pyx_builtin_IndexError = __Pyx_GetBuiltinName(__pyx_n_s_IndexError); if (!__pyx_builtin_IndexError) __PYX_ERR(0, 780, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(3, 272, __pyx_L1_error)
  __pyx_builtin_RuntimeError = __Pyx_GetBuiltinName(__pyx_n_s_RuntimeError); if (!__pyx_builtin_RuntimeError) __PYX_ERR(3, 856, __pyx_L1_error)
  __pyx_builtin_ImportError = __Pyx_GetBuiltinName(__pyx_n_s_ImportError); if (!__pyx_builtin_ImportError) __PYX_ERR(3, 1038, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":291
 *         current_peak = peak_list[0]
 *         merged_peaks = []
 *         for peak in peak_list[1:]:             # <<<<<<<<<<<<<<
 *             if abs(current_peak.neutral_mass - peak.neutral_mass) < 1e-3 and current_peak.charge == peak.charge:
 *                 current_peak.intensity += peak.intensity
 */
  __pyx_slice_ = PySlice_New(__pyx_int_1, Py_None, Py_None); if (unlikely(!__pyx_slice_)) __PYX_ERR(0, 291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice_);
  __Pyx_GIVEREF(__pyx_slice_);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("self.values cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.values cannot be converted to a Python object for pickling")
 */
  __pyx_tuple__6 = PyTuple_Pack(1, __pyx_kp_s_self_values_cannot_be_converted); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);

  /* "(tree fragment)":4
 *     raise TypeError("self.values cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.values cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__7 = PyTuple_Pack(1, __pyx_kp_s_self_values_cannot_be_converted); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":730
 * 
 * @cython.cdivision
 * cpdef np.ndarray[np.int32_t, ndim=1] quick_charge(FittedPeakCollection peak_set, FittedPeakOrPosition position_spec, int min_charge, int max_charge):             # <<<<<<<<<<<<<<
 *     """An implementation of Hoopman's QuickCharge [1] algorithm for quickly capping charge
 *     state queries
 */
  __pyx_tuple__11 = PyTuple_Pack(1, __pyx_kp_s_No_matching_signature_found); if (unlikely(!__pyx_tuple__11)) __PYX_ERR(0, 730, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__11);
  __Pyx_GIVEREF(__pyx_tuple__11);
  __pyx_tuple__12 = PyTuple_Pack(1, __pyx_kp_s_Function_call_with_ambiguous_arg); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(0, 730, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":780
 *     if index > peaks.get_size():
 *         if peaks.get_size() == 0:
 *             raise IndexError("peak_set cannot be empty!")             # <<<<<<<<<<<<<<
 *         # Assume we have a slice of the peak list, not the actual complete list, so
 *         # adjust the index offset from the first peak in the given list.
 */
  __pyx_tuple__13 = PyTuple_Pack(1, __pyx_kp_s_peak_set_cannot_be_empty); if (unlikely(!__pyx_tuple__13)) __PYX_ERR(0, 780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__13);
  __Pyx_GIVEREF(__pyx_tuple__13);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":810
 *         charges[charge] = 1
 *     if result_size == 0:
 *         result = np.empty(0, dtype=np.int32)             # <<<<<<<<<<<<<<
 *         return result
 *     for j in range(index - 1, -1, -1):
 */
  __pyx_tuple__14 = PyTuple_Pack(1, __pyx_int_0); if (unlikely(!__pyx_tuple__14)) __PYX_ERR(0, 810, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_tuple__14);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":840
 * @cython.binding(True)
 * cpdef set _get_all_peak_charge_pairs(DeconvoluterBase self, FittedPeak peak, double error_tolerance=ERROR_TOLERANCE,
 *                                      object charge_range=(1, 8), int left_search_limit=3, int right_search_limit=3,             # <<<<<<<<<<<<<<
 *                                      bint recalculate_starting_peak=True, bint use_quick_charge=False):
 *         """Construct the set of all unique candidate (monoisotopic peak, charge state) pairs using
 */
  __pyx_tuple__16 = PyTuple_Pack(2, __pyx_int_1, __pyx_int_8); if (unlikely(!__pyx_tuple__16)) __PYX_ERR(0, 840, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__16);
  __Pyx_GIVEREF(__pyx_tuple__16);

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":272
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")             # <<<<<<<<<<<<<<
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 */
  __pyx_tuple__25 = PyTuple_Pack(1, __pyx_kp_u_ndarray_is_not_C_contiguous); if (unlikely(!__pyx_tuple__25)) __PYX_ERR(3, 272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__25);
  __Pyx_GIVEREF(__pyx_tuple__25);

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":276
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")             # <<<<<<<<<<<<<<
 * 
 *             info.buf = PyArray_DATA(self)
 */
  __pyx_tuple__26 = PyTuple_Pack(1, __pyx_kp_u_ndarray_is_not_Fortran_contiguou); if (unlikely(!__pyx_tuple__26)) __PYX_ERR(3, 276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__26);
  __Pyx_GIVEREF(__pyx_tuple__26);

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":306
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"
 */
  __pyx_tuple__27 = PyTuple_Pack(1, __pyx_kp_u_Non_native_byte_order_not_suppor); if (unlikely(!__pyx_tuple__27)) __PYX_ERR(3, 306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__27);
  __Pyx_GIVEREF(__pyx_tuple__27);

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":856
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")             # <<<<<<<<<<<<<<
 * 
 *         if ((child.byteorder == c'>' and little_endian) or
 */
  __pyx_tuple__28 = PyTuple_Pack(1, __pyx_kp_u_Format_string_allocated_too_shor); if (unlikely(!__pyx_tuple__28)) __PYX_ERR(3, 856, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__28);
  __Pyx_GIVEREF(__pyx_tuple__28);

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":880
 *             t = child.type_num
 *             if end - f < 5:
 *                 raise RuntimeError(u"Format string allocated too short.")             # <<<<<<<<<<<<<<
 * 
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 */
  __pyx_tuple__29 = PyTuple_Pack(1, __pyx_kp_u_Format_string_allocated_too_shor_2); if (unlikely(!__pyx_tuple__29)) __PYX_ERR(3, 880, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__29);
  __Pyx_GIVEREF(__pyx_tuple__29);

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1038
 *         _import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
  __pyx_tuple__30 = PyTuple_Pack(1, __pyx_kp_s_numpy_core_multiarray_failed_to); if (unlikely(!__pyx_tuple__30)) __PYX_ERR(3, 1038, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__30);
  __Pyx_GIVEREF(__pyx_tuple__30);

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1044
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
  __pyx_tuple__31 = PyTuple_Pack(1, __pyx_kp_s_numpy_core_umath_failed_to_impor); if (unlikely(!__pyx_tuple__31)) __PYX_ERR(3, 1044, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__31);
  __Pyx_GIVEREF(__pyx_tuple__31);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":730
 * 
 * @cython.cdivision
 * cpdef np.ndarray[np.int32_t, ndim=1] quick_charge(FittedPeakCollection peak_set, FittedPeakOrPosition position_spec, int min_charge, int max_charge):             # <<<<<<<<<<<<<<
 *     """An implementation of Hoopman's QuickCharge [1] algorithm for quickly capping charge
 *     state queries
 */
  __pyx_tuple__32 = PyTuple_Pack(4, __pyx_n_s_peak_set, __pyx_n_s_position_spec, __pyx_n_s_min_charge, __pyx_n_s_max_charge); if (unlikely(!__pyx_tuple__32)) __PYX_ERR(0, 730, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__32);
  __Pyx_GIVEREF(__pyx_tuple__32);
  __pyx_codeobj__33 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__32, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_ms_deisotope__c_deconvoluter_bas, __pyx_n_s_pyx_fuse_0_0quick_charge, 730, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__33)) __PYX_ERR(0, 730, __pyx_L1_error)

  /* "ms_deisotope/_c/deconvoluter_base.pyx":839
 * 
 * @cython.binding(True)
 * cpdef set _get_all_peak_charge_pairs(DeconvoluterBase self, FittedPeak peak, double error_tolerance=ERROR_TOLERANCE,             # <<<<<<<<<<<<<<
 *                                      object charge_range=(1, 8), int left_search_limit=3, int right_search_limit=3,
 *                                      bint recalculate_starting_peak=True, bint use_quick_charge=False):
 */
  __pyx_tuple__34 = PyTuple_Pack(8, __pyx_n_s_self, __pyx_n_s_peak, __pyx_n_s_error_tolerance, __pyx_n_s_charge_range, __pyx_n_s_left_search_limit, __pyx_n_s_right_search_limit, __pyx_n_s_recalculate_starting_peak, __pyx_n_s_use_quick_charge); if (unlikely(!__pyx_tuple__34)) __PYX_ERR(0, 839, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__34);
  __Pyx_GIVEREF(__pyx_tuple__34);
  __pyx_codeobj__35 = (PyObject*)__Pyx_PyCode_New(8, 0, 8, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__34, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_ms_deisotope__c_deconvoluter_bas, __pyx_n_s_get_all_peak_charge_pairs, 839, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__35)) __PYX_ERR(0, 839, __pyx_L1_error)

  /* "ms_deisotope/_c/deconvoluter_base.pyx":929
 * 
 * @cython.binding(True)
 * cpdef int _explore_local(DeconvoluterBase self, peak, error_tolerance=ERROR_TOLERANCE,             # <<<<<<<<<<<<<<
 *                          charge_range=(1, 8), left_search_limit=1, right_search_limit=0,
 *                          charge_carrier=PROTON, truncate_after=TRUNCATE_AFTER,
 */
  __pyx_tuple__36 = PyTuple_Pack(9, __pyx_n_s_self, __pyx_n_s_peak, __pyx_n_s_error_tolerance, __pyx_n_s_charge_range, __pyx_n_s_left_search_limit, __pyx_n_s_right_search_limit, __pyx_n_s_charge_carrier, __pyx_n_s_truncate_after, __pyx_n_s_ignore_below); if (unlikely(!__pyx_tuple__36)) __PYX_ERR(0, 929, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__36);
  __Pyx_GIVEREF(__pyx_tuple__36);
  __pyx_codeobj__37 = (PyObject*)__Pyx_PyCode_New(9, 0, 9, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__36, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_ms_deisotope__c_deconvoluter_bas, __pyx_n_s_explore_local, 929, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__37)) __PYX_ERR(0, 929, __pyx_L1_error)

  /* "ms_deisotope/_c/deconvoluter_base.pyx":1009
 * 
 * @cython.binding(True)
 * cpdef populate_graph(DeconvoluterBase self, error_tolerance=ERROR_TOLERANCE, charge_range=(1, 8),             # <<<<<<<<<<<<<<
 *                      left_search_limit=1, right_search_limit=0, charge_carrier=PROTON,
 *                      truncate_after=TRUNCATE_AFTER, ignore_below=IGNORE_BELOW):
 */
  __pyx_tuple__38 = PyTuple_Pack(8, __pyx_n_s_self, __pyx_n_s_error_tolerance, __pyx_n_s_charge_range, __pyx_n_s_left_search_limit, __pyx_n_s_right_search_limit, __pyx_n_s_charge_carrier, __pyx_n_s_truncate_after, __pyx_n_s_ignore_below); if (unlikely(!__pyx_tuple__38)) __PYX_ERR(0, 1009, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__38);
  __Pyx_GIVEREF(__pyx_tuple__38);
  __pyx_codeobj__39 = (PyObject*)__Pyx_PyCode_New(8, 0, 8, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__38, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_ms_deisotope__c_deconvoluter_bas, __pyx_n_s_populate_graph, 1009, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__39)) __PYX_ERR(0, 1009, __pyx_L1_error)

  /* "ms_deisotope/_c/deconvoluter_base.pyx":1052
 * 
 * @cython.binding(True)
 * cpdef DeconvolutedPeak _make_deconvoluted_peak(self, IsotopicFitRecord fit, double charge_carrier):             # <<<<<<<<<<<<<<
 *     '''Helper method to package a finished :class:`~.IsotopicFitRecord` into a :class:`~.DeconvolutedPeak`
 * 
 */
  __pyx_tuple__40 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_fit, __pyx_n_s_charge_carrier); if (unlikely(!__pyx_tuple__40)) __PYX_ERR(0, 1052, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__40);
  __Pyx_GIVEREF(__pyx_tuple__40);
  __pyx_codeobj__41 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__40, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_ms_deisotope__c_deconvoluter_bas, __pyx_n_s_make_deconvoluted_peak, 1052, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__41)) __PYX_ERR(0, 1052, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __pyx_unpickle_DeconvoluterBase(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_tuple__42 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__42)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__42);
  __Pyx_GIVEREF(__pyx_tuple__42);
  __pyx_codeobj__43 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__42, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_DeconvoluterBase, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__43)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_tuple__44 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__44)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__44);
  __Pyx_GIVEREF(__pyx_tuple__44);
  __pyx_codeobj__45 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__44, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_AveragineDeconvol, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__45)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_tuple__46 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__46)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__46);
  __Pyx_GIVEREF(__pyx_tuple__46);
  __pyx_codeobj__47 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__46, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_MultiAveragineDec, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__47)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_float_5_ = PyFloat_FromDouble(5.); if (unlikely(!__pyx_float_5_)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_1eneg_3 = PyFloat_FromDouble(1e-3); if (unlikely(!__pyx_float_1eneg_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_4 = PyInt_FromLong(4); if (unlikely(!__pyx_int_4)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_8 = PyInt_FromLong(8); if (unlikely(!__pyx_int_8)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_24102755 = PyInt_FromLong(24102755L); if (unlikely(!__pyx_int_24102755)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_165289540 = PyInt_FromLong(165289540L); if (unlikely(!__pyx_int_165289540)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_266290880 = PyInt_FromLong(266290880L); if (unlikely(!__pyx_int_266290880)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_neg_1 = PyInt_FromLong(-1); if (unlikely(!__pyx_int_neg_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  if (__Pyx_ExportFunction("has_multiple_real_peaks", (void (*)(void))__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_has_multiple_real_peaks, "int (PyObject *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("_get_all_peak_charge_pairs", (void (*)(void))__pyx_f_12ms_deisotope_2_c_17deconvoluter_base__get_all_peak_charge_pairs, "PyObject *(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base__get_all_peak_charge_pairs *__pyx_optional_args)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("__pyx_fuse_0_0quick_charge", (void (*)(void))__pyx_fuse_0_0__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_quick_charge, "PyArrayObject *(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, int, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("__pyx_fuse_0_1quick_charge", (void (*)(void))__pyx_fuse_0_1__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_quick_charge, "PyArrayObject *(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, size_t, int, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("__pyx_fuse_0_2quick_charge", (void (*)(void))__pyx_fuse_0_2__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_quick_charge, "PyArrayObject *(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, long, int, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("__pyx_fuse_1_0quick_charge", (void (*)(void))__pyx_fuse_1_0__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_quick_charge, "PyArrayObject *(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, int, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("__pyx_fuse_1_1quick_charge", (void (*)(void))__pyx_fuse_1_1__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_quick_charge, "PyArrayObject *(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *, size_t, int, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("__pyx_fuse_1_2quick_charge", (void (*)(void))__pyx_fuse_1_2__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_quick_charge, "PyArrayObject *(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *, long, int, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase = &__pyx_vtable_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase;
  __pyx_vtable_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase.between = (struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, double, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_between;
  __pyx_vtable_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase.has_peak = (struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, double, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_has_peak;
  __pyx_vtable_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase._has_peak = (struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, double))__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__has_peak;
  __pyx_vtable_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase.match_theoretical_isotopic_distribution = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_match_theoretical_isotopic_distribution *__pyx_optional_args))__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_match_theoretical_isotopic_distribution;
  __pyx_vtable_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase.scale_theoretical_distribution = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, PyObject *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_scale_theoretical_distribution;
  __pyx_vtable_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase._evaluate_theoretical_distribution = (struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *(*)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, PyObject *, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, int, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__evaluate_theoretical_distribution;
  __pyx_vtable_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase.subtraction = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_subtraction *__pyx_optional_args))__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_subtraction;
  __pyx_vtable_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase.fit_incremental_truncation = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, double, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_fit_incremental_truncation;
  __pyx_vtable_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase._check_fit = (int (*)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__check_fit;
  __pyx_vtable_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase._find_next_putative_peak = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, int, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak *__pyx_optional_args))__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak;
  __pyx_vtable_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase._find_previous_putative_peak = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, int, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak *__pyx_optional_args))__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak;
  __pyx_vtable_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase._find_next_putative_peak_inplace = (int (*)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, int, PyObject *, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak_inplace *__pyx_optional_args))__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak_inplace;
  __pyx_vtable_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase._find_previous_putative_peak_inplace = (int (*)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, int, PyObject *, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak_inplace *__pyx_optional_args))__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak_inplace;
  if (PyType_Ready(&__pyx_type_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase) < 0) __PYX_ERR(0, 111, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase.tp_dictoffset && __pyx_type_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase.tp_dict, __pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase) < 0) __PYX_ERR(0, 111, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_DeconvoluterBase, (PyObject *)&__pyx_type_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase) < 0) __PYX_ERR(0, 111, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase) < 0) __PYX_ERR(0, 111, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase = &__pyx_type_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase;
  __pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase = &__pyx_vtable_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase;
  __pyx_vtable_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase.__pyx_base = *__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase;
  __pyx_vtable_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase.fit_theoretical_distribution = (struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *(*)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, double, int, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_fit_theoretical_distribution *__pyx_optional_args))__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_fit_theoretical_distribution;
  __pyx_vtable_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase._fit_peaks_at_charges = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *, PyObject *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase__fit_peaks_at_charges *__pyx_optional_args))__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase__fit_peaks_at_charges;
  __pyx_type_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase.tp_base = __pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase;
  if (PyType_Ready(&__pyx_type_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase) < 0) __PYX_ERR(0, 440, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase.tp_dictoffset && __pyx_type_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase.tp_dict, __pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase) < 0) __PYX_ERR(0, 440, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_AveragineDeconvoluterBase, (PyObject *)&__pyx_type_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase) < 0) __PYX_ERR(0, 440, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase) < 0) __PYX_ERR(0, 440, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase = &__pyx_type_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase;
  __pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase = &__pyx_vtable_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase;
  __pyx_vtable_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase.__pyx_base = *__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase;
  __pyx_vtable_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase.fit_theoretical_distribution = (struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *(*)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, double, int, struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_fit_theoretical_distribution *__pyx_optional_args))__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_fit_theoretical_distribution;
  __pyx_vtable_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase._fit_peaks_at_charges = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *, PyObject *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase__fit_peaks_at_charges *__pyx_optional_args))__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase__fit_peaks_at_charges;
  __pyx_type_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase.tp_base = __pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase;
  if (PyType_Ready(&__pyx_type_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase) < 0) __PYX_ERR(0, 493, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase.tp_dictoffset && __pyx_type_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase.tp_dict, __pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase) < 0) __PYX_ERR(0, 493, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_MultiAveragineDeconvoluterBase, (PyObject *)&__pyx_type_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase) < 0) __PYX_ERR(0, 493, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase) < 0) __PYX_ERR(0, 493, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase = &__pyx_type_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase;
  __pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator = &__pyx_vtable_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator;
  __pyx_vtable_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator._create = (struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *(*)(int, int))__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator__create;
  __pyx_vtable_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator.from_tuple = (struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *(*)(PyObject *))__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_from_tuple;
  __pyx_vtable_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator.from_quickcharge = (struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *(*)(PyObject *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *))__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_from_quickcharge;
  __pyx_vtable_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator.sequence_from_quickcharge = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_sequence_from_quickcharge;
  __pyx_vtable_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator.make_sequence = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_make_sequence;
  __pyx_vtable_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator.release_sequence = (void (*)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *))__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_release_sequence;
  __pyx_vtable_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator.set_bounds = (void (*)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *, int, int))__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_set_bounds;
  __pyx_vtable_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator.has_more = (int (*)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_has_more;
  __pyx_vtable_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator.reset = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_reset;
  __pyx_vtable_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator.get_next_value = (int (*)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_get_next_value;
  __pyx_vtable_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator.get_size = (size_t (*)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *))__pyx_f_12ms_deisotope_2_c_17deconvoluter_base_14ChargeIterator_get_size;
  if (PyType_Ready(&__pyx_type_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator) < 0) __PYX_ERR(0, 600, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator.tp_dictoffset && __pyx_type_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  if (__Pyx_SetVtable(__pyx_type_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator.tp_dict, __pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator) < 0) __PYX_ERR(0, 600, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_ChargeIterator, (PyObject *)&__pyx_type_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator) < 0) __PYX_ERR(0, 600, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator) < 0) __PYX_ERR(0, 600, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator = &__pyx_type_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule("ms_peak_picker._c.peak_set"); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase = __Pyx_ImportType(__pyx_t_1, "ms_peak_picker._c.peak_set", "PeakBase", sizeof(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase) __PYX_ERR(4, 2, __pyx_L1_error)
  __pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakBase = (struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakBase*)__Pyx_GetVtable(__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase->tp_dict); if (unlikely(!__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakBase)) __PYX_ERR(4, 2, __pyx_L1_error)
  __pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak = __Pyx_ImportType(__pyx_t_1, "ms_peak_picker._c.peak_set", "FittedPeak", sizeof(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak) __PYX_ERR(4, 12, __pyx_L1_error)
  __pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_FittedPeak = (struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_FittedPeak*)__Pyx_GetVtable(__pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak->tp_dict); if (unlikely(!__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_FittedPeak)) __PYX_ERR(4, 12, __pyx_L1_error)
  __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet = __Pyx_ImportType(__pyx_t_1, "ms_peak_picker._c.peak_set", "PeakSet", sizeof(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet) __PYX_ERR(4, 29, __pyx_L1_error)
  __pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSet = (struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet*)__Pyx_GetVtable(__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet->tp_dict); if (unlikely(!__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSet)) __PYX_ERR(4, 29, __pyx_L1_error)
  __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed = __Pyx_ImportType(__pyx_t_1, "ms_peak_picker._c.peak_set", "PeakSetIndexed", sizeof(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed) __PYX_ERR(4, 81, __pyx_L1_error)
  __pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed = (struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed*)__Pyx_GetVtable(__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed->tp_dict); if (unlikely(!__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed)) __PYX_ERR(4, 81, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(5, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4bool_bool = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "bool", sizeof(PyBoolObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4bool_bool) __PYX_ERR(6, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_7complex_complex = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "complex", sizeof(PyComplexObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_7complex_complex) __PYX_ERR(7, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("numpy"); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5numpy_dtype = __Pyx_ImportType(__pyx_t_1, "numpy", "dtype", sizeof(PyArray_Descr), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_dtype) __PYX_ERR(3, 206, __pyx_L1_error)
  __pyx_ptype_5numpy_flatiter = __Pyx_ImportType(__pyx_t_1, "numpy", "flatiter", sizeof(PyArrayIterObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_flatiter) __PYX_ERR(3, 229, __pyx_L1_error)
  __pyx_ptype_5numpy_broadcast = __Pyx_ImportType(__pyx_t_1, "numpy", "broadcast", sizeof(PyArrayMultiIterObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_broadcast) __PYX_ERR(3, 233, __pyx_L1_error)
  __pyx_ptype_5numpy_ndarray = __Pyx_ImportType(__pyx_t_1, "numpy", "ndarray", sizeof(PyArrayObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_ndarray) __PYX_ERR(3, 242, __pyx_L1_error)
  __pyx_ptype_5numpy_ufunc = __Pyx_ImportType(__pyx_t_1, "numpy", "ufunc", sizeof(PyUFuncObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_ufunc) __PYX_ERR(3, 918, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_peak_picker._c.peak_index"); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_14ms_peak_picker_2_c_10peak_index_PeakIndex = __Pyx_ImportType(__pyx_t_1, "ms_peak_picker._c.peak_index", "PeakIndex", sizeof(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_14ms_peak_picker_2_c_10peak_index_PeakIndex) __PYX_ERR(8, 5, __pyx_L1_error)
  __pyx_vtabptr_14ms_peak_picker_2_c_10peak_index_PeakIndex = (struct __pyx_vtabstruct_14ms_peak_picker_2_c_10peak_index_PeakIndex*)__Pyx_GetVtable(__pyx_ptype_14ms_peak_picker_2_c_10peak_index_PeakIndex->tp_dict); if (unlikely(!__pyx_vtabptr_14ms_peak_picker_2_c_10peak_index_PeakIndex)) __PYX_ERR(8, 5, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("brainpy._c.composition"); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7brainpy_2_c_11composition_PyComposition = __Pyx_ImportType(__pyx_t_1, "brainpy._c.composition", "PyComposition", sizeof(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7brainpy_2_c_11composition_PyComposition) __PYX_ERR(9, 123, __pyx_L1_error)
  __pyx_vtabptr_7brainpy_2_c_11composition_PyComposition = (struct __pyx_vtabstruct_7brainpy_2_c_11composition_PyComposition*)__Pyx_GetVtable(__pyx_ptype_7brainpy_2_c_11composition_PyComposition->tp_dict); if (unlikely(!__pyx_vtabptr_7brainpy_2_c_11composition_PyComposition)) __PYX_ERR(9, 123, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("brainpy._c.isotopic_distribution"); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak = __Pyx_ImportType(__pyx_t_1, "brainpy._c.isotopic_distribution", "TheoreticalPeak", sizeof(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak) __PYX_ERR(10, 63, __pyx_L1_error)
  __pyx_vtabptr_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak = (struct __pyx_vtabstruct_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak*)__Pyx_GetVtable(__pyx_ptype_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak->tp_dict); if (unlikely(!__pyx_vtabptr_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak)) __PYX_ERR(10, 63, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_deisotope._c.averagine"); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_12ms_deisotope_2_c_9averagine_Averagine = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.averagine", "Averagine", sizeof(struct __pyx_obj_12ms_deisotope_2_c_9averagine_Averagine), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_9averagine_Averagine) __PYX_ERR(11, 12, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_9averagine_Averagine = (struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_Averagine*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_9averagine_Averagine->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_9averagine_Averagine)) __PYX_ERR(11, 12, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.averagine", "TheoreticalIsotopicPattern", sizeof(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern) __PYX_ERR(11, 23, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern = (struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern)) __PYX_ERR(11, 23, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_9averagine_AveragineCache = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.averagine", "AveragineCache", sizeof(struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_9averagine_AveragineCache) __PYX_ERR(11, 79, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_9averagine_AveragineCache = (struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_AveragineCache*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_9averagine_AveragineCache->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_9averagine_AveragineCache)) __PYX_ERR(11, 79, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_deisotope._c.scoring"); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitRecord = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "IsotopicFitRecord", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitRecord) __PYX_ERR(12, 16, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_IsotopicFitRecord = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitRecord*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitRecord->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_IsotopicFitRecord)) __PYX_ERR(12, 16, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_FitSelectorBase = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "FitSelectorBase", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_FitSelectorBase) __PYX_ERR(12, 38, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_FitSelectorBase = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FitSelectorBase*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_FitSelectorBase->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_FitSelectorBase)) __PYX_ERR(12, 38, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitterBase = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "IsotopicFitterBase", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitterBase) __PYX_ERR(12, 50, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_IsotopicFitterBase = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitterBase->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_IsotopicFitterBase)) __PYX_ERR(12, 50, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_LeastSquaresFitter = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "LeastSquaresFitter", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_LeastSquaresFitter), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_LeastSquaresFitter) __PYX_ERR(12, 62, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_LeastSquaresFitter = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_LeastSquaresFitter*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_LeastSquaresFitter->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_LeastSquaresFitter)) __PYX_ERR(12, 62, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_GTestFitter = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "GTestFitter", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_GTestFitter), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_GTestFitter) __PYX_ERR(12, 66, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_GTestFitter = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_GTestFitter*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_GTestFitter->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_GTestFitter)) __PYX_ERR(12, 66, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledGTestFitter = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "ScaledGTestFitter", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledGTestFitter), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledGTestFitter) __PYX_ERR(12, 69, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_ScaledGTestFitter = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_ScaledGTestFitter*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledGTestFitter->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_ScaledGTestFitter)) __PYX_ERR(12, 69, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_MSDeconVFitter = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "MSDeconVFitter", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_MSDeconVFitter) __PYX_ERR(12, 72, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_MSDeconVFitter = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_MSDeconVFitter*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_MSDeconVFitter->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_MSDeconVFitter)) __PYX_ERR(12, 72, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "PenalizedMSDeconVFitter", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter) __PYX_ERR(12, 78, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter)) __PYX_ERR(12, 78, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_FunctionScorer = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "FunctionScorer", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FunctionScorer), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_FunctionScorer) __PYX_ERR(12, 84, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_FunctionScorer = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FunctionScorer*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_FunctionScorer->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_FunctionScorer)) __PYX_ERR(12, 84, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_InterferenceDetection = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "InterferenceDetection", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_InterferenceDetection) __PYX_ERR(12, 89, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_InterferenceDetection = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_InterferenceDetection*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_InterferenceDetection->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_InterferenceDetection)) __PYX_ERR(12, 89, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_DistinctPatternFitter = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "DistinctPatternFitter", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_DistinctPatternFitter) __PYX_ERR(12, 95, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_DistinctPatternFitter = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_DistinctPatternFitter*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_DistinctPatternFitter->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_DistinctPatternFitter)) __PYX_ERR(12, 95, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "ScaledPenalizedMSDeconvFitter", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter) __PYX_ERR(12, 102, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter)) __PYX_ERR(12, 102, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_DotProductFitter = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "DotProductFitter", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_DotProductFitter), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_DotProductFitter) __PYX_ERR(12, 112, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_DotProductFitter = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_DotProductFitter*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_DotProductFitter->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_DotProductFitter)) __PYX_ERR(12, 112, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_deisotope._c.peak_set"); if (unlikely(!__pyx_t_1)) __PYX_ERR(13, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_12ms_deisotope_2_c_8peak_set__Index = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.peak_set", "_Index", sizeof(struct __pyx_obj_12ms_deisotope_2_c_8peak_set__Index), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_8peak_set__Index) __PYX_ERR(13, 5, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_8peak_set__Index = (struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set__Index*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_8peak_set__Index->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_8peak_set__Index)) __PYX_ERR(13, 5, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_8peak_set_EnvelopePair = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.peak_set", "EnvelopePair", sizeof(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_EnvelopePair), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_8peak_set_EnvelopePair) __PYX_ERR(13, 14, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_8peak_set_EnvelopePair = (struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_EnvelopePair*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_8peak_set_EnvelopePair->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_EnvelopePair)) __PYX_ERR(13, 14, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_8peak_set_Envelope = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.peak_set", "Envelope", sizeof(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_8peak_set_Envelope) __PYX_ERR(13, 25, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_8peak_set_Envelope = (struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_Envelope*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_8peak_set_Envelope->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_Envelope)) __PYX_ERR(13, 25, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.peak_set", "DeconvolutedPeak", sizeof(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak) __PYX_ERR(13, 40, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak = (struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak)) __PYX_ERR(13, 40, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.peak_set", "DeconvolutedPeakSolution", sizeof(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution) __PYX_ERR(13, 66, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution = (struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution)) __PYX_ERR(13, 66, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.peak_set", "DeconvolutedPeakSet", sizeof(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet) __PYX_ERR(13, 70, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet = (struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet)) __PYX_ERR(13, 70, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.peak_set", "DeconvolutedPeakSetIndexed", sizeof(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed) __PYX_ERR(13, 99, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed = (struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed)) __PYX_ERR(13, 99, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_deisotope._c.peak_dependency_network.intervals"); if (unlikely(!__pyx_t_1)) __PYX_ERR(14, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.peak_dependency_network.intervals", "SpanningMixin", sizeof(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin) __PYX_ERR(14, 1, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin = (struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin)) __PYX_ERR(14, 1, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.peak_dependency_network.intervals", "Interval", sizeof(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval) __PYX_ERR(14, 19, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval = (struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval)) __PYX_ERR(14, 19, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.peak_dependency_network.intervals", "IntervalTreeNode", sizeof(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode) __PYX_ERR(14, 25, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode = (struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode)) __PYX_ERR(14, 25, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_deisotope._c.peak_dependency_network.subgraph"); if (unlikely(!__pyx_t_1)) __PYX_ERR(15, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_8subgraph_FitNode = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.peak_dependency_network.subgraph", "FitNode", sizeof(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_8subgraph_FitNode), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_8subgraph_FitNode) __PYX_ERR(15, 8, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_8subgraph_FitNode = (struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_8subgraph_FitNode*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_8subgraph_FitNode->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_8subgraph_FitNode)) __PYX_ERR(15, 8, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_8subgraph_ConnectedSubgraph = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.peak_dependency_network.subgraph", "ConnectedSubgraph", sizeof(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_8subgraph_ConnectedSubgraph), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_8subgraph_ConnectedSubgraph) __PYX_ERR(15, 28, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_8subgraph_ConnectedSubgraph = (struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_8subgraph_ConnectedSubgraph*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_8subgraph_ConnectedSubgraph->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_8subgraph_ConnectedSubgraph)) __PYX_ERR(15, 28, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_deisotope._c.peak_dependency_network.peak_network"); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_12peak_network_PeakNode = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.peak_dependency_network.peak_network", "PeakNode", sizeof(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_12peak_network_PeakNode), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_12peak_network_PeakNode) __PYX_ERR(16, 8, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_12peak_network_PeakNode = (struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_12peak_network_PeakNode*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_12peak_network_PeakNode->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_12peak_network_PeakNode)) __PYX_ERR(16, 8, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_12peak_network_DependenceCluster = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.peak_dependency_network.peak_network", "DependenceCluster", sizeof(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_12peak_network_DependenceCluster), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_12peak_network_DependenceCluster) __PYX_ERR(16, 30, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_12peak_network_DependenceCluster = (struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_12peak_network_DependenceCluster*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_12peak_network_DependenceCluster->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_12peak_network_DependenceCluster)) __PYX_ERR(16, 30, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_12peak_network_PeakDependenceGraphBase = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.peak_dependency_network.peak_network", "PeakDependenceGraphBase", sizeof(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_12peak_network_PeakDependenceGraphBase), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_12peak_network_PeakDependenceGraphBase) __PYX_ERR(16, 63, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_12peak_network_PeakDependenceGraphBase = (struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_12peak_network_PeakDependenceGraphBase*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_12peak_network_PeakDependenceGraphBase->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_12peak_network_PeakDependenceGraphBase)) __PYX_ERR(16, 63, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __pyx_t_1 = PyImport_ImportModule("ms_peak_picker._c.peak_set"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "INTERVAL_INDEX_SIZE", (void **)&__pyx_vp_14ms_peak_picker_2_c_8peak_set_INTERVAL_INDEX_SIZE, "size_t") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_2 = PyImport_ImportModule("brainpy._c.composition"); if (!__pyx_t_2) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_2, "nist_mass", (void **)&__pyx_vp_7brainpy_2_c_11composition_nist_mass, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_2, "PROTON", (void **)&__pyx_vp_7brainpy_2_c_11composition_PROTON, "double") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_2, "_ElementTable", (void **)&__pyx_vp_7brainpy_2_c_11composition__ElementTable, "struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = PyImport_ImportModule("brainpy._c.isotopic_constants"); if (!__pyx_t_3) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_3, "DEFAULT_ISOTOPIC_CONSTANTS_SIZE", (void **)&__pyx_vp_7brainpy_2_c_18isotopic_constants_DEFAULT_ISOTOPIC_CONSTANTS_SIZE, "size_t") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = PyImport_ImportModule("ms_deisotope._c.averagine"); if (!__pyx_t_4) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_4, "PROTON", (void **)&__pyx_vp_12ms_deisotope_2_c_9averagine_PROTON, "double") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_5 = PyImport_ImportModule("ms_deisotope._c.scoring"); if (!__pyx_t_5) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_5, "INFINITY", (void **)&__pyx_vp_12ms_deisotope_2_c_7scoring_INFINITY, "double") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_6 = PyImport_ImportModule("ms_deisotope._c.peak_set"); if (!__pyx_t_6) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_6, "INTERVAL_INDEX_SIZE", (void **)&__pyx_vp_12ms_deisotope_2_c_8peak_set_INTERVAL_INDEX_SIZE, "size_t") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __pyx_t_1 = PyImport_ImportModule("ms_deisotope._c.averagine"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "neutral_mass", (void (**)(void))&__pyx_f_12ms_deisotope_2_c_9averagine_neutral_mass, "double (double, int, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_neutral_mass *__pyx_optional_args)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "isotopic_shift", (void (**)(void))&__pyx_f_12ms_deisotope_2_c_9averagine_isotopic_shift, "double (int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_isotopic_shift *__pyx_optional_args)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}


#if PY_MAJOR_VERSION < 3
#ifdef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC void
#else
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#endif
#else
#ifdef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initdeconvoluter_base(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initdeconvoluter_base(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_deconvoluter_base(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_deconvoluter_base(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_deconvoluter_base(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  double __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'deconvoluter_base' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_deconvoluter_base(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("deconvoluter_base", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_ms_deisotope___c__deconvoluter_base) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "ms_deisotope._c.deconvoluter_base")) {
      if (unlikely(PyDict_SetItemString(modules, "ms_deisotope._c.deconvoluter_base", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) goto __pyx_L1_error;
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) goto __pyx_L1_error;
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  if (unlikely(__Pyx_modinit_function_export_code() != 0)) goto __pyx_L1_error;
  if (unlikely(__Pyx_modinit_type_init_code() != 0)) goto __pyx_L1_error;
  if (unlikely(__Pyx_modinit_type_import_code() != 0)) goto __pyx_L1_error;
  if (unlikely(__Pyx_modinit_variable_import_code() != 0)) goto __pyx_L1_error;
  if (unlikely(__Pyx_modinit_function_import_code() != 0)) goto __pyx_L1_error;
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "ms_deisotope/_c/deconvoluter_base.pyx":12
 * 
 * from ms_deisotope.constants import (
 *     ERROR_TOLERANCE as _ERROR_TOLERANCE,             # <<<<<<<<<<<<<<
 *     IGNORE_BELOW as _IGNORE_BELOW,
 *     TRUNCATE_AFTER as _TRUNCATE_AFTER)
 */
  __pyx_t_1 = PyList_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_ERROR_TOLERANCE);
  __Pyx_GIVEREF(__pyx_n_s_ERROR_TOLERANCE);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_ERROR_TOLERANCE);
  __Pyx_INCREF(__pyx_n_s_IGNORE_BELOW);
  __Pyx_GIVEREF(__pyx_n_s_IGNORE_BELOW);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_n_s_IGNORE_BELOW);
  __Pyx_INCREF(__pyx_n_s_TRUNCATE_AFTER);
  __Pyx_GIVEREF(__pyx_n_s_TRUNCATE_AFTER);
  PyList_SET_ITEM(__pyx_t_1, 2, __pyx_n_s_TRUNCATE_AFTER);

  /* "ms_deisotope/_c/deconvoluter_base.pyx":11
 * from brainpy._c.isotopic_distribution cimport TheoreticalPeak
 * 
 * from ms_deisotope.constants import (             # <<<<<<<<<<<<<<
 *     ERROR_TOLERANCE as _ERROR_TOLERANCE,
 *     IGNORE_BELOW as _IGNORE_BELOW,
 */
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_ms_deisotope_constants, __pyx_t_1, -1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_ERROR_TOLERANCE); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ERROR_TOLERANCE_2, __pyx_t_1) < 0) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_IGNORE_BELOW); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_IGNORE_BELOW_2, __pyx_t_1) < 0) __PYX_ERR(0, 13, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_TRUNCATE_AFTER); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_TRUNCATE_AFTER_2, __pyx_t_1) < 0) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":29
 * from cpython.set cimport PySet_Add
 * 
 * import numpy as np             # <<<<<<<<<<<<<<
 * cimport numpy as cnp
 * 
 */
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_numpy, 0, -1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_np, __pyx_t_2) < 0) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":32
 * cimport numpy as cnp
 * 
 * cnp.import_array()             # <<<<<<<<<<<<<<
 * 
 * import operator
 */
  __pyx_t_3 = __pyx_f_5numpy_import_array(); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 32, __pyx_L1_error)

  /* "ms_deisotope/_c/deconvoluter_base.pyx":34
 * cnp.import_array()
 * 
 * import operator             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_operator, 0, -1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_operator, __pyx_t_2) < 0) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":37
 * 
 * 
 * cdef double ERROR_TOLERANCE = _ERROR_TOLERANCE             # <<<<<<<<<<<<<<
 * cdef double IGNORE_BELOW = _IGNORE_BELOW
 * cdef double TRUNCATE_AFTER = _TRUNCATE_AFTER
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ERROR_TOLERANCE_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_4 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_12ms_deisotope_2_c_17deconvoluter_base_ERROR_TOLERANCE = __pyx_t_4;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":38
 * 
 * cdef double ERROR_TOLERANCE = _ERROR_TOLERANCE
 * cdef double IGNORE_BELOW = _IGNORE_BELOW             # <<<<<<<<<<<<<<
 * cdef double TRUNCATE_AFTER = _TRUNCATE_AFTER
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_IGNORE_BELOW_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_4 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 38, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_12ms_deisotope_2_c_17deconvoluter_base_IGNORE_BELOW = __pyx_t_4;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":39
 * cdef double ERROR_TOLERANCE = _ERROR_TOLERANCE
 * cdef double IGNORE_BELOW = _IGNORE_BELOW
 * cdef double TRUNCATE_AFTER = _TRUNCATE_AFTER             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_TRUNCATE_AFTER_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_4 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_12ms_deisotope_2_c_17deconvoluter_base_TRUNCATE_AFTER = __pyx_t_4;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":452
 * 
 *     cpdef IsotopicFitRecord fit_theoretical_distribution(self, FittedPeak peak, double error_tolerance, int charge,
 *                                                          double charge_carrier=PROTON, double truncate_after=0.95,             # <<<<<<<<<<<<<<
 *                                                          double ignore_below=0):
 *         cdef:
 */
  __pyx_k__2 = __pyx_v_12ms_deisotope_2_c_9averagine_PROTON;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":451
 *             minimum_intensity, *args, **kwargs)
 * 
 *     cpdef IsotopicFitRecord fit_theoretical_distribution(self, FittedPeak peak, double error_tolerance, int charge,             # <<<<<<<<<<<<<<
 *                                                          double charge_carrier=PROTON, double truncate_after=0.95,
 *                                                          double ignore_below=0):
 */
  __pyx_k__2 = __pyx_v_12ms_deisotope_2_c_9averagine_PROTON;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":464
 *         return self._evaluate_theoretical_distribution(eid, tid, peak, charge)
 * 
 *     cpdef set _fit_peaks_at_charges(self, set peak_charge_set, double error_tolerance, double charge_carrier=PROTON,             # <<<<<<<<<<<<<<
 *                                     double truncate_after=0.95, double ignore_below=0):
 *         cdef:
 */
  __pyx_k__3 = __pyx_v_12ms_deisotope_2_c_9averagine_PROTON;
  __pyx_k__3 = __pyx_v_12ms_deisotope_2_c_9averagine_PROTON;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":504
 * 
 *     cpdef IsotopicFitRecord fit_theoretical_distribution(self, FittedPeak peak, double error_tolerance, int charge,
 *                                                          AveragineCache  averagine, double charge_carrier=PROTON,             # <<<<<<<<<<<<<<
 *                                                          double truncate_after=0.95, double ignore_below=0):
 *         cdef:
 */
  __pyx_k__4 = __pyx_v_12ms_deisotope_2_c_9averagine_PROTON;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":503
 *             minimum_intensity, *args, **kwargs)
 * 
 *     cpdef IsotopicFitRecord fit_theoretical_distribution(self, FittedPeak peak, double error_tolerance, int charge,             # <<<<<<<<<<<<<<
 *                                                          AveragineCache  averagine, double charge_carrier=PROTON,
 *                                                          double truncate_after=0.95, double ignore_below=0):
 */
  __pyx_k__4 = __pyx_v_12ms_deisotope_2_c_9averagine_PROTON;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":516
 *         return self._evaluate_theoretical_distribution(eid, tid, peak, charge)
 * 
 *     cpdef set _fit_peaks_at_charges(self, set peak_charge_set, double error_tolerance, double charge_carrier=PROTON,             # <<<<<<<<<<<<<<
 *                                     double truncate_after=0.95, double ignore_below=0):
 *         cdef:
 */
  __pyx_k__5 = __pyx_v_12ms_deisotope_2_c_9averagine_PROTON;
  __pyx_k__5 = __pyx_v_12ms_deisotope_2_c_9averagine_PROTON;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":730
 * 
 * @cython.cdivision
 * cpdef np.ndarray[np.int32_t, ndim=1] quick_charge(FittedPeakCollection peak_set, FittedPeakOrPosition position_spec, int min_charge, int max_charge):             # <<<<<<<<<<<<<<
 *     """An implementation of Hoopman's QuickCharge [1] algorithm for quickly capping charge
 *     state queries
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 730, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __pyx_FusedFunction_NewEx(&__pyx_fuse_0_0__pyx_mdef_12ms_deisotope_2_c_17deconvoluter_base_17__pyx_fuse_0_0quick_charge, 0, __pyx_n_s_pyx_fuse_0_0quick_charge, NULL, __pyx_n_s_ms_deisotope__c_deconvoluter_bas_2, __pyx_d, ((PyObject *)__pyx_codeobj__33)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 730, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_1, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_2, __pyx_kp_s_PeakSet_FittedPeak, __pyx_t_1) < 0) __PYX_ERR(0, 730, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __pyx_FusedFunction_NewEx(&__pyx_fuse_0_1__pyx_mdef_12ms_deisotope_2_c_17deconvoluter_base_19__pyx_fuse_0_1quick_charge, 0, __pyx_n_s_pyx_fuse_0_1quick_charge, NULL, __pyx_n_s_ms_deisotope__c_deconvoluter_bas_2, __pyx_d, ((PyObject *)__pyx_codeobj__33)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 730, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_1, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_2, __pyx_kp_s_PeakSet_size_t, __pyx_t_1) < 0) __PYX_ERR(0, 730, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __pyx_FusedFunction_NewEx(&__pyx_fuse_0_2__pyx_mdef_12ms_deisotope_2_c_17deconvoluter_base_21__pyx_fuse_0_2quick_charge, 0, __pyx_n_s_pyx_fuse_0_2quick_charge, NULL, __pyx_n_s_ms_deisotope__c_deconvoluter_bas_2, __pyx_d, ((PyObject *)__pyx_codeobj__33)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 730, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_1, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_2, __pyx_kp_s_PeakSet_long, __pyx_t_1) < 0) __PYX_ERR(0, 730, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __pyx_FusedFunction_NewEx(&__pyx_fuse_1_0__pyx_mdef_12ms_deisotope_2_c_17deconvoluter_base_23__pyx_fuse_1_0quick_charge, 0, __pyx_n_s_pyx_fuse_1_0quick_charge, NULL, __pyx_n_s_ms_deisotope__c_deconvoluter_bas_2, __pyx_d, ((PyObject *)__pyx_codeobj__33)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 730, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_1, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_2, __pyx_kp_s_PeakIndex_FittedPeak, __pyx_t_1) < 0) __PYX_ERR(0, 730, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __pyx_FusedFunction_NewEx(&__pyx_fuse_1_1__pyx_mdef_12ms_deisotope_2_c_17deconvoluter_base_25__pyx_fuse_1_1quick_charge, 0, __pyx_n_s_pyx_fuse_1_1quick_charge, NULL, __pyx_n_s_ms_deisotope__c_deconvoluter_bas_2, __pyx_d, ((PyObject *)__pyx_codeobj__33)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 730, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_1, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_2, __pyx_kp_s_PeakIndex_size_t, __pyx_t_1) < 0) __PYX_ERR(0, 730, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __pyx_FusedFunction_NewEx(&__pyx_fuse_1_2__pyx_mdef_12ms_deisotope_2_c_17deconvoluter_base_27__pyx_fuse_1_2quick_charge, 0, __pyx_n_s_pyx_fuse_1_2quick_charge, NULL, __pyx_n_s_ms_deisotope__c_deconvoluter_bas_2, __pyx_d, ((PyObject *)__pyx_codeobj__33)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 730, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_1, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_2, __pyx_kp_s_PeakIndex_long, __pyx_t_1) < 0) __PYX_ERR(0, 730, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __pyx_FusedFunction_NewEx(&__pyx_mdef_12ms_deisotope_2_c_17deconvoluter_base_1quick_charge, 0, __pyx_n_s_quick_charge, NULL, __pyx_n_s_ms_deisotope__c_deconvoluter_bas_2, __pyx_d, ((PyObject *)__pyx_codeobj__33)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 730, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_1, __pyx_empty_tuple);
  ((__pyx_FusedFunctionObject *) __pyx_t_1)->__signatures__ = __pyx_t_2;
  __Pyx_GIVEREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_quick_charge, __pyx_t_1) < 0) __PYX_ERR(0, 730, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":839
 * 
 * @cython.binding(True)
 * cpdef set _get_all_peak_charge_pairs(DeconvoluterBase self, FittedPeak peak, double error_tolerance=ERROR_TOLERANCE,             # <<<<<<<<<<<<<<
 *                                      object charge_range=(1, 8), int left_search_limit=3, int right_search_limit=3,
 *                                      bint recalculate_starting_peak=True, bint use_quick_charge=False):
 */
  __pyx_k__15 = __pyx_v_12ms_deisotope_2_c_17deconvoluter_base_ERROR_TOLERANCE;
  __pyx_k__15 = __pyx_v_12ms_deisotope_2_c_17deconvoluter_base_ERROR_TOLERANCE;
  __pyx_t_5 = __Pyx_CyFunction_NewEx(&__pyx_mdef_12ms_deisotope_2_c_17deconvoluter_base_3_get_all_peak_charge_pairs, 0, __pyx_n_s_get_all_peak_charge_pairs, NULL, __pyx_n_s_ms_deisotope__c_deconvoluter_bas_2, __pyx_d, ((PyObject *)__pyx_codeobj__35)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 839, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_get_all_peak_charge_pairs, __pyx_t_5) < 0) __PYX_ERR(0, 839, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":929
 * 
 * @cython.binding(True)
 * cpdef int _explore_local(DeconvoluterBase self, peak, error_tolerance=ERROR_TOLERANCE,             # <<<<<<<<<<<<<<
 *                          charge_range=(1, 8), left_search_limit=1, right_search_limit=0,
 *                          charge_carrier=PROTON, truncate_after=TRUNCATE_AFTER,
 */
  __pyx_t_5 = PyFloat_FromDouble(__pyx_v_12ms_deisotope_2_c_17deconvoluter_base_ERROR_TOLERANCE); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 929, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_k__17 = __pyx_t_5;
  __Pyx_GIVEREF(__pyx_t_5);
  __pyx_t_5 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":931
 * cpdef int _explore_local(DeconvoluterBase self, peak, error_tolerance=ERROR_TOLERANCE,
 *                          charge_range=(1, 8), left_search_limit=1, right_search_limit=0,
 *                          charge_carrier=PROTON, truncate_after=TRUNCATE_AFTER,             # <<<<<<<<<<<<<<
 *                          ignore_below=IGNORE_BELOW):
 *     """Given a peak, explore the local neighborhood for candidate isotopic fits and add each
 */
  __pyx_t_5 = PyFloat_FromDouble(__pyx_v_12ms_deisotope_2_c_9averagine_PROTON); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 931, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_k__18 = __pyx_t_5;
  __Pyx_GIVEREF(__pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = PyFloat_FromDouble(__pyx_v_12ms_deisotope_2_c_17deconvoluter_base_TRUNCATE_AFTER); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 931, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_k__19 = __pyx_t_5;
  __Pyx_GIVEREF(__pyx_t_5);
  __pyx_t_5 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":932
 *                          charge_range=(1, 8), left_search_limit=1, right_search_limit=0,
 *                          charge_carrier=PROTON, truncate_after=TRUNCATE_AFTER,
 *                          ignore_below=IGNORE_BELOW):             # <<<<<<<<<<<<<<
 *     """Given a peak, explore the local neighborhood for candidate isotopic fits and add each
 *     fit above a threshold to the peak dependence graph.
 */
  __pyx_t_5 = PyFloat_FromDouble(__pyx_v_12ms_deisotope_2_c_17deconvoluter_base_IGNORE_BELOW); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 932, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_k__20 = __pyx_t_5;
  __Pyx_GIVEREF(__pyx_t_5);
  __pyx_t_5 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":929
 * 
 * @cython.binding(True)
 * cpdef int _explore_local(DeconvoluterBase self, peak, error_tolerance=ERROR_TOLERANCE,             # <<<<<<<<<<<<<<
 *                          charge_range=(1, 8), left_search_limit=1, right_search_limit=0,
 *                          charge_carrier=PROTON, truncate_after=TRUNCATE_AFTER,
 */
  __pyx_t_5 = PyFloat_FromDouble(__pyx_v_12ms_deisotope_2_c_17deconvoluter_base_ERROR_TOLERANCE); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 929, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_k__17 = __pyx_t_5;
  __Pyx_GIVEREF(__pyx_t_5);
  __pyx_t_5 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":931
 * cpdef int _explore_local(DeconvoluterBase self, peak, error_tolerance=ERROR_TOLERANCE,
 *                          charge_range=(1, 8), left_search_limit=1, right_search_limit=0,
 *                          charge_carrier=PROTON, truncate_after=TRUNCATE_AFTER,             # <<<<<<<<<<<<<<
 *                          ignore_below=IGNORE_BELOW):
 *     """Given a peak, explore the local neighborhood for candidate isotopic fits and add each
 */
  __pyx_t_5 = PyFloat_FromDouble(__pyx_v_12ms_deisotope_2_c_9averagine_PROTON); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 931, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_k__18 = __pyx_t_5;
  __Pyx_GIVEREF(__pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = PyFloat_FromDouble(__pyx_v_12ms_deisotope_2_c_17deconvoluter_base_TRUNCATE_AFTER); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 931, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_k__19 = __pyx_t_5;
  __Pyx_GIVEREF(__pyx_t_5);
  __pyx_t_5 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":932
 *                          charge_range=(1, 8), left_search_limit=1, right_search_limit=0,
 *                          charge_carrier=PROTON, truncate_after=TRUNCATE_AFTER,
 *                          ignore_below=IGNORE_BELOW):             # <<<<<<<<<<<<<<
 *     """Given a peak, explore the local neighborhood for candidate isotopic fits and add each
 *     fit above a threshold to the peak dependence graph.
 */
  __pyx_t_5 = PyFloat_FromDouble(__pyx_v_12ms_deisotope_2_c_17deconvoluter_base_IGNORE_BELOW); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 932, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_k__20 = __pyx_t_5;
  __Pyx_GIVEREF(__pyx_t_5);
  __pyx_t_5 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":929
 * 
 * @cython.binding(True)
 * cpdef int _explore_local(DeconvoluterBase self, peak, error_tolerance=ERROR_TOLERANCE,             # <<<<<<<<<<<<<<
 *                          charge_range=(1, 8), left_search_limit=1, right_search_limit=0,
 *                          charge_carrier=PROTON, truncate_after=TRUNCATE_AFTER,
 */
  __pyx_t_5 = __Pyx_CyFunction_NewEx(&__pyx_mdef_12ms_deisotope_2_c_17deconvoluter_base_5_explore_local, 0, __pyx_n_s_explore_local, NULL, __pyx_n_s_ms_deisotope__c_deconvoluter_bas_2, __pyx_d, ((PyObject *)__pyx_codeobj__37)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 929, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_explore_local, __pyx_t_5) < 0) __PYX_ERR(0, 929, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":1009
 * 
 * @cython.binding(True)
 * cpdef populate_graph(DeconvoluterBase self, error_tolerance=ERROR_TOLERANCE, charge_range=(1, 8),             # <<<<<<<<<<<<<<
 *                      left_search_limit=1, right_search_limit=0, charge_carrier=PROTON,
 *                      truncate_after=TRUNCATE_AFTER, ignore_below=IGNORE_BELOW):
 */
  __pyx_t_5 = PyFloat_FromDouble(__pyx_v_12ms_deisotope_2_c_17deconvoluter_base_ERROR_TOLERANCE); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1009, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_k__21 = __pyx_t_5;
  __Pyx_GIVEREF(__pyx_t_5);
  __pyx_t_5 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":1010
 * @cython.binding(True)
 * cpdef populate_graph(DeconvoluterBase self, error_tolerance=ERROR_TOLERANCE, charge_range=(1, 8),
 *                      left_search_limit=1, right_search_limit=0, charge_carrier=PROTON,             # <<<<<<<<<<<<<<
 *                      truncate_after=TRUNCATE_AFTER, ignore_below=IGNORE_BELOW):
 *     """Visit each experimental peak and execute :meth:`_explore_local` on it with the provided
 */
  __pyx_t_5 = PyFloat_FromDouble(__pyx_v_12ms_deisotope_2_c_9averagine_PROTON); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1010, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_k__22 = __pyx_t_5;
  __Pyx_GIVEREF(__pyx_t_5);
  __pyx_t_5 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":1011
 * cpdef populate_graph(DeconvoluterBase self, error_tolerance=ERROR_TOLERANCE, charge_range=(1, 8),
 *                      left_search_limit=1, right_search_limit=0, charge_carrier=PROTON,
 *                      truncate_after=TRUNCATE_AFTER, ignore_below=IGNORE_BELOW):             # <<<<<<<<<<<<<<
 *     """Visit each experimental peak and execute :meth:`_explore_local` on it with the provided
 *     parameters, populating the peak dependence graph with all viable candidates.
 */
  __pyx_t_5 = PyFloat_FromDouble(__pyx_v_12ms_deisotope_2_c_17deconvoluter_base_TRUNCATE_AFTER); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1011, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_k__23 = __pyx_t_5;
  __Pyx_GIVEREF(__pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = PyFloat_FromDouble(__pyx_v_12ms_deisotope_2_c_17deconvoluter_base_IGNORE_BELOW); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1011, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_k__24 = __pyx_t_5;
  __Pyx_GIVEREF(__pyx_t_5);
  __pyx_t_5 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":1009
 * 
 * @cython.binding(True)
 * cpdef populate_graph(DeconvoluterBase self, error_tolerance=ERROR_TOLERANCE, charge_range=(1, 8),             # <<<<<<<<<<<<<<
 *                      left_search_limit=1, right_search_limit=0, charge_carrier=PROTON,
 *                      truncate_after=TRUNCATE_AFTER, ignore_below=IGNORE_BELOW):
 */
  __pyx_t_5 = PyFloat_FromDouble(__pyx_v_12ms_deisotope_2_c_17deconvoluter_base_ERROR_TOLERANCE); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1009, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_k__21 = __pyx_t_5;
  __Pyx_GIVEREF(__pyx_t_5);
  __pyx_t_5 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":1010
 * @cython.binding(True)
 * cpdef populate_graph(DeconvoluterBase self, error_tolerance=ERROR_TOLERANCE, charge_range=(1, 8),
 *                      left_search_limit=1, right_search_limit=0, charge_carrier=PROTON,             # <<<<<<<<<<<<<<
 *                      truncate_after=TRUNCATE_AFTER, ignore_below=IGNORE_BELOW):
 *     """Visit each experimental peak and execute :meth:`_explore_local` on it with the provided
 */
  __pyx_t_5 = PyFloat_FromDouble(__pyx_v_12ms_deisotope_2_c_9averagine_PROTON); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1010, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_k__22 = __pyx_t_5;
  __Pyx_GIVEREF(__pyx_t_5);
  __pyx_t_5 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":1011
 * cpdef populate_graph(DeconvoluterBase self, error_tolerance=ERROR_TOLERANCE, charge_range=(1, 8),
 *                      left_search_limit=1, right_search_limit=0, charge_carrier=PROTON,
 *                      truncate_after=TRUNCATE_AFTER, ignore_below=IGNORE_BELOW):             # <<<<<<<<<<<<<<
 *     """Visit each experimental peak and execute :meth:`_explore_local` on it with the provided
 *     parameters, populating the peak dependence graph with all viable candidates.
 */
  __pyx_t_5 = PyFloat_FromDouble(__pyx_v_12ms_deisotope_2_c_17deconvoluter_base_TRUNCATE_AFTER); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1011, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_k__23 = __pyx_t_5;
  __Pyx_GIVEREF(__pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = PyFloat_FromDouble(__pyx_v_12ms_deisotope_2_c_17deconvoluter_base_IGNORE_BELOW); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1011, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_k__24 = __pyx_t_5;
  __Pyx_GIVEREF(__pyx_t_5);
  __pyx_t_5 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":1009
 * 
 * @cython.binding(True)
 * cpdef populate_graph(DeconvoluterBase self, error_tolerance=ERROR_TOLERANCE, charge_range=(1, 8),             # <<<<<<<<<<<<<<
 *                      left_search_limit=1, right_search_limit=0, charge_carrier=PROTON,
 *                      truncate_after=TRUNCATE_AFTER, ignore_below=IGNORE_BELOW):
 */
  __pyx_t_5 = __Pyx_CyFunction_NewEx(&__pyx_mdef_12ms_deisotope_2_c_17deconvoluter_base_7populate_graph, 0, __pyx_n_s_populate_graph, NULL, __pyx_n_s_ms_deisotope__c_deconvoluter_bas_2, __pyx_d, ((PyObject *)__pyx_codeobj__39)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1009, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_populate_graph, __pyx_t_5) < 0) __PYX_ERR(0, 1009, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":1052
 * 
 * @cython.binding(True)
 * cpdef DeconvolutedPeak _make_deconvoluted_peak(self, IsotopicFitRecord fit, double charge_carrier):             # <<<<<<<<<<<<<<
 *     '''Helper method to package a finished :class:`~.IsotopicFitRecord` into a :class:`~.DeconvolutedPeak`
 * 
 */
  __pyx_t_5 = __Pyx_CyFunction_NewEx(&__pyx_mdef_12ms_deisotope_2_c_17deconvoluter_base_9_make_deconvoluted_peak, 0, __pyx_n_s_make_deconvoluted_peak, NULL, __pyx_n_s_ms_deisotope__c_deconvoluter_bas_2, __pyx_d, ((PyObject *)__pyx_codeobj__41)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1052, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_make_deconvoluted_peak, __pyx_t_5) < 0) __PYX_ERR(0, 1052, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_DeconvoluterBase(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_5 = PyCFunction_NewEx(&__pyx_mdef_12ms_deisotope_2_c_17deconvoluter_base_11__pyx_unpickle_DeconvoluterBase, NULL, __pyx_n_s_ms_deisotope__c_deconvoluter_bas_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_DeconvoluterBase, __pyx_t_5) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_DeconvoluterBase__set_state(<DeconvoluterBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_DeconvoluterBase__set_state(DeconvoluterBase __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._slice_cache = __pyx_state[0]; __pyx_result.merge_isobaric_peaks = __pyx_state[1]; __pyx_result.minimum_intensity = __pyx_state[2]; __pyx_result.peaklist = __pyx_state[3]; __pyx_result.scale_method = __pyx_state[4]; __pyx_result.scorer = __pyx_state[5]; __pyx_result.use_subtraction = __pyx_state[6]; __pyx_result.verbose = __pyx_state[7]
 *     if len(__pyx_state) > 8 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_5 = PyCFunction_NewEx(&__pyx_mdef_12ms_deisotope_2_c_17deconvoluter_base_13__pyx_unpickle_AveragineDeconvoluterBase, NULL, __pyx_n_s_ms_deisotope__c_deconvoluter_bas_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_AveragineDeconvol, __pyx_t_5) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_MultiAveragineDeconvoluterBase(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_5 = PyCFunction_NewEx(&__pyx_mdef_12ms_deisotope_2_c_17deconvoluter_base_15__pyx_unpickle_MultiAveragineDeconvoluterBase, NULL, __pyx_n_s_ms_deisotope__c_deconvoluter_bas_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_MultiAveragineDec, __pyx_t_5) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "ms_deisotope/_c/deconvoluter_base.pyx":1
 * # cython: embedsignature=True             # <<<<<<<<<<<<<<
 * 
 * cimport cython
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_5) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "../../Miniconda2/envs/py37/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1046
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init ms_deisotope._c.deconvoluter_base", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init ms_deisotope._c.deconvoluter_base");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_CheckExact(key)) || likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (PyUnicode_GET_SIZE(**name) != PyUnicode_GET_SIZE(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (PyUnicode_GET_SIZE(**argname) != PyUnicode_GET_SIZE(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
    return 0;
}

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (PyCFunction_GET_FLAGS(func) & METH_FASTCALL) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* WriteUnraisableException */
static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
#ifdef _MSC_VER
    else state = (PyGILState_STATE)-1;
#endif
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* SliceObject */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(PyObject* obj,
        Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** _py_start, PyObject** _py_stop, PyObject** _py_slice,
        int has_cstart, int has_cstop, CYTHON_UNUSED int wraparound) {
#if CYTHON_USE_TYPE_SLOTS
    PyMappingMethods* mp;
#if PY_MAJOR_VERSION < 3
    PySequenceMethods* ms = Py_TYPE(obj)->tp_as_sequence;
    if (likely(ms && ms->sq_slice)) {
        if (!has_cstart) {
            if (_py_start && (*_py_start != Py_None)) {
                cstart = __Pyx_PyIndex_AsSsize_t(*_py_start);
                if ((cstart == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstart = 0;
        }
        if (!has_cstop) {
            if (_py_stop && (*_py_stop != Py_None)) {
                cstop = __Pyx_PyIndex_AsSsize_t(*_py_stop);
                if ((cstop == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstop = PY_SSIZE_T_MAX;
        }
        if (wraparound && unlikely((cstart < 0) | (cstop < 0)) && likely(ms->sq_length)) {
            Py_ssize_t l = ms->sq_length(obj);
            if (likely(l >= 0)) {
                if (cstop < 0) {
                    cstop += l;
                    if (cstop < 0) cstop = 0;
                }
                if (cstart < 0) {
                    cstart += l;
                    if (cstart < 0) cstart = 0;
                }
            } else {
                if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                    goto bad;
                PyErr_Clear();
            }
        }
        return ms->sq_slice(obj, cstart, cstop);
    }
#endif
    mp = Py_TYPE(obj)->tp_as_mapping;
    if (likely(mp && mp->mp_subscript))
#endif
    {
        PyObject* result;
        PyObject *py_slice, *py_start, *py_stop;
        if (_py_slice) {
            py_slice = *_py_slice;
        } else {
            PyObject* owned_start = NULL;
            PyObject* owned_stop = NULL;
            if (_py_start) {
                py_start = *_py_start;
            } else {
                if (has_cstart) {
                    owned_start = py_start = PyInt_FromSsize_t(cstart);
                    if (unlikely(!py_start)) goto bad;
                } else
                    py_start = Py_None;
            }
            if (_py_stop) {
                py_stop = *_py_stop;
            } else {
                if (has_cstop) {
                    owned_stop = py_stop = PyInt_FromSsize_t(cstop);
                    if (unlikely(!py_stop)) {
                        Py_XDECREF(owned_start);
                        goto bad;
                    }
                } else
                    py_stop = Py_None;
            }
            py_slice = PySlice_New(py_start, py_stop, Py_None);
            Py_XDECREF(owned_start);
            Py_XDECREF(owned_stop);
            if (unlikely(!py_slice)) goto bad;
        }
#if CYTHON_USE_TYPE_SLOTS
        result = mp->mp_subscript(obj, py_slice);
#else
        result = PyObject_GetItem(obj, py_slice);
#endif
        if (!_py_slice) {
            Py_DECREF(py_slice);
        }
        return result;
    }
    PyErr_Format(PyExc_TypeError,
        "'%.200s' object is unsliceable", Py_TYPE(obj)->tp_name);
bad:
    return NULL;
}

/* py_abs */
#if CYTHON_USE_PYLONG_INTERNALS
static PyObject *__Pyx_PyLong_AbsNeg(PyObject *n) {
    if (likely(Py_SIZE(n) == -1)) {
        return PyLong_FromLong(((PyLongObject*)n)->ob_digit[0]);
    }
#if CYTHON_COMPILING_IN_CPYTHON
    {
        PyObject *copy = _PyLong_Copy((PyLongObject*)n);
        if (likely(copy)) {
            Py_SIZE(copy) = -(Py_SIZE(copy));
        }
        return copy;
    }
#else
    return PyNumber_Negative(n);
#endif
}
#endif

/* PyObjectSetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* GetAttr */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* GetAttr3 */
static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r = __Pyx_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
}

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* set_iter */
static CYTHON_INLINE PyObject* __Pyx_set_iterator(PyObject* iterable, int is_set,
                                                  Py_ssize_t* p_orig_length, int* p_source_is_set) {
#if CYTHON_COMPILING_IN_CPYTHON
    is_set = is_set || likely(PySet_CheckExact(iterable) || PyFrozenSet_CheckExact(iterable));
    *p_source_is_set = is_set;
    if (likely(is_set)) {
        *p_orig_length = PySet_Size(iterable);
        Py_INCREF(iterable);
        return iterable;
    }
#else
    (void)is_set;
    *p_source_is_set = 0;
#endif
    *p_orig_length = 0;
    return PyObject_GetIter(iterable);
}
static CYTHON_INLINE int __Pyx_set_iter_next(
        PyObject* iter_obj, Py_ssize_t orig_length,
        Py_ssize_t* ppos, PyObject **value,
        int source_is_set) {
    if (!CYTHON_COMPILING_IN_CPYTHON || unlikely(!source_is_set)) {
        *value = PyIter_Next(iter_obj);
        if (unlikely(!*value)) {
            return __Pyx_IterFinish();
        }
        (void)orig_length;
        (void)ppos;
        return 1;
    }
#if CYTHON_COMPILING_IN_CPYTHON
    if (unlikely(PySet_GET_SIZE(iter_obj) != orig_length)) {
        PyErr_SetString(
            PyExc_RuntimeError,
            "set changed size during iteration");
        return -1;
    }
    {
        Py_hash_t hash;
        int ret = _PySet_NextEntry(iter_obj, ppos, value, &hash);
        assert (ret != -1);
        if (likely(ret)) {
            Py_INCREF(*value);
            return 1;
        }
    }
#endif
    return 0;
}

/* IsLittleEndian */
static CYTHON_INLINE int __Pyx_Is_Little_Endian(void)
{
  union {
    uint32_t u32;
    uint8_t u8[4];
  } S;
  S.u32 = 0x01020304;
  return S.u8[0] == 4;
}

/* BufferFormatCheck */
static void __Pyx_BufFmt_Init(__Pyx_BufFmt_Context* ctx,
                              __Pyx_BufFmt_StackElem* stack,
                              __Pyx_TypeInfo* type) {
  stack[0].field = &ctx->root;
  stack[0].parent_offset = 0;
  ctx->root.type = type;
  ctx->root.name = "buffer dtype";
  ctx->root.offset = 0;
  ctx->head = stack;
  ctx->head->field = &ctx->root;
  ctx->fmt_offset = 0;
  ctx->head->parent_offset = 0;
  ctx->new_packmode = '@';
  ctx->enc_packmode = '@';
  ctx->new_count = 1;
  ctx->enc_count = 0;
  ctx->enc_type = 0;
  ctx->is_complex = 0;
  ctx->is_valid_array = 0;
  ctx->struct_alignment = 0;
  while (type->typegroup == 'S') {
    ++ctx->head;
    ctx->head->field = type->fields;
    ctx->head->parent_offset = 0;
    type = type->fields->type;
  }
}
static int __Pyx_BufFmt_ParseNumber(const char** ts) {
    int count;
    const char* t = *ts;
    if (*t < '0' || *t > '9') {
      return -1;
    } else {
        count = *t++ - '0';
        while (*t >= '0' && *t <= '9') {
            count *= 10;
            count += *t++ - '0';
        }
    }
    *ts = t;
    return count;
}
static int __Pyx_BufFmt_ExpectNumber(const char **ts) {
    int number = __Pyx_BufFmt_ParseNumber(ts);
    if (number == -1)
        PyErr_Format(PyExc_ValueError,\
                     "Does not understand character buffer dtype format string ('%c')", **ts);
    return number;
}
static void __Pyx_BufFmt_RaiseUnexpectedChar(char ch) {
  PyErr_Format(PyExc_ValueError,
               "Unexpected format string character: '%c'", ch);
}
static const char* __Pyx_BufFmt_DescribeTypeChar(char ch, int is_complex) {
  switch (ch) {
    case 'c': return "'char'";
    case 'b': return "'signed char'";
    case 'B': return "'unsigned char'";
    case 'h': return "'short'";
    case 'H': return "'unsigned short'";
    case 'i': return "'int'";
    case 'I': return "'unsigned int'";
    case 'l': return "'long'";
    case 'L': return "'unsigned long'";
    case 'q': return "'long long'";
    case 'Q': return "'unsigned long long'";
    case 'f': return (is_complex ? "'complex float'" : "'float'");
    case 'd': return (is_complex ? "'complex double'" : "'double'");
    case 'g': return (is_complex ? "'complex long double'" : "'long double'");
    case 'T': return "a struct";
    case 'O': return "Python object";
    case 'P': return "a pointer";
    case 's': case 'p': return "a string";
    case 0: return "end";
    default: return "unparseable format string";
  }
}
static size_t __Pyx_BufFmt_TypeCharToStandardSize(char ch, int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return 2;
    case 'i': case 'I': case 'l': case 'L': return 4;
    case 'q': case 'Q': return 8;
    case 'f': return (is_complex ? 8 : 4);
    case 'd': return (is_complex ? 16 : 8);
    case 'g': {
      PyErr_SetString(PyExc_ValueError, "Python does not define a standard format string size for long double ('g')..");
      return 0;
    }
    case 'O': case 'P': return sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
static size_t __Pyx_BufFmt_TypeCharToNativeSize(char ch, int is_complex) {
  switch (ch) {
    case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(short);
    case 'i': case 'I': return sizeof(int);
    case 'l': case 'L': return sizeof(long);
    #ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(PY_LONG_LONG);
    #endif
    case 'f': return sizeof(float) * (is_complex ? 2 : 1);
    case 'd': return sizeof(double) * (is_complex ? 2 : 1);
    case 'g': return sizeof(long double) * (is_complex ? 2 : 1);
    case 'O': case 'P': return sizeof(void*);
    default: {
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
  }
}
typedef struct { char c; short x; } __Pyx_st_short;
typedef struct { char c; int x; } __Pyx_st_int;
typedef struct { char c; long x; } __Pyx_st_long;
typedef struct { char c; float x; } __Pyx_st_float;
typedef struct { char c; double x; } __Pyx_st_double;
typedef struct { char c; long double x; } __Pyx_st_longdouble;
typedef struct { char c; void *x; } __Pyx_st_void_p;
#ifdef HAVE_LONG_LONG
typedef struct { char c; PY_LONG_LONG x; } __Pyx_st_longlong;
#endif
static size_t __Pyx_BufFmt_TypeCharToAlignment(char ch, CYTHON_UNUSED int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(__Pyx_st_short) - sizeof(short);
    case 'i': case 'I': return sizeof(__Pyx_st_int) - sizeof(int);
    case 'l': case 'L': return sizeof(__Pyx_st_long) - sizeof(long);
#ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(__Pyx_st_longlong) - sizeof(PY_LONG_LONG);
#endif
    case 'f': return sizeof(__Pyx_st_float) - sizeof(float);
    case 'd': return sizeof(__Pyx_st_double) - sizeof(double);
    case 'g': return sizeof(__Pyx_st_longdouble) - sizeof(long double);
    case 'P': case 'O': return sizeof(__Pyx_st_void_p) - sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
/* These are for computing the padding at the end of the struct to align
   on the first member of the struct. This will probably the same as above,
   but we don't have any guarantees.
 */
typedef struct { short x; char c; } __Pyx_pad_short;
typedef struct { int x; char c; } __Pyx_pad_int;
typedef struct { long x; char c; } __Pyx_pad_long;
typedef struct { float x; char c; } __Pyx_pad_float;
typedef struct { double x; char c; } __Pyx_pad_double;
typedef struct { long double x; char c; } __Pyx_pad_longdouble;
typedef struct { void *x; char c; } __Pyx_pad_void_p;
#ifdef HAVE_LONG_LONG
typedef struct { PY_LONG_LONG x; char c; } __Pyx_pad_longlong;
#endif
static size_t __Pyx_BufFmt_TypeCharToPadding(char ch, CYTHON_UNUSED int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(__Pyx_pad_short) - sizeof(short);
    case 'i': case 'I': return sizeof(__Pyx_pad_int) - sizeof(int);
    case 'l': case 'L': return sizeof(__Pyx_pad_long) - sizeof(long);
#ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(__Pyx_pad_longlong) - sizeof(PY_LONG_LONG);
#endif
    case 'f': return sizeof(__Pyx_pad_float) - sizeof(float);
    case 'd': return sizeof(__Pyx_pad_double) - sizeof(double);
    case 'g': return sizeof(__Pyx_pad_longdouble) - sizeof(long double);
    case 'P': case 'O': return sizeof(__Pyx_pad_void_p) - sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
static char __Pyx_BufFmt_TypeCharToGroup(char ch, int is_complex) {
  switch (ch) {
    case 'c':
        return 'H';
    case 'b': case 'h': case 'i':
    case 'l': case 'q': case 's': case 'p':
        return 'I';
    case 'B': case 'H': case 'I': case 'L': case 'Q':
        return 'U';
    case 'f': case 'd': case 'g':
        return (is_complex ? 'C' : 'R');
    case 'O':
        return 'O';
    case 'P':
        return 'P';
    default: {
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
  }
}
static void __Pyx_BufFmt_RaiseExpected(__Pyx_BufFmt_Context* ctx) {
  if (ctx->head == NULL || ctx->head->field == &ctx->root) {
    const char* expected;
    const char* quote;
    if (ctx->head == NULL) {
      expected = "end";
      quote = "";
    } else {
      expected = ctx->head->field->type->name;
      quote = "'";
    }
    PyErr_Format(PyExc_ValueError,
                 "Buffer dtype mismatch, expected %s%s%s but got %s",
                 quote, expected, quote,
                 __Pyx_BufFmt_DescribeTypeChar(ctx->enc_type, ctx->is_complex));
  } else {
    __Pyx_StructField* field = ctx->head->field;
    __Pyx_StructField* parent = (ctx->head - 1)->field;
    PyErr_Format(PyExc_ValueError,
                 "Buffer dtype mismatch, expected '%s' but got %s in '%s.%s'",
                 field->type->name, __Pyx_BufFmt_DescribeTypeChar(ctx->enc_type, ctx->is_complex),
                 parent->type->name, field->name);
  }
}
static int __Pyx_BufFmt_ProcessTypeChunk(__Pyx_BufFmt_Context* ctx) {
  char group;
  size_t size, offset, arraysize = 1;
  if (ctx->enc_type == 0) return 0;
  if (ctx->head->field->type->arraysize[0]) {
    int i, ndim = 0;
    if (ctx->enc_type == 's' || ctx->enc_type == 'p') {
        ctx->is_valid_array = ctx->head->field->type->ndim == 1;
        ndim = 1;
        if (ctx->enc_count != ctx->head->field->type->arraysize[0]) {
            PyErr_Format(PyExc_ValueError,
                         "Expected a dimension of size %zu, got %zu",
                         ctx->head->field->type->arraysize[0], ctx->enc_count);
            return -1;
        }
    }
    if (!ctx->is_valid_array) {
      PyErr_Format(PyExc_ValueError, "Expected %d dimensions, got %d",
                   ctx->head->field->type->ndim, ndim);
      return -1;
    }
    for (i = 0; i < ctx->head->field->type->ndim; i++) {
      arraysize *= ctx->head->field->type->arraysize[i];
    }
    ctx->is_valid_array = 0;
    ctx->enc_count = 1;
  }
  group = __Pyx_BufFmt_TypeCharToGroup(ctx->enc_type, ctx->is_complex);
  do {
    __Pyx_StructField* field = ctx->head->field;
    __Pyx_TypeInfo* type = field->type;
    if (ctx->enc_packmode == '@' || ctx->enc_packmode == '^') {
      size = __Pyx_BufFmt_TypeCharToNativeSize(ctx->enc_type, ctx->is_complex);
    } else {
      size = __Pyx_BufFmt_TypeCharToStandardSize(ctx->enc_type, ctx->is_complex);
    }
    if (ctx->enc_packmode == '@') {
      size_t align_at = __Pyx_BufFmt_TypeCharToAlignment(ctx->enc_type, ctx->is_complex);
      size_t align_mod_offset;
      if (align_at == 0) return -1;
      align_mod_offset = ctx->fmt_offset % align_at;
      if (align_mod_offset > 0) ctx->fmt_offset += align_at - align_mod_offset;
      if (ctx->struct_alignment == 0)
          ctx->struct_alignment = __Pyx_BufFmt_TypeCharToPadding(ctx->enc_type,
                                                                 ctx->is_complex);
    }
    if (type->size != size || type->typegroup != group) {
      if (type->typegroup == 'C' && type->fields != NULL) {
        size_t parent_offset = ctx->head->parent_offset + field->offset;
        ++ctx->head;
        ctx->head->field = type->fields;
        ctx->head->parent_offset = parent_offset;
        continue;
      }
      if ((type->typegroup == 'H' || group == 'H') && type->size == size) {
      } else {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return -1;
      }
    }
    offset = ctx->head->parent_offset + field->offset;
    if (ctx->fmt_offset != offset) {
      PyErr_Format(PyExc_ValueError,
                   "Buffer dtype mismatch; next field is at offset %" CYTHON_FORMAT_SSIZE_T "d but %" CYTHON_FORMAT_SSIZE_T "d expected",
                   (Py_ssize_t)ctx->fmt_offset, (Py_ssize_t)offset);
      return -1;
    }
    ctx->fmt_offset += size;
    if (arraysize)
      ctx->fmt_offset += (arraysize - 1) * size;
    --ctx->enc_count;
    while (1) {
      if (field == &ctx->root) {
        ctx->head = NULL;
        if (ctx->enc_count != 0) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return -1;
        }
        break;
      }
      ctx->head->field = ++field;
      if (field->type == NULL) {
        --ctx->head;
        field = ctx->head->field;
        continue;
      } else if (field->type->typegroup == 'S') {
        size_t parent_offset = ctx->head->parent_offset + field->offset;
        if (field->type->fields->type == NULL) continue;
        field = field->type->fields;
        ++ctx->head;
        ctx->head->field = field;
        ctx->head->parent_offset = parent_offset;
        break;
      } else {
        break;
      }
    }
  } while (ctx->enc_count);
  ctx->enc_type = 0;
  ctx->is_complex = 0;
  return 0;
}
static PyObject *
__pyx_buffmt_parse_array(__Pyx_BufFmt_Context* ctx, const char** tsp)
{
    const char *ts = *tsp;
    int i = 0, number;
    int ndim = ctx->head->field->type->ndim;
;
    ++ts;
    if (ctx->new_count != 1) {
        PyErr_SetString(PyExc_ValueError,
                        "Cannot handle repeated arrays in format string");
        return NULL;
    }
    if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
    while (*ts && *ts != ')') {
        switch (*ts) {
            case ' ': case '\f': case '\r': case '\n': case '\t': case '\v':  continue;
            default:  break;
        }
        number = __Pyx_BufFmt_ExpectNumber(&ts);
        if (number == -1) return NULL;
        if (i < ndim && (size_t) number != ctx->head->field->type->arraysize[i])
            return PyErr_Format(PyExc_ValueError,
                        "Expected a dimension of size %zu, got %d",
                        ctx->head->field->type->arraysize[i], number);
        if (*ts != ',' && *ts != ')')
            return PyErr_Format(PyExc_ValueError,
                                "Expected a comma in format string, got '%c'", *ts);
        if (*ts == ',') ts++;
        i++;
    }
    if (i != ndim)
        return PyErr_Format(PyExc_ValueError, "Expected %d dimension(s), got %d",
                            ctx->head->field->type->ndim, i);
    if (!*ts) {
        PyErr_SetString(PyExc_ValueError,
                        "Unexpected end of format string, expected ')'");
        return NULL;
    }
    ctx->is_valid_array = 1;
    ctx->new_count = 1;
    *tsp = ++ts;
    return Py_None;
}
static const char* __Pyx_BufFmt_CheckString(__Pyx_BufFmt_Context* ctx, const char* ts) {
  int got_Z = 0;
  while (1) {
    switch(*ts) {
      case 0:
        if (ctx->enc_type != 0 && ctx->head == NULL) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return NULL;
        }
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        if (ctx->head != NULL) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return NULL;
        }
        return ts;
      case ' ':
      case '\r':
      case '\n':
        ++ts;
        break;
      case '<':
        if (!__Pyx_Is_Little_Endian()) {
          PyErr_SetString(PyExc_ValueError, "Little-endian buffer not supported on big-endian compiler");
          return NULL;
        }
        ctx->new_packmode = '=';
        ++ts;
        break;
      case '>':
      case '!':
        if (__Pyx_Is_Little_Endian()) {
          PyErr_SetString(PyExc_ValueError, "Big-endian buffer not supported on little-endian compiler");
          return NULL;
        }
        ctx->new_packmode = '=';
        ++ts;
        break;
      case '=':
      case '@':
      case '^':
        ctx->new_packmode = *ts++;
        break;
      case 'T':
        {
          const char* ts_after_sub;
          size_t i, struct_count = ctx->new_count;
          size_t struct_alignment = ctx->struct_alignment;
          ctx->new_count = 1;
          ++ts;
          if (*ts != '{') {
            PyErr_SetString(PyExc_ValueError, "Buffer acquisition: Expected '{' after 'T'");
            return NULL;
          }
          if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
          ctx->enc_type = 0;
          ctx->enc_count = 0;
          ctx->struct_alignment = 0;
          ++ts;
          ts_after_sub = ts;
          for (i = 0; i != struct_count; ++i) {
            ts_after_sub = __Pyx_BufFmt_CheckString(ctx, ts);
            if (!ts_after_sub) return NULL;
          }
          ts = ts_after_sub;
          if (struct_alignment) ctx->struct_alignment = struct_alignment;
        }
        break;
      case '}':
        {
          size_t alignment = ctx->struct_alignment;
          ++ts;
          if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
          ctx->enc_type = 0;
          if (alignment && ctx->fmt_offset % alignment) {
            ctx->fmt_offset += alignment - (ctx->fmt_offset % alignment);
          }
        }
        return ts;
      case 'x':
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        ctx->fmt_offset += ctx->new_count;
        ctx->new_count = 1;
        ctx->enc_count = 0;
        ctx->enc_type = 0;
        ctx->enc_packmode = ctx->new_packmode;
        ++ts;
        break;
      case 'Z':
        got_Z = 1;
        ++ts;
        if (*ts != 'f' && *ts != 'd' && *ts != 'g') {
          __Pyx_BufFmt_RaiseUnexpectedChar('Z');
          return NULL;
        }
        CYTHON_FALLTHROUGH;
      case 'c': case 'b': case 'B': case 'h': case 'H': case 'i': case 'I':
      case 'l': case 'L': case 'q': case 'Q':
      case 'f': case 'd': case 'g':
      case 'O': case 'p':
        if (ctx->enc_type == *ts && got_Z == ctx->is_complex &&
            ctx->enc_packmode == ctx->new_packmode) {
          ctx->enc_count += ctx->new_count;
          ctx->new_count = 1;
          got_Z = 0;
          ++ts;
          break;
        }
        CYTHON_FALLTHROUGH;
      case 's':
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        ctx->enc_count = ctx->new_count;
        ctx->enc_packmode = ctx->new_packmode;
        ctx->enc_type = *ts;
        ctx->is_complex = got_Z;
        ++ts;
        ctx->new_count = 1;
        got_Z = 0;
        break;
      case ':':
        ++ts;
        while(*ts != ':') ++ts;
        ++ts;
        break;
      case '(':
        if (!__pyx_buffmt_parse_array(ctx, &ts)) return NULL;
        break;
      default:
        {
          int number = __Pyx_BufFmt_ExpectNumber(&ts);
          if (number == -1) return NULL;
          ctx->new_count = (size_t)number;
        }
    }
  }
}

/* BufferGetAndValidate */
  static CYTHON_INLINE void __Pyx_SafeReleaseBuffer(Py_buffer* info) {
  if (unlikely(info->buf == NULL)) return;
  if (info->suboffsets == __Pyx_minusones) info->suboffsets = NULL;
  __Pyx_ReleaseBuffer(info);
}
static void __Pyx_ZeroBuffer(Py_buffer* buf) {
  buf->buf = NULL;
  buf->obj = NULL;
  buf->strides = __Pyx_zeros;
  buf->shape = __Pyx_zeros;
  buf->suboffsets = __Pyx_minusones;
}
static int __Pyx__GetBufferAndValidate(
        Py_buffer* buf, PyObject* obj,  __Pyx_TypeInfo* dtype, int flags,
        int nd, int cast, __Pyx_BufFmt_StackElem* stack)
{
  buf->buf = NULL;
  if (unlikely(__Pyx_GetBuffer(obj, buf, flags) == -1)) {
    __Pyx_ZeroBuffer(buf);
    return -1;
  }
  if (unlikely(buf->ndim != nd)) {
    PyErr_Format(PyExc_ValueError,
                 "Buffer has wrong number of dimensions (expected %d, got %d)",
                 nd, buf->ndim);
    goto fail;
  }
  if (!cast) {
    __Pyx_BufFmt_Context ctx;
    __Pyx_BufFmt_Init(&ctx, stack, dtype);
    if (!__Pyx_BufFmt_CheckString(&ctx, buf->format)) goto fail;
  }
  if (unlikely((size_t)buf->itemsize != dtype->size)) {
    PyErr_Format(PyExc_ValueError,
      "Item size of buffer (%" CYTHON_FORMAT_SSIZE_T "d byte%s) does not match size of '%s' (%" CYTHON_FORMAT_SSIZE_T "d byte%s)",
      buf->itemsize, (buf->itemsize > 1) ? "s" : "",
      dtype->name, (Py_ssize_t)dtype->size, (dtype->size > 1) ? "s" : "");
    goto fail;
  }
  if (buf->suboffsets == NULL) buf->suboffsets = __Pyx_minusones;
  return 0;
fail:;
  __Pyx_SafeReleaseBuffer(buf);
  return -1;
}

/* BufferFallbackError */
  static void __Pyx_RaiseBufferFallbackError(void) {
  PyErr_SetString(PyExc_ValueError,
     "Buffer acquisition failed on assignment; and then reacquiring the old buffer failed too!");
}

/* BufferIndexError */
  static void __Pyx_RaiseBufferIndexError(int axis) {
  PyErr_Format(PyExc_IndexError,
     "Out of bounds on buffer access (axis %d)", axis);
}

/* PyObjectCallNoArg */
  #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* RaiseException */
  #if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* DictGetItem */
  #if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            if (unlikely(PyTuple_Check(key))) {
                PyObject* args = PyTuple_Pack(1, key);
                if (likely(args)) {
                    PyErr_SetObject(PyExc_KeyError, args);
                    Py_DECREF(args);
                }
            } else {
                PyErr_SetObject(PyExc_KeyError, key);
            }
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#endif

/* SetItemInt */
  static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v) {
    int r;
    if (!j) return -1;
    r = PyObject_SetItem(o, j, v);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v, int is_list,
                                               CYTHON_NCP_UNUSED int wraparound, CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = (!wraparound) ? i : ((likely(i >= 0)) ? i : i + PyList_GET_SIZE(o));
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o)))) {
            PyObject* old = PyList_GET_ITEM(o, n);
            Py_INCREF(v);
            PyList_SET_ITEM(o, n, v);
            Py_DECREF(old);
            return 1;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_ass_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return -1;
                    PyErr_Clear();
                }
            }
            return m->sq_ass_item(o, i, v);
        }
    }
#else
#if CYTHON_COMPILING_IN_PYPY
    if (is_list || (PySequence_Check(o) && !PyDict_Check(o)))
#else
    if (is_list || PySequence_Check(o))
#endif
    {
        return PySequence_SetItem(o, i, v);
    }
#endif
    return __Pyx_SetItemInt_Generic(o, PyInt_FromSsize_t(i), v);
}

/* PyObjectGetMethod */
  static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (descr != NULL) {
        *method = descr;
        return 0;
    }
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(name));
#endif
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod0 */
  static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
    PyObject *method = NULL, *result = NULL;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_CallOneArg(method, obj);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) goto bad;
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
}

/* RaiseNeedMoreValuesToUnpack */
  static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* RaiseTooManyValuesToUnpack */
  static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* UnpackItemEndCheck */
  static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    } else {
        return __Pyx_IterFinish();
    }
    return 0;
}

/* RaiseNoneIterError */
  static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* UnpackTupleError */
  static void __Pyx_UnpackTupleError(PyObject *t, Py_ssize_t index) {
    if (t == Py_None) {
      __Pyx_RaiseNoneNotIterableError();
    } else if (PyTuple_GET_SIZE(t) < index) {
      __Pyx_RaiseNeedMoreValuesError(PyTuple_GET_SIZE(t));
    } else {
      __Pyx_RaiseTooManyValuesError(index);
    }
}

/* UnpackTuple2 */
  static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
        PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2, int decref_tuple) {
    PyObject *value1 = NULL, *value2 = NULL;
#if CYTHON_COMPILING_IN_PYPY
    value1 = PySequence_ITEM(tuple, 0);  if (unlikely(!value1)) goto bad;
    value2 = PySequence_ITEM(tuple, 1);  if (unlikely(!value2)) goto bad;
#else
    value1 = PyTuple_GET_ITEM(tuple, 0);  Py_INCREF(value1);
    value2 = PyTuple_GET_ITEM(tuple, 1);  Py_INCREF(value2);
#endif
    if (decref_tuple) {
        Py_DECREF(tuple);
    }
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
#if CYTHON_COMPILING_IN_PYPY
bad:
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
#endif
}
static int __Pyx_unpack_tuple2_generic(PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2,
                                       int has_known_size, int decref_tuple) {
    Py_ssize_t index;
    PyObject *value1 = NULL, *value2 = NULL, *iter = NULL;
    iternextfunc iternext;
    iter = PyObject_GetIter(tuple);
    if (unlikely(!iter)) goto bad;
    if (decref_tuple) { Py_DECREF(tuple); tuple = NULL; }
    iternext = Py_TYPE(iter)->tp_iternext;
    value1 = iternext(iter); if (unlikely(!value1)) { index = 0; goto unpacking_failed; }
    value2 = iternext(iter); if (unlikely(!value2)) { index = 1; goto unpacking_failed; }
    if (!has_known_size && unlikely(__Pyx_IternextUnpackEndCheck(iternext(iter), 2))) goto bad;
    Py_DECREF(iter);
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
unpacking_failed:
    if (!has_known_size && __Pyx_IterFinish() == 0)
        __Pyx_RaiseNeedMoreValuesError(index);
bad:
    Py_XDECREF(iter);
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
}

/* dict_iter */
  static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* iterable, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_source_is_dict) {
    is_dict = is_dict || likely(PyDict_CheckExact(iterable));
    *p_source_is_dict = is_dict;
    if (is_dict) {
#if !CYTHON_COMPILING_IN_PYPY
        *p_orig_length = PyDict_Size(iterable);
        Py_INCREF(iterable);
        return iterable;
#elif PY_MAJOR_VERSION >= 3
        static PyObject *py_items = NULL, *py_keys = NULL, *py_values = NULL;
        PyObject **pp = NULL;
        if (method_name) {
            const char *name = PyUnicode_AsUTF8(method_name);
            if (strcmp(name, "iteritems") == 0) pp = &py_items;
            else if (strcmp(name, "iterkeys") == 0) pp = &py_keys;
            else if (strcmp(name, "itervalues") == 0) pp = &py_values;
            if (pp) {
                if (!*pp) {
                    *pp = PyUnicode_FromString(name + 4);
                    if (!*pp)
                        return NULL;
                }
                method_name = *pp;
            }
        }
#endif
    }
    *p_orig_length = 0;
    if (method_name) {
        PyObject* iter;
        iterable = __Pyx_PyObject_CallMethod0(iterable, method_name);
        if (!iterable)
            return NULL;
#if !CYTHON_COMPILING_IN_PYPY
        if (PyTuple_CheckExact(iterable) || PyList_CheckExact(iterable))
            return iterable;
#endif
        iter = PyObject_GetIter(iterable);
        Py_DECREF(iterable);
        return iter;
    }
    return PyObject_GetIter(iterable);
}
static CYTHON_INLINE int __Pyx_dict_iter_next(
        PyObject* iter_obj, CYTHON_NCP_UNUSED Py_ssize_t orig_length, CYTHON_NCP_UNUSED Py_ssize_t* ppos,
        PyObject** pkey, PyObject** pvalue, PyObject** pitem, int source_is_dict) {
    PyObject* next_item;
#if !CYTHON_COMPILING_IN_PYPY
    if (source_is_dict) {
        PyObject *key, *value;
        if (unlikely(orig_length != PyDict_Size(iter_obj))) {
            PyErr_SetString(PyExc_RuntimeError, "dictionary changed size during iteration");
            return -1;
        }
        if (unlikely(!PyDict_Next(iter_obj, ppos, &key, &value))) {
            return 0;
        }
        if (pitem) {
            PyObject* tuple = PyTuple_New(2);
            if (unlikely(!tuple)) {
                return -1;
            }
            Py_INCREF(key);
            Py_INCREF(value);
            PyTuple_SET_ITEM(tuple, 0, key);
            PyTuple_SET_ITEM(tuple, 1, value);
            *pitem = tuple;
        } else {
            if (pkey) {
                Py_INCREF(key);
                *pkey = key;
            }
            if (pvalue) {
                Py_INCREF(value);
                *pvalue = value;
            }
        }
        return 1;
    } else if (PyTuple_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyTuple_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyTuple_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else if (PyList_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyList_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyList_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else
#endif
    {
        next_item = PyIter_Next(iter_obj);
        if (unlikely(!next_item)) {
            return __Pyx_IterFinish();
        }
    }
    if (pitem) {
        *pitem = next_item;
    } else if (pkey && pvalue) {
        if (__Pyx_unpack_tuple2(next_item, pkey, pvalue, source_is_dict, source_is_dict, 1))
            return -1;
    } else if (pkey) {
        *pkey = next_item;
    } else {
        *pvalue = next_item;
    }
    return 1;
}

/* pyfrozenset_new */
  static CYTHON_INLINE PyObject* __Pyx_PyFrozenSet_New(PyObject* it) {
    if (it) {
        PyObject* result;
#if CYTHON_COMPILING_IN_PYPY
        PyObject* args;
        args = PyTuple_Pack(1, it);
        if (unlikely(!args))
            return NULL;
        result = PyObject_Call((PyObject*)&PyFrozenSet_Type, args, NULL);
        Py_DECREF(args);
        return result;
#else
        if (PyFrozenSet_CheckExact(it)) {
            Py_INCREF(it);
            return it;
        }
        result = PyFrozenSet_New(it);
        if (unlikely(!result))
            return NULL;
        if (likely(PySet_GET_SIZE(result)))
            return result;
        Py_DECREF(result);
#endif
    }
#if CYTHON_USE_TYPE_SLOTS
    return PyFrozenSet_Type.tp_new(&PyFrozenSet_Type, __pyx_empty_tuple, NULL);
#else
    return PyObject_Call((PyObject*)&PyFrozenSet_Type, __pyx_empty_tuple, NULL);
#endif
}

/* py_set_discard_unhashable */
  static int __Pyx_PySet_DiscardUnhashable(PyObject *set, PyObject *key) {
    PyObject *tmpkey;
    int rv;
    if (likely(!PySet_Check(key) || !PyErr_ExceptionMatches(PyExc_TypeError)))
        return -1;
    PyErr_Clear();
    tmpkey = __Pyx_PyFrozenSet_New(key);
    if (tmpkey == NULL)
        return -1;
    rv = PySet_Discard(set, tmpkey);
    Py_DECREF(tmpkey);
    return rv;
}

/* py_set_discard */
  static CYTHON_INLINE int __Pyx_PySet_Discard(PyObject *set, PyObject *key) {
    int found = PySet_Discard(set, key);
    if (unlikely(found < 0)) {
        found = __Pyx_PySet_DiscardUnhashable(set, key);
    }
    return found;
}

/* Import */
  static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if (strchr(__Pyx_MODULE_NAME, '.')) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
  static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* HasAttr */
  static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!__Pyx_PyBaseString_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_GetAttr(o, n);
    if (unlikely(!r)) {
        PyErr_Clear();
        return 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}

/* GetTopmostException */
  #if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
  #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* GetException */
  #if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* PyObject_GenericGetAttrNoDict */
  #if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
  #if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetVTable */
  static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* SetupReduce */
  static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStr(meth, __pyx_n_s_name);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    if (_PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate)) goto GOOD;
#else
    if (PyObject_HasAttr(type_obj, __pyx_n_s_getstate)) goto GOOD;
#endif
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_cython); if (unlikely(!reduce_cython)) goto BAD;
            ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto BAD;
            ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto BAD;
            setstate = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate_cython); if (unlikely(!setstate_cython)) goto BAD;
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto BAD;
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto GOOD;
BAD:
    if (!PyErr_Occurred())
        PyErr_Format(PyExc_RuntimeError, "Unable to initialize pickling for %s", ((PyTypeObject*)type_obj)->tp_name);
    ret = -1;
GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* TypeImport */
  #ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* GetVTable */
  static void* __Pyx_GetVtable(PyObject *dict) {
    void* ptr;
    PyObject *ob = PyObject_GetItem(dict, __pyx_n_s_pyx_vtable);
    if (!ob)
        goto bad;
#if PY_VERSION_HEX >= 0x02070000
    ptr = PyCapsule_GetPointer(ob, 0);
#else
    ptr = PyCObject_AsVoidPtr(ob);
#endif
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* FetchCommonType */
  static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* fake_module;
    PyTypeObject* cached_type = NULL;
    fake_module = PyImport_AddModule((char*) "_cython_" CYTHON_ABI);
    if (!fake_module) return NULL;
    Py_INCREF(fake_module);
    cached_type = (PyTypeObject*) PyObject_GetAttrString(fake_module, type->tp_name);
    if (cached_type) {
        if (!PyType_Check((PyObject*)cached_type)) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s is not a type object",
                type->tp_name);
            goto bad;
        }
        if (cached_type->tp_basicsize != type->tp_basicsize) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s has the wrong size, try recompiling",
                type->tp_name);
            goto bad;
        }
    } else {
        if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
        PyErr_Clear();
        if (PyType_Ready(type) < 0) goto bad;
        if (PyObject_SetAttrString(fake_module, type->tp_name, (PyObject*) type) < 0)
            goto bad;
        Py_INCREF(type);
        cached_type = type;
    }
done:
    Py_DECREF(fake_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* CythonFunction */
  #include <structmember.h>
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *closure)
{
    if (unlikely(op->func_doc == NULL)) {
        if (op->func.m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(op->func.m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(op->func.m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp = op->func_doc;
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    op->func_doc = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    if (unlikely(op->func_name == NULL)) {
#if PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(op->func.m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(op->func.m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = op->func_name;
    Py_INCREF(value);
    op->func_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = op->func_qualname;
    Py_INCREF(value);
    op->func_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_self(__pyx_CyFunctionObject *m, CYTHON_UNUSED void *closure)
{
    PyObject *self;
    self = m->func_closure;
    if (self == NULL)
        self = Py_None;
    Py_INCREF(self);
    return self;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    tmp = op->func_dict;
    Py_INCREF(value);
    op->func_dict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(CYTHON_UNUSED __pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyTuple_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_tuple;
    op->defaults_tuple = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->defaults_tuple;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_kwdict;
    op->defaults_kwdict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->defaults_kwdict;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value || value == Py_None) {
        value = NULL;
    } else if (!PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    tmp = op->func_annotations;
    op->func_annotations = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->func_annotations;
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "__self__", (getter)__Pyx_CyFunction_get_self, 0, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
    {(char *) "__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), PY_WRITE_RESTRICTED, 0},
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, CYTHON_UNUSED PyObject *args)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromString(m->func.m_ml->ml_name);
#else
    return PyString_FromString(m->func.m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func.m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_New(PyTypeObject *type, PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    __pyx_CyFunctionObject *op = PyObject_GC_New(__pyx_CyFunctionObject, type);
    if (op == NULL)
        return NULL;
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
    op->func.m_ml = ml;
    op->func.m_self = (PyObject *) op;
    Py_XINCREF(closure);
    op->func_closure = closure;
    Py_XINCREF(module);
    op->func.m_module = module;
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
    op->func_classobj = NULL;
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    PyObject_GC_Track(op);
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
    Py_CLEAR(m->func.m_module);
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
    Py_CLEAR(m->func_classobj);
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyObject_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    PyObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
    Py_VISIT(m->func.m_module);
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
    Py_VISIT(m->func_classobj);
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject *__Pyx_CyFunction_descr_get(PyObject *func, PyObject *obj, PyObject *type)
{
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    if (m->flags & __Pyx_CYFUNCTION_STATICMETHOD) {
        Py_INCREF(func);
        return func;
    }
    if (m->flags & __Pyx_CYFUNCTION_CLASSMETHOD) {
        if (type == NULL)
            type = (PyObject *)(Py_TYPE(obj));
        return __Pyx_PyMethod_New(func, type, (PyObject *)(Py_TYPE(type)));
    }
    if (obj == Py_None)
        obj = NULL;
    return __Pyx_PyMethod_New(func, obj, type);
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    Py_ssize_t size;
    switch (f->m_ml->ml_flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void*)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 0))
                return (*meth)(self, NULL);
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags in "
                        "__Pyx_CyFunction_Call. METH_OLDARGS is no "
                        "longer supported!");
        return NULL;
    }
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    return __Pyx_CyFunction_CallMethod(func, ((PyCFunctionObject*)func)->m_self, arg, kw);
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
        argc = PyTuple_GET_SIZE(args);
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
    0,
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_CallAsMethod,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_CyFunction_descr_get,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1
    0,
#endif
};
static int __pyx_CyFunction_init(void) {
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
    if (unlikely(__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_Malloc(size);
    if (unlikely(!m->defaults))
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* FusedFunction */
  static PyObject *
__pyx_FusedFunction_New(PyTypeObject *type, PyMethodDef *ml, int flags,
                        PyObject *qualname, PyObject *self,
                        PyObject *module, PyObject *globals,
                        PyObject *code)
{
    __pyx_FusedFunctionObject *fusedfunc =
        (__pyx_FusedFunctionObject *) __Pyx_CyFunction_New(type, ml, flags, qualname,
                                                           self, module, globals, code);
    if (!fusedfunc)
        return NULL;
    fusedfunc->__signatures__ = NULL;
    fusedfunc->type = NULL;
    fusedfunc->self = NULL;
    return (PyObject *) fusedfunc;
}
static void
__pyx_FusedFunction_dealloc(__pyx_FusedFunctionObject *self)
{
    PyObject_GC_UnTrack(self);
    Py_CLEAR(self->self);
    Py_CLEAR(self->type);
    Py_CLEAR(self->__signatures__);
    __Pyx__CyFunction_dealloc((__pyx_CyFunctionObject *) self);
}
static int
__pyx_FusedFunction_traverse(__pyx_FusedFunctionObject *self,
                             visitproc visit,
                             void *arg)
{
    Py_VISIT(self->self);
    Py_VISIT(self->type);
    Py_VISIT(self->__signatures__);
    return __Pyx_CyFunction_traverse((__pyx_CyFunctionObject *) self, visit, arg);
}
static int
__pyx_FusedFunction_clear(__pyx_FusedFunctionObject *self)
{
    Py_CLEAR(self->self);
    Py_CLEAR(self->type);
    Py_CLEAR(self->__signatures__);
    return __Pyx_CyFunction_clear((__pyx_CyFunctionObject *) self);
}
static PyObject *
__pyx_FusedFunction_descr_get(PyObject *self, PyObject *obj, PyObject *type)
{
    __pyx_FusedFunctionObject *func, *meth;
    func = (__pyx_FusedFunctionObject *) self;
    if (func->self || func->func.flags & __Pyx_CYFUNCTION_STATICMETHOD) {
        Py_INCREF(self);
        return self;
    }
    if (obj == Py_None)
        obj = NULL;
    meth = (__pyx_FusedFunctionObject *) __pyx_FusedFunction_NewEx(
                    ((PyCFunctionObject *) func)->m_ml,
                    ((__pyx_CyFunctionObject *) func)->flags,
                    ((__pyx_CyFunctionObject *) func)->func_qualname,
                    ((__pyx_CyFunctionObject *) func)->func_closure,
                    ((PyCFunctionObject *) func)->m_module,
                    ((__pyx_CyFunctionObject *) func)->func_globals,
                    ((__pyx_CyFunctionObject *) func)->func_code);
    if (!meth)
        return NULL;
    Py_XINCREF(func->func.func_classobj);
    meth->func.func_classobj = func->func.func_classobj;
    Py_XINCREF(func->__signatures__);
    meth->__signatures__ = func->__signatures__;
    Py_XINCREF(type);
    meth->type = type;
    Py_XINCREF(func->func.defaults_tuple);
    meth->func.defaults_tuple = func->func.defaults_tuple;
    if (func->func.flags & __Pyx_CYFUNCTION_CLASSMETHOD)
        obj = type;
    Py_XINCREF(obj);
    meth->self = obj;
    return (PyObject *) meth;
}
static PyObject *
_obj_to_str(PyObject *obj)
{
    if (PyType_Check(obj))
        return PyObject_GetAttr(obj, __pyx_n_s_name);
    else
        return PyObject_Str(obj);
}
static PyObject *
__pyx_FusedFunction_getitem(__pyx_FusedFunctionObject *self, PyObject *idx)
{
    PyObject *signature = NULL;
    PyObject *unbound_result_func;
    PyObject *result_func = NULL;
    if (self->__signatures__ == NULL) {
        PyErr_SetString(PyExc_TypeError, "Function is not fused");
        return NULL;
    }
    if (PyTuple_Check(idx)) {
        PyObject *list = PyList_New(0);
        Py_ssize_t n = PyTuple_GET_SIZE(idx);
        PyObject *string = NULL;
        PyObject *sep = NULL;
        int i;
        if (!list)
            return NULL;
        for (i = 0; i < n; i++) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            PyObject *item = PyTuple_GET_ITEM(idx, i);
#else
            PyObject *item = PySequence_ITEM(idx, i);
#endif
            string = _obj_to_str(item);
#if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
            Py_DECREF(item);
#endif
            if (!string || PyList_Append(list, string) < 0)
                goto __pyx_err;
            Py_DECREF(string);
        }
        sep = PyUnicode_FromString("|");
        if (sep)
            signature = PyUnicode_Join(sep, list);
__pyx_err:
;
        Py_DECREF(list);
        Py_XDECREF(sep);
    } else {
        signature = _obj_to_str(idx);
    }
    if (!signature)
        return NULL;
    unbound_result_func = PyObject_GetItem(self->__signatures__, signature);
    if (unbound_result_func) {
        if (self->self || self->type) {
            __pyx_FusedFunctionObject *unbound = (__pyx_FusedFunctionObject *) unbound_result_func;
            Py_CLEAR(unbound->func.func_classobj);
            Py_XINCREF(self->func.func_classobj);
            unbound->func.func_classobj = self->func.func_classobj;
            result_func = __pyx_FusedFunction_descr_get(unbound_result_func,
                                                        self->self, self->type);
        } else {
            result_func = unbound_result_func;
            Py_INCREF(result_func);
        }
    }
    Py_DECREF(signature);
    Py_XDECREF(unbound_result_func);
    return result_func;
}
static PyObject *
__pyx_FusedFunction_callfunction(PyObject *func, PyObject *args, PyObject *kw)
{
     __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
    int static_specialized = (cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD &&
                              !((__pyx_FusedFunctionObject *) func)->__signatures__);
    if (cyfunc->flags & __Pyx_CYFUNCTION_CCLASS && !static_specialized) {
        return __Pyx_CyFunction_CallAsMethod(func, args, kw);
    } else {
        return __Pyx_CyFunction_Call(func, args, kw);
    }
}
static PyObject *
__pyx_FusedFunction_call(PyObject *func, PyObject *args, PyObject *kw)
{
    __pyx_FusedFunctionObject *binding_func = (__pyx_FusedFunctionObject *) func;
    Py_ssize_t argc = PyTuple_GET_SIZE(args);
    PyObject *new_args = NULL;
    __pyx_FusedFunctionObject *new_func = NULL;
    PyObject *result = NULL;
    PyObject *self = NULL;
    int is_staticmethod = binding_func->func.flags & __Pyx_CYFUNCTION_STATICMETHOD;
    int is_classmethod = binding_func->func.flags & __Pyx_CYFUNCTION_CLASSMETHOD;
    if (binding_func->self) {
        Py_ssize_t i;
        new_args = PyTuple_New(argc + 1);
        if (!new_args)
            return NULL;
        self = binding_func->self;
#if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
        Py_INCREF(self);
#endif
        Py_INCREF(self);
        PyTuple_SET_ITEM(new_args, 0, self);
        for (i = 0; i < argc; i++) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            PyObject *item = PyTuple_GET_ITEM(args, i);
            Py_INCREF(item);
#else
            PyObject *item = PySequence_ITEM(args, i);  if (unlikely(!item)) goto bad;
#endif
            PyTuple_SET_ITEM(new_args, i + 1, item);
        }
        args = new_args;
    } else if (binding_func->type) {
        if (argc < 1) {
            PyErr_SetString(PyExc_TypeError, "Need at least one argument, 0 given.");
            return NULL;
        }
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        self = PyTuple_GET_ITEM(args, 0);
#else
        self = PySequence_ITEM(args, 0);  if (unlikely(!self)) return NULL;
#endif
    }
    if (self && !is_classmethod && !is_staticmethod) {
        int is_instance = PyObject_IsInstance(self, binding_func->type);
        if (unlikely(!is_instance)) {
            PyErr_Format(PyExc_TypeError,
                         "First argument should be of type %.200s, got %.200s.",
                         ((PyTypeObject *) binding_func->type)->tp_name,
                         self->ob_type->tp_name);
            goto bad;
        } else if (unlikely(is_instance == -1)) {
            goto bad;
        }
    }
#if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
    Py_XDECREF(self);
    self = NULL;
#endif
    if (binding_func->__signatures__) {
        PyObject *tup;
        if (is_staticmethod && binding_func->func.flags & __Pyx_CYFUNCTION_CCLASS) {
            tup = PyTuple_Pack(3, args,
                               kw == NULL ? Py_None : kw,
                               binding_func->func.defaults_tuple);
            if (unlikely(!tup)) goto bad;
            new_func = (__pyx_FusedFunctionObject *) __Pyx_CyFunction_CallMethod(
                func, binding_func->__signatures__, tup, NULL);
        } else {
            tup = PyTuple_Pack(4, binding_func->__signatures__, args,
                               kw == NULL ? Py_None : kw,
                               binding_func->func.defaults_tuple);
            if (unlikely(!tup)) goto bad;
            new_func = (__pyx_FusedFunctionObject *) __pyx_FusedFunction_callfunction(func, tup, NULL);
        }
        Py_DECREF(tup);
        if (unlikely(!new_func))
            goto bad;
        Py_XINCREF(binding_func->func.func_classobj);
        Py_CLEAR(new_func->func.func_classobj);
        new_func->func.func_classobj = binding_func->func.func_classobj;
        func = (PyObject *) new_func;
    }
    result = __pyx_FusedFunction_callfunction(func, args, kw);
bad:
#if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
    Py_XDECREF(self);
#endif
    Py_XDECREF(new_args);
    Py_XDECREF((PyObject *) new_func);
    return result;
}
static PyMemberDef __pyx_FusedFunction_members[] = {
    {(char *) "__signatures__",
     T_OBJECT,
     offsetof(__pyx_FusedFunctionObject, __signatures__),
     READONLY,
     0},
    {0, 0, 0, 0, 0},
};
static PyMappingMethods __pyx_FusedFunction_mapping_methods = {
    0,
    (binaryfunc) __pyx_FusedFunction_getitem,
    0,
};
static PyTypeObject __pyx_FusedFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "fused_cython_function",
    sizeof(__pyx_FusedFunctionObject),
    0,
    (destructor) __pyx_FusedFunction_dealloc,
    0,
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    0,
    0,
    0,
    &__pyx_FusedFunction_mapping_methods,
    0,
    (ternaryfunc) __pyx_FusedFunction_call,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    0,
    (traverseproc) __pyx_FusedFunction_traverse,
    (inquiry) __pyx_FusedFunction_clear,
    0,
    0,
    0,
    0,
    0,
    __pyx_FusedFunction_members,
    __pyx_CyFunction_getsets,
    &__pyx_CyFunctionType_type,
    0,
    __pyx_FusedFunction_descr_get,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1
    0,
#endif
};
static int __pyx_FusedFunction_init(void) {
    __pyx_FusedFunctionType = __Pyx_FetchCommonType(&__pyx_FusedFunctionType_type);
    if (__pyx_FusedFunctionType == NULL) {
        return -1;
    }
    return 0;
}

/* CLineInTraceback */
  #ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
  static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, (size_t)new_max*sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
  #include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

#if PY_MAJOR_VERSION < 3
static int __Pyx_GetBuffer(PyObject *obj, Py_buffer *view, int flags) {
    if (PyObject_CheckBuffer(obj)) return PyObject_GetBuffer(obj, view, flags);
        if (__Pyx_TypeCheck(obj, __pyx_ptype_5numpy_ndarray)) return __pyx_pw_5numpy_7ndarray_1__getbuffer__(obj, view, flags);
    PyErr_Format(PyExc_TypeError, "'%.200s' does not have the buffer interface", Py_TYPE(obj)->tp_name);
    return -1;
}
static void __Pyx_ReleaseBuffer(Py_buffer *view) {
    PyObject *obj = view->obj;
    if (!obj) return;
    if (PyObject_CheckBuffer(obj)) {
        PyBuffer_Release(view);
        return;
    }
    if ((0)) {}
        else if (__Pyx_TypeCheck(obj, __pyx_ptype_5numpy_ndarray)) __pyx_pw_5numpy_7ndarray_3__releasebuffer__(obj, view);
    view->obj = NULL;
    Py_DECREF(obj);
}
#endif


  /* CIntFromPyVerify */
  #define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* Declarations */
  #if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return ::std::complex< float >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return x + y*(__pyx_t_float_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      __pyx_t_float_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
  #if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabsf(b.real) >= fabsf(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                float r = b.imag / b.real;
                float s = (float)(1.0) / (b.real + b.imag * r);
                return __pyx_t_float_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            float r = b.real / b.imag;
            float s = (float)(1.0) / (b.imag + b.real * r);
            return __pyx_t_float_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            float denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_float_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrtf(z.real*z.real + z.imag*z.imag);
          #else
            return hypotf(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
            __pyx_t_float_complex z;
            float r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    float denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(a, a);
                    case 3:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, a);
                    case 4:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if (b.imag == 0) {
                    z.real = powf(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2f(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_float(a);
                theta = atan2f(a.imag, a.real);
            }
            lnr = logf(r);
            z_r = expf(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cosf(z_theta);
            z.imag = z_r * sinf(z_theta);
            return z;
        }
    #endif
#endif

/* Declarations */
  #if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return ::std::complex< double >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return x + y*(__pyx_t_double_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      __pyx_t_double_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
  #if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabs(b.real) >= fabs(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                double r = b.imag / b.real;
                double s = (double)(1.0) / (b.real + b.imag * r);
                return __pyx_t_double_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            double r = b.real / b.imag;
            double s = (double)(1.0) / (b.imag + b.real * r);
            return __pyx_t_double_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            double denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_double_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrt(z.real*z.real + z.imag*z.imag);
          #else
            return hypot(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
            __pyx_t_double_complex z;
            double r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    double denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(a, a);
                    case 3:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, a);
                    case 4:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if (b.imag == 0) {
                    z.real = pow(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_double(a);
                theta = atan2(a.imag, a.real);
            }
            lnr = log(r);
            z_r = exp(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cos(z_theta);
            z.imag = z_r * sin(z_theta);
            return z;
        }
    #endif
#endif

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__NPY_TYPES(enum NPY_TYPES value) {
    const enum NPY_TYPES neg_one = (enum NPY_TYPES) ((enum NPY_TYPES) 0 - (enum NPY_TYPES) 1), const_zero = (enum NPY_TYPES) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum NPY_TYPES) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(enum NPY_TYPES) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum NPY_TYPES) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum NPY_TYPES) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum NPY_TYPES) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(enum NPY_TYPES),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
  static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
  static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *x) {
    const size_t neg_one = (size_t) ((size_t) 0 - (size_t) 1), const_zero = (size_t) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(size_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(size_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (size_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(size_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 2 * PyLong_SHIFT) {
                            return (size_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 3 * PyLong_SHIFT) {
                            return (size_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 4 * PyLong_SHIFT) {
                            return (size_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (size_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(size_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(size_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(size_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(size_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) ((((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) ((((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) ((((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(size_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            size_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (size_t) -1;
        }
    } else {
        size_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (size_t) -1;
        val = __Pyx_PyInt_As_size_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to size_t");
    return (size_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to size_t");
    return (size_t) -1;
}

/* CIntFromPy */
  static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* ImportNumPyArray */
  static PyObject* __Pyx__ImportNumPyArray(void) {
    PyObject *numpy_module, *ndarray_object = NULL;
    numpy_module = __Pyx_Import(__pyx_n_s_numpy, NULL, 0);
    if (likely(numpy_module)) {
        ndarray_object = PyObject_GetAttrString(numpy_module, "ndarray");
        Py_DECREF(numpy_module);
    }
    if (unlikely(!ndarray_object)) {
        PyErr_Clear();
    }
    if (unlikely(!ndarray_object || !PyObject_TypeCheck(ndarray_object, &PyType_Type))) {
        Py_XDECREF(ndarray_object);
        Py_INCREF(Py_None);
        ndarray_object = Py_None;
    }
    return ndarray_object;
}
static CYTHON_INLINE PyObject* __Pyx_ImportNumPyArrayTypeIfAvailable(void) {
    if (unlikely(!__pyx_numpy_ndarray)) {
        __pyx_numpy_ndarray = __Pyx__ImportNumPyArray();
    }
    Py_INCREF(__pyx_numpy_ndarray);
    return __pyx_numpy_ndarray;
}

/* FastTypeChecks */
  #if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* CheckBinaryVersion */
  static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* FunctionExport */
  static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(__pyx_m, (char *)"__pyx_capi__");
    if (!d) {
        PyErr_Clear();
        d = PyDict_New();
        if (!d)
            goto bad;
        Py_INCREF(d);
        if (PyModule_AddObject(__pyx_m, (char *)"__pyx_capi__", d) < 0)
            goto bad;
    }
    tmp.fp = f;
#if PY_VERSION_HEX >= 0x02070000
    cobj = PyCapsule_New(tmp.p, sig, 0);
#else
    cobj = PyCObject_FromVoidPtrAndDesc(tmp.p, (void *)sig, 0);
#endif
    if (!cobj)
        goto bad;
    if (PyDict_SetItemString(d, name, cobj) < 0)
        goto bad;
    Py_DECREF(cobj);
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(cobj);
    Py_XDECREF(d);
    return -1;
}

/* VoidPtrImport */
  #ifndef __PYX_HAVE_RT_ImportVoidPtr
#define __PYX_HAVE_RT_ImportVoidPtr
static int __Pyx_ImportVoidPtr(PyObject *module, const char *name, void **p, const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, name);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C variable %.200s",
                PyModule_GetName(module), name);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C variable %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), name, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    *p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C variable %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), name, sig, desc);
        goto bad;
    }
    *p = PyCObject_AsVoidPtr(cobj);}
#endif
    if (!(*p))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* FunctionImport */
  #ifndef __PYX_HAVE_RT_ImportFunction
#define __PYX_HAVE_RT_ImportFunction
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, funcname);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C function %.200s",
                PyModule_GetName(module), funcname);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    tmp.p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, desc);
        goto bad;
    }
    tmp.p = PyCObject_AsVoidPtr(cobj);}
#endif
    *f = tmp.fp;
    if (!(*f))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* InitStrings */
  static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
